--|$Log:	keymap_.a,v $
--|Revision 8.0  91/01/17  10:37:30  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Key_Bindings
--
----------------------------------------------------------------------------
--
--
-- Abstract
--
-- The Key_Bindings package provides the data structures and procedures needed
-- to bind the keyboard keys to both history and line editing functions, and to
-- user supplied functions.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------


with Tty_Operations;

with Text_Io;
use Text_Io;

pragma ELABORATE (Tty_Operations);

package body Key_Bindings is

   type Keymap_Array;
   type Keymap_Link is access Keymap_Array;

   type Discriminator is (More_Characters_Needed, No_More_Characters_Needed);

   type Key_Info (Discr : Discriminator := No_More_Characters_Needed) is
      record
         case Discr is
            when More_Characters_Needed  =>
               Next_Keymap : Keymap_Link := null;
            when No_More_Characters_Needed    =>
               Command  : Commands   := Unbound;
               Optional : String_Ptr := null;
         end case;
      end record;

   type Keymap_Array is array (Character'FIRST .. Character'LAST) of Key_Info;
   Top_Keymap           : Keymap_Link;

   Line_Separator_Count : Integer    := 0;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Error
   --
   --    Last Mod:    
   --    Author:      R. Pollack
   --
   --  <Description>
   --
   --    This procedure centralizes outputting error messages.
   --
   --  --------------------------------------------------------------------

   procedure Error (S : in String) is
   begin
      Put ("Key Binding Error: ");
      Put_Line (S);
   end Error;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Top_Binding
   --
   --    Last Mod:    
   --    Author:      R. Pollack
   --
   --  <Description>
   --
   --    This procedure is used to find a key binding when there is a requirement
   --    that the escape sequence be only one character long.
   --  --------------------------------------------------------------------

   function Top_Binding (C : in Character) return Commands is
   begin
      if Top_Keymap (C).Discr = More_Characters_Needed then
         return Error;
      else
         return Top_Keymap (C).Command;
      end if;
   end Top_Binding;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Make_Binding
   --
   --    Last Mod:    
   --    Author:    R. Pollack  
   --    Art:       C. Smotherman
   --
   --  <Description>
   --    This procedure comes in two flavors; the first simply allows the user to
   --    make the call with a character as the first argument; the second procedure
   --    does the work.
   --
   --    This procedure traces a character sequence through a series of arrays;
   --    for the earlier characters in the sequence, if needed, it creates a new array
   --    and puts a pointer to the new array in the previous array, but if the
   --    needed array is already there, just proceeds to the next character in the
   --    sequence. For the last character it creates an array if necessary and
   --    writes the command name in the proper entry. There is a requirement that
   --    no character sequence used for the first argument be a prefix of another
   --    character sequence, used in a different Make_Binding call. If the user
   --    in the example below had chosen to make the last call:
   --        Make_Binding( "ac", move);
   --    they would have caused the binding for "aca" to be lost.

   --    The picture below is a shrunk down example of the arrays that would be built
   --    by three calls to Make_Binding: Make_Binding( "aab", insert);
   --                                    Make_Binding( "aca", delete);
   --                                    Make_Binding( "ca", move);
   --
   --   -----     _____            _____
   --   | a |---> | a |----------> | a |
   --   |   |     |   |            | b |- Command = insert
   --   |   |     | b |            | c |
   --   | b |     |   |       ____ -----
   --   |   |     | c |-----> | a |- Command = delete
   --   |   |     -----       | b |
   --   |   |     _____       | c |
   --   | c |---> | a |       -----
   --   -----     | b |- Command = move
   --             | c |
   --             -----

   --  --------------------------------------------------------------------

   procedure Make_Binding (C               : in Character;
                           Binding         : in Commands;
                           Optional_String : in String := "") is
   begin
      Make_Binding (C & "", Binding, Optional_String);
   end Make_Binding;

   procedure Make_Binding (Char_Seq        : in String;
                           Binding         : in Commands;
                           Optional_String : in String := "") is

      K : Keymap_Link;
      Next_K                     : Keymap_Link;
      Unbinding_A_Line_Separator : Boolean;

   begin

      -- Check for constraints on all bindings

      if Char_Seq'LENGTH < 1 then
         Error ("Attempt to bind null character sequence");
         return ;
      end if;

      Unbinding_A_Line_Separator := Char_Seq'LENGTH = 1 and then
                                    Top_Binding (Char_Seq (Char_Seq'FIRST)) =
                                    Line_Separator;

      -- Check for constraints on special bindings

      case Binding is
         when Line_Separator   =>
            if Char_Seq'LENGTH > 1 then
               Error ("Line separation may only be bound to single character");
               return ;
            end if;
         when others =>
            if Unbinding_A_Line_Separator and then Line_Separator_Count < 2
                then
               Error ("Attempt to unbind only line separator");
               return ;
            end if;
      end case;

      K := Top_Keymap;
      for I in Char_Seq'FIRST .. Char_Seq'LAST - 1 loop

         if K (Char_Seq (I)).Discr = No_More_Characters_Needed then
            Next_K := new Keymap_Array;
            K (Char_Seq (I)) := (Discr       => More_Characters_Needed,
                                 Next_Keymap => Next_K);
         end if;
         K := K (Char_Seq (I)).Next_Keymap;
      end loop;

      K (Char_Seq (Char_Seq'LAST)) :=
         (Discr   => No_More_Characters_Needed,
          Command => Binding,
          Optional => new String'(Optional_String));

      if Binding = Line_Separator then
         Line_Separator_Count := Line_Separator_Count + 1;
      end if;
      if Unbinding_A_Line_Separator then
         Line_Separator_Count := Line_Separator_Count - 1;
      end if;
   end Make_Binding;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Find_Binding
   --
   --    Last Mod:    
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --    This procedure traces a character sequence through the arrays until
   --    it comes to the entry which identifies the command.
   --  --------------------------------------------------------------------

   procedure Find_Binding (First_Char      : in Character;
                           Binding         : out Commands;
                           Optional_String : out String_Ptr) is

      Char : Character;
      K    : Keymap_Link;

   begin
      K    := Top_Keymap;
      Char := First_Char;
      loop
         if K (Char).Discr = No_More_Characters_Needed then
            Binding := K (Char).Command;
            Optional_String := K (Char).Optional;
            exit ;
         else
            K    := K (Char).Next_Keymap;
            Char := Tty_Operations.Getch;
         end if;
      end loop;
   end Find_Binding;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Make_Default_Bindings
   --
   --    Last Mod:    
   --    Author:      R. Pollack
   --
   --  <Description>
   --
   --    This procedure establishes a basic set of bindings to enable the user to
   --    take advantage of the history and line editing features of Sanctified_Input.
   --  --------------------------------------------------------------------

   procedure Make_Default_Bindings is
   begin
      for C in ' ' .. '~' loop
         Make_Binding (C, Self_Insert);
      end loop;

      Make_Binding (Ascii.Cr, Line_Separator);
      Make_Binding (Ascii.Lf, Line_Separator);

      Make_Binding (Tty_Operations.Terminal_Erase_Character, Delete);

   end Make_Default_Bindings;

begin
   Top_Keymap := new Keymap_Array;
end Key_Bindings;
