--|$Log:	sinput_.a,v $
--|Revision 8.0  91/01/17  10:37:34  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Subunits for Package - Sanctified_Input
--
----------------------------------------------------------------------------
--
--
-- Abstract
--
-- Sanctified_Input provides procedures to accept user input; keep a history
-- of the input; and manipulate the lines in the history, redisplaying them
-- and resubmitting them to the calling program, without the user having to
-- retype them.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Key_Bindings;
use Key_Bindings;

with Text_Io;
with Unchecked_Deallocation;
with Tty_Operations;


pragma ELABORATE (Key_Bindings);

package body Sanctified_Input is

   Max_Number_Of_Lines : Integer := 10;
   -- Making the Lines buffer hold one extra line makes it easier to keep track
   -- of when we have circled all the way around the list
   Number_Of_Lines : Integer := Max_Number_Of_Lines + 1;
   subtype Line_Length_Range is
      Natural range 0 .. Tty_Operations.Max_Line_Length;
   type Line_Ptr is access String;
   Current_Character : Character;
   Binding           : Commands;
   Optional          : String_Ptr;

   type Line_Entry is
      record
         Line_Number : Natural;
         Line        : Line_Ptr;
      end record;
   type Storage is array (Positive range <>) of Line_Entry;
   type Storage_Ptr is access Storage;
   History     : Storage_Ptr;
   Line_Chosen : Integer;
   -- Line_Chosen is used to keep track of scrolling thru the history
   Next_Line : Natural;
   -- Next_Line is the number of the next line to be added to History
   The_Next_Line, Place_In_History : Integer;
   Pos_In_Line                     : Line_Length_Range := 0;
   -- Pos_in_Line describes the number of spaces between the beginning of the
   --  line and the space the cursor is on
   Prior_Line_Buf, Next_Line_Buf : String (1 .. Tty_Operations.Max_Line_Length);
   Prior_Line_Len, Next_Line_Len : Line_Length_Range;

   procedure Free is new Unchecked_Deallocation (String, Line_Ptr);
   procedure Free_History is new Unchecked_Deallocation (Storage, Storage_Ptr);


   function Maximum_Line_Length return Natural is
   begin
      return Tty_Operations.Max_Line_Length;
   end;

   -----------------------------------------------------------------------
   --  <Subprogram Name>
   --    Rel_Line
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --    This function serves as a counter that makes the numbers from
   --    1 to Number_Of_Lines one circular loop
   --
   --  --------------------------------------------------------------------

   function Rel_Line (Line_Number : in Natural) return Positive is
      Relative_Line_Number : Positive;
      Place                : Natural;
   begin
      if (Line_Number = 0) then
         Relative_Line_Number := Number_Of_Lines;
      else
         Place := Line_Number mod Number_Of_Lines;
         if (Place = 0) then
            Relative_Line_Number := Number_Of_Lines;
         else
            Relative_Line_Number := Place;
         end if;
      end if;
      return Relative_Line_Number;

   exception
      when others =>
         Tty_Operations.Put ("Error in rel_line ");
         raise;
   end Rel_Line;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Do_History_Back
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --    This procedure moves the pointer for the history mechanism back by one
   --    line, and displays the line it is then pointing to
   --
   --==========================================================================

   procedure Do_History_Back is
   begin
      if (History (Place_In_History).Line_Number > 1 and
         not (The_Next_Line = Rel_Line (Place_In_History - 1)) and
      -- that is, we haven't come all the way around the list
         History (Rel_Line (Place_In_History - 1)).Line /= null) then
    -- the 3rd test is necessary in case the user has increased the history size
         Line_Chosen := Rel_Line (Place_In_History - 1);
         Place_In_History := Line_Chosen;
         Prior_Line_Len := History (Place_In_History).Line'LENGTH;
         Prior_Line_Buf (Prior_Line_Buf'RANGE) := (others => ' ');
         Prior_Line_Buf (1 .. Prior_Line_Len)  :=
            History (Place_In_History).Line.all;
         Tty_Operations.Clear_Line;
         Tty_Operations.Put (Prior_Line_Buf (1 .. Prior_Line_Len));
         Pos_In_Line := Prior_Line_Len;
      end if;
   exception
      when others =>
         Tty_Operations.Put ("Error in History_Back ");
   end Do_History_Back;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Do_History_Forward
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --    This procedure moves the pointer for the history mechanism forward
   --    by one line, and displays the line it is then pointing to
   --
   --==========================================================================

   procedure Do_History_Forward is
   begin
      if (History (Place_In_History).Line_Number < Next_Line) then
         Tty_Operations.Clear_Line;
         Line_Chosen      := Rel_Line (Place_In_History + 1);
         Place_In_History := Rel_Line (Place_In_History + 1);
         if (Place_In_History = The_Next_Line) then
            Pos_In_Line := Next_Line_Len;
            if (Next_Line_Len > 0) then
               Tty_Operations.Put (Next_Line_Buf (1 .. Next_Line_Len));
            end if;
         else
            Prior_Line_Buf (Prior_Line_Buf'RANGE) := (others => ' ');
            Prior_Line_Len := History (Place_In_History).Line'LENGTH;
            Pos_In_Line := Prior_Line_Len;
            Prior_Line_Buf (1 .. Prior_Line_Len)  :=
               History (Place_In_History).Line (1 .. Prior_Line_Len);
            Tty_Operations.Put (Prior_Line_Buf (1 .. Prior_Line_Len));
         end if;
      end if;
   exception
      when others =>
         Tty_Operations.Put ("Error in History_Forward ");
   end Do_History_Forward;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Current_Line
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --   These routines provide the current line of input (the one the user
   --   has not yet completed by typing a line separator)
   --==========================================================================

   function Get_Current_Line return String is
   begin

      if (Line_Chosen = The_Next_Line) then
         return Next_Line_Buf (1 .. Next_Line_Len);
      else
         return Prior_Line_Buf (1 .. Prior_Line_Len);
      end if;
   end Get_Current_Line;


   procedure Get_Current_Line (Line   : out String;
                               Length : out Natural) is
   begin

      if (Line_Chosen = The_Next_Line) then
         Line (Line'FIRST .. Line'FIRST + Next_Line_Len - 1) :=
            Next_Line_Buf (1 .. Next_Line_Len);
         Length := Next_Line_Len;
      else
         Line (Line'FIRST .. Line'FIRST + Prior_Line_Len - 1) :=
            Prior_Line_Buf (1 .. Prior_Line_Len);
         Length := Prior_Line_Len;
      end if;
   end Get_Current_Line;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Refresh_Current_Line, Refresh_Current_Line_and_Prompt
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --   Refresh_Current_Line clears, then rewrites the current line of input,
   --   starting with the first space after the prompt;
   --   Refresh_Current_Line_and_Prompt writes the prompt and current line of
   --   input (possibly blank) onto a new line (later it will rewrite onto the
   --   same physical line)
   --==========================================================================

   procedure Refresh_Current_Line is
      Line   : String (1 .. Tty_Operations.Max_Line_Length);
      Length : Natural;
   begin
      Get_Current_Line (Line, Length);
      Terminal_Set;
      Kill_Line;
      Insert_String (Line (Line'FIRST .. Line'FIRST + Length - 1));
   end Refresh_Current_Line;


   -- Generally the user should be at the beginning of a line
   -- when calling this procedure
   procedure Refresh_Current_Line_And_Prompt is
      Line   : String (1 .. Tty_Operations.Max_Line_Length);
      Length : Natural;
   begin
      Get_Current_Line (Line, Length);
      Terminal_Set;
      Kill_Line;
      Terminal_Reset;
      Text_Io.Put (Prompt);
      Terminal_Set;
      Insert_String (Line (Line'FIRST .. Line'FIRST + Length - 1));
   end Refresh_Current_Line_And_Prompt;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Rewrite_Current_Line, Rewrite_Current_Line_and_Prompt
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
  --   These procedures are identical to the Refresh versions above, except that
  --   they accept an argument to be used as the new version of the current line
   --==========================================================================

   procedure Rewrite_Current_Line (Str : in String) is
      Line : String (1 .. Tty_Operations.Max_Line_Length);
   begin
      Terminal_Set;
      Kill_Line;
      Insert_String (Str (Str'FIRST .. Str'LAST));
   end Rewrite_Current_Line;


   -- Generally the user should be at the beginning of a line
   -- when calling this procedure
   procedure Rewrite_Current_Line_And_Prompt (Str : in String) is
   begin
      Terminal_Set;
      Kill_Line;
      Terminal_Reset;
      Text_Io.Put (Prompt);
      Terminal_Set;
      Insert_String (Str (Str'FIRST .. Str'LAST));
   end Rewrite_Current_Line_And_Prompt;


   ---==========================================================================
   --  <Subprogram Name>
   --
   --    Goto_Beg_of_Line, Goto_End_of_Line
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --    These procedures move the cursor to the beginning, or end, of the
   --    current line
   --
   --==========================================================================

   procedure Goto_Beg_Of_Line is
   begin
      Tty_Operations.Move_Left (Pos_In_Line);
      Pos_In_Line := 0;
   end Goto_Beg_Of_Line;


   procedure Goto_End_Of_Line is
   begin
      if (Line_Chosen = The_Next_Line) then
         Tty_Operations.Move_Right (Next_Line_Len - Pos_In_Line);
         Pos_In_Line := Next_Line_Len;
      else
         Tty_Operations.Move_Right (Prior_Line_Len - Pos_In_Line);
         Pos_In_Line := Prior_Line_Len;
      end if;
   end Goto_End_Of_Line;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Display_History_Line
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
  --   This procedure prints one line from the History array and its line number
   --
   --==========================================================================

   package Int_Io is new Text_Io.Integer_Io (Integer);

   procedure Display_History_Line (Index : Integer) is
      Line : String (1 .. Tty_Operations.Max_Line_Length + 8);
   begin
      Line (Line'RANGE) := (others => ' ');
      Int_Io.Put (Line (1 .. 6), History (Index).Line_Number);
      Line (7 .. 8) := "  ";
      for I in 1 .. History (Index).Line'LENGTH loop
         Line (I + 8) := History (Index).Line (I);
      end loop;
      Tty_Operations.Put (Line (1 .. History (Index).Line'LENGTH + 8));
      Tty_Operations.New_Line;

   exception
      when others =>
         Tty_Operations.Put ("Error in Display_History_Line");
   end Display_History_Line;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Show_History
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
  --   This procedure arranges for a list of lines to be displayed on the screen
   --
   --==========================================================================


   procedure Show_History is
      J : Integer;
   begin
      -- this is a first cut, will display all possible lines

      Tty_Operations.New_Line;

      J := 1;
      for I in Rel_Line (The_Next_Line + 1) .. Number_Of_Lines loop
         if (History (I).Line /= null) then
            Display_History_Line (I);
         end if;
         J := J + 1;
         if (J > Max_Number_Of_Lines) then
            goto Bottom;
         end if;
      end loop;

      if (Rel_Line (The_Next_Line - 1) < Rel_Line (The_Next_Line + 1)) then
         for I in 1 .. Rel_Line (The_Next_Line - 1) loop
            if (History (I).Line /= null) then
               Display_History_Line (I);
            end if;
            J := J + 1;
         end loop;
         if (J > Max_Number_Of_Lines) then
            goto Bottom;
         end if;
      end if;

   <<Bottom>>
      null;
      Refresh_Current_Line_And_Prompt;

   exception
      when others =>
         Tty_Operations.Put ("Error in Show_History");
   end Show_History;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Show_History_Limit
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --   This procedure shows the user the constraints on saving input text.
   --==========================================================================

   procedure Show_History_Limit is
      Line : String (1 .. 48);
   begin

      Tty_Operations.New_Line;
      Line (1 .. 17) := "History size is: ";
      Int_Io.Put (Line (18 .. 21), Max_Number_Of_Lines);
      Line (22 .. 38) := "  (max length of ";
      Int_Io.Put (Line (39 .. 42), Tty_Operations.Max_Line_Length);
      Line (43 .. 48) := " each)";
      Tty_Operations.Put (Line (1 .. 48));
      Tty_Operations.New_Line;
      Refresh_Current_Line_And_Prompt;

   exception
      when others =>
         Tty_Operations.Put ("Error in Show_History_Limit");
   end Show_History_Limit;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    ask_for_integer
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --   This function asks the user for an integer and then restores the
   --   current line if successful; if not successful in getting an
   --   integer it prints a message and returns -1
   --==========================================================================

   function Ask_For_Integer (Request : in String) return Integer is
      Line        : String (1 .. Tty_Operations.Max_Line_Length);
      Length      : Natural;
      The_Integer : Integer := - 1;
   begin
      Get_Current_Line (Line, Length);
      Kill_Line;
      Terminal_Reset;
      Text_Io.Put (Request);
      begin
         Int_Io.Get (The_Integer);
      exception
         when others =>
            Text_Io.New_Line;
            Text_Io.Put
                ("Exception encountered while trying to read an integer");
      end;
      Terminal_Set;
-- the carriage return the user ended the integer with prepares us for the inser
      --+t
      Insert_String (Line (Line'FIRST .. Line'FIRST + Length - 1));
      return The_Integer;
   end Ask_For_Integer;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_History_Limit
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --   This procedure processes a user request to change the number of lines of
   --   history, by temporarily storing the appropriate amount of the history
  -- information elsewhere, then destroying the old history array and creating a
   -- new one of the requested size, & moving the information into the new array
   --==========================================================================

   function Min (X, Y : Integer) return Integer is
   begin
      if (Y > X) then
         return X;
      else
         return Y;
      end if;
   end Min;

   procedure Set_History_Limit is
      Temp : Storage_Ptr;
      I, Index, New_Limit, Limit : Integer;
      Position, Former_Max       : Integer;
   begin
      New_Limit :=
         Ask_For_Integer ("How many lines of history do you want to save? ");

      if (New_Limit = Max_Number_Of_Lines) then
         Refresh_Current_Line_And_Prompt;
      elsif (New_Limit < 2) then
         Tty_Operations.Beep;
      else
         Temp := new Storage (1 .. New_Limit + 1);

         declare
            subtype Temp_Type is Storage (1 .. New_Limit + 1);
         begin
            Temp (1 .. New_Limit + 1) := Temp_Type'(others => (0, null));

            -- was :
            --
            --   Temp (1 .. New_Limit + 1) := (others => (0, null));
            --
            -- change required for TeleSoft compiler
         end;

         Limit := Min (Number_Of_Lines, New_Limit + 1);
         Limit := Min (Limit, Next_Line);

         Index                := The_Next_Line;
         I                    := 1;
         Temp (I).Line_Number := History (The_Next_Line).Line_Number;
         while (I < Limit) loop
            I                    := I + 1;
            Index                := Rel_Line (Index - 1);
            Temp (I).Line_Number := History (Index).Line_Number;
            Temp (I).Line        := History (Index).Line;
         end loop;

         Free_History (History);
         History := new Storage (1 .. New_Limit + 1);

         declare
            subtype Temp_Type is Storage (1 .. New_Limit + 1);
         begin
            History (1 .. New_Limit + 1) := Temp_Type'(others => (0, null));

            -- was :
            --
            --   History (1 .. New_Limit + 1) := (others => (0, null));
            --
            -- change required for TeleSoft compiler
         end;

         Former_Max          := Max_Number_Of_Lines;
         Max_Number_Of_Lines := New_Limit;
         Number_Of_Lines     := New_Limit + 1;

         I := 1;
         while (I <= New_Limit + 1) loop
            if (Temp (I).Line /= null) then
               Position := Rel_Line (Temp (I).Line_Number);
               History (Position).Line_Number := Temp (I).Line_Number;
               History (Position).Line        := Temp (I).Line;
            elsif (Temp (I).Line_Number /= 0) then
               Position := Rel_Line (Temp (I).Line_Number);
               History (Position).Line_Number := Temp (I).Line_Number;
            end if;

            I := I + 1;
         end loop;
         Free_History (Temp);

         if (Line_Chosen > Number_Of_Lines or Line_Chosen = The_Next_Line) then
            Line_Chosen := Rel_Line (Next_Line);
         end if;
         if (Place_In_History > Number_Of_Lines or
            Place_In_History = The_Next_Line) then
            Place_In_History := Rel_Line (Next_Line);
         end if;
         The_Next_Line := Rel_Line (Next_Line);

         Tty_Operations.Put ("History size reset to " &
                             Integer'IMAGE (New_Limit));
         Tty_Operations.New_Line;
         Refresh_Current_Line_And_Prompt;
      end if;

   exception
      when others =>
         Tty_Operations.Put ("Error in Set_History_Limit");
   end Set_History_Limit;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Goto_Start_of_History, Goto_End_of_History
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --   These procedures move the pointer for the history mechanism back to the
   --   earliest available line of history, or down to the current line of input
   --   being gathered, and display that line on the screen.
   --
   --==========================================================================

   procedure Goto_Start_Of_History is
      Pos : Integer;
   begin
      if (Next_Line = 1) then
         null;
      else
         Pos := The_Next_Line;
         while (History (Rel_Line (Pos + 1)).Line = null) loop
            Pos := Rel_Line (Pos + 1);
         end loop;
         Line_Chosen := Rel_Line (Pos + 2);
         -- we set Line_Chosen to the earliest line of
         --history, + 1, so we can use do_history_back
         Place_In_History := Line_Chosen;
         Do_History_Back;
      end if;

   exception
      when others =>
         Tty_Operations.Put ("Error in Goto_Start_of_History");
   end Goto_Start_Of_History;


   procedure Goto_End_Of_History is
   begin
      Line_Chosen      := Rel_Line (The_Next_Line - 1);
      Place_In_History := Line_Chosen;
      Do_History_Forward;

   exception
      when others =>
         Tty_Operations.Put ("Error in Goto_End_of_History");
   end Goto_End_Of_History;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Move_Left, Move_Right
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --    These routines move the cursor one space to the left, or right, on the
   --    current line
   --
   --==========================================================================

   procedure Move_Left is
   begin
      if (Pos_In_Line > 0) then
         Pos_In_Line := Pos_In_Line - 1;
         Tty_Operations.Move_Left;
      end if;
   exception
      when others =>
         Tty_Operations.Put ("Error in Move_Left");
   end Move_Left;

   procedure Move_Right is
   begin
      if (Line_Chosen = The_Next_Line) then
         if (Pos_In_Line < Next_Line_Len) then
            Pos_In_Line := Pos_In_Line + 1;
            Tty_Operations.Move_Right;
         end if;
      else
         if (Pos_In_Line < Prior_Line_Len) then
            Pos_In_Line := Pos_In_Line + 1;
            Tty_Operations.Move_Right;
         end if;
      end if;
   exception
      when others =>
         Tty_Operations.Put ("Error in Move_Right");
   end Move_Right;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Delete
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --    This procedure deletes the character to the left of the cursor and
   --    updates the screen display
   --
   --==========================================================================

   procedure Delete is
      Line_Len : Line_Length_Range;
   begin
      if (Line_Chosen = The_Next_Line) then
         Line_Len := Next_Line_Len;
      else
         Line_Len := Prior_Line_Len;
      end if;

      if (Line_Len = 0) then
         null;
      elsif (Pos_In_Line = 0) then
         null;
      else
         Pos_In_Line := Pos_In_Line - 1;
         if (Line_Chosen = The_Next_Line) then
            Next_Line_Len := Next_Line_Len - 1;
            if (Pos_In_Line = Next_Line_Len) then
               null;
            else
               Next_Line_Buf (Pos_In_Line + 1 .. Next_Line_Len) :=
                  Next_Line_Buf (Pos_In_Line + 2 .. Next_Line_Len + 1);
            end if;
         else -- we're on a historic line, apply this ascii to Prior_Line_Buf
            Prior_Line_Len := Prior_Line_Len - 1;
            if (Pos_In_Line = Prior_Line_Len) then
               null;
            else
               Prior_Line_Buf (Pos_In_Line + 1 .. Prior_Line_Len) :=
                  Prior_Line_Buf (Pos_In_Line + 2 .. Prior_Line_Len + 1);
            end if;
         end if;
         Tty_Operations.Delete (1);
      end if;

   exception
      when others =>
         Tty_Operations.Put ("Error in Delete");
   end Delete;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Delete_This_Char
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --    This routine moves the cursor one position to the right and then
   --    calls Delete - the effect being to delete the character the cursor is
   --    sitting on
   --==========================================================================

   procedure Delete_This_Char is -- delete the character we are sitting on
   begin
      if (Line_Chosen = The_Next_Line) then
         if (Pos_In_Line = Next_Line_Len) then
            null;
         else
            Pos_In_Line := Pos_In_Line + 1;
            Tty_Operations.Move_Right (1);
            Delete;
         end if;
      else
         if (Pos_In_Line = Prior_Line_Len) then
            null;
         else
            Pos_In_Line := Pos_In_Line + 1;
            Tty_Operations.Move_Right (1);
            Delete;
         end if;
      end if;

   exception
      when others =>
         Tty_Operations.Put ("Error in Delete_This_Char");
   end Delete_This_Char;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Insert_Char, Insert_String
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --    Insert_Char inserts a character into the current line, to the left
   --    of the position the cursor occupies; Insert_String pushes a string
   --    through insert_char
   --==========================================================================

   procedure Insert_Char is
   begin
      if (Line_Chosen = The_Next_Line and then
         Next_Line_Len = Tty_Operations.Max_Line_Length) then
         null; -- can't do self-insert on already full line
      else
         Pos_In_Line := Pos_In_Line + 1;
         if (Line_Chosen = The_Next_Line) then
            -- apply this ascii to Next_Line_Buf
            Next_Line_Len := Next_Line_Len + 1;
            Next_Line_Buf (Pos_In_Line + 1 .. Next_Line_Len) :=
               Next_Line_Buf (Pos_In_Line .. Next_Line_Len - 1);
            Next_Line_Buf (Pos_In_Line) := Current_Character;
         else -- we're on a historic line, apply this ascii to Prior_Line_Buf
            Prior_Line_Len := Prior_Line_Len + 1;
            Prior_Line_Buf (Pos_In_Line + 1 .. Prior_Line_Len) :=
               Prior_Line_Buf (Pos_In_Line .. Prior_Line_Len - 1);
            Prior_Line_Buf (Pos_In_Line) := Current_Character;
         end if;
         Tty_Operations.Insert_Char (Current_Character);
      end if;

   exception
      when others =>
         Tty_Operations.Put ("Error in Insert_Char");
   end Insert_Char;


   procedure Insert_String (The_String : in String) is
   begin
      for I in The_String'FIRST .. The_String'LAST loop
         Current_Character := The_String (I);
         Insert_Char;
      end loop;
   end Insert_String;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Kill_Line
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --   Kill_Line really wipes the line out, both on the screen and in memory
   --==========================================================================

   procedure Kill_Line is
   begin
      Pos_In_Line := 0;
      if (Line_Chosen = The_Next_Line) then
         Next_Line_Len := 0;
      else
         Prior_Line_Len := 0;
      end if;
      Tty_Operations.Clear_Line;
   end Kill_Line;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Current_Column, Get_Current_Character
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --  These functions return the 1-based number of the column the cursor is
   --  in, and the character in that column (ascii.nul if the cursor is at
   --  the end of the line)
   --==========================================================================

   function Get_Current_Column return Natural is
   begin
      return Pos_In_Line + 1;
   end Get_Current_Column;


   function Get_Current_Character return Character is
   begin
      if (Line_Chosen = The_Next_Line) then
         if (Next_Line_Len = 0) then
            return Ascii.Nul;
         elsif (Pos_In_Line = Next_Line_Len) then
            return Ascii.Nul;
         else
            return Next_Line_Buf (Pos_In_Line + 1);
         end if;
      else
         if (Prior_Line_Len = 0) then
            return Ascii.Nul;
         elsif (Pos_In_Line = Prior_Line_Len) then
            return Ascii.Nul;
         else
            return Prior_Line_Buf (Pos_In_Line + 1);
         end if;
      end if;
   end Get_Current_Character;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Quoted_Insert
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --   This procedure allows the user to insert a character that is bound to
   --   something other than self-insert
   --==========================================================================

   procedure Quoted_Insert is
   begin
      Current_Character := Tty_Operations.Getch;
      Insert_Char;
   end Quoted_Insert;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Apply_Binding
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --    This procedure is pretty straightforward!
   --    (ok, given a command, it invokes the routines which will accomplish
   --    the command)
   --
   --==========================================================================

   procedure Apply_Binding is
   begin

      case Binding is
         when Self_Insert =>
            Insert_Char;

         when Move_Left =>
            Move_Left;

         when Move_Right =>
            Move_Right;

         when History_Back =>
            Do_History_Back;

         when History_Forward =>
            Do_History_Forward;

         when Beginning_Of_Line =>
            Goto_Beg_Of_Line;

         when End_Of_Line =>
            Goto_End_Of_Line;

         when Start_Of_History =>
            Goto_Start_Of_History;

         when End_Of_History =>
            Goto_End_Of_History;

         when Show_History =>
            Show_History;

         when Show_History_Limit =>
            Show_History_Limit;

         when Set_History_Limit =>
            Set_History_Limit;

         when Interpret_String =>
            Interpret_String (Optional.all);

         when Delete =>
            Delete;

         when Delete_This_Char =>
            Delete_This_Char;

         when Kill_Line =>
            Kill_Line;

         when Quoted_Insert =>
            Quoted_Insert;

         when Refresh_Current_Line =>
            Refresh_Current_Line;

         when Refresh_Current_Line_And_Prompt =>
            Tty_Operations.New_Line;
            Refresh_Current_Line_And_Prompt;

         when others =>
            Tty_Operations.Beep;
      end case;

   exception
      when others =>
         Tty_Operations.Put ("Error in Apply_Binding");
   end Apply_Binding;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Store_Line
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --    This procedure adds a completed line of user input to the history
   --    mechanism
   --
   --==========================================================================

   procedure Store_Line (Line   : out String;
                         Length : out Natural) is
   begin
      if (History (The_Next_Line).Line /= null) then
         Free (History (The_Next_Line).Line);
      end if;

      if (Line_Chosen = The_Next_Line) then
         Line (1 .. Next_Line_Len) := Next_Line_Buf (1 .. Next_Line_Len);
         Length := Next_Line_Len;
         History (The_Next_Line).Line       := new String (1 .. Next_Line_Len);
         History (The_Next_Line).Line.all := Next_Line_Buf (1 .. Next_Line_Len);
         Next_Line_Buf (1 .. Next_Line_Len) := (others => ' ');
         Next_Line_Len := 0;
      else
         Line (1 .. Prior_Line_Len) := Prior_Line_Buf (1 .. Prior_Line_Len);
         Length := Prior_Line_Len;
         History (The_Next_Line).Line     := new String (1 .. Prior_Line_Len);
         History (The_Next_Line).Line.all := Prior_Line_Buf (1 ..
                                                             Prior_Line_Len);
         Prior_Line_Buf (1 .. Prior_Line_Len) := (others => ' ');
         Prior_Line_Len := 0;
      end if;

   exception
      when others =>
         Tty_Operations.Put ("Error in Store_Line ");
   end Store_Line;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Line_of_Input
   --
   --    Last Mod:
   --    Author:      C. Smotherman
   --
   --  <Description>
   --
   --    This procedure is the cpu of Sanctified_Input; it provides the loop
   --    which collects a line of input, and then does the processing necessary
   --    to prepare to receive the next line of input.
   --
   --     When the user starts typing a new line, then does history_back, the
   --     input to the new line is saved; however, if a user edits a historic
   --     line and then does either history_back or history_forward, the
   --     editing of the historic line is lost.

   --     Ordinarily the pointer into the history mechanism is incremented each
   --     time a line is input, so that history_back would cause the line just
   --     entered to appear on the screen. However, once the user uses one of
   --     the commands which resets that pointer, it will continue to point to
   --     whatever line the user positioned it on, until the user enters a
   --     completely fresh line of input, at which time the pointer is reset to
   --     point to the most recently entered line.
   --==========================================================================

   procedure Get_Line_Of_Input (Line   : out String;
                                Length : out Natural) is
   begin
      Line_Chosen    := The_Next_Line;
      Prior_Line_Len := 0;
      Next_Line_Len  := 0;
      Pos_In_Line    := 0;

      loop
         Current_Character := Tty_Operations.Getch;
         Find_Binding (Current_Character, Binding, Optional);
         exit when Binding = Line_Separator;
         Apply_Binding;
      end loop;

      Tty_Operations.New_Line;

      -- ignore empty statements;
      -- reset place_in_history if they have entered a brand line
      if (Line_Chosen = The_Next_Line) then
         if (Next_Line_Len = 0) then
            Length := 0;
            goto Bottom;
         else
            Place_In_History := Rel_Line (The_Next_Line + 1);
         end if;
      else
         if (Prior_Line_Len = 0) then
            Line_Chosen := The_Next_Line;
            Length      := 0;
            goto Bottom;
         end if;
      end if;

      Store_Line (Line, Length);
      Next_Line := Next_Line + 1;
      The_Next_Line := Rel_Line (Next_Line);
      History (The_Next_Line).Line_Number := Next_Line;

   <<Bottom>>
      null;

   exception
      when others =>
         Tty_Operations.Put ("Error in Get_Line_Of_Input ");
   end Get_Line_Of_Input;

   --==========================================================================
   --  <Subprogram Name>
   --
   --   Get_Line
   --
   --    Last Mod:
   --    Author:     C. Smotherman
   --
   --  <Description>
   --
   --   This procedure is a wrapper for Get_Line_of_Input. It allows the
   --   calling program to see Sanctified_Input's input routine in the same way
   --   it sees Text_Io's input routine.
   --==========================================================================

   procedure Get_Line (File : in Text_Io.File_Type;
                       Item : out String;
                       Last : out Natural) is
   begin
      Get_Line_Of_Input (Item, Last);
   end Get_Line;

   procedure Initialize is
   begin

      Make_Default_Bindings;
      History := new Storage (1 .. Number_Of_Lines);

      declare
         subtype History_Type is Storage (1 .. Number_Of_Lines);
      begin
         History (1 .. Number_Of_Lines) := History_Type'(others => (0, null));

         -- was :
         --
         --   History (1 .. Number_Of_Lines) := (others => (0, null));
         --
         -- change required for TeleSoft compiler
      end;

      History (1).Line_Number := 1;
      Next_Line               := 1;
      The_Next_Line           := 1;
      Place_In_History        := 1;
   end Initialize;

begin
   Initialize;
end Sanctified_Input;
