--|$Log:	tty_ops_.a,v $
--|Revision 8.0  91/01/17  10:40:45  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Tty_Operations
--
----------------------------------------------------------------------------
--
--
-- Abstract
--
-- This package offers a number of procedures for positioning the
-- terminal (tty) cursor and manipulating the terminal screen directly.
-- It also permits characters to be read in "raw" mode, that is,
-- as soon as keys are struck; this makes it possible to respond to
-- such things as arrow keys and function keys.
--
-- Those who elect to use this package should use only the operations
-- of this package should use only the operations provided by this
-- package while the terminal is in raw mode.  That is, while the
-- terminal is in raw mode, the caller should not read or write
-- to or from the terminal with Text_IO.  The function Getch causes
-- the terminal to enter this mode; the procedures New_Line and
-- Terminal_Reset cause it to return to normal mode.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------
with System;
use System;
with Unix_Types;
use Unix_Types;
with Unchecked_Conversion;
with Text_Io;

pragma ELABORATE (Unix_Types);

package body Tty_Operations is

   type String_Ptr is access String;

   Null_Cstring : constant Cstring := 0;

   -- Boolean terminal capability flags

   Have_Termcap            : Boolean := False;
   Have_Backspace_Cap      : Boolean := False;
   Have_Insert_Cap         : Boolean := False;
   Need_Insert_End         : Boolean := False;
   Have_Delete_Cap         : Boolean := False;
   Need_Delete_Mode        : Boolean := False;
   Have_Backspace_Wrap_Cap : Boolean := False;
   Have_Last_Column        : Boolean := False;
   Have_Auto_Wrap_Cap      : Boolean := False;
   Have_Up_Cap             : Boolean := False;
   Have_Right_Cap          : Boolean := False;

   -- Integer terminal capability flags

   Last_Column : Integer := 80;

   -- Terminal control sequences

   Backspace_Ctlseq       : String_Ptr;
   Insert_Ctlseq          : String_Ptr;
   End_Insert_Ctlseq      : String_Ptr;
   Delete_Ctlseq          : String_Ptr;
   Delete_Mode_Ctlseq     : String_Ptr;
   End_Delete_Mode_Ctlseq : String_Ptr;
   Move_Up_Ctlseq         : String_Ptr;
   Move_Right_Ctlseq      : String_Ptr;

   -- Other declarations needed to use termcap

   Termcap   : Cstring;
   Term_Name : Cstring;

   Am_Word   : constant Cstring := Convert_To_Cstring ("am");
   Bc_Word   : constant Cstring := Convert_To_Cstring ("bc");
   Bs_Word   : constant Cstring := Convert_To_Cstring ("bs");
   Bw_Word   : constant Cstring := Convert_To_Cstring ("bw");
   Co_Word   : constant Cstring := Convert_To_Cstring ("co");
   Dc_Word   : constant Cstring := Convert_To_Cstring ("dc");
   Dm_Word   : constant Cstring := Convert_To_Cstring ("dm");
   Ed_Word   : constant Cstring := Convert_To_Cstring ("ed");
   Ei_Word   : constant Cstring := Convert_To_Cstring ("ei");
   Ic_Word   : constant Cstring := Convert_To_Cstring ("ic");
   Im_Word   : constant Cstring := Convert_To_Cstring ("im");
   Le_Word   : constant Cstring := Convert_To_Cstring ("le");
   Nd_Word   : constant Cstring := Convert_To_Cstring ("nd");
   Up_Word   : constant Cstring := Convert_To_Cstring ("up");
   Term_Word : constant Cstring := Convert_To_Cstring ("TERM");

   -- Bit masks for screen options

   M_Cbreak : constant := 16#02#;
   M_Crmode : constant := 16#10#;
   M_Echo   : constant := 16#08#;
   M_Flush  : constant := 16#08#;
   M_Raw    : constant := 16#20#;

   type Sgtty_Structure is
      record
         Sg_Ispeed : Char; -- input speed
         Sg_Ospeed : Char; -- output speed
         Sg_Erase  : Char; -- erase character
         Sg_Kill   : Char; -- kill character
         Sg_Flags  : Short; -- mode flags   -- different for each UNIX
      end record;

   Stdin_Fileno : constant Thirty_Two_Int := 0;
   Tty          : Sgtty_Structure;
   Res_Flg      : Thirty_Two_Int;

   Terminal_Is_Set_Up : Boolean := False;

   type Flag_Array is array (0 .. 15) of Boolean;
   pragma PACK (Flag_Array);

   function Array_Short is new Unchecked_Conversion (Flag_Array, Short);
   function Short_Array is new Unchecked_Conversion (Short, Flag_Array);

   function Unpack (S : Short) return Flag_Array renames Short_Array;
   function Pack (F : Flag_Array) return Short renames Array_Short;

   Line : String (1 .. Max_Line_Length);

   Cursor_Pos    : Positive := Line'FIRST;
   Line_End      : Natural  := Line'FIRST - 1;
   Column_Offset : Integer  := 0;

   -- Cursor_Pos is the position of the cursor, expressed as a subscript
   -- into the Line buffer; therefore, its minimum value is Line'FIRST

   -- Line_End is the subscript of the last character in Line that has
   -- actually been written to the screen; since it may be that no
   -- characters have been written, its minimum value is Line'FIRST - 1

   -- Column_Offset is the offset required to determine the "true"
   -- cursor position, assuming a 1-relative positioning scheme.
   -- That is, if when Cursor_Pos is 1 the cursor is in the leftmost
   -- position of the screen (column 1), then Column_Offset should be 0.
   -- The "true" column should always be given by the expression
   -- (Cursor_Pos + Column_Offset).
   -- The main use of Column_Offset is in backspacing over multiple lines.

   function Ascii_Character (C : in Char) return Character;
   procedure C_Gtty (Fd          : in Thirty_Two_Int;
                     Tty_Pointer : in Address);
   function C_Read (Fd     : in Thirty_Two_Int;
                    Buff   : in Address;
                    Nbytes : in Thirty_Two_Int) return Integer;
   procedure C_Stty (Fd          : in Thirty_Two_Int;
                     Tty_Pointer : in Address);
   procedure Echo;
   function Getenv (Name : in Cstring) return Cstring;
   procedure Get_Terminal_Capabilities;
   procedure Goto_End_Of_Line;
   procedure Gtty (Fd          : in Thirty_Two_Int;
                   Tty_Pointer : in Address);
   procedure Nl;
   procedure Noraw;
   procedure Noecho;
   procedure Nonl;
   procedure Raw;
   procedure Resetty;
   procedure Savetty;
   procedure Show_Terminal_Capabilities;
   procedure Stty (Fd          : in Thirty_Two_Int;
                   Tty_Pointer : in Address);
   function Tgetent (Bp   : in Cstring;
                     Name : in Cstring) return Thirty_Two_Int;
   function Tgetflag (Id : in Cstring) return Boolean;
   function Tgetnum (Id : in Cstring) return Integer;
   function Tgetstr (Id : in Cstring) return String_Ptr;
   function True_Column (Cursor_Pos : in Positive) return Positive;

   --=========================================================================
   --  <Subprogram Name>
   --    Ascii_Character
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The function Ascii_Character converts a system-dependent
   --    character to an Ada character
   --
   --=========================================================================
   function Ascii_Character (C : in Char) return Character is
   begin
      if C >= 0 and then Integer (C) < 128 then
         return Character'VAL (C);
      else
         return Ascii.Del;
      end if;
   end Ascii_Character;

   --=========================================================================
   --  <Subprogram Name>
   --    Backspace
   --
   --    Last Mod:    Mon Apr 16 18:22:20 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Backspace moves the cursor one or more positions
   --    to the left.
   --
   --=========================================================================
   procedure Backspace is
   begin
      if Cursor_Pos > Line'FIRST then
         if Have_Backspace_Cap then
            if True_Column (Cursor_Pos) > 1 or Have_Backspace_Wrap_Cap then
               Text_Io.Put (Backspace_Ctlseq.all);
            elsif Have_Up_Cap and Have_Right_Cap then
               Text_Io.Put (Move_Up_Ctlseq.all);
               for I in 1 .. Last_Column - 1 loop
                  Text_Io.Put (Move_Right_Ctlseq.all);
               end loop;
            end if;
         else
            Text_Io.Put (Line (Cursor_Pos - 1));
         end if;

         Cursor_Pos := Cursor_Pos - 1;
      end if;
   end Backspace;

   procedure Backspace (N : in Integer) is
   begin
      for I in 1 .. N loop
         Backspace;
      end loop;
   end Backspace;

   --=========================================================================
   --  <Subprogram Name>
   --    Beep
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Beep sounds the terminal alarm signal.
   --    This is normally used to indicate an input error, and so
   --    may be modified to flash the screen instead.
   --
   --=========================================================================
   procedure Beep is
   begin
      Text_Io.Put (Ascii.Bel);
   end Beep;

   pragma INTERFACE (C, C_Gtty);
   pragma LINKNAME (C_Gtty, "gtty");

   pragma INTERFACE (C, C_Read);
   pragma LINKNAME (C_Read, "read");

   pragma INTERFACE (C, C_Stty);
   pragma LINKNAME (C_Stty, "stty");

   --=========================================================================
   --  <Subprogram Name>
   --    Clear_Line
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Clear_Line moves the cursor to the beginning
   --    of the current line and clears all characters to the right
   --    of this point.
   --
   --=========================================================================
   procedure Clear_Line is
   begin
      Goto_End_Of_Line;
      for I in Line'FIRST .. Line_End loop
         Backspace;
         Tty_Operations.Put (' ');
         Backspace;
      end loop;
      Line_End := Line'FIRST - 1;
   end Clear_Line;

   --=========================================================================
   --  <Subprogram Name>
   --    Delete
   --
   --    Last Mod:    Mon Apr 16 18:22:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Delete removes one or more characters from
   --    the cursor's left, moving the cursor leftward as it does so.
   --
   --=========================================================================
   procedure Delete (N : in Integer := 1) is

      -- Delete N characters from the left of the cursor

      N1               : Integer;
      Reposition_Count : Integer := Line_End - Cursor_Pos + 1;

   begin

      if N < 1 then
         return;
      end if;

      -- Do not attempt to delete more characters than there are

      N1 := Cursor_Pos - Line'FIRST;
      if N < N1 then
         N1 := N;
      end if;

      -- Delete the characters from the buffer

      Line (Cursor_Pos - N1 .. Line_End - N1) := Line (Cursor_Pos .. Line_End);
      for I in Line_End - N1 + 1 .. Line_End loop
         Line (I) := ' ';
      end loop;

      -- Update the characters on the screen

      if Have_Delete_Cap and Line_End + Column_Offset + 1 < Last_Column then

         -- Use terminal's delete capability to remove characters
         -- from an unwrapped line

         for I in 1 .. N1 loop
            Backspace;
            if Need_Delete_Mode then
               Text_Io.Put (Delete_Mode_Ctlseq.all);
            end if;
            Text_Io.Put (Delete_Ctlseq.all);
            if Need_Delete_Mode then
               Text_Io.Put (End_Delete_Mode_Ctlseq.all);
            end if;
         end loop;
      elsif Have_Backspace_Cap then

         -- If terminal does not have delete capability, or if
         -- line is wrapped, redraw

         Backspace (N1);
         Tty_Operations.Put (Line (Cursor_Pos .. Line_End));
         Backspace (N1);
         Backspace (Reposition_Count);
      else
         for I in 1 .. N1 loop
            Text_Io.Put ('\');
         end loop;
      end if;

      -- Adjust the line length
      Line_End := Line_End - N1;
   end Delete;

   --=========================================================================
   --  <Subprogram Name>
   --    Echo
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Echo causes the terminal to echo characters
   --    as they are typed.  Such echoing is a feature of normal
   --    terminal mode.
   --
   --=========================================================================
   procedure Echo is
   begin
      Tty.Sg_Flags := Pack (Unpack (Tty.Sg_Flags) or Unpack (M_Echo));
      Stty (Stdin_Fileno, Tty'ADDRESS);
   end Echo;

   --=========================================================================
   --  <Subprogram Name>
   --    Getch
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The function Getch returns the next character of user input,
   --    as soon as the key is struck.  That is, this function does
   --    whatever is necessary to remove operating system interference
   --    (e.g., line buffering) from terminal input.
   --
   --=========================================================================
   function Getch return Character is

      Count : Integer;
      C     : Char := Character'POS (Ascii.Eot);

   begin
      if not Terminal_Is_Set_Up then
         Terminal_Set;
         if Cursor_Pos = Line'FIRST then
            Column_Offset := Integer (Text_Io.Col) - Cursor_Pos;
         end if;
      end if;

      Count := C_Read (Stdin_Fileno, C'ADDRESS, Thirty_Two_Int (1));
      if Count < 1 then
         raise Text_Io.End_Error;
      else
         return Ascii_Character (C);
      end if;
   end Getch;

   pragma INTERFACE (C, Getenv);

   --=========================================================================
   --  <Subprogram Name>
   --    Get_Terminal_Capabilities
   --
   --    Last Mod:    Mon Apr 16 18:22:42 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Get_Terminal_Capabilities obtains knowledge of
   --    the ability of the terminal to perform certain functions, such
   --    as the ability to backspace the cursor or the ability to
   --    insert new characters, as well as knowledge of the sequences
   --    required to cause the terminal to perform these functions.
   --    The information obtained by this procedure is stored in
   --    global variables.
   --
   --=========================================================================
   procedure Get_Terminal_Capabilities is
      Debug_Termcap : constant Boolean := False;
   begin
      Term_Name := Getenv (Term_Word);

      if Debug_Termcap then
         Text_Io.Put ("$TERM = " & Convert_To_Astring (Term_Name));
         New_Line;
      end if;

      declare
         Temporary_Ada_String : String (1 .. 1024) := (others => Ascii.Nul);
         Result               : Thirty_Two_Int;
      begin
         Termcap := Convert_To_Cstring (Temporary_Ada_String);
         if Tgetent (Termcap, Term_Name) > 0 then
            Have_Termcap := True;
         end if;
      end;

      if Have_Termcap then

         if Tgetflag (Bs_Word) then

            -- ASCII backspace character does backspacing

            Have_Backspace_Cap   := True;
            Backspace_Ctlseq     := new String (1 .. 1);
            Backspace_Ctlseq (1) := Ascii.Bs;
         else

            -- Check for explicit backspace sequence

            Backspace_Ctlseq := Tgetstr (Bc_Word);
            if Backspace_Ctlseq /= null then
               Have_Backspace_Cap := True;
            else

               -- Try for move-left sequence

               Backspace_Ctlseq := Tgetstr (Le_Word);
               if Backspace_Ctlseq /= null then
                  Have_Backspace_Cap := True;
               end if;
            end if;
         end if;

         Insert_Ctlseq := Tgetstr (Ic_Word);
         if Insert_Ctlseq /= null then

            -- Terminal has single-character insert capability

            Have_Insert_Cap := True;
         else

            -- Try for insert mode

            Insert_Ctlseq     := Tgetstr (Im_Word);
            End_Insert_Ctlseq := Tgetstr (Ei_Word);
            if Insert_Ctlseq /= null and End_Insert_Ctlseq /= null then
               Have_Insert_Cap := True;
               Need_Insert_End := True;
            end if;
         end if;

         -- The "dc" capability deletes the character under the cursor;
         -- therefore, if we do not have the capability to backspace,
         -- then, for practical purposes, we do not have the capability
         -- to delete (it is unlikely that this will arise in real life).

         if Have_Backspace_Cap then
            Delete_Ctlseq := Tgetstr (Dc_Word);
            if Delete_Ctlseq /= null then

               -- Terminal has single-character delete capability

               Have_Delete_Cap := True;

               -- See whether terminal must be in delete mode to delete

               Delete_Mode_Ctlseq     := Tgetstr (Dm_Word);
               End_Delete_Mode_Ctlseq := Tgetstr (Ed_Word);
               if Delete_Mode_Ctlseq /= null and End_Delete_Mode_Ctlseq /= null
                   then
                  Need_Delete_Mode := True;
               end if;
            end if;
         end if;

         if Tgetflag (Bw_Word) then

            -- Terminal backspace sequence will return to previous
            -- line, if necessary

            Have_Backspace_Wrap_Cap := True;
         end if;

         declare
            Co : Integer := Tgetnum (Co_Word);
         begin
            if Co >= 0 then
               Last_Column      := Co;
               Have_Last_Column := True;
            end if;
         end;

         if Tgetflag (Am_Word) then

            -- Terminal feeds a new line automatically when cursor
            -- reaches end of current line

            Have_Auto_Wrap_Cap := True;
         end if;

         Move_Up_Ctlseq := Tgetstr (Up_Word);
         if Move_Up_Ctlseq /= null then

            -- Terminal has capability to move up

            Have_Up_Cap := True;
         end if;

         Move_Right_Ctlseq := Tgetstr (Nd_Word);
         if Move_Right_Ctlseq /= null then

            -- Terminal has capability to move right non-destructively

            Have_Right_Cap := True;
         end if;

         if Debug_Termcap then
            Show_Terminal_Capabilities;
         end if;

      end if;

      -- Subtract 1 from Last_Column to handle problems with line wrap

      Last_Column := Last_Column - 1;

   end Get_Terminal_Capabilities;

   --=========================================================================
   --  <Subprogram Name>
   --    Goto_End_Of_Line
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Goto_End_Of_Line positions the cursor immediately
   --    to the right of the righmost character on the current line.
   --
   --=========================================================================
   procedure Goto_End_Of_Line is
   begin
      if Cursor_Pos < Line_End + 1 then
         Move_Right (Line_End + 1 - Cursor_Pos);
      end if;
   end Goto_End_Of_Line;

   --=========================================================================
   --  <Subprogram Name>
   --    Gtty
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Gtty obtains the current settings of some
   --    system-dependent and user-dependent terminal options, chiefly
   --    so that these options can later be restored.
   --
   --=========================================================================
   procedure Gtty (Fd          : in Thirty_Two_Int;
                   Tty_Pointer : in Address) is
   begin
      C_Gtty (Thirty_Two_Int (Fd), Tty_Pointer);
   end Gtty;

   --=========================================================================
   --  <Subprogram Name>
   --    Insert_Char
   --
   --    Last Mod:    Mon Apr 16 18:23:06 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Insert_Char inserts a single character to the
   --    left of the cursor, leaving the cursor positioned on the
   --    same character (not in the same spot) as before the operation.
   --    The usual purpose of this procedure will be to add characters
   --    to the end of a line, but this is not its only capability.
   --
   --=========================================================================
   procedure Insert_Char (C : in Character) is

      Reposition_Count : Integer := Line_End - Cursor_Pos + 1;

   begin

      -- Insert a character to the left of the cursor
      -- The cursor should wind up to the right of the new character

      if Cursor_Pos <= Line'LAST then

         if Reposition_Count = 0 then

            -- If we are extending the line, just write the new character
            -- (the Put routine will copy it into the Line buffer)

            Line (Cursor_Pos) := C;
            Tty_Operations.Put (C);
         else

            -- Rearrange the Line buffer to reflect the insertion

            Line (Cursor_Pos + 1 .. Line_End + 1) := Line (Cursor_Pos ..
                                                           Line_End);
            Line (Cursor_Pos) := C;

            if Have_Insert_Cap and Line_End + Column_Offset + 2 < Last_Column
                then

               -- If the terminal can do insertions,
               -- let the terminal do this one,
               -- unless the new line will have to wrap on the screen

               Text_Io.Put (Insert_Ctlseq.all);
               Tty_Operations.Put (C);
               if Need_Insert_End then
                  Text_Io.Put (End_Insert_Ctlseq.all);
               end if;
               Line_End := Line_End + 1;

            else

               -- Otherwise, redraw the line to the right of the
               -- cursor, and move back to where we started;
               -- if line wrapping is required, this is one way
               -- (albeit far from the most efficient way) to
               -- accomplish it

               if Have_Backspace_Cap then
                  Tty_Operations.Put (Line (Cursor_Pos .. Line_End + 1));
                  Backspace (Reposition_Count);
               else
                  Text_Io.Put (C);
               end if;
            end if;
         end if;

      end if;
   end Insert_Char;

   --=========================================================================
   --  <Subprogram Name>
   --    Move_Left
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Move_Left moves the cursor one or more positions
   --    to the left, or until the beginning of the current line.
   --
   --=========================================================================
   procedure Move_Left (N : in Integer := 1) is
   begin
      Backspace (N);
   end Move_Left;

   --=========================================================================
   --  <Subprogram Name>
   --    Move_Right
   --
   --    Last Mod:    Mon Apr 16 18:23:21 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Move_Right moves the cursor one or more positions
   --    to the right, but never more than one position past the last
   --    character the user has typed on the current line.  That is,
   --    this procedure may not be used to extend a line's length.
   --
   --=========================================================================
   procedure Move_Right (N : in Integer := 1) is
   begin
      for I in 1 .. N loop
         exit when Cursor_Pos > Line_End;
         Tty_Operations.Put (Line (Cursor_Pos));
      end loop;
   end Move_Right;

   --=========================================================================
   --  <Subprogram Name>
   --    New_Line
   --
   --    Last Mod:    Mon Apr 16 18:23:43 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure New_Line causes the current line to be scrolled
   --    upward and the cursor to be positioned at the beginning of
   --    a new line.
   --
   --    NOTE: This procedure causes the terminal to return to
   --          normal I/O mode.
   --
   --=========================================================================
   procedure New_Line is
   begin
      if Line_End + Column_Offset + 1 > Last_Column then

         -- If this line is wrapped on the screen, move to
         -- the end, so that scrolling will still show the
         -- whole thing

         Goto_End_Of_Line;
      end if;

      Text_Io.Put (Ascii.Cr);
      Text_Io.Put (Ascii.Lf);

      Cursor_Pos := Line'FIRST;
      Line_End   := Line'FIRST - 1;

      Terminal_Reset;
   end New_Line;

   --=========================================================================
   --  <Subprogram Name>
   --    Nl
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Nl causes the terminal carriage-return to be
   --    read as a line-feed character, and causes the writing of a
   --    line-feed to send a carriage-return and a line-feed.
   --    This character translation is the terminal's normal mode.
   --
   --=========================================================================
   procedure Nl is
   begin
      Tty.Sg_Flags := Pack (Unpack (Tty.Sg_Flags) or Unpack (M_Crmode));
      Stty (Stdin_Fileno, Tty'ADDRESS);
   end Nl;

   --=========================================================================
   --  <Subprogram Name>
   --    Noecho
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Noecho causes the terminal not to echo characters
   --    as they are typed.
   --
   --=========================================================================
   procedure Noecho is
   begin
      Tty.Sg_Flags := Pack (Unpack (Tty.Sg_Flags) and not Unpack (M_Echo));
      Stty (Stdin_Fileno, Tty'ADDRESS);
   end Noecho;

   --=========================================================================
   --  <Subprogram Name>
   --    Nonl
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Nonl causes the line-feed character to have no
   --    special treatment on input or output.
   --
   --=========================================================================
   procedure Nonl is
   begin
      Tty.Sg_Flags := Pack (Unpack (Tty.Sg_Flags) and not Unpack (M_Crmode));
      Stty (Stdin_Fileno, Tty'ADDRESS);
   end Nonl;

   --=========================================================================
   --  <Subprogram Name>
   --    Noraw
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Noraw causes the terminal to accumulate complete
   --    lines of input before sending any characters to the program.
   --    This is a feature of the terminal's normal mode.
   --
   --=========================================================================
   procedure Noraw is
   begin
      Tty.Sg_Flags := Pack (Unpack (Tty.Sg_Flags) and not Unpack (M_Raw));
      Stty (Stdin_Fileno, Tty'ADDRESS);
   end Noraw;

   --=========================================================================
   --  <Subprogram Name>
   --    Put
   --
   --    Last Mod:    Mon Apr 16 18:23:58 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Put writes a character or string to the terminal,
   --    starting at the current cursor position.  The cursor is left
   --    one position to the right of the last character written.
   --
   --=========================================================================
   procedure Put (C : in Character) is
   begin
      if Cursor_Pos < Line'LAST + 1 then

         -- If the cursor is already past the maximum number
         -- of characters on a line, then we cannot add any
         -- new characters

         -- If not, add the new character by writing it,
         -- doing line wrap, if necessary (we never assume
         -- that the terminal does automatic line wrapping,
         -- because some termcap entries lie about this)

         Text_Io.Put (C);
         if True_Column (Cursor_Pos) >= Last_Column then
            Text_Io.Put (Ascii.Cr);
            Text_Io.Put (Ascii.Lf);
         end if;

         -- Be sure the character is in our copy of the current line

         Line (Cursor_Pos) := C;

         -- Advance our record of the cursor position

         Cursor_Pos := Cursor_Pos + 1;

         -- If the character to the left of the cursor position
         -- is now past the end of the line,
         -- we have made the line longer

         if Cursor_Pos - 1 > Line_End then
            Line_End := Line_End + 1;
         end if;
      end if;
   end Put;

   procedure Put (S : in String) is
   begin
      for I in S'RANGE loop
         Put (S (I));
      end loop;
   end Put;

   --=========================================================================
   --  <Subprogram Name>
   --    Raw
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Raw causes the terminal to transfer input
   --    characters to the program as soon as they are typed.
   --
   --=========================================================================
   procedure Raw is
   begin
      Tty.Sg_Flags := Pack (Unpack (Tty.Sg_Flags) or Unpack (M_Raw));
      Stty (Stdin_Fileno, Tty'ADDRESS);
   end Raw;

   --=========================================================================
   --  <Subprogram Name>
   --    Resetty
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Resetty restores certain terminal options
   --    to the values they had when the package was elaborated, i.e.,
   --    to normal mode.
   --
   --=========================================================================
   procedure Resetty is
   begin
      Tty.Sg_Flags := Short (Res_Flg);
      Stty (Stdin_Fileno, Tty'ADDRESS);
   end Resetty;

   --=========================================================================
   --  <Subprogram Name>
   --    Savetty
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Savetty is called during package elaboration
   --    to save the values of certain terminal options.
   --
   --=========================================================================
   procedure Savetty is
   begin
      Gtty (Stdin_Fileno, Tty'ADDRESS);
      Res_Flg := Thirty_Two_Int (Tty.Sg_Flags);
   end Savetty;

   --=========================================================================
   --  <Subprogram Name>
   --    Show_Terminal_Capabilities
   --
   --    Last Mod:    Mon Apr 16 18:25:12 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Show_Terminal_Capabilities displays the
   --    terminal capabilities used by history and line editing.
   --    It is currently used only for debugging.
   --
   --=========================================================================
   procedure Show_Terminal_Capabilities is
   begin
      if Have_Termcap then

         Text_Io.Put ("$TERMCAP = " & Convert_To_Astring (Termcap));
         New_Line;

         Text_Io.Put ("Terminal can");
         if not Have_Backspace_Cap then
            Text_Io.Put ("not");
         end if;
         Text_Io.Put (" backspace");
         New_Line;

         Text_Io.Put ("Terminal can");
         if not Have_Insert_Cap then
            Text_Io.Put ("not");
         end if;
         Text_Io.Put (" insert");
         New_Line;

         Text_Io.Put ("Terminal can");
         if not Have_Delete_Cap then
            Text_Io.Put ("not");
         end if;
         Text_Io.Put (" delete");
         New_Line;

         Text_Io.Put ("Terminal backspace does");
         if not Have_Backspace_Wrap_Cap then
            Text_Io.Put (" not");
         end if;
         Text_Io.Put (" wrap");
         New_Line;

         Text_Io.Put ("Terminal");
         if Have_Last_Column then
            Text_Io.Put (" has");
         else
            Text_Io.Put (" assumed to have");
         end if;
         Text_Io.Put (Integer'IMAGE (Last_Column));
         Text_Io.Put (" columns");
         New_Line;

         Text_Io.Put ("Terminal does");
         if not Have_Auto_Wrap_Cap then
            Text_Io.Put (" not");
         end if;
         Text_Io.Put (" wrap automatically");
         New_Line;

         Text_Io.Put ("Terminal can");
         if not Have_Up_Cap then
            Text_Io.Put ("not");
         end if;
         Text_Io.Put (" move up");
         New_Line;

         Text_Io.Put ("Terminal can");
         if not Have_Right_Cap then
            Text_Io.Put ("not");
         end if;
         Text_Io.Put (" move right");
         New_Line;

      end if;
   end Show_Terminal_Capabilities;

   --=========================================================================
   --  <Subprogram Name>
   --    Stty
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Stty changes certain system_dependent and
   --    and user-dependent terminal options.
   --
   --=========================================================================
   procedure Stty (Fd          : in Thirty_Two_Int;
                   Tty_Pointer : in Address) is
   begin
      C_Stty (Thirty_Two_Int (Fd), Tty_Pointer);
   end Stty;

   --=========================================================================
   --  <Subprogram Name>
   --    Terminal_Erase_Character
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The function Terminal_Erase_Character returns the
   --    character normally used for deleting (erasing) a character on
   --    the screen.  This is used to permit the ACE delete function to be
   --    bound to the same key that the user employs for this function
   --    outside ACE.
   --
   --=========================================================================
   function Terminal_Erase_Character return Character is
   begin
      return Ascii_Character (Tty.Sg_Erase);
   end Terminal_Erase_Character;

   --=========================================================================
   --  <Subprogram Name>
   --    Terminal_Reset
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Terminal_Reset returns the terminal to normal
   --    I/O mode.
   --
   --=========================================================================
   procedure Terminal_Reset is
   begin
      Noraw;
      Echo;
      Nl;
      Resetty;
      Terminal_Is_Set_Up := False;
   end Terminal_Reset;

   --=========================================================================
   --  <Subprogram Name>
   --    Terminal_Set
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Terminal_Set causes the terminal to enter the mode
   --    necessary for Getch operations.
   --
   --=========================================================================
   procedure Terminal_Set is
   begin
      Nonl;
      Noecho;
      Raw;
      Terminal_Is_Set_Up := True;
   end Terminal_Set;

   pragma INTERFACE (C, Tgetent);

   --=========================================================================
   --  <Subprogram Name>
   --    Tgetflag
   --
   --    Last Mod:    Wed Apr 11 15:10:58 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The function Tgetflag obtains the value of Boolean
   --    terminal capabilities.
   --
   --=========================================================================
   function Tgetflag (Id : in Cstring) return Boolean is

      function C_Tgetflag (Id : in Cstring) return Thirty_Two_Int;
      pragma INTERFACE (C, C_Tgetflag);
      pragma LINKNAME (C_Tgetflag, "tgetflag");

   begin
      return C_Tgetflag (Id) /= 0;
   end Tgetflag;

   --=========================================================================
   --  <Subprogram Name>
   --    Tgetnum
   --
   --    Last Mod:    Wed Apr 11 15:10:58 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The function Tgetnum obtains the value of numeric
   --    terminal capabilities.
   --
   --=========================================================================
   function Tgetnum (Id : in Cstring) return Integer is

      function C_Tgetnum (Id : in Cstring) return Thirty_Two_Int;
      pragma INTERFACE (C, C_Tgetnum);
      pragma LINKNAME (C_Tgetnum, "tgetnum");

   begin
      return Integer (C_Tgetnum (Id));
   end Tgetnum;

   --=========================================================================
   --  <Subprogram Name>
   --    Tgetstr
   --
   --    Last Mod:    Wed Apr 11 15:10:58 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The function Tgetstr returns the control sequences required
   --    for certain terminal capabilities.
   --
   --=========================================================================
   function Tgetstr (Id : in Cstring) return String_Ptr is

      function C_Tgetstr (Id   : in Cstring;
                          Area : in Address) return Cstring;
      pragma INTERFACE (C, C_Tgetstr);
      pragma LINKNAME (C_Tgetstr, "tgetstr");

      Ada_String_1          : String (1 .. 1024) := (others => Ascii.Nul);
      Return_String         : Cstring;
      Return_String_Address : Cstring;
      Return_Ptr            : String_Ptr;

   begin
      Return_String         := Convert_To_Cstring (Ada_String_1);
      Return_String_Address := Return_String;

      Return_String := C_Tgetstr (Id, Return_String_Address'ADDRESS);

      if Return_String = Null_Cstring then
         return null;
      else
         declare
            Ada_String_2 : constant String :=
               Convert_To_Astring (Return_String);
         begin
            if Ada_String_2 = "" then
               Return_Ptr := null;
            else
               Return_Ptr     := new String (Ada_String_2'RANGE);
               Return_Ptr.all := Ada_String_2;
            end if;
         end;
         return Return_Ptr;
      end if;
   end Tgetstr;

   --=========================================================================
   --  <Subprogram Name>
   --    True_Column
   --
   --    Last Mod:    Mon Apr 16 18:24:39 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The function True_Column returns the terminal column under the
   --    cursor, counting the leftmost column as 1.
   --
   --=========================================================================
   function True_Column (Cursor_Pos : in Positive) return Positive is
   begin
      return ((Cursor_Pos + Column_Offset - 1) mod Last_Column) + 1;
   end True_Column;

begin
   Savetty;
   Get_Terminal_Capabilities;
end Tty_Operations;
