--|$Log:	tty_ops.a,v $
--|Revision 8.0  91/01/17  10:59:41  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Tty_Operations
--
----------------------------------------------------------------------------
--
--
-- Abstract
--
-- This package offers a number of procedures for positioning the
-- terminal (tty) cursor and manipulating the terminal screen directly.
-- It also permits characters to be read in "raw" mode, that is,
-- as soon as keys are struck; this makes it possible to respond to
-- such things as arrow keys and function keys.
--
-- Those who elect to use this package should use only the operations
-- of this package should use only the operations provided by this
-- package while the terminal is in raw mode.  That is, while the
-- terminal is in raw mode, the caller should not read or write
-- to or from the terminal with Text_IO.  The function Getch causes
-- the terminal to enter this mode; the procedures New_Line and
-- Terminal_Reset cause it to return to normal mode.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------
with System;
use System;
with Unix_Types;
use Unix_Types;

package Tty_Operations is

   Max_Line_Length : Integer := 1024;

   --=========================================================================
   --  <Subprogram Name>
   --    Beep
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Beep sounds the terminal alarm signal.
   --    This is normally used to indicate an input error, and so
   --    may be modified to flash the screen instead.
   --
   --=========================================================================
   procedure Beep;
   --=========================================================================
   --  <Subprogram Name>
   --    Clear_Line
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Clear_Line moves the cursor to the beginning
   --    of the current line and clears all characters to the right
   --    of this point.
   --
   --=========================================================================
   procedure Clear_Line;
   --=========================================================================
   --  <Subprogram Name>
   --    Delete
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Delete removes one or more characters from
   --    the cursor's left, moving the cursor leftward as it does so.
   --
   --=========================================================================
   procedure Delete (N : in Integer := 1);
   --=========================================================================
   --  <Subprogram Name>
   --    Getch
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The function Getch returns the next character of user input,
   --    as soon as the key is struck.  That is, this function does
   --    whatever is necessary to remove operating system interference
   --    (e.g., line buffering) from terminal input.
   --
   --=========================================================================
   function Getch return Character;
   --=========================================================================
   --  <Subprogram Name>
   --    Insert_Char
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Insert_Char inserts a single character to the
   --    left of the cursor, leaving the cursor positioned on the
   --    same character (not in the same spot) as before the operation.
   --    The usual purpose of this procedure will be to add characters
   --    to the end of a line, but this is not its only capability.
   --
   --=========================================================================
   procedure Insert_Char (C : in Character);
   --=========================================================================
   --  <Subprogram Name>
   --    Move_Left
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Move_Left moves the cursor one or more positions
   --    to the left, or until the beginning of the current line.
   --
   --=========================================================================
   procedure Move_Left (N : in Integer := 1);
   --=========================================================================
   --  <Subprogram Name>
   --    Move_Right
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Move_Right moves the cursor one or more positions
   --    to the right, but never more than one position past the last
   --    character the user has typed on the current line.  That is,
   --    this procedure may not be used to extend a line's length.
   --
   --=========================================================================
   procedure Move_Right (N : in Integer := 1);
   --=========================================================================
   --  <Subprogram Name>
   --    New_Line
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure New_Line causes the current line to be scrolled
   --    upward and the cursor to be positioned at the beginning of
   --    a new line.
   --
   --    NOTE: This procedure causes the terminal to return to
   --          normal I/O mode.
   --
   --=========================================================================
   procedure New_Line;
   --=========================================================================
   --  <Subprogram Name>
   --    Put
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Put writes a character or string to the terminal,
   --    starting at the current cursor position.  The cursor is left
   --    one position to the right of the last character written.
   --
   --=========================================================================
   procedure Put (C : in Character);
   procedure Put (S : in String);
   --=========================================================================
   --  <Subprogram Name>
   --    Terminal_Erase_Character
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The function Terminal_Erase_Character returns the
   --    character normally used for deleting (erasing) a character on
   --    the screen.  This is used to permit the ACE delete function to be
   --    bound to the same key that the user employs for this function
   --    outside ACE.
   --
   --=========================================================================
   function Terminal_Erase_Character return Character;
   --=========================================================================
   --  <Subprogram Name>
   --    Terminal_Reset
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Terminal_Reset returns the terminal to normal
   --    I/O mode.
   --
   --=========================================================================
   procedure Terminal_Reset;
   --=========================================================================
   --  <Subprogram Name>
   --    Terminal_Set
   --
   --    Last Mod:    Tue Apr 10 16:35:29 EDT 1990 (rhp)
   --    Author:      Robert H. Pollack
   --
   --  <Description>
   --
   --    The procedure Terminal_Set causes the terminal to enter the mode
   --    necessary for Getch operations.
   --
   --=========================================================================
   procedure Terminal_Set;

end Tty_Operations;
