--|$Log:	aprgut.a,v $
--|Revision 8.0  91/01/17  10:40:54  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Ace_Program_Unit
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    This package defines the ACE abstract data types for packages
--    and subprograms that are constructed within the command environment.
--    The types for packages and subprograms are defined; and the operations
--    which may be applied to objects to these types are also defined.
--    These types and operations allow ACE packages and subprograms (Ada code)
--    to be manipulated as data.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Io_Exceptions;

with Dump_Routines;
use Dump_Routines;

with Context_Control;

package Ace_Program_Unit is

   -- ------
   -- Limit on number of open ACE program units (packages, subprograms)
   -- ------
   Program_Unit_Limit : constant := 32;

   -- ------
   -- The Abstract Data Type (ADT) for Package objects and Subprogram Objects
   -- ------
   type Package_Type is new Integer range 0 .. Program_Unit_Limit;
   type Subprogram_Type is new Integer range 0 .. Program_Unit_Limit;

   -- ------
   -- The possible modes for an ACE program unit
   -- ------
   type Program_Unit_Mode is (In_Prog_Unit, Out_Prog_Unit);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create
   --
   --  <Description>
   --
   --    This procedure establishes a new ACE package, and associates this
   --    package with the given ACE package object.  The Name must not clash
   --    with the currently existing set of visible identifiers within ACE.
   --    The mode of the package object is set to the given Program_Unit_Mode.
   --    The Form parameter is currently unused.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Name_Error is raised if the Name parameter is the name of an
   --      existing Ada object within ACE.
   --    Use_Error is raised if the maximum number of open ACE package
   --      objects has been reached.
   --
   --  <Exceptions Handled>
   --
   --
   --==========================================================================

   procedure Create (Context     : in out Context_Control.Context_Definition;
                     Ace_Package : in out Package_Type;
                     Mode        : in Program_Unit_Mode := Out_Prog_Unit;
                     Name        : in String            := "";
                     Form        : in String            := "");

   --
   -- No "Create" supported for subprograms currently
   --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Open
   --
   --  <Description>
   --
   --    This procedure associates an ACE package object with the named
   --    package that must be currently defined within ACE.  The mode of
   --    the package object is set to the given Program_Unit_Mode.
   --    The Form parameter is currently unused.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Name_Error is raised if the Name parameter is not the name of an
   --      existing package.
   --    Use_Error is raised if the maximum number of open ACE package
   --      objects has been reached.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   procedure Open (Context     : in Context_Control.Context_Definition;
                   Ace_Package : in out Package_Type;
                   Mode        : in Program_Unit_Mode;
                   Name        : in String;
                   Form        : in String := "");


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Open
   --
   --  <Description>
   --
   --    This procedure associates an ACE subprogram object with the named
   --    subprogram that must be currently defined within ACE.  The mode of
   --    the subprogram object is set to the given Program_Unit_Mode.
   --    The Form parameter is currently unused.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Name_Error is raised if the Name parameter is not the name of an
   --      existing subprogram.
   --    Use_Error is raised if the maximum number of open ACE subprogram
   --      objects has been reached.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   procedure Open (Context        : in out Context_Control.Context_Definition;
                   Ace_Subprogram : in out Subprogram_Type;
                   Mode           : in Program_Unit_Mode;
                   Name           : in String;
                   Form           : in String := "");


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Close
   --
   --  <Description>
   --
   --    This procedure closes an ACE package object.  This terminates the
   --    association between the ACE package object and the actual package.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the package object is not open.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   procedure Close (Ace_Package : in out Package_Type);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Close
   --
   --  <Description>
   --
   --    This procedure closes an ACE subprogram object.  This terminates
   --    the association between the ACE subprogram object and the actual
   --    subprogram.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the subprogram object is not open.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   procedure Close (Ace_Subprogram : in out Subprogram_Type);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Delete
   --
   --  <Description>
   --
   --    This procedure deletes an ACE package object.  The deletion is
   --    accomplished by marking the symbol table entry associated with the
   --    object as "Removed".
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the package object is not open.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   procedure Delete (Ace_Package : in out Package_Type);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Delete
   --
   --  <Description>
   --
   --    This procedure deletes an ACE subprogram object.  The deletion is
   --    accomplished by marking the symbol table entry associated with the
   --    object as "Removed".
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the subprogram object is not open.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   procedure Delete (Ace_Subprogram : in out Subprogram_Type);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Mode
   --
   --  <Description>
   --
   --    This function returns the program unit mode associated with an
   --    open ACE package object.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the package object is not open.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   function Mode (Ace_Package : in Package_Type) return Program_Unit_Mode;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Mode
   --
   --  <Description>
   --
   --    This function returns the program unit mode associated with an
   --    open ACE subprogram object.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the subprogram object is not open.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   function Mode (Ace_Subprogram : in Subprogram_Type) return Program_Unit_Mode;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Name
   --
   --  <Description>
   --
   --    This function returns the name associated with an ACE package
   --    object.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the package object is not open.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   function Name (Ace_Package : in Package_Type) return String;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Name
   --
   --  <Description>
   --
   --    This function returns the name associated with an ACE subprogram
   --    object.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the subprogram object is not open.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   function Name (Ace_Subprogram : in Subprogram_Type) return String;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Form
   --
   --  <Description>
   --
   --    This function returns the form string associated with an ACE package
   --    object.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the package object is not open.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   function Form (Ace_Package : in Package_Type) return String;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Form
   --
   --  <Description>
   --
   --    This function returns the form string associated with an ACE
   --    subprogram object.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the subprogram object is not open.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   function Form (Ace_Subprogram : in Subprogram_Type) return String;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Open
   --
   --  <Description>
   --
   --    This function returns an indication of whether the ACE package
   --    object is open.  True is returned when the object is open; otherwise
   --    False is returned.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    <none>
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   function Is_Open (Ace_Package : in Package_Type) return Boolean;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Open
   --
   --  <Description>
   --
   --    This function returns an indication of whether the ACE subprogram
   --    object is open.  True is returned when the object is open; otherwise
   --    False is returned.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    <none>
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   function Is_Open (Ace_Subprogram : in Subprogram_Type) return Boolean;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Put
   --
   --  <Description>
   --
   --    This procedures moves an ACE subprogram object into an ACE
   --    package object.  The parameters Action (specification or
   --    implementation) and Visibility (visible or hidden) control the
   --    location within the package into which the subprogram is inserted.
   --    The location for a package includes the package specification
   --    (visible and private), and the package body.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if either the subprogram object or package
   --      object is not open.
   --    Mode_Error is raised if the mode of the package object does permit
   --      writing to the package.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   type Action_Type is (Specification, Implementation);
   type Visibility_Type is (Visible, Hidden);

   procedure Put (Context         : in out Context_Control.Context_Definition;
                  Subprogram_Item : in Subprogram_Type;
                  Into_Package    : in Package_Type;
                  Action          : in Action_Type;
                  Visibility      : in Visibility_Type);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    List
   --
   --  <Description>
   --
   --    This procedure displays an ACE package object.  The Ada code
   --    description is the medium in which the package is displayed.
   --    The list method controls whether the specification, the body,
   --    or both of these are displayed.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the package object is not open.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================


   procedure List (Ace_Package : in Package_Type;
                   List_Method : in List_Mode := Both);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    List
   --
   --  <Description>
   --
   --    This procedure displays an ACE subprogram object.  The Ada code
   --    description is the medium in which the subprogram is displayed.
   --    The list method controls whether the specification, the body,
   --    or both of these are displayed.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the subprogram object is not open.
   --
   --  <Exceptions Handled>
   --
   --    <none>
   --==========================================================================

   procedure List (Ace_Subprogram : in Subprogram_Type;
                   List_Method    : in List_Mode := Both);


   -- Exceptions from standard Text_Io that are raised by this package
   Status_Error : exception renames Io_Exceptions.Status_Error;
   Mode_Error   : exception renames Io_Exceptions.Mode_Error;
   Name_Error   : exception renames Io_Exceptions.Name_Error;
   Use_Error    : exception renames Io_Exceptions.Use_Error;

private
   type Program_Unit_Info is
      record
         Symtbl_Entry : Ref_Symbol_Table_Entry;
         Mode         : Program_Unit_Mode;
      end record;

   Package_Entries    : array (Package_Type'FIRST .. Package_Type'LAST) of
   Program_Unit_Info := (others => (Symtbl_Entry => null,
                                    Mode         => Out_Prog_Unit));

   Subprogram_Entries : array (Subprogram_Type'FIRST .. Subprogram_Type'LAST)
    of Program_Unit_Info := (others => (Symtbl_Entry => null,
                                        Mode         => Out_Prog_Unit));

end Ace_Program_Unit;







