--|$Log:	dir.a,v $
--|Revision 8.0  91/01/17  10:41:05  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Directory_Objects
--
---------------------------------------------------------------
--
--
-- Abstract
--
--  Directory_Objects is an interface from ACE's environment to operations
--  that make directories "look like" files.  On Unix this can, be assumed
--  to be true, and these routines are not needed, but on MS-DOS and other
--  machines this may not be true.  In any case, we retain this layer
--  for Unix for consistency.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------
with Ace_Universal_Types;
use Ace_Universal_Types;

-- Only used in the definition of type directory objects
with Ace_Text_Io;
use Ace_Text_Io;

package Directory_Objects is

   -- Directory_Object object values should never overlap with File_Type object
   -- values.  That is why the Directory_Object starts were File_Type ends,
   -- therefore any time a File_Type routine is called with a Directory_Object
   -- values (which can happen, because Text_Io functions are derived for
   -- the directory_object in the environment) an exception will be raised.
   -- The exception should inform the user of his/her error.  Perhaps this
   -- information really belongs in the body of this package? [TBD] wpl

   Directory_Limit : constant := 100;

   subtype Directory_Object is Ace_Integer range 0 .. Directory_Limit;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create.
   --
   --  <Description>
   --
   --    Create a named directory, and return an object associated with
   --    that directory.
   --
   --  <Exceptions Raised>
   --
   --    Same as Text_Io's exceptions
   --
   --==========================================================================
   procedure Create (Directory : in out Directory_Object;
                     Name      : in String := "";
                     Form      : in String := "");

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Open
   --
   --  <Description>
   --
   --    Open associates an object with an existing directory.
   --
   --  <Exceptions Raised>
   --
   --    The same as Text_Io's open.
   --
   --==========================================================================
   procedure Open (Directory : in out Directory_Object;
                   Name      : in String;
                   Form      : in String := "");

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Close
   --
   --  <Description>
   --
   --    Close disassociates an object from a directory file.
   --
   --  <Exceptions Raised>
   --
   --    The same as Text_Io's name.
   --
   --==========================================================================
   procedure Close (Directory : in out Directory_Object);

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Name
   --
   --  <Description>
   --
   --    From a directory object return the name of the directory as a
   --    String.
   --
   --  <Exceptions Raised>
   --
   --    The same as Text_Io's name.
   --
   --==========================================================================
   function Name (Directory : in Directory_Object) return String;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Form
   --
   --  <Description>
   --
   --    Return the Form String that was used in the Open or Create that
   --    was used to associated the object with an existing object.
   --
   --  <Exceptions Raised>
   --
   --    Same as Text_Io's Form.
   --
   --==========================================================================
   function Form (Directory : in Directory_Object) return String;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Open
   --
   --  <Description>
   --
   --    This routine returns a boolean true if the object has been
   --    associated with a directory.
   --
   --  <Exceptions Raised>
   --
   --    Same as Text_Io's Is_Open.
   --
   --==========================================================================
   function Is_Open (Directory : in Directory_Object) return Boolean;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Home_Directory
   --
   --  <Description>
   --
   --    Returns an object that has been associated with the home
   --    directory (defined as the directory in which ACE was originally
   --    started).
   --
   --==========================================================================
   function Home_Directory return Directory_Object;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Current_Directory
   --
   --  <Description>
   --
   --    Returns an object that has been associated with the current
   --    working directory.
   --
   --==========================================================================
   function Current_Directory return Directory_Object;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    List
   --
   --  <Description>
   --
   --    List will output a list of files that are contained in the
   --    directory object.
   --
   --  <Exceptions Raised>
   --
   --    Same as Text_Io's Set_Input.
   --
   --==========================================================================
   procedure List (Directory : in Directory_Object);

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Directory
   --
   --  <Description>
   --
   --    This routines will change the current directory to the specified
   --    directory.
   --
   --  <Exceptions Raised>
   --
   --    Same as Text_Io's Set_Input.
   --
   --==========================================================================
   procedure Set_Directory (Directory : in Directory_Object);


end Directory_Objects;
