--|$Log:	drtns.a,v $
--|Revision 8.0  91/01/17  10:41:07  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Declaration_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    This package contains the set of routines that handle the processing
--    of Ada declarations during parsing of input by the Ada Command
--    Language Interpreter.  These routines perform semantic checks,
--    build up symbol table entries associated with entities (objects,
--    types, subprograms, packages, ...), and build statement database nodes
--    associated with the "declaration" of these entities.
--
--    The externally visible routines within the package are invoked by the
--    ACLI parser during the syntactic analysis.  The AdaYACC grammar file
--    is the appropriate base to determine the context in which these
--    routines are invoked.  The actions which these routines perform are
--    based upon the context in they are invoked.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Statement_Database;
use Statement_Database;

with Context_Control;

package Declaration_Routines is


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Object_Decl
   --
   --  <Description>
   --
   --    The object declaration statement database node is built for the
   --    list of identifiers.  The information as to its type, whether its
   --    value is to remain constant, and the initialization code are used
   --    in constructing the object declaration.
   --
   --==========================================================================

   procedure Create_Object_Decl
       (Context          : in out Context_Control.Context_Definition;
        Id_List_Node     : in Ref_Statement_Database;
        Constant_Token   : in Boolean;
        Subtype_Ind_Node : in Ref_Statement_Database;
        Init_Node        : in Ref_Statement_Database;
        Value            : out Ref_Statement_Database);

   -- Notes:
   -- Id_List_Node: always a list of ids (1 or more)
   -- Constant_Token: TRUE if the keyword CONSTANT appeared in the statement
   -- Init_Node: may be null


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Number_Decl
   --
   --  <Description>
   --
   --    An object declaration statement database node is built for the
   --    list of constants.  The information as to its type and the
   --    initialization code are used in constructing the constant declaration.
   --
   --==========================================================================

   procedure Create_Number_Decl
       (Context      : in out Context_Control.Context_Definition;
        Id_List_Node : in Ref_Statement_Database;
        Init_Node    : in Ref_Statement_Database;
        Value        : out Ref_Statement_Database);

   -- Notes:
   -- Id_List_Node: always a list of ids (1 or more)


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Link_Id_Onto_Front_Of_List
   --
   --  <Description>
   --
   --    The identifier is linked onto the front of a list of identifiers.
   --    A list item statement database node is created with the identifier
   --    as the left element and the identifier list as the right element.
   --    The "Holder" field of the statement database list item node contains
   --    a reference to the last element in the list and this invariant is
   --    maintained by this routine.
   --
   --==========================================================================

   function Link_Id_Onto_Front_Of_List
       (Id_Node      : in Ref_Statement_Database;
        Id_List_Node : in Ref_Statement_Database) return Ref_Statement_Database;

   -- Notes:
   -- Id_Node: always a single id
   -- Id_List_Node: may be null, id, or a list of ids


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Link_Id_Onto_Rear_Of_List
   --
   --  <Description>
   --
   --
   --    The identifier is linked onto the rear of a list of identifiers.
   --    A list item statement database node is created with the identifier
   --    as the left element and null as the right element.  The right element
   --    of the last entry on the id list will point to the newly created
   --    list item node.
   --    The "Holder" field of the statement database list item node contains
   --    a reference to the last element in the list and this invariant is
   --    maintained by this routine.
   --
   --==========================================================================

   function Link_Id_Onto_Rear_Of_List (Id_Node      : in Ref_Statement_Database;
                                       Id_List_Node : in Ref_Statement_Database
                                       ) return Ref_Statement_Database;

   -- Notes:
   -- Id_Node: always a single id
   -- Id_List_Node: may be null, id, or a list of ids


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Type_Decl
   --
   --  <Description>
   --
   --    A Statement_Record/Type_Decl is created for the identifier and its
   --    associated type and discriminant information.
   --
   --    The discriminant node is optional.
   --
   --==========================================================================

   function Create_Type_Decl
       (Context           : in Context_Control.Context_Definition;
        Id_Node           : in Ref_Statement_Database;
        Discriminant_Node : in Ref_Statement_Database;
        Type_Defn_Node    : in Ref_Statement_Database) return
         Ref_Statement_Database;

   -- Notes:
   -- Discriminant_Node: may be null


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Subtype_Decl
   --
   --  <Description>
   --
   --    A statement database subtype declaration node is constructed for the
   --    identifier.  The subtype indication information has already been
   --    constructed and is passed in via a subtype indication node.
   --
   --==========================================================================

   function Create_Subtype_Decl
       (Context          : in Context_Control.Context_Definition;
        Id_Node          : in Ref_Statement_Database;
        Subtype_Ind_Node : in Ref_Statement_Database) return
         Ref_Statement_Database;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Build_Subtype_Indication
   --
   --  <Description>
   --
   --    A symbol table entry reference statement database node is constructed
   --    which contains the subtype indication information.  The subtype
   --    indication information is formulated form the type mark and optional
   --    constraint information.
   --
   --==========================================================================

   function Build_Subtype_Indication
       (Type_Mark_Node  : in Ref_Statement_Database;
        Constraint_Node : in Ref_Statement_Database) return
         Ref_Statement_Database;

   -- Notes:
   -- Constraint_Node: may be null
   -- 


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Build_Unconstrained_Array_Decl
   --
   --  <Description>
   --
   --
   --==========================================================================
   function Build_Unconstrained_Array_Decl
       (First_Index : in Ref_Statement_Database;
        List_Of_Indexes  : in Ref_Statement_Database;
        Components  : in Ref_Statement_Database) return
         Ref_Statement_Database;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Build_Unconstrained_Array_Decl
   --
   --  <Description>
   --
   --
   --==========================================================================
   function Build_Constrained_Array_Decl
       (List_Of_Indexes  : in Ref_Statement_Database;
        Components  : in Ref_Statement_Database) return
         Ref_Statement_Database;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Resolve_Type_Mark
   --
   --  <Description>
   --
   --    The symbolic name is resolved to the kind of type mark which it
   --    represents (type, subtype, or derived type).  A symbol table entry
   --    reference statement database node is constructed for the type mark.
   --
   --==========================================================================

   function Resolve_Type_Mark
       (Expanded_Name_Node : in Ref_Statement_Database) return
         Ref_Statement_Database;

   -- Notes:
   -- Expanded_Name_Node: may be an identifier or an expanded name


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Mark_Derived_Type
   --
   --  <Description>
   --
   --    The subtype indication node has its type mark changed to that of
   --    a derived type.
   --
   --==========================================================================

   function Mark_Derived_Type
       (Subtype_Ind_Node : in Ref_Statement_Database) return
         Ref_Statement_Database;


   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Create_Record_Type_Entry
   --
   --    Last Mod:    Fri Aug 18 11:43:07 1989 (wpl)
   --    Author:      C. Oei <wpl@huey>
   --
   --  <Description>
   -- A record type entry is created for the specified record components
   -- list.
   --
   --  --------------------------------------------------------------------
   function Create_Record_Type_Entry
       (Context                     : in Context_Control.Context_Definition;
        Record_Components_List_Node : in Ref_Statement_Database) return
         Ref_Statement_Database;

   -- Notes:
   -- Record_Components_List_Node: may be null

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Link_Record_Component_Entries
   --
   --    Last Mod:    Fri Aug 18 11:43:17 1989
   --    Author:      C. Oei <wpl@huey>
   --
   --  <Description>
   -- The symbol table entries for the components of the record are linked
   -- together.
   --
   --  --------------------------------------------------------------------
   function Link_Record_Component_Entries
       (Component_List_Node : in Ref_Statement_Database;
        Component_Node      : in Ref_Statement_Database) return
         Ref_Statement_Database;

   -- Notes:
   -- Component_List_Node: may be null

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Build_Record_Component
   --
   --    Last Mod:    Fri Aug 18 11:43:51 1989
   --    Author:      C. Oei <wpl@huey>
   --
   --  <Description>
   -- A record component symbol table entry is built for each record component
   -- identifier.
   --
   --  --------------------------------------------------------------------
   function Build_Record_Component
       (Context          : in Context_Control.Context_Definition;
        Id_List_Node     : in Ref_Statement_Database;
        Subtype_Ind_Node : in Ref_Statement_Database;
        Init_Node        : in Ref_Statement_Database) return
         Ref_Statement_Database;

   -- Notes:
   -- Id_List: always a list of ids (1 or more)
   -- Init_Node: may be null


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Build_Integer_Type
   --
   --  <Description>
   --
   --    An integer type definition node is constructed from the range
   --    definition information.
   --
   --==========================================================================

   function Build_Integer_Type
       (Range_Definition_Node : in Ref_Statement_Database) return
         Ref_Statement_Database;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Exception_Declaration
   --
   --  <Description>
   --
   --    A symbol table reference entry statement database node is constructed
   --    for the an exception declaration
   --
   --==========================================================================
   procedure Create_Exception_Declaration
       (Context        : in Context_Control.Context_Definition;
        Exception_Name : in Ref_Statement_Database;
        Value          : out Ref_Statement_Database);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Build_Range_Constraint
   --
   --  <Description>
   --
   --    A symbol table reference entry statement database node is constructed
   --    for the constraint information.
   --
   --==========================================================================

   function Build_Range_Constraint (Constraint_Info : in Ref_Statement_Database
                                    ) return Ref_Statement_Database;


   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Create_Enum_Type_Def_Entry
   --
   --    Last Mod:    Fri Aug 18 11:44:35 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   -- An enumeration type definition symbol table entry is built for the
   -- list of enumeration literals within the enumeration type node.
   --
   --  --------------------------------------------------------------------
   function Create_Enum_Type_Def_Entry
       (Context        : in Context_Control.Context_Definition;
        Enum_Type_Node : in Ref_Statement_Database) return
         Ref_Statement_Database;


   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Create_Enum_Literal_Entry
   --
   --    Last Mod:    Fri Aug 18 11:44:51 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   -- An enumeration literal symbol table entry is built of the enumeration
   -- literal.
   --
   --  --------------------------------------------------------------------
   function Create_Enum_Literal_Entry
       (Context           : in Context_Control.Context_Definition;
        Enum_Literal_Node : in Ref_Statement_Database) return
         Ref_Statement_Database;


   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Link_Enum_Literal_On_Front
   --
   --    Last Mod:    Fri Aug 18 11:45:15 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   -- The first enumeration literal entry is linked onto the front of the
   -- list of enumeration literals.  The pred_node has its successor link set
   -- to the head item in the list_node; the head item in the list node has
   -- its predecessor link set to pred_node.
   --
   --  --------------------------------------------------------------------
   function Link_Enum_Literal_On_Front
       (Pred_Node, List_Node : in Ref_Statement_Database) return
         Ref_Statement_Database;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Link_Enum_Literal_On_End
   --
   --    Last Mod:    Fri Aug 18 11:45:30 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   -- The current enumeration literal entry is linked onto the end of the
   -- list of enumeration literals.  The succ_node has its predecessor link set
   -- to the tail item in the list_node; the tail item in the list node has
   -- its successor link set to succ_node.
   --
   --  --------------------------------------------------------------------
   function Link_Enum_Literal_On_End
       (List_Node, Succ_Node : in Ref_Statement_Database) return
         Ref_Statement_Database;


   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Build_Subprogram_Spec_Decl
   --
   --    Last Mod:    Fri Aug 18 11:45:44 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   -- The subprogram symbol table entry that has been already created is
   -- inserted into a statement to create a subprogram specification.
   --
   --  --------------------------------------------------------------------
   function Build_Subprogram_Spec_Decl
       (Context         : in Context_Control.Context_Definition;
        Subprogram_Node : in Ref_Statement_Database) return
         Ref_Statement_Database;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Create_Subprogram_Spec
   --
   --    Last Mod:    Fri Aug 18 11:46:03 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   -- A subprogram (function or procedure) specification is built for the
   -- identifier/designator.  The specification contains the symbol table
   -- entry for a subprogram.
   --
   --  --------------------------------------------------------------------
   function Create_Subprogram_Spec
       (Context            : in Context_Control.Context_Definition;
        Id_Desig_Node      : in Ref_Statement_Database;
        Formal_Params_Node : in Ref_Statement_Database;
        Type_Mark_Node     : in Ref_Statement_Database) return
         Ref_Statement_Database;

   -- Notes:
   -- Id_Desig_Node: an identifier for a procedure,
   --                a designator for a function
   -- Formal_Params_Node: may be null
   -- Type_Mark_Node: null => procedure, non-null => function

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Build_Formal_Parameter
   --
   --    Last Mod:    Fri Aug 18 11:47:15 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   -- A formal parameter symbol table entry is built for each identifier
   -- in the id list.  The symbol table entries are linked together and
   -- returned as a single statement database node.
   --
   --  --------------------------------------------------------------------
   function Build_Formal_Parameter
       (Id_List_Node    : in Ref_Statement_Database;
        Param_Mode_Node : in Ref_Statement_Database;
        Type_Mark_Node  : in Ref_Statement_Database;
        Init_Node       : in Ref_Statement_Database) return
         Ref_Statement_Database;

   -- Notes:
   -- Id_List_Node: always a list of ids (1 or more)
   -- Param_Mode_Node: indicates default in, explicit in, out, or in out
   -- Init_Node: may be null

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Link_Formal_Param_On_Front
   --
   --    Last Mod:    Fri Aug 18 11:47:43 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   -- A list of formal parameters are created from the parameter and the list
   -- of parameters.  A parameter node contains a parameter specification
   -- which may contain more than one formal parameter identifiers.
   -- The parameter is linked onto the front of the parameter list.
   --
   --  --------------------------------------------------------------------
   function Link_Formal_Param_On_Front
       (Param_Node      : in Ref_Statement_Database;
        Param_List_Node : in Ref_Statement_Database) return
         Ref_Statement_Database;

   -- Notes:
   -- Param_Node: a single parameter specification
   -- Param_List_Node: may be null or a list of parameter specifications

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Link_Formal_Param_On_Rear
   --
   --    Last Mod:    Fri Aug 18 11:48:01 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   -- A list of formal parameters are created from the parameter and the list
   -- of parameters.  A parameter node contains a parameter specification
   -- which may contain more than one formal parameter identifiers.
   -- The parameter is linked onto the rear of the parameter list.
   --
   --  --------------------------------------------------------------------
   function Link_Formal_Param_On_Rear
       (Param_Node      : in Ref_Statement_Database;
        Param_List_Node : in Ref_Statement_Database) return
         Ref_Statement_Database;

   -- Notes:
   -- Param_Node: a single parameter specification
   -- Param_List_Node: may be null or a list of parameter specifications

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Build_Param_Mode
   --
   --    Last Mod:    Fri Aug 18 11:48:29 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   -- The tokens for the mode of a formal parameter (nothing, in, out, or
   -- in out) are converted into a statement database node that can be
   -- passed back on to a subsequent formal parameter symbol table entry
   -- building routine.
   --
   --  --------------------------------------------------------------------
   function Build_Param_Mode (Context : in Context_Control.Context_Definition;
                              First_Token, Second_Token : in Integer :=
      - 1) return Ref_Statement_Database;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Start_Of_Subprogram_Body
   --
   --    Last Mod:    Fri Aug 18 11:48:44 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   -- The validity of the name of the subprogram is checked and data
   -- allocation offsets are assigned to each of the subprogram's parameters.
   --
   --  --------------------------------------------------------------------
   procedure Start_Of_Subprogram_Body
       (Context         : in out Context_Control.Context_Definition;
        Subprogram_Node : in Ref_Statement_Database);

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Build_Subprogram_Body_Decl
   --
   --    Last Mod:    Fri Aug 18 11:49:12 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   -- The subprogram declaration node is built for the subprogram body. All
   -- the relevant information associated with the subprogram body is
   -- inserted into the symbol table entry for the subprogram.
   --
   --  --------------------------------------------------------------------
   procedure Build_Subprogram_Body_Decl
       (Context         : in out Context_Control.Context_Definition;
        Subprogram_Node : in Ref_Statement_Database;
        Decl_Node       : in Ref_Statement_Database;
        Stmts_Node      : in Ref_Statement_Database;
        Exception_Node  : in Ref_Statement_Database;
        End_Label_Node  : in Ref_Statement_Database;
        Tree            : out Ref_Statement_Database);

   -- Notes:
   -- Decl_Node: may be null
   -- Exception_Node: may be null
   -- End_Label_Node: may be null

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Start_Of_Package
   --
   --    Last Mod:    Fri Aug 18 11:49:28 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --   The validity of the name of the package is checked and appropriate
   --   actions are executed.  For a package specification, a symbol table
   --   entry for the package is created.   For a package body, the
   --   specification for the package is located and the symbol table
   --   entries associated with the package spec are pushed onto the
   --   symbol table.
   --
   --  --------------------------------------------------------------------
   procedure Start_Of_Package
       (Context : in out Context_Control.Context_Definition;
        Id_Node : in Ref_Statement_Database;
        Is_Body : in Boolean;
        Tree    : out Ref_Statement_Database);

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Create_Package_Spec_Decl
   --
   --    Last Mod:    Fri Aug 18 11:49:44 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --   The package spec declaration node is built for the package spec.
   --   All the relevant information associated with the package spec
   --   is inserted into the symbol table entry for the package.
   --
   --  --------------------------------------------------------------------
   procedure Create_Package_Spec_Decl
       (Context        : in out Context_Control.Context_Definition;
        Package_Node   : in Ref_Statement_Database;
        Spec_Decl_Node : in Ref_Statement_Database;
        End_Label_Node : in Ref_Statement_Database;
        Tree           : out Ref_Statement_Database);

   -- Notes:
   --   Spec_Decl_Node: may be null
   --   End_Label_Node: may be null

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Create_Package_Body_Decl
   --
   --    Last Mod:    Fri Aug 18 11:50:06 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --   The package body declaration node is built for the package body.
   --   All the relevant information associated with the package body
   --   is inserted into the symbol table entry for the package.
   --
   --  --------------------------------------------------------------------
   procedure Create_Package_Body_Decl
       (Context        : in out Context_Control.Context_Definition;
        Package_Node   : in Ref_Statement_Database;
        Body_Decl_Node : in Ref_Statement_Database;
        Stmts_Node     : in Ref_Statement_Database;
        End_Label_Node : in Ref_Statement_Database;
        Tree           : out Ref_Statement_Database);

   -- Notes:
   --   Body_Decl_Node: may be null
   --   Stmts_Node    : may be null
   --   End_Label_Node: may be null

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Create_Use_Node
   --
   --    Last Mod:    Fri Aug 18 11:50:22 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --   The id names are verified to be packages.  A Use statement is
   --   built.
   --
   --  --------------------------------------------------------------------
   function Create_Use_Node
       (Context      : in Context_Control.Context_Definition;
        Id_List_Node : in Ref_Statement_Database) return Ref_Statement_Database;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_List_Item_Node
   --
   --  <Description>
   --
   --    A list item node is constructed with the two items being the
   --    left and right elements of the list item.
   --   This routine is used to "merge" the sequence of statements following
   --   BEGIN in a package body with the exception handler part of the
   --   package body.  These items are merged into a single statement
   --   database node so that they may be passed up in the Ada grammar to
   --   a higher level rule.
   --
   --==========================================================================

   function Create_List_Item_Node
       (Item1 : in Ref_Statement_Database;
        Item2 : in Ref_Statement_Database) return Ref_Statement_Database;

   -- -------------------------------------------------------------------------

end Declaration_Routines;
