--|$Log:	dump.a,v $
--|Revision 8.0  91/01/17  10:41:10  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Dump_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    This package contains a set of routines that dumps out in textual
--    form Ada code.  Ada code read by ACE is parsed and translated into
--    an internal representation.  The internal representation consists of
--    statement database nodes and symbol table entries.  The dump routines
--    traverse the statement database nodes and symbol table entries and
--    reconstruct the Ada text from the ACE internal data structures.
--
--    These dump routines also support a set of controls over the form in
--    which the text is created.  This includes whether statement numbers
--    are to be displayed, whether statements should be dumped for tracing
--    purposes, whether the bodies associated with subprograms and packages
--    should be displayed, ...
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Text_Io;
use Text_Io;

with Literals;
use Literals;

with Statement_Database;
use Statement_Database;

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Context_Control;

package Dump_Routines is

   type List_Mode is (Both, Specification, Implementation);

   type Obj_Declarations is
      (Any,
       Object_Declaration,
       Type_Mark,
       Type_Declaration,
       Subtype_Declaration,
       Subprogram_Declaration,
       Package_Declaration,
       Procedure_Declaration,
       Function_Declaration,
       Label_Declaration);

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Dump
   --
   --  <Description>
   --
   --    This procedure sets whether statement dumping is to be effect.
   --    The Switch parameter controls whether statement dumping is to be
   --    on or off.
   --
   --==========================================================================

   procedure Dump (Switch : Boolean);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Dump_To_Text_Windows
   --
   --  <Description>
   --
   --    This procedure sets whether statement dumping is being performed to
   --    a text window or not.  The method used to indent statements is
   --    affected by whether dumping is being performed to a text window or
   --    not.
   --
   --==========================================================================

   procedure Dump_To_Text_Windows (Switch : Boolean);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Dump_Statement_Numbers
   --
   --  <Description>
   --
   --    This procedure controls whether statement numbers associated with the
   --    Ada statements should be dumped when the statement is displayed.
   --
   --==========================================================================

   procedure Dump_Statement_Numbers (Switch : Boolean);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Statement_Dump_On
   --
   --  <Description>
   --
   --    This function returns a boolean indication of whether statement
   --    dumping is turned on.
   --
   --==========================================================================

   function Is_Statement_Dump_On return Boolean;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Dump_Statement_Database
   --
   --  <Description>
   --
   --    This procedure reconstructs the Ada text associated with an
   --    ACE statement database node.  The text is written to the specified
   --    File, and the display of bodies for packages and subprograms is
   --    controlled by the Dumping_Mode parameter.
   --
   --==========================================================================

   procedure Dump_Statement_Database (Ref          : Ref_Statement_Database;
                                      File         : File_Type;
                                      Dumping_Mode : List_Mode := Specification
                                      );

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Dump_For_Statement_Trace
   --
   --  <Description>
   --
   --    This procedure reconstructs the Ada text associated with a
   --    a single Ada statement from an ACE statement database node for
   --    the purpose of tracing Ada statements as they are executed.
   --    Since the statement is being displayed to support a trace,
   --    the complete text of a compound statement is not displayed when
   --    first encountered.  For example, only the loop scheme will be
   --    displayed when a loop statement is passed in.  The body of the
   --    loop would be displayed on a one-by-one basis as each statement
   --    within the loop is executed.  The text is written to the specified
   --    File.
   --
   --==========================================================================

   procedure Dump_For_Statement_Trace (Ref  : Ref_Statement_Database;
                                       File : File_Type);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Dump_Symtbl
   --
   --  <Description>
   --
   --    This procedure reconstructs the Ada text associated with an
   --    ACE symbol table entry.  The text is written to the specified File,
   --    and the display of bodies for packages and subprograms is controlled
   --    by the Dumping_Mode parameter.
   --
   --==========================================================================

   procedure Dump_Symtbl (Ref          : Ref_Symbol_Table_Entry;
                          File         : File_Type;
                          Dumping_Mode : List_Mode := Specification;
                          Dump_Help    : Boolean   := True);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Dump_Entire_Symtbl
   --
   --  <Description>
   --
   --    This procedure displays each and every entry in the entire ACE
   --    symbol table.  Each symbol table entry is reconstructed into the
   --    Ada code that caused the symbol table entry to be generated.
   --
   --==========================================================================

   procedure Dump_Entire_Symtbl (Context   : Context_Control.Context_Definition;
                                 Dump_Mode : in List_Mode := Specification);

   -- -------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Dump_Help_Info
   --
   --  <Description>
   --
   --   This procedure outputs the help comments associated with a symbol
   --   table entry.  The STRIP_HYPH parameter will control whether the help
   --   comment delimiter is pre-pended to the help information.
   --
   --==========================================================================

   procedure Dump_Help_Info (Ref           : in Ref_Symbol_Table_Entry;
                             Strip_Comment : in Boolean;
                             File          : in File_Type);

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Dump_Enclosing_Pkgs
   --
   -- <Description>
   --
   -- This procedure dumps the names of the enclosing packages for the
   -- given symbol table entry.  The STRIP_HYPH parameter will control
   -- whether the help comment delimiter is pre-pended to the help information.
   --==========================================================================

   procedure Dump_Enclosing_Pkgs (Ref  : in Ref_Symbol_Table_Entry;
                                  File : in File_Type);

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Dump_Objs
   --
   --  <Description>
   --
   --    Dump_Objs will list the names of visible objects on the symbol
   --  table.
   --
   --==========================================================================
   procedure Dump_Objs (Context   : Context_Control.Context_Definition;
                        Obj_Type  : in Obj_Declarations;
                        Specifier : in Ref_String;
                        File      : in File_Type);

end Dump_Routines;
