--|$Log:	sdb.a,v $
--|Revision 8.0  91/01/17  10:41:22  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Statement_Database
--
---------------------------------------------------------------------------
--
-- Abstract
--
--   These structures comprise the statement database where Ada
--   statements are stored for access and manipulation by the various
--   components of the ACLI.
--   The statement database essentially stores two types of nodes:
--   statements (which include Ada statements, and declarations),
--   and expressions.
--   The statement evaluator walks statement and declaration nodes.
--   Enumeration types named "Unknown"  (and hence nodes constructed of
--   type "Unknown") are used solely by the tree builder as temporary
--   storage.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------
with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Storage_Management;
use Storage_Management;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Literals;
use Literals;

with Help_Pkg;
use Help_Pkg;

package Statement_Database is

   -- -----
   -- Enumerates the two types of nodes within the statement database.
   -- -----
   type Statement_Database_Enumeration is
      (Statement_Node_Type,
       Expression_Node_Type,
       Unknown);

   -- -----
   -- Enumerates the types of statement nodes appropriate for representing
   -- Ada statements and declarations (subsequent to semantic analysis).
   -- -----
   type Statement_Enumeration is
      (Object_Declaration,
       Type_Declaration,
       Subtype_Declaration,
       Exception_Declaration,
       Procedure_Spec_Declaration,
       Procedure_Declaration,
       Function_Spec_Declaration,
       Function_Declaration,
       Package_Declaration,
       Package_Body_Declaration,
       Assign_Statement,
       Loop_Statement,
       For_Statement,
       If_Statement,
       Case_Statement,
       Case_Alternative_Part,
       Exception_Alternative,
       Labeled_Statement,
       Goto_Statement,
       Exit_Statement,
       Return_Statement,
       Block_Statement,
       With_Statement,
       Use_Statement,
       Procedure_Call,
       Pragma_Statement,
       Null_Statement,
       Unknown);

   -- -----
   -- Enumerates the types of expression nodes appropriate for representing
   -- Ada expressions.
   -- -----
   type Expression_Enumeration is
      (Dot_Operator,
       Left_Parenthesis_Operator,
       List_Item,
       Aggregate_Component,
       Range_Expression,
       Attribute_Tick,
       Terminal_Node,
       Constant_Node,
       String_Node,
       Character_Node,
       Array_Reference,
       Function_Call,
       Type_Conversion,
       Record_Reference,
       Dot_Notation_Variable_Reference,
       Unknown);

   -- -----
   -- Enumerates the two types of ACE Ada expressions that can contain
   -- a "dot" operator.
   -- Access types are not being implemented in the initial contract.
   -- -----
   type Dot_Node_Enumeration is
      (Record_Reference,
       Dot_Notation_Variable_Reference,
       Unknown);

   -- -----
   -- Enumerates the three types of ACE Ada expressions that can contain
   -- a "left parenthesis" operator.
   -- -----
   type Lp_Node_Enumeration is
      (Array_Reference,
       Function_Call,
       Type_Conversion,
       Unknown);

   -- -----
   -- Data structure to hold Comments
   --
   type Comments;

   type Ref_Comment is access Comments;

   type String_Ptr is access String;

   type Comments is
      record
         Comment : String_Ptr;
         Next    : Ref_Comment;
      end record;
   --
   -- -----

   -- -----
   -- Forward reference to the statement database node and
   -- corresponding visible type that references the statement database node.
   -- -----
   type Statement_Database_Record (Node_Kind : Statement_Database_Enumeration);
   type Ref_Statement_Database is access Statement_Database_Record;

   -- --------------------------------------------------------------------------

   -- -----
   -- The structure of a statement node used for representing
   -- Ada statements and declarations (subsequent to semantic analysis).
   -- -----
   type Statement_Record (Statement_Kind : Statement_Enumeration) is
      record

         Next_Statement : Ref_Statement_Database;
         Holder : Ref_Statement_Database; -- for treebuilder
         -- used for internal reference and as a point of reference
         -- during interactive debugging
         Statement_Number                     : Ace_Natural;

         Suspend_Evaluation_At_This_Statement : Boolean := False;

         case Statement_Kind is

            when Object_Declaration =>
               Name_List       : Ref_Statement_Database;
               -- list of symtbl entries
               Init_Expression : Ref_Statement_Database; -- expr

            when Exception_Declaration   =>

               Exception_Name_List : Ref_Statement_Database;

            when Type_Declaration           |
                 Subtype_Declaration        |
                 Procedure_Spec_Declaration |
                 Procedure_Declaration      |
                 Function_Spec_Declaration  |
                 Function_Declaration       |
                 Package_Declaration        |
                 Package_Body_Declaration     =>

               Symbol_Table_Entry : Ref_Symbol_Table_Entry; -- id

            when Assign_Statement   =>
               Lhs : Ref_Statement_Database; -- expr
               Rhs : Ref_Statement_Database; -- expr

            when Loop_Statement     =>
               Loop_Name   : Ref_Statement_Database; -- Needed for exits
               Loop_Scheme : Ref_Statement_Database; -- while, for, null.
               Loop_Body   : Ref_Statement_Database; -- stmt

            when For_Statement =>
               For_Variable : Ref_Statement_Database; -- terminal symbol
               For_Range    : Ref_Statement_Database; -- expr

            when If_Statement  =>
               If_Condition : Ref_Statement_Database; -- expr
               True_Task    : Ref_Statement_Database; -- stmt
               False_Task   : Ref_Statement_Database; -- stmt

            when Case_Statement     =>
               Case_Expression        : Ref_Statement_Database; -- expr
               Case_Alternatives_List : Ref_Statement_Database; -- "stmts"

            when Case_Alternative_Part   =>
               Case_Alternative_Choice     : Ref_Statement_Database; -- expr
               Case_Alternative_Statements : Ref_Statement_Database; -- stmts
               Next_Case_Alternative       : Ref_Statement_Database; -- "stmts"

            when Exception_Alternative   =>
               Exception_Alternative_Choice     : Ref_Statement_Database;
               -- expr
               Exception_Alternative_Statements : Ref_Statement_Database;
               -- stmts
               Next_Exception_Alternative       : Ref_Statement_Database;
               -- "stmts"

            when Labeled_Statement  =>
               Label_Entry : Ref_Statement_Database; -- id (label name)

            when Goto_Statement     =>
               Goto_Label : Ref_Statement_Database; -- id (goto label)

            when Exit_Statement     =>
               Exit_Loop_Name : Ref_Statement_Database; -- id (loop name)
               Exit_Condition : Ref_Statement_Database; -- expr

            when Return_Statement   =>
               Return_Expression    : Ref_Statement_Database; -- expr
               Subprogram_Reference : Ref_Symbol_Table_Entry;
               -- subprogram id which "owns" this return stmt

            when Block_Statement    =>
               Block_Name         : Ref_Statement_Database; -- terminal symbol
               Block_Declarations : Ref_Statement_Database; -- decls
               Block_Statements   : Ref_Statement_Database; -- stmts
               Block_Exceptions   : Ref_Statement_Database; -- exceptions

            when With_Statement     | Use_Statement     =>
               Unit_Name_List : Ref_Statement_Database; -- id (unit name(s))

            when Procedure_Call     =>
               -- procedure name, formal parameters, location of stmts, etc.
               Procedure_Info                   : Ref_Statement_Database; -- id

               -- exprs -- list of actual parameters
               Actual_Procedure_Parameters_List : Ref_Statement_Database;

            when Pragma_Statement   =>
               Pragma_Name : Ref_Statement_Database; -- id
               Pragma_Arg  : Ref_Statement_Database; -- list

            when Null_Statement     | Unknown =>
               null;

         end case;
      end record;

   -- --------------------------------------------------------------------------

   -- -----
   -- The structure of a expression node used for representing
   -- Ada expressions.
   -- Used by the tree builder and semantic analyzer,
   -- and subsequently interpreted by the statement evaluator.
   -- -----
   type Expression_Record (Expression_Kind : Expression_Enumeration) is
      record

         Node_Value             : Ref_Data_Storage; -- for stmt evaluator
         Node_Type              : Ref_Symbol_Table_Entry; -- for type checker
         Is_Node_Type_Universal : Boolean; -- for type checker
         Next_Expression        : Ref_Statement_Database; -- for tree builder
         -- used for default parameter initialization code
         Is_Default_Expression  : Boolean      := False;
         Reference_Count        : Ace_Positive := Ace_Positive'FIRST;

         case Expression_Kind is

            when Dot_Operator  =>

               -- for record references, dot notation variable references

               Dot_Node_Kind        : Dot_Node_Enumeration;

               Dot_Object           : Ref_Statement_Database; -- id|expr
               Dot_Component        : Ref_Statement_Database; -- id|expr

               -- expanded name symtbl entry
               Expn_Terminal_Symbol : Ref_Symbol_Table_Entry := null;
               Expn_Terminal_Sdb    : Ref_Statement_Database;

            when Left_Parenthesis_Operator    =>

               -- for arrays, function calls, type conversions

               Lp_Node_Kind  : Lp_Node_Enumeration;

               -- id|expr --
               --    arrays: array structure information (name, etc.) -- id|expr
               --    function: function information (name, etc.) -- id
               --        [used for std binary & unary expressions]
               --    type conv: type name -- id
               Lp_Object     : Ref_Statement_Database;

               -- exprs: list of item(s), actual parameters
               Lp_Items_List : Ref_Statement_Database;

            when List_Item     =>

               -- used for array indices, subprogram parameters,
               --    enumeration definitions

               Item_Expression : Ref_Statement_Database; -- expr
               Next_List_Item  : Ref_Statement_Database; -- next item

            when Aggregate_Component     =>
               Aggregate_Choice     : Ref_Statement_Database; -- expr
               Aggregate_Expression : Ref_Statement_Database; -- expr

            when Range_Expression   =>
               Range_Name  : Ref_Statement_Database; -- expr
               Lower_Bound : Ref_Statement_Database; -- expr
               Upper_Bound : Ref_Statement_Database; -- expr

            when Attribute_Tick     =>
               Attribute_Prefix : Ref_Statement_Database; -- id
               Attribute        : Ref_Statement_Database;

            when String_Node   =>
               -- for string literals
               String_Symbol : Ref_String;

            when Character_Node     =>
               -- for character literals
               Character_Symbol : Character;

            when Constant_Node =>
               -- for constants
               Constant_Symbol : Ref_String;

            when Terminal_Node =>
               -- for identifiers, etc.
               Symbol_Name     : Ref_String;
               Terminal_Symbol : Ref_Symbol_Table_Entry;

               -- these next 3 are placeholders for "(" operator
            when Array_Reference                 |
                 Function_Call                   |
                 Type_Conversion                 |
            -- these next 2 are placeholders for "." operator
                 Record_Reference                |
                 Dot_Notation_Variable_Reference |
                 Unknown  =>
               null;

         end case;
      end record;

   -- --------------------------------------------------------------------------

   -- -----
   -- References to the two types of nodes of the statement database.
   -- -----
   type Ref_Statement_Record is access Statement_Record;
   type Ref_Expression_Record is access Expression_Record;

   -- -----
   -- The structure of a the main statement database node.
   -- -----
   type Statement_Database_Record (Node_Kind : Statement_Database_Enumeration)
                                  is
      record

         Parent_Node    : Ref_Statement_Database;

         -- Comment Holder is used to store comments in the statement
         -- database
         Comment_Holder : Ref_Comment;

         case Node_Kind is
            when Statement_Node_Type     =>
               Statement_Node : Ref_Statement_Record;
            when Expression_Node_Type    =>
               Expression_Node : Ref_Expression_Record;
            when Unknown  =>
               null;
         end case;
      end record;

   -- -----
   -- array of statement db nodes (w/ stmt #'s) for debugger breaking
   Stmt_List_Range_Limit : constant Ace_Natural := 5000;
   --- This representation must change?  Or we must reclaim
   --- statements (i.e., garbage collection).
   subtype Stmt_List_Range is
      Ace_Natural range Ace_Natural'FIRST .. Stmt_List_Range_Limit;
   Stmt_List             : array (Stmt_List_Range) of Ref_Statement_Database;
   --
   -- -----

end Statement_Database;
