--|$Log:	stget.a,v $
--|Revision 8.0  91/01/17  10:41:27  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Symtbl_Entry_Get_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    This package contains the set of routines that retrieves fields of symbol
--    table entries.  All retrievals of components from symbol table entries
--    should be made through the routines provided within this package.
--    This package (in conjunction with a similar package for storing into
--    symbol table entry components) provides the mechanism for hiding the
--    makeup and names of symbol table entry components.  Changes to the
--    structure and component names for symbol table entries would only
--    affect this package of routines, rather than all the other components
--    that deal with symbol table information.
--
--    The routines within the package are ordered with respect to the
--    definition of components for a symbol table entry.  The routines that
--    retrieve components that are common to all symbol table entries come
--    first, followed by components that are applicable to different variants
--    of the symbol table entry.
--
--    Due to the Ada requirement that all record component names be distinct
--    (even those within different variants), a single routine is used to
--    retrieve a similar component that is named differently within several
--    variants.  This routine is defined within the section associated with
--    the first variant that contains the particular data item.  After each
--    routine is listed the record component associated with the routine.
--    Groupings of routines are introduced by the variant which contains
--    the components.
--
--    Components whose values are that of an enumerated type "hide" the
--    enumeration literals from external packages.  Instead of acquiring
--    the value of the enumerated type, a set of routines are provided that
--    have incorporated the enumeration literal into the routine name and
--    return a boolean indication (e.g. Is_Mode_In, Is_Mode_Out,
--    Is_Mode_In_Out, ... [which return a boolean] rather than Get_Mode which
--    would return the mode enumeration type.)
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Statement_Database;
use Statement_Database;

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Help_Pkg;
use Help_Pkg;

with Storage_Management;
use Storage_Management;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Literals;
use Literals;

package Symtbl_Entry_Get_Routines is

   -- --------------------------------------------- --
   -- Components Common to All Symbol Table Entries --
   -- --------------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Entry_Kind
   --
   --  <Description>
   --
   --    Return the kind of entry associated with the symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Entry_Kind
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Type_Entry_Enumeration;
   -- record component name: Entry_Kind

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Next_Entry
   --
   --  <Description>
   --
   --    Return the next entry that is linked with to this symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Next_Entry
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Next_Entry

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Symbolic_Name
   --
   --  <Description>
   --
   --    Return the symbolic name associated with the symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Symbolic_Name
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_String;
   -- record component name: Symbol_Name

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Level
   --
   --  <Description>
   --
   --    Return the static level associated with the symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Level
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural;
   -- record component name: Item_Level

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Enclosing_Package
   --
   --  <Description>
   --
   --    Return the enclosing package symbol table entry associated with
   --    this symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Enclosing_Package
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Enclosing_Package

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Visible
   --
   --  <Description>
   --
   --    Return whether the item is visible
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Visible
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Visibility

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Hidden
   --
   --  <Description>
   --
   --    Return whether the item is hidden
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Hidden (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Visibility

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Removed
   --
   --  <Description>
   --
   --    Return whether the item has been "removed" from the symbol table
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Removed
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Visibility

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Statement_Number
   --
   --  <Description>
   --
   --    Return the statement number associated with the symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Statement_Number
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural;
   -- record component name: Statement_Number

   -- -------------------------------------------------------------------------

   -- -------------------------------------- --
   -- Components for Variant : Package_Entry --
   -- -------------------------------------- --
   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Context_Clause_Body
   --
   --  <Description>
   --
   --    Return the context clause associated with a package body
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Context_Clause_Body
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database;
   -- record component name: Context_Clause

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Context_Clause
   --
   --  <Description>
   --
   --    Return the context clause associated with a compilation unit
   --    specification
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Context_Clause
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database;
   -- record component name: Context_Clause

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Package_Spec_Decls
   --
   --  <Description>
   --
   --    Return the declarations associated with the package specification
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Package_Spec_Decls
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database;
   -- record component name: Package_Spec_Decls

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Package_Body_Decls
   --
   --  <Description>
   --
   --    Return the declarations associated with the package body
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Package_Body_Decls
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database;
   -- record component name: Package_Body_Decls


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Has_A_Body
   --
   --  <Description>
   --
   --    Return an indication of whether the package has a body defined
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Has_A_Body
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Has_Body

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Head_Of_Statements_List
   --
   --  <Description>
   --
   --    Return the head of the list of statements for the executable
   --    statements associated with a package symbol table entry, a
   --    subprogram symbol table entry, or a label symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Head_Of_Statements_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database;
   -- record component name: Package_Stmts
   -- record component name: Statements_List
   --   (for variant: Procedure_Entry, Function_Entry)
   -- record component name: Label_Stmt
   --   (for variant: Label_Stmt)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Derived_Subprogram_List
   --
   --  <Description>
   --
   --    Return the head of the list of derived subprograms that are
   --    associated with a package.  Also, for types and derived types,
   --    return the head of the list of subprograms which are derivable
   --    from this type (or derived type).
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Derived_Subprogram_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return
         Ref_List_Of_Subprograms;
   -- record component name: Pkg_Derived_Subprogram_List
   -- record component name: Type_Derived_Subprogram_List
   --   (for variant: Type_Entry)
   -- record component name: DType_Derived_Subprogram_List
   --   (for variant: Derived_Type_Entry)

   -- -------------------------------------------------------------------------

   -- ---------------------------------------- --
   -- Components for Variant : Procedure_Entry --
   --                        : Function_Entry  --
   -- ---------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Parameter_Count
   --
   --  <Description>
   --
   --    Return the number of parameters associated with a subprogram
   --    symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Parameter_Count
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural;
   -- record component name: Parameter_Count

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Has_Default_Parameters
   --
   --  <Description>
   --
   --    Return the number of parameters associated with a subprogram
   --    symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Has_Default_Parameters
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Default_Parameter_Count

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Number_Of_Default_Parameters
   --
   --  <Description>
   --
   --    Return the number of parameters associated with a subprogram
   --    symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Number_Of_Default_Parameters
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural;
   -- record component name: Default_Parameter_Count

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Head_Of_Parameters_List
   --
   --  <Description>
   --
   --    Return the first parameter's symbol table entry associated with a
   --    subprogram symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Head_Of_Parameters_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Head_Param_List

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Tail_Of_Parameters_List
   --
   --  <Description>
   --
   --    Return the last parameter's symbol table entry associated with a
   --    subprogram symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Tail_Of_Parameters_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Tail_Param_List

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Return_Type
   --
   --  <Description>
   --
   --    Return the return value's symbol table entry associated with a
   --    subprogram (only appropriate for functions) symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Return_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Return_Type

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Head_Of_Declarations_List
   --
   --  <Description>
   --
   --    Return the head of the list of statements for the declaration
   --    statements associated with a subprogram symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Head_Of_Declarations_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database;
   -- record component name: Declarations_List

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Value_Offset
   --
   --  <Description>
   --
   --    Return the reference to the data storage location in which
   --    the value associated with the item (object, function return
   --    or parameter) is stored
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Value_Offset
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural;
   -- record component name: Function_Value_Offset
   -- record component name: Object_Value_Offset
   --   (for variant: Object_Entry)
   -- record component name: Param_Value_Offset
   --   (for variant: Formal_Param_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Subprogram_Built_In
   --
   --  <Description>
   --
   --    Return whether the subprogram is a built-in routine within ACE
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Subprogram_Built_In
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Is_Built_In_Subprogram

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Built_In_Instance_Number
   --
   --  <Description>
   --
   --    Return the instance # used for evaluating built-ins within ACE
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Built_In_Instance_Number
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural;
   -- record component name: Is_Built_In_Subprogram

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Derived_Subprogram_Parent
   --
   --  <Description>
   --
   --    Return the reference to the parent of a derived subprogram.
   --    (A non-derived subprogram has a null parent.)
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Derived_Subprogram_Parent
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Derived_Subprogram_Parent

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Function_Entry
   --
   --  <Description>
   --
   --    Return the reference to the parameterless function call symtbl entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Function_Entry
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Func_Entry

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Enumeration_Literal_Entry
   --
   --  <Description>
   --
   --    Return the reference to the enumeration literal entry if there is one
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Enumeration_Literal_Entry
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Enum_Lit_Entry

   -- -------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Type_Definition
   --
   --  <Description>
   --
   --    Return the Basic type of a type, sub-type, or derived-type
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Type_Definition
       (Node : Ref_Symbol_Table_Entry) return Basic_Type_Definitions;
   -- -------------------------------------------------------------------------
   -- ------------------------------------- --
   -- Components for Variant : Object_Entry --
   -- ------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Type
   --
   --  <Description>
   --
   --    Return the reference to the type symbol table entry associated with
   --    either an object, a parameter, an enumeration literal, or a record
   --    component
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Object_Type
   -- record component name: Param_Type
   --   (for variant: Formal_Param_Entry)
   -- record component name: Enumeration_Type
   --   (for variant: Enumeration_Literal_Entry)
   -- record component name: Component_Type
   --   (for variant: Record_Component_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Has_Constraint
   --
   --  <Description>
   --
   --    Return whether the object, subtype, or the derived type has a
   --    constraint definition associated with it
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Has_Constraint
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Object_Constraint
   -- record component name: Constraint_Info
   --   (for variant: Subtype_Entry, Derived_Type_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Mode_Default_In
   --
   --  <Description>
   --
   --    Return whether the mode for the item (object or parameter) is
   --    IN by default
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Mode_Default_In
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Object_Mode
   -- record component name: Param_Mode
   --   (for variant: Formal_Param_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Mode_Explicit_In
   --
   --  <Description>
   --
   --    Return whether the mode for the item (object or parameter) is
   --    IN (Mode_In), explicitly stated by user
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Mode_Explicit_In
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Object_Mode
   -- record component name: Param_Mode
   --   (for variant: Formal_Param_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Mode_In
   --
   --  <Description>
   --
   --    Return whether the mode for the item (object or parameter) is IN,
   --    either explicitly (In_Mode) or implicitly (Default_In_Mode)
   --
   --    For type checking, the return value of a function and an
   --    enumeration literal are "read only" items, and their mode can be
   --    considered as IN.  This routine returns True for function and
   --    enumeration literal entries.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Mode_In
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Object_Mode
   -- record component name: Param_Mode
   --   (for variant: Formal_Param_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Mode_Out
   --
   --  <Description>
   --
   --    Return whether the mode for the item (object or parameter) is OUT
   --
   --    For type checking, the return value of a function and an
   --    enumeration literal are "read only" items, and their mode can be
   --    considered as IN.  This routine returns False for function and
   --    enumeration literal entries.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Mode_Out
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Object_Mode
   -- record component name: Param_Mode
   --   (for variant: Formal_Param_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Mode_In_Out
   --
   --  <Description>
   --
   --    Return whether the mode for the item (object or parameter) is IN OUT
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Mode_In_Out
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Object_Mode
   -- record component name: Param_Mode
   --   (for variant: Formal_Param_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Are_Modes_Equal
   --
   --  <Description>
   --
   --    Return whether the modes for the two items (object or parameter)
   --    are identical
   --
   --    This handles any combination of objects and parameters.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Are_Modes_Equal
       (Symtbl_Entry_One : in Ref_Symbol_Table_Entry;
        Symtbl_Entry_Two : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Object_Mode
   -- record component name: Param_Mode
   --   (for variant: Formal_Param_Entry)

   -- -------------------------------------------------------------------------

   -- ------------------------------------------- --
   -- Components for Variant : Formal_Param_Entry --
   -- ------------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Initialization_Code
   --
   --  <Description>
   --
   --    Return the statement database node associated with the
   --    initialization code for the formal parameter or a record component
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Initialization_Code
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database;
   -- record component name: Param_Init
   -- record component name: Component_Init
   --   (for variant: Record_Component_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Next_Parameter
   --
   --  <Description>
   --
   --    Return the symbol table entry for the next parameter following this
   --    parameter in the specification of the subprogram
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Next_Parameter
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Next_Param

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Preceded_By_Comma
   --
   --  <Description>
   --
   --    Return whether the formal parameter was preceded by a comma
   --    in the textual definition of the subprogram declaration
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Preceded_By_Comma
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Preceded_By_Comma

   -- -------------------------------------------------------------------------

   -- ----------------------------------- --
   -- Components for Variant : Type_Entry --
   -- ----------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_State_Incomplete
   --
   --  <Description>
   --
   --    Return whether the state of the this type identifier is an
   --    incomplete type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Type_State_Incomplete
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Concreteness

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_State_Full
   --
   --  <Description>
   --
   --    Return whether the state of the this type identifier is a
   --    full type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Type_State_Full
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Concreteness

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_State_Private
   --
   --  <Description>
   --
   --    Return whether the state of the this type identifier is a
   --    private type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Type_State_Private
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Concreteness

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_State_Limited_Private
   --
   --  <Description>
   --
   --    Return whether the state of the this type identifier is a
   --    limited private type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Type_State_Limited_Private
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Concreteness

   -- -------------------------------------------------------------------------

   -- ------------------------------------------- --
   -- Components for Variant : Subtype_Entry      --
   --                        : Derived_Type_Entry --
   -- ------------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Entry_Base_Type
   --
   --  <Description>
   --
   --    Return the base type definition associated with the subtype or the
   --    derived type symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Entry_Base_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Base_Type

   -- -------------------------------------------------------------------------

   -- -------------------------------------------------- --
   -- Components for Variant : Enumeration_Literal_Entry --
   -- -------------------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Literal_Value
   --
   --  <Description>
   --
   --    Return the internal numeric value assigned to the enumeration literal
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Literal_Value
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural;
   -- record component name: Literal_Value

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Successor
   --
   --  <Description>
   --
   --    Return the symbol table entry for the successor to this
   --    enumeration literal
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Successor
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Succ_Enum_Literal

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Predecessor
   --
   --  <Description>
   --
   --    Return the symbol table entry for the predecessor to this
   --    enumeration literal
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Predecessor
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Pred_Enum_Literal

   -- -------------------------------------------------------------------------

   -- ----------------------------------------------- --
   -- Components for Variant : Record_Component_Entry --
   -- ----------------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Next_Record_Component
   --
   --  <Description>
   --
   --    Return the symbol table entry for the next record component following
   --    this record component
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Next_Record_Component (Symtbl_Entry : in Ref_Symbol_Table_Entry
                                       ) return Ref_Symbol_Table_Entry;
   -- record component name: Next_Record_Component

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Simple_Name_Construct
   --
   --  <Description>
   --
   --    Get the construct associated with the simple name.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Simple_Name_Construct (Symtbl_Entry : in Ref_Symbol_Table_Entry
                                       ) return Ref_Statement_Database;
   -- Simple_Name_Construct


   -- -------------------------------------------------------------------------

   -- ------
   -- A Type_Entry variant symbol table entry contains a reference to a
   -- Type_Definition_Node.  The Type_Definition_Node contains informaton
   -- associated with particular types, with variants supporting the
   -- different types.  Access to the information in the Type_Definition_Node
   -- is made through the following set of routines.
   -- ------

   -- ------------------------------- --
   -- Record ==> Type_Definition_Node --
   --   Component for ALL Variants    --
   -- ------------------------------- --


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_An_Enumeration_Type
   --
   --  <Description>
   --
   --    Returns whether the type entry defines an enumeration type
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Type_An_Enumeration_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Type_Kind

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_An_Integer_Type
   --
   --  <Description>
   --
   --    Returns whether the type entry defines an integer type
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Type_An_Integer_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Type_Kind

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_A_Real_Type
   --
   --  <Description>
   --
   --    Returns whether the type entry defines a real type
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Type_A_Real_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Type_Kind

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_A_Record_Type
   --
   --  <Description>
   --
   --   Returns whether the type entry defines a record type
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Type_A_Record_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Type_Kind

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_An_Array_Type
   --
   --  <Description>
   --
   --    Returns whether the type entry defines an array type
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Type_An_Array_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Type_Kind

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_An_Access_Type
   --
   --  <Description>
   --
   --    Returns whether the type entry defines an access type
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Type_An_Access_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Type_Kind


   -- ----------------------------------------- --
   --    Record ==> Type_Definition_Node        --
   -- Components for Variant : Enumeration_Type --
   -- ----------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Literal_Count
   --
   --  <Description>
   --
   --    Return the number of enumeration literals associated with an
   --    enumeration type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Literal_Count
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural;
   -- record component name: Literal_Count

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_First_Literal
   --
   --  <Description>
   --
   --    Return the symbol table entry for the first enumeration literal
   --    associated with an enumeration type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_First_Literal
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: First_Literal

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Last_Literal
   --
   --  <Description>
   --
   --    Return the symbol table entry for the last enumeration literal
   --    associated with an enumeration type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Last_Literal
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Last_Literal



   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Lower_Bound
   --
   --  <Description>
   --
   --    Return the lower bound associated with an integer type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Lower_Bound
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Integer;
   -- record component name: Lower_Bound

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Upper_Bound
   --
   --  <Description>
   --
   --    Return the upper bound associated with an integer type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Upper_Bound
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Integer;
   -- record component name: Upper_Bound

   -- ------------------------------------ --
   --    Record ==> Type_Definition_Node   --
   -- Components for Variant : Record_Type --
   -- ------------------------------------ --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Record_Component
   --
   --  <Description>
   --
   --    Return the first record component associated with a record type
   --    definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Record_Component
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Record_Component

   -- ----------------------------------- --
   --    Record ==> Type_Definition_Node  --
   -- Components for Variant : Array_Type --
   -- ----------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Constrained
   --
   --  <Description>
   --
   --    Return whether the array is a constrained array for an array type
   --    definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Is_Constrained
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;
   -- record component name: Is_Constrained

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Index_Type
   --
   --  <Description>
   --
   --    Return the symbol table entry for the index associated with the
   --    array type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Index_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database;
   -- record component name: Array_Index_Type

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Element_Type
   --
   --  <Description>
   --
   --    Return the symbol table entry for the element associated with the
   --    array type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Element_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database;
   -- record component name: Array_Element_Type

   -- -------------------------------------------------------------------------

   -- ------
   -- A Constraint_Entry variant symbol table entry contains a reference to a
   -- Constraint_Node.  (Also, the Object_Entry variant symbol table entry
   -- contains a component Object_Constraint and the Subtype_Entry &
   -- Derived_Type_Entry contains a component Constraint_Info.  Both of these
   -- are references to a Constraint_Entry symbol table entry.  The following
   -- routines support references with the Constraint_Entry, Object_Entry,
   -- Subtype_Entry, or the Derived_Type_Entry passed in as the starting
   -- symbol table entry.)  The Constraint_Node contains information
   -- associated with constraints, with variants supporting the different kinds
   -- of constraints.  Access to the information in the Constraint_Node is made
   -- through the following set of routines.
   -- ------

   -- ----------------------------------------- --
   --    Record ==> Constraint_Node             --
   -- Components for Variant : Range_Constraint --
   --                        : Index_Constraint --
   -- ----------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Lower_Bound
   --
   --  <Description>
   --
   --    Return the lower bound associated with an array type definition,
   --    or a constraint definition
   --
   --    Note: a constraint definition may exist for an object, a subtype, a
   --    derived type, or a constraint being built during tree building.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Lower_Bound
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database;
   -- record component name:  Lower_Bound

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Lower_Bound_Value_Offset
   --
   --  <Description>
   --
   --    Return the storage management offset location in which the lower
   --    bound associated with an array type definition, or a constraint
   --    definition is stored
   --
   --    Note: a constraint definition may exist for an object, a subtype, a
   --          derived type, or a constraint being built during tree building.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Lower_Bound_Value_Offset
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural;
   -- record component name: Lower_Bound_Value_Offset

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Upper_Bound
   --
   --  <Description>
   --
   --    Return the upper bound associated with an array type definition,
   --    or a constraint definition
   --
   --   Note: a constraint definition may exist for an object, a subtype, a
   --         derived type, or a constraint being built during tree building.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Upper_Bound
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database;
   -- record component name: Upper_Bound

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Upper_Bound_Value_Offset
   --
   --  <Description>
   --
   --    Return the storage management offset location in which the upper
   --    bound associated with an array type definition, or a constraint
   --    definition is stored
   --
   --    Note: a constraint definition may exist for an object, a subtype, a
   --          derived type, or a constraint being built during tree building.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Upper_Bound_Value_Offset
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural;
   -- record component name: Upper_Bound_Value_Offset

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Next_Constraint
   --
   --  <Description>
   --
   --    Return the next constraint node associated with the index of an
   --    array
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Next_Constraint
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;
   -- record component name: Next_Constraint

   -- -------------------------------------------------------------------------

   -- ------
   -- The following group of routines are associated with acquiring information
   -- based upon several or more components of a record.  This may involve
   -- determining information based upon the content of a component, or
   -- following links from one symbol table entry to another.
   -- ------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Base_Type
   --
   --  <Description>
   --
   --    Returns the base type associated with a type, subtype, or derived
   --    type symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Base_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Root_Type
   --
   --  <Description>
   --
   --    For a symbol table entry, follows base pointers until it reaches
   --    the root type entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   function Get_Root_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_A_Derived_Subprogram
   --
   --  <Description>
   --
   --    This procedure determines whether a symbol table entry is a derived
   --    subprogram.
   --    Return True if the symbol table entry is a derived subprogram;
   --    otherwise returns False.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   function Is_A_Derived_Subprogram
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean;

   -- -------------------------------------------------------------------------
   function Get_Help_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Help;

private


   -- pragma inline(Get_Entry_Kind);
   pragma INLINE (Get_Next_Entry);
   -- pragma inline(Get_Symbolic_Name);
   -- pragma inline(Get_Level);
   -- pragma inline(Get_Enclosing_Package);
   -- pragma inline(Is_Visible);
   -- pragma inline(Is_Hidden);
   -- pragma inline(Is_Removed);
   -- pragma inline(Get_Statement_Number);
   -- pragma inline(Get_Context_Clause_Body);
   -- pragma inline(Get_Context_Clause);
   -- pragma inline(Get_Package_Spec_Decls);
   -- pragma inline(Get_Package_Body_Decls);
   -- pragma inline(Has_A_Body);
   -- pragma inline(Get_Head_Of_Statements_List);
   -- pragma inline(Get_Derived_Subprogram_List);
   -- pragma inline(Get_Parameter_Count);
   -- pragma inline(Has_Default_Parameters);
   -- pragma inline(Get_Number_Of_Default_Parameters);
   -- pragma inline(Get_Head_Of_Parameters_List);
   -- pragma inline(Get_Tail_Of_Parameters_List);
   -- pragma inline(Get_Return_Type);
   -- pragma inline(Get_Head_Of_Declarations_List);
   -- pragma inline(Get_Value_Offset);
   -- pragma inline(Is_Subprogram_Built_In);
   -- pragma inline(Get_Built_In_Instance_Number);
   -- pragma inline(Get_Derived_Subprogram_Parent);
   -- pragma inline(Get_Function_Entry);
   -- pragma inline(Get_Enumeration_Literal_Entry);
   -- pragma inline(Get_Type_Definition);
   -- pragma inline(Get_Type);
   -- pragma inline(Has_Constraint);
   -- pragma inline(Is_Mode_Default_In);
   -- pragma inline(Is_Mode_Explicit_In);
   -- pragma inline(Is_Mode_In);
   -- pragma inline(Is_Mode_Out);
   -- pragma inline(Is_Mode_In_Out);
   -- pragma inline(Are_Modes_Equal);
   -- pragma inline(Get_Initialization_Code);
   -- pragma inline(Get_Next_Parameter);
   -- pragma inline(Is_Preceded_By_Comma);
   -- pragma inline(Is_Type_State_Incomplete);
   -- pragma inline(Is_Type_State_Full);
   -- pragma inline(Is_Type_State_Private);
   -- pragma inline(Is_Type_State_Limited_Private);
   -- pragma inline(Get_Entry_Base_Type);
   -- pragma inline(Get_Literal_Value);
   -- pragma inline(Get_Successor);
   -- pragma inline(Get_Predecessor);
   -- pragma inline(Get_Next_Record_Component);
   -- pragma inline(Get_Simple_Name_Construct);
   -- pragma inline(Is_Type_An_Enumeration_Type);
   -- pragma inline(Is_Type_An_Integer_Type);
   -- pragma inline(Is_Type_A_Real_Type);
   -- pragma inline(Is_Type_A_Record_Type);
   -- pragma inline(Is_Type_An_Array_Type);
   -- pragma inline(Is_Type_An_Access_Type);
   -- pragma inline(Get_Literal_Count);
   -- pragma inline(Get_First_Literal);
   -- pragma inline(Get_Last_Literal);
   -- pragma inline(Get_Lower_Bound);
   -- pragma inline(Get_Upper_Bound);
   -- pragma inline(Get_Record_Component);
   -- pragma inline(Is_Constrained);
   -- pragma inline(Get_Index_Type);
   -- pragma inline(Get_Element_Type);
   -- pragma inline(Get_Lower_Bound);
   -- pragma inline(Get_Lower_Bound_Value_Offset);
   -- pragma inline(Get_Upper_Bound);
   -- pragma inline(Get_Upper_Bound_Value_Offset);
   -- pragma inline(Get_Next_Constraint);
   -- pragma inline(Get_Base_Type);
   -- pragma inline(Get_Root_Type);
   -- pragma inline(Is_A_Derived_Subprogram);
   -- pragma inline(Get_Help_List);

end Symtbl_Entry_Get_Routines;
