--|$Log:	strt.a,v $
--|Revision 8.0  91/01/17  10:41:31  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Symbol_Table_Routines
--
--
-- Abstract
--
--    The symbol table manager constructs and manages the information
--    associated with identifiers within the Ada statements given to the
--    Ada command language interpreter.  The symbol table manager consists of
--    a set of data types (the package Symbol_Table_Structures) and a
--    set of operations (this package: Symbol_Table_Routines).  The other
--    components of the interpreter do not directly manipulate the symbol
--    table; the symbol table routines perform the manipulation.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Literals;
use Literals;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Context_Control;
use Context_Control;

package Symbol_Table_Routines is

   -- Saved Symbol table entries for association with constants.
   Integer_Entry   : Ref_Symbol_Table_Entry;
   String_Entry    : Ref_Symbol_Table_Entry;
   Character_Entry : Ref_Symbol_Table_Entry;
   Boolean_Entry   : Ref_Symbol_Table_Entry;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Symbol_Table_Entry
   --
   --  <Description>
   --
   --    Create_Entry creates a symbol table entry for the specified Name
   --    using the Entry_Type given.  A reference to the symbol table entry
   --    created is returned as the result.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   function Create_Symbol_Table_Entry
       (Name       : in Ref_String;
        Entry_Type : in Type_Entry_Enumeration) return Ref_Symbol_Table_Entry;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Duplicate_Symbol_Table_Entry
   --
   --  <Description>
   --
   --    Duplicate_Entry creates a new symbol table entry for the specified
   --    Name using the template original symbol table entry.  The appropriate
   --    information from the original symbol table entry is copied to the new
   --    symbol table entry, and unique components for the new symbol table
   --    entry are initialized as required.  The new entry is returned.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   function Duplicate_Symbol_Table_Entry
       (Original_Entry : in Ref_Symbol_Table_Entry;
        Name           : in Ref_String) return Ref_Symbol_Table_Entry;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Add_Entry_To_Symbol_Table
   --
   --  <Description>
   --
   --    Add_Entry_To_Symbol_Table inserts the specified symbol table entry and
   --    any subordinate entry associated with the entry into the symbol table.
   --    The addition of this entry may cause other entries to no longer be
   --    directly visible.  Those entries are marked as hidden.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Duplicate_Symbol_Table_Entry_Error is raised if the symbol table
   --      entry already exists within the symbol table.
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   procedure Add_Entry_To_Symbol_Table
       (Context      : in out Context_Definition;
        Symtbl_Entry : in Ref_Symbol_Table_Entry);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Add_Duplicate_Of_Entry_To_Symbol_Table
   --
   --  <Description>
   --
   --    Add_Entry_To_Symbol_Table inserts the specified symbol table entry and
   --    any subordinate entry associated with the entry into the symbol table.
   --    The addition of this entry may cause other entries to no longer be
   --    directly visible.  Those entries are marked as hidden.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   procedure Add_Duplicate_Of_Entry_To_Symbol_Table
       (Context      : in out Context_Definition;
        Symtbl_Entry : in Ref_Symbol_Table_Entry);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Symbol_Table_Entry
   --
   --  <Description>
   --
   --    Find_Symbol_Table_Entry searches the symbol table for the given
   --    name. The first occurrence found within the table is returned;
   --    otherwise null is returned.  Find_Symbol_Table_Entry invoked with
   --    the type of the name (Entry_Type) searches only for the Name
   --    defined by the specified type.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   function Find_Symbol_Table_Entry (Context    : in Context_Definition;
                                     Name       : in Ref_String;
                                     Entry_Type : in Type_Entry_Enumeration :=
      Unknown) return Ref_Symbol_Table_Entry;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Next_Symbol_Table_Entry
   --
   --  <Description>
   --
   --    Find_Next_Symbol_Table_Entry searches the symbol table for the
   --    next occurrence of the name, starting from the specified entry.
   --    The next occurrence of the name found within the symbol table is
   --    returned; otherwise null is returned.
   --    Find_Next_Symbol_Table_Entry invoked with the type of the name
   --    (Entry_Type) searches only for the Name defined by the specified
   --    type.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   function Find_Next_Symbol_Table_Entry
       (Reference_Entry : in Ref_Symbol_Table_Entry;
        Entry_Type      : in Type_Entry_Enumeration := Unknown) return
         Ref_Symbol_Table_Entry;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Entry_Within_Scope
   --
   --  <Description>
   --
   --    Find_Entry_Within_Scope searches a portion of the symbol table for
   --    the given name.  The portion of the symbol table to be searched
   --    is determined by whether parsing is being conducted within a
   --    declarative region.  Only entries defined directly within the
   --    declarative region are considered (i.e. entries introduced by use
   --    statements are ignored.)  The first occurrence found within the table
   --    is returned; otherwise null is returned.
   --    Find_Entry_Within_Scope invoked with the type of the name (Entry_Type)
   --    searches only for the Name defined by the specified type.
   --
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   function Find_Entry_Within_Scope (Context    : in Context_Definition;
                                     Name       : in Ref_String;
                                     Entry_Type : in Type_Entry_Enumeration :=
      Unknown) return Ref_Symbol_Table_Entry;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Next_Entry_Within_Scope
   --
   --  <Description>
   --
   --    Find_Next_Entry_Within_Scope searches a portion of the symbol table
   --    for the next occurrence of the given name, starting from the specified
   --    entry.  The portion of the symbol table to be searched is determined
   --    by whether parsing is being conducted within a declarative region.
   --    Only entries defined directly within the declarative region are
   --    considered (i.e. entries introduced by use statements are ignored.)
   --    The first occurrence found within the table is returned;
   --    otherwise null is returned.
   --    Find_Next_Entry_Within_Scope invoked with the type of the name
   --    (Entry_Type) searches only for the name defined by the specified type.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   function Find_Next_Entry_Within_Scope
       (Context         : in Context_Control.Context_Definition;
        Reference_Entry : in Ref_Symbol_Table_Entry;
        Entry_Type      : in Type_Entry_Enumeration := Unknown) return
         Ref_Symbol_Table_Entry;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Matching_Name
   --
   --  <Description>
   --
   --    Find_Matching_Name searches the symbol table for the given
   --    name.  This routine will look at ALL names within the symbol table,
   --    even those that are not directly visible (i.e. hidden by the USE of
   --    packages).  The first occurrence found within the table is returned;
   --    otherwise null is returned.  Find_Matching_Name invoked with
   --    the type of the name (Entry_Type) searches only for the Name
   --    defined by the specified type.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   function Find_Matching_Name (Context    : in Context_Definition;
                                Name       : in Ref_String;
                                Entry_Type : in Type_Entry_Enumeration :=
      Unknown) return Ref_Symbol_Table_Entry;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Next_Matching_Name
   --
   --  <Description>
   --
   --    Find_Next_Matching_Name searches the symbol table for the
   --    next occurrence of the name, starting from the specified entry.
   --    This routine will look at ALL names within the symbol table,
   --    even those that are not directly visible (i.e. hidden by the USE of
   --    packages).  The next occurrence found within the table is returned;
   --    otherwise null is returned.  Find_Next_Matching_Name invoked with
   --    the type of the name (Entry_Type) searches only for the Name
   --    defined by the specified type.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   function Find_Next_Matching_Name
       (Reference_Entry : in Ref_Symbol_Table_Entry;
        Entry_Type      : in Type_Entry_Enumeration := Unknown) return
         Ref_Symbol_Table_Entry;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Delete_Entry_From_Symbol_Table
   --
   --  <Description>
   --
   --    Delete_Entry_From_Symbol_Table removes the specified symbol table
   --    entry from the symbol table if it currently is in the symbol table.
   --    The entry is actually deleted from the list, as opposed to having
   --    its "visibility" field marked as removed.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   procedure Delete_Entry_From_Symbol_Table
       (Context      : in out Context_Definition;
        Symtbl_Entry : in Ref_Symbol_Table_Entry);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Add_Package_Decls_To_Symbol_Table
   --
   --  <Description>
   --
   --    Adds declarations onto the symbol table so that they may be
   --    visible during parsing of the package body, or that the body
   --    declarations may be visible for debugging purposes.  When parsing
   --    the package body for the package specification, the objects that were
   --    made visible by use statements within the package specification must
   --    be made visible also for the package body.  The Process_Use_Stmts
   --    should be set to True when the package spec declarations are being
   --    added to the symbol table for the parsing of the package body.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================
   type Pkg_Decl_Kind is (Pkg_Spec, Pkg_Body);

   procedure Add_Package_Decls_To_Symbol_Table
       (Context              : in out Context_Control.Context_Definition;
        Package_Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Decl_Kind            : in Pkg_Decl_Kind := Pkg_Spec;
        Process_Use_Stmts    : in Boolean       := False);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Delete_Package_Decls_From_Symbol_Table
   --
   --  <Description>
   --
   --    Removes declarations from the symbol table associated with a
   --    package.  This, in essence, is an "undo" of executing an Ada
   --    use statement.
   --    Symbol table entries are actually unlinked from the symbol table.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   procedure Delete_Package_Decls_From_Symbol_Table
       (Context              : in out Context_Control.Context_Definition;
        Package_Symtbl_Entry : in Ref_Symbol_Table_Entry);


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Unhide_Entry
   --
   --  <Description>
   --
   --    This procedure will traverse the symbol table and look for hidden
   --    subprograms with the same provie as the Symtbl_Entry parameter. If
   --    only one is found it will be unhidden, if more than one is found,
   --    it they will remain obscured.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    none.
   --
   --  <Exceptions Handled>
   --
   --    none.
   --==========================================================================

   procedure Unhide_Entry (Context      : in Context_Control.Context_Definition;
                           Symtbl_Entry : in Ref_Symbol_Table_Entry);

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Top_Of_Symtbl
   --
   --  <Description>
   --
   --    Returns the head of the symbol table.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   function Get_Top_Of_Symtbl
       (Context : in Context_Definition) return Ref_Symbol_Table_Entry renames
      Context_Control.Get_Top_Of_Symtbl;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Top_Of_Symtbl
   --
   --  <Description>
   --
   --    Sets the head of the symbol table to the specified symbol table
   --    entry.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   procedure Set_Top_Of_Symtbl (Context : in out Context_Definition;
                                Item    : in Ref_Symbol_Table_Entry) renames
      Context_Control.Set_Top_Of_Symtbl;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Saved_Main_Level_Symbol_Table_Head
   --
   --  <Description>
   --
   --    This procedure retains the value of a symbol table entry such that
   --    it may be used by the Reset_From_Subprogram_Or_Package_Declaration
   --    procedure.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   procedure Set_Saved_Main_Level_Symbol_Table_Head
       (Context      : in out Context_Definition;
        Symtbl_Entry : in Ref_Symbol_Table_Entry) renames Context_Control.
         Set_Saved_Main_Level_Symbol_Table_Head;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Saved_Main_Level_Symbol_Table_Head
   --
   --  <Description>
   --
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   function Get_Saved_Main_Level_Symbol_Table_Head
       (Context : in Context_Definition) return Ref_Symbol_Table_Entry renames
      Context_Control.Get_Saved_Main_Level_Symbol_Table_Head;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Enum_Node
   --
   --  <Description>
   --
   --     Get_Enum_Node will return the Ref_Symbol_Table_Entry of the
   --    enumeration literal node of type "enum_type"  whose literal value is
   --    "enum_literal_value"
   --
   --==========================================================================

   function Get_Enum_Node
       (Enum_Type          : in Ref_Symbol_Table_Entry;
        Enum_Literal_Value : in Ace_Integer) return Ref_Symbol_Table_Entry;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Initialize_Symbol_Table
   --
   --  <Description>
   --
   --    This procedure initializes the symbol table.
   --
   --==========================================================================

   procedure Initialize_Symbol_Table;


   --    pragma inline(Create_Symbol_Table_Entry);
   --    pragma inline(Duplicate_Symbol_Table_Entry);
   --    pragma inline(Add_Entry_To_Symbol_Table);
   --    pragma inline(Add_Duplicate_Of_Entry_To_Symbol_Table);
   --    pragma inline(Find_Symbol_Table_Entry);
   --    pragma inline(Find_Next_Symbol_Table_Entry);
   --    pragma inline(Find_Entry_Within_Scope);
   --    pragma inline(Find_Next_Entry_Within_Scope);
   --    pragma inline(Find_Matching_Name);
   --    pragma inline(Find_Next_Matching_Name);
   --    pragma inline(Delete_Entry_From_Symbol_Table);
   --    pragma inline(Add_Package_Decls_To_Symbol_Table);
   --    pragma inline(Delete_Package_Decls_From_Symbol_Table);
   --    pragma inline(Get_Top_Of_Symtbl);
   --    pragma inline(Set_Top_Of_Symtbl);
   --    pragma inline(Set_Saved_Main_Level_Symbol_Table_Head);
   --    pragma inline(Get_Saved_Main_Level_Symbol_Table_Head);
   --    pragma inline(Get_Enum_Node);
   --    pragma inline(Initialize_Symbol_Table);

end Symbol_Table_Routines;
