--|$Log:	sts.a,v $
--|Revision 8.0  91/01/17  10:41:32  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Symbol_Table_Structures
--
---------------------------------------------------------------
--
--
-- Abstract
--    Symbol table management consists of the construction and management of
--    the information associated with identifiers within the Ada statements
--    handled by the Ada Command Language Interpreter.  Symbol table management
--    consists of a set of data types for symbol table entries (this package:
--    Symbol_Table_Structures), a set of operations to manipulate the
--    components of the symbol table entries (the packages
--    Symtbl_Entry_Get_Routines and Symtbl_Entry_Set_Routines) and a set
--    of routines which organize and access symbol table entries to support
--    the scoping and visibility rules of Ada (the package
--    Symbol_Table_Routines).
--
--    These data structures describe symbol table entries.  An entry contains
--    the information relevant to the function served by the symbolic name in
--    the Ada code.  An initial overall definition of these structures was
--    performed at the initial design stages of ACE.  As additional
--    functionality is added to ACE, a more detailed design is performed and
--    the necessary modifications to the structures are performed.  (i.e.
--    it is known that some of the structures need to be revised to retain
--    the specific attributes properly).
--
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Ace_Universal_Types;
use Ace_Universal_Types;

with Literals;
use Literals;

with Help_Pkg;
use Help_Pkg;

package Symbol_Table_Structures is

   -- -------------------------------------------------------------------------
   -- This package consists of three major structure definitions:
   --    Symbol Table Entries, Type Definition Nodes, and Constraint Nodes.
   -- The symbol table entry is the primary data structure that makes use
   -- of the other types of nodes as retainers of secondary information
   -- associated with some symbol table entries.
   -- -------------------------------------------------------------------------

   -- --------------------------------------------------------------------------
   -- ***** WORKAROUND *****
   -- The symbol table structures contain references to the statement
   -- database structures.  Also, the statement database structures contain
   -- references to the symbol table structures.  Do not want to place these
   -- two distinct objects within the same package.  Therefore, have defined
   -- a "fake" reference within this package to a statement database node
   -- that will get converted to a proper statement database node.

   type Ref_Statement_Database_Fake is access Integer;
   --    A true Ref_Statement_Database will be assigned to objects of this
   --    type through unchecked conversion routine.
   -- --------------------------------------------------------------------------

   -- -----
   -- Associated with each identifier is the usage of the name.
   -- Information retained relevant to the identifier is based upon
   -- its usage.
   -- -----
   type Type_Entry_Enumeration is
      (Package_Entry,
       Procedure_Entry,
       Function_Entry,
       Object_Entry,
       Exception_Entry,
       Formal_Param_Entry,
       Type_Entry,
       Subtype_Entry,
       Derived_Type_Entry,
       Constraint_Entry,
       Enumeration_Literal_Entry,
       Label_Entry,
       Record_Component_Entry,
       Simple_Name_Entry,
       Unknown);

   -- -----
   -- The type definition node contains references to the symbol table node.
   -- Provide an incomplete type definition of the symbol table node so that
   -- a reference type to a symbol table node may be defined prior to the
   -- definition of the type definition node
   -- -----
   type Symbol_Table_Entry (Entry_Kind : Type_Entry_Enumeration);
   type Ref_Symbol_Table_Entry is access Symbol_Table_Entry;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::
   -- : Constraint Node :
   -- :::::::::::::::::::

   -- -----
   -- Constraint type enumerates the possible kinds of constraints
   -- -----
   type Constraint_Enumeration is
      (Range_Constraint,
       Index_Constraint,
       Floating_Point_Constraint,
       Fixed_Point_Constraint,
       Discriminant_Constraint);

   -- -----
   -- Constraint information tmay be associated with either a subtype,
   -- derived type, or object.  The constraint node contains the relevant
   -- constraint information based upon the kind of constraint.
   -- -----

   type Constraint_Node (Constraint_Kind : Constraint_Enumeration) is
      record
         case Constraint_Kind is

            -- Range and Index constraints for enumeration types are retained
            -- by their numeric representation

            when Range_Constraint   | Index_Constraint  =>
               Lower_Bound              : Ref_Statement_Database_Fake := null;
               Lower_Bound_Value_Offset : Ace_Natural                 := 0;
               Upper_Bound              : Ref_Statement_Database_Fake := null;
               Upper_Bound_Value_Offset : Ace_Natural                 := 0;
               Next_Constraint          : Ref_Symbol_Table_Entry      := null;
               -- to constraint_entry

               -- The following three constraint types are not planned to be
               -- implemented in the six month prototype

            when Floating_Point_Constraint    =>
               null;
            when Fixed_Point_Constraint  =>
               null;
            when Discriminant_Constraint =>
               null;

         end case;
      end record;

   type Ref_Constraint_Node is access Constraint_Node;

   -- --------------------------------------------------------------------------

   -- ::::::::::::::::::::::::
   -- : Type Definition Node :
   -- ::::::::::::::::::::::::

   -- -----
   -- Basic type definitions are the categories of types for which a
   -- type definition record will exist within the symbol table
   -- -----
   type Basic_Type_Definitions is
      (Enumeration_Type,
       Integer_Type,
       Real_Type,
       Record_Type,
       Array_Type,
       Access_Type);


   -- -----
   -- For each type defined by the user, the symbol table manager constructs
   -- and manipulates the information relevant to the particular type.
   -- The information retained for a type is based upon the kind of the type.
   -- The type information may contain values or references to other
   -- identifiers
   -- -----
   type Type_Definition_Node (Type_Kind : Basic_Type_Definitions) is
      record
         case Type_Kind is


            when Enumeration_Type   =>
               Literal_Count : Ace_Natural            := 0;
               First_Literal : Ref_Symbol_Table_Entry := null;
               Last_Literal  : Ref_Symbol_Table_Entry := null; -- ?? Needed

            when Integer_Type  =>
               Lower_Bound : Ace_Integer := 0;
               Upper_Bound : Ace_Integer := 0;

            when Real_Type     =>
               null; -- not supported

            when Access_Type   =>
               null; -- not supported

            when Record_Type   =>
               -- reference to first component name
               Record_Component : Ref_Symbol_Table_Entry := null;

            when Array_Type    =>
               Is_Constrained     : Boolean                := False;
               Array_Index_Type   : Ref_Statement_Database_Fake := null;
               Array_Element_Type : Ref_Statement_Database_Fake := null;

         end case;
      end record;

   type Ref_Type_Definition_Node is access Type_Definition_Node;

   -- -------------------------------------------------------------------------

   -- -----
   -- Linked list, where the information stored within the linked list
   -- is a symbol table entry (at present, used for subprogram entries)
   -- -----
   type List_Of_Subprograms;
   type Ref_List_Of_Subprograms is access List_Of_Subprograms;
   type List_Of_Subprograms is
      record
         Subprogram_Entry      : Ref_Symbol_Table_Entry;
         Next_Subprogram_Entry : Ref_List_Of_Subprograms;
      end record;

   -- -------------------------------------------------------------------------

   -- ::::::::::::::::::::::
   -- : Symbol Table Entry :
   -- ::::::::::::::::::::::

   -- -----
   -- Formal paramter mode type enumerates the possible access attribute
   -- associated with formal subprogram parameters
   -- -----
   type Mode_Enumeration is (Default_In_Mode, In_Mode, In_Out_Mode, Out_Mode);

   -- -----
   -- The state attribute for type name provides a categorization of
   -- the kind of type definition given
   -- -----
   type Type_State_Attr is (Incomplete, Full, Only_Private, Limited_Private);

   -- -----
   -- The name status for an item identifies the visibility of the name,
   -- The default value is Visible.  Use statements for packages may hide
   -- (Hidden) the name.  For dynamic redefinition of identifiers within
   -- the ACLI, Removed states that the name has been undefined.
   -- -----
   type Name_Visibility_Status is (Visible, Hidden, Removed);

   -- -----
   -- The symbol table node is the major element within the symbol table.
   -- Each identifier within a user's Ada statement is inserted into the
   -- symbol table with relevant information to handle future references to
   -- the identifier.
   -- Common information is retained for each identifier.  The majority of
   -- the common information is designed for internal usage by the interpreter
   -- and symbol table manager.
   -- The specific information for an identifier varies according to the usage
   -- of the identifier.  Some symbol table nodes are completely
   -- self-sufficient; others contain references to other nodes within the
   -- symbol table.
   --
   -- Within the common information, the Next_Entry field is used internally
   -- by the symbol table manager to keep a linked list of symbol table
   -- nodes.  This implementation of the symbol table as a linked list is
   -- isolated within the symbol table manager.  Other components of the ACLI
   -- are insulated from the implementation mechanism of the symbol table
   -- manager.
   -- -----

   type Symbol_Table_Entry (Entry_Kind : Type_Entry_Enumeration) is
      record

         -- Next entry in symbol table
         Next_Entry        : Ref_Symbol_Table_Entry := null;

         -- Identifier name
         Symbol_Name       : Ref_String             := Create;

         Item_Level        : Ace_Natural            := 0;

         Enclosing_Package : Ref_Symbol_Table_Entry := null;

         Visibility        : Name_Visibility_Status := Visible;

         Statement_Number  : Ace_Natural            := 0;

         --      Help_Comment      : Ref_Help               := null;

         case Entry_Kind is

            when Package_Entry =>
               Pkg_Spec_Context_Clause : Ref_Statement_Database_Fake := null;
               Pkg_Body_Context_Clause : Ref_Statement_Database_Fake := null;
               Package_Spec_Decls : Ref_Statement_Database_Fake      := null;
               Package_Body_Decls : Ref_Statement_Database_Fake      := null;
               --
               Has_Body                    : Boolean                 := False;
               --
               Package_Stmts : Ref_Statement_Database_Fake           := null;
               Pkg_Derived_Subprogram_List : Ref_List_Of_Subprograms := null;
               Pkg_Help                    : Ref_Help                := null;

            when Procedure_Entry    | Function_Entry    =>
               Parameter_Count           : Ace_Natural                 := 0;
               Default_Parameter_Count   : Ace_Natural                 := 0;
               Head_Param_List           : Ref_Symbol_Table_Entry      := null;
               --    null if no params
               Tail_Param_List           : Ref_Symbol_Table_Entry      := null;
               --    null if no params
               Return_Type               : Ref_Symbol_Table_Entry      := null;
               --    only for functions
               Subprg_Context_Clause     : Ref_Statement_Database_Fake := null;
               Subprg_Help               : Ref_Help                    := null;
               Declarations_List         : Ref_Statement_Database_Fake := null;
               Statements_List           : Ref_Statement_Database_Fake := null;
               Exception_Handler         : Ref_Statement_Database_Fake := null;
               --    prologue or first stmt of subprogram
               Function_Value_Offset     : Ace_Natural                 := 0;
               Is_Built_In_Subprogram    : Ace_Natural                 :=
                  Ace_Natural'LAST;
               --    signifies that subprogram is to be handled
               --    internally w/in ACE
               Has_Explicit_Decl         : Boolean                     := False;
               -- signifies that a subprogram declaraion was explicitly
               -- made.
               Derived_Subprogram_Parent : Ref_Symbol_Table_Entry      := null;
               --    for a derived subprogram, a pointer to the subprogram
               --    for which a body will exist

               -- when not null, this points to the enumeration literal
               -- symtbl entry
               Enum_Lit_Entry            : Ref_Symbol_Table_Entry      := null;


            when Exception_Entry    =>
               null;
               -- We need no other info other than it's name
               -- and a pointer to this entry??

            when Object_Entry  =>
               Object_Type         : Ref_Symbol_Table_Entry := null;
               Object_Constraint   : Ref_Symbol_Table_Entry := null;
               --    may be null
               Object_Mode         : Mode_Enumeration       := In_Out_Mode;
               Object_Value_Offset : Ace_Natural            := 0;

            when Formal_Param_Entry =>
               Param_Type         : Ref_Symbol_Table_Entry      := null;
               Param_Mode         : Mode_Enumeration := Default_In_Mode;
               Param_Init         : Ref_Statement_Database_Fake := null;
               Next_Param         : Ref_Symbol_Table_Entry      := null;
               Param_Value_Offset : Ace_Natural                 := 0;
               Preceded_By_Comma  : Boolean                     := False;

            when Type_Entry    =>
               Concreteness                 : Type_State_Attr          := Full;
               Type_Definition              : Ref_Type_Definition_Node := null;
               Type_Derived_Subprogram_List : Ref_List_Of_Subprograms  := null;
               Type_Help                    : Ref_Help                 := null;

            when Subtype_Entry | Derived_Type_Entry     =>
               Base_Type                     : Ref_Symbol_Table_Entry  := null;
               Other_Type_Help               : Ref_Help                := null;
               Constraint_Info               : Ref_Symbol_Table_Entry  := null;
               -- may be null
               Dtype_Derived_Subprogram_List : Ref_List_Of_Subprograms := null;
               --   only applicable for derived types

            when Constraint_Entry   =>
               Constraint_Ref : Ref_Constraint_Node := null;

            when Enumeration_Literal_Entry    =>
               Literal_Value     : Ace_Natural            := 0;
               Enumeration_Type  : Ref_Symbol_Table_Entry := null;
               Succ_Enum_Literal : Ref_Symbol_Table_Entry := null;
               --    null for last literal
               Pred_Enum_Literal : Ref_Symbol_Table_Entry := null;
               --    null for first literal

               -- references the parameterless function call symtbl entry
               Func_Entry        : Ref_Symbol_Table_Entry := null;

            when Label_Entry   =>
               Label_Stmt : Ref_Statement_Database_Fake := null;

            when Record_Component_Entry  =>
               Component_Type        : Ref_Symbol_Table_Entry      := null;
               Component_Init        : Ref_Statement_Database_Fake := null;
               Next_Record_Component : Ref_Symbol_Table_Entry      := null;

            when Simple_Name_Entry  =>
               Simple_Name_Construct : Ref_Statement_Database_Fake := null;

            when Unknown  =>
               null;

         end case;
      end record;

end Symbol_Table_Structures;
