--|$Log:	stset.a,v $
--|Revision 8.0  91/01/17  10:41:33  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Symtbl_Entry_Set_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--
--    This package contains the set of routines that stores values into fields
--    of symbol table entries.  All storage of values into components of
--    symbol table entries should be made through the routines provided within
--    this package.  This package (in conjunction with a similar package for
--    retrieving from symbol table entry components) provides the mechanism for
--    hiding the makeup and names of symbol table entry components.  Changes to
--    the structure and component names for symbol table entries would only
--    affect this package of routines, rather than all the other components
--    that deal with symbol table information.
--
--    The routines within the package are ordered with respect to the
--    definition of components for a symbol table entry.  The routines that
--    store values into components that are common to all symbol table entries
--    come first, followed by components that are applicable to different
--    variants of the symbol table entry.
--
--    Due to the Ada requirement that all record component names be distinct
--    (even those within different variants), a single routine is used to
--    store a value into  a similar component that is named differently within
--    several variants.  This routine is defined within the section associated
--    with the first variant that contains the particular data item.  After
--    each routine is listed the record component associated with the routine.
--    Groupings of routines are introduced by the variant which contains
--    the components.
--
--    Components whose values are that of an enumerated type "hide" the
--    enumeration literals from external packages.  Instead of storing
--    the value of the enumerated type, a set of routines are provided that
--    have incorporated the enumeration literal into the routine name.
--    (e.g. Set_Mode_To_In, Set_Mode_To_Out, Set_Mode_To_In_Out, ... rather
--    than Set_Mode which would accept the enumeration type as an input
--    parameter.)
--

-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Statement_Database;
use Statement_Database;

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Storage_Management;
use Storage_Management;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Literals;
use Literals;

package Symtbl_Entry_Set_Routines is

   -- --------------------------------------------- --
   -- Components Common to All Symbol Table Entries --
   -- --------------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Entry_Kind
   --
   --  <Description>
   --
   --    Set the kind of entry associated with the symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Entry_Kind (Symtbl_Entry : in out Ref_Symbol_Table_Entry;
                             Item         : in Type_Entry_Enumeration);
   -- record component name: Entry_Kind

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Next_Entry
   --
   --  <Description>
   --
   --    Set the next entry field of the Symtbl_Entry node to Item
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Next_Entry (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                             Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Next_Entry

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Symbolic_Name
   --
   --  <Description>
   --
   --    Set the symbolic name associated with the symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Symbolic_Name (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                Item         : in Ref_String);
   -- record component name: Symbol_Name

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Level
   --
   --  <Description>
   --
   --    Set the static level associated with the symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Level (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                        Item         : in Ace_Natural);
   -- record component name: Item_Level

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Enclosing_Package
   --
   --  <Description>
   --
   --    Set the enclosing package field of the Symtbl_Entry node to Item
   --    (which should be a package symbol table entry)
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Enclosing_Package (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                    Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Enclosing_Package

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Make_Entry_Visible
   --
   --  <Description>
   --
   --    Set the visibility field of the symbol table entry such that the
   --    item is visible
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Make_Entry_Visible (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Visibility

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Make_Entry_Hidden
   --
   --  <Description>
   --
   --    Set the visibility field of the symbol table entry such that the
   --    item is hidden
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Make_Entry_Hidden (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Visibility

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Make_Entry_Removed
   --
   --  <Description>
   --
   --    Set the visibility field of the symbol table entry such that the
   --    item is "removed" from the symbol table
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Make_Entry_Removed (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Visibility

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Statement_Number
   --
   --  <Description>
   --
   --    Set the statement number associated with the symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Statement_Number (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                   Item         : in Ace_Natural);
   -- record component name: Statement_Number

   -- -------------------------------------------------------------------------

   -- -------------------------------------- --
   -- Components for Variant : Package_Entry --
   -- -------------------------------------- --
   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Context_Clause_Body
   --
   --  <Description>
   --
   --    Set the context clause associated with a compilation unit
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Context_Clause_Body (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                      Item         : in Ref_Statement_Database);

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Context_Clause
   --
   --  <Description>
   --
   --    Set the context clause associated with a compilation unit
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Context_Clause (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                 Item         : in Ref_Statement_Database);
   -- record component name: Context_Clause

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Package_Spec_Decls
   --
   --  <Description>
   --
   --    Set the reference to the statement database node associated with the
   --    declarations within the specification portion of the package
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Package_Spec_Decls (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                     Item         : in Ref_Statement_Database);
   -- record component name: Package_Spec_Decls

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Package_Body_Decls
   --
   --  <Description>
   --
   --    Set the reference to the statement database node associated with the
   --    declarations within the body  portion of the package
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Package_Body_Decls (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                     Item         : in Ref_Statement_Database);
   -- record component name: Package_Body_Decls


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Has_Body
   --
   --  <Description>
   --
   --    Set the fact that the package has a body
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Has_Body (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Has_Body

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Head_Of_Statements_List
   --
   --  <Description>
   --
   --    Set the head of the list of statements for the executable
   --    statements associated with a package symbol table entry, a
   --    subprogram symbol table entry, or a label symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Head_Of_Statements_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Statement_Database);
   -- record component name: Package_Stmts
   -- record component name: Statements_List
   --   (for variant: Procedure_Entry, Function_Entry)
   -- record component name: Label_Stmt
   --   (for variant: Label_Stmt)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Derived_Subprogram_List
   --
   --  <Description>
   --
   --    Set the head of the list of derived subprograms that are
   --    associated with a package.  Also, for types and derived types,
   --    set the head of the list of subprograms which are derivable
   --    from this type (or derived type).
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Derived_Subprogram_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_List_Of_Subprograms);
   -- record component name: Pkg_Derived_Subprogram_List
   -- record component name: Type_Derived_Subprogram_List
   --   (for variant: Type_Entry)
   -- record component name: DType_Derived_Subprogram_List
   --   (for variant: Derived_Type_Entry)

   -- -------------------------------------------------------------------------

   -- ---------------------------------------- --
   -- Components for Variant : Procedure_Entry --
   --                        : Function_Entry  --
   -- ---------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Subprg_Has_Decl
   --
   --  <Description>
   --
   --    Sets flag to indicate subprogram has an explicit declaration
   --
   --  <Called By>
   --
   --    Build_Subprogram_Spec_Decl
   --
   --==========================================================================

   procedure Set_Subprg_Has_Decl (Symtbl_Entry : in Ref_Symbol_Table_Entry);

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Parameter_Count
   --
   --  <Description>
   --
   --    Set the number of parameters associated with a subprogram
   --    symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Parameter_Count (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                  Item         : in Ace_Natural);
   -- record component name: Parameter_Count

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Number_Of_Default_Parameters
   --
   --  <Description>
   --
   --    Set the number of parameters associated with a subprogram
   --    symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Number_Of_Default_Parameters
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ace_Natural);
   -- record component name: Default_Parameter_Count

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Head_Of_Parameters_List
   --
   --  <Description>
   --
   --    Set the reference to the first formal parameter associated with
   --    a subprogram symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Head_Of_Parameters_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Head_Param_List

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Tail_Of_Parameters_List
   --
   --  <Description>
   --
   --   Set the reference to the last formal parameter associated with
   --   a subprogram symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Tail_Of_Parameters_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Tail_Param_List

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Return_Type
   --
   --  <Description>
   --
   --    For a function subprogram, set the reference to the return
   --    type associated with a function subprogram symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Return_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                              Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Return_Type

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Head_Of_Declarations_List
   --
   --  <Description>
   --
   --    Set the head of the list of statements for the declaration
   --    statements associated with a subprogram symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Head_Of_Declarations_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Statement_Database);
   -- record component name: Declarations_List

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Value_Offset
   --
   --  <Description>
   --
   --    Set the reference to the data storage location in which the value
   --    associated with the item (object, parameter, or function) is stored
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Value_Offset (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                               Item         : in Ace_Natural);
   -- record component name: Function_Value_Offset
   -- record component name: Object_Value_Offset
   --   (for variant: Object_Entry)
   -- record component name: Param_Value_Offset
   --   (for variant: Formal_Param_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_As_Built_In_Subprogram
   --
   --  <Description>
   --
   --    Associate an internal numeric identifier with an ACE built-in
   --    subprogram
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_As_Built_In_Subprogram
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ace_Natural);
   -- record component name: Is_Built_In_Subprogram

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Derived_Subprogram_Parent
   --
   --  <Description>
   --
   --    Set the reference to the parent of a derived subprogram.
   --    (A non-derived subprogram has a null parent.)
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Derived_Subprogram_Parent
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Derived_Subprogram_Parent

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Function_Entry
   --
   --  <Description>
   --
   --    Set the reference to the parameterless function call symtbl entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Function_Entry (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                 Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Func_Entry

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Enumeration_Literal_Entry
   --
   --  <Description>
   --
   --    Set the reference to the enumeration literal entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Enumeration_Literal_Entry
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Enum_Lit_Entry

   -- -------------------------------------------------------------------------

   -- ------------------------------------- --
   -- Components for Variant : Object_Entry --
   -- ------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type
   --
   --  <Description>
   --
   --    Set the reference to the type symbol table entry associated with
   --    either an object, a parameter, an enumeration literal, or a record
   --    component
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                       Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Object_Type
   -- record component name: Param_Type
   --   (for variant: Formal_Param_Entry)
   -- record component name: Enumeration_Type
   --   (for variant: Enumeration_Literal_Entry)
   -- record component name: Component_Type
   --   (for variant: Record_Component_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Constraint
   --
   --  <Description>
   --
   --    Set the constraint for the item (object, subtype or derived type) to
   --    refer to a constraint entry already constructed
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Constraint (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                             Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Object_Constraint
   -- record component name: Constraint_Info
   --   (for variant: Subtype_Entry, Derived_Type_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Mode_To_Default_In
   --
   --  <Description>
   --
   --    Set the mode for the item (object or parameter) to DEFAULT_IN
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Mode_To_Default_In (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Object_Mode
   -- record component name: Param_Mode
   --   (for variant: Formal_Param_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Mode_To_In
   --
   --  <Description>
   --
   --    Set the mode for the item (object or parameter) to IN
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Mode_To_In (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Object_Mode
   -- record component name: Param_Mode
   --   (for variant: Formal_Param_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Mode_To_Out
   --
   --  <Description>
   --
   --    Set the mode for the item (object or parameter) to OUT
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Mode_To_Out (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Object_Mode
   -- record component name: Param_Mode
   --   (for variant: Formal_Param_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Mode_To_In_Out
   --
   --  <Description>
   --
   --    Set the mode for the item (object or parameter) to IN OUT
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Mode_To_In_Out (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Object_Mode
   -- record component name: Param_Mode
   --   (for variant: Formal_Param_Entry)

   -- -------------------------------------------------------------------------

   -- ------------------------------------------- --
   -- Components for Variant : Formal_Param_Entry --
   -- ------------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Initialization_Code
   --
   --  <Description>
   --
   --    Set the initialization code location for the formal parameter or
   --    the record component
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Initialization_Code (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                      Item         : in Ref_Statement_Database);
   -- record component name: Param_Init
   -- record component name: Component_Init
   --   (for variant: Record_Component_Entry)

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Next_Parameter
   --
   --  <Description>
   --
   --    Set the symbol table entry for the next parameter following this
   --    parameter in the specification of the subprogram
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Next_Parameter (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                 Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Next_Param

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Preceded_By_Comma
   --
   --  <Description>
   --
   --    Set the symbol table entry for the formal parameter to indicate
   --    that this parameter name was preceded by a comma in the textual
   --    definition of the subprogram declaration
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Preceded_By_Comma (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Preceded_By_Comma

   -- -------------------------------------------------------------------------

   -- ----------------------------------- --
   -- Components for Variant : Type_Entry --
   -- ----------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_State_To_Incomplete
   --
   --  <Description>
   --
   --    Set the state of the this type identifier to be an incomplete
   --    type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Type_State_To_Incomplete
       (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Concreteness

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_State_To_Full
   --
   --  <Description>
   --
   --    Set the state of the this type identifier to be a full type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Type_State_To_Full (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Concreteness

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_State_To_Private
   --
   --  <Description>
   --
   --    Set the state of the this type identifier to be a private
   --    type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Type_State_To_Private
       (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Concreteness

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_State_To_Limited_Private
   --
   --  <Description>
   --
   --    Set the state of the this type identifier to be a limited private
   --    type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Type_State_To_Limited_Private
       (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Concreteness

   -- -------------------------------------------------------------------------

   -- ------------------------------------------- --
   -- Components for Variant : Subtype_Entry      --
   --                        : Derived_Type_Entry --
   -- ------------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Entry_Base_Type
   --
   --  <Description>
   --
   --    Set the base type definition associated with the subtype or the
   --    derived type symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Entry_Base_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                  Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Base_Type

   -- -------------------------------------------------------------------------

   -- -------------------------------------------------- --
   -- Components for Variant : Enumeration_Literal_Entry --
   -- -------------------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Literal_Value
   --
   --  <Description>
   --
   --    Set the internal numeric value assigned to the enumeration literal
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Literal_Value (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                Item         : in Ace_Natural);
   -- record component name: Literal_Value

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Successor
   --
   --  <Description>
   --
   --    Set the symbol table entry for the successor to this enumeration
   --    literal
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Successor (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                            Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Succ_Enum_Literal

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Predecessor
   --
   --  <Description>
   --
   --    Return the symbol table entry for the predecessor to this
   --    enumeration literal
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Predecessor (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                              Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Pred_Enum_Literal

   -- -------------------------------------------------------------------------

   -- ----------------------------------------------- --
   -- Components for Variant : Record_Component_Entry --
   -- ----------------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Next_Record_Component
   --
   --  <Description>
   --
   --    Set the symbol table entry for the next record component following
   --    this record component
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Next_Record_Component
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Next_Record_Component

   -- -------------------------------------------------------------------------
   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Simple_Name_Construct
   --
   --  <Description>
   --
   --    Set the symbol table entry for Simple_Name_Entry to the appropriate
   --    statement database node.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Simple_Name_Construct
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Statement_Database);
   -- Simple_Name_Construct


   -- ------
   -- A Type_Entry variant symbol table entry contains a reference to a
   -- Type_Definition_Node.  The Type_Definition_Node contains informaton
   -- associated with particular types, with variants supporting the
   -- different types.  Access to the information in the Type_Definition_Node
   -- is made through the following set of routines.
   -- ------

   -- ------------------------------- --
   -- Record ==> Type_Definition_Node --
   --   Component for ALL Variants    --
   -- ------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_To_Enumeration_Type
   --
   --  <Description>
   --
   --    Create a type definition node that defines an enumeration type
   --    and associate it with the symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Type_To_Enumeration_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Type_Kind


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_To_Integer_Type
   --
   --  <Description>
   --
   --    Create a type definition node that defines an integer type
   --    and associate it with the symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Type_To_Integer_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry
                                       );
   -- record component name: Type_Kind

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_To_Real_Type
   --
   --  <Description>
   --
   --    Create a type definition node that defines a real type
   --    and associate it with the symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Type_To_Real_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Type_Kind

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_To_Record_Type
   --
   --  <Description>
   --
   --    Create a type definition node that defines a record type
   --    and associate it with the symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Type_To_Record_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Type_Kind

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_To_Array_Type
   --
   --  <Description>
   --
   --    Create a type definition node that defines an array type
   --    and associate it with the symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Type_To_Array_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Type_Kind

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_To_Access_Type
   --
   --  <Description>
   --
   --    Create a type definition node that defines an access type
   --    and associate it with the symbol table entry
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Type_To_Access_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Type_Kind


   -- ----------------------------------------- --
   --    Record ==> Type_Definition_Node        --
   -- Components for Variant : Enumeration_Type --
   -- ----------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Literal_Count
   --
   --  <Description>
   --
   --    Set the number of enumeration literals associated with an
   --    enumeration type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Literal_Count (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                Item         : in Ace_Natural);
   -- record component name: Literal_Count

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_First_Literal
   --
   --  <Description>
   --
   --    Set the symbol table entry for the first enumeration literal
   --    associated with an enumeration type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_First_Literal (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                Item         : in Ref_Symbol_Table_Entry);
   -- record component name: First_Literal

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Last_Literal
   --
   --  <Description>
   --
   --    Set the symbol table entry for the last enumeration literal
   --    associated with an enumeration type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Last_Literal (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                               Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Last_Literal


   -- ------------------------------------- --
   --    Record ==> Type_Definition_Node    --
   -- Components for Variant : Integer_Type --
   -- ------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Lower_Bound
   --
   --  <Description>
   --
   --    Set the lower bound associated with an integer type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Lower_Bound (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                              Item         : in Ace_Integer);
   -- record component name: Lower_Bound

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Upper_Bound
   --
   --  <Description>
   --
   --    Set the upper bound associated with an integer type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Upper_Bound (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                              Item         : in Ace_Integer);
   -- record component name: Upper_Bound

   -- ------------------------------------ --
   --    Record ==> Type_Definition_Node   --
   -- Components for Variant : Record_Type --
   -- ------------------------------------ --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Record_Component
   --
   --  <Description>
   --
   --    Set the first record component associated with a record type
   --    definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Record_Component (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                   Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Record_Component

   -- ----------------------------------- --
   --    Record ==> Type_Definition_Node  --
   -- Components for Variant : Array_Type --
   -- ----------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_As_Constrained
   --
   --  <Description>
   --
   --    Set the array to be a constrained array for an array type
   --    definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_As_Constrained (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Is_Constrained

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_As_Unconstrained
   --
   --  <Description>
   --
   --    Set the array to be an unconstrained array for an array type
   --    definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_As_Unconstrained (Symtbl_Entry : in Ref_Symbol_Table_Entry);
   -- record component name: Is_Constrained

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Index_Type
   --
   --  <Description>
   --
   --    Set the symbol table entry for the index associated with the
   --    array type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Index_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                             Item         : in Ref_Statement_Database);
   -- record component name: Array_Index_Type

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Element_Type
   --
   --  <Description>
   --
   --    Set the symbol table entry for the element associated with the
   --    array type definition
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Element_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                               Item         : in Ref_Statement_Database);
   -- record component name: Array_Element_Type

   -- -------------------------------------------------------------------------

   -- ------
   -- A Constraint_Entry variant symbol table entry contains a reference to a
   -- Constraint_Node.  (Also, the Object_Entry variant symbol table entry
   -- contains a component Object_Constraint and the Subtype_Entry &
   -- Derived_Type_Entry contains a component Constraint_Info.  Both of these
   -- are references to a Constraint_Entry symbol table entry.  The following
   -- routines support references with the Constraint_Entry, Object_Entry,
   -- Subtype_Entry, or the Derived_Type_Entry passed in as the starting
   -- symbol table entry.)  The Constraint_Node contains information
   -- associated with constraints, with variants supporting the different kinds
   -- of constraints.  Access to the information in the Constraint_Node is made
   -- through the following set of routines.
   -- ------

   -- ----------------------------------------- --
   --    Record ==> Constraint_Node             --
   -- Components for Variant : Range_Constraint --
   --                        : Index_Constraint --
   -- ----------------------------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Lower_Bound
   --
   --  <Description>
   --
   --    Set the lower bound associated with an array type definition,
   --    or a constraint definition
   --
   --    Note: a constraint definition may exist for an object, a subtype, a
   --          derived type, or a constraint being built during tree building.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Lower_Bound (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                              Item         : in Ref_Statement_Database);
   -- record component name: Lower_Bound

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Lower_Bound_Value_Offset
   --
   --  <Description>
   --
   --    Set the storage management offset location in which the lower
   --    bound associated with an array type definition, or a constraint
   --    definition is stored
   --
   --    Note: a constraint definition may exist for an object, a subtype, a
   --          derived type, or a constraint being built during tree building.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Lower_Bound_Value_Offset
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ace_Natural);
   -- record component name: Lower_Bound_Value_Offset

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Upper_Bound
   --
   --  <Description>
   --
   --    Set the upper bound associated with an array type definition,
   --    or a constraint definition
   --
   --    Note: a constraint definition may exist for an object, a subtype, a
   --          derived type, or a constraint being built during tree building.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Upper_Bound (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                              Item         : in Ref_Statement_Database);
   -- record component name: Upper_Bound

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Upper_Bound_Value_Offset
   --
   --  <Description>
   --
   --    Set the storage management offset location in which the upper
   --    bound associated with an array type definition, or a constraint
   --    definition is stored
   --
   --    Note: a constraint definition may exist for an object, a subtype, a
   --          derived type, or a constraint being built during tree building.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Upper_Bound_Value_Offset
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ace_Natural);
   -- record component name: Upper_Bound_Value_Offset

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Next_Constraint
   --
   --  <Description>
   --
   --    Set the next constraint node associated with the index of an array
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    [none]
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================

   procedure Set_Next_Constraint (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                  Item         : in Ref_Symbol_Table_Entry);
   -- record component name: Next_Constraint

   -- -------------------------------------------------------------------------
   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Help
   --
   --  <Description>
   --
   --    Set the help field of the symbol table entry to the associated
   --    information.
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    none.
   --
   --  <Exceptions Handled>
   --
   --    none.
   --==========================================================================

   procedure Set_Help (St_Entry : in Ref_Symbol_Table_Entry);



   -- pragma inline(Set_Entry_Kind);
   -- pragma inline(Set_Next_Entry);
   -- pragma inline(Set_Symbolic_Name);
   -- pragma inline(Set_Level);
   -- pragma inline(Set_Enclosing_Package);
   -- pragma inline(Make_Entry_Visible);
   -- pragma inline(Make_Entry_Hidden);
   -- pragma inline(Make_Entry_Removed);
   -- pragma inline(Set_Statement_Number);
   -- pragma inline(Set_Context_Clause_Body);
   -- pragma inline(Set_Context_Clause);
   -- pragma inline(Set_Package_Spec_Decls);
   -- pragma inline(Set_Package_Body_Decls);
   -- pragma inline(Set_Has_Body);
   -- pragma inline(Set_Head_Of_Statements_List);
   -- pragma inline(Set_Derived_Subprogram_List);
   -- pragma inline(Set_Subprg_Has_Decl);
   -- pragma inline(Set_Parameter_Count);
   -- pragma inline(Set_Number_Of_Default_Parameters);
   -- pragma inline(Set_Head_Of_Parameters_List);
   -- pragma inline(Set_Tail_Of_Parameters_List);
   -- pragma inline(Set_Return_Type);
   -- pragma inline(Set_Head_Of_Declarations_List);
   -- pragma inline(Set_Value_Offset);
   -- pragma inline(Set_As_Built_In_Subprogram);
   -- pragma inline(Set_Derived_Subprogram_Parent);
   -- pragma inline(Set_Function_Entry);
   -- pragma inline(Set_Enumeration_Literal_Entry);
   -- pragma inline(Set_Type);
   -- pragma inline(Set_Constraint);
   -- pragma inline(Set_Mode_To_Default_In);
   -- pragma inline(Set_Mode_To_In);
   -- pragma inline(Set_Mode_To_Out);
   -- pragma inline(Set_Mode_To_In_Out);
   -- pragma inline(Set_Initialization_Code);
   -- pragma inline(Set_Next_Parameter);
   -- pragma inline(Set_Preceded_By_Comma);
   -- pragma inline(Set_Type_State_To_Incomplete);
   -- pragma inline(Set_Type_State_To_Full);
   -- pragma inline(Set_Type_State_To_Private);
   -- pragma inline(Set_Type_State_To_Limited_Private);
   -- pragma inline(Set_Entry_Base_Type);
   -- pragma inline(Set_Literal_Value);
   -- pragma inline(Set_Successor);
   -- pragma inline(Set_Predecessor);
   -- pragma inline(Set_Next_Record_Component);
   -- pragma inline(Set_Simple_Name_Construct);
   -- pragma inline(Set_Type_To_Enumeration_Type);
   -- pragma inline(Set_Type_To_Integer_Type);
   -- pragma inline(Set_Type_To_Real_Type);
   -- pragma inline(Set_Type_To_Record_Type);
   -- pragma inline(Set_Type_To_Array_Type);
   -- pragma inline(Set_Type_To_Access_Type);
   -- pragma inline(Set_Literal_Count);
   -- pragma inline(Set_First_Literal);
   -- pragma inline(Set_Last_Literal);
   -- pragma inline(Set_Lower_Bound);
   -- pragma inline(Set_Upper_Bound);
   -- pragma inline(Set_Record_Component);
   -- pragma inline(Set_As_Constrained);
   -- pragma inline(Set_As_Unconstrained);
   -- pragma inline(Set_Index_Type);
   -- pragma inline(Set_Element_Type);
   -- pragma inline(Set_Lower_Bound);
   -- pragma inline(Set_Lower_Bound_Value_Offset);
   -- pragma inline(Set_Upper_Bound);
   -- pragma inline(Set_Upper_Bound_Value_Offset);
   -- pragma inline(Set_Next_Constraint);
   -- pragma inline(Set_Help);

end Symtbl_Entry_Set_Routines;
