--|$Log:	ace_adt.a,v $
--|Revision 8.0  91/01/17  10:43:28  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Ace_Adt
--
---------------------------------------------------------------
--
--
-- Abstract
--
--  This package provides an interface to the interpreter's
--  treebuilding routines and tree executer routines.
---------------------------------------------------------------

with Pre_Parser;
with Statement_Evaluator;
with Misc_Support_Routines;
with Ace_Universal_Types;
with Error_Handler;
with User_Pkg;
with Misc;
with Lexacts;
with Unchecked_Conversion;

package body Ace_Adt is



   -- The convert routines are used to translate from an interpreted
   -- statement database pointer to a compiled pointer.

   function Convert is new Unchecked_Conversion (Ace_Statement_Database,
                                                 Ref_Statement_Database);
   function Convert is new Unchecked_Conversion (Ref_Statement_Database,
                                                 Ace_Statement_Database);

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Error_Recovery
   --
   --  <Description>
   --
   --    Error_Recovery will clean off the symbol table from anything
   --    that was placed on the symbol table during a compile or
   --    interpret execution.
   --
   --==========================================================================
   procedure Error_Recovery
       (Context : in out Context_Control.Context_Definition) is
      use Error_Handler;
   begin
      Reset_From_Subprogram_Or_Package_Declaration (Context);
   end Error_Recovery;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Compile
   --
   --  <Description>
   --
   --    Compile will take a file and produce a statement database
   --    sub-tree.
   --
   --==========================================================================
   procedure Compile (Context : in out Context_Control.Context_Definition;
                      File    : in File_Type;
                      Tree    : in out Ace_Statement_Database;
                      Error   : out Boolean) is
      use Error_Handler;
      use Ace_Universal_Types;


      Real_Tree     : Ref_Statement_Database;
      Current_File  : File_Type;
      Eol_Flag_Save : Integer;

      procedure Cleanup (Str : String) is
         use Error_Handler;
      begin
         Show_Error (3006, Warning, Str);
         Lexacts.Eol_Flag := Eol_Flag_Save;
         Error            := True;
         -- Show_Error (3014, Warning, "Compile");
         Error_Recovery (Context);
         Misc.Reset_Input;
      exception
         when others =>
            null;
      end Cleanup;

   begin
      Current_File := Current_Ace_Input;
      Set_Ace_Input (File);
      User_Pkg.Disallow_Eof;
      Context_Control.Set_To_Zero_Level (Context);
      begin
         Eol_Flag_Save    := Lexacts.Eol_Flag;
         Lexacts.Eol_Flag := 0;
         Pre_Parser.Pre_Parse (Real_Tree);
         Error            := False;
      exception
         when Constraint_Error =>
            Cleanup ("Constraint_Error");

         when Program_Error    =>
            Cleanup ("Program_Error");

         when Storage_Error    =>
            Cleanup ("Storage_Error");

         when Numeric_Error    =>
            Cleanup ("Numeric_Error");

         when Tasking_Error    =>
            Cleanup ("Storage_Error");

         when Status_Error     =>
            Cleanup ("Status_Error");

         when Mode_Error  =>
            Cleanup ("Mode_Error");

         when Name_Error  =>
            Cleanup ("Name_Error");

         when Use_Error   =>
            Cleanup ("Use_Error");

         when Device_Error     =>
            Cleanup ("Device_Error");

         when End_Error   =>
            Cleanup ("End_Error");

         when Data_Error  =>
            Cleanup ("Data_Error");

         when Layout_Error     =>
            Cleanup ("Layout_Error");

         when Tree_Builder_Error    =>
            Cleanup ("Tree_Builder_Error");

         when Parser_Error     =>
            Cleanup ("Parser_Error");

         when Evaluator_Error  =>
            Cleanup ("Evaluator_Error");

         when Scanner_Error    =>
            Cleanup ("Scanner_Error");

         when Main_Error  =>
            Cleanup ("Main_Error");

         when Semantic_Error   =>
            Cleanup ("Semantic_Error");

         when Duplicate_Symbol_Table_Entry_Error   =>
            Cleanup ("Duplicate_Symbol_Table_Entry_Error");

         when Exit_Statement_Encountered =>
            Cleanup ("Exit_Statement_Encountered");

         when User_Error  =>
            Cleanup ("User_Error");

      when Terminate_Current_Ace   =>
         Cleanup ("Terminate_Current_Ace");

      when Interrupt_Error    =>
         Cleanup ("Interrupt_Error");

      when End_Break     =>
         Cleanup ("End_Break");

      when Super_Terminate  =>
         Cleanup ("Super_Terminate");

      when  End_Ace =>
         Cleanup ("End_Ace");

      when others =>
            -- An unexpected exception was raised.  Someone has
            -- not been maintaining ACE correctly.  All exceptions
            -- that can be pass to this point should be named
            -- explicitly in the exception handlers
            Cleanup ("Unknown Exception (Assertion Error)");
            -- raise ; -- For the developers we should exit the program
            -- i.e., pass the exception out, but for user's
            -- we shouldn't.  [TBD]

      end;
      Context_Control.Reset_From_Zero_Level (Context);
      Misc.Reset_Input;
      Set_Ace_Input (Current_File);
      User_Pkg.Allow_Eof;
      Lexacts.Eol_Flag := Eol_Flag_Save;
      Tree := Convert (Real_Tree);
   end Compile;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Compile
   --
   --  <Description>
   --
   --    Compile will take a string and produce a statement database tree.
   --
   --==========================================================================
   procedure Compile (Context : in out Context_Control.Context_Definition;
                      Str     : in String;
                      Tree    : in out Ace_Statement_Database;
                      Error   : out Boolean) is
      use Error_Handler;
      File : File_Type := 1;
   begin
      Create (File, Out_File);
      Put (File, Str);
      Reset (File, In_File);
      Compile (Context, File, Tree, Error);
      Delete (File);
   end Compile;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Delete
   --
   --  <Description>
   --
   --    Delete will remove a tree from the internal data structures.
   --
   --==========================================================================
   procedure Delete (Tree : in Ace_Statement_Database) is
      D_Tree : Ref_Statement_Database := Convert (Tree);
   begin
      Misc_Support_Routines.Free (D_Tree);
   end Delete;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Interpret
   --
   --  <Description>
   --
   --    Interpret will execute the statement database Tree passed to it.
   --
   --==========================================================================
   procedure Interpret (Context : in out Context_Control.Context_Definition;
                        Tree    : in Ref_Statement_Database;
                        Error   : out Boolean) is
      use Error_Handler;
      use Ace_Universal_Types;

      procedure Cleanup (Str : String) is
         use Error_Handler;
      begin
         Show_Error (3006, Warning, Str);
         -- Show_Error (3014, Warning, "Interpret");
         Context_Control.Reset_From_Zero_Level (Context);
         Error := True;
      end Cleanup;

   begin
      Context_Control.Set_To_Zero_Level (Context);
      Statement_Evaluator.Evaluate_Statement (Context, Tree);
      Context_Control.Reset_From_Zero_Level (Context);
      Error := False;
   exception
      when Constraint_Error    =>
         Cleanup ("Constraint_Error");

      when Program_Error  =>
         Cleanup ("Program_Error");

      when Storage_Error  =>
         Cleanup ("Storage_Error");

      when Numeric_Error  =>
         Cleanup ("Numeric_Error");

      when Tasking_Error  =>
         Cleanup ("Storage_Error");

      when Status_Error   =>
         Cleanup ("Status_Error");

      when Mode_Error     =>
         Cleanup ("Mode_Error");

      when Name_Error     =>
         Cleanup ("Name_Error");

      when Use_Error =>
         Cleanup ("Use_Error");

      when Device_Error   =>
         Cleanup ("Device_Error");

      when End_Error =>
         Cleanup ("End_Error");

      when Data_Error     =>
         Cleanup ("Data_Error");

      when Layout_Error   =>
         Cleanup ("Layout_Error");

      when Tree_Builder_Error  =>
         Cleanup ("Tree_Builder_Error");

      when Parser_Error   =>
         Cleanup ("Parser_Error");

      when Evaluator_Error     =>
         Cleanup ("Evaluator_Error");

      when Scanner_Error  =>
         Cleanup ("Scanner_Error");

      when Main_Error     =>
         Cleanup ("Main_Error");

      when Semantic_Error =>
         Cleanup ("Semantic_Error");

      when Duplicate_Symbol_Table_Entry_Error =>
         Cleanup ("Duplicate_Symbol_Table_Entry_Error");

      when Exit_Statement_Encountered    =>
         Cleanup ("Exit_Statement_Encountered");

      when User_Error     =>
         Cleanup ("User_Error");

      when Terminate_Current_Ace   =>
         Cleanup ("Terminate_Current_Ace");

      when Interrupt_Error    =>
         Cleanup ("Interrupt_Error");

      when End_Break     =>
         Cleanup ("End_Break");

      when Super_Terminate  =>
         Cleanup ("Super_Terminate");

      when  End_Ace =>
         Cleanup ("End_Ace");

      when others    =>
         -- An unexpected exception was raised.  Someone has
         -- not been maintaining ACE correctly.  All exceptions
         -- that can be pass to this point should be named
         -- explicitly in the exception handlers
         Cleanup ("Unknown Exception (Assertion Error)");
         -- raise ; -- For the developers we should exit the program
         -- i.e., pass the exception out, but for user's
         -- we shouldn't.  [TBD]
   end Interpret;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Interpret
   --
   --  <Description>
   --
   --    Interpret will execute the statement database Tree passed to it.
   --
   --==========================================================================
   procedure Interpret (Context : in out Context_Control.Context_Definition;
                        Tree    : in Ace_Statement_Database;
                        Error   : out Boolean) is
      use Error_Handler;
   begin
      Interpret (Context, Convert (Tree), Error);
   end Interpret;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Interpret
   --
   --  <Description>
   --
   --    Interpret will compile and execute the string provided.
   --
   --==========================================================================
   procedure Interpret (Context : in out Context_Control.Context_Definition;
                        Str     : in String;
                        Error   : out Boolean) is

      Tree   : Ace_Statement_Database;
      Status : Boolean := False;

   begin
      Compile (Context, Str, Tree, Status);
      if not Status then
         Interpret (Context, Tree, Status);
         if not Status then
            Delete (Tree);
         end if;
      end if;
      Error := Status;
   end Interpret;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Interpret
   --
   --  <Description>
   --
   --    Interpret will compile and execute the file provided.
   --
   --==========================================================================
   procedure Interpret (Context : in out Context_Control.Context_Definition;
                        File    : in File_Type;
                        Error   : out Boolean) is

      Tree   : Ace_Statement_Database;
      Status : Boolean := False;

   begin
      loop
         exit  when Status or End_Of_File (File);
         Compile (Context, File, Tree, Status);
         if not Status then
            Interpret (Context, Tree, Status);
            if not Status then
               Delete (Tree);
            end if;
         end if;
         Error := Status;
      end loop;
   end Interpret;

   procedure Interpret_File
       (Context   : in out Context_Control.Context_Definition;
        File_Name : in String) is
      File_Ptr : File_Type;
      Status   : Boolean;
   begin
      Open (File_Ptr, In_File, File_Name);
      Interpret (Context, File_Ptr, Status);
      Close (File_Ptr);
   end Interpret_File;


end Ace_Adt;
