--|$Log:	amain.a,v $
--|Revision 8.0  91/01/17  10:43:34  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - ACE
--
---------------------------------------------------------------
--
--
-- Abstract
--
--   This is the driving routine for the ACE prototype.
--   One construct entered by the user is put into tree form
--   by the pre_parser and parsers, and passed to the tree
--   evaluator for execution.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Pre_Parser;
use Pre_Parser;

with Text_Io;
use Text_Io;

with Statement_Database;
use Statement_Database;

with Dump_Routines;
use Dump_Routines;

with Statement_Evaluator;
use Statement_Evaluator;

with Error_Handler;
use Error_Handler;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Misc;
use Misc;

with Misc_Support_Routines;
use Misc_Support_Routines;

with Error_Handler;

with System_Dependent_Routines;

with Ace_Text_Io;

package body Ace is

   procedure Cleanup (Tree          : in out Ref_Statement_Database;
                      Error_Message : String) is
   begin
      if (Error_Message /= "") then
         Error_Handler.Show_Error (3006, Warning, Error_Message);
      end if;
      Error_Handler.Reset_From_Subprogram_Or_Package_Declaration
          (Main_Thread_Context);
      Reset_Input;
      Free (Tree);
   end Cleanup;

   procedure Free_Only_Stmts (Item : in out Ref_Statement_Database) is
   begin
      if Item.Node_Kind = Statement_Node_Type then
         case Item.Statement_Node.Statement_Kind is

            when Object_Declaration         |
                 Type_Declaration           |
                 Subtype_Declaration        |
                 Procedure_Spec_Declaration |
                 Procedure_Declaration      |
                 Function_Spec_Declaration  |
                 Function_Declaration       |
                 Package_Declaration        |
                 Package_Body_Declaration     =>
               null;

            when others   =>
               Free (Item, Entire_Substructure_And_Constants);

         end case;
      end if;
   end Free_Only_Stmts;

   procedure Ace_Main is

      -- Overview:
      --   This is the driving routine for the ACE prototype.
      --   One construct entered by the user is put into tree form
      --   by the pre_parser and parsers, and passed to the tree
      --   evaluator for execution.

      Tree : Ref_Statement_Database := Empty;
      -- Object where a pointer to the constructed tree is stored

   begin

      -- We will loop forever.  User will have to interupt the loop.
   Forever:
      loop
         begin

            -- Notes:
            --   Pre_parse is a procedure that will read a few tokens (using the
            --   scanner) and determine what parser should be
            --   used for the current input construct.
            --
            --   The available parsers are:
            --      1) a statement parser,
            --      2) a declaration parser, and
            --      3) a compilation unit parser.

            begin
               Pre_Parse (Tree);

            exception
               when User_Error    |
                    Program_Error |
                    Parser_Error  |
                    Scanner_Error   =>
                  Cleanup (Tree, "");
            end;


            if not Is_Empty (Tree) then

               if Is_Statement_Dump_On then
                  Dump_Statement_Database
                      (Tree, Standard_Output, Specification);
               end if;

               begin
                  Evaluate_Statement (Main_Thread_Context, Tree);
               exception
                  when User_Error   =>
                     null;
               end;

               Free_Only_Stmts (Tree);

               Tree := null;

            end if;
         exception

            when Constraint_Error   =>
               Cleanup (Tree, "Constraint_Error");

            when Program_Error =>
               Cleanup (Tree, "Program_Error");

            when Storage_Error =>
               Cleanup (Tree, "Storage_Error");

            when Numeric_Error =>
               Cleanup (Tree, "Numeric_Error");

            when Tasking_Error =>
               Cleanup (Tree, "Storage_Error");

            when Status_Error  =>
               Cleanup (Tree, "Status_Error");

            when Mode_Error    =>
               Cleanup (Tree, "Mode_Error");

            when Name_Error    =>
               Cleanup (Tree, "Name_Error");

            when Use_Error     =>
               Cleanup (Tree, "Use_Error");

            when Device_Error  =>
               Cleanup (Tree, "Device_Error");

            when End_Error     =>
               Cleanup (Tree, "End_Error");

            when Data_Error    =>
               Cleanup (Tree, "Data_Error");

            when Layout_Error  =>
               Cleanup (Tree, "Layout_Error");

            when Tree_Builder_Error =>
               Cleanup (Tree, "Tree_Builder_Error");

            when Parser_Error  =>
               Cleanup (Tree, "Parser_Error");

            when Evaluator_Error    =>
               Cleanup (Tree, "Evaluator_Error");

            when Scanner_Error =>
               Cleanup (Tree, "Scanner_Error");

            when Main_Error    =>
               Cleanup (Tree, "Main_Error");

            when Semantic_Error     =>
               Cleanup (Tree, "Semantic_Error");

            when Duplicate_Symbol_Table_Entry_Error     =>
               Cleanup (Tree, "Duplicate_Symbol_Table_Entry_Error");

            when Exit_Statement_Encountered   =>
               Cleanup (Tree, "Exit_Statement_Encountered");

            when User_Error    =>
               Cleanup (Tree, "User_Error");

            when Terminate_Current_Ace   =>
               Cleanup (Tree, "");
               raise Interrupt_Error;

            when Interrupt_Error    =>
               if Ace_Level > 0 then
                  Ace_Text_Io.Put_Line (Ace_Text_Io.Current_Ace_Output,
                                        "Return to previous ACE level");

                  Ace_Level := Ace_Level - 1;
               else
                  Ace_Text_Io.Put_Line (Ace_Text_Io.Current_Ace_Output,
                                        "Aborted");
               end if;
               Cleanup (Tree, "");

            when End_Break     =>
               raise ; -- Handled by outer scope

            when Super_Terminate    | End_Ace =>
               raise ; -- These exception is handled by the top level main

            when others   =>
               -- An unexpected exception was raised.  Someone has
               -- not been maintaining ACE correctly.  All exceptions
               -- that can be pass to this point should be named
               -- explicitly in the exception handlers
               Cleanup (Tree, "Unknown Exception (Assertion Error)");
               raise ; -- For the developers we should exit the program
               -- i.e., pass the exception out, but for user's
               -- we shouldn't.  [TBD]
         end;

      end loop Forever;

   exception

      when End_Break =>
         null;

   end Ace_Main;

end Ace;
