--|$Log:	aprgut.a,v $
--|Revision 8.0  91/01/17  10:43:35  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Ace_Program_Unit
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    This package defines the ACE abstract data types for packages
--    and subprograms that are constructed within the command environment.
--    The types for packages and subprograms are defined; and the operations
--    which may be applied to objects to these types are also defined.
--    These types and operations allow ACE packages and subprograms (Ada code)
--    to be manipulated as data.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Symbol_Table_Routines;
use Symbol_Table_Routines;

with Symtbl_Entry_Get_Routines;
use Symtbl_Entry_Get_Routines;

with Symtbl_Entry_Set_Routines;
use Symtbl_Entry_Set_Routines;

with Declaration_Support_Routines;
use Declaration_Support_Routines;

with Misc_Support_Routines;
use Misc_Support_Routines;

with Storage_Management_Routines;
use Storage_Management_Routines;

with Literals;
use Literals;

with Statement_Database;
use Statement_Database;

with Node_Create_Routines;
with Node_Set_Routines;
with Node_Get_Routines;

with Expdname;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Error_Handler;
with Text_Io;
with Context_Control;

package body Ace_Program_Unit is

   -- ------
   -- Types used for parameter of internal procedure: Insert_Into_Package
   -- ------
   type Pkg_Portion_Type is (Spec_Portion, Body_Portion, Private_Portion);
   type Pkg_Position_Type is (At_Start, At_End);

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create
   --
   --  <Description>
   --
   --    This procedure establishes a new ACE package, and associates this
   --    package with the given ACE package object.  The Name must not clash
   --    with the currently existing set of visible identifiers within ACE.
   --    The mode of the package object is set to the given Program_Unit_Mode.
   --    The Form parameter is currently unused.
   --
   --  <Exceptions Raised>
   --
   --    Name_Error is raised if the Name parameter is the name of an
   --      existing Ada object within ACE.
   --    Use_Error is raised if the maximum number of open ACE package
   --      objects has been reached.
   --
   --  <Exceptions Handled>
   --
   --
   --==========================================================================
   procedure Create (Context     : in out Context_Control.Context_Definition;
                     Ace_Package : in out Package_Type;
                     Mode        : in Program_Unit_Mode := Out_Prog_Unit;
                     Name        : in String            := "";
                     Form        : in String            := "") is

      Package_Name         : Ref_String := Convert_To_Ref_String (Name);
      Package_Symtbl_Entry : Ref_Symbol_Table_Entry;

   begin
      for I in Package_Entries'FIRST + 1 .. Package_Entries'LAST loop
         -- Find an unused compilation unit entry
         if not Is_Open (I) then

            -- Make sure there is not a name clash for this item already
            -- in the symbol table
            if Name_Clash (Context, Package_Name) then
               raise Name_Error;
            end if;

            Context_Control.Set_To_Zero_Level (Context);

            -- Build a package symbol table entry, retain the symbol table
            -- value in the Package_Entries information, and add the newly
            -- created package symbol table entry into the symbol table
            Package_Symtbl_Entry := Create_Symbol_Table_Entry (Package_Name,
                                                               Package_Entry);
            Set_Level (Package_Symtbl_Entry, Get_Current_Level (Context));
            Package_Entries (I).Symtbl_Entry := Package_Symtbl_Entry;
            Add_Entry_To_Symbol_Table (Context, Package_Symtbl_Entry);

            Context_Control.Reset_From_Zero_Level (Context);

            Ace_Package := I;
            return ;
         end if;
      end loop;

      -- reached the limit on number of open package units
      raise Use_Error;

   end Create;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Open
   --
   --  <Description>
   --
   --    This procedure associates an ACE package object with the named
   --    package that must be currently defined within ACE.  The mode of
   --    the package object is set to the given Program_Unit_Mode.
   --    The Form parameter is currently unused.
   --
   --  <Exceptions Raised>
   --
   --    Name_Error is raised if the Name parameter is not the name of an
   --      existing package.
   --    Use_Error is raised if the maximum number of open ACE package
   --      objects has been reached.
   --
   --==========================================================================
   procedure Open (Context     : in Context_Control.Context_Definition;
                   Ace_Package : in out Package_Type;
                   Mode        : in Program_Unit_Mode;
                   Name        : in String;
                   Form        : in String := "") is

      --NOTE: for files, Status_Error is raised if the file is already open
      -- should we check this (can not open the compilation unit twice)?

      Package_Name         : Ref_String := Convert_To_Ref_String (Name);
      Ref_Form_Str         : Ref_String := Convert_To_Ref_String (Form);
      Package_Symtbl_Entry : Ref_Symbol_Table_Entry;
      Enclosing_Pkg        : Ref_Symbol_Table_Entry;

      use Expdname;

   begin
      for I in Package_Entries'FIRST + 1 .. Package_Entries'LAST loop
         -- Find an unused compilation unit entry
         if not Is_Open (I) then

            -- Find the Package name within the current scope
            Handle_Expanded_Name (Context, Package_Name, Enclosing_Pkg);
            if Is_Empty (Enclosing_Pkg) then
               Package_Symtbl_Entry := Find_Symbol_Table_Entry (Context,
                                                                Package_Name);
            elsif Get_Entry_Kind (Enclosing_Pkg) = Package_Entry then
               declare
                  Last_Stmt       : Ref_Statement_Database;
                  Curr_Decl_Stmt  : Ref_Statement_Database :=
                     Get_Package_Spec_Decls (Enclosing_Pkg);
                  Form_Str        : constant String :=
                     Convert_To_String (Ref_Form_Str);
                  Found_Something : Boolean := False;

               begin
                  Find_Matching_Name
                      (Context, Enclosing_Pkg, Package_Name, Form_Str,
                       Package_Symtbl_Entry, Last_Stmt);
                  while not (Is_Empty (Package_Symtbl_Entry) or Found_Something
                     ) loop
                     if Are_Symbolic_Names_Equal
                            (Get_Symbolic_Name (Package_Symtbl_Entry),
                             Package_Name) and
                        Do_Forms_Match (Package_Symtbl_Entry, Form_Str) then
                        Found_Something := True;
                     else
                        Find_Next_Matching_Name (Context, Curr_Decl_Stmt,
                                                 Package_Name, Form_Str,
                                                 Package_Symtbl_Entry);
                     end if;
                  end loop;

               end;
            else
               Error_Handler.Show_Error
                   (4206, Error_Handler.Fatal,
                    Convert_To_String (Get_Symbolic_Name (Enclosing_Pkg)));
            end if;

            if Is_Empty (Package_Symtbl_Entry) or else
               Get_Entry_Kind (Package_Symtbl_Entry) /= Package_Entry then
               raise Name_Error;
            end if;

            -- Assign the symbol table value to the package_entry field
            Package_Entries (I).Symtbl_Entry := Package_Symtbl_Entry;
            Package_Entries (I).Mode         := Mode;
            Ace_Package := I;
            return ;
         end if;
      end loop;

      -- reached the limit on number of open compilation units
      raise Use_Error;

   end Open;

   -- ----------------------- --
   -- START INTERNAL ROUTINES --
   -- ----------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Specific_Subprogram
   --
   --  <Description>
   --
   --    This procedure looks for the specified program name that matches
   --    the subprogram form.  (The form corresponds to the statement number
   --    associated with the subprogram declaration.)  If a match is found,
   --    then the symbol table entry associated with the subprogram is
   --    returned.  Otherwise, null is returned.
   --
   --  <Called By>
   --
   --    Open  (for Subprograms)
   --
   --  <Exceptions Raised>
   --
   --    Name_Error is raised if the Name parameter is not the name of an
   --      existing subprogram or the form does not match.
   --
   --==========================================================================
   procedure Find_Specific_Subprogram
       (Context                 : in Context_Control.Context_Definition;
        Subprogram_Name         : in Ref_String;
        Subprogram_Form         : in Ace_Universal_Types.Ace_Natural;
        Subprogram_Symtbl_Entry : in out Ref_Symbol_Table_Entry) is

      Found : Boolean := False;
   begin
      Subprogram_Symtbl_Entry := Find_Symbol_Table_Entry (Context,
                                                          Subprogram_Name);
      loop
         exit  when Is_Empty (Subprogram_Symtbl_Entry) or Found;

         if ((Get_Entry_Kind (Subprogram_Symtbl_Entry) = Function_Entry or
            Get_Entry_Kind (Subprogram_Symtbl_Entry) = Procedure_Entry)
             and then
            Is_Empty (Get_Derived_Subprogram_Parent (Subprogram_Symtbl_Entry)))
             and
            Get_Statement_Number (Subprogram_Symtbl_Entry) = Subprogram_Form
             then
            Found := True;
         else
            Subprogram_Symtbl_Entry :=
               Find_Next_Symbol_Table_Entry (Subprogram_Symtbl_Entry);
         end if;
      end loop;

      if Is_Empty (Subprogram_Symtbl_Entry) then
         raise Name_Error;
      end if;

   end Find_Specific_Subprogram;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_First_And_Only_Subprogram
   --
   --  <Description>
   --
   --    This procedure looks for the existence of a single subprogram with
   --    the specified program name.  If only a single subprogram exists with
   --    the name, it is returned.  If more than one subprogram exists,
   --    then the specification of each such subprogram and its associated
   --    statement number (the subprogram's "form") is displayed.  If no
   --    subprogram exists by that name, Name_Error is raised.  (For multiple
   --    or no occurrences of the subprogram, null is returned.)
   --
   --  <Called By>
   --
   --    Open  (for Subprograms)
   --
   --  <Exceptions Raised>
   --
   --    Name_Error is raised if the Name parameter is not the name of an
   --      existing subprogram.
   --
   --==========================================================================
   procedure Find_First_And_Only_Subprogram
       (Context                 : in Context_Control.Context_Definition;
        Subprogram_Name         : in Ref_String;
        Subprogram_Symtbl_Entry : in out Ref_Symbol_Table_Entry) is

      Count : Ace_Universal_Types.Ace_Natural := 0;

      Current_Entry, Saved_Entry : Ref_Symbol_Table_Entry;

   begin
      Current_Entry := Find_Symbol_Table_Entry (Context, Subprogram_Name);
      loop
         exit  when Is_Empty (Current_Entry);

         if (Get_Entry_Kind (Current_Entry) = Function_Entry or
            Get_Entry_Kind (Current_Entry) = Procedure_Entry) and then
            Is_Empty (Get_Derived_Subprogram_Parent (Current_Entry)) then
            Count := Count + 1;

            if Count = 2 then
               -- Display warning message about multiple occurrences
               Error_Handler.Show_Error (2018, Error_Handler.Warning, "");
               Text_Io.New_Line;
            end if;

            if Count > 1 then
               -- Display "form" and specification for the prior subprogram
               Text_Io.Put_Line
                   ("-- FORM ==> " &
                    Ace_Universal_Types.Ace_Natural'IMAGE (
                       Get_Statement_Number (Saved_Entry)));

               -- [TBD] Fix final parameter when Dump_Symtbl routine is revised
               Dump_Routines.Dump_Symtbl (Saved_Entry, Text_Io.Current_Output);
            end if;

            Saved_Entry := Current_Entry;
         end if;

         Current_Entry := Find_Next_Symbol_Table_Entry (Current_Entry);
      end loop;

      if Count > 1 then
         -- Display information about last entry found
         Text_Io.Put_Line
             ("-- FORM ==> " &
              Ace_Universal_Types.Ace_Natural'IMAGE (
                 Get_Statement_Number (Saved_Entry)));

         -- [TBD] Fix final parameter when Dump_Symtbl routine is revised
         Dump_Routines.Dump_Symtbl (Saved_Entry, Text_Io.Current_Output);

         Subprogram_Symtbl_Entry := null;
         raise Name_Error;
      elsif Count < 1 then
         raise Name_Error;
      else
         Subprogram_Symtbl_Entry := Saved_Entry;
      end if;

   end Find_First_And_Only_Subprogram;

   -- --------------------- --
   -- END INTERNAL ROUTINES --
   -- --------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Open
   --
   --  <Description>
   --
   --    This procedure associates an ACE subprogram object with the named
   --    subprogram that must be currently defined within ACE.  The mode of
   --    the subprogram object is set to the given Program_Unit_Mode.
   --    The Form parameter is currently unused.
   --
   --  <Exceptions Raised>
   --
   --    Name_Error is raised if the Name parameter is not the name of an
   --      existing subprogram.
   --    Use_Error is raised if the maximum number of open ACE subprogram
   --      objects has been reached.
   --    Constraint_Error is raised if the Form parameter is ill-formed
   --
   --==========================================================================
   procedure Open (Context        : in out Context_Control.Context_Definition;
                   Ace_Subprogram : in out Subprogram_Type;
                   Mode           : in Program_Unit_Mode;
                   Name           : in String;
                   Form           : in String := "") is

      --NOTE: for files, Status_Error is raised if the file is already open
      -- should we check this (can not open the compilation unit twice)?

      Subprogram_Name         : Ref_String := Convert_To_Ref_String (Name);
      Ref_Form_Str            : Ref_String := Convert_To_Ref_String (Form);
      Subprogram_Symtbl_Entry : Ref_Symbol_Table_Entry;
      Enclosing_Pkg           : Ref_Symbol_Table_Entry;
      Subprogram_Form         : Ace_Universal_Types.Ace_Natural;
      --    needed until default parameters are handled
      use Expdname;
   begin
      if Form = "" then
         Subprogram_Form := 0;
      else
         Subprogram_Form := Ace_Universal_Types.Ace_Natural'VALUE (Form);
      end if;

      Handle_Expanded_Name (Context, Subprogram_Name, Enclosing_Pkg);
      if Is_Empty (Enclosing_Pkg) then
         for I in Subprogram_Entries'FIRST + 1 .. Subprogram_Entries'LAST loop
            -- Find an unused compilation unit entry
            if not Is_Open (I) then
               -- Find the appropriate subprogram, ignoring derived subprograms

               if Subprogram_Form /= 0 then
                  Find_Specific_Subprogram (Context, Subprogram_Name,
                                            Subprogram_Form,
                                            Subprogram_Symtbl_Entry);
               else
                  Find_First_And_Only_Subprogram (Context, Subprogram_Name,
                                                  Subprogram_Symtbl_Entry);
               end if;

               if not Is_Empty (Subprogram_Symtbl_Entry) then
                  -- Retain a reference to the the subprogram symbol table entry
                  Subprogram_Entries (I).Symtbl_Entry :=
                     Subprogram_Symtbl_Entry;
                  Subprogram_Entries (I).Mode         := Mode;
                  Ace_Subprogram := I;
                  return ;
               end if;
            end if;
         end loop;
      elsif Get_Entry_Kind (Enclosing_Pkg) = Package_Entry then
         declare
            Last_Stmt       : Ref_Statement_Database;
            Curr_Decl_Stmt  : Ref_Statement_Database :=
               Get_Package_Spec_Decls (Enclosing_Pkg);
            Form_Str : constant String := Convert_To_String (Ref_Form_Str);
            Found_Something : Boolean := False;

         begin
            for I in Subprogram_Entries'FIRST + 1 .. Subprogram_Entries'LAST
             loop
               -- Find an unused compilation unit entry
               if not Is_Open (I) then
                  Find_Matching_Name (Context, Enclosing_Pkg, Subprogram_Name,
                                      Form_Str, Subprogram_Symtbl_Entry,
                                      Last_Stmt);
                  while not (Is_Empty (Subprogram_Symtbl_Entry) or
                     Found_Something) loop
                     if Are_Symbolic_Names_Equal
                            (Get_Symbolic_Name (Subprogram_Symtbl_Entry),
                             Subprogram_Name) and
                        Do_Forms_Match (Subprogram_Symtbl_Entry, Form_Str) then
                        Found_Something := True;
                     else
                        Find_Next_Matching_Name (Context, Curr_Decl_Stmt,
                                                 Subprogram_Name, Form_Str,
                                                 Subprogram_Symtbl_Entry);
                     end if;
                  end loop;
                  if not Is_Empty (Subprogram_Symtbl_Entry) then
                  -- Retain a reference to the the subprogram symbol table entry
                     Subprogram_Entries (I).Symtbl_Entry :=
                        Subprogram_Symtbl_Entry;
                     Subprogram_Entries (I).Mode         := Mode;
                     Ace_Subprogram := I;
                     return ;
                  end if;
               end if;
            end loop;

         end;
      else
         Error_Handler.Show_Error
             (4206, Error_Handler.Fatal,
              Convert_To_String (Get_Symbolic_Name (Enclosing_Pkg)));
      end if;

      -- reached the limit on number of open compilation units
      raise Use_Error;

   end Open;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Close
   --
   --  <Description>
   --
   --    This procedure closes an ACE package object.  This terminates the
   --    association between the ACE package object and the actual package.
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the package object is not open.
   --
   --==========================================================================
   procedure Close (Ace_Package : in out Package_Type) is
   begin
      if Is_Open (Ace_Package) then
         Package_Entries (Ace_Package).Symtbl_Entry := null;
         Ace_Package := Package_Type'FIRST;
      else
         raise Status_Error;
      end if;
   end Close;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Close
   --
   --  <Description>
   --
   --    This procedure closes an ACE subprogram object.  This terminates
   --    the association between the ACE subprogram object and the actual
   --    subprogram.
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the subprogram object is not open.
   --
   --==========================================================================
   procedure Close (Ace_Subprogram : in out Subprogram_Type) is
   begin
      if Is_Open (Ace_Subprogram) then
         Subprogram_Entries (Ace_Subprogram).Symtbl_Entry := null;
         Ace_Subprogram := Subprogram_Type'FIRST;
      else
         raise Status_Error;
      end if;
   end Close;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Delete
   --
   --  <Description>
   --
   --    This procedure deletes an ACE package object.  The deletion is
   --    accomplished by marking the symbol table entry associated with the
   --    object as "Removed".
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the package object is not open.
   --
   --==========================================================================
   procedure Delete (Ace_Package : in out Package_Type) is
   begin
      if Is_Open (Ace_Package) then
         Make_Entry_Removed (Package_Entries (Ace_Package).Symtbl_Entry);
      else
         raise Status_Error;
      end if;

      Close (Ace_Package);
   end Delete;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Delete
   --
   --  <Description>
   --
   --    This procedure deletes an ACE subprogram object.  The deletion is
   --    accomplished by marking the symbol table entry associated with the
   --    object as "Removed".
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the subprogram object is not open.
   --
   --==========================================================================
   procedure Delete (Ace_Subprogram : in out Subprogram_Type) is
   begin
      if Is_Open (Ace_Subprogram) then
         Make_Entry_Removed (Subprogram_Entries (Ace_Subprogram).Symtbl_Entry);
      else
         raise Status_Error;
      end if;

      Close (Ace_Subprogram);
   end Delete;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Mode
   --
   --  <Description>
   --
   --    This function returns the program unit mode associated with an
   --    open ACE package object.
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the package object is not open.
   --
   --==========================================================================
   function Mode (Ace_Package : in Package_Type) return Program_Unit_Mode is
   begin
      if Is_Open (Ace_Package) then
         return Package_Entries (Ace_Package).Mode;
      else
         raise Status_Error;
      end if;
   end Mode;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Mode
   --
   --  <Description>
   --
   --    This function returns the program unit mode associated with an
   --    open ACE subprogram object.
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the subprogram object is not open.
   --
   --==========================================================================
   function Mode (Ace_Subprogram : in Subprogram_Type) return Program_Unit_Mode
                 is
   begin
      if Is_Open (Ace_Subprogram) then
         return Subprogram_Entries (Ace_Subprogram).Mode;
      else
         raise Status_Error;
      end if;
   end Mode;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Name
   --
   --  <Description>
   --
   --    This function returns the name associated with an ACE package
   --    object.
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the package object is not open.
   --
   --==========================================================================
   function Name (Ace_Package : in Package_Type) return String is

   begin
      if Is_Open (Ace_Package) then
         return
            Convert_To_String
                (Get_Symbolic_Name (Package_Entries (Ace_Package).Symtbl_Entry)
                 );
      else
         raise Status_Error;
      end if;
   end Name;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Name
   --
   --  <Description>
   --
   --    This function returns the name associated with an ACE subprogram
   --    object.
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the subprogram object is not open.
   --
   --==========================================================================
   function Name (Ace_Subprogram : in Subprogram_Type) return String is

   begin
      if Is_Open (Ace_Subprogram) then
         return
            Convert_To_String
                (Get_Symbolic_Name
                     (Subprogram_Entries (Ace_Subprogram).Symtbl_Entry));
      else
         raise Status_Error;
      end if;
   end Name;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Form
   --
   --  <Description>
   --
   --    This function returns the form string associated with an ACE package
   --    object.
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the package object is not open.
   --
   --==========================================================================
   function Form (Ace_Package : in Package_Type) return String is

      Stmt_No : Ace_Universal_Types.Ace_Natural;

   begin
      if Is_Open (Ace_Package) then
         Stmt_No :=
            Get_Statement_Number (Package_Entries (Ace_Package).Symtbl_Entry);
         return Ace_Universal_Types.Ace_Natural'IMAGE (Stmt_No);
      else
         raise Status_Error;
      end if;
   end Form;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Form
   --
   --  <Description>
   --
   --    This function returns the form string associated with an ACE
   --    subprogram object.
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the subprogram object is not open.
   --
   --==========================================================================
   function Form (Ace_Subprogram : in Subprogram_Type) return String is

      Stmt_No : Ace_Universal_Types.Ace_Natural;

   begin
      if Is_Open (Ace_Subprogram) then
         Stmt_No :=
            Get_Statement_Number
                (Subprogram_Entries (Ace_Subprogram).Symtbl_Entry);
         return Ace_Universal_Types.Ace_Natural'IMAGE (Stmt_No);
      else
         raise Status_Error;
      end if;
   end Form;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Open
   --
   --  <Description>
   --
   --    This function returns an indication of whether the ACE package
   --    object is open.  True is returned when the object is open; otherwise
   --    False is returned.
   --
   --==========================================================================
   function Is_Open (Ace_Package : in Package_Type) return Boolean is
   begin
      return not Is_Empty (Package_Entries (Ace_Package).Symtbl_Entry);
   end Is_Open;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Open
   --
   --  <Description>
   --
   --    This function returns an indication of whether the ACE subprogram
   --    object is open.  True is returned when the object is open; otherwise
   --    False is returned.
   --
   --==========================================================================
   function Is_Open (Ace_Subprogram : in Subprogram_Type) return Boolean is
   begin
      return not Is_Empty (Subprogram_Entries (Ace_Subprogram).Symtbl_Entry);
   end Is_Open;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Build_Subprogram_Spec_Decl
   --
   --  <Description>
   --
   --    This procedures constructs a subprogram specification declaration
   --     node (an ACEstatement database node) for the subprogram associated
   --     with the ACE subprogram object.
   --
   --  <Called By>
   --
   --    Move
   --
   --==========================================================================
   -- INTERNAL ROUTINE --

   function Build_Subprogram_Spec_Decl
       (Subprogram_Index : in Subprogram_Type;
        Package_Index    : in Package_Type) return Ref_Statement_Database is

      Decl_Node               : Ref_Statement_Database;
      Subprogram_Symtbl_Entry : Ref_Symbol_Table_Entry;

   begin
      -- Create the Procedure or Function Declaration Node

      Subprogram_Symtbl_Entry :=
         Subprogram_Entries (Subprogram_Index).Symtbl_Entry;
      case Get_Entry_Kind (Subprogram_Symtbl_Entry) is
         when Procedure_Entry  =>
            Decl_Node :=
               Node_Create_Routines.Create_Node (Procedure_Spec_Declaration);
         when Function_Entry   =>
            Decl_Node :=
               Node_Create_Routines.Create_Node (Function_Spec_Declaration);
         when others =>
            -- error
            null;
      end case;

      -- Set the subprogram to refer to the package in which it will now
      -- be enclosed within
      Set_Enclosing_Package (Subprogram_Symtbl_Entry,
                             Package_Entries (Package_Index).Symtbl_Entry);

      Node_Set_Routines.Set_Symbol_Table_Entry (Decl_Node,
                                                Subprogram_Symtbl_Entry);
      return Decl_Node;
   end Build_Subprogram_Spec_Decl;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Build_Subprogram_Decl
   --
   --  <Description>
   --
   --    This procedures constructs a subprogram declaration node (an ACE
   --    statement database node) for the subprogram associated with the ACE
   --    subprogram object.
   --
   --  <Called By>
   --
   --    Move
   --
   --==========================================================================
   -- INTERNAL ROUTINE --

   function Build_Subprogram_Decl
       (Subprogram_Index : in Subprogram_Type;
        Package_Index    : in Package_Type) return Ref_Statement_Database is

      Decl_Node               : Ref_Statement_Database;
      Subprogram_Symtbl_Entry : Ref_Symbol_Table_Entry;

   begin
      -- Create the Procedure or Function Declaration Node

      Subprogram_Symtbl_Entry :=
         Subprogram_Entries (Subprogram_Index).Symtbl_Entry;
      case Get_Entry_Kind (Subprogram_Symtbl_Entry) is
         when Procedure_Entry  =>
            Decl_Node :=
               Node_Create_Routines.Create_Node (Procedure_Declaration);
         when Function_Entry   =>
            Decl_Node :=
               Node_Create_Routines.Create_Node (Function_Declaration);
         when others =>
            -- error
            null;
      end case;

      -- Set the subprogram to refer to the package in which it will now
      -- be enclosed within
      Set_Enclosing_Package (Subprogram_Symtbl_Entry,
                             Package_Entries (Package_Index).Symtbl_Entry);

      Node_Set_Routines.Set_Symbol_Table_Entry (Decl_Node,
                                                Subprogram_Symtbl_Entry);
      return Decl_Node;
   end Build_Subprogram_Decl;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Insert_Into_Package
   --
   --  <Description>
   --
   --    This procedures appends a subprogram declaration node (an ACE
   --    statement database node) onto the symbol table entry associated with
   --    the ACE package object.  The Pkg_Portion parameter (Spec, Body, or
   --    Private) controls the location of the package in which to insert the
   --    subprogram declaration.  The Pkg_Position parameter (At_Start or
   --    At_End) controls the position within the respective spec or body in
   --    which to place the subprogram.  (For example, a subprogram that is
   --    local to the package body will be inserted at the head of the package
   --    body.)
   --
   --    The declarations of Pkg_Portion_Type and Pkg_Position_Type are located
   --    at the beginning of this package body.
   --
   --  <Called By>
   --
   --    Move
   --
   --==========================================================================
   -- INTERNAL ROUTINE --

   procedure Insert_Into_Package (Package_Index   : in Package_Type;
                                  Pkg_Portion     : in Pkg_Portion_Type;
                                  Pkg_Position    : in Pkg_Position_Type;
                                  Subprogram_Decl : in Ref_Statement_Database)
                                 is

      Package_Symtbl_Entry : Ref_Symbol_Table_Entry;
      Current_Stmt         : Ref_Statement_Database;
      Previous_Stmt        : Ref_Statement_Database;

   begin
      Package_Symtbl_Entry := Package_Entries (Package_Index).Symtbl_Entry;

      -- Acquire the head of the statement list for the appropriate
      -- portion of the package
      if Pkg_Portion = Spec_Portion then
         Current_Stmt := Get_Package_Spec_Decls (Package_Symtbl_Entry);
      elsif Pkg_Portion = Body_Portion then
         Current_Stmt := Get_Package_Body_Decls (Package_Symtbl_Entry);
      else
         null; -- [TBD] Private not yet implemented
      end if;

      -- Insertion is based upon the position in which the subprogram is to
      -- be placed within the respective portion of the package
      if Pkg_Position = At_Start then
         Node_Set_Routines.Set_Next_Statement (Subprogram_Decl, Current_Stmt);
         if Pkg_Portion = Spec_Portion then
            Set_Package_Spec_Decls (Package_Symtbl_Entry, Subprogram_Decl);
         elsif Pkg_Portion = Body_Portion then
            Set_Package_Body_Decls (Package_Symtbl_Entry, Subprogram_Decl);
         else
            null; -- [TBD] Private not yet implemented
         end if;

      else -- Pkg_Position = At_End

         -- Locate the current end of the statement list for the appropriate
         -- portion of the package
         Previous_Stmt := null;
         while not Is_Empty (Current_Stmt) loop
            Previous_Stmt := Current_Stmt;
            Current_Stmt := Node_Get_Routines.Get_Next_Statement (Current_Stmt);
         end loop;

         -- Link the subprogram declaration node onto the end of the statement
         -- list for the appropriate portion of the package.  (Different method
         -- must be employed in the list is currently empty.)
         if not Is_Empty (Previous_Stmt) then
            Node_Set_Routines.Set_Next_Statement (Previous_Stmt,
                                                  Subprogram_Decl);
         else
            if Pkg_Portion = Spec_Portion then
               Set_Package_Spec_Decls (Package_Symtbl_Entry, Subprogram_Decl);
            elsif Pkg_Portion = Body_Portion then
               Set_Package_Body_Decls (Package_Symtbl_Entry, Subprogram_Decl);
            else
               null; -- [TBD] setting for private
            end if;
         end if;
      end if;

      -- If the item was inserted into the body portion of the package,
      -- mark the package as now having a body
      if Pkg_Portion = Body_Portion then
         Set_Has_Body (Package_Symtbl_Entry);
      end if;

   end Insert_Into_Package;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Put
   --
   --  <Description>
   --
   --    This procedures moves an ACE subprogram object into an ACE
   --    package object.  The parameters Action (specification or
   --    implementation) and Visibility (visible or hidden) control the
   --    location within the package into which the subprogram is inserted.
   --    The location for a package includes the package specification
   --    (visible and private), and the package body.
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if either the subprogram object or package
   --      object is not open.
   --    Mode_Error is raised if the mode of the package object does permit
   --      writing to the package.
   --
   --==========================================================================
   procedure Put (Context         : in out Context_Control.Context_Definition;
                  Subprogram_Item : in Subprogram_Type;
                  Into_Package    : in Package_Type;
                  Action          : in Action_Type;
                  Visibility      : in Visibility_Type) is

      -- Location of where the subprogram is to be inserted into the package
      -- based upon the Action and Visiblity values.
      --
      --         \             VISIBILITY
      --          \    Visible             Hidden
      --  ACTION   \
      --            |--------------------------------------------|
      --    Spec    |    spec      |     private portion of spec |
      --            |--------------------------------------------|
      --    Impl    | spec & body  |       body only             |
      --            |--------------------------------------------|

      Decl_Node : Ref_Statement_Database;

   begin
      -- Both the subprogram Item to be inserted and the package Into which
      -- it is to be inserted must be open.
      if not Is_Open (Subprogram_Item) then
         raise Status_Error;
      elsif not Is_Open (Into_Package) then
         raise Status_Error;
      end if;

      -- Mode of the package must permit writing to the package.
      if Mode (Into_Package) = In_Prog_Unit then
         raise Mode_Error;
      end if;

      -- Perform actions based upon the values of Action and Visibility
      case Action is
         when Specification    =>
            case Visibility is
               when Visible    =>
                  Decl_Node := Build_Subprogram_Spec_Decl (Subprogram_Item,
                                                           Into_Package);
                  -- Insert subprogram spec decl at the end of package spec
                  Insert_Into_Package (Into_Package, Spec_Portion, At_End,
                                       Decl_Node);

               when Hidden     =>
                  -- [TBD] Private portion of spec not yet implemented
                  null;
            end case;

         when Implementation   =>
            case Visibility is
               when Visible    =>
                  Decl_Node := Build_Subprogram_Spec_Decl (Subprogram_Item,
                                                           Into_Package);
                  -- Insert subprogram spec decl at the end of package spec; and
                  -- Insert subprogram body decl ar the end of package body
                  -- (NOTE: must create a different decl node for the second
                  --        insert)
                  Insert_Into_Package (Into_Package, Spec_Portion, At_End,
                                       Decl_Node);
                  Decl_Node := Build_Subprogram_Decl (Subprogram_Item,
                                                      Into_Package);
                  Insert_Into_Package (Into_Package, Body_Portion, At_End,
                                       Decl_Node);

               when Hidden     =>
                  Decl_Node := Build_Subprogram_Decl (Subprogram_Item,
                                                      Into_Package);
                  -- Insert subprogram body decl at the start of package body
                  Insert_Into_Package (Into_Package, Body_Portion, At_Start,
                                       Decl_Node);

            end case;
      end case;

      -- Remove the subprogram from the symbol table (since it is now
      -- within a package)

      Delete_Entry_From_Symbol_Table
          (Context, Subprogram_Entries (Subprogram_Item).Symtbl_Entry);

      -- super-kludge... but we don't want the entry to really be deleted
      --   just off of the runtime-stack... [raf]
      Make_Entry_Visible (Subprogram_Entries (Subprogram_Item).Symtbl_Entry);

      -- If the package is "use"d, add the subprogram entry onto the
      -- symbol table as if it occurred during execution of the use for
      -- the package [TBD]
      -- Currently, no fool-proof method for determining whether a package
      -- has been "use"d.

   end Put;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    List
   --
   --  <Description>
   --
   --    This procedure displays an ACE package object.  The Ada code
   --    description is the medium in which the package is displayed.
   --    The list method controls whether the specification, the body,
   --    or both of these are displayed.
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the package object is not open.
   --
   --==========================================================================
   procedure List (Ace_Package : in Package_Type;
                   List_Method : in List_Mode := Both) is

   begin
      if Is_Open (Ace_Package) then
         -- [TBD]
         -- Must revise when Dump_Symtbl is redone
         --         Dump_Routines.Dump_Symtbl
         --             (Package_Entries (Ace_Package).Symtbl_Entry,
         --              Text_Io.Current_Output, List_Method);

         case List_Method is
            when Specification =>
               Dump_Routines.Dump_Symtbl
                   (Package_Entries (Ace_Package).Symtbl_Entry,
                    Text_Io.Current_Output, Specification);
            when Implementation     =>
               -- [TBD]
               -- Current version of Dump_Symtbl does not support dumping of
               -- just the body associated with a package.
               -- Can get either the spec, or the spec and the body.
               -- For now, dump the spec and body
               Dump_Routines.Dump_Symtbl
                   (Package_Entries (Ace_Package).Symtbl_Entry,
                    Text_Io.Current_Output, Implementation);
            when Both     =>
               Dump_Routines.Dump_Symtbl
                   (Package_Entries (Ace_Package).Symtbl_Entry,
                    Text_Io.Current_Output, Both);
         end case;
      else
         raise Status_Error;
      end if;
   end List;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    List
   --
   --  <Description>
   --
   --    This procedure displays an ACE subprogram object.  The Ada code
   --    description is the medium in which the subprogram is displayed.
   --    The list method controls whether the specification, the body,
   --    or both of these are displayed.
   --
   --  <Exceptions Raised>
   --
   --    Status_Error is raised if the subprogram object is not open.
   --
   --==========================================================================
   procedure List (Ace_Subprogram : in Subprogram_Type;
                   List_Method    : in List_Mode := Both) is

   begin
      if Is_Open (Ace_Subprogram) then
         -- [TBD]
         -- Must revise when Dump_Symtbl is redone
         case List_Method is
            when Specification =>
               Dump_Routines.Dump_Symtbl
                   (Subprogram_Entries (Ace_Subprogram).Symtbl_Entry,
                    Text_Io.Current_Output, Specification);

            when Implementation     =>
               Dump_Routines.Dump_Symtbl
                   (Subprogram_Entries (Ace_Subprogram).Symtbl_Entry,
                    Text_Io.Current_Output, Implementation);
            when Both     =>
               -- [TBD]
               -- Current version of Dump_Symtbl does not support dumping of
               -- both the spec & the body of a function.
               -- Can get either the spec or the body.
               -- For now dump the implementation
               Dump_Routines.Dump_Symtbl
                   (Subprogram_Entries (Ace_Subprogram).Symtbl_Entry,
                    Text_Io.Current_Output, Implementation);
         end case;
      else
         raise Status_Error;
      end if;
   end List;

end Ace_Program_Unit;
