--|$Log:	astd.a,v $
--|Revision 8.0  91/01/17  10:43:37  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Ace_Standard
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    This package defines the standard environment manipulation operations
--    that are applicable to objects defined within ACE.
--    The user is automatically given access to these operations
--    when the ACE system elaborates itself.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Misc_Support_Routines;
use Misc_Support_Routines;

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Symbol_Table_Routines;
use Symbol_Table_Routines;

with Symtbl_Entry_Get_Routines;
use Symtbl_Entry_Get_Routines;

with Symtbl_Entry_Set_Routines;
use Symtbl_Entry_Set_Routines;

with Error_Handler;

package body Ace_Standard is

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Delete
   --
   --  <Description>
   --
   --    This procedure deletes the named ACE object.  The deletion is
   --    accomplished by marking the symbol table entry associated with the
   --    object as "Removed".
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Name_Error is raised if no object exists with the given name
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   procedure Delete (Context     : in out Context_Control.Context_Definition;
                     Object_Name : in Ref_String;
                     Form        : String) is

      Object_Symtbl_Entry : Ref_Symbol_Table_Entry;
      Deleted_Something   : Boolean := False;
   begin
      -- The object, if it exists, will have its visibility field set
      -- to "Removed" (rather than physically removing it from the symbol
      -- table).
      Object_Symtbl_Entry := Find_Symbol_Table_Entry (Context, Object_Name);
      if Is_Empty (Object_Symtbl_Entry) then
         raise Name_Error;
      elsif Form = "" then
         Make_Entry_Removed (Object_Symtbl_Entry);
      else
         loop
            if (Form = "all") or else
               (Integer'VALUE (Form) =
               Integer (Get_Statement_Number (Object_Symtbl_Entry))) then
               -- Found an object to delete
               Make_Entry_Removed (Object_Symtbl_Entry);
               Unhide_Entry (Context, Object_Symtbl_Entry);
               Deleted_Something := True;
               if (Form /= "all") then
                  exit ;
               end if;
            end if;
            Object_Symtbl_Entry :=
               Find_Next_Symbol_Table_Entry (Object_Symtbl_Entry);
            exit  when Is_Empty (Object_Symtbl_Entry);
            -- no more objects on the symbol table.
         end loop;
         if (not Deleted_Something) then
            raise Name_Error;
         end if;
      end if;
   end Delete;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Defuse
   --
   --  <Description>
   --
   --    This procedure deletes an ACE subprogram object.  The deletion is
   --    accomplished by marking the symbol table entry associated with the
   --    object as "Removed".
   --
   --  <Called By>
   --
   --    External
   --
   --  <Exceptions Raised>
   --
   --    Name_Error is raised if the object does not exist or the object is
   --      not a package.
   --
   --  <Exceptions Handled>
   --
   --    [none]
   --==========================================================================

   procedure Defuse (Context      : in out Context_Control.Context_Definition;
                     Package_Name : in Ref_String;
                     Form         : String) is

      -- Form is not used.

      Package_Symtbl_Entry : Ref_Symbol_Table_Entry;

   begin
      Package_Symtbl_Entry := Find_Symbol_Table_Entry (Context, Package_Name);
      if Is_Empty (Package_Symtbl_Entry) or else
         Get_Entry_Kind (Package_Symtbl_Entry) /= Package_Entry then
         raise Name_Error;
      end if;
      Delete_Package_Decls_From_Symbol_Table (Context, Package_Symtbl_Entry);
   exception
      when Constraint_Error    =>
         Error_Handler.Show_Error (4205, Error_Handler.Fatal,
                                   Convert_To_String (Package_Name));
   end Defuse;

end Ace_Standard;
