--|$Log:	cexpr.a,v $
--|Revision 8.0  91/01/17  10:43:43  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Subunit for Package - Semantic_Analyzer
--
---------------------------------------------------------------
--
--
-- Abstract
--
--  Separate unit for Semantic_Analyzer
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------
--==========================================================================
--
--  <Subprogram Name>
--
--   Check_Expression
--
--  <Description>
--
--   Checks the given expression node structure against correct Ada
--   semantics.
--   Resolves references to symbol table entries.
--
--==========================================================================
with Node_Set_Routines;
use Node_Set_Routines;

with Storage_Management_Routines;
use Storage_Management_Routines;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Symbol_Table_Routines;

separate (Semantic_Analyzer)

procedure Check_Expression (Context : in out Context_Control.Context_Definition;
                            Node       : in Ref_Statement_Database;
                            Entry_Kind : Type_Entry_Enumeration := Unknown) is

   Subprogram_Name                 : constant String := "Check_Expression";

   Entry_Kind_Actual               : Type_Entry_Enumeration := Entry_Kind;
   Handle_Function_Entries_Special : Boolean                := True;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Check_Attribute
   --
   --  <Description>
   --
   --   Checks the given attribute node structure against
   --   correct Ada semantics.
   --
   --==========================================================================
   procedure Check_Attribute (Node : in Ref_Statement_Database) is separate;

   procedure Check_Dot_Notation (Node : in Ref_Statement_Database) is separate;

   procedure Check_Array_Reference (Node : in Ref_Statement_Database) is separate;

begin
   -- Notes in general:
   --    - set node type when/if appropriate
   --    - check for correct usage of MODEs of items

   case Check_Expr_Node (Node) is

      when Terminal_Node  =>
         -- assert (Is_Empty (Get_Symbol_Table_Entry (Node)))

         if Entry_Kind_Actual = Function_Entry then
            Handle_Function_Entries_Special := False;
            Entry_Kind_Actual               := Unknown;
         end if;

         -- symbol named should be a previously declared identifier
         Set_Symbol_Table_Entry
             (Node,
              Find_Symbol_Table_Entry (Context, Get_Symbolic_Name (Node),
                                       Entry_Kind_Actual));
         if (Is_Empty (Get_Symbol_Table_Entry (Node))) then
            -- symbol table entry was _not_ found
            -- symbol is undeclared
            Show_Error (1100, Fatal,
                        Convert_To_String (Get_Symbolic_Name (Node)));

         else -- symbol table entry was found

            case Get_Entry_Kind (Get_Symbol_Table_Entry (Node)) is
               when Function_Entry  =>
                  -- check if terminal node is really ...
                  --   a parameterless function call
                  if Handle_Function_Entries_Special then
                     Set_Node_Kind (Node, Function_Call);
                  end if;
                  -- check if function call node is *ultimately* ...
                  --    an enumeration literal entry
                  if (Check_Expr_Node (Node) = Function_Call) and then
                     (not Is_Empty
                         (Get_Enumeration_Literal_Entry
                              (Get_Symbol_Table_Entry
                                   (Get_Subprogram_Info (Node))))) then
                     -- is a enumeration literal entry dressed as
                     -- a parameterless function call
                     --== Set_As_Universal_Type (Node);
                     -- set node_type (good as place as any)
                     if Is_Empty (Get_Type (Node)) then
                        Set_Type
                            (Node,
                             Get_Return_Type
                                 (Get_Symbol_Table_Entry
                                      (Get_Subprogram_Info (Node))));
                     end if;
                  end if;
                  if Handle_Function_Entries_Special then
                     Check_Subprogram_Call (Context, Node);
                  end if;
                  -- when Enumeration_Literal_Entry =>
               when Object_Entry       |
                    Formal_Param_Entry |
                    Record_Component_Entry    =>
                  -- set node_type for non-functions
                  Set_Type (Node, Get_Type (Get_Symbol_Table_Entry (Node)));
               when others     =>
                  -- only Derived_Entry and Subtype_Entry have types,
                  -- associated with them; the others are absolutely invalid
                  -- so ignore them
                  null;
            end case;
         end if;

      when Constant_Node  =>
         -- constant is a numeric literal
         begin
         Set_Value
             (Node,
              Create_Storage_For_Integer
                  (Ace_Integer'VALUE (
                      Convert_To_String (Get_Symbolic_Name (Node)))));
         exception
           when Constraint_Error =>
              Show_Error (3016, Fatal, "Numeric Type");
         end;
         Set_Type (Node, Symbol_Table_Routines.Integer_Entry);
         Set_As_Universal_Type (Node);

      when String_Node    =>
         -- it's a string literal
         Set_Value (Node, Create_Storage_For_String (Get_Symbolic_Name (Node)));
         Set_Type (Node, Symbol_Table_Routines.String_Entry);
         Set_As_Universal_Type (Node);

      when Character_Node =>
         -- it's a character literal
         declare
            Str : constant String :=
               Convert_To_String (Get_Symbolic_Name (Node));
         begin
            Set_Value (Node, Create_Storage_For_Character (Str (Str'FIRST)));
            Set_Type (Node, Symbol_Table_Routines.Character_Entry);
            Set_As_Universal_Type (Node);
         end;

      when Left_Parenthesis_Operator     =>
         -- assert (not Is_Empty (Get_Object (Node)))
         -- it's either an array reference, type conversion,
         --    or function call (built-in or user-defined)

         -----
         -- check if it is a function designator
         if (Check_Expr_Node (Get_Object (Node)) = String_Node) then
            -- turn string_node into terminal_node
            declare
               New_Node : Ref_Statement_Database := Create_Node (Terminal_Node);
               Temp     : Ref_Statement_Database;
            begin
               Set_Symbolic_Name (New_Node,
                                  Get_Symbolic_Name (Get_Object (Node)));
               Set_Object (Node, New_Node);
               -- free old Get_Object (Node)
               -- Temp := Get_Object (Node);
               -- free (temp);
            end;
         end if;
         -----

         -- lookup symbol in symtbl;
         -- warning: Object could be an expr for array reference
         Check_Expression (Context, Get_Object (Node), Function_Entry);

         -- determine what kind of symbol it is, and mark and check it
         case Get_Entry_Kind (Get_Symbol_Table_Entry (Get_Object (Node))) is
            when Function_Entry     =>
               Set_Node_Kind (Node, Function_Call);
               Check_Subprogram_Call (Context, Node);
            when Type_Entry    | Subtype_Entry     | Derived_Type_Entry     =>
               Set_Node_Kind (Node, Type_Conversion);
               -- check that there is one and only one argument
               if Is_Empty (Get_Expression (Get_List_Of_Items (Node))) or
                  not Is_Empty (Get_Next_List_Item (Get_List_Of_Items (Node)))
                   then
                  Show_Error
                      (1986, Fatal,
                       Convert_To_String
                           (Get_Symbolic_Name (Get_Object (Node))));
               else
                  Check_Expression (Context,
                                    Get_Expression (Get_List_Of_Items (Node)));
                  -- check that type is convertable [tbd]
                  -- check that arg is at least of IN_MODE [tbd]
                  Set_Type (Node, Get_Symbol_Table_Entry (Get_Object (Node)));
               end if;
            when Object_Entry  =>
               Set_Node_Kind (Node, Array_Reference);
               -- not implementing this yet
               Check_Array_Reference (Node);
            when Procedure_Entry    =>
               Show_Error
                   (3024, Fatal,
                    Convert_To_String (Get_Symbolic_Name (Get_Object (Node))));
            when others   =>
               Show_Error (1004, Internal, Subprogram_Name);
         end case;

      when Dot_Operator   =>
         if Is_Empty (Get_Symbol_Table_Entry (Node)) then
            Check_Dot_Notation (Node);
         end if;

         if Entry_Kind_Actual /= Unknown and then
            Get_Entry_Kind (Get_Symbol_Table_Entry (Node)) /= Entry_Kind_Actual
             then
            Show_Error
                (3026, Fatal,
                 Convert_To_String
                     (Get_Symbolic_Name (Get_Symbol_Table_Entry (Node))));
         end if;
         -- set node_type
         --
         case Get_Entry_Kind (Get_Symbol_Table_Entry (Node)) is
            when Object_Entry              |
                 Function_Entry            |
                 Formal_Param_Entry        |
                 Enumeration_Literal_Entry |
                 Record_Component_Entry  =>
               Set_Type (Node, Get_Type (Get_Symbol_Table_Entry (Node)));
            when others   =>
               null; -- ignore
         end case;

      when List_Item =>
         declare
            Current_Node : Ref_Statement_Database;
            Count : Natural := 0;
         begin
            -- semanticly check each item
            Current_Node := Node;
            while (not Is_Empty (Current_Node)) loop
               Count := Count + 1;
               -- suggest a type prior to checking?
               Check_Expression (Context, Get_Expression (Current_Node));
               Set_Type (Node, Get_Type (Get_Expression (Current_Node)));
               Current_Node := Get_Next_List_Item (Current_Node);
            end loop;

            if Count > 1 then
               -- This is a list with more than one element, so we
               -- may treat it as an aggr.
               -- Create array type of "Count" elements and of type of
               -- the first component. TBD not right for records.
               null;
            end if;

         end;
         -- afterwards, node_type = type of last list expr checked

      when Aggregate_Component =>
         -- [finish later]
         -- for each Aggregate_Choice_Component,
         -- make sure that choice item is "correct"
         -- suggest a type prior to checking?
         Check_Expression (Context, Get_Aggregate_Expression (Node));
         -- set node_type [tbd]

      when Range_Expression    =>
         -- suggest a type prior to checking?
         Check_Expression (Context, Get_Lower_Bound (Node));
         Check_Expression (Context, Get_Upper_Bound (Node));
         if (Is_Same_Type (Get_Lower_Bound (Node), Get_Upper_Bound (Node)))
             then
            Set_Type (Node, Get_Type (Get_Lower_Bound (Node)));
         else
            Show_Error (1111, Fatal,
                        "upper differs from lower bound range expression type");
         end if;

      when Attribute_Tick =>
         Check_Attribute (Node);

      when Array_Reference  |
           Type_Conversion  |
           Record_Reference |
           Dot_Notation_Variable_Reference    =>
         -- can't be these here!
         Show_Error (1000, Internal, Subprogram_Name);

      when Function_Call  =>
         -- This line is only commented out for the Demo!! [TBD]
         -- When an object declaration has an initialization
         -- expression, and the expression is an unqualified
         -- function, this error message is erroneously displayed.
         -- wpl 11/23/88
         -- Show_Error (1000, Internal, Subprogram_Name);
         null;

      when Unknown   =>
         Show_Error (1003, Internal, Subprogram_Name);
   end case;

end Check_Expression;
