--|$Log:	context.a,v $
--|Revision 8.0  91/01/17  10:44:12  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Context_Definition
--
---------------------------------------------------------------
--
-- Abstract
--
--  This package contains the global data that the symbol table routines
--  and statement evaluator use to retain state (i.e., the context in which
--  statements are executed).  Currently, it is a collection of unrelated
--  data, but as ACE evolves a clearer picture of what is actually
--  needed to retain state should emerge (i.e., some of this information
--  is not inherently global).
--

with Symtbl_Entry_Get_Routines;
with Symtbl_Entry_Set_Routines;
with Text_Io;
use Text_Io;
with Reader_Writer;

package body Context_Control is

   package Int_Io is new Integer_Io (Integer);
   use Int_Io;
   -- used by Statement_Evaluator to tell if the current ACE level was
   -- caused by a subprogram call.

   Max_Call_Level : constant Ace_Natural := 1024;

   subtype Subprogram_Call_Index_Type is
      Ace_Natural range Ace_Natural'FIRST .. Ace_Natural'FIRST + Max_Call_Level;

   type Subprogram_Call_Stack is array (Subprogram_Call_Index_Type) of Boolean;

   -- used by the Statement_Evaluator to keep a list of the routines that
   -- are currently hidden at level, so that they may be re-introduced
   -- at the end of each level (i.e. scoping)
   type Hidden_Entry_List is array (Ace_Max_Levels) of Ref_List_Of_Entries;

   -- Data structures to implement a stack for the symtbl.  This
   -- should really be implemented as a generic. [TBD]

   type Top_Of_Symtbl_List;

   type Ref_Top_Of_Symtbl_List is access Top_Of_Symtbl_List;

   type Top_Of_Symtbl_List is
      record
         Head : Ref_Symbol_Table_Entry;
         Next : Ref_Top_Of_Symtbl_List;
      end record;

   -- Data structures to implement a stack for the display.  This
   -- should really be implemented as a generic. [TBD]

   type Display_Items;

   type Ref_Display_Items is access Display_Items;

   type Display_Items is
      record
         Item : Ace_Natural;
         Next : Ref_Display_Items;
      end record;

   -- Data structures to implement a stack for the storage pointers.  This
   -- should really be implemented as a generic. [TBD]

   type Storage_Pointers;

   type Ref_Storage_Pointers is access Storage_Pointers;

   type Storage_Pointers is
      record
         Pointer : Ace_Natural;
         Next    : Ref_Storage_Pointers;
      end record;

   type Symbol_Table_For_Dr is array (Ace_Max_Levels) of Ref_Symbol_Table_Entry;

   -- used by the Statement_Evaluator to know if a return statement is
   -- valid at a level.
   type Valid_Returns is array (Ace_Max_Levels) of Boolean;

   -- the Data Storage Stack (DSS) is used to store all object values in
   -- ACE.
   subtype Dss_Index_Type is
      Ace_Natural range Ace_Natural'FIRST .. Ace_Natural'FIRST + 1024;

   type Data_Storage_Stack is array (Dss_Index_Type) of Ref_Data_Storage;

   -- Offset and Display are used to calculate the position in the DSS
   -- that is associated with a particular object.
   type Offsets is array (Ace_Max_Levels) of Ace_Natural;

   type Displays is array (Ace_Max_Levels) of Ace_Natural;

   ----------------------------------------------------------------------
   -- Main record definition for context.
   ----------------------------------------------------------------------

   type Context_Record is
      record
         In_Interpret : Boolean := False;

         -- We retain the symbol table info in the context
         Head_Symbol_Table_Stack            : Ref_Top_Of_Symtbl_List :=
            new Top_Of_Symtbl_List'(Head => null,
                                    Next => null);
         Head_Symbol_Table_Stack_Save       : Ref_Top_Of_Symtbl_List :=
            new Top_Of_Symtbl_List'(Head => null,
                                    Next => null);

         Top_Of_Symtbl_Save                 : Ref_Symbol_Table_Entry;

         -- Used for major error recovery (i.e., we don't know where we are,
         -- so we set the head_symbol_table to this and return control to
         -- the user).
         Saved_Main_Level_Symbol_Table_Head : Ref_Symbol_Table_Entry := null;

         -- What runtime scoping level is.
         Call_Level : Ace_Natural := Ace_Natural'FIRST;
         Call_Level_Save                    : Ace_Natural := Ace_Natural'FIRST;

         -- for debugger routines so they know what the (real) currently
         -- executing statement is.
         Current_Statement                  : Ref_Statement_Database := null;
         In_Subprogram_Call                 : Subprogram_Call_Stack :=
            (others => False);

         -- Declarative Region working objects
         Context_Clause_Holder              : Ref_Symbol_Table_Entry := null;

         -- when parsing these keep the scoping levels.
         Dregion_Count : Ace_Natural := Ace_Natural'FIRST;
         Dr_Count_Keep : Ace_Natural := Ace_Natural'FIRST;

         Declarative_Region : Symbol_Table_For_Dr := (others => null);

         Declarative_Region_Entry           : Symbol_Table_For_Dr :=
            (others => null);

         Current_Level : Ace_Max_Levels := Ace_Max_Levels'FIRST;
         Saved_Level : Ace_Max_Levels := Ace_Max_Levels'FIRST;

         -- --
         -- current offset
         -- --
         Current_Offset                     : Ace_Natural := Ace_Natural'FIRST;
         Offset_Stack : Offsets := (others => Ace_Natural'FIRST);

         Saved_Offset : Ace_Natural := Ace_Natural'FIRST;
         Saved_Offset_Stack : Offsets := (others => Ace_Natural'FIRST);

         -- --
         -- data storage stack
         -- --
         Last_Dss_Sp : Ace_Natural := Ace_Natural'FIRST;
         -- Dss_Threshold holds the threshold for the DSS that is
         -- held in this context, otherwise we need to search the
         -- previous context (this implements the cactus stack).
         Dss_Threshold : Ace_Natural := Ace_Natural'FIRST;
         Dss : Data_Storage_Stack;
         Dss_Save : Data_Storage_Stack;

         Storage_Pointers_Stack             : Ref_Storage_Pointers;
         Display_Items_Stack                : Ref_Display_Items;

         -- Needs to be a copy of the previous context, so that a
         -- cactus stack is maintained. (see Fetch_Value routine
         -- for more information on cactus stack)
         Display : Displays;

         -- Context Links; we use this link to search different paths on
         -- the cactus stack for the DSS.
         Previous_Context                   : Context_Definition     := null;

         -- Enclosing Scope is used for exception handling, so we know what
         -- exception handler to execute.
         Enclosing_Scope                    : Ref_Statement_Database := null;

         -- Used by the debugger, not sure if it needs to be included in
         -- the context.
         Have_Encountered_Valid_Return_Stmt : Valid_Returns;
         Hidden_Entries : Hidden_Entry_List := (others => null);
      end record;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Create
   --
   --    Last Mod:    Wed Aug 16 13:56:05 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --   This routine will create a new context based on a previous context.
   --   It is called when a new task is elaborated.
   --
   --  --------------------------------------------------------------------
   function Create (Context : in Context_Definition) return Context_Definition
                   is
      Context_Holder : Context_Definition;
      Threshold      : Ace_Natural;
   begin
      if Context = null then
         -- The first context (the main thread of control) starts fresh.
         return new Context_Record;
      else

         Threshold := Get_Last_Storage_Location_Pointer (Context);

         -- If we are starting a new context then we must
         -- make a copy of the old one, and build on it.  This
         -- is what makes the stack look like a cactus.  Remember that
         -- a copy of Context.all needs to be made.
         Context_Holder                  := new Context_Record'(Context.all);

         -- Link the contexts together
         Context_Holder.Previous_Context := Context;

         -- Remember the last storage location, so we know when we need
         -- to traverse the Previous context link for a data value
         Context_Holder.Dss_Threshold    := Threshold;

         Context_Holder.Display          := Context.Display;

         return Context_Holder;
      end if;
   end Create;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Create
   --
   --    Last Mod:    Wed Aug 16 13:51:28 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Create return Context_Definition is
      Context : Context_Definition := new Context_Record;
   begin
      -- The first context (the main thread of control) starts fresh.
      return Context;
   end Create;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Enclosing_Scope
   --
   --    Last Mod:    Wed Aug 16 13:51:39 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Enclosing_Scope
       (Context : in Context_Definition) return Ref_Statement_Database is
   begin
      return Context.Enclosing_Scope;
   end Get_Enclosing_Scope;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Enclosing_Scope
   --
   --    Last Mod:    Wed Aug 16 13:43:14 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Enclosing_Scope (Context : in out Context_Definition;
                                  Item    : in Ref_Statement_Database) is
   begin
      Context.Enclosing_Scope := Item;
   end Set_Enclosing_Scope;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Saved_Level
   --
   --    Last Mod:    Wed Aug 16 13:44:03 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Saved_Level
       (Context : in Context_Definition) return Ace_Max_Levels is
   begin
      return Context.Saved_Level;
   end Get_Saved_Level;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Saved_Level
   --
   --    Last Mod:    Wed Aug 16 13:46:44 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Saved_Level (Context : in out Context_Definition;
                              Item    : in Ace_Max_Levels) is

   begin
      Context.Saved_Level := Item;
   end Set_Saved_Level;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Current_Level
   --
   --    Last Mod:    Wed Aug 16 13:48:57 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Current_Level
       (Context : in Context_Definition) return Ace_Max_Levels is

   begin
      return Context.Current_Level;
   end Get_Current_Level;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Current_Level
   --
   --    Last Mod:    Wed Aug 16 13:48:41 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Current_Level (Context : in out Context_Definition;
                                Item    : in Ace_Max_Levels) is

   begin
      Context.Current_Level := Item;
   end Set_Current_Level;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Saved_Offset
   --
   --    Last Mod:    Wed Aug 16 13:49:25 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Saved_Offset
       (Context : in Context_Definition) return Ace_Natural is

   begin -- GET_CURRENT_OFFSET
      return Context.Saved_Offset;
   end Get_Saved_Offset;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Saved_Offset
   --
   --    Last Mod:    Wed Aug 16 13:49:25 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Saved_Offset (Context : in out Context_Definition;
                               Item    : in Ace_Natural) is

   begin --
      Context.Saved_Offset := Item;
   end Set_Saved_Offset;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Current_Offset
   --
   --    Last Mod:    Wed Aug 16 13:49:25 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Current_Offset
       (Context : in Context_Definition) return Ace_Natural is

   begin -- GET_CURRENT_OFFSET
      return Context.Current_Offset;
   end Get_Current_Offset;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Current_Offset
   --
   --    Last Mod:    Wed Aug 16 13:49:25 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Current_Offset (Context : in out Context_Definition;
                                 Item    : in Ace_Natural) is

   begin --
      Context.Current_Offset := Item;
   end Set_Current_Offset;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Saved_Offset_Stack
   --
   --    Last Mod:    Wed Aug 16 13:49:25 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Saved_Offset_Stack
       (Context : in Context_Definition;
        Level   : in Ace_Max_Levels) return Ace_Natural is

   begin -- GET_OFFSET_STACK
      return Context.Saved_Offset_Stack (Level);
   end Get_Saved_Offset_Stack;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Saved_Offset_Stack
   --
   --    Last Mod:    Wed Aug 16 13:49:25 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Saved_Offset_Stack (Context : in out Context_Definition;
                                     Level   : in Ace_Max_Levels;
                                     Item    : in Ace_Natural) is

   begin -- SET_OFFSET_STACK
      Context.Saved_Offset_Stack (Level) := Item;
   end Set_Saved_Offset_Stack;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Offset_Stack
   --
   --    Last Mod:    Wed Aug 16 13:49:26 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Offset_Stack (Context : in Context_Definition;
                              Level   : in Ace_Max_Levels) return Ace_Natural
                             is

   begin -- GET_OFFSET_STACK
      return Context.Offset_Stack (Level);
   end Get_Offset_Stack;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Offset_Stack
   --
   --    Last Mod:    Wed Aug 16 13:49:26 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Offset_Stack (Context : in out Context_Definition;
                               Level   : in Ace_Max_Levels;
                               Item    : in Ace_Natural) is

   begin -- SET_OFFSET_STACK
      Context.Offset_Stack (Level) := Item;
   end Set_Offset_Stack;

   --=====================================================================
   --  <Subprogram Name>
   --
   --   Get_Data_Storage_Stack
   --   Set_Data_Storage_Stack
   --
   --  <Description>
   --
   --   These routines provide access to various storage management
   --   data structures.
   --
   --=====================================================================
   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Data_Storage_Stack
   --
   --    Last Mod:    Wed Aug 16 13:49:26 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Data_Storage_Stack
       (Context  : in Context_Definition;
        Location : in Ace_Natural) return Ref_Data_Storage is

   begin -- GET_DATA_STORAGE_STACK
      return Context.Dss (Location);
   end Get_Data_Storage_Stack;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Data_Storage_Stack
   --
   --    Last Mod:    Wed Aug 16 13:49:26 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Data_Storage_Stack (Context  : in out Context_Definition;
                                     Location : in Ace_Natural;
                                     Item     : in Ref_Data_Storage) is

   begin -- SET_DATA_STORAGE_STACK
      Context.Dss (Location) := Item;
   exception
      when Constraint_Error =>
         raise Storage_Error;
   end Set_Data_Storage_Stack;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Threshold
   --
   --    Last Mod:    Wed Aug 16 13:49:27 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Threshold (Context : in Context_Definition) return Ace_Natural
                          is

   begin -- GET_THRESHOLD
      return Context.Dss_Threshold;
   end Get_Threshold;

   --=====================================================================
   --   Get_Previous_Context
   --
   --  <Description>
   --
   --   This routine will provide access to the previous context (i.e., the
   --   point at which the current context diverged.
   --
   --=====================================================================
   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Previous_Context
   --
   --    Last Mod:    Wed Aug 16 13:49:27 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Previous_Context
       (Context : in Context_Definition) return Context_Definition is

   begin -- GET_PREVIOUS_CONTEXT
      return Context.Previous_Context;
   end Get_Previous_Context;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Last_Storage_Location_Pointer
   --
   --    Last Mod:    Wed Aug 16 13:49:27 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Last_Storage_Location_Pointer
       (Context : in Context_Definition) return Ace_Natural is
   begin
      return (Context.Last_Dss_Sp);
   end Get_Last_Storage_Location_Pointer;

   -- --
   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Last_Storage_Location_Pointer
   --
   --    Last Mod:    Wed Aug 16 13:49:27 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Last_Storage_Location_Pointer
       (Context : in out Context_Definition;
        Item    : in Ace_Natural) is
   begin
      Context.Last_Dss_Sp := Item;
   end Set_Last_Storage_Location_Pointer;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Push_Last_Storage_Location_Pointer
   --
   --    Last Mod:    Fri Oct 20 17:19:20 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Push_Last_Storage_Location_Pointer
       (Context : in out Context_Definition) is
   begin
      Context.Storage_Pointers_Stack :=
         new Storage_Pointers'(Pointer => Context.Last_Dss_Sp,
                               Next    => Context.Storage_Pointers_Stack);
   end Push_Last_Storage_Location_Pointer;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Pop_Last_Storage_Location_Pointer
   --
   --    Last Mod:    Fri Oct 20 17:19:24 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Pop_Last_Storage_Location_Pointer
       (Context : in out Context_Definition) is
      Tmp : Ref_Storage_Pointers;
   begin
      Tmp := Context.Storage_Pointers_Stack;
      Set_Last_Storage_Location_Pointer
          (Context, Context.Storage_Pointers_Stack.Pointer);
      Context.Storage_Pointers_Stack := Context.Storage_Pointers_Stack.Next;
      --[TBD]      Free (Tmp);
   end Pop_Last_Storage_Location_Pointer;

   -- --
   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Display_For_Level
   --
   --    Last Mod:    Wed Aug 16 13:49:27 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Display_For_Level (Context : in Context_Definition;
                                   Level   : in Ace_Natural) return Ace_Natural
                                  is
   begin
      return (Context.Display (Level));
   end Get_Display_For_Level;

   -- --
   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Display_For_Level
   --
   --    Last Mod:    Wed Aug 16 13:49:27 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Display_For_Level (Context     : in out Context_Definition;
                                    Level, Item : in Ace_Natural) is
   begin
      Context.Display (Level) := Item;
   end Set_Display_For_Level;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Push_Display_For_Level
   --
   --    Last Mod:    Fri Oct 20 17:19:05 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Push_Display_For_Level (Context : in out Context_Definition;
                                     Level   : in Ace_Natural) is
   begin
      Context.Display_Items_Stack :=
         new Display_Items'(Item => Context.Display (Level),
                            Next => Context.Display_Items_Stack);
   end Push_Display_For_Level;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Pop_Display_For_Level
   --
   --    Last Mod:    Fri Oct 20 17:19:01 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Pop_Display_For_Level (Context : in out Context_Definition;
                                    Level   : in Ace_Natural) is
      Tmp : Ref_Display_Items;
   begin
      Tmp := Context.Display_Items_Stack;
      Set_Display_For_Level (Context, Level, Context.Display_Items_Stack.Item);
      Context.Display_Items_Stack := Context.Display_Items_Stack.Next;
      --[TBD]      Free (Tmp);
   end Pop_Display_For_Level;

   --=====================================================================
   --  <Subprogram Name>
   --
   --   Set_Current_Statement
   --   Get_Current_Statement
   --
   --  <Description>
   --
   --   These routines set and get the currently executing statement
   --   w.r.t. the debuggers perspective
   --
   --=====================================================================
   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Current_Statement
   --
   --    Last Mod:    Wed Aug 16 13:49:27 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Current_Statement (Context : in out Context_Definition;
                                    Node    : Ref_Statement_Database) is
   begin
      Context.Current_Statement := Node;
   end Set_Current_Statement;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Current_Statement
   --
   --    Last Mod:    Wed Aug 16 13:49:27 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Current_Statement
       (Context : in Context_Definition) return Ref_Statement_Database is
   begin
      return Context.Current_Statement;
   end Get_Current_Statement;

   --=====================================================================
   --  <Subprogram Name>
   --
   --   Set_In_Subprogram_Call
   --   UnSet_In_Subprogram_Call
   --   Is_In_Subprogram_Call
   --
   --  <Description>
   --
   --   These routines provide access to a marker of when a routine
   --   is being executed within a subprogram call.
   --
   --=====================================================================
   -- --------------------------------------------------------------------
   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_In_Subprogram_Call
   --
   --    Last Mod:    Wed Aug 16 13:49:28 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_In_Subprogram_Call (Context : in out Context_Definition) is
   begin
      Context.In_Subprogram_Call (Context.Call_Level) := True;
   end Set_In_Subprogram_Call;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Unset_In_Subprogram_Call
   --
   --    Last Mod:    Wed Aug 16 13:49:28 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Unset_In_Subprogram_Call (Context : in out Context_Definition) is
   begin
      Context.In_Subprogram_Call (Context.Call_Level) := False;
   end Unset_In_Subprogram_Call;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Is_In_Subprogram_Call
   --
   --    Last Mod:    Wed Aug 16 13:49:28 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Is_In_Subprogram_Call
       (Context : in Context_Definition) return Boolean is
   begin
      return not Context.In_Interpret and then
             Context.In_Subprogram_Call (Context.Call_Level);
   end Is_In_Subprogram_Call;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Increment_Call_Level
   --
   --    Last Mod:    Wed Aug 16 13:49:28 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Increment_Call_Level (Context : in out Context_Definition) is
   begin
      Context.Call_Level := Context.Call_Level + 1;
   end Increment_Call_Level;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Decrement_Call_Level
   --
   --    Last Mod:    Wed Aug 16 13:49:28 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Decrement_Call_Level (Context : in out Context_Definition) is
   begin
      Context.Call_Level := Context.Call_Level - 1;
   end Decrement_Call_Level;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Call_Level
   --
   --    Last Mod:    Wed Aug 16 13:49:28 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Call_Level (Context : in Context_Definition) return Ace_Natural
                           is
   begin
      return Context.Call_Level;
   end Get_Call_Level;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Top_Of_Symtbl
   --
   --  <Description>
   --
   --    Returns the head of the symbol table.
   --
   --==========================================================================
   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Top_Of_Symtbl
   --
   --    Last Mod:    Wed Aug 16 13:49:28 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Top_Of_Symtbl
       (Context : in Context_Definition) return Ref_Symbol_Table_Entry is

   begin
      return Context.Head_Symbol_Table_Stack.Head;
   end Get_Top_Of_Symtbl;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Top_Of_Symtbl
   --
   --    Last Mod:    Wed Aug 16 13:49:28 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --    Sets the head of the symbol table to the specified symbol table
   --    entry.
   --  --------------------------------------------------------------------
   procedure Set_Top_Of_Symtbl (Context : in out Context_Definition;
                                Item    : in Ref_Symbol_Table_Entry) is

   begin
      Context.Head_Symbol_Table_Stack.Head := Item;
   end Set_Top_Of_Symtbl;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Push_Top_Of_Symtbl
   --
   --    Last Mod:    Wed Oct 18 12:14:46 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --    Push a new top of symbol_table.
   --
   --  --------------------------------------------------------------------
   procedure Push_Top_Of_Symtbl (Context : in out Context_Definition;
                                 Item    : in Ref_Symbol_Table_Entry) is

   begin
      Context.Head_Symbol_Table_Stack :=
         new Top_Of_Symtbl_List'(Head => Item,
                                 Next => Context.Head_Symbol_Table_Stack);
   end Push_Top_Of_Symtbl;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Pop_Top_Of_Symtbl
   --
   --    Last Mod:    Wed Oct 18 12:14:31 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Pop_Top_Of_Symtbl (Context : in out Context_Definition) is
      Tmp : Ref_Top_Of_Symtbl_List;
   begin
      Tmp := Context.Head_Symbol_Table_Stack;
      Context.Head_Symbol_Table_Stack := Context.Head_Symbol_Table_Stack.Next;
      --[TBD]      Free (Tmp);
   end Pop_Top_Of_Symtbl;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_To_Zero_Level
   --
   --    Last Mod:    Wed Oct 18 14:19:27 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --    Sets the state of the interpret to the zero level (the level
   --    at which the users types code (i.e., the prompt level).
   --
   --  --------------------------------------------------------------------
   procedure Set_To_Zero_Level (Context : in out Context_Definition) is

      -- Find level zero by traversing the symtbl_table to its end
      -- and return it.
      function Traverse_List
          (List : Ref_Top_Of_Symtbl_List) return Ref_Top_Of_Symtbl_List is
      begin
         if List.Next = null then
            return List;
         else
            return Traverse_List (List.Next);
         end if;
      end Traverse_List;

   begin
      -- currently, we can't nest calls to this routine...[TBD]
      -- because information will get lost, since all of these
      -- items are not stored in a stack...

      if Context.In_Interpret then
         raise User_Error;
      end if;

      -- Set In_Interpret flag (see In_Subprogram_Call)
      Context.In_Interpret                 := True;

      -- save top_of_symtbl for each level.
      Context.Head_Symbol_Table_Stack_Save := Context.Head_Symbol_Table_Stack;

      -- save zeroeth level top_of_symtbl
      Context.Top_Of_Symtbl_Save           :=
         Traverse_List (Context.Head_Symbol_Table_Stack).Head;

      -- set the top to the zeroeth level
      Context.Head_Symbol_Table_Stack :=
         Traverse_List (Context.Head_Symbol_Table_Stack);

      -- set call level for execution (saving old value)
      Context.Call_Level_Save              := Context.Call_Level;
      Context.Call_Level                   := Ace_Natural'FIRST;

      -- Save Dregion
      Context.Dr_Count_Keep                := Context.Dregion_Count;
      -- Set Dregion_Count to zeroeth level
      Context.Dregion_Count                := Ace_Natural'FIRST;

      -- save current level and offset values
      Context.Saved_Level                  := Context.Current_Level;
      Context.Saved_Offset                 := Context.Current_Offset;

      -- if at level zero already, make sure we save its proper offset.
      if Context.Current_Level = Ace_Natural'FIRST then
         Context.Offset_Stack (Ace_Natural'FIRST) := Context.Current_Offset;
      end if;

      -- set current level and offset to level zero
      Context.Current_Level      := Ace_Natural'FIRST;
      Context.Current_Offset := Context.Offset_Stack (Ace_Natural'FIRST);

      Context.Saved_Offset_Stack := Context.Offset_Stack;

      -- save storage array! (Costly save), should be some kind of
      -- heap that is copied instead of an array.
      Context.Dss_Save           := Context.Dss;

   end Set_To_Zero_Level;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Reset_From_Zero_Level
   --
   --    Last Mod:    Wed Oct 18 14:19:35 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --    This routine resets the interpreter to the level at which
   --    set_to_zero_Level was called.
   --
   --  --------------------------------------------------------------------
   procedure Reset_From_Zero_Level (Context : in out Context_Definition) is

      First_Offset : constant Ace_Natural :=
         Context.Saved_Offset_Stack (Ace_Natural'FIRST);
      Storage_Difference : constant Ace_Natural := Context.Current_Offset -
                                                   First_Offset;

      -- Used to link in added entries
      Old_Head, New_Head, Before_Old_Head : Ref_Symbol_Table_Entry;

      function Find_Entry_Before
          (Symtbl_Entry : in Ref_Symbol_Table_Entry) return
            Ref_Symbol_Table_Entry is
         Current_Entry : Ref_Symbol_Table_Entry :=
            Context.Head_Symbol_Table_Stack_Save.Head;
         Before        : Ref_Symbol_Table_Entry;
      begin
         while (Current_Entry /= null) loop
            if Current_Entry = Symtbl_Entry then
               return Before;
            end if;
            Before := Current_Entry;
            Current_Entry :=
               Symtbl_Entry_Get_Routines.Get_Next_Entry (Current_Entry);
         end loop;
      end Find_Entry_Before;

      -- traverse symtbl and push all values from level 1 on
      -- back onto the symtbl stack.

      procedure Traverse_Symtbl_List (List : in out Ref_Top_Of_Symtbl_List) is
      begin
         if List.Next.Next /= null then -- traverse everyone but the zeroeth.
            Traverse_Symtbl_List (List.Next);
         end if;
         Push_Top_Of_Symtbl (Context, List.Head);
         --[TBD] Free (List);
      end Traverse_Symtbl_List;

      -- traverse symtbl and change the pushed levels
      procedure Replace_Symtbl_List (List : in out Ref_Top_Of_Symtbl_List;
                                     Head : in Ref_Symbol_Table_Entry) is
      begin
         if List.Next = null then
            return ;
         end if;
         if List.Next.Next /= null then -- traverse everyone but the zeroeth.
            Replace_Symtbl_List (List.Next, Head);
         end if;
         List.Head := Head;
      end Replace_Symtbl_List;

      -- traverse Display_Stack and add the Storage_Difference
      -- to correct for the added values at level zero.
      procedure Traverse_Display_List (List : in out Ref_Display_Items) is
      begin
         List.Item := List.Item + Storage_Difference;
         if List.Next /= null then
            Traverse_Display_List (List.Next);
         end if;
      end Traverse_Display_List;

      -- traverse last_location_pointer_stack and add the Storage_Difference
      -- to correct for the added values at level zero.
      procedure Traverse_Storage_List (List : in out Ref_Storage_Pointers) is
      begin
         List.Pointer := List.Pointer + Storage_Difference;
         if List.Next /= null then
            Traverse_Storage_List (List.Next);
         end if;
      end Traverse_Storage_List;

   begin
      if First_Offset /= Context.Current_Offset then
         -- There was storage created, so we need to reset all the displays,
         -- and the DSS (move them "Storage_Difference" up).
         for I in Context.Current_Offset ..
            Context.Dss'LAST - Storage_Difference loop
            Context.Dss (I) := Context.Dss_Save (I - Storage_Difference);
         end loop;
         for I in Context.Display'FIRST + 1 .. Context.Display'LAST loop
            Context.Display (I) := Context.Display (I) + Storage_Difference;
         end loop;
         -- Move display.
         Traverse_Display_List (Context.Display_Items_Stack);
         -- Move Values.
         Traverse_Storage_List (Context.Storage_Pointers_Stack);
      end if;

      -- restore offsets stack (Keeping new zero level) [TBD] is this needed?
      Context.Offset_Stack                     := Context.Saved_Offset_Stack;
      Context.Offset_Stack (Ace_Natural'FIRST) := Context.Current_Offset;

      -- if we were at zeroth level then leave offset alone.
      if Context.Saved_Level /= Ace_Natural'FIRST then
         Context.Current_Offset := Context.Saved_Offset;
         Context.Current_Level  := Context.Saved_Level;
      end if;

      -- restore Dregion
      Context.Dregion_Count := Context.Dr_Count_Keep;

      -- restore calling level
      Context.Call_Level    := Context.Call_Level_Save;

      -- Must link the new head of the zeroeth level into the
      -- symbol table, so that the added entries are visible during
      -- the current execution.

      Old_Head              := Context.Top_Of_Symtbl_Save;

      New_Head              := Context.Head_Symbol_Table_Stack.Head;

      Before_Old_Head := Find_Entry_Before (Old_Head);

      if Before_Old_Head /= null then
         -- null means currently at the zeroeth level
         Symtbl_Entry_Set_Routines.Set_Next_Entry (Before_Old_Head, New_Head);
      else
         Replace_Symtbl_List (Context.Head_Symbol_Table_Stack_Save, New_Head);
      end if;

      -- restore symtbl (Keeping new zero level)
      if Context.Head_Symbol_Table_Stack_Save.Next /= null then
         Traverse_Symtbl_List (Context.Head_Symbol_Table_Stack_Save);
      end if;

      -- unset In_Interpret Flag
      Context.In_Interpret := False;

   end Reset_From_Zero_Level;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Saved_Main_Level_Symbol_Table_Head
   --
   --    Last Mod:    Wed Aug 16 13:49:29 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --    This procedure retains the value of a symbol table entry such that
   --    it may be used by the Reset_From_Subprogram_Or_Package_Declaration
   --    procedure.
   --  --------------------------------------------------------------------
   procedure Set_Saved_Main_Level_Symbol_Table_Head
       (Context      : in out Context_Definition;
        Symtbl_Entry : in Ref_Symbol_Table_Entry) is

   begin
      Context.Saved_Main_Level_Symbol_Table_Head := Symtbl_Entry;
   end Set_Saved_Main_Level_Symbol_Table_Head;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Saved_Main_Level_Symbol_Table_Head
   --
   --    Last Mod:    Wed Aug 16 13:49:29 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --    This procedure retains the value of a symbol table entry such that
   --    it may be used by the Reset_From_Subprogram_Or_Package_Declaration
   --    procedure.
   --  --------------------------------------------------------------------
   function Get_Saved_Main_Level_Symbol_Table_Head
       (Context : in Context_Definition) return Ref_Symbol_Table_Entry is

   begin
      return Context.Saved_Main_Level_Symbol_Table_Head;
   end Get_Saved_Main_Level_Symbol_Table_Head;

   --==========================================================================
   --  <Subprogram Name>
   --
   --  Reset_If_Symtbl_Head
   --
   --  <Description>
   --
   --    Will look through list of saved symbol table heads and replace
   --   any pointer to "Old_Entry" with the "New_Head".
   --
   --==========================================================================
   procedure Reset_If_Symtbl_Head (Context   : in out Context_Definition;
                                   Old_Entry : Ref_Symbol_Table_Entry;
                                   New_Head  : Ref_Symbol_Table_Entry) is
      procedure Modify_Matching_Symtbl_If_Head
          (List      : in out Ref_Top_Of_Symtbl_List;
           Old_Entry : in Ref_Symbol_Table_Entry;
           New_Entry : in Ref_Symbol_Table_Entry) is
      begin
         if List.Next /= null then -- traverse everyone
            Modify_Matching_Symtbl_If_Head (List.Next, Old_Entry, New_Entry);
         end if;
         if (List.Head = Old_Entry) then
            List.Head := New_Entry;
         end if;
      end Modify_Matching_Symtbl_If_Head;

   begin

      Modify_Matching_Symtbl_If_Head (Context.Head_Symbol_Table_Stack,
                                      Old_Entry, New_Head);

   end Reset_If_Symtbl_Head;

   --=====================================================================
   --  <Subprogram Name>
   --
   --   Get_Context_Clause_Holder
   --   Set_Context_Clause_Holder
   --
   --   Get_Dregion_Count
   --   Set_Dregion_Count
   --
   --   Get_Dr_Count_Keep
   --   Set_Dr_Count_Keep
   --
   --   Get_Declarative_Region
   --   Set_Declarative_Region
   --
   --   Get_Declarative_Region_Entry
   --   Set_Declarative_Region_Entry
   --
   --   Get_Have_Encountered_Valid_Return_Stmt
   --   Set_Have_Encountered_Valid_Return_Stmt
   --
   --   Get_Hidden_Entries
   --   Set_Hidden_Entries
   --
   --  <Description>
   --
   --   These routines provide access to various symbol table markers for
   --   the declarative region routines.
   --
   --=====================================================================
   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Context_Clause_Holder
   --
   --    Last Mod:    Wed Aug 16 13:49:29 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Context_Clause_Holder
       (Context : in Context_Definition) return Ref_Symbol_Table_Entry is
   begin
      return Context.Context_Clause_Holder;
   end Get_Context_Clause_Holder;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Context_Clause_Holder
   --
   --    Last Mod:    Wed Aug 16 13:49:29 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Context_Clause_Holder (Context : in out Context_Definition;
                                        Item    : in Ref_Symbol_Table_Entry) is
   begin
      Context.Context_Clause_Holder := Item;
   end Set_Context_Clause_Holder;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Dregion_Count
   --
   --    Last Mod:    Wed Aug 16 13:49:29 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Dregion_Count
       (Context : in Context_Definition) return Ace_Natural is
   begin
      return Context.Dregion_Count;
   end Get_Dregion_Count;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Dregion_Count
   --
   --    Last Mod:    Wed Aug 16 13:49:29 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Dregion_Count (Context : in out Context_Definition;
                                Item    : in Ace_Natural) is
   begin
      Context.Dregion_Count := Item;
   end Set_Dregion_Count;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Dr_Count_Keep
   --
   --    Last Mod:    Wed Aug 16 13:49:29 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Dr_Count_Keep
       (Context : in Context_Definition) return Ace_Natural is
   begin
      return Context.Dr_Count_Keep;
   end Get_Dr_Count_Keep;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Dr_Count_Keep
   --
   --    Last Mod:    Wed Aug 16 13:49:29 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Dr_Count_Keep (Context : in out Context_Definition;
                                Item    : in Ace_Natural) is
   begin
      Context.Dr_Count_Keep := Item;
   end Set_Dr_Count_Keep;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Declarative_Region
   --
   --    Last Mod:    Wed Aug 16 13:49:30 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Declarative_Region
       (Context : in Context_Definition;
        Level   : in Ace_Max_Levels) return Ref_Symbol_Table_Entry is
   begin
      return Context.Declarative_Region (Level);
   end Get_Declarative_Region;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Declarative_Region
   --
   --    Last Mod:    Wed Aug 16 13:49:30 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Declarative_Region (Context : in out Context_Definition;
                                     Level   : in Ace_Max_Levels;
                                     Item    : in Ref_Symbol_Table_Entry) is
   begin
      Context.Declarative_Region (Level) := Item;
   end Set_Declarative_Region;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Declarative_Region_Entry
   --
   --    Last Mod:    Wed Aug 16 13:49:30 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Declarative_Region_Entry
       (Context : in Context_Definition;
        Level   : in Ace_Max_Levels) return Ref_Symbol_Table_Entry is
   begin
      return Context.Declarative_Region_Entry (Level);
   end Get_Declarative_Region_Entry;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Declarative_Region_Entry
   --
   --    Last Mod:    Wed Aug 16 13:49:30 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Declarative_Region_Entry (Context : in out Context_Definition;
                                           Level   : in Ace_Max_Levels;
                                           Item    : in Ref_Symbol_Table_Entry)
                                          is
   begin
      Context.Declarative_Region_Entry (Level) := Item;
   end Set_Declarative_Region_Entry;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Have_Encountered_Valid_Return_Stmt
   --
   --    Last Mod:    Wed Aug 16 13:49:30 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Have_Encountered_Valid_Return_Stmt
       (Context : in Context_Definition;
        Level   : in Ace_Max_Levels) return Boolean is
   begin
      return Context.Have_Encountered_Valid_Return_Stmt (Level);
   end Get_Have_Encountered_Valid_Return_Stmt;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Have_Encountered_Valid_Return_Stmt
   --
   --    Last Mod:    Wed Aug 16 13:49:30 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Have_Encountered_Valid_Return_Stmt
       (Context : in out Context_Definition;
        Level   : in Ace_Max_Levels;
        Item    : in Boolean) is
   begin
      Context.Have_Encountered_Valid_Return_Stmt (Level) := Item;
   end Set_Have_Encountered_Valid_Return_Stmt;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Hidden_Entries
   --
   --    Last Mod:    Wed Aug 16 13:49:30 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Get_Hidden_Entries
       (Context : in Context_Definition;
        Level   : in Ace_Max_Levels) return Ref_List_Of_Entries is
   begin
      return Context.Hidden_Entries (Level);
   end Get_Hidden_Entries;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Set_Hidden_Entries
   --
   --    Last Mod:    Wed Aug 16 13:49:31 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   procedure Set_Hidden_Entries (Context : in out Context_Definition;
                                 Level   : in Ace_Max_Levels;
                                 Item    : in Ref_List_Of_Entries) is
   begin
      Context.Hidden_Entries (Level) := Item;
   end Set_Hidden_Entries;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Is_Empty
   -- 
   --    Last Mod:    Wed Apr 11 11:54:00 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   function Is_Empty (Context : in Context_Definition) return Boolean is
   begin
      return Context = null;
   end Is_Empty;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Is_In_Interpret
   -- 
   --    Last Mod:    Wed Apr 11 11:54:04 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   function Is_In_Interpret (Context : in Context_Definition) return Boolean is
   begin
      return True;
   end Is_In_Interpret;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Top_Of_Symtbl_Save
   -- 
   --    Last Mod:    Wed Apr 11 11:54:10 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   function Get_Top_Of_Symtbl_Save
       (Context : in Context_Definition) return Ref_Symbol_Table_Entry is

   begin
      return Context.Top_Of_Symtbl_Save;
   end Get_Top_Of_Symtbl_Save;


   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_Call_Level_Save
   -- 
   --    Last Mod:    Wed Apr 11 11:54:13 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   function Get_Call_Level_Save
       (Context : in Context_Definition) return Ace_Natural is
   begin
      return Ace_Natural'FIRST;
   end Get_Call_Level_Save;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Dump_Subprogram_Call_Stack
   -- 
   --    Last Mod:    Wed Apr 11 11:54:20 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   procedure Dump_Subprogram_Call_Stack (Outfile : Text_Io.File_Type;
                                         Context : in Context_Definition) is
   begin
      Put (Outfile, ":");

      for I in Subprogram_Call_Index_Type'FIRST ..
         Subprogram_Call_Index_Type'LAST loop
         Put (" ");
         Put (Integer (Boolean'POS (Context.In_Subprogram_Call (I))), 1);
      end loop;

   end Dump_Subprogram_Call_Stack;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Dump_Dss
   -- 
   --    Last Mod:    Wed Apr 11 11:54:26 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   procedure Dump_Dss (Outfile : File_Type;
                       Context : in Context_Definition) is
   begin

      for I in Dss_Index_Type loop
         Reader_Writer.Dump_Ref_Data_Storage (Outfile,
                                              Get_Data_Storage_Stack (Context,
                                                                      I));
      end loop;

   end Dump_Dss;


   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Dump_Dss_Save
   -- 
   --    Last Mod:    Wed Apr 11 11:54:33 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   procedure Dump_Dss_Save (Outfile : File_Type;
                            Context : in Context_Definition) is
   begin
      Put ("/ ");
      for I in Dss_Index_Type loop

         Reader_Writer.Dump_Ref_Data_Storage (Outfile, Context.Dss_Save (I));
      end loop;
      Put ("/ ");

   end Dump_Dss_Save;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Dump_Storage_Pointers_Stack
   -- 
   --    Last Mod:    Wed Apr 11 11:54:37 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   procedure Dump_Storage_Pointers_Stack (Outfile : File_Type;
                                          Context : in Context_Definition) is
      Curr_Ptr : Ref_Storage_Pointers := Context.Storage_Pointers_Stack;
   begin

      Put ("& ");
      while Curr_Ptr /= null loop
         Put (Integer (Curr_Ptr.Pointer), 1);
         Put (" ");
         Curr_Ptr := Curr_Ptr.Next;
      end loop;

      Put ("&");
   end Dump_Storage_Pointers_Stack;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Dump_Display_Items_Stack
   -- 
   --    Last Mod:    Wed Apr 11 11:54:43 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   procedure Dump_Display_Items_Stack (Outfile : in File_Type;
                                       Context : in Context_Definition) is
      Curr_Ptr : Ref_Display_Items := Context.Display_Items_Stack;
   begin
      Put ("= ");
      while Curr_Ptr /= null loop
         Put (Integer (Curr_Ptr.Item), 1);
         Put (" ");
         Curr_Ptr := Curr_Ptr.Next;
      end loop;
      Put ("= ");
   end Dump_Display_Items_Stack;


   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Dump_Valid_Returns
   -- 
   --    Last Mod:    Wed Apr 11 11:54:47 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   procedure Dump_Valid_Returns (Outfile : in File_Type;
                                 Context : in Context_Definition) is
   begin
      null;
   end Dump_Valid_Returns;


   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Dump_Hidden_Entries
   -- 
   --    Last Mod:    Wed Apr 11 11:54:52 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   procedure Dump_Hidden_Entries (Context : in Context_Definition) is
   begin
      null;
   end Dump_Hidden_Entries;


   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Get_In_Interpret
   -- 
   --    Last Mod:    Wed Apr 11 11:54:56 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   function Get_In_Interpret (Context : in Context_Definition) return Boolean
                             is
   begin
      return False;
   end Get_In_Interpret;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Dump_Top_Of_Symtbl_Save
   -- 
   --    Last Mod:    Wed Apr 11 11:54:59 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   procedure Dump_Top_Of_Symtbl_Save (Outfile : in File_Type;
                                      Context : in Context_Definition) is
   begin
      null;
   end Dump_Top_Of_Symtbl_Save;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Dump_Top_Of_Symtbl_Stack
   -- 
   --    Last Mod:    Wed Apr 11 11:55:03 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   procedure Dump_Top_Of_Symtbl_Stack (Outfile : in File_Type;
                                       Context : in Context_Definition) is
   begin
      null;
   end Dump_Top_Of_Symtbl_Stack;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Dump_Suprogram_Call_Stack
   -- 
   --    Last Mod:    Wed Apr 11 11:55:07 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   procedure Dump_Suprogram_Call_Stack (Outfile : File_Type;
                                        Context : in Context_Definition) is
   begin
      null;
   end Dump_Suprogram_Call_Stack;


   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Dump_Dregion_Entry
   -- 
   --    Last Mod:    Wed Apr 11 11:55:10 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   procedure Dump_Dregion_Entry (Outfile : File_Type;
                                 Context : in Context_Definition) is
      use Reader_Writer;
   begin
      Put ("< ");
      for I in Ace_Max_Levels loop
         Dump_Symbol_Table_Node (Outfile, Context.Declarative_Region_Entry (I));
      end loop;
      Put ("> ");

   end Dump_Dregion_Entry;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Dump_Dregion
   -- 
   --    Last Mod:    Wed Apr 11 11:55:15 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   procedure Dump_Dregion (Outfile : File_Type;
                           Context : in Context_Definition) is
      use Reader_Writer;
   begin
      Put ("< ");
      for I in Ace_Max_Levels'FIRST .. Ace_Max_Levels'LAST loop
         Dump_Symbol_Table_Node (Outfile, Context.Declarative_Region (1));
      end loop;
      Put (" >");
   end Dump_Dregion;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Dump_Offset_Stack
   -- 
   --    Last Mod:    Wed Apr 11 11:55:21 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   procedure Dump_Offset_Stack (Outfile : File_Type;
                                Context : in Context_Definition) is
   begin
      null;
   end Dump_Offset_Stack;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Dump_Saved_Offset_Stack
   -- 
   --    Last Mod:    Wed Apr 11 11:55:25 1990
   --    Author:      William P Loftus <wpl@huey>
   -- 
   --  <Description> 
   -- 
   --  --------------------------------------------------------------------
   procedure Dump_Saved_Offset_Stack (Outfile : File_Type;
                                      Context : in Context_Definition) is
   begin
      null;
   end Dump_Saved_Offset_Stack;

end Context_Control;
