--|$Log:	create.a,v $
--|Revision 8.0  91/01/17  10:44:15  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Node_Create_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    This package provides a set of CREATE routines for statement database
--    nodes.  This set of routines provides a level of data abstraction and
--    information hiding of the statement database node structures from the
--    tree building portion of the Ada command lanaguage interpreter.
--
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Ace_Universal_Types;
use Ace_Universal_Types;

with Statement_Database;
use Statement_Database;

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Error_Handler;
use Error_Handler;

with Node_Get_Routines;
use Node_Get_Routines;

with Literals;
use Literals;

with Help_Pkg;
use Help_Pkg;

pragma ELABORATE (Node_Get_Routines);

package body Node_Create_Routines is

   -- Holder for comments in interpreted source code, and
   -- this must be separated out when the parsers are made
   -- reentrant.  [TBD]

   Comment_List_Head : Ref_Comment := null;
   Comment_List      : Ref_Comment := null;

   -- -------------------------------------------------------------------------

   -- A unique number is assigned to every statement.
   -- This is done by incrementing a counter as a side-effect
   -- of a function that is called when a statement database
   -- node is created.

   Unique_Counter    : Ace_Natural := 0;
   pragma SHARED (Unique_Counter); -- Tasks might access this
   -- in parallel, so it needs
   -- to be shared.

   -- -------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Record_Statement
   --
   --  <Description>
   --
   --    The function is used to keep an array of all statements (a record
   --    of stmt database node).  Used by debugger "display (stmt)".
   --==========================================================================
   function Record_Statement
       (Stmt : in Ref_Statement_Database) return Ref_Statement_Database is
   begin
      Stmt_List (Get_Statement_Number (Stmt)) := Stmt;
      return Stmt;
   end Record_Statement;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Unique_Number
   --
   --  <Description>
   --
   --    This function side-effects by incrementing Unique_Counter.
   --    The function is used to assign unique numbers to each statement
   --    database node (that are used in debugging).
   --==========================================================================
   function Get_Unique_Number return Ace_Natural is
   begin
      Unique_Counter := Unique_Counter + 1;
      return Unique_Counter;
   end Get_Unique_Number;

   -- -------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Process_Comments
   --
   --  <Description>
   --
   --    This routine stores up comments, so that they may be
   --    associated with the nearest unit.  This needs some
   --    improvement.  Comments sometimes move from their original
   --    position to a position next to the nearest unit.
   --
   --  <Called By>
   --
   --    The parsers.
   --==========================================================================
   procedure Process_Comments (Comment : String) is

   begin
      if Comment_List_Head = null then
         Comment_List      := new Comments'(new String'(Comment),
                                       null);
         Comment_List_Head := Comment_List;
      else
         Comment_List.Next := new Comments'(new String'(Comment),
                                            null);
         Comment_List      := Comment_List.Next;
      end if;
   end Process_Comments;

   -- -------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Comments
   --
   --  <Description>
   --
   --    This routines returns the stored, so that they may be
   --    associated with the nearest unit.  After it is called the
   --    stored comments are emptied, so that they are not associated
   --    twice.
   --==========================================================================
   function Get_Comments return Ref_Comment is

      Current_Comments : Ref_Comment;

      procedure Void_Comments is
      begin
         Comment_List_Head := null;
      end Void_Comments;

   begin
      Current_Comments := Comment_List_Head;
      Void_Comments;
      return Current_Comments;
   end Get_Comments;

   -- -------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Node
   --
   --  <Description>
   --
   --    This routine creates a statement database node (statements only).
   --==========================================================================
   function Create_Node
       (Stmt_Kind : in Statement_Enumeration) return Ref_Statement_Database is

      -- Effect:
      --   A statement database record is created.  The statement database node
      --   contains a statement node of the appropriate statement kind.

      Stmt_Node         : Ref_Statement_Record;
      Hold_Stmt_Db_Node : Ref_Statement_Database;

   begin
      case Stmt_Kind is

         when Object_Declaration    =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Object_Declaration,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Name_List => null,
                   Init_Expression => null);

            -- variants are identical for Type_Declaration, Subtype_Declaration
            -- Procedure_Declaration, Function_Declaration, Package_Declaration
            -- Package_Body_Declaration, and Exception_Declaration
         when Exception_Declaration =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Exception_Declaration,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Exception_Name_List                  => null);

         when Type_Declaration =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Type_Declaration,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Symbol_Table_Entry                   => null);

         when Subtype_Declaration   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Subtype_Declaration,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Symbol_Table_Entry                   => null);

         when Procedure_Spec_Declaration =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Procedure_Spec_Declaration,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Symbol_Table_Entry                   => null);

         when Procedure_Declaration =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Procedure_Declaration,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Symbol_Table_Entry                   => null);

         when Function_Spec_Declaration  =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Function_Spec_Declaration,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Symbol_Table_Entry                   => null);

         when Function_Declaration  =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Function_Declaration,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Symbol_Table_Entry                   => null);

         when Package_Declaration   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Package_Declaration,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Symbol_Table_Entry                   => null);

         when Package_Body_Declaration   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Package_Body_Declaration,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Symbol_Table_Entry                   => null);

         when Assign_Statement =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Assign_Statement,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Lhs => null,
                   Rhs => null);

         when Loop_Statement   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Loop_Statement,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Loop_Name => null,
                   Loop_Scheme => null,
                   Loop_Body => null);

         when For_Statement    =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => For_Statement,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   For_Variable => null,
                   For_Range => null);

         when If_Statement     =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => If_Statement,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   If_Condition => null,
                   True_Task => null,
                   False_Task => null);

         when Case_Statement   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Case_Statement,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Case_Expression => null,
                   Case_Alternatives_List               => null);

         when Case_Alternative_Part =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Case_Alternative_Part,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Case_Alternative_Choice              => null,
                   Case_Alternative_Statements          => null,
                   Next_Case_Alternative                => null);

         when Exception_Alternative =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Exception_Alternative,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Exception_Alternative_Choice         => null,
                   Exception_Alternative_Statements     => null,
                   Next_Exception_Alternative           => null);

         when Labeled_Statement     =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Labeled_Statement,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Label_Entry => null);

         when Goto_Statement   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Goto_Statement,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Goto_Label => null);

         when Exit_Statement   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Exit_Statement,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Exit_Loop_Name => null,
                   Exit_Condition => null);

         when Return_Statement =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Return_Statement,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Return_Expression                    => null,
                   Subprogram_Reference                 => null);

         when Block_Statement  =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Block_Statement,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Block_Name => null,
                   Block_Declarations                   => null,
                   Block_Statements                     => null,
                   Block_Exceptions                     => null);

            -- variants are identical for With_Statement and Use_Statement
         when With_Statement   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => With_Statement,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Unit_Name_List => null);

         when Use_Statement    =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Use_Statement,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Unit_Name_List => null);

         when Procedure_Call   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Procedure_Call,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Procedure_Info => null,
                   Actual_Procedure_Parameters_List     => null);

         when Pragma_Statement =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Pragma_Statement,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number,
                   Pragma_Name => null,
                   Pragma_Arg => null);

            -- variants are identical for Null_Statement and Unknown
         when Null_Statement   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Null_Statement,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number);

         when Unknown     =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Unknown,
                   Next_Statement => null,
                   Holder => null,
                   Statement_Number                     => Get_Unique_Number);

      end case;

      Hold_Stmt_Db_Node :=
         (new Statement_Database_Record'(Node_Kind      => Statement_Node_Type,
                                         Parent_Node    => null,
                                         Comment_Holder => Get_Comments,
                                         Statement_Node => Stmt_Node));
      return Record_Statement (Hold_Stmt_Db_Node);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Create_Node <stmt>");
         raise Tree_Builder_Error;

   end Create_Node;

   -- -------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Node
   --
   --  <Description>
   --
   --   This routine creates a statement database node (expressions only).
   --   A statement database record is created.  The statement database node
   --   contains an expression node of the appropriate expression kind.
   --==========================================================================
   function Create_Node
       (Expr_Kind : in Expression_Enumeration) return Ref_Statement_Database is

      Expr_Node : Ref_Expression_Record;

   begin
      case Expr_Kind is
         when Dot_Operator     =>
            Expr_Node :=
               new Expression_Record'(Expression_Kind        => Dot_Operator,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Dot_Node_Kind          => Unknown,
                                      Dot_Object             => null,
                                      Expn_Terminal_Symbol   => null,
                                      Expn_Terminal_Sdb      => null,
                                      Dot_Component          => null);

         when Left_Parenthesis_Operator  =>
            Expr_Node :=
               new Expression_Record'
                  (Expression_Kind        => Left_Parenthesis_Operator,
                   Node_Value             => null,
                   Reference_Count        => Ace_Positive'FIRST,
                   Node_Type              => null,
                   Is_Default_Expression  => False,
                   Is_Node_Type_Universal => False,
                   Next_Expression        => null,
                   Lp_Node_Kind           => Unknown,
                   Lp_Object              => null,
                   Lp_Items_List          => null);

         when List_Item   =>
            Expr_Node :=
               new Expression_Record'(Expression_Kind        => List_Item,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Item_Expression        => null,
                                      Next_List_Item         => null);

         when Aggregate_Component   =>
            Expr_Node :=
               new Expression_Record'(Expression_Kind => Aggregate_Component,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Aggregate_Choice       => null,
                                      Aggregate_Expression   => null);

         when Range_Expression =>
            Expr_Node :=
               new Expression_Record'(Expression_Kind => Range_Expression,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Range_Name             => null,
                                      Lower_Bound            => null,
                                      Upper_Bound            => null);

         when Attribute_Tick   =>

            Expr_Node :=
               new Expression_Record'(Expression_Kind        => Attribute_Tick,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Attribute_Prefix       => null,
                                      Attribute              => null);

         when Terminal_Node    =>
            Expr_Node :=
               new Expression_Record'(Expression_Kind        => Terminal_Node,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Symbol_Name            => Create,
                                      Terminal_Symbol        => null);

         when String_Node =>
            Expr_Node :=
               new Expression_Record'(Expression_Kind        => String_Node,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      String_Symbol          => Create);

         when Character_Node   =>
            Expr_Node :=
               new Expression_Record'(Expression_Kind        => Character_Node,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Character_Symbol       => ' ');

         when Constant_Node    =>
            Expr_Node :=
               new Expression_Record'(Expression_Kind        => Constant_Node,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Constant_Symbol        => Create);

            -- the Array_Reference, Function_Call, and Type_Conversion are
            -- special cases of the Left_Parenthesis_Operator node kind
         when Array_Reference  =>
            Expr_Node :=
               new Expression_Record'
                  (Expression_Kind        => Left_Parenthesis_Operator,
                   Node_Value             => null,
                   Reference_Count        => Ace_Positive'FIRST,
                   Node_Type              => null,
                   Is_Default_Expression  => False,
                   Is_Node_Type_Universal => False,
                   Next_Expression        => null,
                   Lp_Node_Kind           => Array_Reference,
                   Lp_Object              => null,
                   Lp_Items_List          => null);

         when Function_Call    =>
            Expr_Node :=
               new Expression_Record'
                  (Expression_Kind        => Left_Parenthesis_Operator,
                   Node_Value             => null,
                   Reference_Count        => Ace_Positive'FIRST,
                   Node_Type              => null,
                   Is_Default_Expression  => False,
                   Is_Node_Type_Universal => False,
                   Next_Expression        => null,
                   Lp_Node_Kind           => Function_Call,
                   Lp_Object              => null,
                   Lp_Items_List          => null);

         when Type_Conversion  =>
            Expr_Node :=
               new Expression_Record'
                  (Expression_Kind        => Left_Parenthesis_Operator,
                   Node_Value             => null,
                   Reference_Count        => Ace_Positive'FIRST,
                   Node_Type              => null,
                   Is_Default_Expression  => False,
                   Is_Node_Type_Universal => False,
                   Next_Expression        => null,
                   Lp_Node_Kind           => Type_Conversion,
                   Lp_Object              => null,
                   Lp_Items_List          => null);

            -- the Record_Reference, and Dot_Notation_Variable_Reference are
            -- special cases of the Dot_Operator node kind
         when Record_Reference =>
            Expr_Node :=
               new Expression_Record'(Expression_Kind        => Dot_Operator,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Dot_Node_Kind => Record_Reference,
                                      Dot_Object             => null,
                                      Expn_Terminal_Symbol   => null,
                                      Expn_Terminal_Sdb      => null,
                                      Dot_Component          => null);

         when Dot_Notation_Variable_Reference =>
            Expr_Node :=
               new Expression_Record'
                  (Expression_Kind        => Dot_Operator,
                   Node_Value             => null,
                   Reference_Count        => Ace_Positive'FIRST,
                   Node_Type              => null,
                   Is_Default_Expression  => False,
                   Is_Node_Type_Universal => False,
                   Next_Expression        => null,
                   Dot_Node_Kind          => Dot_Notation_Variable_Reference,
                   Dot_Object             => null,
                   Expn_Terminal_Symbol   => null,
                   Expn_Terminal_Sdb      => null,
                   Dot_Component          => null);

         when Unknown     =>
            Expr_Node :=
               new Expression_Record'(Expression_Kind => Array_Reference,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null);
      end case;

      return
         (new Statement_Database_Record'(Node_Kind => Expression_Node_Type,
                                         Parent_Node     => null,
                                         Comment_Holder  => null,
                                         Expression_Node => Expr_Node));

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Create_Node <expr>");
         raise Tree_Builder_Error;

   end Create_Node;

   -- -------------------------------------------------------------------------
   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Statement_Database_Item
   --
   --  <Description>
   --
   --   The tree builder is separated into two package one package is the
   --   general routines, and the other is the specific routines.
   --   This package contain the specification of the general routines.
   --
   --   The general routines can be separated into three categories:
   --
   --     creating statement database nodes
   --
   --     creating expression nodes
   --
   --     creating statement nodes
   --
   --   These catagories are the only assumptions made about the
   --   underlying structure of the statement database.
   --   These routines are functional in nature returning small sections of
   --   the abstract syntax tree that will be hooked together to form a complete
   --+--   tree by the specific tree building routines.
   --   The general tree building routines interface directly with the statement
   --+--   database's data structures.
   --
   --
   --  <Exceptions Raised>
   --
   --    Tree_Builder_Error
   --
   --  <Exceptions Handled>
   --
   --    Turns all exceptions to Tree_Builder_Error
   --==========================================================================
   function Create_Statement_Database_Item (Item : Ref_Statement_Record := null
                                            ) return Ref_Statement_Database is

   begin
      return new Statement_Database_Record'(Node_Kind => Statement_Node_Type,
                                            Parent_Node    => null,
                                            Comment_Holder => Get_Comments,
                                            Statement_Node => Item);
   exception
      when others    =>
         Show_Error (3008, Internal, "Create_Statement_Database_Item");
         raise Tree_Builder_Error;

   end Create_Statement_Database_Item;

   -- -------------------------------------------------------------------------
   --==========================================================================
   --  <Subprogram Name>
   --
   --   Create_Statement_Database_Item
   --
   --  <Description>
   --
   --   The routine Create_Statement_Database_Item will hook together pointers
   --   to creation one Statement_Database_Record.  The kind of variant that
   --   will be created is based on the parameter Kind, the rest of the items
   --   are the pointers to subtrees to be "hooked" in.
   --
   --  <Exceptions Raised>
   --
   --    Tree_Builder_Error
   --
   --  <Exceptions Handled>
   --
   --    Turns all exceptions to Tree_Builder_Error
   --==========================================================================
   function Create_Statement_Database_Item (Item : Ref_Expression_Record :=
      null) return Ref_Statement_Database is

   begin
      return new Statement_Database_Record'(Node_Kind => Expression_Node_Type,
                                            Parent_Node     => null,
                                            Comment_Holder  => null,
                                            Expression_Node => Item);
   exception
      when others    =>
         Show_Error (3008, Internal, "Create_Statement_Database_Item");
         raise Tree_Builder_Error;

   end Create_Statement_Database_Item;

   -- -------------------------------------------------------------------------
   --==========================================================================
   --  <Subprogram Name>
   --
   --   Create_Statement_Item
   --
   --  <Description>
   --
   --   The routine Create_Statement_Item will hook together pointers
   --   to creation one Statement_Record.  The kind of variant that
   --   will be created is based on the parameter Kind, the rest of the items
   --   are the pointers to subtrees to be "hooked" in.
   --
   --  <Exceptions Raised>
   --
   --    Tree_Builder_Error
   --
   --  <Exceptions Handled>
   --
   --    Turns all exceptions to Tree_Builder_Error
   --==========================================================================
   function Create_Statement_Item
       (Kind : Statement_Enumeration;
        Item1, Item2, Item3, Item4 : Ref_Statement_Database := null) return
         Ref_Statement_Record is

      Stmt_Node : Ref_Statement_Record;

   begin
      case Kind is

         when Object_Declaration         |
              Type_Declaration           |
              Exception_Declaration      |
              Subtype_Declaration        |
              Procedure_Spec_Declaration |
              Procedure_Declaration      |
              Function_Spec_Declaration  |
              Function_Declaration       |
              Package_Declaration        |
              Package_Body_Declaration   =>
            raise Tree_Builder_Error;

         when Exception_Alternative =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Exception_Alternative,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   Exception_Alternative_Choice         => Item1,
                   Exception_Alternative_Statements     => Item2,
                   Next_Exception_Alternative           => Item3);

            return Stmt_Node;

         when Assign_Statement =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Assign_Statement,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   Lhs => Item1,
                   Rhs => Item2);

            return Stmt_Node;

         when Loop_Statement   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Loop_Statement,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   Loop_Scheme => Item1,
                   Loop_Body => Item2,
                   Loop_Name => Item3);

            return Stmt_Node;

         when For_Statement    =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => For_Statement,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   For_Variable => Item1,
                   For_Range => Item2);

            return Stmt_Node;

         when If_Statement     =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => If_Statement,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   If_Condition => Item1,
                   True_Task => Item2,
                   False_Task => Item3);

            return Stmt_Node;

         when Case_Statement   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Case_Statement,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   Case_Expression => Item1,
                   Case_Alternatives_List               => Item2);

            return Stmt_Node;

         when Case_Alternative_Part =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Case_Alternative_Part,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   Case_Alternative_Choice              => Item1,
                   Case_Alternative_Statements          => Item2,
                   Next_Case_Alternative                => Item3);

            return Stmt_Node;

         when Labeled_Statement     =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Labeled_Statement,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   Label_Entry => Item1);

            return Stmt_Node;

         when Goto_Statement   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Goto_Statement,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   Goto_Label => Item1);

            return Stmt_Node;

         when Exit_Statement   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Exit_Statement,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   Exit_Loop_Name => Item1,
                   Exit_Condition => Item2);

            return Stmt_Node;

         when Return_Statement =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Return_Statement,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   Return_Expression                    => Item1,
                   Subprogram_Reference                 => null);

            return Stmt_Node;

         when Block_Statement  =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Block_Statement,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   Block_Declarations                   => Item1,
                   Block_Statements                     => Item2,
                   Block_Name => Item3,
                   Block_Exceptions                     => Item4);

            return Stmt_Node;

         when With_Statement   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => With_Statement,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   Unit_Name_List => Item1);

            return Stmt_Node;

         when Use_Statement    =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Use_Statement,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   Unit_Name_List => Item1);

            return Stmt_Node;

         when Procedure_Call   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Procedure_Call,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   Procedure_Info => Item1,
                   Actual_Procedure_Parameters_List     => Item2);

            return Stmt_Node;

         when Pragma_Statement =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Pragma_Statement,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number,
                   Pragma_Name => Item1,
                   Pragma_Arg => Item2);

            return Stmt_Node;

         when Null_Statement   =>
            Stmt_Node :=
               new Statement_Record'
                  (Suspend_Evaluation_At_This_Statement => False,
                   Statement_Kind => Null_Statement,
                   Holder => null,
                   Next_Statement => null,
                   Statement_Number                     => Get_Unique_Number);

            return Stmt_Node;

         when Unknown     =>
            raise Tree_Builder_Error;
      end case;

   exception
      when others    =>
         Show_Error (3008, Internal, "Create_Statement_Item");
         raise Tree_Builder_Error;

   end Create_Statement_Item;

   -- -------------------------------------------------------------------------
   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Expression_Item
   --
   --  <Description>
   --
   --   The routine Create_Expression_Item will hook together pointers
   --   to creation one Expression_Record.  The kind of variant that
   --   will be created is based on the parameter Kind, the rest of the items
   --   are the pointers to subtrees to be "hooked" in.
   --
   --
   --  <Exceptions Raised>
   --
   --   Tree_Builder_Error
   --
   --  <Exceptions Handled>
   --
   --   Turns all exceptions to Tree_Builder_Error
   --==========================================================================
   function Create_Expression_Item
       (Kind                : Expression_Enumeration;
        Item1, Item2, Item3 : Ref_Statement_Database := null;
        Symbol_Table        : Ref_Symbol_Table_Entry := null;
        Str                 : Ref_String             := Create;
        Enum                : Lp_Node_Enumeration    := Unknown) return
         Ref_Expression_Record is

   begin
      case Kind is
         when Dot_Operator     =>
            return
               new Expression_Record'(Expression_Kind        => Dot_Operator,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Dot_Node_Kind          => Unknown,
                                      Dot_Object             => Item1,
                                      Expn_Terminal_Symbol   => null,
                                      Expn_Terminal_Sdb      => null,
                                      Dot_Component          => Item2);

         when Left_Parenthesis_Operator  =>
            return
               new Expression_Record'
                  (Expression_Kind        => Left_Parenthesis_Operator,
                   Node_Value             => null,
                   Reference_Count        => Ace_Positive'FIRST,
                   Node_Type              => null,
                   Is_Default_Expression  => False,
                   Is_Node_Type_Universal => False,
                   Next_Expression        => null,
                   Lp_Node_Kind           => Enum,
                   Lp_Object              => Item1,
                   Lp_Items_List          => Item2);

         when List_Item   =>
            return
               new Expression_Record'(Expression_Kind        => List_Item,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Item_Expression        => Item1,
                                      Next_List_Item         => Item2);

         when Aggregate_Component   =>
            return
               new Expression_Record'(Expression_Kind => Aggregate_Component,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Aggregate_Choice       => Item1,
                                      Aggregate_Expression   => Item2);

         when Range_Expression =>
            return
               new Expression_Record'(Expression_Kind => Range_Expression,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Lower_Bound            => Item1,
                                      Upper_Bound            => Item2,
                                      Range_Name             => Item3);

         when Attribute_Tick   =>
            return
               new Expression_Record'(Expression_Kind        => Attribute_Tick,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Attribute_Prefix       => Item1,
                                      Attribute              => Item2);

         when Terminal_Node    =>
            return
               new Expression_Record'(Expression_Kind        => Terminal_Node,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Symbol_Name            => Str,
                                      Terminal_Symbol        => Symbol_Table);

         when String_Node =>
            return
               new Expression_Record'(Expression_Kind        => String_Node,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      String_Symbol          => Str);

         when Character_Node   =>
            return
               new Expression_Record'
                  (Expression_Kind        => Character_Node,
                   Node_Value             => null,
                   Reference_Count        => Ace_Positive'FIRST,
                   Node_Type              => null,
                   Is_Default_Expression  => False,
                   Is_Node_Type_Universal => False,
                   Next_Expression        => null,
                   Character_Symbol =>
                      Convert_To_String (Str) (Convert_To_String (Str)'FIRST));

         when Constant_Node    =>
            return
               new Expression_Record'(Expression_Kind        => Constant_Node,
                                      Node_Value             => null,
                                      Reference_Count => Ace_Positive'FIRST,
                                      Node_Type              => null,
                                      Is_Default_Expression  => False,
                                      Is_Node_Type_Universal => False,
                                      Next_Expression        => null,
                                      Constant_Symbol        => Str);

         when Array_Reference                 |
              Function_Call                   |
              Type_Conversion                 |
              Record_Reference                |
              Dot_Notation_Variable_Reference |
              Unknown     =>
            raise Tree_Builder_Error;

      end case;

   exception
      when others    =>
         Show_Error (3008, Internal, "Create_Expression_Item");
         raise Tree_Builder_Error;

   end Create_Expression_Item;

   -- -------------------------------------------------------------------------

end Node_Create_Routines;
