--|$Log:	cstmt.a,v $
--|Revision 8.0  91/01/17  10:44:18  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Subunit for Package - Semantic_Analyzer
--
---------------------------------------------------------------
--
--
-- Abstract
--
--  The following are spearate routines for the package Semantic_Analyzer.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------
-- --------------------------------------------------------------------------
--
-- Effects:
--   Checks the given Ada statement node structure against correct
--   Ada semantics.
--
-- Requires:
--

with Node_Set_Routines;
use Node_Set_Routines;

with Symtbl_Entry_Set_Routines;
use Symtbl_Entry_Set_Routines;

with Storage_Management_Routines;
use Storage_Management_Routines;

with Declarative_Region_Support_Routines;
use Declarative_Region_Support_Routines;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Derived_Subprogram_Support;
use Derived_Subprogram_Support;

--==========================================================================
--  <Subprogram Name>
--
--    Check_Statement
--
--  <Description>
--
--   Checks the given Ada statement node structure against correct
--   Ada semantics.
--
--==========================================================================
separate (Semantic_Analyzer)

procedure Check_Statement (Context : in out Context_Control.Context_Definition;
                           Node    : in Ref_Statement_Database) is

   Subprogram_Name : constant String := "Check_Statement";

   procedure Evaluate_Pragma_Statement (Node : in Ref_Statement_Database) is
   separate;

begin
   -- Notes in general:
   --    - set node type when/if appropriate
   --    - check for correct usage of MODEs of items
   if Is_Empty (Node) then
      return ;
   end if;

   case Check_Stmt_Node (Node) is

      when Exception_Declaration    | Exception_Alternative  =>

         null; -- TBD what semantics should we do?

      when Assign_Statement    =>
         declare
            Lhs : Ref_Statement_Database := Get_Lhs (Node);
            Rhs : Ref_Statement_Database := Get_Rhs (Node);
         begin
            Check_Expression (Context, Lhs);

            -----
            -- check that LHS is assignable
            --
            if Check_Expr_Node (Lhs) = Terminal_Node then
               case Get_Entry_Kind (Get_Symbol_Table_Entry (Lhs)) is
                  when Object_Entry | Formal_Param_Entry     =>
                     -- [tbd] Record_Component_Entry =>
                     null; -- these are ok LHSs
                  when others  =>
                     Show_Error
                         (1116, Fatal,
                          "assignment statement LHS: " &
                          Convert_To_String
                              (Get_Symbolic_Name (Get_Symbol_Table_Entry (Lhs))
                               ));
               end case;
            end if;
            -----

            -----
            -- Note: this next if condition works for references to
            -- identifiers only
            -- redo for record & array references [tbd] -- oei
            if Is_Mode_In (Lhs) then
               Show_Error (1114, Fatal,
                           "mode of LHS should be of some OUT mode");
            end if;
            -----

            -----
            -- RHS type should be same as LHS type, so suggest it
            Set_Type (Rhs, Get_Type (Lhs));
            -----

            Check_Expression (Context, Rhs);

            if Check_Expr_Node (Rhs) = Terminal_Node then
               case Get_Entry_Kind (Get_Symbol_Table_Entry (Rhs)) is
                  when Object_Entry       |
                       Formal_Param_Entry |
                       Function_Entry    =>
                     null;
                  when others  =>
                     Show_Error
                         (3025, Fatal,
                          Convert_To_String
                              (Get_Symbolic_Name (Get_Symbol_Table_Entry (Rhs))
                               ));
               end case;
            end if;

            if Is_Mode_Out (Rhs) then
               Show_Error
                   (1114, Fatal, "mode of RHS should be of some IN mode");
            end if;

            if (not Is_Same_Type (Lhs, Rhs)) then
               Show_Error (1111, Fatal, "types for LHS and RHS do not match");
            end if;
         end;

      when Loop_Statement =>
         if (Is_Empty (Get_Loop_Scheme (Node))) then
            -- endless LOOP
            null; -- nothing to do here
         elsif (Is_Expression_Node (Get_Loop_Scheme (Node))) then
            null; -- checks on WHILE loop condition already performed
         elsif (Is_Statement_Node (Get_Loop_Scheme (Node))) then
            -- FOR stmt
            -- does this need to be done??
            Check_Statement (Context, Get_Loop_Scheme (Node));
            -- check if Loop_Scheme is of type BOOLEAN:
            -- [tbd]
         else
            Show_Error (1001, Internal, "Loop_Statement/Loop_Scheme");
            -- raise some exception
         end if;

      when Procedure_Call =>
         -- search symtbl for proc decl symtbl entry and attach to this node
         -- [tbd] oei
         Check_Expression
             (Context, Get_Subprogram_Info (Node), Procedure_Entry);
         -- Set_Symbol_Table_Entry
         -- (Get_Subprogram_Info (Node),
         -- Find_Symbol_Table_Entry
         -- (Get_Symbolic_Name (Get_Subprogram_Info (Node)),
         -- Procedure_Entry));

         if (Is_Empty (Get_Symbol_Table_Entry (Get_Subprogram_Info (Node))))
             then
            Show_Error
                (1100, Fatal,
                 "specification for procedure not found: " &
                 Convert_To_String
                     (Get_Symbolic_Name (Get_Subprogram_Info (Node))));
         end if;
         Check_Subprogram_Call (Context, Node);

      when Case_Statement =>
         declare
            List_Node : Ref_Statement_Database; -- for lists
         begin
            -- check that all case choices are covered [tbd]
            -- check on CASE EXPRESSION already performed
            List_Node := Get_Alternatives_List (Node);
            while (not Is_Empty (List_Node)) loop
               Check_Expression (Context, Get_Alternative_Choice (List_Node));
               -- need to handle when multiple-ORed expr for a choice (??)
               if (not Is_Same_Type (Get_Expression (Node),
                                     Get_Alternative_Choice (List_Node))) then
                  Show_Error (1111, Fatal,
                              "CASE and CHOICE expression type do not match");
                  exit ;
               end if;
               List_Node := Get_Next_Alternative (List_Node);
            end loop;
         end;

      when Return_Statement    =>
         Set_Subprogram_Reference
             (Node, Get_Declarative_Region_Entry (Context));
         if (Is_Empty (Get_Subprogram_Reference (Node))) then
            Show_Error (1402, Warning, "return statement");
            raise User_Error;
         else
            case Get_Entry_Kind (Get_Declarative_Region_Entry (Context)) is
               when Procedure_Entry =>
                  -- expression must be empty for proc bodies
                  if (not Is_Empty (Get_Expression (Node))) then
                     Show_Error
                         (1401, Warning,
                          Convert_To_String
                              (Get_Symbolic_Name
                                   (Get_Declarative_Region_Entry (Context))));
                  end if;
               when Function_Entry  =>
                  -- expression cannot be empty w/in function bodies
                  if (Is_Empty (Get_Expression (Node))) then
                     Show_Error
                         (1403, Warning,
                          Convert_To_String
                              (Get_Symbolic_Name
                                   (Get_Declarative_Region_Entry (Context))));
                  else
                     -- suggest a type prior to checking?
                     Check_Expression (Context, Get_Expression (Node));
                     -- check that return types match
                     if not Is_Same_Type
                            (Get_Expression (Node),
                             Get_Return_Type
                                 (Get_Declarative_Region_Entry (Context))) then
                        Show_Error (1111, Fatal,
                                    "in return statement expression");
                     end if;
                     -- must have at least one "return" w/in function bodies
                     Valid_Return_Statement_Encountered
                         (Context);
                  end if;
               when others     =>
                  Show_Error (1402, Warning,
                              "return statement only valid within subprograms");
            end case;
         end if;

      when Exit_Statement =>
         if (not Is_Empty (Get_Name (Node))) then
            if ((not Are_Symbolic_Names_Equal
                   (Get_Symbolic_Name (Get_Name (Node)),
                    Convert_To_Ref_String ("ace"))) and
               (not Are_Symbolic_Names_Equal
                   (Get_Symbolic_Name (Get_Name (Node)),
                    Convert_To_Ref_String ("ace_main"))) and
               (not Are_Symbolic_Names_Equal
                   (Get_Symbolic_Name (Get_Name (Node)),
                    Convert_To_Ref_String ("ace_level")))) then
               Show_Error (1988, Warning, "exit statement with loop name");
            end if;
         end if;
         -- condition already checked (if there is one)

      when Block_Statement     =>
         -- the following should be done as a part of the grammar ...
         -- perform operations similar to subprogram environment "pushing"
         -- Incr_Level;
         -- Init_Offset_Value;
         -- save current offset_value

         -- don't need to check DECLs or STMTs because they been checked already
         null;

         -- the following should be done as a part of the grammar ...
         -- perform operations similar to subprogram environment "restoring"
         -- Decr_Level;
         -- reset previous offset_value

         --------------------
      when Object_Declaration  =>

         -- [check constraints ??]

         declare
            Init_Node    : Ref_Statement_Database := Get_Expression (Node);
            List_Node    : Ref_Statement_Database := Get_Name_List (Node);
            Symtbl_Entry : Ref_Symbol_Table_Entry;
         begin
            -----
            -- check init expr
            if (not Is_Empty (Init_Node)) then
               -- suggest a type prior to checking
               Set_Type
                   (Init_Node,
                    Get_Type
                        (Get_Symbol_Table_Entry
                             (Get_Expression (Get_Name_List (Node)))));
               Check_Expression (Context, Init_Node);
               if Is_Mode_Out (Init_Node) then
                  Show_Error
                      (1114, Fatal,
                      "mode of object(s) declaration initialization expression "
                        & "should be of some IN mode");
               end if;
               if (not Is_Same_Type
                      (Init_Node,
                       Get_Type
                           (Get_Symbol_Table_Entry
                                (Get_Expression (Get_Name_List (Node)))))) then
                  Show_Error
                      (1111, Fatal,
                       "initialization expression type does not match" &
                       " object(s) declaration type");
               end if;
            end if;
            -----
            -----
            -- check decl
            while (not Is_Empty (List_Node)) loop
               Symtbl_Entry :=
                  Get_Symbol_Table_Entry (Get_Expression (List_Node));
               Set_Statement_Number (Symtbl_Entry, Get_Statement_Number (Node));
               -----
               -- [if w/in subpgm body, then put symtbl entry on symtbl]
               if Within_Declarative_Region (Context) then
                  Add_Entry_To_Symbol_Table (Context, Symtbl_Entry);
               end if;
               -----
               List_Node := Get_Next_List_Item (List_Node);
            end loop;
            -----
         end;

         --------------------

      when Type_Declaration    | Subtype_Declaration    =>

         -- [check constraints (for type and subtypes only) ??]

         declare
            Symtbl_Entry : Ref_Symbol_Table_Entry :=
               Get_Symbol_Table_Entry (Node);
         begin
            Set_Statement_Number (Symtbl_Entry, Get_Statement_Number (Node));
            -----
            -- [if w/in subpgm body, then put symtbl entry on symtbl]
            if Within_Declarative_Region (Context) then
               Add_Entry_To_Symbol_Table (Context, Symtbl_Entry);
               if Get_Entry_Kind (Symtbl_Entry) = Derived_Type_Entry then
                  Derive_Subprograms (Context, Symtbl_Entry);
               end if;
            end if;
            -----
         end;

      when Package_Declaration        |
           Procedure_Spec_Declaration |
           Procedure_Declaration      |
           Function_Spec_Declaration  |
           Function_Declaration     =>

         Set_Statement_Number (Get_Symbol_Table_Entry (Node),
                               Get_Statement_Number (Node));
         -- if w/in subpgm body, then put symtbl entry on symtbl,
         --    but only if the entry is not already there
         if Within_Declarative_Region (Context) then
            begin
               Add_Entry_To_Symbol_Table (Context,
                                          Get_Symbol_Table_Entry (Node));
            exception
               when Duplicate_Symbol_Table_Entry_Error  =>
                  null;
            end;
         end if;

         if not Is_Empty (Get_Context_Clause (Get_Symbol_Table_Entry (Node)))
             then
            Check_Statement
                (Context, Get_Context_Clause (Get_Symbol_Table_Entry (Node)));
         end if;
         --------------------

      when Package_Body_Declaration =>
         -- if w/in subpgm =>
         --   check that pck_symtbl_entry is already on symtbl; [tbd]
         if not Is_Empty (Get_Context_Clause (Get_Symbol_Table_Entry (Node)))
             then
            Check_Statement
                (Context, Get_Context_Clause (Get_Symbol_Table_Entry (Node)));
         end if;

         --------------------

      when With_Statement =>
         -- assoc package name w/ file containing package spec
         -- for "use" in context_cl add package_symtbl_entry to symbol_table;
         declare
            List_Node : Ref_Statement_Database := Get_Next_Statement (Node);
         begin
            while not Is_Empty (List_Node) loop
               if Check_Stmt_Node (List_Node) = Use_Statement then
                  Check_Statement (Context, List_Node);
               end if;
               List_Node := Get_Next_Statement (List_Node);
            end loop;
         end;
         -- trap duplicate_symbol_table_entry_error exception

      when Use_Statement  =>
         -- add package_spec basic_decl_items to symbol_table
         --    while checking for conflicts
         -- [i.e. make valid ones visible]
         declare
            List_Node : Ref_Statement_Database := Get_Name_List (Node);
         begin
            while (not Is_Empty (List_Node)) loop
               begin
                  Add_Package_Decls_To_Symbol_Table
                      (Context,
                       Get_Symbol_Table_Entry (Get_Expression (List_Node)));
               exception
                  when Duplicate_Symbol_Table_Entry_Error    =>
                     -- raised by add_entry_to_symtbl
                     -- [means that package is already USE-ed]
                     null;
               end;
               List_Node := Get_Next_List_Item (List_Node);
            end loop;
         end;

         --------------------

      when Pragma_Statement    =>
         Evaluate_Pragma_Statement (Node);

         --------------------

      when Null_Statement =>
         null; -- nothing to do here

      when If_Statement   =>
         null; -- checks on IF STMT condition already performed

      when Labeled_Statement   =>
         null; -- ignore these

         --------------------

      when Case_Alternative_Part    =>
         -- can't be this kind here!
         Show_Error (1000, Internal, Subprogram_Name);

         --------------------

      when Goto_Statement | For_Statement     =>
         -- not implementing these yet
         Show_Error (1999, Internal, Subprogram_Name);
         -- FOR stmt:
         --   [convert all possible FOR_RANGE kinds to range w/ UB and LB??]

         --------------------

      when Unknown   =>
         Show_Error (1002, Internal, Subprogram_Name);
   end case;

end Check_Statement;
