--|$Log:	csubpgm.a,v $
--|Revision 8.0  91/01/17  10:44:19  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Subunit for Package - Semantic_Analyzer
--
---------------------------------------------------------------
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Ace_Universal_Types;
use Ace_Universal_Types;

--==========================================================================
--  <Subprogram Name>
--
--    Check_Subprogram_Call
--
--  <Description>
--
--    Checks the semantic correctness of procedure and function calls.
--    Performs overload resolution on said subprograms.
--
--    If this routine is called, then at least one subpgm of the
--    same _name_ was found ... subpgm signatures must be checked
--    for them.
--
--  <Exceptions Raised>
--
--    No_Valid_Entry_Found_Error
--
--==========================================================================
separate (Semantic_Analyzer)

procedure Check_Subprogram_Call
    (Context : in out Context_Control.Context_Definition;
     Node    : in Ref_Statement_Database) is

   Subprogram_Name            : Ref_String :=
      Get_Symbolic_Name (Get_Subprogram_Info (Node));
   Subprogram_Symtbl_Entry    : Ref_Symbol_Table_Entry :=
      Get_Symbol_Table_Entry (Get_Subprogram_Info (Node));
   Is_Function_Call : Boolean := (Get_Entry_Kind (Subprogram_Symtbl_Entry) =
                                 Function_Entry);

   No_Valid_Entry_Found_Error : exception;

   Actual_Argument            : Ref_Statement_Database;
   Argument_Count             : Ace_Natural := 0;

   Formal_Parameter           : Ref_Symbol_Table_Entry;

   -- structures for list of all possible subpgms entries which
   --   might match this call
   type Possible_List_Node;
   type Ref_Possible_List_Node is access Possible_List_Node;
   type Possible_List_Node is
      record
         Subpgm_Entry      : Ref_Symbol_Table_Entry := null;
         Next_Subpgm_Entry : Ref_Possible_List_Node := null;
      end record;

   Good_Match_List                  : Ref_Possible_List_Node;
   Count_Of_Good_Matches            : Ace_Natural := 0;

   Bad_Parameter_Match_List         : Ref_Possible_List_Node;
   Count_Of_Bad_Parameter_Matches   : Ace_Natural := 0;

   Bad_Return_Type_Match_List       : Ref_Possible_List_Node;
   Count_Of_Bad_Return_Type_Matches : Ace_Natural := 0;

   -- --------------------
   function Create_Possible_List_Node return Ref_Possible_List_Node is separate;

   procedure Save_Found_Entry (Node        : in out Ref_Possible_List_Node;
                               Found_Entry : in Ref_Symbol_Table_Entry) is
   separate;

   procedure Check_Infix_Operation (Node : in Ref_Statement_Database) is
   separate;

   procedure Check_Parameters (Symtbl_Entry : in Ref_Symbol_Table_Entry) is
   separate;

   procedure Mark_As_Proper_Entry (Symtbl_Entry : in Ref_Symbol_Table_Entry) is
    separate;

   function Is_One_An_Infix_Operation
       (List_Of_Entries : in Ref_Possible_List_Node) return Boolean is separate;
   -- --------------------

begin

   -----
   -- initialize the lists of overloaded subpgm entries
   Good_Match_List            := Create_Possible_List_Node;
   Bad_Parameter_Match_List   := Create_Possible_List_Node;
   Bad_Return_Type_Match_List := Create_Possible_List_Node;
   -----

   -- [tbd] [named params: reorder args]

   -----
   -- check each actual argument [resolve references to arguments]
   declare
      -- [for functions only]
      -- suggest types from parent to first child, then
      -- from first child to right sibling
      Suggested_Type : Ref_Symbol_Table_Entry;
   begin
      if Is_Function_Call then
         -- initialize suggested type from parent node
         Suggested_Type := Get_Type (Node);
      end if;
      Actual_Argument := Get_Parameter_List (Node);
      while (not Is_Empty (Actual_Argument)) loop
         Argument_Count := Argument_Count + 1;
         if Is_Function_Call then
            -- suggest a possible type that this argument could have
            Set_Type (Get_Expression (Actual_Argument), Suggested_Type);
         end if;
         Check_Expression (Context, Get_Expression (Actual_Argument));
         -----
         -- check that expression is "valid"
         if Check_Expr_Node (Get_Expression (Actual_Argument)) = Terminal_Node
             then
            case Get_Entry_Kind
                (Get_Symbol_Table_Entry (Get_Expression (Actual_Argument))) is
               when Object_Entry       |
                    Formal_Param_Entry |
                    Function_Entry     |
                    Enumeration_Literal_Entry =>
                  -- [tbd] Record_Component_Entry =>
                  null; -- these are ok expressions
               when others     =>
                  Show_Error
                      (1116, Fatal,
                       Convert_To_String
                           (Get_Symbolic_Name
                                (Get_Symbol_Table_Entry
                                     (Get_Expression (Actual_Argument)))));
            end case;
         end if;
         -----
         if Is_Function_Call then
            -- suggest this child's type for later use by its right sibling
            -- (if any)
            Suggested_Type := Get_Type (Get_Expression (Actual_Argument));
         end if;
         Actual_Argument := Get_Next_List_Item (Actual_Argument);
      end loop;
   end;
   -----

   -----
   -- get list of all possible subpgms entries which might match this call
   declare
      Found_Entry : Ref_Symbol_Table_Entry := Subprogram_Symtbl_Entry;
      Current_Good_Entry : Ref_Possible_List_Node := Good_Match_List;
      Current_Bad_Parameter_Entry         : Ref_Possible_List_Node :=
         Bad_Parameter_Match_List;
      Current_Bad_Return_Type_Entry       : Ref_Possible_List_Node :=
         Bad_Return_Type_Match_List;
      Find_Next_Possible_Subprogram_Entry : exception;
   begin
      while (not Is_Empty (Found_Entry)) loop
         begin
            -----
            -- check for match in number of parameters
            if ((Argument_Count = Get_Parameter_Count (Found_Entry)) or
               ((Argument_Count < Get_Parameter_Count (Found_Entry)) and
               Has_Default_Parameters (Found_Entry))) then

               -----
               -- check for match of parameter types and MODEs
               Actual_Argument := Get_Parameter_List (Node);
               Formal_Parameter := Get_Head_Of_Parameters_List (Found_Entry);
               while (not Is_Empty (Actual_Argument)) loop
                  if not Is_Same_Type (Get_Expression (Actual_Argument),
                                       Get_Type (Formal_Parameter)) then
                     -- parameters mismatch
                     -- attach entry to list of bad parameter match entries
                     Save_Found_Entry
                         (Current_Bad_Parameter_Entry, Found_Entry);
                     Count_Of_Bad_Parameter_Matches :=
                        Count_Of_Bad_Parameter_Matches + 1;
                     raise Find_Next_Possible_Subprogram_Entry;
                  end if;
                  -- get next parameter to check
                  Actual_Argument := Get_Next_List_Item (Actual_Argument);
                  Formal_Parameter := Get_Next_Parameter (Formal_Parameter);
               end loop;
               -----

               -----
               -- for default parameters
               --
               if (Argument_Count < Get_Parameter_Count (Found_Entry)) then
                  -- check that remaining formal parameters have
                  -- default init code
                  while (not Is_Empty (Formal_Parameter)) loop
                     if (Is_Empty (Get_Initialization_Code (Formal_Parameter)))
                         then
                        -- formal parameter does not have a default value
                        -- so forget this "found_entry"
                        raise Find_Next_Possible_Subprogram_Entry;
                     end if;
                     Formal_Parameter := Get_Next_Parameter (Formal_Parameter);
                  end loop;
               end if;
               -----

               -----
               -- [for functions only]
               -- check for match of return types
               if Is_Function_Call then
                  -- try to match return type of found entry with
                  --   suggested type for actual function call
                  if not Is_Same_Type (Node, Get_Return_Type (Found_Entry))
                      then
                     -- return types possibly mismatch
                     -- attach entry to list of bad return_type match entries
                     Save_Found_Entry (Current_Bad_Return_Type_Entry,
                                       Found_Entry);
                     Count_Of_Bad_Return_Type_Matches :=
                        Count_Of_Bad_Return_Type_Matches + 1;
                     raise Find_Next_Possible_Subprogram_Entry;
                  end if;
               end if;
               -----

               -----
               -- symtbl entry is a good match so keep it
               Save_Found_Entry (Current_Good_Entry, Found_Entry);
               Count_Of_Good_Matches := Count_Of_Good_Matches + 1;
               -----
            end if;
            -----

            raise Find_Next_Possible_Subprogram_Entry;

         exception
            when Find_Next_Possible_Subprogram_Entry    =>
               Found_Entry :=
                  Find_Next_Symbol_Table_Entry (Found_Entry,
                                                Get_Entry_Kind (Found_Entry));
         end;
      end loop;
   end;
   -----

   -----
   -- try to find the proper match
   if (Count_Of_Good_Matches > 0) then
      if (Count_Of_Good_Matches = 1) then
         -- this is the one!
         Mark_As_Proper_Entry (Good_Match_List.Subpgm_Entry);
         return ;
      else
         -- just pass the first back anyway; other subpgms found
         -- are probably not w/in scope
         -- [tbd] test that modes of params match
         Mark_As_Proper_Entry (Good_Match_List.Subpgm_Entry);
         return ;
         -- ambiguous subpgm specs found
         -- Mark_As_Proper_Entry (null);
         -- Show_Error (1115, Fatal, Convert_To_String (Subprogram_Name));
      end if;

   elsif (Count_Of_Bad_Return_Type_Matches > 0) then
      if (Count_Of_Bad_Return_Type_Matches = 1) then
         -- for now, assume this is the one [tbd] -- fix -- oei
         Mark_As_Proper_Entry (Bad_Return_Type_Match_List.Subpgm_Entry);
         return ;
      else
         -- try to resolve subpgm reference using suggested Node type
         -- [tbd] test that modes of params match
         declare
            Current     : Ref_Possible_List_Node := Bad_Return_Type_Match_List;
            Found_Count : Ace_Natural            := 0;
            Found_Entry : Ref_Symbol_Table_Entry := null; --fix (no null) [tbd]
         begin
            while Current.Subpgm_Entry /= null loop
               if Is_Same_Type (Node, Get_Return_Type (Current.Subpgm_Entry))
                   then
                  Found_Count := Found_Count + 1;
                  Found_Entry := Current.Subpgm_Entry;
               end if;
               Current := Current.Next_Subpgm_Entry;
            end loop;
            if Found_Count = 1 then
               -- Found_Entry is the one!
               Mark_As_Proper_Entry (Found_Entry);
               return ;
            else
               -- cannot resolve combination of overloading yet [tbd]
               Mark_As_Proper_Entry (null);
               Show_Error
                   (1051, Internal,
                    "found >1 (bad) match w/ different return types: " &
                    "could not resolve desired return type (yet): " &
                    Convert_To_String (Subprogram_Name));
            end if;
         end;
      end if;

   elsif (Count_Of_Bad_Parameter_Matches > 0) then
      -- check if call is a valid inheritance use of one of the
      -- predefined infix operations
      if (Is_One_An_Infix_Operation (Bad_Parameter_Match_List)) then
         Check_Infix_Operation (Node);
         return ;
      else
         raise No_Valid_Entry_Found_Error;
      end if;

   else
      raise No_Valid_Entry_Found_Error;
   end if;
   -----

exception

   when No_Valid_Entry_Found_Error  =>
      Mark_As_Proper_Entry (null);
      Show_Error (1112, Fatal, Convert_To_String (Subprogram_Name));

end Check_Subprogram_Call;
