--|$Log:	debug.a,v $
--|Revision 8.0  91/01/17  10:44:23  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - ACE Debugger [body]
--
------------------------------------------------------------
--
--
-- Abstract
--
--  This package contains the routines that perform symbolic
--  debugging in ACE.
--
--
-- Update Log - :
--
-- End of Header
-------------------------------------------------------------------------

with Window_System;
use Window_System;

with Statement_Evaluator;
use Statement_Evaluator;

with Node_Get_Routines;
use Node_Get_Routines;

with Node_Set_Routines;
use Node_Set_Routines;

with Literals;
use Literals;

with Symbol_Table_Routines;
use Symbol_Table_Routines;

with Symtbl_Entry_Get_Routines;
use Symtbl_Entry_Get_Routines;

with Dump_Routines;
use Dump_Routines;

with Ace;
use Ace;

with Ooe;
use Ooe;

with Misc_Support_Routines;
use Misc_Support_Routines;

with Text_Io;
use Text_Io;

with Error_Handler;
use Error_Handler;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Context_Control;
with Unchecked_Deallocation;

package body Debugger is

   -- All routines currently (before the addition of Tasks) should operate
   -- on the Main_Thread_Context (while new routines that are oriented
   -- towards tasks, should operate on any context).
   Context : Context_Control.Context_Definition renames Main_Thread_Context;

   -------
   Place_Observe_Window_Executable : constant String :=
      "place_observe_window.csh";

   Observe_Statements_Filename     : constant String := "STMTS.OBS";
   Observe_Objects_Filename        : constant String := "OBJECTS.OBS";
   Observe_Subprograms_Filename    : constant String := "SUBPGMS.OBS";

   Trace_Statements                : Boolean         := False;
   Observe_Statements              : Boolean         := False;
   Observe_Objects                 : Boolean         := False;
   Observe_Subprograms             : Boolean         := False;
   Observe_Tasks                   : Boolean         := False;
   Symbol_Table_Debugging          : Boolean         := False;
   Overall_Debugging               : Boolean         := False;

   Observe_Statements_File         : File_Type;
   Observe_Objects_File            : File_Type;
   Observe_Subprograms_File        : File_Type;

   Text_Subwindow_Is_Created       : Boolean         := False;

   Interpreting_Within_Subprogram  : Ace_Natural     := Ace_Natural'FIRST;

   -------
   -- flag for determining whether stepping through statement execution
   -- or continuing on as usual
   Is_Stepping_Through_Flag        : Boolean         := False;

   Stepping_Count                  : Ace_Positive    := Ace_Positive'FIRST;
   -------

   -------
   -- flag for determining whether either executing "normally" or as
   -- a result of being w/in a breakpoint
   Within_Breakpoint_Flag          : Boolean         := False;
   -------

   Blanks : constant String := "          ";

   -------
   -- structures for current list of breakpoints set
   --
   type Breakpoint_List_Node;
   type Ref_Breakpoint_List_Node is access Breakpoint_List_Node;
   type Breakpoint_List_Node is
      record
         Statement_Number : Ace_Natural              := Ace_Natural'LAST;
         Next_Breakpoint  : Ref_Breakpoint_List_Node := null;
      end record;

   Head_Of_Breakpoints : Ref_Breakpoint_List_Node := new Breakpoint_List_Node;
   -------

   -------
   -- structures for list of subpgms entries which have been called
   --
   type Entries_List_Node;
   type Ref_Entries_List_Node is access Entries_List_Node;
   type Entries_List_Node is
      record
         Subpgm_Entry      : Ref_Symbol_Table_Entry := null;
         Next_Subpgm_Entry : Ref_Entries_List_Node  := null;
      end record;

   Head_Of_Called_Entries : Ref_Entries_List_Node := null;
   -------

   -- This is a fix to the delay problem with VADS and X on 68030s.
   procedure Really_Do_Delay (Delay_For : Duration) is
   begin
     if Do_Delay then
       delay Delay_For;
     end if;
   end;

   --=======================================================================
   --  <Subprogram Name>
   --
   --    Any_Windows_Created
   --
   --  <Description>
   --
   --  <Exceptions Raised>
   --
   --    none
   --
   --  <Exceptions Handled>
   --
   --    none
   --=======================================================================
   function Any_Windows_Created return Boolean is
   begin
      return Text_Subwindow_Is_Created;
   end Any_Windows_Created;

   --=======================================================================
   --  <Subprogram Name>
   --
   --    Set_Statements_Trace_On
   --    Set_Statements_Trace_Off
   --    Is_Statements_Trace_On
   --
   --  <Description>
   --
   --    These routines turn on, turn off, and query a flag that marks
   --    whether or not the dump routines will output the statement
   --    number that is associated with every statement.
   --
   --  <exceptions raised>
   --
   --    none
   --
   --  <exceptions handled>
   --
   --    none
   --=======================================================================

   procedure Set_Statements_Trace_On is
   begin
      Trace_Statements := True;
   end Set_Statements_Trace_On;

   procedure Set_Statements_Trace_Off is
   begin
      Trace_Statements := False;
   end Set_Statements_Trace_Off;

   function Is_Statements_Trace_On return Boolean is
   begin
      return Trace_Statements;
   end Is_Statements_Trace_On;

   --=======================================================================
   --  <subprogram name>
   --
   --    set_observe_statements_on
   --    set_observe_statements_off
   --    is_statements_observe_on
   --
   --  <description>
   --
   --    these routines control the statements observation window.
   --
   --=======================================================================

   procedure Set_Observe_Statements_On is
   begin
      Observe_Statements := True;
      if Is_Open (Observe_Statements_File) then
         Delete (Observe_Statements_File);
      end if;
      Create (Observe_Statements_File, Out_File, Observe_Statements_Filename);
      Host (Place_Observe_Window_Executable & " statements " &
            Observe_Statements_Filename);
   end Set_Observe_Statements_On;

   procedure Set_Observe_Statements_Off is
   begin
      Observe_Statements := False;
      if Is_Open (Observe_Statements_File) then
         Close (Observe_Statements_File);
      end if;
   end Set_Observe_Statements_Off;

   function Is_Statements_Observe_On return Boolean is
   begin
      return Observe_Statements;
   end Is_Statements_Observe_On;

   --======================================================================
   --  <subprogram name>
   --
   --    set_observe_objects_on
   --    set_observe_objects_off
   --    is_objects_observe_on
   --
   --  <description>
   --
   --    these routines control the objects observation window.
   --
   --=======================================================================
   procedure Set_Observe_Objects_On is
   begin
      Observe_Objects := True;
      if Is_Open (Observe_Objects_File) then
         Delete (Observe_Objects_File);
      end if;
      Create (Observe_Objects_File, Out_File, Observe_Objects_Filename);
      Host (Place_Observe_Window_Executable & " objects " &
            Observe_Objects_Filename);
   end Set_Observe_Objects_On;

   procedure Set_Observe_Objects_Off is
   begin
      Observe_Objects := False;
      if Is_Open (Observe_Objects_File) then
         Close (Observe_Objects_File);
      end if;
   end Set_Observe_Objects_Off;

   function Is_Objects_Observe_On return Boolean is
   begin
      return Observe_Objects;
   end Is_Objects_Observe_On;

   --=======================================================================
   --  <subprogram name>
   --
   --    set_observe_subprograms_on
   --    set_observe_subprograms_off
   --    is_subprograms_observe_on
   --
   --  <description>
   --
   --    these routines control the subprograms observation window.
   --
   --======================================================================
   procedure Set_Observe_Subprograms_On is
   begin
      if Running_Under_Windowing_System then
         Observe_Subprograms := True;
         if not Text_Subwindow_Is_Created then
            Create_Observe_Subprograms_Window;
            Display_Line ("---- ===========================================");
            Display_Line ("   ");
            Text_Subwindow_Is_Created := True;
         end if;
      else
         Show_Error (1987, Warning, "pragma observe");
      end if;
   end Set_Observe_Subprograms_On;

   procedure Set_Observe_Subprograms_Off is
   begin
      Observe_Subprograms := False;
      -- if is_open (observe_subprograms_file) then
      -- close (observe_subprograms_file);
      -- end if;
   end Set_Observe_Subprograms_Off;

   function Is_Subprograms_Observe_On return Boolean is
   begin
      return Observe_Subprograms;
   end Is_Subprograms_Observe_On;

   --=======================================================================
   --  <subprogram name>
   --
   --    set_symbol_table_debugging_on
   --    set_symbol_table_debugging_off
   --    is_symbol_table_debugging_on
   --
   --  <description>
   --
   --    these routines control the whether or not the symbol_table is
   --    set up (retaining symbolic information) for debugging.
   --
   --======================================================================
   procedure Set_Symbol_Table_Debugging_On is
   begin
      Symbol_Table_Debugging := True;
   end Set_Symbol_Table_Debugging_On;

   procedure Set_Symbol_Table_Debugging_Off is
   begin
      Symbol_Table_Debugging := False;
   end Set_Symbol_Table_Debugging_Off;

   function Is_Symbol_Table_Debugging_On return Boolean is
   begin
      return Symbol_Table_Debugging and Overall_Debugging;
   end Is_Symbol_Table_Debugging_On;

   --=======================================================================
   --  <subprogram name>
   --
   --    set_overall_debugging_on
   --    set_overall_debugging_off
   --    is_overall_debugging_on
   --
   --  <description>
   --
   --    these routines control the whether or not the ace is
   --    set up (retaining symbolic information) for debugging.
   --======================================================================

   procedure Set_Overall_Debugging_On is
   begin
      Set_Symbol_Table_Debugging_On;
      Overall_Debugging := True;
   end Set_Overall_Debugging_On;

   procedure Set_Overall_Debugging_Off is
   begin
      Overall_Debugging := False;
   end Set_Overall_Debugging_Off;

   function Is_Overall_Debugging_On return Boolean is
   begin
      return Overall_Debugging;
   end Is_Overall_Debugging_On;

   --=======================================================================
   --  <subprogram name>
   --
   --    observe_statement_if_statements_observation_is_on
   --    observe_object_if_objects_observation_is_on
   --    observe_subprogram_if_subprogram_observation_is_on
   --
   --  <description>
   --
   --    these routines query which obeervation windows are currently
   --    active.
   --
   --======================================================================
   procedure Observe_Statement_If_Statements_Observation_Is_On
       (Node : in Ref_Statement_Database) is
      Was_On : Boolean := False;
   begin
      if Is_Statements_Observe_On then
         Dump_Statement_Numbers (True);
         Dump_For_Statement_Trace (Node, Observe_Statements_File);
         Dump_Statement_Numbers (False);
      end if;
      Observe_Subprogram_If_Subprogram_Observation_Is_On
          (Node);
   end Observe_Statement_If_Statements_Observation_Is_On;

   procedure Observe_Object_If_Objects_Observation_Is_On
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) is
   begin
      if Is_Objects_Observe_On then
         Dump_Symtbl (Symtbl_Entry, Observe_Objects_File);
      end if;
   end Observe_Object_If_Objects_Observation_Is_On;

   procedure Observe_Subprogram_If_Subprogram_Observation_Is_On
       (Node : in Ref_Statement_Database) is
   begin
      if Is_Subprograms_Observe_On and
         (Interpreting_Within_Subprogram > Ace_Natural'FIRST) then
         Show_Statement
             (Blanks (1 ..
                      5 -
                      Ace_Integer'IMAGE (Get_Statement_Number (Node))'LENGTH)
               & Ace_Natural'IMAGE (Get_Statement_Number (Node)) & "|");
         Really_Do_Delay (0.5);
         Synchonize_With_Textwindow;
      end if;
   end Observe_Subprogram_If_Subprogram_Observation_Is_On;

   --=======================================================================
   --  <subprogram name>
   --
   --    initialize_statements_observation_if_statements_observation_is_on1
   --    reset_statements_observation_if_statements_observation_is_on
   --    is_subprograms_observe_on
   --
   --  <description>
   --
   --    these routines control the statements observation window if its
   --    activity is toggled (i.e., turned on and off).
   --
   --======================================================================
   procedure Initialize_Statements_Observation_If_Statements_Observation_Is_On
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      function Is_On_List
          (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is
         Current : Ref_Entries_List_Node := Head_Of_Called_Entries;
      begin
         while Current.Next_Subpgm_Entry /= null loop
            if Current.Subpgm_Entry = Symtbl_Entry then
               return True;
            end if;
            Current := Current.Next_Subpgm_Entry;
         end loop;
         return False;
      end Is_On_List;

   begin
      if Is_Subprograms_Observe_On then
         Interpreting_Within_Subprogram := Interpreting_Within_Subprogram + 1;
         if not Is_On_List (Symtbl_Entry) then
            Head_Of_Called_Entries :=
               new Entries_List_Node'
                  (Subpgm_Entry      => Symtbl_Entry,
                   Next_Subpgm_Entry => Head_Of_Called_Entries);

            Dump_To_Text_Windows (True);
            Dump_Statement_Numbers (True);
            Dump_Symtbl (Symtbl_Entry, Standard_Output, Implementation);
            -- dump bodies
            Dump_Statement_Numbers (False);
            Dump_To_Text_Windows (False);

            Display_Line ("   ");
            Really_Do_Delay (0.1);
            Display_Line ("---- ===========================================");
            Really_Do_Delay (0.1);


            Display_Line ("   ");
            Really_Do_Delay (0.1);

            -- end_of_displaying_lines;
         end if;
      end if;
   end Initialize_Statements_Observation_If_Statements_Observation_Is_On;

   procedure Reset_Statements_Observation_If_Statements_Observation_Is_On
       is
   begin
      if Is_Subprograms_Observe_On then
         Interpreting_Within_Subprogram := Interpreting_Within_Subprogram - 1;
      end if;
   end Reset_Statements_Observation_If_Statements_Observation_Is_On;


   procedure Remove_Entry_From_List (Symtbl_Entry : Ref_Symbol_Table_Entry) is

      Current    : Ref_Entries_List_Node := Head_Of_Called_Entries;
      Prev, Temp : Ref_Entries_List_Node;
      procedure Free is new Unchecked_Deallocation (Entries_List_Node,
                                                    Ref_Entries_List_Node);
   begin
      if (Is_Subprograms_Observe_On and then Current /= null) then
         if Current.Subpgm_Entry = Symtbl_Entry then
         -- First entry is what we want to zap, so reset head and blow away node
            Temp                   := Current.Next_Subpgm_Entry;
            Free (Current);
            Head_Of_Called_Entries := Temp;
         else

            Prev    := Current;
            Current := Current.Next_Subpgm_Entry;
            while Current /= null loop
               if Current.Subpgm_Entry = Symtbl_Entry then
                  Temp                   := Current.Next_Subpgm_Entry;
                  Free (Current);
                  Prev.Next_Subpgm_Entry := Temp;
               else
                  Prev    := Current;
                  Current := Current.Next_Subpgm_Entry;
               end if;
            end loop;

         end if;
      end if;
   end Remove_Entry_From_List;


   -- ====================================================================
   -- workhorse of Symbolic Debugger Routines
   -- ====================================================================

   procedure Break (Current_Statement : in Ref_Statement_Database;
                    Notice_Message    : String := "Breakpoint encountered") is
   begin
      if (not Is_Stepping_Through) then
         Put_Line (Notice_Message);
      end if;
      -- Misc.Prompt ("ACE" & Ace_Natural'Image(Call_Level) & "> ");

      -- state what is the (real/last) current statement being executed
      Context_Control.Set_Current_Statement (Context, Current_Statement);

      Within_Breakpoint_Flag := True;

      if (Is_Stepping_Through) then
         if (Stepping_Count = Ace_Positive'FIRST) then
            -- turn off until explicitly turned on
            Is_Stepping_Through_Flag := False;
         else
            Stepping_Count := Stepping_Count - 1;
            return ;
         end if;
      else
         Ace_Level := Ace_Level + 1;
      end if;

      Ace_Main;

   end Break;

   procedure Continue is
   begin
      -- if (Call_Level = Ace_Natural'FIRST+1) then
      -- Misc.Prompt ("ACE> ");
      -- else
      -- Misc.Prompt ("ACE" & Ace_Natural'Image(Call_Level-1) & "> ");
      -- end if;

      Within_Breakpoint_Flag := False;

      if not (Is_Stepping_Through_Flag or Ace_Level = 0) then

         Ace_Level := Ace_Level - 1;

      end if;

      raise End_Break;

   end Continue;

   function Is_Breakpoint_Set (Node : in Ref_Statement_Database) return Boolean
                              is
   begin
      if Is_Statement_Node (Node) then
         return Should_Suspend_Evaluation_At_This_Statement (Node);
      else
         -- TBD (Can't break at expressions)
         return False;
      end if;
   end Is_Breakpoint_Set;

   procedure Set_Break (Statement_Number : in Ace_Natural) is
      New_Stmt : Ref_Breakpoint_List_Node := new Breakpoint_List_Node;
   begin
      -- mark stmt_db node
      Set_Suspend_Evaluation_At_This_Statement (Stmt_List (Statement_Number));
      -- add new stmt to "break at" to list
      New_Stmt.Statement_Number           := Statement_Number;
      New_Stmt.Next_Breakpoint := Head_Of_Breakpoints.Next_Breakpoint;
      Head_Of_Breakpoints.Next_Breakpoint := New_Stmt;
   end Set_Break;

   -- name is a label or subprogram name
   procedure Set_Break (Name : in String) is
      Found_Subprogram, Found_Label : Ref_Symbol_Table_Entry;
   begin

      -- [TBD]:
      -- verify that named item is only either a label or subprogram [tbd]
      -- if subprogram, check for overloaded names [tbd]

      -- get stmt # for named stmt
      Found_Subprogram :=
         Find_Symbol_Table_Entry (Context, Convert_To_Ref_String (Name));

      if (Is_Empty (Found_Subprogram)) then
         Show_Error (1514, Warning, Name);
         return ;
      end if;

      -- check if label or subprogram name
      if (Get_Entry_Kind (Found_Subprogram) = Label_Entry) then
         -- it's a label
         Set_Break
             (Get_Statement_Number
                  (Get_Head_Of_Statements_List (Found_Subprogram)));
      elsif ((Get_Entry_Kind (Found_Subprogram) = Function_Entry) or
            (Get_Entry_Kind (Found_Subprogram) = Procedure_Entry)) then
         -- if a subprogram, record breakpoint at first decl or executable stmt
         if (Is_Empty (Get_Head_Of_Declarations_List (Found_Subprogram))) then
            if (Is_Empty (Get_Head_Of_Statements_List (Found_Subprogram))) then
               Show_Error (1512, Warning, "Set_Break");
               return ;
            else
               Set_Break
                   (Get_Statement_Number
                        (Get_Head_Of_Statements_List (Found_Subprogram)));
            end if;
         else
            Set_Break
                (Get_Statement_Number
                     (Get_Head_Of_Declarations_List (Found_Subprogram)));
         end if;
      else
         -- actually can break at 1st stmt of a package body [tbd]
         Show_Error (1514, Warning,
                     Convert_To_String (Get_Symbolic_Name (Found_Subprogram)));
         return ;
      end if;
   end Set_Break;

   -------
   procedure Clear_Break (Statement_Number : in Ace_Natural) is
      Current : Ref_Breakpoint_List_Node := Head_Of_Breakpoints.Next_Breakpoint;
      Trail   : Ref_Breakpoint_List_Node := Head_Of_Breakpoints;
   begin
      if Is_Breakpoint_Set (Stmt_List (Statement_Number)) then
         -- unmark stmt_db node
         Unset_Suspend_Evaluation_At_This_Statement
             (Stmt_List (Statement_Number));
         -- remove stmt # from breakpoint list
         while Current /= null loop
            if (Statement_Number = Current.Statement_Number) then
               -- remove "node" from list
               Trail.Next_Breakpoint := Current.Next_Breakpoint;
               -- free (current); -- null out next_breakpoint field first
               return ;
            end if;
            Trail   := Current;
            Current := Current.Next_Breakpoint;
         end loop;
      else
         Show_Error (1511, Warning, Ace_Natural'IMAGE (Statement_Number));
      end if;
   end Clear_Break;

   -- name is a label or subprogram name
   procedure Clear_Break (Name : in String) is
      Found_Subprogram, Found_Label : Ref_Symbol_Table_Entry;
   begin

      -- [TBD]:
      -- verify that named item is only either a label or subprogram [tbd]
      -- if subprogram, check for overloaded names [tbd]

      -- get stmt # for named stmt
      Found_Subprogram :=
         Find_Symbol_Table_Entry (Context, Convert_To_Ref_String (Name));

      if (Is_Empty (Found_Subprogram)) then
         Show_Error (1515, Warning, Name);
         return ;
      end if;

      -- check if label or subprogram name
      if (Get_Entry_Kind (Found_Subprogram) = Label_Entry) then
         -- it's a label
         Clear_Break
             (Get_Statement_Number
                  (Get_Head_Of_Statements_List (Found_Subprogram)));
      elsif ((Get_Entry_Kind (Found_Subprogram) = Function_Entry) or
            (Get_Entry_Kind (Found_Subprogram) = Procedure_Entry)) then
         -- if a subprogram, record breakpoint at first decl or executable stmt
         if (Is_Empty (Get_Head_Of_Declarations_List (Found_Subprogram))) then
            if (Is_Empty (Get_Head_Of_Statements_List (Found_Subprogram))) then
               Show_Error (1512, Warning, "Clear_Break");
               return ;
            else
               Clear_Break
                   (Get_Statement_Number
                        (Get_Head_Of_Statements_List (Found_Subprogram)));
            end if;
         else
            Clear_Break
                (Get_Statement_Number
                     (Get_Head_Of_Declarations_List (Found_Subprogram)));
         end if;
      else
         -- actually can clear_break at 1st stmt of a package body [tbd]
         Show_Error (1515, Warning,
                     Convert_To_String (Get_Symbolic_Name (Found_Subprogram)));
         return ;
      end if;
   end Clear_Break;

   -------
   procedure Display (Statement_Number : in Ace_Natural) is
   begin
      -- get stmt node for given stmt_# and display
      Dump_Statement_Numbers (True);
      Dump_For_Statement_Trace (Stmt_List (Statement_Number),
                                Text_Io.Standard_Output);
      Dump_Statement_Numbers (False);
      Context_Control.Set_Current_Statement (Context,
                                             Stmt_List (Statement_Number));
   end Display;

   procedure Display (Name : in String) is
      Found_Subprogram, Found_Label : Ref_Symbol_Table_Entry;
   begin

      -- [TBD]:
      -- verify that named item is only either a label or subprogram [tbd]
      -- if subprogram, check for overloaded names [tbd]

      -- get stmt # for named stmt
      Found_Subprogram :=
         Find_Symbol_Table_Entry (Context, Convert_To_Ref_String (Name));

      if (Is_Empty (Found_Subprogram)) then
         Show_Error (1513, Warning, Name);
         return ;
      end if;

      -- check if label or subprogram name
      if (Get_Entry_Kind (Found_Subprogram) = Label_Entry) then
         -- it's a label
         Display
             (Get_Statement_Number
                  (Get_Head_Of_Statements_List (Found_Subprogram)));
      elsif ((Get_Entry_Kind (Found_Subprogram) = Function_Entry) or
            (Get_Entry_Kind (Found_Subprogram) = Procedure_Entry)) then
         -- if a subprogram, record breakpoint at first decl or executable stmt
         if (Is_Empty (Get_Head_Of_Declarations_List (Found_Subprogram))) then
            if (Is_Empty (Get_Head_Of_Statements_List (Found_Subprogram))) then
               Show_Error (1512, Warning, "Display");
               return ;
            else
               Display
                   (Get_Statement_Number
                        (Get_Head_Of_Statements_List (Found_Subprogram)));
            end if;
         else
            Display
                (Get_Statement_Number
                     (Get_Head_Of_Declarations_List (Found_Subprogram)));
         end if;
      else
         -- actually can display at 1st stmt of a package body [tbd]
         Show_Error (1513, Warning,
                     Convert_To_String (Get_Symbolic_Name (Found_Subprogram)));
         return ;
      end if;
   end Display;

   procedure Display_Next (Reference_Statement_Number : in Ace_Natural) is
      -- get stmt node for given stmt_#
      Next_Statement : Ref_Statement_Database :=
         Get_Next_Statement (Stmt_List (Reference_Statement_Number));
   begin
      if Is_Empty (Next_Statement) then
         Show_Error (1510, Warning, "Display_Next");
         return ;
      else
         -- display based on its stmt_#
         Display (Get_Statement_Number (Next_Statement));
      end if;
   end Display_Next;

   procedure Display_Previous (Reference_Statement_Number : in Ace_Natural) is
   begin
      -- do similar error chking code as display next above [oei]
      -- get stmt node for given stmt_# and display previous
      -- [this means that parent node must be hooked up (?)]
      null;
   end Display_Previous;

   -------
   procedure Dump_Call_Stack is
   begin
      null;
   end Dump_Call_Stack;

   -- ----------------------------------------------------------------------
   function Is_Stepping_Through return Boolean is
   begin
      return Is_Stepping_Through_Flag;
   end Is_Stepping_Through;

   -- ----------------------------------------------------------------------
   procedure Step (Count : Ace_Positive := 1) is
      -- ignore statement_number formal param for now
   begin
      Stepping_Count           := Count;
      Is_Stepping_Through_Flag := True;
      Continue;
   end Step;

   -- ----------------------------------------------------------------------
   procedure List_Breakpoints is
      Current : Ref_Breakpoint_List_Node := Head_Of_Breakpoints.Next_Breakpoint;
   begin
      -- look through list of breakpoints ...
      while Current /= null loop
         Dump_Statement_Numbers (True);
         Dump_For_Statement_Trace (Stmt_List (Current.Statement_Number),
                                   Text_Io.Standard_Output);
         Dump_Statement_Numbers (False);
         Current := Current.Next_Breakpoint;
      end loop;
   end List_Breakpoints;

   -- ----------------------------------------------------------------------
   function Within_Breakpoint return Boolean is
   begin
      return Within_Breakpoint_Flag;
   end Within_Breakpoint;

   -- ----------------------------------------------------------------------

   procedure Set_Observe_Tasks_On is
   begin
      if Running_Under_Windowing_System then
         Observe_Tasks := True;
         if not Text_Subwindow_Is_Created then
            --            Create_Observe_Tasks_Window;
            Display_Line ("---- ===========================================");
            Display_Line ("   ");
            Text_Subwindow_Is_Created := True;
         end if;
      else
         Show_Error (1987, Warning, "pragma observe");
      end if;
   end Set_Observe_Tasks_On;

   procedure Set_Observe_Tasks_Off is
   begin
      Observe_Tasks := False;
      -- if is_open (observe_Tasks_file) then
      -- close (observe_Tasks_file);
      -- end if;
   end Set_Observe_Tasks_Off;

   function Is_Tasks_Observe_On return Boolean is
   begin
      return Observe_Tasks;
   end Is_Tasks_Observe_On;


begin
   Head_Of_Called_Entries := new Entries_List_Node'(Subpgm_Entry      => null,
                                                    Next_Subpgm_Entry => null);

end Debugger;
