--|$Log:	dir.a,v $
--|Revision 8.0  91/01/17  10:44:46  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Directory_Objects
--
---------------------------------------------------------------
--
--
-- Abstract
--
--  Directory_Objects is an interface from ACE's environment to operations
--  that make directories "look like" files.  On Unix this can, be assumed
--  to be true, and these routines are not needed, but on msdos ando other
--  machines this may not be true.  In any case, we retain this layer
--  for Unix for consistancy.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------
with System_Dependent_Routines;
with Literals;
package body Directory_Objects is
   use System_Dependent_Routines;
   use Literals;

   -- Hold the Name; Currently the Form will always be empty.
   Directory_Storage : array (Directory_Object) of Ref_String;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create.
   --
   --  <Description>
   --
   --    Create a named directory, and return an object associated with
   --    that directory.
   --
   --  <Called By>
   --
   --    Various
   --
   --  <Exceptions Raised>
   --
   --    Same as Text_Io's exceptions
   --
   --  <Exceptions Handled>
   --
   --    None
   --==========================================================================
   procedure Create (Directory : in out Directory_Object;
                     Name      : in String := "";
                     Form      : in String := "") is
   begin

      if Form /= "" then

         -- No Form string for now.
         raise User_Error;

      end if;

      for I in Directory_Object'FIRST .. Directory_Object'LAST - 1 loop

         if not Is_Open (I) then

            System_Dependent_Routines.Mkdir (Convert_To_Ref_String (Name));
            Directory_Storage (I) := Convert_To_Ref_String (Name);
            Directory := I;
            return ;

         end if;

      end loop;

      raise Use_Error;

   end Create;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Open
   --
   --  <Description>
   --
   --    Open associates an object with an existing directory.
   --
   --  <Called By>
   --
   --    Various
   --
   --  <Exceptions Raised>
   --
   --    The same as Text_Io's open.
   --
   --  <Exceptions Handled>
   --
   --    None.
   --==========================================================================
   procedure Open (Directory : in out Directory_Object;
                   Name      : in String;
                   Form      : in String := "") is
   begin

      if Form /= "" then

         -- No Form string for now.
         raise User_Error;

      end if;

      for I in Directory_Object'FIRST .. Directory_Object'LAST - 1 loop

         if not Is_Open (I) then

            Directory_Storage (I) :=
               Convert_To_Ref_String (Get_Directory_Name (Name));
            Directory := I;
            return ;

         end if;

      end loop;

      raise Use_Error;

   end Open;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Close
   --
   --  <Description>
   --
   --    Close disassociates an object from a directory file.
   --
   --  <Called By>
   --
   --    Various
   --
   --  <Exceptions Raised>
   --
   --    The same as Text_Io's name.
   --
   --  <Exceptions Handled>
   --
   --    None.
   --==========================================================================
   procedure Close (Directory : in out Directory_Object) is
   begin
      if Is_Open (Directory) then
         Directory_Storage (Directory) := Create;
         Directory                     := Directory_Object'LAST;
      else
         raise Status_Error;
      end if;
   end Close;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Name
   --
   --  <Description>
   --
   --    From a directory object return the name of the directory as a
   --    String.
   --
   --  <Called By>
   --
   --    Various
   --
   --  <Exceptions Raised>
   --
   --    The same as Text_Io's name.
   --
   --  <Exceptions Handled>
   --
   --    None.
   --==========================================================================
   function Name (Directory : in Directory_Object) return String is
   begin
      if Is_Open (Directory) then
         return Convert_To_String (Directory_Storage (Directory));
      else
         raise Status_Error;
      end if;
   end Name;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Form
   --
   --  <Description>
   --
   --    Return the Form String that was used in the Open or Create that
   --    was used to associated the object with an existing object.
   --
   --  <Called By>
   --
   --    Various
   --
   --  <Exceptions Raised>
   --
   --    Same as Text_Io's Form.
   --
   --  <Exceptions Handled>
   --
   --
   --==========================================================================
   function Form (Directory : in Directory_Object) return String is
   begin
      if Is_Open (Directory) then
         return "";
      else
         raise Status_Error;
      end if;
   end Form;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Open
   --
   --  <Description>
   --
   --    This routine returns a boolean true if the object has been
   --    associated with a directory.
   --
   --  <Called By>
   --
   --    Various
   --
   --  <Exceptions Raised>
   --
   --    Same as Text_Io's Is_Open.
   --
   --  <Exceptions Handled>
   --
   --    None.
   --==========================================================================
   function Is_Open (Directory : in Directory_Object) return Boolean is
   begin
      if Directory_Storage (Directory) /= Create then
         return True;
      else
         return False;
      end if;
   exception
      when Constraint_Error    =>
         return False;
   end Is_Open;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Home_Directory
   --
   --  <Description>
   --
   --    Returns an object that has been associated with the home
   --    directory (defined as the directory in which ACE was originally
   --    started).
   --
   --  <Called By>
   --
   --    Various
   --
   --  <Exceptions Raised>
   --
   --    None.
   --
   --  <Exceptions Handled>
   --
   --    None.
   --==========================================================================
   function Home_Directory return Directory_Object is
   begin
      return System_Dependent_Routines.Home_Directory;
   end Home_Directory;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Current_Directory
   --
   --  <Description>
   --
   --    Returns an object that has been associated with the current
   --    working directory.
   --
   --  <Called By>
   --
   --    Various
   --
   --  <Exceptions Raised>
   --
   --    None.
   --
   --  <Exceptions Handled>
   --
   --    None.
   --==========================================================================
   function Current_Directory return Directory_Object is
   begin
      return System_Dependent_Routines.Current_Directory;
   end Current_Directory;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    List
   --
   --  <Description>
   --
   --    List will output a list of files that are contained in the
   --    directory object.
   --
   --  <Called By>
   --
   --    Various
   --
   --  <Exceptions Raised>
   --
   --    Same as Text_Io's Form.
   --
   --  <Exceptions Handled>
   --
   --    None.
   --==========================================================================
   procedure List (Directory : in Directory_Object) is
   begin
      if Is_Open (Directory) then
         System_Dependent_Routines.Directory_Lister
             (Directory_Storage (Directory));
      else
         raise Status_Error;
      end if;
   end List;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Directory
   --
   --  <Description>
   --
   --    This routines will change the current directory to the specified
   --    directory.
   --
   --  <Called By>
   --
   --    Various
   --
   --  <Exceptions Raised>
   --
   --    Same as Text_Io's Form.
   --
   --  <Exceptions Handled>
   --
   --    None.
   --==========================================================================
   procedure Set_Directory (Directory : in Directory_Object) is
   begin
      if Is_Open (Directory) then
         System_Dependent_Routines.Chdir (Directory_Storage (Directory));
      else
         raise Status_Error;
      end if;
   end Set_Directory;

end Directory_Objects;
