--|$Log:	dpkg.a,v $
--|Revision 8.0  91/01/17  10:44:47  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Subunit for Package - Declaration_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--   The following routines are separate routines for the
--   package Declaration_Routines.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------
-- --------------------------------------------------------------------------
-- INTERNAL ROUTINE
-- --------------------------------------------------------------------------
--==========================================================================
--  <Subprogram Name>
--
--    Mark_Items_With_Parent_Package
--
--  <Description>
--
--   Each item has its symbol table entry marked such that it refers to
--   its parent package (the enclosing package).
--
--
--==========================================================================
separate (Declaration_Routines)

procedure Mark_Items_With_Parent_Package
    (Items          : in Ref_Statement_Database;
     Parent_Package : in Ref_Symbol_Table_Entry) is

   Current           : Ref_Statement_Database := Items;
   Type_Symtbl_Entry : Ref_Symbol_Table_Entry;

begin
   -- For each declaration within the items list, set the enclosing
   -- package field for each symbol table entry

   while (not (Is_Empty (Current))) loop
      case Check_Stmt_Node (Current) is

         when Object_Declaration    =>
            -- Set the pointer to the parent for all objects in the object
            -- declaration
            declare
               List_Node : Ref_Statement_Database := Get_Name_List (Current);
            begin
               while (not Is_Empty (List_Node)) loop
                  Set_Enclosing_Package
                      (Get_Symbol_Table_Entry (Get_Expression (List_Node)),
                       Parent_Package);
                  List_Node := Get_Next_List_Item (List_Node);
               end loop;
            end;

         when Type_Declaration =>
            -- For enumeration types, set the parent package for each
            -- enumeration literal
            Type_Symtbl_Entry := Get_Symbol_Table_Entry (Current);

            if Get_Entry_Kind (Type_Symtbl_Entry) = Type_Entry and then
               Is_Type_An_Enumeration_Type (Type_Symtbl_Entry) then
               declare
                  Literal_Entry : Ref_Symbol_Table_Entry :=
                     Get_First_Literal (Type_Symtbl_Entry);
               begin
                  while Literal_Entry /= null loop
                     Set_Enclosing_Package (Literal_Entry, Parent_Package);
                     Literal_Entry := Get_Successor (Literal_Entry);
                  end loop;
               end; -- block
            end if;

            -- Set the enclosing package for the type
            Set_Enclosing_Package (Type_Symtbl_Entry, Parent_Package);

         when Subtype_Declaration        |
              Package_Declaration        |
              Package_Body_Declaration   |
              Procedure_Spec_Declaration |
              Procedure_Declaration      |
              Function_Spec_Declaration  |
              Function_Declaration  =>
            Set_Enclosing_Package (Get_Symbol_Table_Entry (Current),
                                   Parent_Package);

         when others =>
            null; -- ignore

      end case;

      Current := Get_Next_Statement (Current);
   end loop;

end Mark_Items_With_Parent_Package;

-- --------------------------------------------------------------------------
-- END INTERNAL ROUTINE
-- --------------------------------------------------------------------------

-- --------------------------------------------------------------------------
--==========================================================================
--  <Subprogram Name>
--
--    Start_Of_Package
--
--  <Description>
--
--   The validity of the name of the package is checked and appropriate
--   actions are executed.  For a package specification, a symbol table
--   entry for the package is created.   For a package body, the
--   specification for the package is located and the symbol table
--   entries associated with the package spec are pushed onto the
--   symbol table.
--
--
--==========================================================================
with Context_Control;
with Help_Pkg;
separate (Declaration_Routines)

procedure Start_Of_Package (Context : in out Context_Control.Context_Definition;
                            Id_Node : in Ref_Statement_Database;
                            Is_Body : in Boolean;
                            Tree    : out Ref_Statement_Database) is

   use Help_Pkg;

   Package_Symtbl_Entry : Ref_Symbol_Table_Entry;
   Top_Of_Symbol_Table  : Ref_Symbol_Table_Entry;
   Package_Name         : Ref_String;

begin
   -- Extract the name of the package
   Package_Name := Get_Symbolic_Name (Id_Node);

   -- Save TOS
   Top_Of_Symbol_Table := Get_Top_Of_Symtbl (Context);

   -- For a package spec, perform a name clash.  For a package body,
   -- locate the package spec associated with the package body.
   if not Is_Body then
      if Name_Clash (Context, Package_Name) then
         --E user error: redefinition of existing identifier
         Show_Error (2001, Fatal, Convert_To_String (Package_Name));
      end if;

   else
      -- (Locate_Package_Spec handles all error cases)
      Package_Symtbl_Entry := Locate_Package_Spec (Context, Package_Name);

   end if;

   -- Build a symbol table entry for the package spec.  (For a package body
   -- will use the alread created entry for the package spec.)
   -- Also, set the level of the package name.  Add the symbol table
   -- entry for the package spec onto the symbol table.  (Do not need to add
   -- it for a package body because the package spec symbol table entry
   -- is already on the symbol table.)
   if not Is_Body then
      Package_Symtbl_Entry := Create_Symbol_Table_Entry (Package_Name,
                                                         Package_Entry);
      -- test out associating help here
      Set_Help (Package_Symtbl_Entry);

      Set_Level (Package_Symtbl_Entry, Get_Current_Level (Context));
   end if;

   -- Push a new environment for the package
   Push_Declarative_Region (Context, Package_Symtbl_Entry);

   -- Set error recovery marker to the top of the symbol table
   -- before the package spec.
   Symbol_Table_Routines.Set_Saved_Main_Level_Symbol_Table_Head
       (Context, Top_Of_Symbol_Table);

   -- For a body, push the entries associated with the package spec
   -- onto the symbol table
   if Is_Body then
      Add_Package_Decls_To_Symbol_Table
          (Context,
           Package_Symtbl_Entry => Package_Symtbl_Entry,
           Process_Use_Stmts    => True);
   end if;

   -- Create and return a statement database node that contains the symbol
   -- table entry for the package
   Tree := Create_Entry_Ref_Node (Package_Symtbl_Entry);

end Start_Of_Package;

-- --------------------------------------------------------------------------
with Derived_Subprogram_Support;
use Derived_Subprogram_Support;

--==========================================================================
--  <Subprogram Name>
--
--    Create_Package_Spec_Decl
--
--  <Description>
--
--   The package spec declaration node is built for the package spec.
--   All the relevant information associated with the package spec
--   is inserted into the symbol table entry for the package.
--
--
--==========================================================================
separate (Declaration_Routines)

procedure Create_Package_Spec_Decl
    (Context        : in out Context_Control.Context_Definition;
     Package_Node   : in Ref_Statement_Database;
     Spec_Decl_Node : in Ref_Statement_Database;
     End_Label_Node : in Ref_Statement_Database;
     Tree           : out Ref_Statement_Database) is

   -- Notes:
   --   End_Label_Node: may be null

   Stmt_Db_Node         : Ref_Statement_Database;

   Package_Symtbl_Entry : Ref_Symbol_Table_Entry;

begin
   -- If a designator occurred after the END keyword, check to see that
   -- it matches the designator of the subprogram
   Package_Symtbl_Entry := Get_Symbol_Table_Entry (Package_Node);
   if not Is_Empty (End_Label_Node) and then
      not Are_Symbolic_Names_Equal (Get_Symbolic_Name (Package_Symtbl_Entry),
                                    Get_Symbolic_Name (End_Label_Node)) then
      --E user error: name after "end" doesn't match package name
      Show_Error (2011, Warning,
                  Convert_To_String (Get_Symbolic_Name (Package_Symtbl_Entry)));
   end if;

   -- Store the declarations associated with the package specification
   -- into the symbol table entry for the package
   if not Is_Empty (Spec_Decl_Node) then
      Set_Package_Spec_Decls (Package_Symtbl_Entry, Spec_Decl_Node);
   end if;

   -- Mark each declaration item within the package specification to
   -- refer back to the package name
   Mark_Items_With_Parent_Package (Spec_Decl_Node, Package_Symtbl_Entry);

   -- Determine the subprograms within this package spec that may be
   -- derivable
   Determine_Derivable_Subprograms (Package_Symtbl_Entry);

   -- Pop the environment for the package
   Pop_Declarative_Region (Context);

   -- Constract a package spec declaration node
   Stmt_Db_Node := Create_Node (Package_Declaration);

   Set_Symbol_Table_Entry (Stmt_Db_Node, Package_Symtbl_Entry);

   Tree := Stmt_Db_Node;

end Create_Package_Spec_Decl;

-- --------------------------------------------------------------------------
--==========================================================================
--  <Subprogram Name>
--
--    Create_Package_Body_Decl
--
--  <Description>
--
--   The package body declaration node is built for the package body.
--   All the relevant information associated with the package body
--   is inserted into the symbol table entry for the package.
--
--
--==========================================================================
separate (Declaration_Routines)

procedure Create_Package_Body_Decl
    (Context        : in out Context_Control.Context_Definition;
     Package_Node   : in Ref_Statement_Database;
     Body_Decl_Node : in Ref_Statement_Database;
     Stmts_Node     : in Ref_Statement_Database;
     End_Label_Node : in Ref_Statement_Database;
     Tree           : out Ref_Statement_Database) is

   -- Notes:
   --   Body_Decl_Node : may be null
   --   Stmts_Node: may be null (contains Begin sequence of stmts
   --                            and the exception handler stmts)
   --   End_Label_Node: may be null

   Stmt_Db_Node         : Ref_Statement_Database;

   Package_Symtbl_Entry : Ref_Symbol_Table_Entry;

begin
   -- If a designator occurred after the END keyword, check to see that
   -- it matches the designator of the subprogram
   Package_Symtbl_Entry := Get_Symbol_Table_Entry (Package_Node);
   if not Is_Empty (End_Label_Node) and then
      not Are_Symbolic_Names_Equal (Get_Symbolic_Name (Package_Symtbl_Entry),
                                    Get_Symbolic_Name (End_Label_Node)) then
      --E user error: name after "end" doesn't match package name
      Show_Error (2011, Warning,
                  Convert_To_String (Get_Symbolic_Name (Package_Symtbl_Entry)));
   end if;

   -- Pop the environment for the package.
   Pop_Declarative_Region (Context);

   -- Set the pointer to the declaration items that are associated with
   -- the package body
   if not Is_Empty (Body_Decl_Node) then
      Set_Package_Body_Decls (Package_Symtbl_Entry, Body_Decl_Node);
   end if;

   -- Mark each declaration item within the package body to
   -- refer back to the package name
   Mark_Items_With_Parent_Package (Body_Decl_Node, Package_Symtbl_Entry);

   -- Extract the sequence of statements from the node that contains
   -- both the sequence of statements and the exception handling.
   -- The "List_Item" node of the statement database has been used
   -- as the temporary holding location for the two lists of statements.
   -- The Item_Expression field holds the sequence of statements, and the
   -- Next_List_Item holds the exception handling statements.
   -- [TBD] Exception Handling
   if not Is_Empty (Stmts_Node) then
      Set_Head_Of_Statements_List (Package_Symtbl_Entry,
                                   Get_Expression (Stmts_Node));
   end if;

   -- Indicate that the package specification has a body
   Set_Has_Body (Package_Symtbl_Entry);

   -- Traverse the symbol table entries for the package specification
   -- portion and check that the entries that require a body within
   -- the package body have one.
   All_Specs_Have_Bodies (Get_Package_Spec_Decls (Package_Symtbl_Entry));

   -- Constract a package body declaration node
   Stmt_Db_Node := Create_Node (Package_Body_Declaration);

   -- Associate any help with the package body declaration
   --Set_Help (Package_Symtbl_Entry);

   Set_Symbol_Table_Entry (Stmt_Db_Node, Package_Symtbl_Entry);

   Tree := Stmt_Db_Node;

end Create_Package_Body_Decl;

-- --------------------------------------------------------------------------
--==========================================================================
--  <Subprogram Name>
--
--    Create_Use_Node
--
--  <Description>
--
--    The id names are verified to be packages.  A Use statement is
--    built.
--
--==========================================================================
separate (Declaration_Routines)

function Create_Use_Node
    (Context      : in Context_Control.Context_Definition;
     Id_List_Node : in Ref_Statement_Database) return Ref_Statement_Database is

   Stmt_Db_Node : Ref_Statement_Database;
   Current      : Ref_Statement_Database;

   Id_Name      : Ref_String;
   Id_Entry     : Ref_Symbol_Table_Entry;

begin
   -- Verify that each name in the id list is the name of a package
   Current := Id_List_Node;
   while not Is_Empty (Current) loop

      -- Extract the name
      Id_Name := Get_Symbolic_Name (Get_Expression (Current));

      -- Check that the name exists and that it is a package name
      Id_Entry := Find_Symbol_Table_Entry (Context, Id_Name, Package_Entry);
      if Is_Empty (Id_Entry) then
         --E user error: package name given in use is not defined or is
         --              not a package
         Show_Error (2012, Fatal, Convert_To_String (Id_Name));
      end if;

      -- Add the symbol table reference to the package into the statement
      -- database node
      Set_Symbol_Table_Entry (Get_Expression (Current), Id_Entry);

      -- Get next list node from the id list
      Current := Get_Next_List_Item (Current);

   end loop;

   -- Constract a use statement node
   Stmt_Db_Node := Create_Node (Use_Statement);
   Set_Name_List (Stmt_Db_Node, Id_List_Node);

   return (Stmt_Db_Node);
end Create_Use_Node;

-- --------------------------------------------------------------------------
