--|$Log:	drtns.a,v $
--|Revision 8.0  91/01/17  10:44:50  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Declaration_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    This package contains the set of routines that handle the processing
--    of Ada declarations during parsing of input by the Ada Command
--    Language Interpreter.  These routines perform semantic checks,
--    build up symbol table entries associated with entities (objects,
--    types, subprograms, packages, ...), and build statement database nodes
--    associated with the "declaration" of these entities.
--
--    The externally visible routines within the package are invoked by the
--    ACLI parser during the syntactic analysis.  The AdaYACC grammar file
--    is the appropriate base to determine the context in which these
--    routines are invoked.  The actions which these routines perform are
--    based upon the context in they are invoked.
--
--    The bodies for all of the routines for the package Declaration_Routines
--    are not physically located within this file.  Routines have been
--    grouped according to their function (e.g. packages, subprograms,
--    enumerations, ...) and organized accordingly.  The routines will be
--    identifed as subunits within this package.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Ace_Universal_Types;
use Ace_Universal_Types;

with Statement_Database;
use Statement_Database;

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Symbol_Table_Routines;
use Symbol_Table_Routines;

with Symtbl_Entry_Set_Routines;
use Symtbl_Entry_Set_Routines;

with Symtbl_Entry_Get_Routines;
use Symtbl_Entry_Get_Routines;

with Node_Create_Routines;
use Node_Create_Routines;

with Node_Get_Routines;
use Node_Get_Routines;

with Node_Set_Routines;
use Node_Set_Routines;

with Misc_Support_Routines;
use Misc_Support_Routines;

with Declaration_Support_Routines;
use Declaration_Support_Routines;

with Storage_Management_Routines;
use Storage_Management_Routines;

with Declarative_Region_Support_Routines;
use Declarative_Region_Support_Routines;

with Error_Handler;
use Error_Handler;

with Literals;
use Literals;

with Parserdefs;
with Ace;
with Debugger;

package body Declaration_Routines is

   -- Constants used for the mode of a parameter

   Default_In_Mode : constant String := "      ";
   In_Mode         : constant String := "in    ";
   Out_Mode        : constant String := "out   ";
   In_Out_Mode     : constant String := "in out";

   -- --------------------
   -- INTERNAL ROUTINES --
   -- --------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Entry_Ref_Node
   --
   --  <Description>
   --
   --    A statement database node, with an expression node of kind
   --    Terminal_Node, is created with the Node parameter being assigned
   --    as the symbol table entry value.  This allows symbol table entries
   --    to be passed around as "dummy" statement database nodes.
   --
   --  <Called By>
   --
   --    routines internal to this package
   --
   --==========================================================================
   function Create_Entry_Ref_Node
       (Node : in Ref_Symbol_Table_Entry) return Ref_Statement_Database is

      Stmt_Db_Node : Ref_Statement_Database;

   begin
      Stmt_Db_Node := Create_Node (Terminal_Node);
      Set_Symbol_Table_Entry (Stmt_Db_Node, Node);
      return (Stmt_Db_Node);

   end Create_Entry_Ref_Node;

   -- ------------------------
   -- END INTERNAL ROUTINES --
   -- ------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Type_Decl
   --
   --  <Description>
   --
   --    A Statement_Record/Type_Decl is created for the identifier and its
   --    associated type and discriminant information.
   --
   --    The discriminant node is optional.
   --
   --==========================================================================
   function Create_Type_Decl
       (Context           : in Context_Control.Context_Definition;
        Id_Node           : in Ref_Statement_Database;
        Discriminant_Node : in Ref_Statement_Database;
        Type_Defn_Node    : in Ref_Statement_Database) return
         Ref_Statement_Database is separate;

   -- --------------------
   -- INTERNAL ROUTINES --
   -- --------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Construct_Object_Decl
   --
   --  <Description>
   --
   --    The object declaration statement database node is built for the
   --    list of identifiers.  The information as to its type, whether its
   --    value is to remain constant, and the initialization code are used
   --    in constructing the object declaration.
   --
   --  <Called By>
   --
   --    Create_Object_Decl, Create_Number_Decl
   --
   --==========================================================================
   procedure Construct_Object_Decl
       (Context              : in out Context_Control.Context_Definition;
        Id_List_Node         : in Ref_Statement_Database;
        Is_Constant          : in Boolean;
        Subtype_Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Init_Node            : in Ref_Statement_Database;
        Value                : out Ref_Statement_Database) is

      Id_Name                : Ref_String;

      Object_Symtbl_Entry    : Ref_Symbol_Table_Entry;
      Duplicate_Symtbl_Entry : Ref_Symbol_Table_Entry;

      Current                : Ref_Statement_Database;
      Stmt_Db_Node           : Ref_Statement_Database;

      First_Id               : Boolean := True;

      Value1                 : Ace_Natural;

   begin

      -- Each identifier in the list will be checked for uniqueness.
      Init_For_Duplicate_Check;

      Current := Id_List_Node;
      while not Is_Empty (Current) loop

         -- Extract the identifier name from the id list
         Id_Name := Get_Symbolic_Name (Get_Expression (Current));

         -- Verify that the object identifier does not clash with existing
         -- identifiers
         if Name_Clash (Context, Id_Name) then
            --E user error: redefinition of existing identifier
            Show_Error (2001, Fatal, Convert_To_String (Id_Name));
         end if;

         -- Check whether this identifier name conflicts with other
         -- identifiers in this statement
         if Check_For_Duplicates (Id_Name) then
            --E user error: multiple objects with same name
            Show_Error (2002, Fatal, Convert_To_String (Id_Name));
         end if;

         -- Build the symbol table entry for this object identifier.
         -- Method for building the identifier depends upon whether this
         -- is the first object id or a subsequent object id.
         if First_Id then

            -- Build symbol table entry for the first object identifier
            Object_Symtbl_Entry := Subtype_Symtbl_Entry;
            Set_Entry_Kind (Object_Symtbl_Entry, Object_Entry);
            if Is_Constant then
               Set_Mode_To_In (Object_Symtbl_Entry);
            end if;
            Set_Symbolic_Name (Object_Symtbl_Entry, Id_Name);

            -- Assign the current static nesting level to the object and
            -- acquire an data allocation offset for the object
            Set_Level (Object_Symtbl_Entry, Get_Current_Level (Context));

            Next_Offset_Value (Context, Value1);
            Set_Value_Offset (Object_Symtbl_Entry, Value1);

            -- Place the symbol table entry into the list
            Set_Symbol_Table_Entry (Get_Expression (Current),
                                    Object_Symtbl_Entry);

            First_Id := False;

         else
            -- Create the symbol table entry by duplicating the entry created
            -- for the first object
            Duplicate_Symtbl_Entry :=
               Duplicate_Symbol_Table_Entry (Object_Symtbl_Entry, Id_Name);

            -- Acquire a different data allocation offset for this object
            Next_Offset_Value (Context, Value1);
            Set_Value_Offset (Duplicate_Symtbl_Entry, Value1);

            -- Place the symbol table entry into the list
            Set_Symbol_Table_Entry (Get_Expression (Current),
                                    Duplicate_Symtbl_Entry);

         end if;

         -- Get next list node from the id_list
         Current := Get_Next_List_Item (Current);

      end loop;

      -- Create a statement database node for the object(s) declaration,
      -- which refers to the list of symbol table entries for the object
      -- name(s), and the optional initialization
      Stmt_Db_Node := Create_Node (Object_Declaration);
      Set_Name_List (Stmt_Db_Node, Id_List_Node);
      if not Is_Empty (Init_Node) then
         Set_Expression (Stmt_Db_Node, Init_Node);
      end if;

      Value := Stmt_Db_Node;

   end Construct_Object_Decl;
   -- ------------------------
   -- END INTERNAL ROUTINES --
   -- ------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Object_Decl
   --
   --  <Description>
   --
   --    The object declaration statement database node is built for the
   --    list of identifiers.  The information as to its type, whether its
   --    value is to remain constant, and the initialization code are used
   --    in constructing the object declaration.
   --
   --==========================================================================
   procedure Create_Object_Decl
       (Context          : in out Context_Control.Context_Definition;
        Id_List_Node     : in Ref_Statement_Database;
        Constant_Token   : in Boolean;
        Subtype_Ind_Node : in Ref_Statement_Database;
        Init_Node        : in Ref_Statement_Database;
        Value            : out Ref_Statement_Database) is

      -- Id_List_Node: always id_list (1 or more ids)
      -- Constant_Token: TRUE ==> keyword CONSTANT was present
      -- Init_Node: may be null

   begin
      -- Check that a constant object declaration has initialization code
      -- associated with it
      if Constant_Token and Is_Empty (Init_Node) then
         --E user error: constant object without initialization
         Show_Error (2017, Fatal, "");
      end if;

      Construct_Object_Decl (Context, Id_List_Node, Constant_Token,
                             Get_Symbol_Table_Entry (Subtype_Ind_Node),
                             Init_Node, Value);

   end Create_Object_Decl;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Number_Decl
   --
   --  <Description>
   --
   --    An object declaration statement database node is built for the
   --    list of constants.  The information as to its type and the
   --    initialization code are used in constructing the constant declaration.
   --
   --==========================================================================
   procedure Create_Number_Decl
       (Context      : in out Context_Control.Context_Definition;
        Id_List_Node : in Ref_Statement_Database;
        Init_Node    : in Ref_Statement_Database;
        Value        : out Ref_Statement_Database) is

      -- [TBD]  NOT COMPLETELY SUPPORTED YET

      -- Notes:
      -- Id_List_Node: always id_list (1 or more ids)

      Integer_Name : constant Ref_String := Convert_To_Ref_String ("integer");
      Is_Constant          : constant Boolean := True;

      Subtype_Symtbl_Entry : Ref_Symbol_Table_Entry;

      Stmt_Db_Node         : Ref_Statement_Database;

   begin
      -- Verify that the initialization is a static expression and yields
      -- either a universal_integer or a universal_real
      -- [TBD]

      -- Deduce the type of the number declaration from the initialization
      -- [TBD]

      -- Since reals are not being handled, assume that the constant is
      -- universal_integer.  Locate the type node for "integer" and build
      -- a sub_type symbol table entry.
      Subtype_Symtbl_Entry :=
         Create_Symbol_Table_Entry (Convert_To_Ref_String (Anonymous_Name),
                                    Subtype_Entry);
      Set_Entry_Base_Type (Subtype_Symtbl_Entry,
                           Find_Symbol_Table_Entry (Ace.Main_Thread_Context,
                                                    Integer_Name));

      Construct_Object_Decl (Context, Id_List_Node, Is_Constant,
                             Subtype_Symtbl_Entry, Init_Node, Value);

   end Create_Number_Decl;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Subtype_Decl
   --
   --  <Description>
   --
   --    A statement database subtype declaration node is constructed for the
   --    identifier.  The subtype indication information has already been
   --    constructed and is passed in via a subtype indication node.
   --
   --==========================================================================
   function Create_Subtype_Decl
       (Context          : in Context_Control.Context_Definition;
        Id_Node          : in Ref_Statement_Database;
        Subtype_Ind_Node : in Ref_Statement_Database) return
         Ref_Statement_Database is

      Stmt_Db_Node         : Ref_Statement_Database;
      Subtype_Symtbl_Entry : Ref_Symbol_Table_Entry;

   begin

      -- Set the symbol table entry to be a subtype definition node,
      -- and assign the identifier to the symbol table entry
      Subtype_Symtbl_Entry := Get_Symbol_Table_Entry (Subtype_Ind_Node);
      Set_Entry_Kind (Subtype_Symtbl_Entry, Subtype_Entry);
      Set_Symbol_Table_Entry (Subtype_Ind_Node, Subtype_Symtbl_Entry);

      Set_Symbolic_Name (Subtype_Symtbl_Entry, Get_Symbolic_Name (Id_Node));

      -- Set the static nesting level of the type
      Set_Level (Subtype_Symtbl_Entry, Get_Current_Level (Context));

      -- Associate help info with the symbol table entry
      Set_Help (Subtype_Symtbl_Entry);

      -- Verify that the identifier does not clash with an existing identifier
      if Name_Clash (Context, Get_Symbolic_Name (Id_Node)) then
         --E user error: redefinition of existing identifier
         Show_Error (2001, Fatal,
                     Convert_To_String (Get_Symbolic_Name (Id_Node)));
      end if;

      -- Create a statement database node for the subtype declaration, which
      -- refers to the symbol table entry for the subtype name
      Stmt_Db_Node := Create_Node (Subtype_Declaration);
      Set_Symbol_Table_Entry (Stmt_Db_Node, Subtype_Symtbl_Entry);
      return (Stmt_Db_Node);

   end Create_Subtype_Decl;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Link_Id_Onto_Front_Of_List
   --
   --  <Description>
   --
   --    The identifier is linked onto the front of a list of identifiers.
   --    A list item statement database node is created with the identifier
   --    as the left element and the identifier list as the right element.
   --    The "Holder" field of the statement database list item node contains
   --    a reference to the last element in the list and this invariant is
   --    maintained by this routine.
   --
   --==========================================================================
   function Link_Id_Onto_Front_Of_List
       (Id_Node      : in Ref_Statement_Database;
        Id_List_Node : in Ref_Statement_Database) return Ref_Statement_Database
       is

      -- Notes:
      -- Id_Node is always a single id
      -- Id_List_Node may be null or a list of ids

      New_Node : Ref_Statement_Database;

   begin

      -- Build a new list_item statement database node for the id
      New_Node := Create_Node (List_Item);
      Set_Expression (New_Node, Id_Node);

      -- Link the new node onto the id_list node
      Set_Next_List_Item (New_Node, Id_List_Node);
      if not Is_Empty (Id_List_Node) then
         Set_Holder (New_Node, Get_Holder (Id_List_Node));
      else
         Set_Holder (New_Node, New_Node);
      end if;

      return (New_Node);

   end Link_Id_Onto_Front_Of_List;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Link_Id_Onto_Rear_Of_List
   --
   --  <Description>
   --
   --
   --    The identifier is linked onto the rear of a list of identifiers.
   --    A list item statement database node is created with the identifier
   --    as the left element and null as the right element.  The right element
   --    of the last entry on the id list will point to the newly created
   --    list item node.
   --    The "Holder" field of the statement database list item node contains
   --    a reference to the last element in the list and this invariant is
   --    maintained by this routine.
   --
   --==========================================================================
   function Link_Id_Onto_Rear_Of_List (Id_Node      : in Ref_Statement_Database;
                                       Id_List_Node : in Ref_Statement_Database
                                       ) return Ref_Statement_Database is

      -- Notes:
      -- Id_Node is always a single id
      -- Id_List_Node may be null or a list of ids

      New_Node : Ref_Statement_Database;

   begin

      -- Build a new list_item statement database node for the id
      New_Node := Create_Node (List_Item);
      Set_Expression (New_Node, Id_Node);

      -- Link the new node onto the id_list node.  Method of linking depends
      -- upon whether the list is empty or not.
      if Is_Empty (Id_List_Node) then
         Set_Holder (New_Node, New_Node);

         return (New_Node);

      else -- link onto the rear of existing list
         Set_Next_List_Item (Get_Holder (Id_List_Node), New_Node);
         Set_Holder (Id_List_Node, New_Node);

         return (Id_List_Node);

      end if;

   end Link_Id_Onto_Rear_Of_List;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Build_Subtype_Indication
   --
   --  <Description>
   --
   --    A symbol table entry reference statement database node is constructed
   --    which contains the subtype indication information.  The subtype
   --    indication information is formulated form the type mark and optional
   --    constraint information.
   --
   --==========================================================================
   function Build_Subtype_Indication
       (Type_Mark_Node  : in Ref_Statement_Database;
        Constraint_Node : in Ref_Statement_Database) return
         Ref_Statement_Database is

      -- NOTE: CONSTRAINTS NOT YET IMPLEMENTED [TBD]

      -- Note: Constraint node may be null

      Subtype_Symtbl_Entry : Ref_Symbol_Table_Entry;

   begin
      -- If the constraint exists, then verify that the constraint is
      -- appropriate for the type_mark
      if not Is_Empty (Constraint_Node) then
         -- CONSTRAINTS NOT YET IMPLEMENTED
         -- [TBD]
         null;
      end if;

      -- Assemble the type_mark and constraint information into an appropriate
      -- symbol table entry.  Build a statement database node for the combined
      -- symbol table entry.
      -- NOTE: this should be checked after the implementation of constraints
      --       is finalized
      Subtype_Symtbl_Entry :=
         Create_Symbol_Table_Entry (Convert_To_Ref_String (Anonymous_Name),
                                    Subtype_Entry);
      Set_Entry_Base_Type (Subtype_Symtbl_Entry,
                           Get_Symbol_Table_Entry (Type_Mark_Node));
      if not Is_Empty (Constraint_Node) then
         Set_Constraint (Subtype_Symtbl_Entry,
                         Get_Symbol_Table_Entry (Build_Range_Constraint
                               (Constraint_Node)));
      end if;
      return (Create_Entry_Ref_Node (Subtype_Symtbl_Entry));

   end Build_Subtype_Indication;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Build_Unconstrained_Array_Decl
   --
   --  <Description>
   --
   --
   --==========================================================================
   function Build_Unconstrained_Array_Decl
       (First_Index     : in Ref_Statement_Database;
        List_Of_Indexes : in Ref_Statement_Database;
        Components      : in Ref_Statement_Database) return
         Ref_Statement_Database is

      -- NOTE: CONSTRAINTS NOT YET IMPLEMENTED [TBD]

      Index     : Ref_Statement_Database :=
         Link_Id_Onto_Front_Of_List (First_Index, List_Of_Indexes);

      Type_Node : Ref_Symbol_Table_Entry :=
         Create_Symbol_Table_Entry (Convert_To_Ref_String (Anonymous_Name),
                                    Type_Entry);

   begin
      Set_Type_To_Array_Type (Type_Node);
      Set_As_Unconstrained (Type_Node);
      Set_Index_Type (Type_Node, Index);
      Set_Element_Type (Type_Node, Components);
      return (Create_Entry_Ref_Node (Type_Node));
   end Build_Unconstrained_Array_Decl;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Build_Constrained_Array_Decl
   --
   --  <Description>
   --
   --
   --==========================================================================
   function Build_Constrained_Array_Decl
       (List_Of_Indexes : in Ref_Statement_Database;
        Components      : in Ref_Statement_Database) return
         Ref_Statement_Database is

      -- NOTE: CONSTRAINTS NOT YET IMPLEMENTED [TBD]

      Type_Node : Ref_Symbol_Table_Entry :=
         Create_Symbol_Table_Entry (Convert_To_Ref_String (Anonymous_Name),
                                    Type_Entry);

   begin
      Set_Type_To_Array_Type (Type_Node);
      Set_As_Constrained (Type_Node);
      Set_Index_Type (Type_Node, List_Of_Indexes);
      Set_Element_Type (Type_Node, Components);
      return (Create_Entry_Ref_Node (Type_Node));
   end Build_Constrained_Array_Decl;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Resolve_Type_Mark
   --
   --  <Description>
   --
   --    The symbolic name is resolved to the kind of type mark which it
   --    represents (type, subtype, or derived type).  A symbol table entry
   --    reference statement database node is constructed for the type mark.
   --
   --==========================================================================
   function Resolve_Type_Mark
       (Expanded_Name_Node : in Ref_Statement_Database) return
         Ref_Statement_Database is

      -- NOTE: Expanded names not currently implemented
      -- [TBD]

      Symbol_Table_Entry : Ref_Symbol_Table_Entry;

   begin

      -- Search symbol table for name/expanded name
      Symbol_Table_Entry := Get_Symbol_Table_Entry (Expanded_Name_Node);
      if Is_Empty (Symbol_Table_Entry) then
         --E user error: undefined identifier
         Show_Error
             (2004, Fatal,
              Convert_To_String (Get_Symbolic_Name (Expanded_Name_Node)));
      end if;

      -- Determine whether name is a type/subtype/derived type, and
      -- return a statement database node that is a reference to the
      -- symbol table entry for the type name
      case Get_Entry_Kind (Symbol_Table_Entry) is
         when Type_Entry  | Subtype_Entry     | Derived_Type_Entry     =>
            return (Create_Entry_Ref_Node (Symbol_Table_Entry));

         when others =>
            --E user error: name must be a type/subtype/derived type
            Show_Error
                (2005, Fatal,
                 Convert_To_String (Get_Symbolic_Name (Expanded_Name_Node)));
      end case;

   end Resolve_Type_Mark;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Mark_Derived_Type
   --
   --  <Description>
   --
   --    The subtype indication node has its type mark changed to that of
   --    a derived type.
   --
   --==========================================================================
   function Mark_Derived_Type
       (Subtype_Ind_Node : in Ref_Statement_Database) return
         Ref_Statement_Database is

      Symtbl_Entry : Ref_Symbol_Table_Entry;

   begin

      -- Subytype indication symbol table entry is marked as a derived type
      Symtbl_Entry := Get_Symbol_Table_Entry (Subtype_Ind_Node);
      Set_Entry_Kind (Symtbl_Entry, Derived_Type_Entry);
      Set_Symbol_Table_Entry (Subtype_Ind_Node, Symtbl_Entry);
      return (Subtype_Ind_Node);

   end Mark_Derived_Type;

   -- ------------------------------------------
   -- Separate Routines for Enumeration Types --
   -- ------------------------------------------

   -- --------------------------------------------------------------------------

   function Create_Enum_Type_Def_Entry
       (Context        : in Context_Control.Context_Definition;
        Enum_Type_Node : in Ref_Statement_Database) return
         Ref_Statement_Database is separate;

   -- --------------------------------------------------------------------------

   function Link_Enum_Literal_On_Front
       (Pred_Node, List_Node : in Ref_Statement_Database) return
         Ref_Statement_Database is separate;

   -- --------------------------------------------------------------------------

   function Link_Enum_Literal_On_End
       (List_Node, Succ_Node : in Ref_Statement_Database) return
         Ref_Statement_Database is separate;

   -- --------------------------------------------------------------------------

   function Create_Enum_Literal_Entry
       (Context           : in Context_Control.Context_Definition;
        Enum_Literal_Node : in Ref_Statement_Database) return
         Ref_Statement_Database is separate;

   -- --------------------------------------------------------------------------

   -- -------------------------------------
   -- Separate Routines for Record Types --
   -- -------------------------------------
   -- --------------------------------------------------------------------------

   function Create_Record_Type_Entry
       (Context                     : in Context_Control.Context_Definition;
        Record_Components_List_Node : in Ref_Statement_Database) return
         Ref_Statement_Database is separate;

   -- --------------------------------------------------------------------------

   function Link_Record_Component_Entries
       (Component_List_Node : in Ref_Statement_Database;
        Component_Node      : in Ref_Statement_Database) return
         Ref_Statement_Database is separate;

   -- --------------------------------------------------------------------------

   function Build_Record_Component
       (Context          : in Context_Control.Context_Definition;
        Id_List_Node     : in Ref_Statement_Database;
        Subtype_Ind_Node : in Ref_Statement_Database;
        Init_Node        : in Ref_Statement_Database) return
         Ref_Statement_Database is separate;

   -- --------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Build_Integer_Type
   --
   --  <Description>
   --
   --    An integer type definition node is constructed from the range
   --    definition information.
   --
   --==========================================================================
   function Build_Integer_Type
       (Range_Definition_Node : in Ref_Statement_Database) return
         Ref_Statement_Database is

      -- NOTE: THIS ROUTINE IS STILL UNDER DEVELOPMENT
      --       COMPLETE IMPLEMENTATION NOT YET DONE  [TBD]

      Anon_Type_Entry   : Ref_Symbol_Table_Entry;
      Int_Subtype_Entry : Ref_Symbol_Table_Entry;

   begin
      -- NOTE: no RANGE attributes handled currently [TBD]

      -- Verify that each bound of the range is a static expression of some
      -- integer type [TBD]

      -- Build the anonymous integer type definition (a derived type
      -- that has the predefined integer as its base type)
      Anon_Type_Entry :=
         Create_Symbol_Table_Entry (Convert_To_Ref_String (Anonymous_Name),
                                    Derived_Type_Entry);
      Set_Entry_Base_Type
          (Anon_Type_Entry,
           Find_Symbol_Table_Entry (Ace.Main_Thread_Context,
                                    Convert_To_Ref_String ("integer")));

      -- Build the integer subtype symbol table entry.  The base type for the
      -- subtype is the just created anonymous derived type and the constraint
      -- has been passed in as the parameter Range_Definition_Node.
      Int_Subtype_Entry :=
         Create_Symbol_Table_Entry (Convert_To_Ref_String (Anonymous_Name),
                                    Subtype_Entry);
      Set_Entry_Base_Type (Int_Subtype_Entry, Anon_Type_Entry);
      Set_Constraint (Int_Subtype_Entry,
                      Get_Symbol_Table_Entry (Range_Definition_Node));

      -- Reuse the Range_Definition_Node to pass back the integer type
      -- definition node that has been built.
      Set_Symbol_Table_Entry (Range_Definition_Node, Int_Subtype_Entry);

      return (Range_Definition_Node);
   end Build_Integer_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Build_Exception_Declaration
   --
   --  <Description>
   --
   --    A symbol table reference entry statement database node is constructed
   --    for the an exception declaration
   --
   --==========================================================================
   procedure Create_Exception_Declaration
       (Context        : in Context_Control.Context_Definition;
        Exception_Name : in Ref_Statement_Database;
        Value          : out Ref_Statement_Database) is

      Exception_Symtbl_Entry : Ref_Symbol_Table_Entry;
      Duplicate_Symtbl_Entry : Ref_Symbol_Table_Entry;

      Id_Name                : Ref_String;

      Current                : Ref_Statement_Database;
      Stmt_Db_Node           : Ref_Statement_Database;
      First_Id               : Boolean := True;

   begin
      -- Each identifier in the list will be checked for uniqueness.
      Init_For_Duplicate_Check;

      Current := Exception_Name;

      while not Is_Empty (Current) loop

         -- Extract the identifier name from the id list
         Id_Name := Get_Symbolic_Name (Get_Expression (Current));

         -- Verify that the object identifier does not clash with existing
         -- identifiers
         if Name_Clash (Context, Id_Name) then
            --E user error: redefinition of existing identifier
            Show_Error (2001, Fatal, Convert_To_String (Id_Name));
         end if;

         -- Check whether this identifier name conflicts with other
         -- identifiers in this statement
         if Check_For_Duplicates (Id_Name) then
            --E user error: multiple objects with same name
            Show_Error (2002, Fatal, Convert_To_String (Id_Name));
         end if;

         -- Build the symbol table entry for this object identifier.
         -- Method for building the identifier depends upon whether this
         -- is the first object id or a subsequent object id.
         if First_Id then

            -- Build symbol table entry for the first object identifier
            Exception_Symtbl_Entry :=
               Create_Symbol_Table_Entry
                   (Get_Symbolic_Name (Get_Expression (Exception_Name)),
                    Exception_Entry);

            -- Assign the current static nesting level to the object and
            -- acquire an data allocation offset for the object
            Set_Level (Exception_Symtbl_Entry, Get_Current_Level (Context));

            -- Place the symbol table entry into the list
            Set_Symbol_Table_Entry (Get_Expression (Current),
                                    Exception_Symtbl_Entry);

            First_Id := False;

         else
            -- Create the symbol table entry by duplicating the entry created
            -- for the first object
            Duplicate_Symtbl_Entry :=
               Duplicate_Symbol_Table_Entry (Exception_Symtbl_Entry, Id_Name);

            -- Place the symbol table entry into the list
            Set_Symbol_Table_Entry (Get_Expression (Current),
                                    Duplicate_Symtbl_Entry);
         end if;

         -- Get next list node from the id_list
         Current := Get_Next_List_Item (Current);

      end loop;

      -- Create a statement database node for the object(s) declaration,
      -- which refers to the list of symbol table entries for the object
      -- name(s), and the optional initialization
      Stmt_Db_Node := Create_Node (Exception_Declaration);
      Set_Name_List (Stmt_Db_Node, Exception_Name);

      Value := Stmt_Db_Node;

   end Create_Exception_Declaration;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Build_Range_Constraint
   --
   --  <Description>
   --
   --    A symbol table reference entry statement database node is constructed
   --    for the constraint information.
   --
   --==========================================================================
   function Build_Range_Constraint (Constraint_Info : in Ref_Statement_Database
                                    ) return Ref_Statement_Database is

      -- NOTE: THIS ROUTINE IS STILL UNDER DEVELOPMENT
      --       COMPLETE IMPLEMENTATION NOT YET DONE  [TBD]

      Constraint_Symbol_Table_Entry : Ref_Symbol_Table_Entry;

   begin

      -- Create a symbol table entry for the constraint
      Constraint_Symbol_Table_Entry :=
         Create_Symbol_Table_Entry (Convert_To_Ref_String (Anonymous_Name),
                                    Constraint_Entry);

      -- Need stset routine TBD
      Constraint_Symbol_Table_Entry.Constraint_Ref :=
         new Constraint_Node (Range_Constraint);
      
      -- Assimilate constraint information into a symbol table entry
      -- Set the upper and lower bound of the discrete range

      if Check_Expr_Node (Constraint_Info) = List_Item then
        Set_Lower_Bound (Constraint_Symbol_Table_Entry,
                         Get_Lower_Bound(Get_Expression (Constraint_Info)));
        Set_Upper_Bound (Constraint_Symbol_Table_Entry,
                         Get_Upper_Bound(Get_Expression (Constraint_Info)));
      else
        Set_Upper_Bound (Constraint_Symbol_Table_Entry,
                         Get_Upper_Bound (Constraint_Info));
        Set_Lower_Bound (Constraint_Symbol_Table_Entry,
                         Get_Lower_Bound (Constraint_Info));
      end if;

      -- Create a statement database nodes which contains the constraint symbol
      -- table entry
      return (Create_Entry_Ref_Node (Constraint_Symbol_Table_Entry));

   end Build_Range_Constraint;

   -- -------------------------------------------------------------------------

   -- ------------------------------------
   -- Separate Routines for Subprograms --
   -- ------------------------------------
   -- -------------------------------------------------------------------------

   function Build_Subprogram_Spec_Decl
       (Context         : in Context_Control.Context_Definition;
        Subprogram_Node : in Ref_Statement_Database) return
         Ref_Statement_Database is separate;

   -- -------------------------------------------------------------------------

   function Create_Subprogram_Spec
       (Context            : in Context_Control.Context_Definition;
        Id_Desig_Node      : in Ref_Statement_Database;
        Formal_Params_Node : in Ref_Statement_Database;
        Type_Mark_Node     : in Ref_Statement_Database) return
         Ref_Statement_Database is separate;

   -- -------------------------------------------------------------------------

   function Build_Formal_Parameter
       (Id_List_Node    : in Ref_Statement_Database;
        Param_Mode_Node : in Ref_Statement_Database;
        Type_Mark_Node  : in Ref_Statement_Database;
        Init_Node       : in Ref_Statement_Database) return
         Ref_Statement_Database is separate;

   -- -------------------------------------------------------------------------

   function Link_Formal_Param_On_Front
       (Param_Node      : in Ref_Statement_Database;
        Param_List_Node : in Ref_Statement_Database) return
         Ref_Statement_Database is separate;

   -- -------------------------------------------------------------------------

   function Link_Formal_Param_On_Rear
       (Param_Node      : in Ref_Statement_Database;
        Param_List_Node : in Ref_Statement_Database) return
         Ref_Statement_Database is separate;

   -- -------------------------------------------------------------------------

   function Build_Param_Mode (Context : in Context_Control.Context_Definition;
                              First_Token, Second_Token : in Integer :=
      - 1) return Ref_Statement_Database is separate;

   -- -------------------------------------------------------------------------

   procedure Start_Of_Subprogram_Body
       (Context         : in out Context_Control.Context_Definition;
        Subprogram_Node : in Ref_Statement_Database) is separate;

   -- --------------------------------------------------------------------------

   procedure Build_Subprogram_Body_Decl
       (Context         : in out Context_Control.Context_Definition;
        Subprogram_Node : in Ref_Statement_Database;
        Decl_Node       : in Ref_Statement_Database;
        Stmts_Node      : in Ref_Statement_Database;
        Exception_Node  : in Ref_Statement_Database;
        End_Label_Node  : in Ref_Statement_Database;
        Tree            : out Ref_Statement_Database) is separate;

   -- -------------------------------------------------------------------------

   -- ---------------------------------
   -- Separate Routines for Packages --
   -- ---------------------------------

   -- -------------------------------------------------------------------------

   procedure Mark_Items_With_Parent_Package
       (Items          : in Ref_Statement_Database;
        Parent_Package : in Ref_Symbol_Table_Entry) is separate;

   -- --------------------------------------------------------------------------

   procedure Start_Of_Package
       (Context : in out Context_Control.Context_Definition;
        Id_Node : in Ref_Statement_Database;
        Is_Body : in Boolean;
        Tree    : out Ref_Statement_Database) is separate;

   -- --------------------------------------------------------------------------

   procedure Create_Package_Spec_Decl
       (Context        : in out Context_Control.Context_Definition;
        Package_Node   : in Ref_Statement_Database;
        Spec_Decl_Node : in Ref_Statement_Database;
        End_Label_Node : in Ref_Statement_Database;
        Tree           : out Ref_Statement_Database) is separate;

   -- --------------------------------------------------------------------------

   procedure Create_Package_Body_Decl
       (Context        : in out Context_Control.Context_Definition;
        Package_Node   : in Ref_Statement_Database;
        Body_Decl_Node : in Ref_Statement_Database;
        Stmts_Node     : in Ref_Statement_Database;
        End_Label_Node : in Ref_Statement_Database;
        Tree           : out Ref_Statement_Database) is separate;

   -- --------------------------------------------------------------------------

   function Create_Use_Node
       (Context      : in Context_Control.Context_Definition;
        Id_List_Node : in Ref_Statement_Database) return Ref_Statement_Database
       is separate;

   -- --------------------------------------------------------------------------

   -- -----------------------------
   -- Separate Routines for Task --
   -- -----------------------------

   -- -------------------------------------------------------------------------

   --  procedure Mark_Items_With_Parent_Task
   --      (Items       : in Ref_Statement_Database;
   --      Parent_Task : in Ref_Symbol_Table_Entry) is separate;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_List_Item_Node
   --
   --  <Description>
   --
   --    A list item node is constructed with the two items being the
   --    left and right elements of the list item.
   --
   --==========================================================================
   function Create_List_Item_Node
       (Item1 : in Ref_Statement_Database;
        Item2 : in Ref_Statement_Database) return Ref_Statement_Database is

      -- Notes:
      --   This routine is used to "merge" the sequence of statements following
      --   BEGIN in a package body with the exception handler part of the
      --   package body.  These items are merged into a single statement
      --   database node so that they may be passed up in the Ada grammar to
      --   a higher level rule.

      Stmt_Db_Node : Ref_Statement_Database;

   begin

      Stmt_Db_Node := Create_Node (List_Item);
      Set_Expression (Stmt_Db_Node, Item1);
      Set_Next_List_Item (Stmt_Db_Node, Item2);
      return (Stmt_Db_Node);

   end Create_List_Item_Node;

   -- --------------------------------------------------------------------------

end Declaration_Routines;
