--|$Log:	dspprt.a,v $
--|Revision 8.0  91/01/17  10:44:53  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Declaration_Support_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--   This package is a collection of routines that provide support during
--   the parsing process of Ada statements by the Ada Command Language
--   Interpreter.  These routines provide ancillary support and are not
--   called directly by the parser.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Ace_Universal_Types;
use Ace_Universal_Types;

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Symbol_Table_Routines;
use Symbol_Table_Routines;

with Symtbl_Entry_Get_Routines;
use Symtbl_Entry_Get_Routines;

with Symtbl_Entry_Set_Routines;
use Symtbl_Entry_Set_Routines;

with Statement_Database;
use Statement_Database;

with Node_Get_Routines;
use Node_Get_Routines;

with Misc_Support_Routines;
use Misc_Support_Routines;

with Storage_Management_Routines;
use Storage_Management_Routines;

with Error_Handler;
use Error_Handler;

with Literals;
use Literals;

with Unchecked_Deallocation;

with Text_Io; -- WPL
with Debugger; -- RAF

package body Declaration_Support_Routines is

   -- ------
   -- Objects and types shared by the procedures Init_For_Duplicate_Check and
   -- Check_For_Duplicates
   -- ------

   type Id_Element;
   type Link_Id_Element is access Id_Element;

   type Id_Element is
      record
         Id_Name      : Ref_String;
         Next_Element : Link_Id_Element;
      end record;

   Head_Id_List : Link_Id_Element := null;

   -- ----------------- --
   -- INTERNAL ROUTINES --
   -- ----------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Match_Parameters
   --
   --  <Description>
   --
   --    The two parameter lists are compared parameter by parameter to
   --    determine whether the specifications are equivalent.  If all
   --    parameters are equivalent, then True is returned; otherwise False.
   --
   --  <Called By>
   --
   --    Subprogram_Specs_Conform
   --
   --==========================================================================
   function Match_Parameters
       (Param_List_One : in Ref_Symbol_Table_Entry;
        Param_List_Two : in Ref_Symbol_Table_Entry) return Boolean is

      -- Precondition:
      -- The number of the parameters has already been verified to be
      -- equal.

      Param_Entry_One : Ref_Symbol_Table_Entry;
      Param_Entry_Two : Ref_Symbol_Table_Entry;

   begin
      Param_Entry_One := Param_List_One;
      Param_Entry_Two := Param_List_Two;

      while not Is_Empty (Param_Entry_One) loop
         -- 1) Compare parameter names
         -- 2) Check the modes of the parameters
         -- 3) Check the type of the parameters
         -- 4) Check the initialization code [TBD]
         -- 5) Check the grouping of identifiers within a parameter spec
         --    (i.e. lexically, was the name preceded by a comma)
         if not Are_Symbolic_Names_Equal (Get_Symbolic_Name (Param_Entry_One),
                                          Get_Symbolic_Name (Param_Entry_Two))
             or not

            Are_Modes_Equal (Param_Entry_One, Param_Entry_Two) or

            Get_Type (Param_Entry_One) /= Get_Type (Param_Entry_Two) or
            (

         -- For now, just check if both have initialization code
         -- Will have to check if the code is the same if the future
         -- [TBD]
            Is_Empty (Get_Initialization_Code (Param_Entry_One)) xor
            Is_Empty (Get_Initialization_Code (Param_Entry_Two))) or
            (

            Is_Preceded_By_Comma (Param_Entry_One) xor
            Is_Preceded_By_Comma (Param_Entry_Two)) then

            return False;
         end if;

         -- Check the next pair of parameters
         Param_Entry_One := Get_Next_Parameter (Param_Entry_One);
         Param_Entry_Two := Get_Next_Parameter (Param_Entry_Two);
      end loop;

      -- Parameter specifications match if this point is reached.
      return True;

   end Match_Parameters;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Subprogram_Specs_Conform
   --
   --  <Description>
   --
   --    The subprogram specifications are compared against each other.
   --    If the two specifications conform, then True is returned;
   --    otherwise False.
   --
   --  <Called By>
   --
   --    Name_Clash_For_Subprogram
   --
   --==========================================================================
   function Subprogram_Specs_Conform
       (Subprogram_Spec_One : in Ref_Symbol_Table_Entry;
        Subprogram_Spec_Two : in Ref_Symbol_Table_Entry) return Boolean is

      -- Precondition: It is assumed that the specifications both are for a
      --        procedure or a function

   begin
      -- Check that the specs are equivalent
      return
         (Are_Symbolic_Names_Equal (Get_Symbolic_Name (Subprogram_Spec_One),
                                    Get_Symbolic_Name (Subprogram_Spec_Two))
          and
         (Get_Parameter_Count (Subprogram_Spec_One) =
         Get_Parameter_Count (Subprogram_Spec_Two) and then
         Match_Parameters (Get_Head_Of_Parameters_List (Subprogram_Spec_One),
                           Get_Head_Of_Parameters_List (Subprogram_Spec_Two)))
          and
         Get_Return_Type (Subprogram_Spec_One) =
         Get_Return_Type (Subprogram_Spec_Two));

   end Subprogram_Specs_Conform;

   -- --------------------- --
   -- END INTERNAL ROUTINES --
   -- --------------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Subprogram_Profiles_Match
   --
   --  <Description>
   --
   --    The subprogram profiles (parameters and return value for function)
   --    are compared against each other.
   --    If the two profiles match, then True is returned; otherwise False.
   --
   --==========================================================================
   function Subprogram_Profiles_Match
       (Subprogram_Spec_One : in Ref_Symbol_Table_Entry;
        Subprogram_Spec_Two : in Ref_Symbol_Table_Entry) return Boolean is

      -- Precondition: It is assumed that the specifications are either a
      --               procedure or a function

      Param_Entry_One : Ref_Symbol_Table_Entry;
      Param_Entry_Two : Ref_Symbol_Table_Entry;

   begin
      -- Subprograms must both be a procedure or a function
      if Get_Entry_Kind (Subprogram_Spec_One) /=
         Get_Entry_Kind (Subprogram_Spec_Two) then
         return False;
      end if;

      -- Parameter count must be the same
      if Get_Parameter_Count (Subprogram_Spec_One) /=
         Get_Parameter_Count (Subprogram_Spec_Two) then
         return False;
      end if;

      -- Base type of each parameter must be the same
      Param_Entry_One := Get_Head_Of_Parameters_List (Subprogram_Spec_One);
      Param_Entry_Two := Get_Head_Of_Parameters_List (Subprogram_Spec_Two);
      while not Is_Empty (Param_Entry_One) loop

         if Get_Base_Type (Get_Type (Param_Entry_One)) /=
            Get_Base_Type (Get_Type (Param_Entry_Two)) then
            return False;
         end if;

         Param_Entry_One := Get_Next_Parameter (Param_Entry_One);
         Param_Entry_Two := Get_Next_Parameter (Param_Entry_Two);
      end loop;

      -- Base type of the Return type must be the same
      if Get_Entry_Kind (Subprogram_Spec_One) = Function_Entry and then
         (Get_Base_Type (Get_Return_Type (Subprogram_Spec_One)) /=
         Get_Base_Type (Get_Return_Type (Subprogram_Spec_Two))) then
         return False;
      end if;

      -- Profiles match if this point is reached.
      return True;

   end Subprogram_Profiles_Match;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Init_For_Duplicate_Check
   --
   --  <Description>
   --
   --    This routine works in conjuction with Check_For_Duplicates.
   --    The local state information is initialized to empty for the subsequent
   --    calls made to Check_For_Duplicates.  Any information remaining from
   --    a previous series of call to Init_For_Duplicate_Check and
   --    Check_For_Duplicates is cleaned up.
   --
   --==========================================================================
   procedure Init_For_Duplicate_Check is

      Current : Link_Id_Element;
      procedure Free is new Unchecked_Deallocation
          (Id_Element, Link_Id_Element);

   begin
      -- if any list remains from a previous usage, discard all the elements
      while Head_Id_List /= null loop
         Current      := Head_Id_List;
         Head_Id_List := Head_Id_List.Next_Element;
         Free (Current);
      end loop;

   end Init_For_Duplicate_Check;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Check_For_Duplicates
   --
   --  <Description>
   --
   --    The identifier name is added to the running list of identifiers.
   --    The identifier is checked against all the current identifiers within
   --    the list to determine whether it is identical to another identifier
   --    within the list.
   --
   --==========================================================================
   function Check_For_Duplicates (Id_Name : in Ref_String) return Boolean is

      Current : Link_Id_Element;

   begin
      -- Compare the name of the id against all previous ids that have already
      -- been checked
      Current := Head_Id_List;
      while Current /= null loop
         if Are_Symbolic_Names_Equal (Id_Name, Current.Id_Name) then
            return True;
         else
            Current := Current.Next_Element;
         end if;
      end loop;

      -- The identifier name does not duplicate any other identifier currently
      -- held. Add this identifier onto the list.  Order of identifiers is not
      -- important - just add identifier onto head of the list.
      Current      := new Id_Element'(Id_Name      => Id_Name,
                                 Next_Element => Head_Id_List);
      Head_Id_List := Current;

      return False;

   end Check_For_Duplicates;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Name_Clash
   --
   --  <Description>
   --
   --    The identifier name is searched for within the symbol table to
   --    determine whether the name already exists.  For a name to clash, the
   --    name must exist at the current static subprogram nesting level.
   --
   --==========================================================================
   function Name_Clash (Context : in Context_Control.Context_Definition;
                        Id_Name : in Ref_String) return Boolean is

      Found_Symtbl_Entry : Ref_Symbol_Table_Entry;

   begin
      -- Search for the identifier in the appropriate portion of the
      -- symbol table.  If identifier is found then a name clash exists,
      -- unless the identifier is a derived subprogram.
      Found_Symtbl_Entry := Find_Entry_Within_Scope (Context, Id_Name);
      if Is_Empty (Found_Symtbl_Entry) or else
         Is_A_Derived_Subprogram (Found_Symtbl_Entry) then
         return False;
      else
         return True;
      end if;

   end Name_Clash;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Name_Clash_For_Enum_Literal
   --
   --  <Description>
   --
   --    The enumeration literal name is searched for within the symbol table
   --    to determine whether the name already exists.  Note that an
   --    enumeration literal may be overloaded (i.e. the name may already be
   --    used as an enumeration literal).  For an enumeration literal name to
   --    clash, the name must exist as something other than an enumeration
   --    literal at the current static subprogram nesting level.
   --
   --==========================================================================
   function Name_Clash_For_Enum_Literal
       (Context      : in Context_Control.Context_Definition;
        Literal_Name : in Ref_String) return Boolean is

      Found_Symtbl_Entry : Ref_Symbol_Table_Entry;

   begin

      -- Search for the enumeration literal in the appropriate portion of the
      -- symbol table
      Found_Symtbl_Entry := Find_Entry_Within_Scope (Context, Literal_Name);

      -- If the identifier is not found then a name clash does not exist.
      -- If an identifier is found at the current static nesting level that
      -- is an enumeration literal or a derived subprogram, then a name clash
      -- does not exists.  Otherwise, the identifier already exists and does
      -- not define an enumeration literal.
      if Is_Empty (Found_Symtbl_Entry) or else
         (Is_A_Derived_Subprogram (Found_Symtbl_Entry) or
         Get_Entry_Kind (Found_Symtbl_Entry) = Enumeration_Literal_Entry) then
         return False;
      else
         return True;
      end if;

   end Name_Clash_For_Enum_Literal;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Name_Clash_For_Subprogram
   --
   --  <Description>
   --
   --    The subprogram name is searched for within the symbol table to
   --    determine whether the name already exists.  If the name is used
   --    for something other than a subprogram, then a name clash exists.
   --    If the name exists as a subprogram name, then the signatures of
   --    the subprograms must be checked.  A name clash exists if the
   --    signature of a subprogram spec matches the signature of another
   --    spec or body, or if the signature of a subprogram body matches the
   --    signature of another body.
   --    If the specification is found for a subprogram body, then the
   --    symbol table entry for the specification is returned; otherwise,
   --    null is returned.
   --
   --==========================================================================
   function Name_Clash_For_Subprogram
       (Context                  : in Context_Control.Context_Definition;
        Subprogram_Entry         : in Ref_Symbol_Table_Entry;
        Subprogram_Entry_Is_Body : in Boolean) return Ref_Symbol_Table_Entry is

      Found_Symtbl_Entry         : Ref_Symbol_Table_Entry;
      Found_Symtbl_Entry_Is_Body : Boolean;

   begin

      -- Search for the subprogram name in the appropriate portion of the
      -- symbol table
      Found_Symtbl_Entry :=
         Find_Entry_Within_Scope
             (Context, Get_Symbolic_Name (Subprogram_Entry));

      -- If the identifier is not found or is the name of a derived subprogram
      -- then a name clash does not exist.
      -- If the first identifier found at the current static nesting level
      -- is a not subprogram, then there is a name clash.
      -- If an identifier is found at the current static nesting level that
      -- is a subprogram, then the subprograms specs must be compared.
      loop
         if Is_Empty (Found_Symtbl_Entry) or else
            Is_A_Derived_Subprogram (Found_Symtbl_Entry) then
            return null;
         elsif Get_Entry_Kind (Found_Symtbl_Entry) /= Function_Entry and
               Get_Entry_Kind (Found_Symtbl_Entry) /= Procedure_Entry then
            --E user error: attempt to redefine existing identifier
            Show_Error
                (2001, Fatal,
                 Convert_To_String (Get_Symbolic_Name (Subprogram_Entry)));
         end if;

         -- "Comparison" between Subprogram_Entry and Found_Symtbl_Entry
         -- depend upon the whether each is a subprogram decl or body.
         -- If (a) Subprogram_Entry is a decl or
         --    (b) Subprogram_Entry is a body and Found_Symtbl_Entry is a body
         -- and the parameter and result profiles are equal for the
         -- Subprogram_Entry and Found_Symtbl_Entry, then Subprogram_Entry
         -- clashes with Found_Symtbl_Entry.
         -- If Subprogram_Entry is a body and Found_Symtbl_Entry is a decl,
         -- and the specs conform, then the decl for the body has been located.
         -- If the specs do not conform, then the parameter and result
         -- profiles must be checked.  If profiles match, then there is a
         -- clash.
         -- If no clash is detected on the first Found_Symtbl_Entry, then
         -- searching and comparison must continue for all functions/procedures
         -- at that level.
         if Get_Entry_Kind (Found_Symtbl_Entry) =
            Get_Entry_Kind (Subprogram_Entry) then

            if Subprogram_Entry_Is_Body and
               Is_Empty (Get_Head_Of_Statements_List (Found_Symtbl_Entry)) and
               Subprogram_Specs_Conform (Subprogram_Entry, Found_Symtbl_Entry)
                then
               -- Decl has been found for the body; No clash exists.
               -- Return the subprogram spec entry.
               return Found_Symtbl_Entry;
            end if;

            if Subprogram_Profiles_Match (Subprogram_Entry, Found_Symtbl_Entry)
                then
               -- Name clash exists.
               --E user error: subprograms with same name that have a matching
               --              profile (parameters & return value)
               --- we should replace the old one.
               Show_Error
                   (2009, Diagnostic,
                    Convert_To_String (Get_Symbolic_Name (Subprogram_Entry)));
               -- we'll return the old symbol table entry, but first have to
               -- make sure that the parameters and declarations of the new
               -- entry are tacked on.
               -- we probably should be doing some garbage collection here...

               Set_Head_Of_Parameters_List
                   (Found_Symtbl_Entry,
                    Get_Head_Of_Parameters_List (Subprogram_Entry));
               Set_Tail_Of_Parameters_List
                   (Found_Symtbl_Entry,
                    Get_Tail_Of_Parameters_List (Subprogram_Entry));

               --Set_Head_Of_Declarations_List
               --    (Found_Symtbl_Entry,
               --     Get_Head_Of_Declarations_List (Subprogram_Entry));

               --Set_Head_Of_Statements_List (Found_Symtbl_Entry, null);

               -- Remove subprogram from debugger list, if pragma
               -- observe subprograms window is on.
               Debugger.Remove_Entry_From_List (Found_Symtbl_Entry);
               return Found_Symtbl_Entry;
            end if;
         end if;

         -- To reach here, the first match found must be of the appropriate
         -- type (subprogram).  Therefore, continue looking but only for the
         -- proper entry kind (function or subprogram).
         -- Search for the subprogram name in the appropriate portion of the
         -- symbol table
         Found_Symtbl_Entry :=
            Find_Next_Entry_Within_Scope (Context, Found_Symtbl_Entry,
                                          Get_Entry_Kind (Subprogram_Entry));

      end loop;

   end Name_Clash_For_Subprogram;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Locate_Package_Spec
   --
   --  <Description>
   --
   --    The current scope is searched to locate the package specification
   --    associated with the package name.
   --
   --==========================================================================
   function Locate_Package_Spec
       (Context      : in Context_Control.Context_Definition;
        Package_Name : in Ref_String) return Ref_Symbol_Table_Entry is

      Package_Symtbl_Entry : Ref_Symbol_Table_Entry;

   begin
      Package_Symtbl_Entry := Find_Entry_Within_Scope (Context, Package_Name);

      if Is_Empty (Package_Symtbl_Entry) then
         --E user error: no package specification for the package body
         --              within the scope
         Show_Error (2010, Fatal, Convert_To_String (Package_Name));

      elsif Get_Entry_Kind (Package_Symtbl_Entry) /= Package_Entry then
         --E user error: name given does not refer to a package
         Show_Error (2001, Fatal, Convert_To_String (Package_Name));

      elsif Has_A_Body (Package_Symtbl_Entry) then
         --E user error: package already has a package body
         Show_Error (2013, Diagnostic, Convert_To_String (Package_Name));
         return Package_Symtbl_Entry;
      else
         return Package_Symtbl_Entry;
      end if;

   end Locate_Package_Spec;

   -- ---------------- --
   -- INTERNAL ROUTINE --
   -- ---------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Package_Spec_Requires_Body
   --
   --  <Description>
   --
   --    Determines whether a package specification requires a package
   --    body.  Returns True if the spec requires a body; otherwise
   --    returns False.
   --
   --==========================================================================
   function Package_Spec_Requires_Body
       (Package_Symtbl_Entry : Ref_Symbol_Table_Entry) return Boolean is

      Current : Ref_Statement_Database;

   begin
      -- Acquire the specifications associated with the package
      Current := Get_Package_Spec_Decls (Package_Symtbl_Entry);

      -- Check each item within the package spec declarations
      while not Is_Empty (Current) loop

         case Check_Stmt_Node (Current) is

            when Procedure_Spec_Declaration |
                 Function_Spec_Declaration  |
                 Procedure_Declaration      |
                 Function_Declaration    =>
               return True;

            when Package_Declaration     =>
               if Package_Spec_Requires_Body (Get_Symbol_Table_Entry (Current))
                   then
                  return True;
               end if;

               -- when "Task_Declaration"
               -- when "Generic_Declaration"

            when others   =>
               null;

         end case;

         Current := Get_Next_Statement (Current);
      end loop;

      -- To reach this point, no declaration within the spec requires a
      -- body
      return False;

   end Package_Spec_Requires_Body;

   -- -----------------------
   -- END INTERNAL ROUTINE --
   -- -----------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    All_Specs_Have_Bodies
   --
   --  <Description>
   --
   --    The package is examined to determine whether all declarations
   --    within the package spec have a body within the package body.
   --
   --==========================================================================
   procedure All_Specs_Have_Bodies
       (Package_Spec_Decls : in Ref_Statement_Database) is

      -- Notes:
      -- Task and generic declarations not implemented yet.
      -- Subprograms written in another language, and subprograms stubs
      -- not implemented yet.

      Current      : Ref_Statement_Database := Package_Spec_Decls;

      Symtbl_Entry : Ref_Symbol_Table_Entry;

   begin
      -- Check all the items within the package spec
      while not Is_Empty (Current) loop

         case Check_Stmt_Node (Current) is

            when  -- Procedure_Declaration   | Function_Declaration |
                 Procedure_Spec_Declaration   | Function_Spec_Declaration   =>
               Symtbl_Entry := Get_Symbol_Table_Entry (Current);
               if Is_Empty (Get_Head_Of_Statements_List (Symtbl_Entry))
                   and then
                  (not Is_Subprogram_Built_In (Symtbl_Entry) and
                  not Is_A_Derived_Subprogram (Symtbl_Entry)) then
                  Show_Error
                      (2014, Warning,
                       Convert_To_String (Get_Symbolic_Name (Symtbl_Entry)));
               end if;

            when Package_Declaration     =>
               Symtbl_Entry := Get_Symbol_Table_Entry (Current);
               if not Has_A_Body (Symtbl_Entry) and then
                  Package_Spec_Requires_Body (Symtbl_Entry) then
                  Show_Error
                      (2014, Warning,
                       Convert_To_String (Get_Symbolic_Name (Symtbl_Entry)));

               end if;

               -- when "Task_declaration"
               -- when "Generic_declaration"

            when others   =>
               null;

         end case;

         Current := Get_Next_Statement (Current);
      end loop;

   end All_Specs_Have_Bodies;

   -- -----------------------------------------------------------------------

end Declaration_Support_Routines;
