--|$Log:	dsubprg.a,v $
--|Revision 8.0  91/01/17  10:44:54  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Subunits for Package - Declaration_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--   The following routines are separate routines for the package
--   Declaration_Routines.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------
-- -------------------------------------------------------------------------
--==========================================================================
--  <Subprogram Name>
--
--    Build_Subprogram_Spec_Decl
--
--  <Description>
--
--    The subprogram symbol table entry that has been already created is
--    inserted into a statement to create a subprogram specification.
--
--
--==========================================================================
separate (Declaration_Routines)

function Build_Subprogram_Spec_Decl
    (Context         : in Context_Control.Context_Definition;
     Subprogram_Node : in Ref_Statement_Database) return Ref_Statement_Database
    is

   Stmt_Db_Node     : Ref_Statement_Database;
   Subprogram_Entry : Ref_Symbol_Table_Entry;

   Ignore           : Ref_Symbol_Table_Entry;
   Spec             : constant Boolean := False;

begin
   Subprogram_Entry := Get_Symbol_Table_Entry (Subprogram_Node);

   -- Verify that the subprogram name does not clash with existing
   -- identifiers.
   -- (Name_Clash_For_Subprogram takes care of reporting errors)
   Ignore := Name_Clash_For_Subprogram (Context, Subprogram_Entry, Spec);

   -- if subprogram already exists, remove from display list, in case
   -- pragma subprograms observe is on.

   if not Is_Empty (Ignore) then
      Debugger.Remove_Entry_From_List (Ignore);
   end if;
   -- Determine whether the subprogram is either a procedure or a function,
   -- and then build the statement database procedure/function declaration
   -- node
   if Get_Entry_Kind (Subprogram_Entry) = Procedure_Entry then
      Stmt_Db_Node := Create_Node (Procedure_Spec_Declaration);
   else
      Stmt_Db_Node := Create_Node (Function_Spec_Declaration);
   end if;

   -- associate any help with the subprogram spec.
   --   Set_Help (Subprogram_Entry);

   -- Indicate that there is an explicit Subprogram Declaration
   Set_Subprg_Has_Decl (Subprogram_Entry);

   Set_Symbol_Table_Entry (Stmt_Db_Node, Subprogram_Entry);
   return (Stmt_Db_Node);

end Build_Subprogram_Spec_Decl;

-- -------------------------------------------------------------------------
--==========================================================================
--  <Subprogram Name>
--
--    Create_Subprogram_Spec
--
--  <Description>
--
--    A subprogram (function or procedure) specification is built for the
-- identifier/designator.  The specification contains the symbol table
-- entry for a subprogram.
--
--==========================================================================
separate (Declaration_Routines)

function Create_Subprogram_Spec
    (Context            : in Context_Control.Context_Definition;
     Id_Desig_Node      : in Ref_Statement_Database;
     Formal_Params_Node : in Ref_Statement_Database;
     Type_Mark_Node     : in Ref_Statement_Database) return
      Ref_Statement_Database is

   -- Effects:
   -- A subprogram (function or procedure) specification is built for the
   -- identifier/designator.  The specification contains the symbol table
   -- entry for a subprogram.

   -- Notes:
   -- Id_Desig_Node: an identifier for a procedure,
   --                a designator for a function
   -- Formal_Params_Node: may be null
   -- Type_Mark_Node: null => procedure, non-null => function

   Subprogram_Entry    : Ref_Symbol_Table_Entry;
   Param_Entry         : Ref_Symbol_Table_Entry;
   First_Param_Entry   : Ref_Symbol_Table_Entry;

   Subprogram_Name     : Ref_String;

   Param_Count         : Ace_Natural := 0;

   Default_Param_Count : Ace_Natural := 0;

begin
   -- Acquire the subprogram name
   Subprogram_Name := Get_Symbolic_Name (Id_Desig_Node);
   Ada_Id (Subprogram_Name);

   -- Create a Procedure or Function entry
   if Is_Empty (Type_Mark_Node) then
      Subprogram_Entry := Create_Symbol_Table_Entry (Subprogram_Name,
                                                     Procedure_Entry);
   else
      Subprogram_Entry := Create_Symbol_Table_Entry (Subprogram_Name,
                                                     Function_Entry);
      -- Set the return type for a function
      Set_Return_Type (Subprogram_Entry,
                       Get_Symbol_Table_Entry (Type_Mark_Node));
   end if;

   -- Set the static nesting level of the subprogram specification
   Set_Level (Subprogram_Entry, Get_Current_Level (Context));

   -- associate any help with the subprogram spec.
   Set_Help (Subprogram_Entry);

   -- Link in the formal parameters, count the number of
   -- parameters, and check that there are no duplicate
   -- names among the subprogram name and the name of all the formal
   -- parameters.
   if not Is_Empty (Formal_Params_Node) then
      Param_Entry := Get_Symbol_Table_Entry (Formal_Params_Node);
      Set_Head_Of_Parameters_List (Subprogram_Entry, Param_Entry);

      Init_For_Duplicate_Check;
      while not Is_Empty (Param_Entry) loop
         -- check that parameter name does not conflict with other
         -- parameter names and the subprogram name
         if Check_For_Duplicates (Get_Symbolic_Name (Param_Entry)) then
            --E user error: id duplicated within subprogram spec stmt
            Show_Error (2002, Fatal,
                        Convert_To_String (Get_Symbolic_Name (Param_Entry)));
         end if;

         -- For functions, check that the mode of the parameter is either
         -- explicit in or default in mode
         if not Is_Empty (Type_Mark_Node) and then not Is_Mode_In (Param_Entry)
             then
            --E user error: function parameter not in mode
            Show_Error (2007, Fatal,
                        Convert_To_String (Get_Symbolic_Name (Param_Entry)));
         end if;

         -- Increment the parameter count
         Param_Count := Param_Count + 1;

         -- check if this is a default parameter
         if not Is_Empty (Get_Initialization_Code (Param_Entry)) then
            -- Increment the default parameter count
            Default_Param_Count := Default_Param_Count + 1;
         end if;

         -- Set the subprogram entry's parameter list tail to point to
         -- the current parameter
         Set_Tail_Of_Parameters_List (Subprogram_Entry, Param_Entry);

         -- Get the next parameter entry
         Param_Entry := Get_Next_Parameter (Param_Entry);

      end loop;
   end if;

   -- Set the subprogram entry's parameter count
   Set_Parameter_Count (Subprogram_Entry, Param_Count);

   -- Set the subprogram entry's default parameter count
   Set_Number_Of_Default_Parameters (Subprogram_Entry, Default_Param_Count);

   -- Create a statement database node containing the symbol table entry
   -- for the subprogram entry
   return (Create_Entry_Ref_Node (Subprogram_Entry));

end Create_Subprogram_Spec;

-- -------------------------------------------------------------------------
--==========================================================================
--  <Subprogram Name>
--
--    Build_Formal_Parameter
--
--  <Description>
--
--    A formal parameter symbol table entry is built for each identifier
--    in the id list.  The symbol table entries are linked together and
--    returned as a single statement database node.
--
--
--==========================================================================
with Semantic_Analyzer;

separate (Declaration_Routines)

function Build_Formal_Parameter
    (Id_List_Node    : in Ref_Statement_Database;
     Param_Mode_Node : in Ref_Statement_Database;
     Type_Mark_Node  : in Ref_Statement_Database;
     Init_Node       : in Ref_Statement_Database) return Ref_Statement_Database
    is

   -- Notes:
   -- Id_List_Node: always a list of ids (1 or more)
   -- Param_Mode_Node: indicates default in, explicit in, out, or in out
   -- Init_Node: may be null

   Current        : Ref_Statement_Database;

   Param_Name     : Ref_String;
   Mode_Descrip   : String (1 .. 6);

   Param_Entry    : Ref_Symbol_Table_Entry;
   First_Param    : Ref_Symbol_Table_Entry := null;
   Previous_Param : Ref_Symbol_Table_Entry := null;

begin
   -- Determine whether the initialization code is appropriate for the type
   if not Is_Empty (Init_Node) then
      if not Semantic_Analyzer.Is_Same_Type
             (Init_Node, Get_Symbol_Table_Entry (Type_Mark_Node)) then
         Show_Error
             (1117, Fatal,
              Convert_To_String
                  (Get_Symbolic_Name (Get_Symbol_Table_Entry (Type_Mark_Node)))
              );
      end if;
   end if;

   -- Extract description of parameter mode
   Mode_Descrip := Convert_To_String (Get_Symbolic_Name (Param_Mode_Node));

   -- For each identifier in the id list, build up a formal parameter
   -- symbol table entry.
   Current := Id_List_Node;
   while not Is_Empty (Current) loop

      -- Extract the parameter name from the id list
      Param_Name := Get_Symbolic_Name (Get_Expression (Current));

      -- Build symbol table entry for the parameter
      Param_Entry := Create_Symbol_Table_Entry (Param_Name, Formal_Param_Entry);
      -- set type, mode, next_param, value_offset
      Set_Type (Param_Entry, Get_Symbol_Table_Entry (Type_Mark_Node));

      if Mode_Descrip = Default_In_Mode then
         Set_Mode_To_Default_In (Param_Entry);
      elsif Mode_Descrip = In_Mode then
         Set_Mode_To_In (Param_Entry);
      elsif Mode_Descrip = Out_Mode then
         Set_Mode_To_Out (Param_Entry);
      elsif Mode_Descrip = In_Out_Mode then
         Set_Mode_To_In_Out (Param_Entry);
      else
         raise Tree_Builder_Error;
      end if;

      if not Is_Empty (Previous_Param) then
         Set_Next_Parameter (Previous_Param, Param_Entry);
      end if;
      Previous_Param := Param_Entry;

      -- Set the initialization code, if it exists
      if not Is_Empty (Init_Node) then
         Set_Initialization_Code (Param_Entry, Init_Node);
         Set_Is_A_Default_Expression (Init_Node);
         -- Check that parameter is IN mode; only IN mode parameters
         -- may have an initialization expression
         if not Is_Mode_In (Param_Entry) then
            --E user error: init expr for out/in out parameter
            Show_Error (2008, Fatal, Convert_To_String (Param_Name));
         end if;
      end if;

      -- Retain a reference to the first parameter entry.  For all other
      -- parameters, record that it is preceded by a comma textually.
      if Is_Empty (First_Param) then
         First_Param := Param_Entry;
      else
         Set_Preceded_By_Comma (Param_Entry);
      end if;

      -- Get next parameter id from id_list
      Current := Get_Next_List_Item (Current);

   end loop;

   -- The next entry field of the first parameter symbol table entry
   -- will be used as a holder to retain a reference to the last parameter
   -- in the list
   Set_Next_Entry (First_Param, Param_Entry);

   -- Create a statement database node that contains the list of parameter
   -- symbol table entries
   return (Create_Entry_Ref_Node (First_Param));

end Build_Formal_Parameter;

-- -------------------------------------------------------------------------
--==========================================================================
--  <Subprogram Name>
--
--    Link_Formal_Param_On_Front
--
--  <Description>
--
-- A list of formal parameters are created from the parameter and the list
-- of parameters.  A parameter node contains a parameter specification
-- which may contain more than one formal parameter identifiers.
-- The parameter is linked onto the front of the parameter list.
--
--
--==========================================================================
separate (Declaration_Routines)

function Link_Formal_Param_On_Front
    (Param_Node      : in Ref_Statement_Database;
     Param_List_Node : in Ref_Statement_Database) return Ref_Statement_Database
    is

   -- Notes:
   -- Param_Node: a single parameter specification
   -- Param_List_Node: may be null or a list of parameter specifications

   Param_Node_First_Entry      : Ref_Symbol_Table_Entry;
   Param_List_Node_First_Entry : Ref_Symbol_Table_Entry;

begin
   -- If the list is empty, nothing needs to be done
   if Is_Empty (Param_List_Node) then
      return (Param_Node);
   end if;

   Param_Node_First_Entry := Get_Symbol_Table_Entry (Param_Node);
   Param_List_Node_First_Entry := Get_Symbol_Table_Entry (Param_List_Node);

   -- Set the next parameter link in the last parameter id in param node.
   -- The first parameter in param node has a reference to the last
   -- parameter retained in the Next_Entry field.
   Set_Next_Parameter (Get_Next_Entry (Param_Node_First_Entry),
                       Param_List_Node_First_Entry);

   -- Update the Next_Entry field in the first entry of the param node
   -- to point to the last entry of the param list node (which is retained
   -- in the Next_Entry field of the first entry of the param list node).
   Set_Next_Entry (Param_Node_First_Entry,
                   Get_Next_Entry (Param_List_Node_First_Entry));

   return (Param_Node);

end Link_Formal_Param_On_Front;

-- -------------------------------------------------------------------------
--==========================================================================
--  <Subprogram Name>
--
--    Link_Formal_Param_On_Rear
--
--  <Description>
--
--    A list of formal parameters are created from the parameter and the list
--    of parameters.  A parameter node contains a parameter specification
--    which may contain more than one formal parameter identifiers.
--    The parameter is linked onto the rear of the parameter list.
--
--
--==========================================================================
separate (Declaration_Routines)

function Link_Formal_Param_On_Rear (Param_Node      : in Ref_Statement_Database;
                                    Param_List_Node : in Ref_Statement_Database
                                    ) return Ref_Statement_Database is

   -- Notes:
   -- Param_Node: a single parameter specification
   -- Param_List_Node: may be null or a list of parameter specifications

   Param_Node_First_Entry      : Ref_Symbol_Table_Entry;
   Param_List_Node_First_Entry : Ref_Symbol_Table_Entry;

begin
   -- If the list is empty, nothing needs to be done
   if Is_Empty (Param_List_Node) then
      return (Param_Node);
   end if;

   Param_Node_First_Entry := Get_Symbol_Table_Entry (Param_Node);
   Param_List_Node_First_Entry := Get_Symbol_Table_Entry (Param_List_Node);

   -- Set the next parameter link in the last parameter id in param list
   -- node.
   -- The first parameter in param list node has a reference to the last
   -- parameter retained in the Next_Entry field.
   Set_Next_Parameter (Get_Next_Entry (Param_List_Node_First_Entry),
                       Param_Node_First_Entry);

   -- Update the Next_Entry field in the first entry of the param list node
   -- to point to the last entry of the param node (which is retained
   -- in the Next_Entry field of the first entry of the param node).
   Set_Next_Entry (Param_List_Node_First_Entry,
                   Get_Next_Entry (Param_Node_First_Entry));

   return (Param_List_Node);

end Link_Formal_Param_On_Rear;

-- -------------------------------------------------------------------------
--==========================================================================
--  <Subprogram Name>
--
--    Build_Param_Mode
--
--  <Description>
--
--   Build up a string that contains the mode associated with a
--   formal parameter specification.
--
--
--==========================================================================
separate (Declaration_Routines)

function Build_Param_Mode (Context : in Context_Control.Context_Definition;
                           First_Token, Second_Token : in Integer :=
   - 1) return Ref_Statement_Database is

   Stmt_Db_Node : Ref_Statement_Database;
   Mode_String  : String (1 .. 6);

begin

   -- From the token values, build up the mode definition for a parameter:
   --    Default In, In, Out, or In Out
   if First_Token = - 1 then
      -- Default IN
      Mode_String := Default_In_Mode;
   elsif First_Token = Parserdefs.Out_Tk then
      -- OUT
      Mode_String := Out_Mode;
   elsif First_Token = Parserdefs.In_Tk and Second_Token = - 1 then
      -- IN
      Mode_String := In_Mode;
   elsif First_Token = Parserdefs.In_Tk and Second_Token = Parserdefs.Out_Tk
          then
      -- IN OUT
      Mode_String := In_Out_Mode;
   else
      raise Tree_Builder_Error;
   end if;

   -- Build a statement database node that contains the string with the
   -- parameter mode.
   Stmt_Db_Node := Create_Node (Terminal_Node);
   Set_Symbolic_Name (Stmt_Db_Node, Convert_To_Ref_String (Mode_String));

   return (Stmt_Db_Node);
end Build_Param_Mode;

-- --------------------------------------------------------------------------
--==========================================================================
--  <Subprogram Name>
--
--    Start_Of_Subprogram_Body
--
--  <Description>
--
--   The validity of the name of the subprogram is checked and data
--   allocation offsets are assigned to each of the subprogram's
--   parameters.
--
--
--==========================================================================
separate (Declaration_Routines)

procedure Start_Of_Subprogram_Body
    (Context         : in out Context_Control.Context_Definition;
     Subprogram_Node : in Ref_Statement_Database) is

   Subprogram_Entry      : Ref_Symbol_Table_Entry;
   Subprogram_Spec_Entry : Ref_Symbol_Table_Entry;
   Param_Entry           : Ref_Symbol_Table_Entry;

   Is_Body               : constant Boolean := True;
   Value                 : Ace_Natural;

begin
   -- At this point it has been determined that a subprogram body
   -- definition is beginning.

   Subprogram_Entry := Get_Symbol_Table_Entry (Subprogram_Node);

   -- For the Subprogram body, if a subprogram specification exists and
   -- the specification matches one for one, then this is the body that
   -- goes along with the specification.  If the names match but the
   -- specifications do not, then the subprogram name has been overloaded.
   -- If the name matches something that is not a subprogram, then there
   -- is an illegal redefinition of an existing identifier.
   -- (Name_Clash_For_Subprogram takes care of reporting errors)
   Subprogram_Spec_Entry := Name_Clash_For_Subprogram (Context,
                                                       Subprogram_Entry,
                                                       Is_Body);
   if not Is_Empty (Subprogram_Spec_Entry) then
      -- Since the subprogram spec symbol table entry exists for this
      -- subprogram body, will use the subprogram spec symbol table entry.
      -- [TBD] At this point the symbol table entry for the subprogram
      --       body can be discarded
      Set_Symbol_Table_Entry (Subprogram_Node, Subprogram_Spec_Entry);
      Subprogram_Entry := Subprogram_Spec_Entry;
   end if;

   -- Push a new environment for the subprogram
   Push_Declarative_Region (Context, Subprogram_Entry);

   -- Assign a data allocation level and offset to each formal parameter
   -- of the subprogram, and for the return value of a function if the
   -- subprogram is a function.  (The level for the function has already
   -- been set during the parsing of the subprogram specification.)
   if Get_Entry_Kind (Subprogram_Entry) = Function_Entry then
      Next_Offset_Value (Context, Value);
      Set_Value_Offset (Subprogram_Entry, Value);
   end if;

   Param_Entry := Get_Head_Of_Parameters_List (Subprogram_Entry);
   while not Is_Empty (Param_Entry) loop
      -- Assign the current static nesting level to the parameter and
      -- acquire an data allocation offset for the parameter
      Set_Level (Param_Entry, Get_Current_Level (Context));
      Next_Offset_Value (Context, Value);
      Set_Value_Offset (Param_Entry, Value);

      -- Get the next parameter
      Param_Entry := Get_Next_Parameter (Param_Entry);
   end loop;

end Start_Of_Subprogram_Body;

-- --------------------------------------------------------------------------
--==========================================================================
--  <Subprogram Name>
--
--    Build_Subprogram_Body_Decl
--
--  <Description>
--
--   The subprogram declaration node is built for the subprogram body.
--   All the relevant information associated with the subprogram body
--   is inserted into the symbol table entry for the subprogram.
--
--
--==========================================================================
separate (Declaration_Routines)

procedure Build_Subprogram_Body_Decl
    (Context         : in out Context_Control.Context_Definition;
     Subprogram_Node : in Ref_Statement_Database;
     Decl_Node       : in Ref_Statement_Database;
     Stmts_Node      : in Ref_Statement_Database;
     Exception_Node  : in Ref_Statement_Database;
     End_Label_Node  : in Ref_Statement_Database;
     Tree            : out Ref_Statement_Database) is

   -- Notes:
   --   Decl_Node: may be null
   --   Exception_Node: may be null
   --   End_Label_Node: may be null

   Stmt_Db_Node     : Ref_Statement_Database;
   Subprogram_Entry : Ref_Symbol_Table_Entry;

begin
   -- If a designator occurred after the END keyword, check to see that
   -- it matches the designator of the subprogram
   Subprogram_Entry := Get_Symbol_Table_Entry (Subprogram_Node);
   if not Is_Empty (End_Label_Node) and then
      not Are_Symbolic_Names_Equal (Get_Symbolic_Name (Subprogram_Entry),
                                    Get_Symbolic_Name (End_Label_Node)) then
      --E user error: name after "end" doesn't match subprogram name
      Show_Error (2006, Warning,
                  Convert_To_String (Get_Symbolic_Name (Subprogram_Entry)));
   end if;

   -- Determine whether the subprogram is either a procedure or a function,
   -- and then build the statement database procedure/function declaration
   -- node
   if Get_Entry_Kind (Subprogram_Entry) = Procedure_Entry then
      Stmt_Db_Node := Create_Node (Procedure_Declaration);
   else
      Stmt_Db_Node := Create_Node (Function_Declaration);
   end if;

   -- Set the location of the declarations for the subprogram
   Set_Head_Of_Declarations_List (Subprogram_Entry, Decl_Node);

   -- Set the location of the statements for the subprogram
   Set_Head_Of_Statements_List (Subprogram_Entry, Stmts_Node);

   -- associate help info with the symbol table entry
   --Set_Help (Subprogram_Entry);

   -- Set the location of the exception handlers for the subprogram
   if not Is_Empty (Exception_Node) then
      Set_Exception_Handler (Stmt_Db_Node, Exception_Node);
   end if;

   Set_Symbol_Table_Entry (Stmt_Db_Node, Subprogram_Entry);

   -- End of subprogram body, pop off the environment
   Pop_Declarative_Region (Context);

   Tree := Stmt_Db_Node;

end Build_Subprogram_Body_Decl;

-- --------------------------------------------------------------------------
