--|$Log:	dsubsup.a,v $
--|Revision 8.0  91/01/17  10:44:55  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Derived_Subprogram_Support
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    This package contains a set of routines that support the generation
--    of derived subprograms.  During the processing of package specifications,
--    a list of derivable subprograms is associated with each appropriate
--    type.  When additional types are derived from these types, the list
--    of derivable subprograms is traversed and a derived subprogram entry
--    is created and entered appropriately at the location where it was
--    derived.
--
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Ace_Universal_Types;
use Ace_Universal_Types;

with Literals;
use Literals;

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Statement_Database;
use Statement_Database;

with Symbol_Table_Routines;
use Symbol_Table_Routines;

with Symtbl_Entry_Get_Routines;
use Symtbl_Entry_Get_Routines;

with Symtbl_Entry_Set_Routines;
use Symtbl_Entry_Set_Routines;

with Node_Get_Routines;
use Node_Get_Routines;

with Misc_Support_Routines;
use Misc_Support_Routines;

with Declaration_Support_Routines;
use Declaration_Support_Routines;

with Declarative_Region_Support_Routines;
use Declarative_Region_Support_Routines;

package body Derived_Subprogram_Support is

   -- ---------------- --
   -- INTERNAL ROUTINE --
   -- ---------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Derived_Subprogram_Will_Be_Hidden
   --
   --  <Description>
   --
   --    This function determines whether the derived subprogram will be
   --    hidden or visible.  This function returns True if the derived
   --    subprogram will be hidded; False otherwise.
   --
   --  <Called By>
   --
   --    Derive_Subprograms
   --
   --==========================================================================
   function Derived_Subprogram_Will_Be_Hidden
       (Context          : Context_Control.Context_Definition;
        Subprogram_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      Found_Symtbl_Entry : Ref_Symbol_Table_Entry;

   begin
      Found_Symtbl_Entry :=
         Find_Entry_Within_Scope
             (Context, Get_Symbolic_Name (Subprogram_Entry));

      -- If name does not exist, the derived subprogram will not be hidden.
      -- If name exists and is not a function or procedure entry, then derived
      -- subprogram will be hidden.  Otherwise, it must be a function or
      -- procedure, and the profiles CAN'T match because the derived subprogram
      -- is being created for a type declaration, where the type declaration
      -- name did not clash.  Hence, no subprogram declaration can exist which
      -- has a profile that contains the type declaration name.
      if Is_Empty (Found_Symtbl_Entry) then
         return (False);
      elsif Get_Entry_Kind (Found_Symtbl_Entry) /= Function_Entry and
            Get_Entry_Kind (Found_Symtbl_Entry) /= Procedure_Entry then
         return (True);
      else
         return (False);
      end if;

   end Derived_Subprogram_Will_Be_Hidden;

   -- ---------------- --
   -- EXTERNAL ROUTINE --
   -- ---------------- --

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Add_Subprogram_Entry_To_List
   --
   --  <Description>
   --
   --    This procedure adds the subprogra
   --    hidden or visible.  This function returns True if the derived
   --    subprogram will be hidded; False otherwise.
   --    This procedure adds a derived subprogram onto a subprogram list.
   --    Add the subprogram onto the derivable subprogram list for the type or
   --    derived type.  Also, add the derived program onto the list for a
   --    package spec which contains all the subprograms derived within the
   --    package spec.  Do not add onto the list if the subprogram entry
   --    already exists on the list (will be at the head of the list).
   --
   --  <Called By>
   --
   --    Determine_Derivable_Subprograms, Derive_Subprograms
   --
   --==========================================================================
   procedure Add_Subprogram_Entry_To_List
       (Derived_Subprogram_Entry : in Ref_Symbol_Table_Entry;
        Symtbl_Entry             : in Ref_Symbol_Table_Entry) is

      Head_Pointer            : Ref_List_Of_Subprograms;
      Current                 : Ref_List_Of_Subprograms;
      Derived_Subprogram_Name : Ref_String;

   begin
      case Get_Entry_Kind (Symtbl_Entry) is
         when Package_Entry    | Type_Entry   | Derived_Type_Entry     =>
            null;
         when others =>
            -- raise ERROR
            null;
      end case;

      Head_Pointer := Get_Derived_Subprogram_List (Symtbl_Entry);

      -- Ensure that the derived subprogram has not already been added to the
      -- list
      if Head_Pointer /= null and then
         Head_Pointer.Subprogram_Entry = Derived_Subprogram_Entry then
         return ;
      end if;

      -- For Types and Derived Types, a subprogram entry may be added that is
      -- to replace a derived subprogram entry.  For each item already in the
      -- list, check that the name matches and if so, whether the profiles
      -- match.  If they do, replace the existing subprogram with the new
      -- subprogram
      case Get_Entry_Kind (Symtbl_Entry) is
         when Type_Entry  | Derived_Type_Entry     =>
            Current := Head_Pointer;
            Derived_Subprogram_Name :=
               Get_Symbolic_Name (Derived_Subprogram_Entry);
            while Current /= null loop
               if Are_Symbolic_Names_Equal
                      (Derived_Subprogram_Name,
                       Get_Symbolic_Name (Current.Subprogram_Entry)) and then
                  Subprogram_Profiles_Match (Derived_Subprogram_Entry,
                                             Current.Subprogram_Entry) then
                  Current.Subprogram_Entry := Derived_Subprogram_Entry;
                  return ;
               end if;

               Current := Current.Next_Subprogram_Entry;
            end loop;
         when others =>
            null;
      end case;

      Head_Pointer :=
         new List_Of_Subprograms'
            (Subprogram_Entry      => Derived_Subprogram_Entry,
             Next_Subprogram_Entry => Head_Pointer);
      Set_Derived_Subprogram_List (Symtbl_Entry, Head_Pointer);

   end Add_Subprogram_Entry_To_List;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Determine_Derivable_Subprograms
   --
   --  <Description>
   --
   --    This procedure determines the derivable subprograms associated with
   --    a package specification.  For each subprogram within the package
   --    specification declarations, mark the type definitions within the
   --    package spec that are referenced in either the formal parameters or
   --    the return value of a function.
   --
   --==========================================================================
   procedure Determine_Derivable_Subprograms
       (Package_Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      Current_Node : Ref_Statement_Database :=
         Get_Package_Spec_Decls (Package_Symtbl_Entry);
      Subprogram_Entry, Current_Param, Type_Symtbl_Entry :
      Ref_Symbol_Table_Entry;

   begin
      while (not (Is_Empty (Current_Node))) loop
         case Check_Stmt_Node (Current_Node) is
            when Procedure_Spec_Declaration |
                 Function_Spec_Declaration  |
                 Procedure_Declaration      |
                 Function_Declaration    =>

               -- Determine whether the type declaration for the parameters
               -- and the function return type is a derived type that was
               -- declared within the same package
               Subprogram_Entry := Get_Symbol_Table_Entry (Current_Node);
               Current_Param := Get_Head_Of_Parameters_List (Subprogram_Entry);
               while not Is_Empty (Current_Param) loop
                  Type_Symtbl_Entry := Get_Type (Current_Param);
                  if (Get_Entry_Kind (Type_Symtbl_Entry) = Derived_Type_Entry
                      or Get_Entry_Kind (Type_Symtbl_Entry) = Type_Entry) and
                     Get_Enclosing_Package (Type_Symtbl_Entry) =
                     Package_Symtbl_Entry then
                     Add_Subprogram_Entry_To_List (Subprogram_Entry,
                                                   Type_Symtbl_Entry);
                  end if;

                  -- Get the next parameter that is associated with next
                  -- parameter specification
                  Current_Param := Get_Next_Parameter (Current_Param);
                  while (not Is_Empty (Current_Param) and then
                     Is_Preceded_By_Comma (Current_Param)) loop
                     Current_Param := Get_Next_Parameter (Current_Param);
                  end loop;

               end loop;

               if Check_Stmt_Node (Current_Node) = Function_Declaration then
                  Type_Symtbl_Entry := Get_Return_Type (Subprogram_Entry);
                  if (Get_Entry_Kind (Type_Symtbl_Entry) = Derived_Type_Entry
                      or Get_Entry_Kind (Type_Symtbl_Entry) = Type_Entry) and
                     Get_Enclosing_Package (Type_Symtbl_Entry) =
                     Package_Symtbl_Entry then
                     Add_Subprogram_Entry_To_List (Subprogram_Entry,
                                                   Type_Symtbl_Entry);
                  end if;
               end if;

            when others   =>
               -- Renaming Declarations and Generic Instantiations must be
               -- handled similarly in the future
               -- [TBD]
               null;

         end case;

         Current_Node := Get_Next_Statement (Current_Node);
      end loop;

   end Determine_Derivable_Subprograms;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Derive_Subprograms
   --
   --  <Description>
   --
   --    This procedure creates a derived subprogram for the list of derivable
   --    subprograms associated with the parent type.  The parent type in the
   --    specification for the subprogram should be substituted with the
   --    derived type for each occurrence of the parent type in the formal
   --    parameters or the return type.
   --
   --==========================================================================
   procedure Derive_Subprograms
       (Context                   : in out Context_Control.Context_Definition;
        Derived_Type_Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      Parent_Type_Symtbl_Entry         : Ref_Symbol_Table_Entry;

      Current_Subprogram               : Ref_List_Of_Subprograms;

      Subprogram_Entry                 : Ref_Symbol_Table_Entry;
      Parameter_Entry                  : Ref_Symbol_Table_Entry;

      Derived_Subprogram_Entry         : Ref_Symbol_Table_Entry;
      Derived_Parameter_Entry          : Ref_Symbol_Table_Entry;
      Previous_Derived_Parameter_Entry : Ref_Symbol_Table_Entry;
      First_Derived_Parameter_Entry    : Ref_Symbol_Table_Entry;

      Declarative_Region               : Ref_Symbol_Table_Entry;

   begin
      -- Acquire the derived type's direct parent, and then chase back until
      -- a type or derived type is reached.
      Parent_Type_Symtbl_Entry :=
         Get_Entry_Base_Type (Derived_Type_Symtbl_Entry);
      Parent_Type_Symtbl_Entry := Get_Base_Type (Parent_Type_Symtbl_Entry);

      Current_Subprogram :=
         Get_Derived_Subprogram_List (Parent_Type_Symtbl_Entry);
      while Current_Subprogram /= null loop

         Subprogram_Entry := Current_Subprogram.Subprogram_Entry;

         -- Duplicate the Subprogram_Entry, substituting derived type for
         -- parent_type.
         Derived_Subprogram_Entry :=
            Duplicate_Symbol_Table_Entry (Subprogram_Entry,
                                          Get_Symbolic_Name (Subprogram_Entry));

         -- Make sure the entry is visible
         Make_Entry_Visible (Derived_Subprogram_Entry);

         -- Set the enclosing package field properly for this subprogram
         -- being derived
         if Within_Declarative_Region (Context) then
            Declarative_Region := Get_Declarative_Region_Entry (Context);
            if Get_Entry_Kind (Declarative_Region) = Package_Entry then
               Set_Enclosing_Package (Derived_Subprogram_Entry,
                                      Declarative_Region);
            else
               Set_Enclosing_Package (Derived_Subprogram_Entry, null);
            end if;
         else
            Set_Enclosing_Package (Derived_Subprogram_Entry, null);
         end if;

         -- Substitute derived type for parent type in function return value
         -- if appropriate
         if Get_Entry_Kind (Subprogram_Entry) = Function_Entry and then
            Get_Return_Type (Subprogram_Entry) = Parent_Type_Symtbl_Entry then
            Set_Return_Type (Derived_Subprogram_Entry,
                             Derived_Type_Symtbl_Entry);
         end if;

         -- Initialize variables used to hook formal parameter entries onto
         -- the subprogram entry
         Previous_Derived_Parameter_Entry := null;
         First_Derived_Parameter_Entry    := null;

         -- Substitute derived type for parent type in parameter entries
         -- if appropriate
         Parameter_Entry := Get_Head_Of_Parameters_List (Subprogram_Entry);
         while not Is_Empty (Parameter_Entry) loop
            Derived_Parameter_Entry :=
               Duplicate_Symbol_Table_Entry
                   (Parameter_Entry, Get_Symbolic_Name (Parameter_Entry));
            if Get_Type (Parameter_Entry) = Parent_Type_Symtbl_Entry then
               Set_Type (Derived_Parameter_Entry, Derived_Type_Symtbl_Entry);
            end if;

            -- Retain a reference to the first derived parameter entry
            if Is_Empty (First_Derived_Parameter_Entry) then
               First_Derived_Parameter_Entry := Derived_Parameter_Entry;
            end if;

            if not Is_Empty (Previous_Derived_Parameter_Entry) then
               Set_Next_Parameter (Previous_Derived_Parameter_Entry,
                                   Derived_Parameter_Entry);
            end if;
            Previous_Derived_Parameter_Entry := Derived_Parameter_Entry;

            -- Get the next parameter that is associated with the subprogram
            Parameter_Entry := Get_Next_Parameter (Parameter_Entry);

         end loop;

         -- Set the head and tail pointers for the derived subprogram entry
         Set_Head_Of_Parameters_List (Derived_Subprogram_Entry,
                                      First_Derived_Parameter_Entry);
         Set_Tail_Of_Parameters_List (Derived_Subprogram_Entry,
                                      Derived_Parameter_Entry);

         -- Set the derived subprogram to point back to the original parent
         -- for which a body of the subprogram must exist at run-time
         if Is_Empty (Get_Derived_Subprogram_Parent (Subprogram_Entry)) then
            Set_Derived_Subprogram_Parent (Derived_Subprogram_Entry,
                                           Subprogram_Entry);
         end if;

         -- Add the derived subprogram onto the list of subprograms that will
         -- be derivable for this derived type.
         Add_Subprogram_Entry_To_List (Derived_Subprogram_Entry,
                                       Derived_Type_Symtbl_Entry);

         -- Note: If the name of the derived subprogram clashes with an item
         -- within the current scope, then the derived subprogram would be
         -- hidden within the entire scope (RM 8.3(17)).  Do not need to add it
         -- to the symbol table (since it is hidden), or to associate it with
         -- the possible enclosing package specification.
         if not Derived_Subprogram_Will_Be_Hidden (Context, Subprogram_Entry)
             then
            Add_Entry_To_Symbol_Table (Context, Derived_Subprogram_Entry);

            -- If within a declarative region that is a package spec, add the
            -- derived subprogram onto the derived subprograms list associated
            -- with the symbol table entry for the package
            if Within_Declarative_Region (Context) then
               Declarative_Region := Get_Declarative_Region_Entry (Context);
               if Get_Entry_Kind (Declarative_Region) = Package_Entry then
                  Add_Subprogram_Entry_To_List (Derived_Subprogram_Entry,
                                                Declarative_Region);
               end if;
            end if;
         end if;

         -- Get the next derivable subprogram
         Current_Subprogram := Current_Subprogram.Next_Subprogram_Entry;

      end loop;

   end Derive_Subprograms;

   -- -------------------------------------------------------------------------

end Derived_Subprogram_Support;
