--|$Log:	dump.a,v $
--|Revision 8.0  91/01/17  10:44:57  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Dump_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    This package contains a set of routines that dumps out in textual
--    form Ada code.  Ada code read by ACE is parsed and translated into
--    an internal representation.  The internal representation consists of
--    statement database nodes and symbol table entries.  The dump routines
--    traverse the statement database nodes and symbol table entries and
--    reconstruct the Ada text from the ACE internal data structures.
--
--    These dump routines also support a set of controls over the form in
--    which the text is created.  This includes whether statement numbers
--    are to be displayed, whether statements should be dumped for tracing
--    purposes, whether the bodies associated with subprograms and packages
--    should be displayed, ...
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Ace_Universal_Types;
use Ace_Universal_Types;

with Literals;
use Literals;

with Symtbl_Entry_Get_Routines;
use Symtbl_Entry_Get_Routines;

with Help_Pkg;
use Help_Pkg;

with Symbol_Table_Routines;
use Symbol_Table_Routines;

with Node_Get_Routines;
use Node_Get_Routines;

with Misc_Support_Routines;
use Misc_Support_Routines;

with Window_System;
use Window_System;

with Error_Handler;
use Error_Handler;

with Storage_Management_Routines;
use Storage_Management_Routines;

with Symbol_Table_Structures;
with Unchecked_Deallocation;

package body Dump_Routines is

   -- These globals prevent the dump routines from being reentrant [TBD]

   We_Are_Dumping                : Boolean := False;

   Dump_Statement_Number         : Boolean := False;
   Dump_Statement_Numbers_Was_On : Boolean := False;

   Dump_Text_Window              : Boolean := False;

   Quote_Infix                   : Boolean := True;

   Line : String (1 .. 256);
   Line_Length                   : Natural := 1;

   Indented : Integer := 0;

   Blanks : constant String                := "          ";

   -- forward reference
   procedure Dump_Expression (Ref             : Ref_Statement_Database;
                              File            : File_Type;
                              Dumping_Mode    : List_Mode;
                              Dump_Init_Exprs : Boolean := False);

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Dump
   --
   --  <Description>
   --
   --    This procedure sets whether statement dumping is to be effect.
   --    The Switch parameter controls whether statement dumping is to be
   --    on or off.
   --==========================================================================
   procedure Dump (Switch : Boolean) is
   begin
      We_Are_Dumping := Switch;
   end Dump;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Dump_To_Text_Windows
   --
   --  <Description>
   --
   --    This procedure sets whether statement dumping is being performed to
   --    a text window or not.  The method used to indent statements is
   --    affected by whether dumping is being performed to a text window or
   --    not.
   --==========================================================================
   procedure Dump_To_Text_Windows (Switch : Boolean) is
   begin
      Dump_Text_Window := Switch;
   end Dump_To_Text_Windows;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Dump_Statement_Numbers
   --
   --  <Description>
   --
   --    This procedure controls whether statement numbers associated with the
   --    Ada statements should be dumped when the statement is displayed.
   --==========================================================================
   procedure Dump_Statement_Numbers (Switch : Boolean) is
   begin
      if (Dump_Statement_Number and Switch) then
         -- show that Dump_Statement_Numbers was on
         -- before this surrounding on/off "dump"
         Dump_Statement_Numbers_Was_On := True;
      end if;
      if (Dump_Statement_Numbers_Was_On and (not Switch)) then
         -- leave "listing" of statement numbers on because it was on
         -- before this surrounding on/off "dump"
         return ;
      end if;
      Dump_Statement_Number := Switch;
   end Dump_Statement_Numbers;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Statement_Dump_On
   --
   --  <Description>
   --
   --    This function returns a boolean indication of whether statement
   --    dumping is turned on.
   --==========================================================================
   function Is_Statement_Dump_On return Boolean is
   begin
      return We_Are_Dumping;
   end Is_Statement_Dump_On;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Indent
   --
   --  <Description>
   --
   --    This routine set the indentation level in by three.
   --==========================================================================
   procedure Indent is
   begin
      Indented := Indented + 3;
   end Indent;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Exdent
   --
   --  <Description>
   --
   --    This routine set the indentation level out by three.
   --==========================================================================
   procedure Exdent is
   begin
      Indented := Indented - 3;
   end Exdent;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Print_Indent
   --
   --  <Description>
   --
   --    This routine outputs spaces to indent a line to the appropriate
   --    level (which is set by indent and exdent).
   --==========================================================================
   procedure Print_Indent (File : File_Type) is
   begin
      for I in 1 .. Indented loop
         if Dump_Text_Window then
            Line_Length             := Line_Length + 1;
            Line (1 .. Line_Length) := Line (1 .. Line_Length - 1) & " ";
         else
            Put (File, " ");
         end if;
      end loop;
   end Print_Indent;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Spacing_For_Statement_Number
   --
   --  <Description>
   --
   --    This routine outputs spaces to keep output lines spaced correctly (it
   --    outputs the correct amount of spaces for statement numbers)
   --==========================================================================
   function Spacing_For_Statement_Number return String is
   begin
      if Dump_Statement_Number then
         return "     | ";
      else
         return "";
      end if;
   end Spacing_For_Statement_Number;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Output
   --
   --  <Description>
   --
   --    This routine outputs a given string, or it saves the string for
   --    output to an observe window.
   --==========================================================================
   procedure Output (File : File_Type;
                     Str  : String) is
   begin
      if Dump_Text_Window then
         Line_Length             := Line_Length + Str'LENGTH;
         Line (1 .. Line_Length) := Line (1 .. Line_Length - Str'LENGTH) & Str;
      else
         Put (File, Str);
      end if;
   end Output;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Outputln
   --
   -- <Description>
   --
   -- This routine outputs a newline to a file, or it sends a saved string
   -- to an observe window.
   --==========================================================================
   procedure Outputln (File : File_Type) is
   begin
      if Dump_Text_Window then
         Display_Line (Line (2 .. Line_Length));
         Line_Length := 1;
      else
         New_Line (File);
      end if;
   end Outputln;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Outputln
   --
   -- <Description>
   --
   -- This routine outputs a given string to a file, or it sends the string
   -- to an observe window.
   --

   --==========================================================================
   procedure Outputln (File : File_Type;
                       Str  : String) is
   begin
      if Dump_Text_Window then
         Line_Length             := Line_Length + Str'LENGTH;
         Line (1 .. Line_Length) := Line (1 .. Line_Length - Str'LENGTH) & Str;
         Display_Line (Line (2 .. Line_Length));
         Line_Length             := 1;
      else
         Put_Line (File, Str);
      end if;
   end Outputln;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- In_Fix
   --
   -- <Description>
   --
   -- This routine returns a boolean indicating whether or not an
   -- operator is infix or not.
   --==========================================================================
   function Is_Infix (Str : String) return Boolean is
   begin
      if Str'LENGTH > 3 then
         return False;
      elsif Str'LENGTH = 3 then
         if Str = "mod" then
            return True;
         elsif Str = "rem" then
            return True;
         elsif Str = "and" then
            return True;
         elsif Str = "xor" then
            return True;
         else
            return False;
         end if;
      elsif Str'LENGTH = 2 then
         if Str = "or" then
            return True;
         elsif Str = ">=" then
            return True;
         elsif Str = "<=" then
            return True;
         elsif Str = "/=" then
            return True;
         elsif Str = "**" then
            return True;
         else
            return False;
         end if;
      elsif Str'LENGTH = 1 then
         case Str (Str'FIRST) is
            when '+' |
                 '-' |
                 '*' |
                 '/' |
                 '&' |
                 '|' |
                 '=' |
                 '<' |
                 '>' =>
               return True;
            when others   =>
               return False;
         end case;
      end if;
   end Is_Infix;

   function Double_Quotes (Str : String) return String is
      Quotes : Natural := 0;
   begin
      for I in Str'FIRST .. Str'LAST loop
         if Str (I) = '"' then
            Quotes := Quotes + 1;
         end if;
      end loop;
      declare
         Return_String : String (Str'FIRST .. Str'LAST + Quotes);
         I             : Positive := Str'FIRST;
         Quotes_Again  : Natural  := 0;
      begin
         while I <= Str'LAST loop
            Return_String (I + Quotes_Again) := Str (I);
            if Str (I) = '"' then
               Quotes_Again                     := Quotes_Again + 1;
               Return_String (I + Quotes_Again) := '"';
            end if;
            I := I + 1;
         end loop;
         return Return_String;
      end;
   end Double_Quotes;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Dump_Type_Name
   --
   -- <Description>
   --
   -- This routine output the type associated with an object.
   --==========================================================================
   procedure Dump_Type_Name (File : File_Type;
                             Ref  : Ref_Symbol_Table_Entry) is
   begin
      case Ref.Entry_Kind is
         when Object_Entry     =>
            Output (File, Convert_To_String (Ref.Object_Type.Symbol_Name));
         when Type_Entry  =>
            Output (File, Convert_To_String (Ref.Symbol_Name));
         when Formal_Param_Entry    =>
            Output (File, Convert_To_String (Ref.Param_Type.Symbol_Name));
         when others =>
            null;
      end case;
   end Dump_Type_Name;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Dump_Symbol_Name
   --
   -- <Description>
   --
   -- This routine outputs the "name" associated with a symbol table
   -- entry.
   --==========================================================================
   procedure Dump_Symbol_Name (File : File_Type;
                               Ref  : Ref_Symbol_Table_Entry) is
   begin
      if Is_Infix (Convert_To_String (Ref.Symbol_Name)) and Quote_Infix then
         Output (File, """");
      end if;
      Output (File, Convert_To_String (Ref.Symbol_Name));
      if Is_Infix (Convert_To_String (Ref.Symbol_Name)) and Quote_Infix then
         Output (File, """");
      end if;
   end Dump_Symbol_Name;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Dump_Constraint_Info
   --
   -- <Description>
   --
   -- This routine outputs constraint info for a symbol table entry
   --==========================================================================
   procedure Dump_Constraint_Info (File : File_Type;
                                   Ref  : Ref_Symbol_Table_Entry) is
   begin
      Output (File, " range ");
      Output (File, Ace_Integer'IMAGE (Get_Lower_Bound (Ref)));
      Output (File, " .. ");
      Output (File, Ace_Integer'IMAGE (Get_Upper_Bound (Ref)));
   end Dump_Constraint_Info;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Dump_Enumeration_List
   --
   -- <Description>
   --
   -- This routine outputs an enumeration list associated with
   -- a symbol table entry.
   --

   --==========================================================================
   procedure Dump_Enumeration_List (File : File_Type;
                                    Ref  : Ref_Symbol_Table_Entry) is
      Iterator : Ref_Symbol_Table_Entry;

   begin
      Iterator := Get_First_Literal (Ref);
      while (not Is_Empty (Iterator)) loop
         Dump_Symbol_Name (File, Iterator);
         Iterator := Get_Successor (Iterator);
         if not Is_Empty (Iterator) then
            Output (File, ", ");
         end if;
      end loop;
   end Dump_Enumeration_List;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Dump_Parameter_List
   --
   -- <Description>
   --
   -- This routine output a parameter list.
   --==========================================================================
   procedure Dump_Parameter_List (File : File_Type;
                                  Ref  : Ref_Symbol_Table_Entry) is

      Iterator  : Ref_Symbol_Table_Entry;
      Iterator2 : Ref_Symbol_Table_Entry;

      procedure Dump_Mode (File  : File_Type;
                           Param : Ref_Symbol_Table_Entry) is
      begin
         if Is_Mode_Default_In (Param) then
            null; -- nothing is output-ed here.
         elsif Is_Mode_Explicit_In (Param) then
            Output (File, "in ");
         elsif Is_Mode_Out (Param) then
            Output (File, "out ");
         elsif Is_Mode_In_Out (Param) then
            Output (File, "in out ");
         end if;
      end Dump_Mode;

      procedure Dump_Init_Code (File  : File_Type;
                                Param : Ref_Symbol_Table_Entry) is
      begin
         if (not Is_Empty (Get_Initialization_Code (Param))) then
            Output (File, " := ");
            Dump_Expression (Get_Initialization_Code (Param), File,
                             Specification, True);
         end if;
      end Dump_Init_Code;

   begin
      Iterator := Get_Head_Of_Parameters_List (Ref);
      if Is_Empty (Iterator) then
         return ;
      else
         Output (File, " (");
         while (not Is_Empty (Iterator)) loop
            Dump_Symbol_Name (File, Iterator);
            -- Loop thru all parameters that are separated by a ","
            loop
               Iterator2 := Get_Next_Parameter (Iterator);
               exit  when Is_Empty (Iterator2) or else
                          not Is_Preceded_By_Comma (Iterator2);
               Output (File, ", ");
               Dump_Symbol_Name (File, Iterator2);
               Iterator := Iterator2;
            end loop;
            Output (File, " : ");
            Dump_Mode (File, Iterator);
            Dump_Type_Name (File, Iterator);
            Dump_Init_Code (File, Iterator);
            if not Is_Empty (Iterator2) then
               Output (File, "; ");
            end if;
            Iterator := Iterator2;
         end loop;
         Output (File, ")");
      end if;
   end Dump_Parameter_List;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Dump_Symtbl
   --
   -- <Description>
   --
   -- This procedure reconstructs the Ada text associated with an
   -- ACE symbol table entry.  The text is written to the specified File,
   -- and the display of bodies for packages and subprograms is controlled
   -- by the Dumping_Mode parameter.
   --==========================================================================
   procedure Dump_Symtbl (Ref          : Ref_Symbol_Table_Entry;
                          File         : File_Type;
                          Dumping_Mode : List_Mode := Specification;
                          Dump_Help    : Boolean   := True) is
   begin
      if (Dumping_Mode = Both) then
         Dump_Symtbl (Ref, File, Specification, Dump_Help);
         Dump_Symtbl (Ref, File, Implementation, Dump_Help);
      else

         if Is_Empty (Ref) then
            Print_Indent (File);
            Outputln (File, "null;");

         else
            case Get_Entry_Kind (Ref) is

               when Object_Entry    =>
                  Dump_Symbol_Name (File, Ref);
                  Output (File, " : ");
                  Dump_Type_Name (File, Ref);
                  if Has_Constraint (Ref) then
                     Output (File, " (");
                     Dump_Statement_Database (Get_Lower_Bound (Ref), File,
                                              Dumping_Mode);
                     Output (File, " .. ");
                     Dump_Statement_Database (Get_Upper_Bound (Ref), File,
                                              Dumping_Mode);
                     Output (File, ")");
                  end if;
                  Outputln (File, ";");

               when Type_Entry | Derived_Type_Entry     =>

                  if (Dump_Help) then
                     Dump_Help_Info (Ref, False, File);
                  end if;

                  Output (File, "type ");
                  Dump_Symbol_Name (File, Ref);
                  Output (File, " is ");
                  if Get_Entry_Kind (Ref) = Type_Entry then
                     if Is_Type_An_Enumeration_Type (Ref) then
                        Output (File, "(");
                        Dump_Enumeration_List (File, Ref);
                        Outputln (File, ");");
                     elsif Is_Type_An_Integer_Type (Ref) then
                        Output (File, " range ");
                        --dump_statement_statement_database
                        --(Get_Lower_Bound(Get_Symbol_Table_Entry(Ref));
                        --dump_statement_statement_database
                        --(Get_Upper_Bound(Get_Symbol_Table_Entry(Ref));
                        Output
                            (File, Ace_Integer'IMAGE (Get_Lower_Bound (Ref)));
                        Output (File, " .. ");
                        Output
                            (File, Ace_Integer'IMAGE (Get_Upper_Bound (Ref)));
                        Outputln (File, ";");
                     else
                        Outputln (File, "<Unimplemented Type>;");
                     end if;
                  elsif Get_Entry_Kind (Ref) = Derived_Type_Entry then
                     Output (File, "new ");
                     Dump_Symbol_Name (File, Get_Entry_Base_Type (Ref));
                     if Has_Constraint (Ref) then
                        Dump_Constraint_Info (File, Ref);
                     end if;
                     Outputln (File, ";");
                  end if;

               when Subtype_Entry   =>

                  if (Dump_Help) then
                     Dump_Help_Info (Ref, False, File);
                  end if;

                  Output (File, "subtype ");
                  Dump_Symbol_Name (File, Ref);
                  Output (File, " is ");
                  Dump_Symbol_Name (File, Get_Entry_Base_Type (Ref));
                  if Has_Constraint (Ref) then
                     Dump_Constraint_Info (File, Ref);
                  end if;
                  Outputln (File, ";");

               when Procedure_Entry =>

                  if (Dumping_Mode = Specification) then

                     if (Dump_Help) then
                        Dump_Help_Info (Ref, False, File);
                     end if;

                     if Get_Context_Clause (Ref) /= null then
                        --Output (File, Spacing_For_Statement_Number);
                        Dump_Statement_Database
                            (Get_Context_Clause (Ref), File, Specification);
                     end if;

                     Output (File, Spacing_For_Statement_Number);
                     Output (File, "procedure ");
                     Dump_Symbol_Name (File, Ref);
                     Dump_Parameter_List (File, Ref);
                     if Is_Subprogram_Built_In (Ref) then
                        Outputln (File, ";");
                        Output (File, Spacing_For_Statement_Number);
                        Output (File, "-- Form => ");
                        Outputln
                            (File,
                             """" &
                             Ace_Integer'IMAGE (Get_Statement_Number (Ref)) &
                             """");
                        Output (File, Spacing_For_Statement_Number);
                        Output (File, "pragma builtin (");
                        Dump_Symbol_Name (File, Ref);
                        Output (File, ", ");
                        Output
                            (File,
                             Ace_Integer'IMAGE (
                                Get_Built_In_Instance_Number (Ref)));
                        Outputln (File, ");");
                        Outputln (File);
                     else
                        Outputln (File, ";");
                        Output (File, Spacing_For_Statement_Number);
                        Output (File, "-- Form => ");
                        Outputln
                            (File,
                             """" &
                             Ace_Integer'IMAGE (Get_Statement_Number (Ref)) &
                             """");
                        Outputln (File);
                     end if;

                  elsif (Dumping_Mode = Implementation) and then
                        (not Is_Subprogram_Built_In (Ref)) and then
                        not Is_Empty (Get_Head_Of_Statements_List (Ref)) then
                     if Get_Context_Clause (Ref) /= null then
                        --Output (File, Spacing_For_Statement_Number);
                        Dump_Statement_Database
                            (Get_Context_Clause (Ref), File, Specification);
                     end if;

                     Output (File, Spacing_For_Statement_Number);
                     Output (File, "procedure ");
                     Dump_Symbol_Name (File, Ref);
                     Dump_Parameter_List (File, Ref);

                     Outputln (File, " is");
                     Output (File, Spacing_For_Statement_Number);
                     Output (File, "-- Form => ");
                     Outputln
                         (File,
                          """" & Ace_Integer'IMAGE (Get_Statement_Number (Ref))
                           & """");
                     Indent;
                     Dump_Statement_Database
                         (Get_Head_Of_Declarations_List (Ref), File,
                          Dumping_Mode);
                     Exdent;
                     Output (File, Spacing_For_Statement_Number);
                     Outputln (File, "begin");
                     Indent;
                     Dump_Statement_Database
                         (Get_Head_Of_Statements_List (Ref), File, Dumping_Mode
                          );
                     Exdent;
                     Output (File, Spacing_For_Statement_Number);
                     Output (File, "end ");
                     Dump_Symbol_Name (File, Ref);
                     Outputln (File, ";");
                     Outputln (File);

                  end if;

               when Function_Entry  =>

                  if (Dumping_Mode = Specification) then

                     if (Dump_Help) then
                        Dump_Help_Info (Ref, False, File);
                     end if;

                     if Get_Context_Clause (Ref) /= null then
                        --Output (File, Spacing_For_Statement_Number);
                        Dump_Statement_Database
                            (Get_Context_Clause (Ref), File, Specification);
                     end if;

                     Output (File, Spacing_For_Statement_Number);
                     Output (File, "function ");
                     Dump_Symbol_Name (File, Ref);
                     Dump_Parameter_List (File, Ref);
                     Output (File, " return ");
                     Dump_Symbol_Name (File, Get_Return_Type (Ref));
                     if Is_Subprogram_Built_In (Ref) then
                        Outputln (File, ";");
                        Output (File, Spacing_For_Statement_Number);
                        Output (File, "-- Form => ");
                        Outputln
                            (File,
                             """" &
                             Ace_Integer'IMAGE (Get_Statement_Number (Ref)) &
                             """");
                        Output (File, Spacing_For_Statement_Number);
                        Output (File, "pragma builtin (");
                        Dump_Symbol_Name (File, Ref);
                        Output (File, ", ");
                        Output
                            (File,
                             Ace_Integer'IMAGE (
                                Get_Built_In_Instance_Number (Ref)));
                        Outputln (File, ");");
                        Outputln (File);
                     else
                        Outputln (File, ";");
                        Output (File, Spacing_For_Statement_Number);
                        Output (File, "-- Form => ");
                        Outputln
                            (File,
                             """" &
                             Ace_Integer'IMAGE (Get_Statement_Number (Ref)) &
                             """");
                        Outputln (File);
                     end if;
                  elsif (Dumping_Mode = Implementation) and then
                        not (Is_Subprogram_Built_In (Ref) and then
                        not Is_Empty (Get_Head_Of_Statements_List (Ref))) then
                     if Get_Context_Clause (Ref) /= null then
                        --Output (File, Spacing_For_Statement_Number);
                        Dump_Statement_Database
                            (Get_Context_Clause (Ref), File, Specification);
                     end if;

                     Output (File, Spacing_For_Statement_Number);
                     Output (File, "function ");
                     Dump_Symbol_Name (File, Ref);
                     Dump_Parameter_List (File, Ref);
                     Output (File, " return ");
                     Dump_Symbol_Name (File, Get_Return_Type (Ref));
                     Outputln (File, " is");
                     Output (File, Spacing_For_Statement_Number);
                     Output (File, "-- Form => ");
                     Outputln
                         (File,
                          """" & Ace_Integer'IMAGE (Get_Statement_Number (Ref))
                           & """");
                     Indent;
                     Dump_Statement_Database
                         (Get_Head_Of_Declarations_List (Ref), File,
                          Dumping_Mode);
                     Exdent;
                     Output (File, Spacing_For_Statement_Number);
                     Outputln (File, "begin");
                     Indent;
                     Dump_Statement_Database
                         (Get_Head_Of_Statements_List (Ref), File, Dumping_Mode
                          );
                     Exdent;
                     Output (File, Spacing_For_Statement_Number);
                     Output (File, "end ");
                     Dump_Symbol_Name (File, Ref);
                     Outputln (File, ";");
                     Outputln (File);

                  end if;

               when Package_Entry   =>
                  if (Dumping_Mode = Specification) then

                     if (Dump_Help) then
                        Dump_Help_Info (Ref, False, File);
                     end if;

                     if Get_Context_Clause (Ref) /= null then
                        --Output (File, Spacing_For_Statement_Number);
                        Dump_Statement_Database
                            (Get_Context_Clause (Ref), File, Specification);
                     end if;

                     Output (File, Spacing_For_Statement_Number);
                     Output (File, "package ");
                     Dump_Symbol_Name (File, Ref);
                     Outputln (File, " is");
                     Output (File, Spacing_For_Statement_Number);
                     Output (File, "-- Form => ");
                     Outputln
                         (File,
                          """" & Ace_Integer'IMAGE (Get_Statement_Number (Ref))
                           & """");
                     Outputln (File);
                     Indent;
                     Dump_Statement_Database (Get_Package_Spec_Decls (Ref),
                                              File, Specification);
                     Exdent;
                     Output (File, Spacing_For_Statement_Number);
                     Output (File, "end ");
                     Dump_Symbol_Name (File, Ref);
                     Outputln (File, ";");

                  elsif (Dumping_Mode = Implementation) and Has_A_Body (Ref)
                         then
                     Print_Indent (File);
                     Output (File, Spacing_For_Statement_Number);
                     if Get_Context_Clause_Body (Ref) /= null then
                        --Output (File, Spacing_For_Statement_Number);
                        Dump_Statement_Database
                            (Get_Context_Clause_Body (Ref), File, Specification
                             );
                     end if;
                     Print_Indent (File);
                     Output (File, Spacing_For_Statement_Number);
                     Output (File, "package body ");
                     Dump_Symbol_Name (File, Ref);
                     Outputln (File, " is");
                     Outputln (File);
                     Indent;
                     Dump_Statement_Database (Get_Package_Body_Decls (Ref),
                                              File, Both);
                     Exdent;
                     if not Is_Empty (Get_Head_Of_Statements_List (Ref)) then
                        Print_Indent (File);
                        Output (File, Spacing_For_Statement_Number);
                        Outputln (File, "begin");
                        Indent;
                        Dump_Statement_Database
                            (Get_Head_Of_Statements_List (Ref), File, Both);
                        Exdent;
                     end if;
                     Output (File, Spacing_For_Statement_Number);
                     Output (File, "end ");
                     Dump_Symbol_Name (File, Ref);
                     Outputln (File, ";");
                  end if;

               when Enumeration_Literal_Entry =>
                  Output (File, "-- Enumeration Literal : """);
                  Dump_Symbol_Name (File, Ref);
                  Output (File, """ of type ");
                  Dump_Symbol_Name (File, Get_Type (Ref));
                  Outputln (File);

               when others     =>
                  -- Show_Error (3009, Internal, "Dump_Symtbl");
                  null;

            end case;
         end if;
      end if;
   end Dump_Symtbl;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Dump_Comments
   --
   -- <Description>
   --
   -- This routine outputs a list of comments associated with
   -- a statement database node.
   --==========================================================================
   procedure Dump_Comments (Ref  : Ref_Statement_Database;
                            File : File_Type) is

      Comments_List : Ref_Comment := Get_Comments_List (Ref);

   begin
      while not Is_Empty (Comments_List) loop
         Output (File, Spacing_For_Statement_Number);
         Print_Indent (File);
         Outputln (File, Get_Comment (Comments_List));
         Comments_List := Get_Comments_List (Comments_List);
      end loop;
   end Dump_Comments;

   --==========================================================================
   -- <Subprogram Name>
   --
   --  Dump_Help_Info
   --
   -- <Description>
   --
   -- This proceudre outputs the help comments associated with a symbol
   -- table entry.  The STRIP_COMMENT parameter will control whether the help
   -- comment delimiter is pre-pended to the help information.
   --==========================================================================
   procedure Dump_Help_Info (Ref           : in Ref_Symbol_Table_Entry;
                             Strip_Comment : in Boolean;
                             File          : in File_Type) is

      Help_Info : Ref_Help;
   begin
      Help_Info := Get_Help_List (Ref);
      if Is_Empty (Help_Info) then
         null;
      else
         Outputln (File);
         loop
            Print_Indent (File);
            Outputln (File, Get_Help_Info (Help_Info, Strip_Comment));
            Help_Info := Get_Help_List (Help_Info);
            exit  when Is_Empty (Help_Info);
         end loop;
      end if;

   end Dump_Help_Info;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Dump_For_Statement_Trace
   --
   -- <Description>
   --
   -- This procedure reconstructs the Ada text associated with a
   -- a single Ada statement from an ACE statement database node for
   -- the purpose of tracing Ada statements as they are executed.
   -- Since the statement is being displayed to support a trace,
   -- the complete text of a compound statement is not displayed when
   -- first encountered.  For example, only the loop scheme will be
   -- displayed when a loop statement is passed in.  The body of the
   -- loop would be displayed on a one-by-one basis as each statement
   -- within the loop is executed.  The text is written to the specified
   -- File.
   --==========================================================================
   procedure Dump_For_Statement_Trace (Ref  : Ref_Statement_Database;
                                       File : File_Type) is
      Dumping_Mode : List_Mode := Specification;

   begin
      -- KLUDGE!!!
      -- ACE represents enumeration types as a parameterless function that
      -- returns the specific value of the enumeration.  When we are tracing
      -- a programs execution then we should not trace the return statement
      -- associated with the enumeration value.  This code is here, instead
      -- of in the case statement, because we do not want to dump the
      -- statement number.
      if Ref.Statement_Node.Statement_Kind = Return_Statement then
         if not Is_Empty
                (Get_Enumeration_Literal_Entry (Get_Subprogram_Reference (Ref))
                 ) then
            return ;
         end if;
      end if;

      if Is_Empty (Ref) then
         Print_Indent (File);
         Outputln (File, "null;");
      else
         if Dump_Statement_Number then
            Output
                (File,
                 Blanks (1 ..
                         5 -
                         Ace_Integer'IMAGE (Get_Statement_Number (Ref))'LENGTH)
                  & Ace_Integer'IMAGE (Get_Statement_Number (Ref)) & "| ");
         end if;

         case Ref.Statement_Node.Statement_Kind is

            when Exception_Declaration   =>
               Dump_Statement_Database
                   (Get_Name_List (Ref), File, Dumping_Mode);
               Output (File, " : ");
               Outputln (File, "exception;");

            when Object_Declaration =>
               Dump_Statement_Database
                   (Get_Name_List (Ref), File, Dumping_Mode);
               Output (File, " : ");
               Dump_Type_Name
                   (File,
                    Get_Symbol_Table_Entry
                        (Get_Expression (Get_Name_List (Ref))));
               if Has_Constraint
                      (Get_Symbol_Table_Entry
                           (Get_Expression (Get_Name_List (Ref)))) then
                  Output (File, " (");
                  Dump_Statement_Database
                      (Get_Lower_Bound
                           (Get_Symbol_Table_Entry
                                (Get_Expression (Get_Name_List (Ref)))), File,
                       Dumping_Mode);
                  Output (File, " .. ");
                  Dump_Statement_Database
                      (Get_Upper_Bound
                           (Get_Symbol_Table_Entry
                                (Get_Expression (Get_Name_List (Ref)))), File,
                       Dumping_Mode);
                  Output (File, ")");
               end if;
               if not Is_Empty (Get_Expression (Ref)) then
                  Output (File, " := ");
                  Dump_Statement_Database (Get_Expression (Ref), File,
                                           Dumping_Mode);
               end if;
               Outputln (File, ";");

            when Type_Declaration   =>
               Output (File, "type ");
               Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
               Output (File, " is ");
               if Get_Entry_Kind (Get_Symbol_Table_Entry (Ref)) = Type_Entry
                   then
                  if Is_Type_An_Enumeration_Type (Get_Symbol_Table_Entry (Ref))
                      then
                     Output (File, "(");
                     Dump_Enumeration_List (File, Get_Symbol_Table_Entry (Ref));
                     Outputln (File, ");");

                  elsif Is_Type_An_Integer_Type (Get_Symbol_Table_Entry (Ref))
                         then
                     Output (File, " range ");
                     --dump_statement_statement_database
                     -- (Get_Lower_Bound(Get_Symbol_Table_Entry(Ref));
                     --dump_statement_statement_database
                     -- (Get_Upper_Bound(Get_Symbol_Table_Entry(Ref));
                     Output
                         (File,
                          Ace_Integer'IMAGE (
                             Get_Lower_Bound (Get_Symbol_Table_Entry (Ref))));
                     Output (File, " .. ");
                     Output
                         (File,
                          Ace_Integer'IMAGE (
                             Get_Upper_Bound (Get_Symbol_Table_Entry (Ref))));
                     Outputln (File, ";");

                  else
                     Outputln (File, "<Unimplemented Type>;");
                  end if;

               elsif Get_Entry_Kind (Get_Symbol_Table_Entry (Ref)) =
                     Derived_Type_Entry then
                  Output (File, "new ");
                  Dump_Symbol_Name
                      (File, Get_Entry_Base_Type (Get_Symbol_Table_Entry (Ref))
                       );
                  if Has_Constraint (Get_Symbol_Table_Entry (Ref)) then
                     Dump_Constraint_Info (File, Get_Symbol_Table_Entry (Ref));
                  end if;
                  Outputln (File, ";");
               end if;

            when Subtype_Declaration     =>
               Output (File, "subtype ");
               Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
               Output (File, " is ");
               Dump_Symbol_Name
                   (File, Get_Entry_Base_Type (Get_Symbol_Table_Entry (Ref)));
               if Has_Constraint (Get_Symbol_Table_Entry (Ref)) then
                  Dump_Constraint_Info (File, Get_Symbol_Table_Entry (Ref));
               end if;
               Outputln (File, ";");

            when Procedure_Spec_Declaration   | Procedure_Declaration  =>
               Output (File, "procedure ");
               Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
               Dump_Parameter_List (File, Get_Symbol_Table_Entry (Ref));
               Outputln (File, ";");
               if Is_Subprogram_Built_In (Get_Symbol_Table_Entry (Ref)) then
                  Output (File, Spacing_For_Statement_Number);
                  Output (File, "pragma builtin (");
                  Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
                  Output (File, ", ");
                  Output
                      (File,
                       Ace_Integer'IMAGE (
                          Get_Built_In_Instance_Number
                              (Get_Symbol_Table_Entry (Ref))));
                  Outputln (File, ");");
               end if;

            when Function_Spec_Declaration    | Function_Declaration   =>
               Output (File, "function ");
               Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
               Dump_Parameter_List (File, Get_Symbol_Table_Entry (Ref));
               Output (File, " return ");
               Dump_Symbol_Name
                   (File, Get_Return_Type (Get_Symbol_Table_Entry (Ref)));
               Outputln (File, ";");
               if Is_Subprogram_Built_In (Get_Symbol_Table_Entry (Ref)) then
                  Output (File, Spacing_For_Statement_Number);
                  Output (File, "pragma builtin (");
                  Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
                  Output (File, ", ");
                  Output
                      (File,
                       Ace_Integer'IMAGE (
                          Get_Built_In_Instance_Number
                              (Get_Symbol_Table_Entry (Ref))));
                  Outputln (File, ");");
               end if;

            when Package_Declaration     =>
               Output (File, "package ");
               Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
               Outputln (File);

            when Package_Body_Declaration     =>
               Output (File, "package body ");
               Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
               Outputln (File);

            when Assign_Statement   =>
               Print_Indent (File);
               Dump_Statement_Database (Get_Lhs (Ref), File, Dumping_Mode);
               Output (File, " := ");
               Dump_Statement_Database (Get_Rhs (Ref), File, Dumping_Mode);
               Outputln (File, ";");

            when Loop_Statement     =>
               declare
                  Statement_Number_On_While : Boolean := True;
               begin
                  if not Is_Empty (Get_Name (Ref)) then
                     Print_Indent (File);
                     Dump_Statement_Database (Get_Name (Ref), File,
                                              Dumping_Mode);
                     Outputln (File, " :");
                     Statement_Number_On_While := False;
                  end if;

                  if not Is_Empty (Get_Loop_Scheme (Ref)) then
                     if Is_Expression_Node (Get_Loop_Scheme (Ref)) then
                        if not Statement_Number_On_While then
                           Output (File, Spacing_For_Statement_Number);
                        end if;
                        Print_Indent (File);
                        Output (File, "while ");
                        Dump_Statement_Database (Get_Loop_Scheme (Ref), File,
                                                 Dumping_Mode);
                        Output (File, " ");
                     else
                        Dump_Statement_Database (Get_Loop_Scheme (Ref), File,
                                                 Dumping_Mode);
                     end if;
                  end if;
                  Print_Indent (File);
                  Outputln (File, "loop");
               end;

            when For_Statement =>
               Output (File, "for ");
               Dump_Statement_Database (Get_For_Variable (Ref), File,
                                        Dumping_Mode);
               Output (File, " in ");
               Dump_Statement_Database
                   (Get_For_Range (Ref), File, Dumping_Mode);
               Output (File, " ");

            when If_Statement  =>
               Output (File, "if ");
               Dump_Statement_Database
                   (Get_Condition (Ref), File, Dumping_Mode);
               Outputln (File, " then");

            when Exception_Alternative   =>
               Output (File, "when ");
               Dump_Statement_Database (Get_Alternative_Choice (Ref), File,
                                        Dumping_Mode);
               Outputln (File, " => ");

            when Case_Statement     =>
               Output (File, "case ");
               Dump_Statement_Database (Get_Expression (Ref), File,
                                        Dumping_Mode);
               Outputln (File, " is");

            when Case_Alternative_Part   =>
               Output (File, "when ");
               Dump_Statement_Database (Get_Alternative_Choice (Ref), File,
                                        Dumping_Mode);
               Outputln (File, " => ");

            when Labeled_Statement  =>
               Output (File, "<<");
               Dump_Statement_Database (Get_Label_Entry (Ref), File,
                                        Dumping_Mode);
               Outputln (File, ">>");

            when Goto_Statement     =>
               Output (File, "goto ");
               Dump_Statement_Database (Get_Goto_Label (Ref), File,
                                        Dumping_Mode);
               Outputln (File, ";");

            when Exit_Statement     =>
               Output (File, "exit ");
               Dump_Statement_Database (Get_Name (Ref), File, Dumping_Mode);
               if not Is_Empty (Get_Condition (Ref)) then
                  Output (File, " when ");
                  Dump_Statement_Database (Get_Condition (Ref), File,
                                           Dumping_Mode);
               end if;
               Outputln (File, ";");

            when Return_Statement   =>
               Output (File, "return ");
               if not Is_Empty (Get_Expression (Ref)) then
                  Dump_Statement_Database (Get_Expression (Ref), File,
                                           Dumping_Mode);
               end if;
               Outputln (File, ";");

            when Block_Statement    =>
               if not Is_Empty (Get_Block_Declarations (Ref)) then
                  Outputln (File, "declare");
               else
                  Outputln (File, "begin");
               end if;

            when With_Statement     =>
               Output (File, "with ");
               Dump_Statement_Database
                   (Get_Name_List (Ref), File, Dumping_Mode);
               Outputln (File, ";");

            when Use_Statement =>
               Output (File, "use ");
               Dump_Statement_Database
                   (Get_Name_List (Ref), File, Dumping_Mode);
               Outputln (File, ";");

            when Procedure_Call     =>
               Dump_Statement_Database (Get_Subprogram_Info (Ref), File,
                                        Dumping_Mode);
               if not Is_Empty (Get_Parameter_List (Ref)) and then
                  not Is_A_Default_Expression
                      (Get_Expression (Get_Parameter_List (Ref))) then
                  Output (File, "(");
                  Dump_Statement_Database (Get_Parameter_List (Ref), File,
                                           Dumping_Mode);
                  Output (File, ")");
               end if;
               Outputln (File, ";");

            when Pragma_Statement   =>
               Output (File, "pragma ");
               Dump_Statement_Database (Get_Name (Ref), File, Dumping_Mode);
               if not Is_Empty (Get_Argument_List (Ref)) then
                  Output (File, "(");
                  Dump_Statement_Database (Get_Argument_List (Ref), File,
                                           Dumping_Mode);
                  Output (File, ")");
               end if;
               Outputln (File, ";");

            when Null_Statement     =>
               Outputln (File, "null;");

            when Unknown  =>
               null;

         end case;
      end if;
   end Dump_For_Statement_Trace;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Dump_Statement
   --
   -- <Description>
   --
   -- This routine outputs a statement.
   --==========================================================================
   procedure Dump_Statement (Ref          : Ref_Statement_Database;
                             File         : File_Type;
                             Dumping_Mode : List_Mode) is
   begin
      if not (Is_Removed (Ref)) then
         Dump_Comments (Ref, File);
         if Ref = null then
            Print_Indent (File);
            Outputln (File, "null;");

         else

            if Dump_Statement_Number then
               Output
                   (File,
                    Blanks (
                       1 ..
                       5 -
                       Ace_Integer'IMAGE (Get_Statement_Number (Ref))'LENGTH)
                     & Ace_Integer'IMAGE (Get_Statement_Number (Ref)) & "| ");
            end if;

            case Ref.Statement_Node.Statement_Kind is

               when Exception_Declaration     =>
                  Dump_Statement_Database (Get_Name_List (Ref), File,
                                           Dumping_Mode);
                  Output (File, " : ");
                  Outputln (File, "exception;");

               when Object_Declaration   =>
                  Print_Indent (File);
                  Dump_Statement_Database (Get_Name_List (Ref), File,
                                           Dumping_Mode);
                  Output (File, " : ");
                  Dump_Type_Name
                      (File,
                       Get_Symbol_Table_Entry
                           (Get_Expression (Get_Name_List (Ref))));
                  if Has_Constraint
                         (Get_Symbol_Table_Entry
                              (Get_Expression (Get_Name_List (Ref)))) then
                     Output (File, " (");
                     Dump_Statement_Database
                         (Get_Lower_Bound
                              (Get_Symbol_Table_Entry
                                   (Get_Expression (Get_Name_List (Ref)))),
                          File, Dumping_Mode);
                     Output (File, " .. ");
                     Dump_Statement_Database
                         (Get_Upper_Bound
                              (Get_Symbol_Table_Entry
                                   (Get_Expression (Get_Name_List (Ref)))),
                          File, Dumping_Mode);
                     Output (File, ")");
                  end if;
                  if not Is_Empty (Get_Expression (Ref)) then
                     Output (File, " := ");
                     Dump_Statement_Database (Get_Expression (Ref), File,
                                              Dumping_Mode);
                  end if;
                  Outputln (File, ";");

               when Type_Declaration     =>
                  if (Dumping_Mode = Specification) then
                     Dump_Help_Info (Get_Symbol_Table_Entry (Ref), False, File);
                  end if;

                  Print_Indent (File);
                  Output (File, "type ");
                  Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
                  Output (File, " is ");
                  if Get_Entry_Kind (Get_Symbol_Table_Entry (Ref)) = Type_Entry
                      then
                     if Is_Type_An_Enumeration_Type
                            (Get_Symbol_Table_Entry (Ref)) then
                        Output (File, "(");
                        Dump_Enumeration_List (File,
                                               Get_Symbol_Table_Entry (Ref));
                        Outputln (File, ");");
                     elsif Is_Type_An_Integer_Type
                               (Get_Symbol_Table_Entry (Ref)) then
                        Output (File, " range ");
                        --dump_statement_statement_database
                        -- (Get_Lower_Bound(Get_Symbol_Table_Entry(Ref));
                        --dump_statement_statement_database
                        -- (Get_Upper_Bound(Get_Symbol_Table_Entry(Ref));
                        Output
                            (File,
                             Ace_Integer'IMAGE (
                                Get_Lower_Bound (Get_Symbol_Table_Entry (Ref)))
                             );
                        Output (File, " .. ");
                        Outputln
                            (File,
                             Ace_Integer'IMAGE (
                                Get_Upper_Bound (Get_Symbol_Table_Entry (Ref)))
                             );
                        Outputln (File, ";");
                     else
                        Outputln (File, "<Unimplemented Type>;");
                     end if;
                  elsif Get_Entry_Kind (Get_Symbol_Table_Entry (Ref)) =
                        Derived_Type_Entry then
                     Output (File, "new ");
                     Dump_Symbol_Name
                         (File,
                          Get_Entry_Base_Type (Get_Symbol_Table_Entry (Ref)));
                     if Has_Constraint (Get_Symbol_Table_Entry (Ref)) then
                        Dump_Constraint_Info (File,
                                              Get_Symbol_Table_Entry (Ref));
                     end if;
                     Outputln (File, ";");
                  end if;

               when Subtype_Declaration  =>
                  if (Dumping_Mode = Specification) then
                     Dump_Help_Info (Get_Symbol_Table_Entry (Ref), False, File);
                  end if;

                  Print_Indent (File);
                  Output (File, "subtype ");
                  Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
                  Output (File, " is ");
                  Dump_Symbol_Name
                      (File, Get_Entry_Base_Type (Get_Symbol_Table_Entry (Ref))
                       );
                  if Has_Constraint (Get_Symbol_Table_Entry (Ref)) then
                     Dump_Constraint_Info (File, Get_Symbol_Table_Entry (Ref));
                  end if;
                  Outputln (File, ";");

               when Procedure_Spec_Declaration     =>
                  Dump_Help_Info (Get_Symbol_Table_Entry (Ref), False, File);

                  if Get_Context_Clause (Get_Symbol_Table_Entry (Ref)) /= null
                      then
                     Print_Indent (File);
                     Dump_Statement_Database
                         (Get_Context_Clause (Get_Symbol_Table_Entry (Ref)),
                          File, Specification);
                  end if;

                  Print_Indent (File);
                  Output (File, "procedure ");
                  Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
                  Dump_Parameter_List (File, Get_Symbol_Table_Entry (Ref));

                  Outputln (File, ";");
                  Outputln (File);

               when Procedure_Declaration     =>
                  if (Dumping_Mode /= Specification) and
                     not Is_Empty
                         (Get_Head_Of_Statements_List
                              (Get_Symbol_Table_Entry (Ref))) then

                     if Get_Context_Clause (Get_Symbol_Table_Entry (Ref)) /=
                        null then
                        Print_Indent (File);
                        Dump_Statement_Database
                            (Get_Context_Clause (Get_Symbol_Table_Entry (Ref)),
                             File, Specification);
                     end if;

                     Print_Indent (File);
                     Output (File, "procedure ");
                     Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
                     Dump_Parameter_List (File, Get_Symbol_Table_Entry (Ref));

                     Outputln (File, " is");
                     Outputln (File);
                     if not Is_Empty
                            (Get_Head_Of_Declarations_List
                                 (Get_Symbol_Table_Entry (Ref))) then
                        Indent;
                        Dump_Statement_Database
                            (Get_Head_Of_Declarations_List
                                 (Get_Symbol_Table_Entry (Ref)), File,
                             Dumping_Mode);
                        Exdent;
                        Outputln (File);
                     end if;
                     Output (File, Spacing_For_Statement_Number);
                     Print_Indent (File);
                     Outputln (File, "begin");
                     Indent;
                     Dump_Statement_Database
                         (Get_Head_Of_Statements_List
                              (Get_Symbol_Table_Entry (Ref)), File,
                          Dumping_Mode);
                     Exdent;
                     --if not Is_Empty () then
                     -- Exceptions
                     --end if;
                     Output (File, Spacing_For_Statement_Number);
                     Print_Indent (File);
                     Output (File, "end ");
                     Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
                     Outputln (File, ";");
                     Outputln (File);

                  end if;

               when Function_Spec_Declaration =>
                  Dump_Help_Info (Get_Symbol_Table_Entry (Ref), False, File);

                  if Get_Context_Clause (Get_Symbol_Table_Entry (Ref)) /= null
                      then
                     Print_Indent (File);
                     Dump_Statement_Database
                         (Get_Context_Clause (Get_Symbol_Table_Entry (Ref)),
                          File, Specification);
                  end if;

                  Print_Indent (File);
                  Output (File, "function ");
                  Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
                  Dump_Parameter_List (File, Get_Symbol_Table_Entry (Ref));
                  Output (File, " return ");
                  Dump_Symbol_Name
                      (File, Get_Return_Type (Get_Symbol_Table_Entry (Ref)));
                  Outputln (File, ";");
                  Outputln (File);

               when Function_Declaration =>
                  if (Dumping_Mode /= Specification) and
                     not Is_Empty
                         (Get_Head_Of_Statements_List
                              (Get_Symbol_Table_Entry (Ref))) then

                     if Get_Context_Clause (Get_Symbol_Table_Entry (Ref)) /=
                        null then
                        Print_Indent (File);
                        Dump_Statement_Database
                            (Get_Context_Clause (Get_Symbol_Table_Entry (Ref)),
                             File, Specification);
                     end if;

                     Print_Indent (File);
                     Output (File, "function ");
                     Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
                     Dump_Parameter_List (File, Get_Symbol_Table_Entry (Ref));
                     Output (File, " return ");
                     Dump_Symbol_Name
                         (File, Get_Return_Type (Get_Symbol_Table_Entry (Ref)));

                     Outputln (File, " is");
                     Outputln (File);
                     if not Is_Empty
                            (Get_Head_Of_Declarations_List
                                 (Get_Symbol_Table_Entry (Ref))) then
                        Indent;
                        Dump_Statement_Database
                            (Get_Head_Of_Declarations_List
                                 (Get_Symbol_Table_Entry (Ref)), File,
                             Dumping_Mode);
                        Exdent;
                        Outputln (File);
                     end if;
                     Output (File, Spacing_For_Statement_Number);
                     Print_Indent (File);
                     Outputln (File, "begin");
                     Indent;
                     Dump_Statement_Database
                         (Get_Head_Of_Statements_List
                              (Get_Symbol_Table_Entry (Ref)), File,
                          Dumping_Mode);
                     Exdent;
                     Output (File, Spacing_For_Statement_Number);
                     Print_Indent (File);
                     Output (File, "end ");
                     Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
                     Outputln (File, ";");
                     Outputln (File);

                  end if;

               when Package_Declaration  =>
                  if (Dumping_Mode = Specification) then
                     if Get_Context_Clause (Get_Symbol_Table_Entry (Ref)) /=
                        null then
                        Print_Indent (File);
                        Dump_Statement_Database
                            (Get_Context_Clause (Get_Symbol_Table_Entry (Ref)),
                             File, Specification);
                     end if;

                     Dump_Help_Info (Get_Symbol_Table_Entry (Ref), False, File);

                     Print_Indent (File);
                     Output (File, "package ");
                     Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
                     Outputln (File, " is");
                     Outputln (File);
                     Indent;
                     Dump_Statement_Database
                         (Get_Package_Spec_Decls (Get_Symbol_Table_Entry (Ref)),
                          File, Specification);

                     -- Dump out help info here so it gets re-associated
                     -- to the package declaration.

                     Exdent;
                     Output (File, Spacing_For_Statement_Number);
                     Print_Indent (File);
                     Output (File, "end ");
                     Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
                     Outputln (File, ";");
                     Outputln (File);
                  end if;
               when Package_Body_Declaration  =>
                  if (Dumping_Mode = Implementation) and
                     Has_A_Body (Get_Symbol_Table_Entry (Ref)) then
                     if Get_Context_Clause_Body (Get_Symbol_Table_Entry (Ref))
                         /= null then
                        Print_Indent (File);
                        Dump_Statement_Database
                            (Get_Context_Clause_Body
                                 (Get_Symbol_Table_Entry (Ref)), File);
                     end if;
                     Print_Indent (File);
                     Output (File, "package body ");
                     Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
                     Outputln (File, " is");
                     Outputln (File);
                     Indent;
                     Dump_Statement_Database
                         (Get_Package_Body_Decls (Get_Symbol_Table_Entry (Ref)),
                          File, Implementation);
                     Exdent;
                     Outputln (File);
                     if not Is_Empty
                            (Get_Head_Of_Statements_List
                                 (Get_Symbol_Table_Entry (Ref))) then
                        Output (File, Spacing_For_Statement_Number);
                        Print_Indent (File);
                        Outputln (File, "begin");
                        Indent;
                        Dump_Statement_Database
                            (Get_Head_Of_Statements_List
                                 (Get_Symbol_Table_Entry (Ref)), File,
                             Implementation);
                        Exdent;
                     end if;
                     Output (File, Spacing_For_Statement_Number);
                     Print_Indent (File);
                     Output (File, "end ");
                     Dump_Symbol_Name (File, Get_Symbol_Table_Entry (Ref));
                     Outputln (File, ";");
                     Outputln (File);
                  end if;

               when Assign_Statement     =>
                  Print_Indent (File);
                  Dump_Statement_Database (Get_Lhs (Ref), File, Dumping_Mode);
                  Output (File, " := ");
                  Dump_Statement_Database (Get_Rhs (Ref), File, Dumping_Mode);
                  Outputln (File, ";");

               when Loop_Statement  =>
                  declare
                     Statement_Number_On_While : Boolean := True;
                  begin
                     if not Is_Empty (Get_Name (Ref)) then
                        Print_Indent (File);
                        Dump_Statement_Database (Get_Name (Ref), File,
                                                 Dumping_Mode);
                        Outputln (File, " :");
                        Statement_Number_On_While := False;
                     end if;

                     Print_Indent (File);
                     if not Is_Empty (Get_Loop_Scheme (Ref)) then
                        if Is_Expression_Node (Get_Loop_Scheme (Ref)) then
                           if not Statement_Number_On_While then
                              Output (File, Spacing_For_Statement_Number);
                           end if;
                           Output (File, "while ");
                           Dump_Statement_Database
                               (Get_Loop_Scheme (Ref), File, Dumping_Mode);
                           Output (File, " ");
                        else
                           Dump_Statement_Database
                               (Get_Loop_Scheme (Ref), File, Dumping_Mode);
                        end if;
                     end if;
                     Outputln (File, "loop");
                     Indent;
                     Dump_Statement_Database
                         (Get_Head_Of_Statements_List (Ref), File, Dumping_Mode
                          );
                     Exdent;
                     Output (File, Spacing_For_Statement_Number);
                     Print_Indent (File);
                     Output (File, "end loop ");
                     Dump_Statement_Database (Get_Name (Ref), File,
                                              Dumping_Mode);
                     Outputln (File, ";");
                  end;

               when For_Statement   =>
                  Output (File, "for ");
                  Dump_Statement_Database (Get_For_Variable (Ref), File,
                                           Dumping_Mode);
                  Output (File, " in ");
                  Dump_Statement_Database (Get_For_Range (Ref), File,
                                           Dumping_Mode);
                  Output (File, " ");

               when If_Statement    =>
                  Print_Indent (File);
                  Output (File, "if ");
                  Dump_Statement_Database (Get_Condition (Ref), File,
                                           Dumping_Mode);
                  Outputln (File, " then");
                  Indent;
                  if not Is_Empty (Get_Head_Of_True_Task_Statements_List (Ref))
                      then
                     Dump_Statement_Database
                         (Get_Head_Of_True_Task_Statements_List (Ref), File,
                          Dumping_Mode);
                  else
                     Print_Indent (File);
                     Outputln (File, "null;");
                  end if;
                  Exdent;
                  if not Is_Empty
                         (Get_Head_Of_False_Task_Statements_List (Ref)) then
                     Output (File, Spacing_For_Statement_Number);
                     Print_Indent (File);
                     Outputln (File, "else");
                     Indent;
                     Dump_Statement_Database
                         (Get_Head_Of_False_Task_Statements_List (Ref), File,
                          Dumping_Mode);
                     Exdent;
                  end if;
                  Output (File, Spacing_For_Statement_Number);
                  Print_Indent (File);
                  Outputln (File, "end if;");

               when Exception_Alternative     =>
                  Print_Indent (File);
                  Output (File, "when ");
                  Dump_Statement_Database (Get_Alternative_Choice (Ref), File,
                                           Dumping_Mode);
                  Outputln (File, " => ");
                  Indent;
                  Dump_Statement_Database (Get_Alternative_Statements (Ref),
                                           File, Dumping_Mode);
                  Exdent;
                  if not Is_Empty (Get_Next_Alternative (Ref)) then
                     Dump_Statement_Database
                         (Get_Next_Alternative (Ref), File, Dumping_Mode);
                  end if;

               when Case_Statement  =>
                  Print_Indent (File);
                  Output (File, "case ");
                  Dump_Statement_Database (Get_Expression (Ref), File,
                                           Dumping_Mode);
                  Outputln (File, " is");
                  Indent;
                  Dump_Statement_Database (Get_Alternatives_List (Ref), File,
                                           Dumping_Mode);
                  Exdent;
                  Output (File, Spacing_For_Statement_Number);
                  Print_Indent (File);
                  Outputln (File, "end case;");

               when Case_Alternative_Part     =>
                  Print_Indent (File);
                  Output (File, "when ");
                  Dump_Statement_Database (Get_Alternative_Choice (Ref), File,
                                           Dumping_Mode);
                  Outputln (File, " => ");
                  Indent;
                  Dump_Statement_Database (Get_Alternative_Statements (Ref),
                                           File, Dumping_Mode);
                  Exdent;
                  if not Is_Empty (Get_Next_Alternative (Ref)) then
                     Dump_Statement_Database
                         (Get_Next_Alternative (Ref), File, Dumping_Mode);
                  end if;

               when Labeled_Statement    =>
                  Print_Indent (File);
                  Output (File, "<<");
                  Dump_Statement_Database (Get_Label_Entry (Ref), File,
                                           Dumping_Mode);
                  Outputln (File, ">>");

               when Goto_Statement  =>
                  Print_Indent (File);
                  Output (File, "goto ");
                  Dump_Statement_Database (Get_Goto_Label (Ref), File,
                                           Dumping_Mode);
                  Outputln (File, ";");

               when Exit_Statement  =>
                  Print_Indent (File);
                  Output (File, "exit ");
                  Dump_Statement_Database (Get_Name (Ref), File, Dumping_Mode);
                  if not Is_Empty (Get_Condition (Ref)) then
                     Output (File, " when ");
                     Dump_Statement_Database (Get_Condition (Ref), File,
                                              Dumping_Mode);
                  end if;
                  Outputln (File, ";");

               when Return_Statement     =>
                  Print_Indent (File);
                  Output (File, "return ");
                  if not Is_Empty (Get_Expression (Ref)) then
                     Dump_Statement_Database (Get_Expression (Ref), File,
                                              Dumping_Mode);
                  end if;
                  Outputln (File, ";");

               when Block_Statement =>
                  if not Is_Empty (Get_Block_Declarations (Ref)) then
                     Print_Indent (File);
                     Outputln (File, "declare");
                     Indent;
                     Dump_Statement_Database (Get_Block_Declarations (Ref),
                                              File, Dumping_Mode);
                     Exdent;
                  end if;

                  Print_Indent (File);
                  Outputln (File, "begin");
                  Indent;
                  Dump_Statement_Database (Get_Head_Of_Statements_List (Ref),
                                           File, Dumping_Mode);
                  Exdent;
                  Print_Indent (File);
                  Outputln (File, "end;");
                  Outputln (File);

               when With_Statement  =>
                  Print_Indent (File);
                  Output (File, "with ");
                  Dump_Statement_Database (Get_Name_List (Ref), File,
                                           Dumping_Mode);
                  Outputln (File, ";");

               when Use_Statement   =>
                  Print_Indent (File);
                  Output (File, "use ");
                  Dump_Statement_Database (Get_Name_List (Ref), File,
                                           Dumping_Mode);
                  Outputln (File, ";");

               when Procedure_Call  =>
                  Print_Indent (File);
                  Dump_Statement_Database (Get_Subprogram_Info (Ref), File,
                                           Dumping_Mode);
                  if not Is_Empty (Get_Parameter_List (Ref)) and then
                     not Is_A_Default_Expression
                         (Get_Expression (Get_Parameter_List (Ref))) then
                     Output (File, "(");
                     Dump_Statement_Database (Get_Parameter_List (Ref), File,
                                              Dumping_Mode);
                     Output (File, ")");
                  end if;
                  Outputln (File, ";");

               when Pragma_Statement     =>
                  Print_Indent (File);
                  Output (File, "pragma ");
                  Dump_Statement_Database (Get_Name (Ref), File, Dumping_Mode);
                  if not Is_Empty (Get_Argument_List (Ref)) then
                     Output (File, "(");
                     Dump_Statement_Database (Get_Argument_List (Ref), File,
                                              Dumping_Mode);
                     Output (File, ")");
                  end if;
                  Outputln (File, ";");

               when Null_Statement  =>
                  Print_Indent (File);
                  Outputln (File, "null;");

               when Unknown    =>
                  null;

            end case;
         end if;
      end if;

      if not Is_Empty (Get_Next_Statement (Ref)) then
         Dump_Statement_Database (Get_Next_Statement (Ref), File, Dumping_Mode);
      end if;

   end Dump_Statement;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Dump_Expression
   --
   -- <Description>
   --
   -- This routine outputs an expression.
   --==========================================================================
   procedure Dump_Expression (Ref             : Ref_Statement_Database;
                              File            : File_Type;
                              Dumping_Mode    : List_Mode;
                              Dump_Init_Exprs : Boolean := False) is
   begin
      case Get_Expression_Kind (Ref) is

         when Dot_Operator     =>
            Dump_Statement_Database (Get_Object (Ref), File, Dumping_Mode);
            Output (File, ".");
            Dump_Statement_Database (Get_Dot_Component (Ref), File,
                                     Dumping_Mode);

         when Left_Parenthesis_Operator  =>
            if Is_Infix
                   (Convert_To_String
                        (Get_Symbolic_Name (Get_Subprogram_Info (Ref)))) then
               Output (File, "(");
               Dump_Statement_Database
                   (Get_Expression (Get_Parameter_List (Ref)), File,
                    Dumping_Mode);
               Quote_Infix := False;
               Output (File, " ");
               Dump_Statement_Database (Get_Object (Ref), File, Dumping_Mode);
               Output (File, " ");
               Quote_Infix := True;
               Dump_Statement_Database
                   (Get_Next_List_Item (Get_Parameter_List (Ref)), File,
                    Dumping_Mode);
               Output (File, ")");
            else
               Dump_Statement_Database (Get_Object (Ref), File, Dumping_Mode);
               -- check for default expressions;
               -- parts below are only valid until named parameters are
               -- implemented [tbd];
               if (Get_Object (Ref) = null or else
                  Get_Parameter_List (Ref) /= null) and then
                  (Dump_Init_Exprs or
                  not (Is_A_Default_Expression
                      (Get_Expression (Get_Parameter_List (Ref))))) then
                  Output (File, "(");
                  Dump_Statement_Database (Get_List_Of_Items (Ref), File,
                                           Dumping_Mode);
                  Output (File, ")");
               end if;
            end if;

         when List_Item   =>

            if not (Is_A_Default_Expression (Get_Expression (Ref))) then

               Dump_Statement_Database (Get_Expression (Ref), File,
                                        Dumping_Mode);

               if (Get_Next_List_Item (Ref) /= null) and then
                  not (Is_A_Default_Expression
                      (Get_Expression (Get_Next_List_Item (Ref)))) then
                  Output (File, ", ");
                  Dump_Statement_Database (Get_Next_List_Item (Ref), File,
                                           Dumping_Mode);
               end if;
            end if;

         when Aggregate_Component   =>
            Dump_Statement_Database (Get_Aggregate_Choice (Ref), File,
                                     Dumping_Mode);
            Output (File, " => ");
            Dump_Statement_Database (Get_Aggregate_Choice (Ref), File,
                                     Dumping_Mode);

         when Range_Expression =>
            if Get_Name_Range (Ref) /= null then
               Dump_Statement_Database (Get_Name_Range (Ref), File,
                                        Dumping_Mode);
               Output (File, " range ");
            end if;
            Dump_Statement_Database (Get_Lower_Bound (Ref), File, Dumping_Mode);
            Output (File, "..");
            Dump_Statement_Database (Get_Upper_Bound (Ref), File, Dumping_Mode);

         when Attribute_Tick   =>
            Dump_Statement_Database (Get_Attribute_Prefix (Ref), File,
                                     Dumping_Mode);
            Output (File, "'");
            Dump_Statement_Database (Get_Attribute (Ref), File, Dumping_Mode);

         when String_Node =>
            Output
                (File,
                 '"' &
                 Double_Quotes (Convert_To_String (Get_Symbolic_Name (Ref))) &
                 '"');

         when Character_Node   =>
            Output (File,
                    ''' & Convert_To_String (Get_Symbolic_Name (Ref)) & ''');

         when Constant_Node    =>
            Output (File, Convert_To_String (Get_Symbolic_Name (Ref)));

         when Terminal_Node    =>
            if Is_Infix (Convert_To_String (Get_Symbolic_Name (Ref))) and
               Quote_Infix then
               Output (File, """");
               Output (File, Convert_To_String (Get_Symbolic_Name (Ref)));
               Output (File, """");
            else
               Output (File, Convert_To_String (Get_Symbolic_Name (Ref)));
            end if;

         when Array_Reference                 |
              Function_Call                   |
              Type_Conversion                 |
              Record_Reference                |
              Dot_Notation_Variable_Reference |
              Unknown     =>
            null;

      end case;
   end Dump_Expression;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Dump_Enclosing_Pkgs
   --
   -- <Description>
   --
   -- This procedure dumps the names of the enclosing packages for the
   -- given symbol table entry.  The STRIP_HYPH parameter will control
   -- whether the helpcomment delimiter is pre-pended to the help information.
   --==========================================================================
   procedure Dump_Enclosing_Pkgs (Ref  : in Ref_Symbol_Table_Entry;
                                  File : in File_Type) is

      procedure Write_Enclosing_Pkgs (Ref : Ref_Symbol_Table_Entry) is
         Encls_Pkg : Ref_Symbol_Table_Entry := Get_Enclosing_Package (Ref);
      begin
         if Is_Empty (Encls_Pkg) then
            null;
         else
            Write_Enclosing_Pkgs (Encls_Pkg);
            Output (File,
                    Convert_To_String (Get_Symbolic_Name (Encls_Pkg)) & '.');
         end if;
      end Write_Enclosing_Pkgs;

   begin
      if not Is_Empty (Get_Enclosing_Package (Ref)) then
         Output (File, "Declared in ");
         Write_Enclosing_Pkgs (Ref);
         Outputln (File);
      end if;

   end Dump_Enclosing_Pkgs;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Dump_Statement_Database
   --
   -- <Description>
   --
   -- This procedure reconstructs the Ada text associated with an
   -- ACE statement database node.  The text is written to the specified
   -- File, and the display of bodies for packages and subprograms is
   -- controlled by the Dumping_Mode parameter.
   --==========================================================================
   procedure Dump_Statement_Database (Ref          : Ref_Statement_Database;
                                      File         : File_Type;
                                      Dumping_Mode : List_Mode := Specification
                                      ) is
   begin
      if not Is_Empty (Ref) then
         case Ref.Node_Kind is
            when Statement_Node_Type     =>
               Dump_Statement (Ref, File, Dumping_Mode);
            when Expression_Node_Type    =>
               Dump_Expression (Ref, File, Dumping_Mode);
            when Unknown  =>
               null;
         end case;
      end if;
   end Dump_Statement_Database;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Dump_Entire_Symtbl
   --
   -- <Description>
   --
   -- This procedure displays each and every entry in the entire ACE
   -- symbol table.  Each symbol table entry is reconstructed into the
   -- Ada code that caused the symbol table entry to be generated.
   --==========================================================================
   procedure Dump_Entire_Symtbl (Context   : Context_Control.Context_Definition;
                                 Dump_Mode : in List_Mode := Specification) is
      Current_Entry : Ref_Symbol_Table_Entry;
   begin
      Put_Line ("Entire Symbol Table Dump =>");
      Current_Entry := Get_Top_Of_Symtbl (Context);
      while not Is_Empty (Current_Entry) loop
         Dump_Symtbl (Current_Entry, Standard_Output, Dump_Mode);
         New_Line;
         Current_Entry := Get_Next_Entry (Current_Entry);
      end loop;
   end Dump_Entire_Symtbl;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Dump_Objs
   --
   --  <Description>
   --
   --    Dump_Objs will list the names of visible objects on the symbol
   --  table.
   --==========================================================================
   procedure Dump_Objs (Context   : Context_Control.Context_Definition;
                        Obj_Type  : in Obj_Declarations;
                        Specifier : in Ref_String;
                        File      : in File_Type) is
      Current_Entry   : Ref_Symbol_Table_Entry;
      Curr_Entry_Kind : Type_Entry_Enumeration;

      use Symbol_Table_Structures;

      type Obj_Node;
      type Ref_Obj_Node is access Obj_Node;

      type Obj_Node is
         record
            Obj_Name : Ref_String;
            Next_Obj : Ref_Obj_Node;
            Count    : Integer := 1;
         end record;

      Obj_Name_List : Ref_Obj_Node;

      function Is_Empty (Node : in Ref_Obj_Node) return Boolean is
      begin
         return (Node = null);
      end Is_Empty;

      -- This procedure will write the names of the objects one the name list
      -- and destroy the list.
      procedure Dump_Names (Obj_Names_Head : in out Ref_Obj_Node;
                            File           : in File_Type) is
         Curr_Node : Ref_Obj_Node := Obj_Names_Head;

         -- This procedure destroys the object names list
         procedure Delete_Obj_Names (Zap_Obj_Names : in out Ref_Obj_Node) is
            procedure Unchecked_Del_Obj is new Unchecked_Deallocation
                (Obj_Node, Ref_Obj_Node);
            Temp_Ptr : Ref_Obj_Node;
         begin -- Delete_Obj_Names
            Temp_Ptr := Zap_Obj_Names;
            while Temp_Ptr /= null loop
               Zap_Obj_Names := Temp_Ptr.Next_Obj;
               Unchecked_Del_Obj (Temp_Ptr);
               Temp_Ptr      := Zap_Obj_Names;
            end loop;
         end Delete_Obj_Names;

      begin -- Dump_Names
         while not Is_Empty (Curr_Node) loop
            -- place infix names within quotes
            if Is_Infix (Convert_To_String (Curr_Node.Obj_Name)) then
               Outputln (File,
                         """" & Convert_To_String (Curr_Node.Obj_Name) & """");
            else
               Outputln (File, Convert_To_String (Curr_Node.Obj_Name));
            end if;
            Curr_Node := Curr_Node.Next_Obj;
         end loop;
         Delete_Obj_Names (Obj_Names_Head);
      end Dump_Names;

      -- This routine will add a name to the obj_list if it does not
      -- already exst, and increment the counter if it is already on.
      procedure Add_Name (Obj_List : in out Ref_Obj_Node;
                          New_Name : in Ref_String) is
         Curr_Node : Ref_Obj_Node := Obj_List;
         Prev_Node : Ref_Obj_Node := null;

         -- This routine will increment the name counter for the node
         procedure Increment_Name_Count (Node : in Ref_Obj_Node) is
         begin -- Increment_Name_Count
            Node.Count := Node.Count + 1;
         end Increment_Name_Count;

         -- procedure Insert adds a new name node to the list
         procedure Insert (Name         : Ref_String;
                           Insert_Point : in out Ref_Obj_Node) is
            New_Obj_Node : Ref_Obj_Node := new Obj_Node;

         begin -- Insert
            New_Obj_Node.Obj_Name := Name;
            if Insert_Point = null then
               New_Obj_Node.Next_Obj := null;
            else
               New_Obj_Node.Next_Obj := Insert_Point;
            end if;

            Insert_Point := New_Obj_Node;

         end Insert;

      begin -- Add name
         if (Curr_Node = null) then
            Insert (New_Name, Obj_List);
         else
            while (Curr_Node /= null and then
               Convert_To_String (Curr_Node.Obj_Name) <
               Convert_To_String (New_Name)) loop
               Prev_Node := Curr_Node;
               Curr_Node := Curr_Node.Next_Obj;
            end loop;

            if (Is_Empty (Curr_Node)) then
               -- append to end of list
               Insert (New_Name, Prev_Node.Next_Obj);

            elsif Are_Symbolic_Names_Equal (Curr_Node.Obj_Name, New_Name) then
               -- duplicate entry
               Increment_Name_Count (Curr_Node);

            elsif (Is_Empty (Prev_Node)) then
               -- head of list
               Insert (New_Name, Obj_List);

            else
               Insert (New_Name, Prev_Node.Next_Obj);

            end if;
         end if;

      end Add_Name;

   begin -- Dump_Objs
      Current_Entry := Get_Top_Of_Symtbl (Context);
      while not Is_Empty (Current_Entry) loop
         if Is_Visible (Current_Entry) then
            Curr_Entry_Kind := Get_Entry_Kind (Current_Entry);
            case Obj_Type is
               when Any   =>
                  Add_Name (Obj_Name_List, Get_Symbolic_Name (Current_Entry));

               when Object_Declaration   =>
                  if (Type_Entry_Enumeration (Curr_Entry_Kind) = Object_Entry)
                      and then
                     (Convert_To_String (Specifier) = "" or else
                     Are_Symbolic_Names_Equal
                         (Specifier,
                          Get_Symbolic_Name (Get_Type (Current_Entry)))) then
                     Add_Name
                         (Obj_Name_List, Get_Symbolic_Name (Current_Entry));

                  end if;

               when Type_Mark  =>
                  if (Type_Entry_Enumeration (Curr_Entry_Kind) = Type_Entry
                      or else
                     Type_Entry_Enumeration (Curr_Entry_Kind) =
                     Derived_Type_Entry or else
                     Type_Entry_Enumeration (Curr_Entry_Kind) = Subtype_Entry)
                      and then
                     (Convert_To_String (Specifier) = "" or else
                     Are_Symbolic_Names_Equal
                         (Specifier,
                          Get_Symbolic_Name (Get_Base_Type (Current_Entry))))
                      then
                     Add_Name
                         (Obj_Name_List, Get_Symbolic_Name (Current_Entry));
                  end if;

               when Type_Declaration     =>
                  if (Type_Entry_Enumeration (Curr_Entry_Kind) = Type_Entry
                      or else
                     Type_Entry_Enumeration (Curr_Entry_Kind) =
                     Derived_Type_Entry) and then
                     (Convert_To_String (Specifier) = "" or else
                     Are_Symbolic_Names_Equal
                         (Specifier,
                          Get_Symbolic_Name (Get_Base_Type (Current_Entry))))
                      then
                     Add_Name
                         (Obj_Name_List, Get_Symbolic_Name (Current_Entry));
                  end if;

               when Subtype_Declaration  =>
                  if (Type_Entry_Enumeration (Curr_Entry_Kind) = Subtype_Entry)
                      and then
                     (Convert_To_String (Specifier) = "" or else
                     Are_Symbolic_Names_Equal
                         (Specifier,
                          Get_Symbolic_Name (Get_Base_Type (Current_Entry))))
                      then

                     Add_Name
                         (Obj_Name_List, Get_Symbolic_Name (Current_Entry));
                  end if;

               when Package_Declaration  =>
                  if (Type_Entry_Enumeration (Curr_Entry_Kind) = Package_Entry)
                      then
                     Add_Name
                         (Obj_Name_List, Get_Symbolic_Name (Current_Entry));
                  end if;

               when Subprogram_Declaration    =>
                  if (Type_Entry_Enumeration (Curr_Entry_Kind) =
                     Procedure_Entry or else
                     Type_Entry_Enumeration (Curr_Entry_Kind) = Function_Entry)
                      then
                     Add_Name
                         (Obj_Name_List, Get_Symbolic_Name (Current_Entry));
                  end if;

               when Procedure_Declaration     =>
                  if (Type_Entry_Enumeration (Curr_Entry_Kind) =
                     Procedure_Entry) then
                     Add_Name
                         (Obj_Name_List, Get_Symbolic_Name (Current_Entry));
                  end if;

               when Function_Declaration =>
                  if (Type_Entry_Enumeration (Curr_Entry_Kind) = Function_Entry
                     ) and then
                     (Convert_To_String (Specifier) = "" or else
                     Convert_To_String (Specifier) =
                     Convert_To_String
                         (Get_Symbolic_Name (Get_Return_Type (Current_Entry))))
                      then

                     Add_Name
                         (Obj_Name_List, Get_Symbolic_Name (Current_Entry));
                  end if;

               when Label_Declaration    =>
                  if (Type_Entry_Enumeration (Curr_Entry_Kind) = Label_Entry)
                      then
                     Add_Name
                         (Obj_Name_List, Get_Symbolic_Name (Current_Entry));
                  end if;
            end case;
         end if;
         Current_Entry := Get_Next_Entry (Current_Entry);

      end loop;
      Dump_Names (Obj_Name_List, Current_Output);
   end Dump_Objs;

end Dump_Routines;
