--|$Log:	help.a,v $
--|Revision 8.0  91/01/17  10:45:46  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Help_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--     This package contains support routines for the help comments
--  that may be associated with ACE objects.  This routines are
--  patterned after similar routines for comments.
--
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------
-- ==========================================================================

with Error_Handler;
use Error_Handler;

package body Help_Pkg is

   -- Holder for help comments in interpreted source code.

   Help_List_Head : Ref_Help := null;
   Help_List      : Ref_Help := null;

   type Help_Q_Entry;

   type Help_Q_Ptr is access Help_Q_Entry;

   type Help_Q_Entry is
      record
         Help_Comment : Ref_Help;
         Next         : Help_Q_Ptr;
      end record;

   type Help_Q is
      record
         Front, Rear : Help_Q_Ptr;
      end record;

   Pkg_Help_Q : Help_Q;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Process_Help
   --
   --  <Description>
   --
   --    This routine stores up help comments, so that they may be
   --    associated with the nearest unit.  This needs some
   --    improvement.  Help comments sometimes move from their original
   --    position to a position next to the nearest unit.
   --
   --  <Called By>
   --
   --    The parsers.
   --
   --==========================================================================
   procedure Process_Help (Comment : String) is

   begin
      if Help_List_Head = null then
         Help_List      := new Help'(new String'(Comment),
                                null);
         Help_List_Head := Help_List;
      else
         Help_List.Next := new Help'(new String'(Comment),
                                     null);
         Help_List      := Help_List.Next;
      end if;
   end Process_Help;

   -- -------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Help
   --
   --  <Description>
   --
   --    This routine returns the stored help information, so that it may be
   --    associated with the nearest unit.  After it is called the
   --    stored help comments are emptied, so that they are not associated
   --    twice.
   --
   --==========================================================================
   function Get_Help return Ref_Help is

      Current_Help : Ref_Help;

      procedure Void_Help is
      begin
         Help_List_Head := null;
      end Void_Help;

   begin
      Current_Help := Help_List_Head;
      Void_Help;
      return Current_Help;
   end Get_Help;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Help_Info
   --
   --  <Description>
   --
   --    Get_Help_Info will return a line of help information from the given
   --   Help_Node.  The Strip flag will instruct the routine to strip off
   --   the help_comment symbol "--?" from the string.
   --==========================================================================
   function Get_Help_Info (Help_Node : in Ref_Help;
                           Strip     : in Boolean) return String is
   begin
      if Strip then
         return (Help_Node.Comment.all (4 .. Help_Node.Comment.all'LAST));
      else
         return (Help_Node.Comment.all);
      end if;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Help_Info");
         raise ;
   end Get_Help_Info;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Help_List
   --
   --  <Description>
   --
   --    Get_Help_List returns the next help_comment in the list.
   --
   --==========================================================================
   function Get_Help_List (Node : in Ref_Help) return Ref_Help is
   begin
      return Node.Next;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "get_help_list2");
         raise ;
   end Get_Help_List;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Clear_Help
   --
   --  <Description>
   --
   --    Clear_Help will clear out the help buffer.
   --
   --==========================================================================
   procedure Clear_Help is
      Help_Tmp : Ref_Help;
   begin
      -- should free up the storage while we're at it, but for now...
      Help_Tmp := Get_Help;
   end Clear_Help;

   -- -------------------------------------------------------------------------

   --=========================================================================
   --  <Subprogram Name>
   --
   --    Nq_Pkg_Help
   --
   --  <Description>
   --
   --    Holds help comments for a package until a symbol table node is created.
   --
   --==========================================================================
   procedure Nq_Pkg_Help (Help_Comments : in Ref_Help) is
      New_Help : Help_Q_Ptr := new Help_Q_Entry'(Help_Comments, null);
   begin
      if (Pkg_Help_Q.Front = null) then
         Pkg_Help_Q.Front := New_Help;
         Pkg_Help_Q.Rear  := New_Help;
      else
         Pkg_Help_Q.Rear.Next := New_Help;
         Pkg_Help_Q.Rear      := New_Help;
      end if;
   end Nq_Pkg_Help;

   function Dq_Pkg_Help return Ref_Help is
      Help_Holder : Ref_Help;
   begin
      if (Pkg_Help_Q.Front = null) then
         return null;
      else
         Help_Holder      := Pkg_Help_Q.Front.Help_Comment;
         --? should free memory
         Pkg_Help_Q.Front := Pkg_Help_Q.Front.Next;
         return Help_Holder;
      end if;
   end Dq_Pkg_Help;

end Help_Pkg;
