--|$Log:	miscs.a,v $
--|Revision 8.0  91/01/17  10:45:53  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Misc_Support_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    This package contains a collection of miscellaneous routines that
--    did not fall logically within any of the other packages for the
--    Ada Command Language Interpreter.  These routines include conversion
--    routines between a reference to a statement database node and a fake
--    reference to a statement database node; routines to determine whether
--    a particular data structure (e.g. statement database node, symbol table
--    entry, ...) is empty; routines to acquire the discriminant value from
--    a record; and routine to duplicate a statement database node.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Statement_Database;
use Statement_Database;

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Help_Pkg;
use Help_Pkg;

with Node_Get_Routines;
use Node_Get_Routines;

with Node_Set_Routines;
use Node_Set_Routines;

with Node_Create_Routines;
use Node_Create_Routines;

with Symtbl_Entry_Get_Routines;
use Symtbl_Entry_Get_Routines;

with Symtbl_Entry_Set_Routines;
use Symtbl_Entry_Set_Routines;

with Symbol_Table_Routines;
use Symbol_Table_Routines;

with Declaration_Support_Routines;
use Declaration_Support_Routines;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Literals;
use Literals;

with Error_Handler;
use Error_Handler;

with Unchecked_Conversion;

with Unchecked_Deallocation;

with Context_Control;
use  Context_Control;

-- following "with" only needed for "Not Implemented" messages
with Text_Io;

package body Misc_Support_Routines is

   Anonymous_Counter : Natural := 0;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Convert_Fake
   --
   --  <Description>
   --
   --    A fake reference to the statement database is converted into an
   --    actual reference to the statement database through use of the Ada
   --    UNCHECKED_CONVERSION generic subprogram.
   --==========================================================================
   function Convert_Fake
       (Node : in Ref_Statement_Database_Fake) return Ref_Statement_Database is

      function From_Fake_To_Ref_Stmt_Db is new Unchecked_Conversion
          (Ref_Statement_Database_Fake, Ref_Statement_Database);

   begin
      return (From_Fake_To_Ref_Stmt_Db (Node));

   end Convert_Fake;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Convert_To_Fake
   --
   --  <Description>
   --
   --    An actual reference to the statement database is converted into a
   --    fake reference to the statement database through use of the Ada
   --    UNCHECKED_CONVERSION generic subprogram.
   --==========================================================================
   function Convert_To_Fake
       (Node : in Ref_Statement_Database) return Ref_Statement_Database_Fake is

      function From_Ref_Stmt_Db_To_Fake is new Unchecked_Conversion
          (Ref_Statement_Database, Ref_Statement_Database_Fake);

   begin
      return (From_Ref_Stmt_Db_To_Fake (Node));

   end Convert_To_Fake;

   -- -------------------------------------------------------------------------

   -- ::::::::::::::::::::::::::::::::::::::::::::::::::::
   -- : Tests for Emptiness of different data structures :
   -- ::::::::::::::::::::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Empty
   --
   --  <Description>
   --
   --    Returns True if the statement database node is empty;
   --    otherwise returns False.
   --==========================================================================
   function Is_Empty (Node : in Ref_Statement_Database) return Boolean is
   begin
      return (Node = null);
   end Is_Empty;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Empty
   --
   --  <Description>
   --
   --    Returns True if the symbol table entry is empty;
   --    otherwise returns False.
   --==========================================================================
   function Is_Empty (Node : in Ref_Symbol_Table_Entry) return Boolean is
   begin
      return (Node = null);
   end Is_Empty;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Empty
   --
   --  <Description>
   --
   --    Returns True if the comment list entry is empty;
   --    otherwise returns False.
   --==========================================================================
   function Is_Empty (Node : in Ref_Comment) return Boolean is
   begin
      return (Node = null);
   end Is_Empty;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Empty
   --
   --  <Description>
   --
   --    Returns True if the data storage location is empty;
   --    otherwise returns False.
   --==========================================================================
   function Is_Empty (Node : in Ref_Data_Storage) return Boolean is
   begin
      return (Node = null) or else (Node.all = null);
   end Is_Empty;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Empty
   --
   --  <Description>
   --
   --    Returns True if the string pool element is empty;
   --    otherwise returns False.
   --==========================================================================
   function Is_Empty (Node : in Ref_String) return Boolean is
   begin
      return (Node = Create);
   end Is_Empty;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Empty
   --
   --  <Description>
   --
   --    Returns True if the Ace_Natural integer element is empty;
   --    otherwise returns False.
   --==========================================================================
   function Is_Empty (Node : in Ace_Natural) return Boolean is
   begin
      return (Node = 0);
   end Is_Empty;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Empty
   --
   --  <Description>
   --
   --    Returns True if the help  node is empty;
   --    otherwise returns False.
   --==========================================================================
   function Is_Empty (Node : in Ref_Help) return Boolean is
   begin
      return (Node = null);
   end Is_Empty;

   --==========================================================================
   -- return an empty value (usually a null)
   --==========================================================================
   function Empty return Ref_Statement_Database is
   begin
      return null;
   end Empty;

   function Empty return Ref_Symbol_Table_Entry is
   begin
      return null;
   end Empty;

   function Empty return Ref_Comment is
   begin
      return null;
   end Empty;

   function Empty return Ref_Data_Storage is
   begin
      return null;
   end Empty;

   function Empty return Ref_String is
   begin
      return Create;
   end Empty;

   function Empty return Ace_Natural is
   begin
      return Ace_Natural'FIRST;
   end Empty;

   function Empty return Ref_Help is
   begin
      return null;
   end Empty;

   --==========================================================================
   -- free node or entire substructure
   --==========================================================================
   -- [tbd] should implement reference counts for safety factor
   -- in releasing nodes; RC's will also help in doing quicker blind free's
   -- i.e., w/ local variables off the DSS -- oei

   -----
   -- stmt db record
   -----
   procedure Free (Item  : in out Ref_Statement_Database;
                   Depth : Depth_Kind := One_Level) is

      procedure Free (Item  : in out Ref_Statement_Record;
                      Depth : Depth_Kind := One_Level);
      procedure Free (Item  : in out Ref_Expression_Record;
                      Depth : Depth_Kind := One_Level);

      -----
      -- Statement_Record
      --
      procedure Free (Item  : in out Ref_Statement_Record;
                      Depth : Depth_Kind := One_Level) is
         procedure Unchecked_Free is new Unchecked_Deallocation
             (Statement_Record, Ref_Statement_Record);
      begin
         if (Item /= null) and
            ((Depth = Entire_Substructure or
            Depth = Entire_Substructure_And_Constants)) then

            -- Free (Item.Holder, Depth);

            case Item.Statement_Kind is

               when Exception_Alternative     =>
                  null;
               when Exception_Declaration     =>
                  null;
               when Object_Declaration   =>
                  Free (Item.Name_List, Depth);
                  Free (Item.Init_Expression, Depth);

               when Type_Declaration           |
                    Subtype_Declaration        |
                    Procedure_Spec_Declaration |
                    Procedure_Declaration      |
                    Function_Spec_Declaration  |
                    Function_Declaration       |
                    Package_Declaration        |
                    Package_Body_Declaration  =>

                  Free (Item.Symbol_Table_Entry, Depth);

               when Assign_Statement     =>
                  Free (Item.Lhs, Depth);
                  Free (Item.Rhs, Depth);

               when Loop_Statement  =>
                  Free (Item.Loop_Name, Depth);
                  Free (Item.Loop_Scheme, Depth);
                  Free (Item.Loop_Body, Depth);

               when For_Statement   =>
                  Free (Item.For_Variable, Depth);
                  Free (Item.For_Range, Depth);

               when If_Statement    =>
                  Free (Item.If_Condition, Depth);
                  Free (Item.True_Task, Depth);
                  Free (Item.False_Task, Depth);

               when Case_Statement  =>
                  Free (Item.Case_Expression, Depth);
                  Free (Item.Case_Alternatives_List, Depth);

               when Case_Alternative_Part     =>
                  Free (Item.Case_Alternative_Choice, Depth);
                  Free (Item.Case_Alternative_Statements, Depth);
                  Free (Item.Next_Case_Alternative, Depth);

               when Labeled_Statement    =>
                  Free (Item.Label_Entry, Depth);

               when Goto_Statement  =>
                  Free (Item.Goto_Label, Depth);

               when Exit_Statement  =>
                  Free (Item.Exit_Loop_Name, Depth);
                  Free (Item.Exit_Condition, Depth);

               when Return_Statement     =>
                  Free (Item.Return_Expression, Depth);
                  Free (Item.Subprogram_Reference, Depth);

               when Block_Statement =>
                  Free (Item.Block_Name, Depth);
                  Free (Item.Block_Declarations, Depth);
                  Free (Item.Block_Statements, Depth);
                  Free (Item.Block_Exceptions, Depth);

               when With_Statement  | Use_Statement     =>
                  Free (Item.Unit_Name_List, Depth);

               when Procedure_Call  =>
                  Free (Item.Procedure_Info, Depth);
                  Free (Item.Actual_Procedure_Parameters_List, Depth);

               when Pragma_Statement     =>
                  Free (Item.Pragma_Name, Depth);
                  Free (Item.Pragma_Arg, Depth);

               when Null_Statement  | Unknown =>
                  null;
            end case;
         end if;

         Unchecked_Free (Item);

      end Free;
      -----

      -----
      -- Expression_Record
      --
      procedure Free (Item  : in out Ref_Expression_Record;
                      Depth : Depth_Kind := One_Level) is
         procedure Unchecked_Free is new Unchecked_Deallocation
             (Expression_Record, Ref_Expression_Record);
      begin

         if (Item /= null) and
            ((Depth = Entire_Substructure or
            Depth = Entire_Substructure_And_Constants)) then

            case Item.Expression_Kind is
               when Character_Node  | String_Node  | Constant_Node     =>
                  if Depth = One_Level_And_Constants or
                     Depth = Entire_Substructure_And_Constants then
                     Free (Item.Node_Value, Depth);
                  end if;
               when Left_Parenthesis_Operator =>
                  if Item.Lp_Node_Kind /= Function_Call then
                     Free (Item.Node_Value, Depth);
                  end if;
               when others     =>
                  Free (Item.Node_Value, Depth);
            end case;

            -- Free (Item.Node_Type, Depth);
            -- Free (Item.Next_Expression, Depth);

            case Item.Expression_Kind is

               when Dot_Operator    =>
                  Free (Item.Dot_Object, Depth);
                  Free (Item.Dot_Component, Depth);
                  -- Free (Item.Expn_Terminal_Symbol, Depth);

               when Left_Parenthesis_Operator =>
                  Free (Item.Lp_Object, Depth);
                  Free (Item.Lp_Items_List, Depth);

               when List_Item  =>
                  Free (Item.Item_Expression, Depth);
                  Free (Item.Next_List_Item, Depth);

               when Aggregate_Component  =>
                  Free (Item.Aggregate_Choice, Depth);
                  Free (Item.Aggregate_Expression, Depth);

               when Range_Expression     =>
                  Free (Item.Range_Name, Depth);
                  Free (Item.Lower_Bound, Depth);
                  Free (Item.Upper_Bound, Depth);

               when Attribute_Tick  =>
                  Free (Item.Attribute_Prefix, Depth);
                  Free (Item.Attribute, Depth);

               when String_Node     =>
                  Free (Item.String_Symbol, Depth);

               when Constant_Node   =>
                  Free (Item.Constant_Symbol, Depth);

               when Terminal_Node   =>
                  Free (Item.Symbol_Name, Depth);
                  -- Free (Item.Terminal_Symbol, Depth);

               when Array_Reference                 |
                    Function_Call                   |
                    Type_Conversion                 |
                    Record_Reference                |
                    Dot_Notation_Variable_Reference |
                    Character_Node                  |
                    Unknown    =>
                  null;

            end case;
         end if;

         Unchecked_Free (Item);

      end Free;

      procedure Unchecked_Free is new Unchecked_Deallocation
          (Statement_Database_Record, Ref_Statement_Database);
   begin
      -----
      -- check the number of references to this node
      -- if it is an expression node
      --
      if (Item /= null) and then (Item.Node_Kind = Expression_Node_Type)
          and then (Item.Expression_Node.Reference_Count > Ace_Positive'FIRST)
          then
         Item.Expression_Node.Reference_Count :=
            Item.Expression_Node.Reference_Count - 1;
         return ;
      end if;
      -----

      if (Item /= null) and
         ((Depth = Entire_Substructure or
         Depth = Entire_Substructure_And_Constants)) then

         -- Free (Item.Parent_Node, Depth);
         Free (Item.Comment_Holder, Depth);

         case Item.Node_Kind is
            when Statement_Node_Type     =>
               Free (Item.Statement_Node, Depth);
            when Expression_Node_Type    =>
               Free (Item.Expression_Node, Depth);
            when Unknown  =>
               null;
         end case;

      end if;

      Unchecked_Free (Item);

   end Free;
   -----

   -----
   -- symbol table
   -----
   procedure Free (Item  : in out Ref_Symbol_Table_Entry;
                   Depth : Depth_Kind := One_Level) is

      Temp : Ref_Statement_Database;

      -----
      -- List_Of_Subprograms
      --
      procedure Free (Item  : in out Ref_List_Of_Subprograms;
                      Depth : Depth_Kind := One_Level) is
         procedure Unchecked_Free is new Unchecked_Deallocation
             (List_Of_Subprograms, Ref_List_Of_Subprograms);
      begin
         if (Item /= null) and
            ((Depth = Entire_Substructure or
            Depth = Entire_Substructure_And_Constants)) then
            Free (Item.Subprogram_Entry, Depth);
            Free (Item.Next_Subprogram_Entry, Depth);
         end if;
         Unchecked_Free (Item);
      end Free;
      -----

      -----
      -- Type_Definition_Node
      --
      procedure Free (Item  : in out Ref_Type_Definition_Node;
                      Depth : Depth_Kind := One_Level) is
         procedure Unchecked_Free is new Unchecked_Deallocation
             (Type_Definition_Node, Ref_Type_Definition_Node);
      begin
         if (Item /= null) and
            ((Depth = Entire_Substructure or
            Depth = Entire_Substructure_And_Constants)) then
            case Item.Type_Kind is
               when Integer_Type    | Real_Type    | Access_Type  =>
                  null;
               when Enumeration_Type     =>
                  Free (Item.First_Literal, Depth);
                  -- Free (Item.Last_Literal, Depth);
               when Record_Type     =>
                  Free (Item.Record_Component, Depth);
               when Array_Type =>
                  -- Free (Item.Array_Index_Type, Depth);
                  -- Free (Item.Array_Element_Type, Depth);
                  null;
            end case;
         end if;
         Unchecked_Free (Item);
      end Free;
      -----

      -----
      -- Constraint_Node
      --
      procedure Free (Item  : in out Ref_Constraint_Node;
                      Depth : Depth_Kind := One_Level) is
         procedure Unchecked_Free is new Unchecked_Deallocation
             (Constraint_Node, Ref_Constraint_Node);
      begin
         if (Item /= null) and
            ((Depth = Entire_Substructure or
            Depth = Entire_Substructure_And_Constants)) then
            case Item.Constraint_Kind is
               when Floating_Point_Constraint |
                    Fixed_Point_Constraint    |
                    Discriminant_Constraint   =>
                  null;
               when Range_Constraint     | Index_Constraint  =>
                  Temp := Convert_Fake (Item.Lower_Bound);
                  Free (Temp, Depth);
                  Temp := Convert_Fake (Item.Upper_Bound);
                  Free (Temp, Depth);
                  Free (Item.Next_Constraint, Depth);
            end case;
         end if;
         Unchecked_Free (Item);
      end Free;
      -----

      procedure Unchecked_Free is new Unchecked_Deallocation
          (Symbol_Table_Entry, Ref_Symbol_Table_Entry);

   begin
      if (Item /= null) and
         ((Depth = Entire_Substructure or
         Depth = Entire_Substructure_And_Constants)) then

         -- Free (Item.Next_Entry, Depth);
         Free (Item.Symbol_Name, Depth);
         -- Free (Item.Enclosing_Package, Depth);

         case Item.Entry_Kind is

            when Exception_Entry    =>
               null;

            when Package_Entry =>
               -- Temp := Get_Context_Clause (Item);
               -- Free (Temp, Depth);
               -- Temp := Get_Context_Clause_Body (Item);
               -- Free (Temp, Depth);
               Temp := Get_Package_Spec_Decls (Item);
               Free (Temp, Depth);
               Temp := Get_Package_Body_Decls (Item);
               Free (Temp, Depth);
               Temp := Get_Head_Of_Statements_List (Item);
               Free (Temp, Depth);
               Free (Item.Pkg_Help, Depth);

               Free (Item.Pkg_Derived_Subprogram_List, Depth);

            when Procedure_Entry    | Function_Entry    =>
               Free (Item.Head_Param_List, Depth);
               -- Free (Item.Tail_Param_List, Depth);
               -- Free (Item.Return_Type, Depth);
               -- Temp := Get_Context_Clause (Item);
               -- Free (Temp, Depth);
               Temp := Get_Head_Of_Declarations_List (Item);
               Free (Temp, Depth);
               Temp := Get_Head_Of_Statements_List (Item);
               Free (Temp, Depth);
               -- Free (Item.Derived_Subprogram_Parent, Depth);
               Free (Item.Subprg_Help, Depth);

            when Object_Entry  =>
               -- Free (Item.Object_Type, Depth);
               Free (Item.Object_Constraint, Depth);

            when Formal_Param_Entry =>
               -- Free (Item.Param_Type, Depth);
               -- Temp := Get_Initialization_Code (Item);
               -- Free (Temp, Depth);
               Free (Item.Next_Param, Depth);

            when Type_Entry    =>
               Free (Item.Type_Definition, Depth);
               Free (Item.Type_Derived_Subprogram_List, Depth);
               Free (Item.Type_Help, Depth);

            when Subtype_Entry | Derived_Type_Entry     =>
               -- Free (Item.Base_Type, Depth);
               Free (Item.Constraint_Info, Depth);
               Free (Item.Dtype_Derived_Subprogram_List, Depth);
               Free (Item.Other_Type_Help, Depth);

            when Constraint_Entry   =>
               Free (Item.Constraint_Ref, Depth);

            when Enumeration_Literal_Entry    =>
               -- Free (Item.Enumeration_Type, Depth);
               Free (Item.Succ_Enum_Literal, Depth);
               -- Free (Item.Pred_Enum_Literal, Depth);

            when Label_Entry   =>
               Temp := Get_Head_Of_Statements_List (Item);
               Free (Temp, Depth);

            when Record_Component_Entry  =>
               -- Free (Item.Component_Type, Depth);
               -- Temp := Get_Initialization_Code (Item);
               -- Free (Temp, Depth);
               Free (Item.Next_Record_Component, Depth);

            when Simple_Name_Entry  =>
               Temp := Get_Simple_Name_Construct (Item);
               Free (Temp, Depth);

            when Unknown  =>
               null;

         end case;
      end if;

      Unchecked_Free (Item);

   end Free;
   -----

   -----
   -- storage structure
   procedure Free (Item  : in out Ref_Data_Storage;
                   Depth : Depth_Kind := Entire_Substructure_And_Constants) is

      Assertion_Error : exception;

      procedure Unchecked_Free is new Unchecked_Deallocation
          (Data_Storage, Indirect_Data_Storage);
      procedure Unchecked_Free is new Unchecked_Deallocation
          (Indirect_Data_Storage, Ref_Data_Storage);

   begin
      if (Item /= null) and
         ((Depth = Entire_Substructure or
         Depth = Entire_Substructure_And_Constants)) then
         -- free next array item
         Free (Item.all.Next_Item, Depth);
         -- free next record component item
         Free (Item.all.Next_Component, Depth);
         if Item.all.Data_Type = String_Data then
            Free (Item.all.String_Item, Depth);
         end if;
         if Item.all = null then
            raise Assertion_Error;
         end if;
         Unchecked_Free (Item.all);
      end if;
      --Unchecked_Free (Item);
   end Free;
   -----

   -----
   -- help comments
   --
   procedure Free (Item  : in out Ref_Help;
                   Depth : Depth_Kind := One_Level) is
      procedure Unchecked_Free is new Unchecked_Deallocation
          (Help_Pkg.Help, Ref_Help);
   begin
      if (Item /= null) and
         ((Depth = Entire_Substructure or
         Depth = Entire_Substructure_And_Constants)) then
         Free (Item.Next, Depth);
      end if;
      Unchecked_Free (Item);
   end Free;
   -----

   -----
   -- strings
   --
   procedure Free (Item  : in out Ref_String;
                   Depth : Depth_Kind := One_Level) is
   begin
      -- do nothing for now, until reference counts are implmented
      -- string could be used somewhere else!
      -- Destroytree (Item);
      null;
   end Free;
   -----

   -----
   -- comments
   --
   procedure Free (Item  : in out Ref_Comment;
                   Depth : Depth_Kind := One_Level) is
      procedure Unchecked_Free is new Unchecked_Deallocation
          (Comments, Ref_Comment);
   begin
      if (Item /= null) and
         ((Depth = Entire_Substructure or
         Depth = Entire_Substructure_And_Constants)) then
         Free (Item.Next, Depth);
      end if;
      Unchecked_Free (Item);
   end Free;
   -----

   -- -------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Check_Stmt_Node
   --
   --  <Description>
   --
   --    Returns the kind of statement that is represented in the
   --    statement database node
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Check_Stmt_Node
       (Node : in Ref_Statement_Database) return Statement_Enumeration is

   begin
      return Node.Statement_Node.Statement_Kind;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Check_Stmt_Node");
         raise Program_Error;

   end Check_Stmt_Node;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Check_Expr_Node
   --
   --  <Description>
   --
   --    Returns the kind of expression that is represented in the
   --    statement database node
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Check_Expr_Node
       (Node : in Ref_Statement_Database) return Expression_Enumeration is

   begin
      if Node.Expression_Node.Expression_Kind = Dot_Operator and then
         Node.Expression_Node.Dot_Node_Kind /= Unknown then
         case Node.Expression_Node.Dot_Node_Kind is
            when Record_Reference   =>
               return Record_Reference;
            when Dot_Notation_Variable_Reference   =>
               return Dot_Notation_Variable_Reference;
            when Unknown  =>
               null;
         end case;

      elsif Node.Expression_Node.Expression_Kind = Left_Parenthesis_Operator
             and then Node.Expression_Node.Lp_Node_Kind /= Unknown then
         case Node.Expression_Node.Lp_Node_Kind is
            when Array_Reference    =>
               return Array_Reference;
            when Function_Call =>
               return Function_Call;
            when Type_Conversion    =>
               return Type_Conversion;
            when Unknown  =>
               null;
         end case;

      else
         return Node.Expression_Node.Expression_Kind;

      end if;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Check_Expr_Node");
         raise Program_Error;

   end Check_Expr_Node;

   -- -------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Duplicate_Node
   --
   --  <Description>
   --
   --    A duplicate copy of the statement database node is created.
   --    This routine currently only supports the duplication of statement
   --    database nodes that are expresssion nodes of kind terminal_node.
   --==========================================================================
   function Duplicate_Node
       (Node : in Ref_Statement_Database) return Ref_Statement_Database is
   begin

      case Check_Expr_Node (Node) is

         when Terminal_Node    =>
            declare
               New_Node : Ref_Statement_Database;
            begin
               New_Node := Create_Node (Terminal_Node);
               Set_Symbolic_Name (New_Node, Get_Symbolic_Name (Node));
               Set_Symbol_Table_Entry (New_Node, Get_Symbol_Table_Entry (Node));
               return New_Node;
            end;

         when others =>
            -- !! NOT YET REQUIRED - NOT YET IMPLEMENTED !!
            Text_Io.Put_Line
                ("*** Function - Duplicate_Node - Not Fully Implemented ***");
            return null;
      end case;

   end Duplicate_Node;

   -- -------------------------------------------------------------------------

   function Is_Declaration (Node : in Ref_Statement_Database) return Boolean is
   begin
      if Node.Node_Kind = Statement_Node_Type then
         case Check_Stmt_Node (Node) is
            when Type_Declaration           |
            -- Exception_Declaration      | -- not fully implemented yet
                 Subtype_Declaration        |
                 Procedure_Spec_Declaration |
                 Procedure_Declaration      |
                 Function_Spec_Declaration  |
                 Function_Declaration       |
                 Package_Declaration        |
                 Package_Body_Declaration     =>
               return True;
            when others   =>
               return False;
         end case;
      else
         return False;
      end if;
   end Is_Declaration;


   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Next_Matching_Name
   --
   --  <Description>
   --
   --    Find_Next_Matching_Name traverses the statement database for the
   --    next occurrence of the name, starting at the current statement
   --    (curr_stmt).  This routine will look at all names, even those not
   --    directly visible.  If a null form string is given the next occurrence
   --    is returned via the Found_Node parameter; otherwise the next node
   --    matching the form string is returned. When invoked with the type
   --    of the name (Entry_Type), only the name defined by the specified type
   --    will be found.
   --==========================================================================
   procedure Find_Next_Matching_Name
       (Context    : in Context_Definition;
        Curr_Stmt  : in out Ref_Statement_Database;
        Name       : in Ref_String;
        Form_Str   : in String;
        Found_Node : out Ref_Symbol_Table_Entry;
        Entry_Type : in Type_Entry_Enumeration := Unknown) is
      Curr_Decl_Node : Ref_Statement_Database := Curr_Stmt;

   begin
      while not Is_Empty (Curr_Decl_Node) loop
         if Is_Declaration (Curr_Decl_Node) then
            Found_Node := Get_Symbol_Table_Entry (Curr_Decl_Node);
            Curr_Stmt := Get_Next_Statement (Curr_Decl_Node);
            return ;
         elsif (Check_Stmt_Node (Curr_Decl_Node) = Object_Declaration) then
            declare
               Curr_Name_Node : Ref_Statement_Database :=
                  Get_Name_List (Curr_Decl_Node);
            begin
               while not Is_Empty (Curr_Name_Node) loop
                  if Are_Symbolic_Names_Equal
                         (Get_Symbolic_Name
                              (Get_Symbol_Table_Entry
                                   (Get_Expression (Curr_Name_Node))), Name)
                      and
                     (Entry_Type = Unknown or else
                     Entry_Type =
                     Get_Entry_Kind
                         (Get_Symbol_Table_Entry
                              (Get_Expression (Curr_Name_Node)))) then
                     Found_Node :=
                        Get_Symbol_Table_Entry
                            (Get_Expression (Curr_Name_Node));
                     Curr_Stmt := Get_Next_Statement (Curr_Decl_Node);
                     return ;
                  else
                     Curr_Name_Node := Get_Next_List_Item (Curr_Name_Node);
                  end if;
               end loop;
            end;
         end if;
         Curr_Decl_Node := Get_Next_Statement (Curr_Decl_Node);
      end loop;

      if not (Is_Empty (Curr_Decl_Node)) then
         Found_Node := Get_Symbol_Table_Entry (Curr_Decl_Node);
      else
         Found_Node := null;
      end if;
   exception
      when Constraint_Error    =>
         Show_Error (3018, Fatal, Form_Str);


   end Find_Next_Matching_Name;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Matching_Name
   --
   --  <Description>
   --
   --    This version of Find_Matching_name searches for a Symbol_Table_entry
   --     for a given expanded name, that is, within some enclosing package
   --   which might not be currently on the Symbol_Table. This routine will
   --   look at any name in the enclosing package, including a hidden or
   --   deleted one by traversing the statement database. If a form_string is
   --   provided, the first entry matching it will be returned, otherwise the
   --     first occurance will be returned.
   --   The symbol table node will be returned via the "found_node" parameter.
   --   The last statement checked will be returned via the Last_Stmt parameter.
   --==========================================================================
   procedure Find_Matching_Name (Context       : in Context_Definition;
                                 Enclosing_Pkg : in Ref_Symbol_Table_Entry;
                                 Name          : in Ref_String;
                                 Form_Str      : in String;
                                 Found_Node    : out Ref_Symbol_Table_Entry;
                                 Last_Stmt     : out Ref_Statement_Database;
                                 Entry_Type    : in Type_Entry_Enumeration :=
      Unknown) is
      Curr_Decl_Node : Ref_Statement_Database;
   begin
      if Get_Entry_Kind (Enclosing_Pkg) = Package_Entry then
         Curr_Decl_Node := Get_Package_Spec_Decls (Enclosing_Pkg);
         Find_Next_Matching_Name (Context, Curr_Decl_Node, Name, Form_Str,
                                  Found_Node, Entry_Type);
      else
         Show_Error (4206, Fatal,
                     Convert_To_String (Get_Symbolic_Name (Enclosing_Pkg)));
      end if;

   end Find_Matching_Name;

   function Anonymous_Name return String is
     Str : constant String := Natural'IMAGE (Anonymous_Counter);
   begin
     Anonymous_Counter := Anonymous_Counter + 1;
     return "<Anonymous_Type_" & Str (Str'First+1 .. Str'Last) & ">";
   end;

end Misc_Support_Routines;
