--|$Log:	pre_parser.a,v $
--|Revision 8.0  91/01/17  10:45:56  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Pre_Parser
--
---------------------------------------------------------------
--
--
-- Abstract
--
--   The Pre_parser package controls the routines needed for
--   parsing dynamic Ada constructs.
--   Routines can call Pre_Parse passing a Ref_Statement_Database object
--   and a abstract syntax tree of the input entered (through routines
--   in the body of this package that are not exported) will be passed
--   out of the Pre_Parse routine.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------
with Compilation_Unit_Parser;
with Seq_Of_Stmts_Parser;
with Decl_Part_Parser;

with User_Pkg;

with Lexacts;

with Misc;

with Context_Control;

with Ace;

with Directory_Objects;

with System_Dependent_Routines;
use System_Dependent_Routines;

with Error_Handler;

with Declarative_Region_Support_Routines;

with Lex_Pkg;
use Lex_Pkg;

with Text_Io;
use Text_Io;

with Parserdefs;
use Parserdefs;

with Yyval_Error;
use Yyval_Error;

with Ace_Universal_Types;
use Ace_Universal_Types;

pragma ELABORATE (System_Dependent_Routines);
pragma ELABORATE (User_Pkg);
pragma ELABORATE (Directory_Objects);

package body Pre_Parser is
   use Ascii;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Parse_Compilation_Unit
   --
   --  <Description>
   --
   --    This routine calls the compilation unit parser.
   --    If an error occurs then it will cleanup, and raise Parse_Error.
   --
   --  <Called By>
   --
   --    Pre_Parse
   --
   --  <Exceptions Raised>
   --
   --    Parse_Error
   --
   --  <Exceptions Handled>
   --
   --    Parse_Error
   --
   --==========================================================================
   procedure Parse_Compilation_Unit
       (Tree : out Statement_Database.Ref_Statement_Database) is
   begin
      Compilation_Unit_Parser.Yyparse;
      Tree := Lval;
   exception
      when Parser_Error | Compilation_Unit_Parser.Syntax_Error  =>
         Misc.Reset_Input;
         Tree := null;
         raise Parser_Error;
   end Parse_Compilation_Unit;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Parse_Types
   --
   --  <Description>
   --
   --    This routine calls the types parser.  If an error occurs then
   --    it will cleanup, and raise Parse_Error.
   --
   --  <Called By>
   --
   --    Pre_Parse
   --
   --  <Exceptions Raised>
   --
   --    Parse_Error
   --
   --  <Exceptions Handled>
   --
   --    Parse_Error
   --
   --==========================================================================
   procedure Parse_Types (Tree : out Statement_Database.Ref_Statement_Database)
                         is
   begin
      Decl_Part_Parser.Yyparse;
      Tree := Lval;
   exception
      when Parser_Error | Decl_Part_Parser.Syntax_Error  =>
         Misc.Reset_Input;
         Tree := null;
         raise Parser_Error;
   end Parse_Types;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Parse_Statements
   --
   --  <Description>
   --
   --    This routine calls the statement parser.  If an error occurs then
   --    it will cleanup, and raise Parse_Error.
   --
   --  <Called By>
   --
   --    Pre_Parse
   --
   --  <Exceptions Raised>
   --
   --    Parse_Error
   --
   --  <Exceptions Handled>
   --
   --    Parse_Error
   --
   --==========================================================================
   procedure Parse_Statements
       (Tree : out Statement_Database.Ref_Statement_Database) is
   begin
      Seq_Of_Stmts_Parser.Yyparse;
      Tree := Lval;
   exception
      when Parser_Error | Seq_Of_Stmts_Parser.Syntax_Error =>
         Misc.Reset_Input;
         Tree := null;
         raise Parser_Error;
   end Parse_Statements;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Unput_Token
   --
   --  <Description>
   --
   --    This routine will put back on the input queue a string of
   --    characters, so the next time input is needed for the interpreter
   --    the "unputted" string will be read.
   --
   --==========================================================================
   procedure Unput_Token (Str : String) is
   begin
      User_Pkg.Unput (' ');
      for I in reverse 1 .. Str'LENGTH loop
         User_Pkg.Unput (Str (I));
      end loop;
   end Unput_Token;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Pre_Parse
   --
   --  <Description>
   --
   --   Routines can call Pre_Parse passing a Ref_Statement_Database object
   --   and a abstract syntax tree of the input entered (through routines
   --   in the body of this package that are not exported) will be passed
   --   out of the Pre_Parse routine.
   --
   --
   --  <Exceptions Handled>
   --
   --   Parser_Error, but Parser_Error is still propagated to a higher block.
   --
   --==========================================================================

   procedure Pre_Parse (Tree : out Statement_Database.Ref_Statement_Database)
                       is
      type Token_Index is  range 1 .. Lmax;

      Token            : Integer;

      Save_Token       : String (1 .. Lmax);
      Save_Token_Lng   : Token_Index;

      Save_Token_2     : String (1 .. Lmax);
      Save_Token_Lng_2 : Token_Index;
   begin

      -- All error checks on the previous construct have
      -- finished, so we can assume that the current head is a
      -- safe place to recover to when there is an error.
      Context_Control.
        Set_Saved_Main_Level_Symbol_Table_Head (Ace.Main_Thread_Context,
           Context_Control.Get_Top_Of_Symtbl (Ace.Main_Thread_Context));

      -- Output a prompt for the user.
      -- This should really depend on a flag since we do not want to
      -- prompt a file (in the case of redirection).
      Misc.Prompt (Misc.Main_Prompt);

      -- get a token from the input stream.
      Token := Yylex;

      -- This case statement is used to determine what parser to
      -- use.  In most cases only one token in needed to determine the
      -- correct parser. For example with the token PROCEDURE, the
      -- parser for compilation units is used.  When the token is an
      -- IDENTIFIER more tokens are needed to determine the correct
      -- parser.  The construct could be an object delaration, an
      -- assignment statement, or a loop label.  In the worst case, three
      -- tokens are needed to select the correct parser.
      --
      -- After the correct parser is selected, the tokens that
      -- were read in are put back in the input queue.

      case Token is -- One;

         when Procedure_Tk |
              Package_Tk   |
              Function_Tk  |
              Generic_Tk   |
              With_Tk     =>
            Unput_Token (Yytext (1 .. Yyleng));
            -- It's a compilation Unit.
            Parse_Compilation_Unit (Tree);

         when For_Tk     |
              Case_Tk    |
              While_Tk   |
              If_Tk      |
              Loop_Tk    |
              Declare_Tk |
              Begin_Tk   |
              Exit_Tk    |
              Return_Tk  |
              Accept_Tk  |
              Goto_Tk    |
              Delay_Tk   |
              Select_Tk  |
              Null_Tk    |
              Abort_Tk   |
              L_Lbl_Tk   |
              Raise_Tk    =>
            Unput_Token (Yytext (1 .. Yyleng));
            -- It's a statement
            Parse_Statements (Tree);

         when Type_Tk     | Subtype_Tk   | Pragma_Tk    | Task_Tk | Use_Tk  =>
            Unput_Token (Yytext (1 .. Yyleng));
            -- It's a type declaration
            Parse_Types (Tree);

         when Identifier  =>
            Save_Token := Yytext;
            Save_Token_Lng := Token_Index (Yyleng);
            Token      := Yylex;

            -- We are now two tokens deep.
            case Token is -- Two;

               when Character'POS (',')  =>
                  Unput_Token (Yytext (1 .. Yyleng));
                  Unput_Token (Save_Token (1 .. Integer (Save_Token_Lng)));
                  -- It's a type declaration
                  Parse_Types (Tree);

               when Character'POS (':')  =>
                  Save_Token_2 := Yytext;
                  Save_Token_Lng_2 := Token_Index (Yyleng);
                  Lexacts.Push_Eol;
                  Token        := Yylex;

                  -- We are now three tokens deep.
                  case Token is -- Three;

                     when For_Tk    | While_Tk     | Loop_Tk | Begin_Tk     =>
                        Unput_Token (Yytext (1 .. Yyleng));
                        Unput_Token (Save_Token_2 (1 ..
                                                   Integer (Save_Token_Lng_2)));
                        Unput_Token
                            (Save_Token (1 .. Integer (Save_Token_Lng)));
                        Lexacts.Pop_Eol;
                        -- It's a statement
                        Parse_Statements (Tree);

                     when others    =>
                        Unput_Token (Yytext (1 .. Yyleng));
                        Unput_Token (Save_Token_2 (1 ..
                                                   Integer (Save_Token_Lng_2)));
                        Unput_Token
                            (Save_Token (1 .. Integer (Save_Token_Lng)));
                        Lexacts.Pop_Eol;
                        -- It's a type declaration
                        Parse_Types (Tree);

                  end case; -- Three;

               when others     =>
                  Unput_Token (Yytext (1 .. Yyleng));
                  Unput_Token (Save_Token (1 .. Integer (Save_Token_Lng)));
                  Parse_Statements (Tree);

            end case; -- Two;

         when 0 => -- EOF or LF
            null;

         when others =>
            Error_Handler.Show_Error (3015, Error_Handler.Warning,
                                      "`" & Yytext (1 .. Yyleng) & "'");
            Misc.Reset_Input;

      end case; -- One;

      -- reset declarative region to where it was before call to Pre_Parse
      -- Declarative_Region_Support_Routines.Reset_DR_To_Previous_Level;

   end Pre_Parse;

begin
   declare
      Start_Up_File : constant String := """" & Command_Line_Argument & """";
      Start_Default : constant String := """" &
         Directory_Objects.Name(Home_Directory) & "/startup.ace""";
   begin
      -- If the command line argument is present then use the the argument
      -- as the name to interpret upon ACE startup. Otherwise, use the default
      if Start_Up_File'LENGTH > 2 then
         Unput_Token ("eval(" & Start_Up_File & ");" & Ascii.Lf);
      else
         Unput_Token ("eval(" & Start_Default & ");" & Ascii.Lf);
      end if;
   end;
end Pre_Parser;
