--|$Log:	semant.a,v $
--|Revision 8.0  91/01/17  10:45:58  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Semantic_Analyzer
--
---------------------------------------------------------------
--
-- Abstract -
--
--   This is the semantic analyzer for the ACE ACLI.
--   Checks the semantics of Ada statements, declarations and expressions.
--   Builds symbol table entries for declarations.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Statement_Database;
use Statement_Database;

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Node_Get_Routines;
use Node_Get_Routines;

with Node_Set_Routines;
use Node_Set_Routines;

with Node_Create_Routines;
use Node_Create_Routines;

with Symtbl_Entry_Get_Routines;
use Symtbl_Entry_Get_Routines;

with Symbol_Table_Routines;
use Symbol_Table_Routines;

with Misc_Support_Routines;
use Misc_Support_Routines;

with Error_Handler;
use Error_Handler;

with Literals;
use Literals;

with Ace;

pragma ELABORATE (Node_Create_Routines);
pragma ELABORATE (Node_Set_Routines);

package body Semantic_Analyzer is

   Type_Symtbl_Entry_Holder_Node           : constant Ref_Statement_Database :=
      Create_Node (Constant_Node);
   -----
   Universal_Type_Symtbl_Entry_Holder_Node : constant Ref_Statement_Database
    := Create_Node (Constant_Node);
   -----

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Same_Type
   --
   --  <Description>
   --
   --    return true iff the type of both symbol table entries are
   --    the same.
   --==========================================================================
   function Is_Same_Type
       (Entry_1, Entry_2 : in Ref_Symbol_Table_Entry) return Boolean is
   begin
      if ((not Is_Empty (Entry_1)) and (not Is_Empty (Entry_2))) then
         return (Get_Base_Type (Entry_1) = Get_Base_Type (Entry_2));
      else
         return False;
      end if;
   end Is_Same_Type;

   --
   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Same_Type
   --
   --  <Description>
   --
   --    return true iff the type of both symbol table entries are
   --    the same.
   --
   --==========================================================================
   function Is_Same_Type
       (Node_1, Node_2 : in Ref_Statement_Database) return Boolean is
      Entry_1 : Ref_Symbol_Table_Entry := Get_Type (Node_1);
      Entry_2 : Ref_Symbol_Table_Entry := Get_Type (Node_2);
   begin
      if ((not Is_Empty (Entry_1)) and (not Is_Empty (Entry_2))) then
         if Is_Type_Universal (Node_1) then
            return Is_Same_Type (Entry_1, Get_Root_Type (Entry_2));
         elsif Is_Type_Universal (Node_2) then
            return Is_Same_Type (Get_Root_Type (Entry_1), Entry_2);
         else
            return Is_Same_Type (Entry_1, Entry_2);
         end if;
      else
         return False;
      end if;
   end Is_Same_Type;

   --
   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Same_Type
   --
   --  <Description>
   --
   --    return true iff the type of both symbol table entries are
   --    the same.
   --
   --==========================================================================
   --
   function Is_Same_Type
       (Node              : in Ref_Statement_Database;
        Type_Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is
   begin
      Set_Type (Type_Symtbl_Entry_Holder_Node, Type_Symtbl_Entry);
      return Is_Same_Type (Node, Type_Symtbl_Entry_Holder_Node);
   end Is_Same_Type;

   --
   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Mode_In
   --
   --  <Description>
   --
   --    return true iff the mode of the object is "In"
   --
   --==========================================================================
   function Is_Mode_In (Node : in Ref_Statement_Database) return Boolean is
   separate;

   --
   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Mode_Out
   --
   --  <Description>
   --
   --    return true iff the mode of the object is "Out"
   --
   --==========================================================================
   function Is_Mode_Out (Node : in Ref_Statement_Database) return Boolean is
   separate;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Check_Subprogram_Call
   --
   --  <Description>
   --
   --    This routine checks the semantics of subprogram calls.
   --
   --==========================================================================
   procedure Check_Subprogram_Call
       (Context : in out Context_Control.Context_Definition;
        Node    : in Ref_Statement_Database) is separate;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Must_Be_Boolean_Type
   --
   --  <Description>
   --
   --    The given expression must be of some "boolean" type.
   --
   --==========================================================================
   procedure Must_Be_Boolean_Type (Node : in Ref_Statement_Database) is
   separate;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Check_Expression
   --
   --  <Description>
   --
   --   Checks the given expression node structure against
   --   correct Ada semantics.
   --
   --==========================================================================
   procedure Check_Expression
       (Context    : in out Context_Control.Context_Definition;
        Node       : in Ref_Statement_Database;
        Entry_Kind : Type_Entry_Enumeration := Unknown) is separate;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Check_Statement
   --
   --  <Description>
   --
   --   Checks the given Ada statement node structure against correct
   --   Ada semantics.
   --
   --==========================================================================
   procedure Check_Statement
       (Context : in out Context_Control.Context_Definition;
        Node    : in Ref_Statement_Database) is separate;


begin

   Set_As_Universal_Type (Universal_Type_Symtbl_Entry_Holder_Node);

end Semantic_Analyzer;
