--|$Log:	smgmt.a,v $
--|Revision 8.0  91/01/17  10:46:21  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Storage_Management_Routines
--
------------------------------------------------------------
--
--
-- Abstract
--
-- The Storage_Management_Routines package provides a set of routines
-- to the statement evaluator of the ACLI for storage management.
-- These routines handle the creation, deletion, and manipulation of
-- storage associated with data objects and the invocation of subprograms.
--
-- Update Log - :
--
-- End of Header
-------------------------------------------------------------------------

with Storage_Management;
use Storage_Management;

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Symtbl_Entry_Get_Routines;
use Symtbl_Entry_Get_Routines;

with Symtbl_Entry_Set_Routines;
use Symtbl_Entry_Set_Routines;

with Symbol_Table_Routines;
use Symbol_Table_Routines;

with Misc_Support_Routines;
use Misc_Support_Routines;

with Error_Handler;
use Error_Handler;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Text_Io;
use Text_Io;

with Unchecked_Deallocation;

package body Storage_Management_Routines is

   --=======================================================================
   --  <Subprogram Name>
   --
   --    Allocate_Storage
   --
   --  <Description>
   --
   --    Allocate_Storage creates a data storage location based upon the type
   -- of symtbl entry item.
   --
   --=======================================================================

   procedure Allocate_Storage
       (Context      : in out Context_Control.Context_Definition;
        Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      Location : Ace_Natural;

   begin
      case Get_Entry_Kind (Symtbl_Entry) is

         when Function_Entry   | Object_Entry | Formal_Param_Entry     =>
            declare
               Type_Entry_Found : Ref_Symbol_Table_Entry;
            begin

               -- find the "root" type of the item that needs storage allocated
               case Get_Entry_Kind (Get_Type (Symtbl_Entry)) is

                  when Type_Entry   =>
                     Type_Entry_Found := Get_Type (Symtbl_Entry);
                  when Subtype_Entry     | Derived_Type_Entry     =>
                     Type_Entry_Found :=
                        Get_Root_Type (Get_Type (Symtbl_Entry));
                  when Function_Entry            |
                       Object_Entry              |
                       Formal_Param_Entry        |
                       Procedure_Entry           |
                       Constraint_Entry          |
                       Enumeration_Literal_Entry |
                       Package_Entry             |
                       Label_Entry               |
                       Record_Component_Entry    |
                       Simple_Name_Entry         |
                       Exception_Entry           |
                       Unknown =>
                     -- can't be !!
                     Show_Error (0, Internal,
                                 "allocate_storage: enumeration not expected");
                     return ;
               end case;

               -- note:
               -- entry's level and offset are set during semantic analysis
               if (Is_Type_An_Enumeration_Type (Type_Entry_Found) or else
                  Is_Type_An_Integer_Type (Type_Entry_Found)) then
                  Location :=
                     Get_Display_For_Level (Context, Get_Level (Symtbl_Entry))
                      + Get_Value_Offset (Symtbl_Entry);
                  Context_Control.Set_Data_Storage_Stack
                      (Context, Location, Create_Storage_For_Integer);
                  if (Location >
                     Context_Control.Get_Last_Storage_Location_Pointer
                         (Context)) then
                     Context_Control.Set_Last_Storage_Location_Pointer
                         (Context, Location);
                  end if;
               elsif (Is_Type_A_Record_Type (Type_Entry_Found)) then
                  null;
                  -- "alloc storage" for each record component;
                  Show_Error (0, Internal, "[tbd]: alloc storage for records");
               elsif (Is_Type_An_Array_Type (Type_Entry_Found)) then
                  Location :=
                     Get_Display_For_Level (Context, Get_Level (Symtbl_Entry))
                      + Get_Value_Offset (Symtbl_Entry);
                  --Context_Control.Set_Data_Storage_Stack
                  --    (Context, Location, Create_Storage_For_Array(L,H));
                  if (Location >
                     Context_Control.Get_Last_Storage_Location_Pointer
                         (Context)) then
                     Context_Control.Set_Last_Storage_Location_Pointer
                         (Context, Location);
                  end if;
               else
                  -- either not implementing or error
                  -- -- array
                  -- -- real
                  -- *** let's assume it's a string for now! [tbd] -- oei
                  Location :=
                     Get_Display_For_Level (Context, Get_Level (Symtbl_Entry))
                      + Get_Value_Offset (Symtbl_Entry);
                  Context_Control.Set_Data_Storage_Stack
                      (Context, Location, Create_Storage_For_String);
                  if (Location >
                     Context_Control.Get_Last_Storage_Location_Pointer
                         (Context)) then
                     Context_Control.Set_Last_Storage_Location_Pointer
                         (Context, Location);
                  end if;
               end if;

               -- what about these??
               ---- character
               ---- string
               ---- task???

            end;

         when Constraint_Entry =>
            case Symtbl_Entry.Constraint_Ref.Constraint_Kind is

               when Range_Constraint     | Index_Constraint  =>
                  -- evaluate lb & ub
                  -- allocate storage and store value
                  null; -- lower_bound_value_offset;
                  null; -- upper_bound_value_offset;

               when Floating_Point_Constraint |
                    Fixed_Point_Constraint    |
                    Discriminant_Constraint   =>
                  null;

            end case;

         when Type_Entry  | Subtype_Entry     | Derived_Type_Entry     =>
            null;

         when Exception_Entry  =>
            null;

         when Procedure_Entry           |
              Enumeration_Literal_Entry |
              Package_Entry             |
              Label_Entry               |
              Record_Component_Entry    |
              Simple_Name_Entry         |
              Unknown     =>
            null;

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "allocate_storage");
         raise ;

   end Allocate_Storage;

   -- -----------------------------------------------------------------------
   --=======================================================================
   --  <subprogram name>
   --
   --    deallocate_storage
   --
   --  <description>
   --
   --  deallocate_storage frees the data storage location associated with
   --  the specified symtbl entry item.
   --
   --======================================================================
   procedure Deallocate_Storage
       (Context      : in out Context_Control.Context_Definition;
        Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      Location : Ace_Natural := Ace_Natural'FIRST;

   begin
      case Get_Entry_Kind (Symtbl_Entry) is


         when Function_Entry   | Object_Entry | Formal_Param_Entry     =>
            declare
               Type_Entry_Found : Ref_Symbol_Table_Entry;
            begin

               -- find the "root" type of the item that needs storage allocated
               case Get_Entry_Kind (Get_Type (Symtbl_Entry)) is


                  when Type_Entry   =>
                     Type_Entry_Found := Get_Type (Symtbl_Entry);
                  when Subtype_Entry     | Derived_Type_Entry     =>
                     Type_Entry_Found :=
                        Get_Root_Type (Get_Type (Symtbl_Entry));
                  when Function_Entry            |
                       Object_Entry              |
                       Formal_Param_Entry        |
                       Procedure_Entry           |
                       Constraint_Entry          |
                       Enumeration_Literal_Entry |
                       Package_Entry             |
                       Label_Entry               |
                       Record_Component_Entry    |
                       Simple_Name_Entry         |
                       Exception_Entry           |
                       Unknown =>
                     -- can't be !!
                     Show_Error (0, Internal,
                                 "allocate_storage: enumeration not expected");
                     return ;
               end case;

               -- note:
               --
               -- entry's level and offset are set during semantic analysis
               if (Is_Type_An_Enumeration_Type (Type_Entry_Found) or else
                  Is_Type_An_Integer_Type (Type_Entry_Found)) then
                  Location :=
                     Get_Display_For_Level (Context, Get_Level (Symtbl_Entry))
                      + Get_Value_Offset (Symtbl_Entry);
                  declare
                     Old_Obj : Ref_Data_Storage :=
                        Context_Control.Get_Data_Storage_Stack (Context,
                                                                Location);
                  begin
                     Free (Old_Obj);
                  end;

               elsif (Is_Type_A_Record_Type (Type_Entry_Found)) then
                  null;

                  -- "dealloc storage" for each record component;
                  Show_Error (0, Internal, "[tbd]: alloc storage for records");
               else
                  -- either not implementing or error
                  -- -- array
                  -- -- real
                  -- *** let's assume it's a string for now! [tbd] -- oei
                  Location :=
                     Get_Display_For_Level (Context, Get_Level (Symtbl_Entry))
                      + Get_Value_Offset (Symtbl_Entry);
                  declare
                     Old_Obj : Ref_Data_Storage :=
                        Context_Control.Get_Data_Storage_Stack (Context,
                                                                Location);
                  begin
                     Free (Old_Obj);
                  end;

               end if;

               --- array
               --- real

            end;


         when Constraint_Entry =>
            case Symtbl_Entry.Constraint_Ref.Constraint_Kind is

               when Range_Constraint     | Index_Constraint  =>
                  -- deallocate storage
                  null; -- lower_bound_value_offset;
                  null; -- upper_bound_value_offset;

               when Floating_Point_Constraint |
                    Fixed_Point_Constraint    |
                    Discriminant_Constraint   =>
                  null;

            end case;

         when Type_Entry  | Subtype_Entry     | Derived_Type_Entry     =>
            null;

         when Exception_Entry  =>
            null;

         when Procedure_Entry           |
              Enumeration_Literal_Entry |
              Package_Entry             |
              Label_Entry               |
              Simple_Name_Entry         |
              Record_Component_Entry    |
              Unknown     =>
            null;

      end case;


   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "deallocate_storage");
         raise ;

   end Deallocate_Storage;

   -- -----------------------------------------------------------------------
   --=======================================================================
   --  <subprogram name>
   --
   --    fetch_value
   --
   --  <description>
   --
   --   fetch_value returns the value from the storage stack given
   --   a symtbl entry.
   --
   --======================================================================
   function Fetch_Value
       (Context      : in Context_Control.Context_Definition;
        Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Data_Storage is

      Storage_Location : Ace_Natural;

      function Fetch_Value_From_Correct_Context
          (Context          : in Context_Control.Context_Definition;
           Storage_Location : in Ace_Natural) return Ref_Data_Storage is
         Threshold : Ace_Natural := Context_Control.Get_Threshold (Context);
      begin
         if Threshold /= 0 and then Threshold >= Storage_Location then
            return
               Fetch_Value_From_Correct_Context
                   (Context_Control.Get_Previous_Context (Context),
                    Storage_Location);
         else
            return Context_Control.Get_Data_Storage_Stack (Context,
                                                           Storage_Location);
         end if;

      end Fetch_Value_From_Correct_Context;
      -- pragma INLINE (Fetch_Value_From_Correct_Context);
      -- When this is Inline-ed an exception propagation error occurs, i.e.,
      -- Constraint_Error is not re-raised as Evaluator_Error (see exception
      -- block on Fetch_Value).  This is a VADS 5.5 t bug. wpl

   begin
      if Is_Removed (Symtbl_Entry) then
         raise Invalid_Reference;
      end if;
      case Get_Entry_Kind (Symtbl_Entry) is

         when Object_Entry     | Formal_Param_Entry     =>
            Storage_Location :=
               Get_Display_For_Level (Context, Get_Level (Symtbl_Entry)) +
               Get_Value_Offset (Symtbl_Entry);
            return Fetch_Value_From_Correct_Context (Context, Storage_Location);

         when Function_Entry   =>
            Storage_Location :=
               Context_Control.Get_Display_For_Level
                   (Context, Get_Level (Symtbl_Entry) + 1) +
               Get_Value_Offset (Symtbl_Entry);
            return Fetch_Value_From_Correct_Context (Context, Storage_Location);

         when Enumeration_Literal_Entry  =>
            -- note that ...
            -- this ref_data_storage node is typically copied after this,
            --    so that the node is not "free"-d afterwards [fix]
            return
               Create_Storage_For_Integer (Get_Literal_Value (Symtbl_Entry));

         when Exception_Entry  =>
            null;

         when Constraint_Entry |
              Type_Entry       |
              Subtype_Entry    |
              Derived_Type_Entry    =>
            null;

         when Procedure_Entry        |
              Package_Entry          |
              Label_Entry            |
              Record_Component_Entry |
              Simple_Name_Entry      |
              Unknown     =>
            null;

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "fetch_value");
         raise Evaluator_Error;

   end Fetch_Value;

   function Get_Data_Type (Item : Ref_Data_Storage) return Data_Representation
                          is
   begin
      return Item.all.Data_Type;
   end Get_Data_Type;


   --=======================================================================
   --  <subprogram name>
   --
   --    store_value
   --
   --  <description>
   --
   --  store_value places the value from the data storage structure into the
   --  item.
   --
   --======================================================================

   procedure Store_Value (Context : in out Context_Control.Context_Definition;
                          Symtbl_Entry : in Ref_Symbol_Table_Entry;
                          Value        : in Ref_Data_Storage) is

      Storage_Location : Ace_Natural;

      procedure Store_Value_In_Correct_Context
          (Context          : in out Context_Control.Context_Definition;
           Storage_Location : in Ace_Natural;
           Value            : in Ref_Data_Storage) is

         Threshold : Ace_Natural := Context_Control.Get_Threshold (Context);
         Context_Holder : Context_Control.Context_Definition;

      begin
         if Threshold /= 0 and then Threshold >= Storage_Location then
            Context_Holder := Context_Control.Get_Previous_Context (Context);
            Store_Value_In_Correct_Context (Context_Holder, Storage_Location,
                                            Value);
         else
            Context_Control.Set_Data_Storage_Stack (Context, Storage_Location,
                                                    Value);
         end if;
      end Store_Value_In_Correct_Context;

      pragma INLINE (Store_Value_In_Correct_Context);

   begin
      case Get_Entry_Kind (Symtbl_Entry) is


         when Object_Entry     | Formal_Param_Entry     =>

            Storage_Location :=
               Get_Display_For_Level (Context, Get_Level (Symtbl_Entry)) +
               Get_Value_Offset (Symtbl_Entry);

            Store_Value_In_Correct_Context (Context, Storage_Location, Value);

         when Function_Entry   =>

            Storage_Location :=
               Context_Control.Get_Display_For_Level
                   (Context, Get_Level (Symtbl_Entry) + 1) +
               Get_Value_Offset (Symtbl_Entry);

            Store_Value_In_Correct_Context (Context, Storage_Location, Value);

         when Exception_Entry  =>
            null;

         when Constraint_Entry |
              Type_Entry       |
              Subtype_Entry    |
              Derived_Type_Entry    =>
            null;

         when Procedure_Entry           |
              Enumeration_Literal_Entry |
              Package_Entry             |
              Label_Entry               |
              Record_Component_Entry    |
              Simple_Name_Entry         |
              Unknown     =>
            null;

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "store_value");
         raise ;

   end Store_Value;

   -- ----------------------------------------------------------------------

   --=======================================================================
   --  <subprogram name>
   --
   --    copy_value
   --
   --  <description>
   --
   --   copies current data storage node (succeeding nodes (next_item
   --   and next_component nodes) are not copied).
   --
   --====================================================================
   function Copy_Value (Value_Node : Ref_Data_Storage) return Ref_Data_Storage
                       is
      Temp : Ref_Data_Storage;
   begin
      case Value_Node.all.Data_Type is
         when Integer_Data     =>
            return
               new Indirect_Data_Storage'(
                  new Data_Storage'
                     (Data_Type      => Integer_Data,
                      Int_Item       => Value_Node.all.Int_Item,
                      Next_Item      => Value_Node.all.Next_Item,
                      Next_Component => Value_Node.all.Next_Component));
         when Char_Data   =>
            return
               new Indirect_Data_Storage'(
                  new Data_Storage'
                     (Data_Type      => Char_Data,
                      Char_Item      => Value_Node.all.Char_Item,
                      Next_Item      => Value_Node.all.Next_Item,
                      Next_Component => Value_Node.all.Next_Component));
         when String_Data =>
            return
               new Indirect_Data_Storage'(
                  new Data_Storage'
                     (Data_Type      => String_Data,
                      String_Item    => Value_Node.all.String_Item,
                      Next_Item      => Value_Node.all.Next_Item,
                      Next_Component => Value_Node.all.Next_Component));
         when Float_Data  =>
            return
               new Indirect_Data_Storage'(
                  new Data_Storage'
                     (Data_Type      => Float_Data,
                      Float_Item     => Value_Node.all.Float_Item,
                      Next_Item      => Value_Node.all.Next_Item,
                      Next_Component => Value_Node.all.Next_Component));

         when Array_Data  =>
--            Temp := 
            return
               new Indirect_Data_Storage'(
                  new Data_Storage'
                     (Data_Type      => Array_Data,
                      Array_Item     => Value_Node.all.Array_Item,
                      Next_Item      => Value_Node.all.Next_Item,
                      Next_Component => Value_Node.all.Next_Component));
--            Copy (Value_Node.Array_Item, Temp.Array_Item);
--            return Temp;
           
         when Void   =>
            return
               new Indirect_Data_Storage'(
                  new Data_Storage'
                     (Data_Type      => Void,
                      Next_Item      => Value_Node.all.Next_Item,
                      Next_Component => Value_Node.all.Next_Component));
      end case;
   end Copy_Value;

   --+++++++++++++++++++++
   -- decl rtns (start)
   --+++++++++++++++++++++

   --=======================================================================
   --  <subprogram name>
   --
   --    get_current_level, incr_level, decr_level,
   --    reset_leveloffset_to_top_level,
   --    reset_leveloffset_to_previous_dr_level,
   --    get_next_offset_value, init_offset_value.
   --
   --  <description>
   --
   --    level and offset manipulation and access routines
   --
   --=======================================================================

   -- --
   procedure Incr_Level (Context : in out Context_Control.Context_Definition)
                        is
   begin
      Context_Control.Set_Offset_Stack
          (Context, Context_Control.Get_Current_Level (Context),
           Context_Control.Get_Current_Offset (Context));
      Context_Control.Set_Current_Level
          (Context, Context_Control.Get_Current_Level (Context) + 1);
      Init_Offset_Value (Context);
   end Incr_Level;

   -- --
   procedure Decr_Level (Context : in out Context_Control.Context_Definition)
                        is
   begin
      Context_Control.Set_Current_Level
          (Context, Context_Control.Get_Current_Level (Context) - 1);
      Context_Control.Set_Current_Offset
          (Context,
           Context_Control.Get_Offset_Stack (Context,
                                             Get_Current_Level (Context)));
   end Decr_Level;

   -- --
   -- reset level/offset to top level (level zero)
   --
   procedure Reset_Leveloffset_To_Top_Level
       (Context : in out Context_Control.Context_Definition) is
   begin
      -- save this/"current" level/offset values
      Context_Control.Set_Saved_Level
          (Context, Context_Control.Get_Current_Level (Context));
      Context_Control.Set_Saved_Offset
          (Context, Context_Control.Get_Current_Offset (Context));
      for I in Context_Control.Get_Saved_Level (Context) ..
         Ace_Natural'FIRST + 1 loop
         Context_Control.Set_Saved_Offset_Stack
             (Context, I, Context_Control.Get_Offset_Stack (Context, I));
      end loop;
      -- "set" to level zero
      Context_Control.Set_Current_Level (Context, Ace_Natural'FIRST);
      Context_Control.Set_Current_Offset
          (Context,
           Context_Control.Get_Offset_Stack (Context, Ace_Natural'FIRST));

   end Reset_Leveloffset_To_Top_Level;

   -- reset level/offset to previous declarative region level
   --
   procedure Reset_Leveloffset_To_Previous_Dr_Level
       (Context : in out Context_Control.Context_Definition) is
   begin
      -- save "current"/newest level zero offset
      Context_Control.Set_Offset_Stack
          (Context, Ace_Natural'FIRST,
           Context_Control.Get_Current_Offset (Context));
      --
      -- reset "previous" level/offset
      for I in Ace_Natural'FIRST + 1 ..
         Context_Control.Get_Saved_Level (Context) loop
         Context_Control.Set_Offset_Stack
             (Context, I, Context_Control.Get_Saved_Offset_Stack (Context, I));
      end loop;
      Context_Control.Set_Current_Offset
          (Context, Context_Control.Get_Saved_Offset (Context));
      Context_Control.Set_Current_Level
          (Context, Context_Control.Get_Saved_Level (Context));
   end Reset_Leveloffset_To_Previous_Dr_Level;

   -- --
   procedure Next_Offset_Value
       (Context : in out Context_Control.Context_Definition;
        Value   : out Ace_Natural) is
   begin
      Context_Control.Set_Current_Offset
          (Context, Context_Control.Get_Current_Offset (Context) + 1);
      Value := (Context_Control.Get_Current_Offset (Context) - 1);
   end Next_Offset_Value;

   -- --
   procedure Init_Offset_Value
       (Context : in out Context_Control.Context_Definition) is
   begin
      Context_Control.Set_Current_Offset (Context, Ace_Natural'FIRST);
   end Init_Offset_Value;

   --+++++++++++++++++++++
   -- decl rtns (end)
   --+++++++++++++++++++++

   --+++++++++++++++++++++
   -- eval rtns (start)
   --+++++++++++++++++++++

   --=======================================================================
   --  <subprogram name>
   --
   --    get_last_storage_location_pointer, set_last_storage_location_pointer
   --    get_display_for_level, set_display_for_level
   --
   --  <description>
   --
   --   dss access & manipulation routines (for subprogram calls)
   --   display access & manipulation routines (for subprogram calls)
   --
   --=======================================================================


   --+++++++++++++++++++++
   -- eval rtns (end)
   --+++++++++++++++++++++

   --=======================================================================
   --  <subprogram name>
   --
   --    dump_dss, dump_display
   --
   --  <description>
   --
   --    output routines for display and dss (used in debuggin).
   --
   --======================================================================
   procedure Dump_Dss (Context : in Context_Control.Context_Definition) is
   begin
      Put_Line ("entire data storage stack dump =>");
      for I in Ace_Natural'FIRST ..
         (Context_Control.Get_Last_Storage_Location_Pointer (Context)) loop
         Put (Ace_Natural'IMAGE (I));
         Put (" -> ");
         if (Is_Empty (Context_Control.Get_Data_Storage_Stack (Context, I)))
             then
            Put_Line (" [null]");
         else
            Put_Line
                (Ace_Natural'IMAGE (
                    Get_Integer_Value
                        (Context_Control.Get_Data_Storage_Stack (Context, I))));
         end if;
      end loop;
   end Dump_Dss;

   procedure Dump_Display (Context : in Context_Control.Context_Definition) is
   begin
      Put ("current_level = ");
      Put_Line (Ace_Natural'IMAGE (Get_Current_Level (Context)));
      Put_Line ("entire display dump =>");
      for I in Ace_Max_Levels'FIRST .. Ace_Max_Levels'LAST loop
         Put (Ace_Max_Levels'IMAGE (I));
         Put (" -> ");
         Put_Line
             (Ace_Natural'IMAGE (
                 Get_Integer_Value
                     (Context_Control.Get_Data_Storage_Stack (Context, I))));
      end loop;
   end Dump_Display;

   -- ---------------------------------------------------------------------
   --=======================================================================
   --  <subprogram name>
   --
   --    create_storage_for_integer, create_storage_for_character,
   --    create_storage_for_string, Create_Storage_For_Array.
   --
   --  <description>
   --
   --    routines to create storage for the three basic types in ace.
   --
   --=======================================================================

   function Create_Storage_For_Integer
       (Value : in Ace_Integer := 0) return Ref_Data_Storage is

   begin
      return
         (new Indirect_Data_Storage'(
            new Data_Storage'(Data_Type      => Integer_Data,
                              Int_Item       => Value,
                              Next_Item      => null,
                              Next_Component => null)));

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "create_storage_for_integer");
         raise ;

   end Create_Storage_For_Integer;

   function Create_Storage_For_Array
       (Lower, Upper : in Index) return Ref_Data_Storage is

   begin
      return
         (new Indirect_Data_Storage'(
            new Data_Storage'(Data_Type      => Array_Data,
                              Array_Item       => (Lower, Upper, Array_Heap.Create),
                              Next_Item      => null,
                              Next_Component => null)));

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "create_storage_for_array");
         raise ;

   end Create_Storage_For_Array;

   -- -----------------------------------------------------------------------
   function Create_Storage_For_Character
       (Value : in Character := ' ') return Ref_Data_Storage is

   begin
      return
         (new Indirect_Data_Storage'(new Data_Storage'(Data_Type => Char_Data,
                                                       Char_Item      => Value,
                                                       Next_Item      => null,
                                                       Next_Component => null))
                );

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "create_storage_for_character");
         raise ;

   end Create_Storage_For_Character;

   -- -----------------------------------------------------------------------
   function Create_Storage_For_String
       (Value : in Ref_String := Create) return Ref_Data_Storage is

   begin
      return
         (new Indirect_Data_Storage'(
            new Data_Storage'(Data_Type      => String_Data,
                              String_Item    => Value,
                              Next_Item      => null,
                              Next_Component => null)));

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "create_storage_for_string");
         raise ;

   end Create_Storage_For_String;

   --=======================================================================
   --  <subprogram name>
   --
   --    get_integer_value, get_character_value,
   --    get_string_value.
   --
   --  <description>
   --
   --    routines to get the object stored for the three basic types in ace.
   --
   --=======================================================================

   function Get_Integer_Value (Node : in Ref_Data_Storage) return Ace_Integer
                              is

   begin
      return (Node.all.Int_Item);

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "get_integer_value");
         raise ;

   end Get_Integer_Value;

   -- -----------------------------------------------------------------------
   function Get_Character_Value (Node : in Ref_Data_Storage) return Character
                                is

   begin
      return (Node.all.Char_Item);

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "get_character_value");
         raise ;

   end Get_Character_Value;

   -- -----------------------------------------------------------------------
   function Get_String_Value (Node : in Ref_Data_Storage) return Ref_String is

   begin
      return (Node.all.String_Item);

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "get_string_value");
         raise ;

   end Get_String_Value;

   function Get_Array_Value (Node : in Ref_Data_Storage) return Array_Definition is

   begin
      return (Node.all.Array_Item);

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "get_array_value");
         raise ;

   end Get_Array_Value;

   -----------------
   -- -----------------------------------------------------------------------
   --=======================================================================
   --  <subprogram name>
   --
   --    set_integer_value, set_character_value,
   --    set_string_value, set_array_value.
   --
   --  <description>
   --
   --    routines to set the object stored for the three basic types in ace.
   --
   --======================================================================

   procedure Set_Integer_Value (Node : in Ref_Data_Storage;
                                Item : in Ace_Integer) is

   begin
      Node.all.Int_Item := Item;

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "Set_Integer_Value");
         raise ;

   end Set_Integer_Value;

   -- -----------------------------------------------------------------------
   procedure Set_Character_Value (Node : in Ref_Data_Storage;
                                  Item : in Character) is

   begin
      Node.all.Char_Item := Item;

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "Set_Character_Value");
         raise ;

   end Set_Character_Value;

   -- -----------------------------------------------------------------------
   procedure Set_String_Value (Node : in Ref_Data_Storage;
                               Item : in Ref_String) is

   begin
      Node.all.String_Item := Item;

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "Set_String_Value");
         raise ;

   end Set_String_Value;

   -- -----------------------------------------------------------------------
   procedure Set_Array_Value (Node : in Ref_Data_Storage;
                               Item : in Array_Definition) is

   begin
      Node.all.Array_Item := Item;

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "Set_Array_Value");
         raise ;

   end Set_Array_Value;

   -----------------
   -- -----------------------------------------------------------------------
   --=======================================================================
   --  <Subprogram Name>
   --
   --    Get_Next_Item, Get_Next_Component.
   --
   --  <Description>
   --
   --    Get routines for sequentially traversing the storage structure.
   --
   --=======================================================================
   function Get_Next_Item (Node : in Ref_Data_Storage) return Ref_Data_Storage
                          is

   begin
      return (Node.all.Next_Item);

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "Get_Next_Item");
         raise ;

   end Get_Next_Item;

   -- -----------------------------------------------------------------------
   function Get_Next_Component
       (Node : in Ref_Data_Storage) return Ref_Data_Storage is

   begin
      return (Node.all.Next_Component);

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "Get_Next_Component");
         raise ;

   end Get_Next_Component;

   -----------------
   -- -----------------------------------------------------------------------
   --=======================================================================
   --  <Subprogram Name>
   --
   --    Set_Next_Item, Set_Next_Component.
   --
   --  <Description>
   --
   --    Set routines for sequentially traversing and setting
   --    the storage structure.
   --
   --=======================================================================
   procedure Set_Next_Item (Node : in Ref_Data_Storage;
                            Item : in Ref_Data_Storage) is

   begin
      Node.all.Next_Item := Item;

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "Set_Next_Item");
         raise ;

   end Set_Next_Item;

   -- -----------------------------------------------------------------------
   procedure Set_Next_Component (Node : in Ref_Data_Storage;
                                 Item : in Ref_Data_Storage) is

   begin
      Node.all.Next_Component := Item;

   exception
      when Constraint_Error    =>
         Show_Error (1200, Internal, "Set_Next_Component");
         raise ;

   end Set_Next_Component;

   -- -----------------------------------------------------------------------

   -- Creation

   procedure Create (Arry                     : in out Array_Definition;
                     Lower_Bound, Upper_Bound : in Index) is
   begin
      Arry.Upper_Bound := Upper_Bound;
      Arry.Lower_Bound := Lower_Bound;
      Arry.Tree        := Array_Heap.Create;
   end Create;

   -- Add and Find Array Values

   procedure Store_Value (Arry     : in out Array_Definition;
                          Position : in Index;
                          Data     : in Ref_Data_Storage) is
      Temp1 : Array_Heap.Tree;
      Temp2 : Boolean;
      Node : Array_Index;
   begin
      if Position in Arry.Lower_Bound .. Arry.Upper_Bound then
         Node := (Data, Position);
         Array_Heap.Insertnode (Node, Arry.Tree, Temp1, Temp2);
      else
         raise Constraint_Error;
      end if;
   end Store_Value;

   function Fetch_Value (Arry     : in Array_Definition;
                          Position : in Index) return
                          Ref_Data_Storage is
      Temp : Array_Index;
   begin
      if Position in Arry.Lower_Bound .. Arry.Upper_Bound then
         Array_Heap.Find ((null, Position), Arry.Tree, Temp);
         return Temp.Item;
      else
         raise Constraint_Error;
      end if;
   end Fetch_Value;

   -- Slices and Copies

   procedure Slice (From        : in Array_Definition;
                    To          : in out Array_Definition;
                    Start, Stop : in Index) is
      Temp : Ref_Data_Storage;
   begin
      for I in Start .. Stop loop
         Temp := Fetch_Value (From, I);
         Store_Value (To, I, Temp);
      end loop;
   end Slice;


   procedure Copy (From : in Array_Definition;
                   To   : in out Array_Definition) is
   begin
      Array_Heap.Copy (From.Tree, To.Tree);
   end Copy;

   -- Attribute functions

   function Attr_Address (Arry : in Array_Definition) return System.Address is

      Dummy : System.Address;

   begin -- ATTR_ADDRESS
      Put_Line ("Body stub for ARRAY_STORAGE.ATTR_ADDRESS");
      return Dummy;

   exception
      when others    =>
         Put_Line ("Unhandled exception in ARRAY_STORAGE.ATTR_ADDRESS");
         raise ;
   end Attr_Address;


   function Attr_Constrained (Arry : in Array_Definition) return Boolean is

      Dummy : Boolean := True;

   begin -- ATTR_CONSTRAINED
      Put_Line ("Body stub for ARRAY_STORAGE.ATTR_CONSTRAINED");
      return Dummy;

   exception
      when others    =>
         Put_Line ("Unhandled exception in ARRAY_STORAGE.ATTR_CONSTRAINED");
         raise ;
   end Attr_Constrained;

   function Attr_First (Arry : in Array_Definition) return Index is
   begin
      return Arry.Lower_Bound;
   end Attr_First;

   function Attr_Last (Arry : in Array_Definition) return Index is
   begin
      return Arry.Upper_Bound;
   end Attr_Last;

   function Attr_Length (Arry : in Array_Definition) return Ace_Integer is
   begin
      return
         (Ace_Integer (Arry.Upper_Bound) - Ace_Integer (Arry.Lower_Bound)) + 1;
   end Attr_Length;


   -- function Attr_Range (Arry : in Array_Definition) return <range>;

   -- Pragma Procedures

   procedure Pack (Arry : in out Array_Definition) is

   begin -- PACK
      Put_Line ("Body stub for ARRAY_STORAGE.PACK");

   exception
      when others    =>
         Put_Line ("Unhandled exception in ARRAY_STORAGE.PACK");
         raise ;
   end Pack;

   -- Garbage Collection
   procedure Free (Arry : in out Array_Definition) is

   begin 
     Array_Heap.Destroytree (Arry.Tree);
     -- Free upper and lower bound too!
   end Free;


end Storage_Management_Routines;
