--|$Log:	staddpk.a,v $
--|Revision 8.0  91/01/17  10:46:23  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Subunit for Package - Symbol_Table_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--  The following routines are separate units for Symbol_Table_Routines.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Statement_Database;
use Statement_Database;

with Node_Get_Routines;
use Node_Get_Routines;

--==========================================================================
--  <Subprogram Name>
--
--    Add_Package_Decls_To_Symbol_Table
--
--  <Description>
--
--    Adds declarations onto the symbol table so that they may be
--    visible during parsing of the package body, or that the body
--    declarations may be visible for debugging purposes.  When parsing
--    the package body for the package specification, the objects that were
--    made visible by use statements within the package specification must
--    be made visible also for the package body.  The Process_Use_Stmts
--    should be set to True when the package spec declarations are being
--    added to the symbol table for the parsing of the package body.
--
--==========================================================================

separate (Symbol_Table_Routines)

procedure Add_Package_Decls_To_Symbol_Table
    (Context              : in out Context_Control.Context_Definition;
     Package_Symtbl_Entry : in Ref_Symbol_Table_Entry;
     Decl_Kind            : in Pkg_Decl_Kind := Pkg_Spec;
     Process_Use_Stmts    : in Boolean       := False) is

   Derived_Subprograms : Ref_List_Of_Subprograms :=
      Get_Derived_Subprogram_List (Package_Symtbl_Entry);
   Current_Subprogram  : Ref_Symbol_Table_Entry;

   Current_Node        : Ref_Statement_Database;

begin
   -- Add the subprograms derived within this package onto the symbol
   -- table when adding package spec declarations.
   if Decl_Kind = Pkg_Spec then
      while Derived_Subprograms /= null loop
         Current_Subprogram := Derived_Subprograms.Subprogram_Entry;
         if Is_Visible (Current_Subprogram) then
            Add_Entry_To_Symbol_Table (Context, Current_Subprogram);
         end if;
         Derived_Subprograms := Derived_Subprograms.Next_Subprogram_Entry;
      end loop;
   end if;

   -- Acquire the package declarations for the specification or the
   -- body
   if Decl_Kind = Pkg_Spec then
      Current_Node := Get_Package_Spec_Decls (Package_Symtbl_Entry);
   else
      Current_Node := Get_Package_Body_Decls (Package_Symtbl_Entry);
   end if;

   -- Add the declarations within the visible portion of the package
   -- onto the symbol table
   while not (Is_Empty (Current_Node)) loop
      case Check_Stmt_Node (Current_Node) is

         when Object_Declaration    =>
            declare
               List_Node : Ref_Statement_Database :=
                  Get_Name_List (Current_Node);
            begin
               while (not Is_Empty (List_Node)) loop
                  Add_Entry_To_Symbol_Table
                      (Context,
                       Get_Symbol_Table_Entry (Get_Expression (List_Node)));
                  List_Node := Get_Next_List_Item (List_Node);
               end loop;
            end;

         when Type_Declaration           |
              Subtype_Declaration        |
              Package_Declaration        |
              Procedure_Declaration      |
              Procedure_Spec_Declaration |
              Function_Declaration       |
              Function_Spec_Declaration  =>
            -- can ignore Package_Body_Declaration: spec already on symtbl
            Add_Entry_To_Symbol_Table (Context,
                                       Get_Symbol_Table_Entry (Current_Node));

         when Use_Statement    =>
            -- Only add objects introduced by a use when constructing
            -- the environment to parse the respective package body.
            if Process_Use_Stmts then
               declare
                  List_Node : Ref_Statement_Database :=
                     Get_Name_List (Current_Node);
               begin
                  while (not Is_Empty (List_Node)) loop
                     Add_Package_Decls_To_Symbol_Table
                         (Context,
                          Get_Symbol_Table_Entry (Get_Expression (List_Node)));
                     List_Node := Get_Next_List_Item (List_Node);
                  end loop;
               end;

            end if;

         when others =>
            null; -- ignore
      end case;

      Current_Node := Get_Next_Statement (Current_Node);
   end loop;

exception
   when Duplicate_Symbol_Table_Entry_Error    =>
      null;

end Add_Package_Decls_To_Symbol_Table;

--==========================================================================
--  <Subprogram Name>
--
--    Delete_Package_Decls_From_Symbol_Table
--
--  <Description>
--
--    Removes declarations from the symbol table associated with a
--    package.  This, in essence, is an "undo" of executing an Ada
--    use statement.
--    Symbol table entries are actually unlinked from the symbol table.
--
--==========================================================================
with Statement_Database;
use Statement_Database;

with Node_Get_Routines;
use Node_Get_Routines;

separate (Symbol_Table_Routines)

procedure Delete_Package_Decls_From_Symbol_Table
    (Context              : in out Context_Control.Context_Definition;
     Package_Symtbl_Entry : in Ref_Symbol_Table_Entry) is

   -- NOTE:
   --    The method for removal of the package decls is based upon the
   --    method in which they were placed upon the symbol table by the
   --    Add_Package_Decls_To_Symbol_Table.  Since the entries are
   --    contiguous on the symbol table, find the last and first symbol
   --    table entry.  Then, set the next_entry of the item prior to last
   --    to point the next_entry of first.

   -- Transform:
   --  HEAD -> ... -> [Prior] -> [Last] -> ... [First] -> [x] -> ...
   -- to:
   --  HEAD -> ... -> [Prior] -> [x] -> ...

   First_Entry, Last_Entry : Ref_Symbol_Table_Entry := null;

   Derived_Subprograms     : Ref_List_Of_Subprograms :=
      Get_Derived_Subprogram_List (Package_Symtbl_Entry);
   Current_Subprogram      : Ref_Symbol_Table_Entry;

   Current_Node            : Ref_Statement_Database;

begin
   -- Derived subprograms for the package may be the first entry
   -- (but can not be the last)
   if Derived_Subprograms /= null then
      First_Entry := Derived_Subprograms.Subprogram_Entry;
   end if;

   -- The package specification declarations may yield the first entry
   -- and must yield the last
   Current_Node := Get_Package_Spec_Decls (Package_Symtbl_Entry);

   while not (Is_Empty (Current_Node)) loop
      case Check_Stmt_Node (Current_Node) is

         when Object_Declaration    =>
            declare
               List_Node : Ref_Statement_Database :=
                  Get_Name_List (Current_Node);
            begin
               if First_Entry = null then
                  First_Entry :=
                     Get_Symbol_Table_Entry (Get_Expression (List_Node));
               end if;
               while (not Is_Empty (List_Node)) loop
                  Last_Entry :=
                     Get_Symbol_Table_Entry (Get_Expression (List_Node));
                  List_Node := Get_Next_List_Item (List_Node);
               end loop;
            end;

         when Type_Declaration           |
              Subtype_Declaration        |
              Package_Declaration        |
              Procedure_Spec_Declaration |
              Function_Spec_Declaration  =>
            -- can ignore Package_Body_Declaration: spec already on symtbl
            begin
               if First_Entry = null then
                  First_Entry := Get_Symbol_Table_Entry (Current_Node);
               end if;
               Last_Entry := Get_Symbol_Table_Entry (Current_Node);
               Unhide_Entry (Context, Last_Entry);
            end;

         when Use_Statement    =>
            null;

         when others =>
            null; -- ignore
      end case;

      Current_Node := Get_Next_Statement (Current_Node);
   end loop;

   -- Having determined the first and last symbol table entries associated
   -- with the package spec, find the entry prior to last and "unlink"
   -- all the entries from [Last] -> ... -> [First] -> ...
   if not Is_Empty (Last_Entry) then
      declare
         Previous_Symtbl_Entry, Current_Symtbl_Entry : Ref_Symbol_Table_Entry
          := Get_Top_Of_Symtbl (Context);
         Node_Type : Statement_Enumeration;
      begin
         while not (Is_Empty (Current_Symtbl_Entry) or else
            Current_Symtbl_Entry = Last_Entry) loop
            Previous_Symtbl_Entry := Current_Symtbl_Entry;
            Current_Symtbl_Entry := Get_Next_Entry (Current_Symtbl_Entry);
         end loop;

         if (Is_Empty (Current_Symtbl_Entry)) then
            raise Constraint_Error;
         elsif Current_Symtbl_Entry /= Get_Top_Of_Symtbl (Context) then
            Set_Next_Entry
                (Previous_Symtbl_Entry, Get_Next_Entry (First_Entry));
         else
            Reset_If_Symtbl_Head (Context, Last_Entry,
                                  Get_Next_Entry (First_Entry));
           --          The next statement can probably be removed, thanks to the
            --          Reset_If_Symtbl_Head procedure. [TBD raf 15NOV89]
        --            Set_Top_Of_Symtbl (Context, Get_Next_Entry (First_Entry));
         end if;
      end;
   end if;

end Delete_Package_Decls_From_Symbol_Table;

-- -------------------------------------------------------------------------
