--|$Log:	stget.a,v $
--|Revision 8.0  91/01/17  10:46:24  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Symtbl_Entry_Get_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    This package contains the set of routines that retrieves fields of symbol
--    table entries.  All retrievals of components from symbol table entries
--    should be made through the routines provided within this package.
--    This package (in conjunction with a similar package for storing into
--    symbol table entry components) provides the mechanism for hiding the
--    makeup and names of symbol table entry components.  Changes to the
--    structure and component names for symbol table entries would only
--    affect this package of routines, rather than all the other components
--    that deal with symbol table information.
--
--    The routines within the package are ordered with respect to the
--    definition of components for a symbol table entry.  The routines that
--    retrieve components that are common to all symbol table entries come
--    first, followed by components that are applicable to different variants
--    of the symbol table entry.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Statement_Database;
use Statement_Database;

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Help_Pkg;
use Help_Pkg;

with Storage_Management;
use Storage_Management;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Misc_Support_Routines;
use Misc_Support_Routines;

with Error_Handler;
use Error_Handler;

package body Symtbl_Entry_Get_Routines is

   -- ::::::::::::::::::::::
   -- : Entry_Kind ==> ALL :
   -- ::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Entry_Kind
   --
   --  <Description>
   --
   --    Return the kind of entry associated with the symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Entry_Kind
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Type_Entry_Enumeration
       is

      -- Valid for: ALL symbol table entries
      --   component ==> Entry_Kind (discriminant)

   begin
      return Symtbl_Entry.Entry_Kind;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Entry_Kind");
         raise Program_Error;

   end Get_Entry_Kind;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Next_Entry
   --
   --  <Description>
   --
   --    Return the next entry that is linked with to this symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Next_Entry
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: ALL symbol table entries
      --   component ==> Next_Entry

   begin
      return Symtbl_Entry.Next_Entry;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Next_Entry");
         raise Program_Error;

   end Get_Next_Entry;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Symbolic_Name
   --
   --  <Description>
   --
   --    Return the symbolic name associated with the symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Symbolic_Name
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_String is

      -- Valid for: ALL symbol table entries
      --   component ==> Symbol_Name

   begin
      return Symtbl_Entry.Symbol_Name;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Symbolic_Name");
         raise Program_Error;

   end Get_Symbolic_Name;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Level
   --
   --  <Description>
   --
   --    Return the static level associated with the symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Level
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural is

      -- Valid for: ALL symbol table entries
      --   component ==> Item_Level

   begin
      return Symtbl_Entry.Item_Level;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Level");
         raise Program_Error;

   end Get_Level;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Enclosing_Package
   --
   --  <Description>
   --
   --    Return the enclosing package symbol table entry associated with
   --    this symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Enclosing_Package
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: ALL symbol table entries
      --   component ==> Enclosing_Package

   begin
      return Symtbl_Entry.Enclosing_Package;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Enclosing_Package");
         raise Program_Error;

   end Get_Enclosing_Package;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Visible
   --
   --  <Description>
   --
   --    Return whether the item is visible
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Visible
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: ALL symbol table entries
      --   component ==> Visibility

   begin
      return (Symtbl_Entry.Visibility = Visible);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Visible");
         raise Program_Error;

   end Is_Visible;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Hidden
   --
   --  <Description>
   --
   --    Return whether the item is hidden
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Hidden (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean
                      is

      -- Valid for: ALL symbol table entries
      --   component ==> Visibility

   begin
      return (Symtbl_Entry.Visibility = Hidden);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Hidden");
         raise Program_Error;

   end Is_Hidden;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Removed
   --
   --  <Description>
   --
   --    Return whether the item has been "removed" from the symbol table
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Removed
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: ALL symbol table entries
      --   component ==> Visibility

   begin
      return (Symtbl_Entry.Visibility = Removed);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Removed");
         raise Program_Error;

   end Is_Removed;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Statement_Number
   --
   --  <Description>
   --
   --    Return the statement number associated with the symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Statement_Number
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural is

      -- Valid for: ALL symbol table entries
      --   component ==> Statement_Number

   begin
      return Symtbl_Entry.Statement_Number;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Statement_Number");
         raise Program_Error;

   end Get_Statement_Number;

   -- -------------------------------------------------------------------------

   -- ::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Package_Entry :
   -- ::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Context_Clause_Body
   --
   --  <Description>
   --
   --    Return the context clause associated with the package implementation
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Context_Clause_Body
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database
       is

      -- Valid for: Package_Entry
      --    component ==> Context_Clause

   begin
      case Symtbl_Entry.Entry_Kind is
         when Package_Entry    =>
            return (Convert_Fake (Symtbl_Entry.Pkg_Body_Context_Clause));
         when others =>
            Show_Error (2000, Internal, "Get_Context_Clause_Body 1");
            raise Program_Error;
      end case;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Context_Clause_Body 2");
         raise Program_Error;
   end Get_Context_Clause_Body;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Context_Clause
   --
   --  <Description>
   --
   --    Return the context clause associated with the package specification
   --    or subprogram.
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Context_Clause
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database
       is

      -- Valid for: Package_Entry
      --    component ==> Context_Clause

   begin
      case Symtbl_Entry.Entry_Kind is
         when Package_Entry    =>
            return (Convert_Fake (Symtbl_Entry.Pkg_Spec_Context_Clause));
         when Procedure_Entry  | Function_Entry    =>
            return (Convert_Fake (Symtbl_Entry.Subprg_Context_Clause));
         when others =>
            Show_Error (2000, Internal, "Get_Context_Clause 1");
            raise Program_Error;
      end case;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Context_Clause 2");
         raise Program_Error;
   end Get_Context_Clause;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Package_Spec_Decls
   --
   --  <Description>
   --
   --    Return the declarations associated with the package specification
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Package_Spec_Decls
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database
       is

      -- Valid for: Package_Entry
      --    component ==> Package_Spec_Decls

   begin
      return (Convert_Fake (Symtbl_Entry.Package_Spec_Decls));

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Package_Spec_Decls");
         raise Program_Error;

   end Get_Package_Spec_Decls;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Package_Body_Decls
   --
   --  <Description>
   --
   --    Return the declarations associated with the package body
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Package_Body_Decls
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database
       is

      -- Valid for: Package_Entry
      --    component ==> Package_Body_Decls

   begin
      return (Convert_Fake (Symtbl_Entry.Package_Body_Decls));

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Package_Body_Decls");
         raise Program_Error;

   end Get_Package_Body_Decls;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Has_A_Body
   --
   --  <Description>
   --
   --    Return an indication of whether the package has a body defined
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Has_A_Body
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Package_Entry
      --    component ==> Has_Body

   begin
      return Symtbl_Entry.Has_Body;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Has_A_Body");
         raise Program_Error;

   end Has_A_Body;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Head_Of_Statements_List
   --
   --  <Description>
   --
   --    Return the head of the list of statements for the executable
   --    statements associated with a package symbol table entry, a
   --    subprogram symbol table entry, or a label symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Head_Of_Statements_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database
       is

      -- Valid for: Package_Entry
      --   component ==> Package_Stmts
      -- Valid for: Procedure_Entry, Function_Entry
      --   component ==> Statements_List
      -- Valid for: Label_Entry
      --   component ==> Label_Stmt

      -- The statements_list component must be converted from a
      -- Ref_Statement_Database_Fake to a Ref_Statement_Database

   begin
      case Symtbl_Entry.Entry_Kind is
         when Procedure_Entry  | Function_Entry    =>
            return (Convert_Fake (Symtbl_Entry.Statements_List));

         when Label_Entry =>
            return (Convert_Fake (Symtbl_Entry.Label_Stmt));

         when Package_Entry    =>
            return (Convert_Fake (Symtbl_Entry.Package_Stmts));

         when others =>
            Show_Error (2000, Internal, "Get_Head_Of_Statements_List1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Head_Of_Statements_List");
         raise Program_Error;

   end Get_Head_Of_Statements_List;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Derived_Subprogram_List
   --
   --  <Description>
   --
   --    Return the head of the list of derived subprograms that are
   --    associated with a package.  Also, for types and derived types,
   --    return the head of the list of subprograms which are derivable
   --    from this type (or derived type).
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Derived_Subprogram_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return
         Ref_List_Of_Subprograms is

      -- Valid for: Package_Entry
      --   component ==> Pkg_Derived_Subprogram_List
      -- Valid for: Type_Entry
      --   component ==> Type_Derived_Subprogram_List
      -- Valid for: Derived_Type_Entry
      --   component ==> DType_Derived_Subprogram_List

   begin
      case Symtbl_Entry.Entry_Kind is
         when Package_Entry    =>
            return Symtbl_Entry.Pkg_Derived_Subprogram_List;

         when Type_Entry  =>
            return Symtbl_Entry.Type_Derived_Subprogram_List;

         when Derived_Type_Entry    =>
            return Symtbl_Entry.Dtype_Derived_Subprogram_List;

         when others =>
            Show_Error (2000, Internal, "Get_Derived_Subprogram_List1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Derived_Subprogram_List");
         raise Program_Error;

   end Get_Derived_Subprogram_List;

   -- -------------------------------------------------------------------------

   -- ::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Procedure_Entry :
   -- :                Function_Entry  :
   -- ::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Parameter_Count
   --
   --  <Description>
   --
   --    Return the number of parameters associated with a subprogram
   --    symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Parameter_Count
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural is
      -- Valid for: Procedure_Entry, Function_Entry
      --    component ==> Parameter_Count
   begin
      return Symtbl_Entry.Parameter_Count;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Parameter_Count");
         raise Program_Error;
   end Get_Parameter_Count;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Number_Of_Default_Parameters
   --
   --  <Description>
   --
   --    Return the number of parameters associated with a subprogram
   --    symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Number_Of_Default_Parameters
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural is
      -- Valid for: Procedure_Entry, Function_Entry
      -- record component name: Default_Parameter_Count
   begin
      return Symtbl_Entry.Default_Parameter_Count;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Number_Of_Default_Parameters");
         raise Program_Error;
   end Get_Number_Of_Default_Parameters;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Has_Default_Parameters
   --
   --  <Description>
   --
   --    Return the number of parameters associated with a subprogram
   --    symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Has_Default_Parameters
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is
      -- Valid for: Procedure_Entry, Function_Entry
      -- record component name: Default_Parameter_Count
   begin
      return Symtbl_Entry.Default_Parameter_Count > Ace_Natural'FIRST;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Has_Default_Parameters");
         raise Program_Error;
   end Has_Default_Parameters;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Head_Of_Parameters_List
   --
   --  <Description>
   --
   --    Return the first parameter's symbol table entry associated with a
   --    subprogram symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Head_Of_Parameters_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: Procedure_Entry, Function_Entry
      --    component ==> Head_Param_List

   begin
      return Symtbl_Entry.Head_Param_List;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Head_Of_Parameters_List");
         raise Program_Error;

   end Get_Head_Of_Parameters_List;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Tail_Of_Parameters_List
   --
   --  <Description>
   --
   --    Return the last parameter's symbol table entry associated with a
   --    subprogram symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Tail_Of_Parameters_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: Procedure_Entry, Function_Entry
      --    component ==> Tail_Param_List

   begin
      return Symtbl_Entry.Tail_Param_List;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Tail_Of_Parameters_List");
         raise Program_Error;

   end Get_Tail_Of_Parameters_List;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Return_Type
   --
   --  <Description>
   --
   --    Return the return value's symbol table entry associated with a
   --    subprogram (only appropriate for functions) symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Return_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: Function_Entry
      --    component ==> Return_Type

      -- Ignored for: Procedure_Entry
      --    component ==> Return_Type -- just return whatever is there

   begin
      return Symtbl_Entry.Return_Type;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Return_Type");
         raise Program_Error;

   end Get_Return_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Head_Of_Declarations_List
   --
   --  <Description>
   --
   --    Return the head of the list of statements for the declaration
   --    statements associated with a subprogram symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Head_Of_Declarations_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database
       is

      -- Valid for: Procedure_Entry, Function_Entry
      --   component ==> Declarations_List

      -- The declarations_list component must be converted from a
      -- Ref_Statement_Database_Fake to a Ref_Statement_Database

   begin
      return (Convert_Fake (Symtbl_Entry.Declarations_List));

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Head_Of_Declarations_List");
         raise Program_Error;

   end Get_Head_Of_Declarations_List;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Value_Offset
   --
   --  <Description>
   --
   --    Return the reference to the data storage location in which
   --    the value associated with the item (object, function return
   --    or parameter) is stored
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Value_Offset
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural is

      -- Valid for: Object_Entry
      --   component ==> Object_Value_Offset
      -- Valid for: Function_Entry
      --   component ==> Function_Value_Offset
      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Value_Offset

      -- Ignored for: Procedure_Entry
      --   component ==> Function_Value_Offset

   begin
      case Symtbl_Entry.Entry_Kind is
         when Function_Entry   =>
            return Symtbl_Entry.Function_Value_Offset;

         when Object_Entry     =>
            return Symtbl_Entry.Object_Value_Offset;

         when Formal_Param_Entry    =>
            return Symtbl_Entry.Param_Value_Offset;

         when Procedure_Entry  =>
            return 0; -- default value

         when others =>
            Show_Error (2000, Internal, "Get_Value_Offset1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Value_Offset");
         raise Program_Error;

   end Get_Value_Offset;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Subprogram_Built_In
   --
   --  <Description>
   --
   --    Return whether the subprogram is a built-in routine within ACE
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Subprogram_Built_In
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Procedure_Entry, Function_Entry
      --   component ==> Is_Built_In_Subprogram

   begin
      return (Symtbl_Entry.Is_Built_In_Subprogram /= Ace_Natural'LAST);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Subprogram_Built_In");
         raise Program_Error;

   end Is_Subprogram_Built_In;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Built_In_Instance_Number
   --
   --  <Description>
   --
   --    Return the instance # used for evaluating built-ins within ACE
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Built_In_Instance_Number
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural is

      -- Valid for: Procedure_Entry, Function_Entry
      --   component ==> Is_Built_In_Subprogram

   begin
      return Symtbl_Entry.Is_Built_In_Subprogram;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Built_In_Instance_Number");
         raise Program_Error;

   end Get_Built_In_Instance_Number;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Derived_Subprogram_Parent
   --
   --  <Description>
   --
   --    Return the reference to the parent of a derived subprogram.
   --    (A non-derived subprogram has a null parent.)
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Derived_Subprogram_Parent
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is
      -- Valid for: Procedure_Entry, Function_Entry
      --    component ==> Derived_Subprogram_Parent
   begin
      return Symtbl_Entry.Derived_Subprogram_Parent;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Derived_Subprogram_Parent");
         raise Program_Error;
   end Get_Derived_Subprogram_Parent;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Function_Entry
   --
   --  <Description>
   --
   --    Return the reference to the parameterless function call symtbl entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Function_Entry
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is
      -- Valid for: Enumeration_Literal_Entry
      --    component ==> Func_Entry
   begin
      return Symtbl_Entry.Func_Entry;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Function_Entry");
         raise Program_Error;
   end Get_Function_Entry;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Enumeration_Literal_Entry
   --
   --  <Description>
   --
   --    Return the reference to the enumeration literal entry if there is one
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Enumeration_Literal_Entry
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is
      -- Valid for: Procedure_Entry, Function_Entry
      --    component ==> Enum_Lit_Entry
   begin
      return Symtbl_Entry.Enum_Lit_Entry;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Enumeration_Literal_Entry");
         raise Program_Error;
   end Get_Enumeration_Literal_Entry;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Object_Entry :
   -- :::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Type
   --
   --  <Description>
   --
   --    Return the reference to the type symbol table entry associated with
   --    either an object, a parameter, an enumeration literal, or a record
   --    component
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: Object_Entry
      --   component ==> Object_Type
      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Type
      -- Valid for: Enumeration_Literal_Entry
      --   component ==> Enumeration_Type
      -- Valid for: Record_Component_Entry
      --   component ==> Component_Type

   begin
      case Symtbl_Entry.Entry_Kind is
         when Object_Entry     =>
            return Symtbl_Entry.Object_Type;

         when Formal_Param_Entry    =>
            return Symtbl_Entry.Param_Type;

         when Enumeration_Literal_Entry  =>
            return Symtbl_Entry.Enumeration_Type;

         when Record_Component_Entry     =>
            return Symtbl_Entry.Component_Type;

         when Function_Entry   =>
            return Symtbl_Entry.Return_Type;

         when others =>
            Show_Error (2000, Internal, "Get_Type1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Type");
         raise Program_Error;

   end Get_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Has_Constraint
   --
   --  <Description>
   --
   --    Return whether the object, subtype, or the derived type has a
   --    constraint definition associated with it
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Has_Constraint
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Object_Entry
      --    component ==> Object_Constraint
      -- Valid for: Subtype_Entry, Derived_Type_Entry
      --    component ==> Constraint_Info

   begin
      case Symtbl_Entry.Entry_Kind is
         when Object_Entry     =>
            return (Symtbl_Entry.Object_Constraint /= null);

         when Subtype_Entry    | Derived_Type_Entry     =>
            return (Symtbl_Entry.Constraint_Info /= null);

         when others =>
            Show_Error (2000, Internal, "Has_Constraint1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Has_Constraint");
         raise Program_Error;

   end Has_Constraint;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Mode_Default_In
   --
   --  <Description>
   --
   --    Return whether the mode for the item (object or parameter) is
   --    IN by default
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Mode_Default_In
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Object_Entry
      --   component ==> Object_Mode
      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Mode

   begin
      case Symtbl_Entry.Entry_Kind is
         when Object_Entry     =>
            return (Symtbl_Entry.Object_Mode = Default_In_Mode);

         when Formal_Param_Entry    =>
            return (Symtbl_Entry.Param_Mode = Default_In_Mode);

         when others =>
            Show_Error (2000, Internal, "Is_Mode_Default_In1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Mode_Default_In");
         raise Program_Error;

   end Is_Mode_Default_In;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Mode_Explicit_In
   --
   --  <Description>
   --
   --    Return whether the mode for the item (object or parameter) is
   --    IN (Mode_In), explicitly stated by user
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Mode_Explicit_In
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Object_Entry
      --   component ==> Object_Mode
      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Mode

   begin
      case Symtbl_Entry.Entry_Kind is
         when Object_Entry     =>
            return (Symtbl_Entry.Object_Mode = In_Mode);

         when Formal_Param_Entry    =>
            return (Symtbl_Entry.Param_Mode = In_Mode);

         when Function_Entry   =>
            return True;

         when others =>
            Show_Error (2000, Internal, "Is_Mode_Explicit_In1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Mode_Explicit_In");
         raise Program_Error;

   end Is_Mode_Explicit_In;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Mode_In
   --
   --  <Description>
   --
   --    Return whether the mode for the item (object or parameter) is IN,
   --    either explicitly (In_Mode) or implicitly (Default_In_Mode)
   --
   --    For type checking, the return value of a function and an
   --    enumeration literal are "read only" items, and their mode can be
   --    considered as IN.  This routine returns True for function and
   --    enumeration literal entries.
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Mode_In
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Object_Entry
      --   component ==> Object_Mode
      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Mode
      -- Valid for: Function_Entry, Enumeration_Literal_Entry
      --   component ==> <<none>>

   begin
      case Symtbl_Entry.Entry_Kind is
         when Object_Entry     =>
            return (Symtbl_Entry.Object_Mode = In_Mode or
                   Symtbl_Entry.Object_Mode = Default_In_Mode);

         when Formal_Param_Entry    =>
            return (Symtbl_Entry.Param_Mode = In_Mode or
                   Symtbl_Entry.Param_Mode = Default_In_Mode);

         when Function_Entry   | Enumeration_Literal_Entry   =>
            return True;

         when others =>
            Show_Error (2000, Internal, "Is_Mode_In1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Mode_In");
         raise Program_Error;

   end Is_Mode_In;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Mode_Out
   --
   --  <Description>
   --
   --    Return whether the mode for the item (object or parameter) is OUT
   --
   --    For type checking, the return value of a function and an
   --    enumeration literal are "read only" items, and their mode can be
   --    considered as IN.  This routine returns False for function and
   --    enumeration literal entries.
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Mode_Out
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Object_Entry
      --   component ==> Object_Mode
      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Mode
      -- Valid for: Function_Entry, Enumeration_Literal_Entry
      --   component ==> <<none>>

   begin
      case Symtbl_Entry.Entry_Kind is
         when Object_Entry     =>
            return (Symtbl_Entry.Object_Mode = Out_Mode);

         when Formal_Param_Entry    =>
            return (Symtbl_Entry.Param_Mode = Out_Mode);

         when Function_Entry   | Enumeration_Literal_Entry   =>
            return False;

         when others =>
            Show_Error (2000, Internal, "Is_Mode_Out1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Mode_Out");
         raise Program_Error;

   end Is_Mode_Out;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Mode_In_Out
   --
   --  <Description>
   --
   --    Return whether the mode for the item (object or parameter) is IN OUT
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Mode_In_Out
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Object_Entry
      --   component ==> Object_Mode
      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Mode

   begin
      case Symtbl_Entry.Entry_Kind is
         when Object_Entry     =>
            return (Symtbl_Entry.Object_Mode = In_Out_Mode);

         when Formal_Param_Entry    =>
            return (Symtbl_Entry.Param_Mode = In_Out_Mode);

         when others =>
            Show_Error (2000, Internal, "Is_Mode_In_Out1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Mode_In_Out");
         raise Program_Error;

   end Is_Mode_In_Out;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Are_Modes_Equal
   --
   --  <Description>
   --
   --    Return whether the modes for the two items (object or parameter)
   --    are identical
   --
   --    This handles any combination of objects and parameters.
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Are_Modes_Equal
       (Symtbl_Entry_One : in Ref_Symbol_Table_Entry;
        Symtbl_Entry_Two : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Object_Entry
      --   component ==> Object_Mode
      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Mode

   begin
      case Symtbl_Entry_One.Entry_Kind is
         when Object_Entry     =>
            case Symtbl_Entry_Two.Entry_Kind is
               when Object_Entry    =>
                  return (Symtbl_Entry_One.Object_Mode =
                         Symtbl_Entry_Two.Object_Mode);

               when Formal_Param_Entry   =>
                  return (Symtbl_Entry_One.Object_Mode =
                         Symtbl_Entry_Two.Param_Mode);

               when others     =>
                  Show_Error (2000, Internal, "Are_Modes_Equal1");
            end case;

         when Formal_Param_Entry    =>
            case Symtbl_Entry_Two.Entry_Kind is
               when Formal_Param_Entry   =>
                  return (Symtbl_Entry_One.Param_Mode =
                         Symtbl_Entry_Two.Param_Mode);

               when Object_Entry    =>
                  return (Symtbl_Entry_One.Param_Mode =
                         Symtbl_Entry_Two.Object_Mode);
               when others     =>
                  Show_Error (2000, Internal, "Are_Modes_Equal2");
            end case;

         when others =>
            Show_Error (2000, Internal, "Are_Modes_Equal1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Are_Modes_Equal");
         raise Program_Error;

   end Are_Modes_Equal;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Formal_Param_Entry :
   -- :::::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Initialization_Code
   --
   --  <Description>
   --
   --    Return the statement database node associated with the
   --    initialization code for the formal parameter or a record component
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Initialization_Code
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database
       is

      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Init
      -- Valid for: Record_Component_Entry
      --   component ==> Component_Init

   begin
      case Symtbl_Entry.Entry_Kind is
         when Formal_Param_Entry    =>
            return (Convert_Fake (Symtbl_Entry.Param_Init));

         when Record_Component_Entry     =>
            return (Convert_Fake (Symtbl_Entry.Component_Init));

         when others =>
            Show_Error (2000, Internal, "Get_Initialization_Code1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Initialization_Code");
         raise Program_Error;

   end Get_Initialization_Code;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Next_Parameter
   --
   --  <Description>
   --
   --    Return the symbol table entry for the next parameter following this
   --    parameter in the specification of the subprogram
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Next_Parameter
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: Formal_Param_Entry
      --   component ==> Next_Param

   begin
      return Symtbl_Entry.Next_Param;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Next_Parameter");
         raise Program_Error;

   end Get_Next_Parameter;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Preceded_By_Comma
   --
   --  <Description>
   --
   --    Return whether the formal parameter was preceded by a comma
   --    in the textual definition of the subprogram declaration
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Preceded_By_Comma
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Formal_Param_Entry
      --   component ==> Preceded_By_Comma

   begin
      return Symtbl_Entry.Preceded_By_Comma;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Preceded_By_Comma");
         raise Program_Error;

   end Is_Preceded_By_Comma;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Type_Entry :
   -- :::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_State_Incomplete
   --
   --  <Description>
   --
   --    Return whether the state of the this type identifier is an
   --    incomplete type definition
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Type_State_Incomplete
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Type_Entry
      --   component ==> Concreteness

   begin
      return (Symtbl_Entry.Concreteness = Incomplete);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Type_State_Incomplete");
         raise Program_Error;

   end Is_Type_State_Incomplete;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_State_Full
   --
   --  <Description>
   --
   --    Return whether the state of the this type identifier is a
   --    full type definition
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Type_State_Full
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Type_Entry
      --   component ==> Concreteness

   begin
      return (Symtbl_Entry.Concreteness = Full);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Type_State_Full");
         raise Program_Error;

   end Is_Type_State_Full;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_State_Private
   --
   --  <Description>
   --
   --    Return whether the state of the this type identifier is a
   --    private type definition
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Type_State_Private
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Type_Entry
      --   component ==> Concreteness

   begin
      return (Symtbl_Entry.Concreteness = Only_Private);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Type_State_Private");
         raise Program_Error;

   end Is_Type_State_Private;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_State_Limited_Private
   --
   --  <Description>
   --
   --    Return whether the state of the this type identifier is a
   --    limited private type definition
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Type_State_Limited_Private
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Type_Entry
      --   component ==> Concreteness

   begin
      return (Symtbl_Entry.Concreteness = Limited_Private);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Type_State_Limited_Private");
         raise Program_Error;

   end Is_Type_State_Limited_Private;

   -- -------------------------------------------------------------------------
   -- ::::::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Type_Entry          :
   -- :                Subtype_Entry       :
   -- :                Derived_Type_Entry  :
   -- ::::::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Type_Definition
   --
   --  <Description>
   --
   --    For a symbol table entry, follows links until the Basic Type
   --    Declaration of a Node
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Type_Definition
       (Node : Ref_Symbol_Table_Entry) return Basic_Type_Definitions is
   begin
      case Get_Entry_Kind (Node) is
         when Type_Entry  =>
            return Node.Type_Definition.Type_Kind;
         when Subtype_Entry    | Derived_Type_Entry     =>
            return (Get_Type_Definition (Node.Base_Type));
         when others =>
            Show_Error (2000, Internal, "Get_Type_Definition1");
      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Type_Definition");
         raise Program_Error;

   end Get_Type_Definition;

   -- -------------------------------------------------------------------------
   -- ::::::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Subtype_Entry       :
   -- :                Derived_Type_Entry  :
   -- ::::::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Entry_Base_Type
   --
   --  <Description>
   --
   --    Return the base type definition associated with the subtype or the
   --    derived type symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Entry_Base_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: Subtype_Entry, Derived_Type_Entry
      --    component ==> Base_Type

   begin
      return Symtbl_Entry.Base_Type;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Entry_Base_Type");
         raise Program_Error;

   end Get_Entry_Base_Type;

   -- -------------------------------------------------------------------------

   -- ::::::::::::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Enumeration_Literal_Entry :
   -- ::::::::::::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Literal_Value
   --
   --  <Description>
   --
   --    Return the internal numeric value assigned to the enumeration literal
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Literal_Value
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural is

      -- Valid for: Enumeration_Literal_Entry
      --    component ==> Literal_Value

   begin
      return Symtbl_Entry.Literal_Value;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Literal_Value");
         raise Program_Error;

   end Get_Literal_Value;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Successor
   --
   --  <Description>
   --
   --    Return the symbol table entry for the successor to this
   --    enumeration literal
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Successor
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: Enumeration_Literal_Entry
      --    component ==> Succ_Enum_Literal

   begin
      return Symtbl_Entry.Succ_Enum_Literal;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Successor");
         raise Program_Error;

   end Get_Successor;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Predecessor
   --
   --  <Description>
   --
   --    Return the symbol table entry for the predecessor to this
   --    enumeration literal
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Predecessor
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: Enumeration_Literal_Entry
      --    component ==> Pred_Enum_Literal

   begin
      return Symtbl_Entry.Pred_Enum_Literal;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Predecessor");
         raise Program_Error;

   end Get_Predecessor;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Record_Component_Entry :
   -- :::::::::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Next_Record_Component
   --
   --  <Description>
   --
   --    Return the symbol table entry for the next record component following
   --    this record component
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Next_Record_Component (Symtbl_Entry : in Ref_Symbol_Table_Entry
                                       ) return Ref_Symbol_Table_Entry is

      -- Valid for: Record_Component_Entry
      --    component ==> Next_Record_Component

   begin
      return Symtbl_Entry.Next_Record_Component;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Next_Record_Component");
         raise Program_Error;

   end Get_Next_Record_Component;

   -- :::::::::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Simple_Name_Entry      :
   -- :   Component - Simple_Name_Construct   :
   -- :::::::::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Simple_Name_Construct
   --
   --  <Description>
   --
   --    Get the construct associated with the simple name.
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Simple_Name_Construct (Symtbl_Entry : in Ref_Symbol_Table_Entry
                                       ) return Ref_Statement_Database is
      -- Valid for: Simple_Name_Entry
      --    component ==> Simple_Name_Construct

   begin
      return Convert_Fake (Symtbl_Entry.Simple_Name_Construct);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Simple_Name_Construct");
         raise Program_Error;
   end Get_Simple_Name_Construct;

   -- -------------------------------------------------------------------------
   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Type_Entry               :
   -- :   Component - Type_Definition.Type_Kind :
   -- :::::::::::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_An_Enumeration_Type
   --
   --  <Description>
   --
   --    Returns whether the type entry defines an enumeration type
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Type_An_Enumeration_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition.Type_Kind

   begin
      return (Symtbl_Entry.Type_Definition.Type_Kind = Enumeration_Type);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Type_An_Enumeration_Type");
         raise Program_Error;

   end Is_Type_An_Enumeration_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_An_Integer_Type
   --
   --  <Description>
   --
   --    Returns whether the type entry defines an integer type
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Type_An_Integer_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition.Type_Kind

   begin
      return (Symtbl_Entry.Type_Definition.Type_Kind = Integer_Type);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Type_An_Integer_Type");
         raise Program_Error;

   end Is_Type_An_Integer_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_A_Real_Type
   --
   --  <Description>
   --
   --    Returns whether the type entry defines a real type
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Type_A_Real_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition.Type_Kind

   begin
      return (Symtbl_Entry.Type_Definition.Type_Kind = Real_Type);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Type_A_Real_Type");
         raise Program_Error;

   end Is_Type_A_Real_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_A_Record_Type
   --
   --  <Description>
   --
   --   Returns whether the type entry defines a record type
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Type_A_Record_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition.Type_Kind

   begin
      return (Symtbl_Entry.Type_Definition.Type_Kind = Record_Type);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Type_A_Record_Type");
         raise Program_Error;

   end Is_Type_A_Record_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_An_Array_Type
   --
   --  <Description>
   --
   --    Returns whether the type entry defines an array type
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Type_An_Array_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition.Type_Kind

   begin
      return (Symtbl_Entry.Type_Definition.Type_Kind = Array_Type);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Type_An_Array_Type");
         raise Program_Error;

   end Is_Type_An_Array_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Type_An_Access_Type
   --
   --  <Description>
   --
   --    Returns whether the type entry defines an access type
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Type_An_Access_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition.Type_Kind

   begin
      return (Symtbl_Entry.Type_Definition.Type_Kind = Access_Type);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Type_An_Access_Type");
         raise Program_Error;

   end Is_Type_An_Access_Type;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Type_Entry     :
   -- :   Component - Type_Definition :
   -- :              Enumeration_Type :
   -- :::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Literal_Count
   --
   --  <Description>
   --
   --    Return the number of enumeration literals associated with an
   --    enumeration type definition
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Literal_Count
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Enumeration_Type

   begin
      return Symtbl_Entry.Type_Definition.Literal_Count;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Literal_Count");
         raise Program_Error;

   end Get_Literal_Count;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_First_Literal
   --
   --  <Description>
   --
   --    Return the symbol table entry for the first enumeration literal
   --    associated with an enumeration type definition
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_First_Literal
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Enumeration_Type

   begin
      return Symtbl_Entry.Type_Definition.First_Literal;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_First_Literal");
         raise Program_Error;

   end Get_First_Literal;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Last_Literal
   --
   --  <Description>
   --
   --    Return the symbol table entry for the last enumeration literal
   --    associated with an enumeration type definition
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Last_Literal
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Enumeration_Type

   begin
      return Symtbl_Entry.Type_Definition.Last_Literal;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Last_Literal");
         raise Program_Error;

   end Get_Last_Literal;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Type_Entry     :
   -- :   Component - Type_Definition :
   -- :                Integer_Type   :
   -- :::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Lower_Bound
   --
   --  <Description>
   --
   --    Return the lower bound associated with an integer type definition
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Lower_Bound
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Integer is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Integer_Type

   begin
      return Symtbl_Entry.Type_Definition.Lower_Bound;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Lower_Bound");
         raise Program_Error;

   end Get_Lower_Bound;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Upper_Bound
   --
   --  <Description>
   --
   --    Return the upper bound associated with an integer type definition
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Upper_Bound
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Integer is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Integer_Type

   begin
      return Symtbl_Entry.Type_Definition.Upper_Bound;

   end Get_Upper_Bound;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Type_Entry     :
   -- :   Component - Type_Definition :
   -- :                Record_Type    :
   -- :::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Record_Component
   --
   --  <Description>
   --
   --    Return the first record component associated with a record type
   --    definition
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Record_Component
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Record_Type

   begin
      return Symtbl_Entry.Type_Definition.Record_Component;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Record_Component");
         raise Program_Error;

   end Get_Record_Component;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Type_Entry     :
   -- :   Component - Type_Definition :
   -- :                Array_Type     :
   -- :::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_Constrained
   --
   --  <Description>
   --
   --    Return whether the array is a constrained array for an array type
   --    definition
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Is_Constrained
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Array_Type

   begin
      return Symtbl_Entry.Type_Definition.Is_Constrained;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Is_Constrained");
         raise Program_Error;

   end Is_Constrained;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Index_Type
   --
   --  <Description>
   --
   --    Return the symbol table entry for the index associated with the
   --    array type definition
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Index_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database
       is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Array_Type

   begin
      return Convert_Fake (Symtbl_Entry.Type_Definition.Array_Index_Type);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Index_Type");
         raise Program_Error;

   end Get_Index_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Element_Type
   --
   --  <Description>
   --
   --    Return the symbol table entry for the element associated with the
   --    array type definition
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Element_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database
       is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Array_Type

   begin
      return Convert_Fake (Symtbl_Entry.Type_Definition.Array_Element_Type);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Element_Type");
         raise Program_Error;

   end Get_Element_Type;

   -- -------------------------------------------------------------------------

   -- ::::::::::::::::::::::::::
   -- : Constraint_Node fields :
   -- ::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Lower_Bound
   --
   --  <Description>
   --
   --    Return the lower bound associated with an array type definition,
   --    or a constraint definition
   --
   --    Note: a constraint definition may exist for an object, a subtype, a
   --    derived type, or a constraint being built during tree building.
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Lower_Bound
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database
       is

      -- Valid for: Constraint_Entry
      --   component ==> Constraint_Ref,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Object_Entry
      --   component ==> Object_Constraint,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Subtype_Entry, Derived_Type_Entry
      --   component ==> Constraint_Info,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint

   begin
      case Symtbl_Entry.Entry_Kind is

         when Constraint_Entry =>
            return (Convert_Fake (Symtbl_Entry.Constraint_Ref.Lower_Bound));

         when Object_Entry     =>
            return
               (Convert_Fake
                   (Symtbl_Entry.Object_Constraint.Constraint_Ref.Lower_Bound));

         when Subtype_Entry    | Derived_Type_Entry     =>
            return
               (Convert_Fake
                   (Symtbl_Entry.Constraint_Info.Constraint_Ref.Lower_Bound));

         when others =>
            Show_Error (2000, Internal, "Get_Lower_Bound_C1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Lower_Bound_C");
         raise Program_Error;

   end Get_Lower_Bound;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Lower_Bound_Value_Offset
   --
   --  <Description>
   --
   --    Return the storage management offset location in which the lower
   --    bound associated with an array type definition, or a constraint
   --    definition is stored
   --
   --    Note: a constraint definition may exist for an object, a subtype, a
   --          derived type, or a constraint being built during tree building.
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Lower_Bound_Value_Offset
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural is

      -- Valid for: Constraint_Entry
      --   component ==> Constraint_Ref,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Object_Entry
      --   component ==> Object_Constraint,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Subtype_Entry, Derived_Type_Entry
      --   component ==> Constraint_Info,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint

   begin
      case Symtbl_Entry.Entry_Kind is

         when Constraint_Entry =>
            return Symtbl_Entry.Constraint_Ref.Lower_Bound_Value_Offset;

         when Object_Entry     =>
            return
               Symtbl_Entry.Object_Constraint.Constraint_Ref.
                  Lower_Bound_Value_Offset;

         when Subtype_Entry    | Derived_Type_Entry     =>
            return
               Symtbl_Entry.Constraint_Info.Constraint_Ref.
                  Lower_Bound_Value_Offset;

         when others =>
            Show_Error (2000, Internal, "Get_Lower_Bound_Value_Offset_1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Lower_Bound_Value_Offset");
         raise Program_Error;

   end Get_Lower_Bound_Value_Offset;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Upper_Bound
   --
   --  <Description>
   --
   --    Return the upper bound associated with an array type definition,
   --    or a constraint definition
   --
   --   Note: a constraint definition may exist for an object, a subtype, a
   --         derived type, or a constraint being built during tree building.
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Upper_Bound
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Statement_Database
       is

      -- Valid for: Constraint_Entry
      --   component ==> Constraint_Ref,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Object_Entry
      --   component ==> Object_Constraint,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Subtype_Entry, Derived_Type_Entry
      --   component ==> Constraint_Info,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint

      -- The upper bound must be converted from a Ref_Statement_Database_Fake
      -- to a Ref_Statement_Database

   begin
      case Symtbl_Entry.Entry_Kind is

         when Constraint_Entry =>
            return (Convert_Fake (Symtbl_Entry.Constraint_Ref.Upper_Bound));

         when Object_Entry     =>
            return
               (Convert_Fake
                   (Symtbl_Entry.Object_Constraint.Constraint_Ref.Upper_Bound));

         when Subtype_Entry    | Derived_Type_Entry     =>
            return
               (Convert_Fake
                   (Symtbl_Entry.Constraint_Info.Constraint_Ref.Upper_Bound));

         when others =>
            Show_Error (2000, Internal, "Get_Upper_Bound_C1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Upper_Bound_C");
         raise Program_Error;

   end Get_Upper_Bound;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Upper_Bound_Value_Offset
   --
   --  <Description>
   --
   --    Return the storage management offset location in which the upper
   --    bound associated with an array type definition, or a constraint
   --    definition is stored
   --
   --    Note: a constraint definition may exist for an object, a subtype, a
   --          derived type, or a constraint being built during tree building.
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Upper_Bound_Value_Offset
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ace_Natural is

      -- Valid for: Constraint_Entry
      --   component ==> Constraint_Ref,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Object_Entry
      --   component ==> Object_Constraint,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Subtype_Entry, Derived_Type_Entry
      --   component ==> Constraint_Info,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint

   begin
      case Symtbl_Entry.Entry_Kind is

         when Constraint_Entry =>
            return Symtbl_Entry.Constraint_Ref.Upper_Bound_Value_Offset;

         when Object_Entry     =>
            return
               Symtbl_Entry.Object_Constraint.Constraint_Ref.
                  Upper_Bound_Value_Offset;

         when Subtype_Entry    | Derived_Type_Entry     =>
            return
               Symtbl_Entry.Constraint_Info.Constraint_Ref.
                  Upper_Bound_Value_Offset;

         when others =>
            Show_Error (2000, Internal, "Get_Upper_Bound_Value_Offset_1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Upper_Bound_Value_Offset");
         raise Program_Error;

   end Get_Upper_Bound_Value_Offset;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Next_Constraint
   --
   --  <Description>
   --
   --    Return the next constraint node associated with the index of an
   --    array
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Next_Constraint
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: Constraint_Entry
      --   component ==> Constraint_Ref,
      --                            where Constraint_Kind = Index_Constraint
      -- Valid for: Object_Entry
      --   component ==> Object_Constraint,
      --                            where Constraint_Kind = Index_Constraint

   begin
      case Symtbl_Entry.Entry_Kind is

         when Constraint_Entry =>
            return Symtbl_Entry.Constraint_Ref.Next_Constraint;

         when Object_Entry     =>
            return
               Symtbl_Entry.Object_Constraint.Constraint_Ref.Next_Constraint;

         when others =>
            Show_Error (2000, Internal, "Get_Next_Constraint_1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Next_Constraint");
         raise Program_Error;

   end Get_Next_Constraint;

   -- -------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Base_Type
   --
   --  <Description>
   --
   --    Returns the base type associated with a type, subtype, or derived
   --    type symbol table entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Base_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: Type_Entry, Subtype_Entry, and Derived_Type_Entry

   begin
      case Get_Entry_Kind (Symtbl_Entry) is

         when Type_Entry  | Derived_Type_Entry     =>
            return Symtbl_Entry;

         when Subtype_Entry    =>
            return (Get_Base_Type (Get_Entry_Base_Type (Symtbl_Entry)));

         when others =>
            Show_Error (2000, Internal, "Get_Base_Type1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Base_Type");
         raise Program_Error;

   end Get_Base_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Root_Type
   --
   --  <Description>
   --
   --    For a symbol table entry, follows base pointers until it reaches
   --    the root type entry
   --
   --  <Exceptions Raised>
   --
   --    Program_Error
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   function Get_Root_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Symbol_Table_Entry
       is

      -- Valid for: Type_Entry, Subtype_Entry, and Derived_Type_Entry

   begin
      case Get_Entry_Kind (Symtbl_Entry) is

         when Type_Entry  =>
            return Symtbl_Entry;

         when Subtype_Entry    | Derived_Type_Entry     =>
            return (Get_Root_Type (Get_Entry_Base_Type (Symtbl_Entry)));

         when others =>
            Show_Error (2000, Internal, "Get_Root_Type1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Get_Root_Type");
         raise Program_Error;

   end Get_Root_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Is_A_Derived_Subprogram
   --
   --  <Description>
   --
   --    This procedure determines whether a symbol table entry is a derived
   --    subprogram.
   --    Return True if the symbol table entry is a derived subprogram;
   --    otherwise returns False.
   --
   --==========================================================================
   function Is_A_Derived_Subprogram
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Boolean is

   begin
      case Get_Entry_Kind (Symtbl_Entry) is
         when Procedure_Entry  | Function_Entry    =>
            return
               (not Is_Empty (Get_Derived_Subprogram_Parent (Symtbl_Entry)));

         when others =>
            return False;
      end case;

   end Is_A_Derived_Subprogram;

   -- -------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Help_List
   --
   --  <Description>
   --
   --    Get_Help_List returns the pointer from the symbol table entry to the
   --  help comment.
   --
   --==========================================================================
   function Get_Help_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) return Ref_Help is
   begin
      case Get_Entry_Kind (Symtbl_Entry) is
         when Package_Entry    =>
            return Symtbl_Entry.Pkg_Help;
         when Procedure_Entry  | Function_Entry    =>
            return Symtbl_Entry.Subprg_Help;
         when Type_Entry  =>
            return Symtbl_Entry.Type_Help;
         when Subtype_Entry    | Derived_Type_Entry     =>
            return Symtbl_Entry.Other_Type_Help;
         when others =>
            return null;
            --Show_Error (1001, Internal, "Get_Help_List");
      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "get_help_list1");
         raise ;
   end Get_Help_List;

end Symtbl_Entry_Get_Routines;
