--|$Log:	strt.a,v $
--|Revision 8.0  91/01/17  10:46:29  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Symbol_Table_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    The symbol table manager constructs and manages the information
--    associated with identifiers within the Ada statements given to the
--    Ada command language interpreter.  The symbol table manager consists of
--    a set of data types (the package Symbol_Table_Structures) and a
--    set of operations (this package: Symbol_Table_Routines).  The other
--    components of the interpreter do not directly manipulate the symbol
--    table; the symbol table routines perform the manipulation.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Symtbl_Entry_Set_Routines;
use Symtbl_Entry_Set_Routines;

with Symtbl_Entry_Get_Routines;
use Symtbl_Entry_Get_Routines;

with Storage_Management_Routines;
use Storage_Management_Routines;

with Misc_Support_Routines;
use Misc_Support_Routines;

with Declarative_Region_Support_Routines;
use Declarative_Region_Support_Routines;

with Declaration_Support_Routines;
use Declaration_Support_Routines;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Literals;
use Literals;

with Error_Handler;
use Error_Handler;

with Text_Io;
use Text_Io;

package body Symbol_Table_Routines is

   -- --------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Symbol_Table_Entry
   --
   --  <Description>
   --
   --    Create_Entry creates a symbol table entry for the specified Name
   --    using the Entry_Type given.  A reference to the symbol table entry
   --    created is returned as the result.
   --
   --==========================================================================
   function Create_Symbol_Table_Entry
       (Name       : in Ref_String;
        Entry_Type : in Type_Entry_Enumeration) return Ref_Symbol_Table_Entry
       is

      New_Symbol_Table_Entry : Ref_Symbol_Table_Entry;

   begin
      New_Symbol_Table_Entry := new Symbol_Table_Entry (Entry_Type);
      Set_Symbolic_Name (New_Symbol_Table_Entry, Name);
      return (New_Symbol_Table_Entry);

   end Create_Symbol_Table_Entry;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Duplicate_Symbol_Table_Entry
   --
   --  <Description>
   --
   --    Duplicate_Entry creates a new symbol table entry for the specified
   --    Name using the template original symbol table entry.  The appropriate
   --    information from the original symbol table entry is copied to the new
   --    symbol table entry, and unique components for the new symbol table
   --    entry are initialized as required.  The new entry is returned.
   --
   --==========================================================================
   function Duplicate_Symbol_Table_Entry
       (Original_Entry : in Ref_Symbol_Table_Entry;
        Name           : in Ref_String) return Ref_Symbol_Table_Entry is

      -- WARNING:
      -- This routine is implemented to only handle "stupid" duplication of
      -- symbol table entries.

      -- Note:
      -- This routine is only currently used for Object_Entry symbol table
      -- entries.  This is the only item that contains an "identifier_list"
      -- where one symbol table entry is constructed from the first.

      Duplicate_Entry : Ref_Symbol_Table_Entry :=
         Create_Symbol_Table_Entry (Name, Get_Entry_Kind (Original_Entry));

   begin
      Duplicate_Entry.all := Original_Entry.all;
      Set_Symbolic_Name (Duplicate_Entry, Name);
      return (Duplicate_Entry);

   end Duplicate_Symbol_Table_Entry;

   -- --------------------------
   -- START INTERNAL ROUTINES --
   -- --------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Hide_Matching_Items
   --
   --  <Description>
   --
   --    This procedure hides the appropriate conflicting symbol table entries.
   --    If the matching entry exists, then the symbol table entry must be
   --    hidden.  Also, the matching entry might have to be hidden
   --    if it resides within a package.
   --    If desired, warning messages could be printed for the items that are
   --    no longer directly visible.  Currently, messages are commented out.
   --
   --  <Called By>
   --
   --    Hide_Conflicting_Entries
   --
   --==========================================================================
   procedure Hide_Matching_Items
       (Context        : in out Context_Control.Context_Definition;
        New_Entry      : in Ref_Symbol_Table_Entry;
        Matching_Entry : in Ref_Symbol_Table_Entry) is

   begin
      -- A matching entry has been found.  Hide the newly added symbol
      -- table entry if it is being introduced by a use statement.
      -- (i.e. it has a completely declared enclosing parent package)
      if (not Is_Empty (Get_Enclosing_Package (New_Entry))) and then
         Get_Enclosing_Package (New_Entry) /=
         Get_Declarative_Region_Entry (Context) then
         Make_Entry_Hidden (New_Entry);
         Push_Hidden_Entry (Context, New_Entry);
         --W warning to user that item is not directly visible
         -- At this time, do not display warning message
         --    Show_Error (2015, Warning,
         --                Convert_To_String (Get_Symbolic_Name (New_Entry)));
      end if;

      -- Also, hide matching entry if it was introduced by a use statement,
      -- and is not already hidden.
      if (Is_Visible (Matching_Entry) and
         ((not Is_Empty (Get_Enclosing_Package (Matching_Entry))) and then
         (Get_Enclosing_Package (Matching_Entry) /=
         Get_Declarative_Region_Entry (Context)))) then
         Make_Entry_Hidden (Matching_Entry);
         Push_Hidden_Entry (Context, Matching_Entry);
         --W warning that previously visible item is no longer directly
         --  visible
         -- At this time, do not display warning message
         --    Show_Error (2016, Warning,
         --             Convert_To_String (Get_Symbolic_Name (Matching_Entry)));

         -- If the matching entry is a derived subprogram that should be hidden
         -- due to the explicit declaration of an object by the same name
         -- within the same package in which the subprogram was derived,
         -- the previous section of code will not hide it.  Also, the derived
         -- subprogram will be made permanently hidden.  (The item being hidden
         -- will not be reported to "Push_Hidden_Entry".
         -- Condition: ("visible" and "derived_subprogram") and
         --         "still within package in which derived_subprogram declared"
      elsif (Is_Visible (Matching_Entry) and
            Is_A_Derived_Subprogram (Matching_Entry)) and
            (not Is_Empty (Get_Enclosing_Package (Matching_Entry)) and then
            (Get_Enclosing_Package (Matching_Entry) =
            Get_Declarative_Region_Entry (Context))) then
         Make_Entry_Hidden (Matching_Entry);

      end if;

   end Hide_Matching_Items;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Unhide_Entry
   --
   --  <Description>
   --
   --    This procedure will traverse the symbol table and look for hidden
   --    subprograms with the same provie as the Symtbl_Entry parameter. If
   --    only one is found it will be unhidden, if more than one is found,
   --    it they will remain obscured.
   --
   --==========================================================================
   procedure Unhide_Entry (Context      : in Context_Control.Context_Definition;
                           Symtbl_Entry : in Ref_Symbol_Table_Entry) is
      Matching_Entry, Next_Matching_Entry : Ref_Symbol_Table_Entry;
   begin

      Matching_Entry := Find_Matching_Name (Context,
                                            Get_Symbolic_Name (Symtbl_Entry),
                                            Get_Entry_Kind (Symtbl_Entry));

      if not (Is_Empty (Matching_Entry)) then
         if (Matching_Entry = Symtbl_Entry or else
            (Is_Removed (Matching_Entry)) or else
            (not Subprogram_Profiles_Match (Matching_Entry, Symtbl_Entry)))
             then
            loop
               Matching_Entry :=
                  Find_Next_Matching_Name (Matching_Entry,
                                           Get_Entry_Kind (Symtbl_Entry));
               exit  when Is_Empty (Matching_Entry) or else
                          (not Is_Removed (Matching_Entry) and
                          Subprogram_Profiles_Match (Matching_Entry,
                                                     Symtbl_Entry));
            end loop;
         end if;

         if not (Is_Empty (Matching_Entry)) then
            Next_Matching_Entry := Matching_Entry;
            loop
               Next_Matching_Entry :=
                  Find_Next_Matching_Name (Next_Matching_Entry,
                                           Get_Entry_Kind (Symtbl_Entry));
               exit  when Is_Empty (Next_Matching_Entry) or else
                          (not (Is_Removed (Next_Matching_Entry) or
                          Next_Matching_Entry = Symtbl_Entry) and
                          Subprogram_Profiles_Match (Next_Matching_Entry,
                                                     Symtbl_Entry));
            end loop;

            if (Is_Empty (Next_Matching_Entry)) then
               -- there are no other matching names, so
               Make_Entry_Visible (Matching_Entry);
            end if;
         end if;
      end if;
   end Unhide_Entry;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Hide_Conflicting_Entries
   --
   --  <Description>
   --
   --    This procedure determines whether there are conflicting entries
   --    within the symbol table, and hides the appropriate symbol table
   --    entries.
   --
   --  <Called By>
   --
   --    Add_Entry_To_Symbol_Table
   --
   --==========================================================================
   procedure Hide_Conflicting_Entries
       (Context   : in out Context_Control.Context_Definition;
        New_Entry : in Ref_Symbol_Table_Entry) is

      Matching_Entry : Ref_Symbol_Table_Entry;

   begin
      -- The Symtbl_Entry has already been added onto the symbol table.
      -- Find the next entry after symtbl_entry that has a matching name.
      -- Additional matching must be performed for enumeration literals
      -- and subprograms

      Matching_Entry := Find_Next_Matching_Name (New_Entry);

      if not Is_Empty (Matching_Entry) and then Is_Removed (Matching_Entry)
          then
         loop
            Matching_Entry := Find_Next_Matching_Name (Matching_Entry);
            exit  when Is_Empty (Matching_Entry) or else
                       not Is_Removed (Matching_Entry);
         end loop;
      end if;

      case Get_Entry_Kind (New_Entry) is
         when Object_Entry       |
              Formal_Param_Entry |
              Type_Entry         |
              Subtype_Entry      |
              Derived_Type_Entry |
              Package_Entry    =>
            if not Is_Empty (Matching_Entry) then
               Hide_Matching_Items (Context, New_Entry, Matching_Entry);
            end if;

         when Enumeration_Literal_Entry  =>
            if not Is_Empty (Matching_Entry) and then
               Get_Entry_Kind (Matching_Entry) /= Enumeration_Literal_Entry
                then
               Hide_Matching_Items (Context, New_Entry, Matching_Entry);
            end if;

         when Procedure_Entry  | Function_Entry    =>
            while not Is_Empty (Matching_Entry) loop
               case Get_Entry_Kind (Matching_Entry) is

                  -- Determine whether matching item is a subprogram. If not,
                  -- hide and exit loop.  If so, then hide if the subprogram
                  -- profiles match.  Otherwise, continue looking for matching
                  -- names.
                  when Procedure_Entry   | Function_Entry    =>
                     if Subprogram_Profiles_Match (New_Entry, Matching_Entry)
                         then
                        Hide_Matching_Items
                            (Context, New_Entry, Matching_Entry);
                        exit ;
                     end if;

                  when others  =>
                     -- Hide the item and exit the loop
                     Hide_Matching_Items (Context, New_Entry, Matching_Entry);
                     exit ;

               end case;

               Matching_Entry := Find_Next_Matching_Name (Matching_Entry);
            end loop;

         when others =>
            null;

      end case;

   end Hide_Conflicting_Entries;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Add_Entry_To_Symbol_Table
   --
   --  <Description>
   --
   --
   --  <Called By>
   --
   --    Internal
   --
   --  <Exceptions Raised>
   --
   --
   --==========================================================================
   procedure Add_Entry_To_Symbol_Table_Level_Zero
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

   begin
      null;
   end Add_Entry_To_Symbol_Table_Level_Zero;

   -- ------------------------
   -- END INTERNAL ROUTINES --
   -- ------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Add_Entry_To_Symbol_Table
   --
   --  <Description>
   --
   --    Add_Entry_To_Symbol_Table inserts the specified symbol table entry and
   --    any subordinate entry associated with the entry into the symbol table.
   --    The addition of this entry may cause other entries to no longer be
   --    directly visible.  Those entries are marked as hidden.
   --
   --  <Exceptions Raised>
   --
   --    Duplicate_Symbol_Table_Entry_Error is raised if the symbol table
   --      entry already exists within the symbol table.
   --
   --==========================================================================
   procedure Add_Entry_To_Symbol_Table
       (Context      : in out Context_Definition;
        Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Notes:
      -- Currently, the only symbol table entry where subordinate entries must
      -- also be inserted is:
      --   Type_Entry, for an Enumeration_Type:
      --                               enumeration literals must be linked in
      --

      Current_Entry : Ref_Symbol_Table_Entry := Get_Top_Of_Symtbl (Context);

   begin

      -- Check that this entry does not already exist on the symbol table
      while (not Is_Empty (Current_Entry)) loop
         if Current_Entry = Symtbl_Entry then
            raise Duplicate_Symbol_Table_Entry_Error;
         end if;
         Current_Entry := Get_Next_Entry (Current_Entry);
      end loop;

      -- Link the symbol table entry onto the head of the symbol table
      Set_Next_Entry (Symtbl_Entry, Get_Top_Of_Symtbl (Context));
      Set_Top_Of_Symtbl (Context, Symtbl_Entry);

      -- do not set the saved_main in this routine because
      -- an error in a package declaration will result in
      -- an empty package being declared.  The saved_main should
      -- only be set at the point where a prompt would be
      -- displayed.

      -- if entering at level 0,
      -- then also (re)set Saved_Main_Level_Symbol_Table_Head
      -- if (Context_Control.Get_Current_Level (Context) = Ace_Max_Levels'FIRST)
      --    then
      --   Set_Saved_Main_Level_Symbol_Table_Head (Context,
      --                                           Get_Top_Of_Symtbl (Context));
      -- end if;

      -- If conflicts now exist, hide the appropriate entries.
      Hide_Conflicting_Entries (Context, Symtbl_Entry);

      -- For an enumeration type entry, check if conflicts arise.
      if Get_Entry_Kind (Symtbl_Entry) = Type_Entry and then
         Is_Type_An_Enumeration_Type (Symtbl_Entry) then
         declare
            Literal_Entry : Ref_Symbol_Table_Entry;
         begin
            Literal_Entry := Get_First_Literal (Symtbl_Entry);
            while Literal_Entry /= null loop

               -- add the substituted parameterless function
               --   symtbl entry instead
               Set_Next_Entry (Get_Function_Entry (Literal_Entry),
                               Get_Top_Of_Symtbl (Context));
               Set_Top_Of_Symtbl (Context, Get_Function_Entry (Literal_Entry));

               Hide_Conflicting_Entries (Context, Literal_Entry);

               Literal_Entry := Get_Successor (Literal_Entry);
            end loop;
         end;
      end if;

   end Add_Entry_To_Symbol_Table;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Add_Duplicate_Of_Entry_To_Symbol_Table
   --
   --  <Description>
   --
   --    Add_Entry_To_Symbol_Table inserts the specified symbol table entry and
   --    any subordinate entry associated with the entry into the symbol table.
   --    The addition of this entry may cause other entries to no longer be
   --    directly visible.  Those entries are marked as hidden.
   --
   --==========================================================================
   procedure Add_Duplicate_Of_Entry_To_Symbol_Table
       (Context      : in out Context_Definition;
        Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Notes:
      -- [TBD] fix hide conflicting entries later

      Current_Entry   : Ref_Symbol_Table_Entry := Get_Top_Of_Symtbl (Context);
      Duplicate_Entry : Ref_Symbol_Table_Entry :=
         Duplicate_Symbol_Table_Entry (Symtbl_Entry,
                                       Get_Symbolic_Name (Symtbl_Entry));

   begin
      -- Link the symbol table entry onto the head of the symbol table
      Set_Next_Entry (Duplicate_Entry, Get_Top_Of_Symtbl (Context));
      Set_Top_Of_Symtbl (Context, Duplicate_Entry);

   end Add_Duplicate_Of_Entry_To_Symbol_Table;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Symbol_Table_Entry
   --
   --  <Description>
   --
   --    Find_Symbol_Table_Entry searches the symbol table for the given
   --    name. The first occurrence found within the table is returned;
   --    otherwise null is returned.  Find_Symbol_Table_Entry invoked with
   --    the type of the name (Entry_Type) searches only for the Name
   --    defined by the specified type.
   --
   --==========================================================================
   function Find_Symbol_Table_Entry (Context    : in Context_Definition;
                                     Name       : in Ref_String;
                                     Entry_Type : in Type_Entry_Enumeration :=
      Unknown) return Ref_Symbol_Table_Entry is

      -- Notes:
      -- Only visible names (not hidden or removed) are looked at.

      Current : Ref_Symbol_Table_Entry;

   begin
      -- Search through the symbol table for the given name.  Return either the
      -- symbol table entry for the name (if found), or null otherwise.
      Current := Get_Top_Of_Symtbl (Context);
      while not Is_Empty (Current) loop
         if Is_Visible (Current) then
            if Entry_Type = Unknown then
               exit  when
                  Are_Symbolic_Names_Equal (Name, Get_Symbolic_Name (Current));
            else
               exit  when
                  Are_Symbolic_Names_Equal (Name, Get_Symbolic_Name (Current))
                   and Entry_Type = Get_Entry_Kind (Current);
            end if;
         end if;
         Current := Get_Next_Entry (Current);
      end loop;

      return (Current);

   end Find_Symbol_Table_Entry;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Next_Symbol_Table_Entry
   --
   --  <Description>
   --
   --    Find_Next_Symbol_Table_Entry searches the symbol table for the
   --    next occurrence of the name, starting from the specified entry.
   --    The next occurrence of the name found within the symbol table is
   --    returned; otherwise null is returned.
   --    Find_Next_Symbol_Table_Entry invoked with the type of the name
   --    (Entry_Type) searches only for the Name defined by the specified
   --    type.
   --
   --==========================================================================
   function Find_Next_Symbol_Table_Entry
       (Reference_Entry : in Ref_Symbol_Table_Entry;
        Entry_Type      : in Type_Entry_Enumeration := Unknown) return
         Ref_Symbol_Table_Entry is

      -- Notes:
      -- Only visible names (not hidden or removed) are looked at.

      Current : Ref_Symbol_Table_Entry := Reference_Entry;
      Name    : Ref_String;

   begin
      if Is_Empty (Current) then
         return (null);
      else
         Name := Get_Symbolic_Name (Reference_Entry);
         Current := Get_Next_Entry (Current);
      end if;

      while not Is_Empty (Current) loop
         if Is_Visible (Current) then
            if Entry_Type = Unknown then
               exit  when
                  Are_Symbolic_Names_Equal (Name, Get_Symbolic_Name (Current));
            else
               exit  when
                  Are_Symbolic_Names_Equal (Name, Get_Symbolic_Name (Current))
                   and Entry_Type = Get_Entry_Kind (Current);
            end if;
         end if;

         Current := Get_Next_Entry (Current);
      end loop;

      return (Current);

   end Find_Next_Symbol_Table_Entry;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Entry_Within_Scope
   --
   --  <Description>
   --
   --    Find_Entry_Within_Scope searches a portion of the symbol table for
   --    the given name.  The portion of the symbol table to be searched
   --    is determined by whether parsing is being conducted within a
   --    declarative region.  Only entries defined directly within the
   --    declarative region are considered (i.e. entries introduced by use
   --    statements are ignored.)  The first occurrence found within the table
   --    is returned; otherwise null is returned.
   --    Find_Entry_Within_Scope invoked with the type of the name (Entry_Type)
   --    searches only for the Name defined by the specified type.
   --
   --
   --==========================================================================
   function Find_Entry_Within_Scope (Context    : in Context_Definition;
                                     Name       : in Ref_String;
                                     Entry_Type : in Type_Entry_Enumeration :=
      Unknown) return Ref_Symbol_Table_Entry is

      -- Notes:
      -- Only visible names (not hidden or removed) are looked at.

      Current           : Ref_Symbol_Table_Entry;
      Search_Terminator : Ref_Symbol_Table_Entry :=
         Get_Declarative_Region_Mark (Context);
      --    Acquire scope information for this search, based upon the
      --    declarative region information.  No open declarative region is
      --    returned as null (the end of the list, which is also the end of
      --    search condition).

   begin
      -- Search through the symbol table for the given name.  Return either the
      -- symbol table entry for the name (if found), or null otherwise.
      -- Ignore hidden and USE introduced entries.
      -- USE introduced entries are detected by comparing their enclosing
      -- package with the current declarative region.
      Current := Get_Top_Of_Symtbl (Context);
      while Current /= Search_Terminator loop
         if (Is_Visible (Current) and
            (Is_Empty (Get_Enclosing_Package (Current)) or else
            Get_Enclosing_Package (Current) =
            Get_Declarative_Region_Entry (Context))) and then
            (Are_Symbolic_Names_Equal (Name, Get_Symbolic_Name (Current)) and
            (Entry_Type = Unknown or Entry_Type = Get_Entry_Kind (Current)))
             then
            return (Current);
         else
            Current := Get_Next_Entry (Current);
         end if;
      end loop;

      -- To reach the point, Current must be outside of the current scope and
      -- therefore the name has not been found.
      return (null);

   end Find_Entry_Within_Scope;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Next_Entry_Within_Scope
   --
   --  <Description>
   --
   --    Find_Next_Entry_Within_Scope searches a portion of the symbol table
   --    for the next occurrence of the given name, starting from the specified
   --    entry.  The portion of the symbol table to be searched is determined
   --    by whether parsing is being conducted within a declarative region.
   --    Only entries defined directly within the declarative region are
   --    considered (i.e. entries introduced by use statements are ignored.)
   --    The first occurrence found within the table is returned;
   --    otherwise null is returned.
   --    Find_Next_Entry_Within_Scope invoked with the type of the name
   --    (Entry_Type) searches only for the name defined by the specified type.
   --
   --==========================================================================
   function Find_Next_Entry_Within_Scope
       (Context         : in Context_Control.Context_Definition;
        Reference_Entry : in Ref_Symbol_Table_Entry;
        Entry_Type      : in Type_Entry_Enumeration := Unknown) return
         Ref_Symbol_Table_Entry is

      -- Notes:
      -- Only visible names (not hidden or removed) are looked at.

      Current           : Ref_Symbol_Table_Entry := Reference_Entry;
      Search_Terminator : Ref_Symbol_Table_Entry :=
         Get_Declarative_Region_Mark (Context);
      --    Acquire scope information for this search, based upon the
      --    declarative region information.  No open declarative region is
      --    returned as null (the end of the list, which is also the end of
      --    search condition).

      Name              : Ref_String;

   begin
      -- Make sure the reference point is not outside of the symbol table
      if Is_Empty (Current) then
         return (null);
      else
         Name := Get_Symbolic_Name (Reference_Entry);
         Current := Get_Next_Entry (Current);
      end if;

      -- Search through the symbol table for the given name.  Return either the
      -- symbol table entry for the name (if found), or null otherwise.
      -- USE introduced entries are detected by comparing their enclosing
      -- package with the current declarative region.
      while Current /= Search_Terminator loop
         if (Is_Visible (Current) and
            (Is_Empty (Get_Enclosing_Package (Current)) or else
            Get_Enclosing_Package (Current) =
            Get_Declarative_Region_Entry (Context))) and then
            (Are_Symbolic_Names_Equal (Name, Get_Symbolic_Name (Current)) and
            (Entry_Type = Unknown or Entry_Type = Get_Entry_Kind (Current)))
             then
            return (Current);
         else
            Current := Get_Next_Entry (Current);
         end if;
      end loop;

      -- To reach the point, Current must be outside of the current scope and
      -- therefore the name has not been found.
      return (null);

   end Find_Next_Entry_Within_Scope;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Matching_Name
   --
   --  <Description>
   --
   --    Find_Matching_Name searches the symbol table for the given
   --    name.  This routine will look at ALL names within the symbol table,
   --    even those that are not directly visible (i.e. hidden by the USE of
   --    packages).  The first occurrence found within the table is returned;
   --    otherwise null is returned.  Find_Matching_Name invoked with
   --    the type of the name (Entry_Type) searches only for the Name
   --    defined by the specified type.
   --
   --==========================================================================
   function Find_Matching_Name (Context    : in Context_Definition;
                                Name       : in Ref_String;
                                Entry_Type : in Type_Entry_Enumeration :=
      Unknown) return Ref_Symbol_Table_Entry is

      -- Notes:
      -- All names (visible and hidden) are looked at.

      Current : Ref_Symbol_Table_Entry;

   begin
      -- Search through the symbol table for the given name.  Return either the
      -- symbol table entry for the name (if found), or null otherwise.
      Current := Get_Top_Of_Symtbl (Context);
      while not Is_Empty (Current) loop
         if Entry_Type = Unknown then
            exit  when Are_Symbolic_Names_Equal (Name,
                                                 Get_Symbolic_Name (Current));
         else
            exit  when
               Are_Symbolic_Names_Equal (Name, Get_Symbolic_Name (Current))
                and Entry_Type = Get_Entry_Kind (Current);
         end if;
         Current := Get_Next_Entry (Current);
      end loop;

      return (Current);

   end Find_Matching_Name;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Find_Next_Matching_Name
   --
   --  <Description>
   --
   --    Find_Next_Matching_Name searches the symbol table for the
   --    next occurrence of the name, starting from the specified entry.
   --    This routine will look at ALL names within the symbol table,
   --    even those that are not directly visible (i.e. hidden by the USE of
   --    packages).  The next occurrence found within the table is returned;
   --    otherwise null is returned.  Find_Next_Matching_Name invoked with
   --    the type of the name (Entry_Type) searches only for the Name
   --    defined by the specified type.
   --
   --==========================================================================
   function Find_Next_Matching_Name
       (Reference_Entry : in Ref_Symbol_Table_Entry;
        Entry_Type      : in Type_Entry_Enumeration := Unknown) return
         Ref_Symbol_Table_Entry is

      -- Notes:
      -- All names (visible and hidden) are looked at.

      Current : Ref_Symbol_Table_Entry := Reference_Entry;
      Name    : Ref_String;

   begin
      if Is_Empty (Current) then
         return (null);
      else
         Name := Get_Symbolic_Name (Reference_Entry);
         Current := Get_Next_Entry (Current);
      end if;

      while not Is_Empty (Current) loop
         if Entry_Type = Unknown then
            exit  when Are_Symbolic_Names_Equal (Name,
                                                 Get_Symbolic_Name (Current));
         else
            exit  when
               Are_Symbolic_Names_Equal (Name, Get_Symbolic_Name (Current))
                and Entry_Type = Get_Entry_Kind (Current);
         end if;

         Current := Get_Next_Entry (Current);
      end loop;

      return (Current);

   end Find_Next_Matching_Name;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Enum_Node
   --
   --  <Description>
   --
   --     Get_Enum_Node will return the Ref_Symbol_Table_Entry of the
   --    enumeration literal node of type "enum_type"  whose literal value is "
   --    enum_literal_value"
   --
   --==========================================================================
   function Get_Enum_Node
       (Enum_Type          : in Ref_Symbol_Table_Entry;
        Enum_Literal_Value : in Ace_Integer) return Ref_Symbol_Table_Entry is

      Literal : Ref_Symbol_Table_Entry;
   begin
      Literal := Get_First_Literal (Enum_Type);
      while not Is_Empty (Literal) loop
         if (Get_Literal_Value (Literal) = Enum_Literal_Value) then
            return (Literal);
         else
            Literal := Get_Successor (Literal);
         end if;
      end loop;
      return null;
   end Get_Enum_Node;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Delete_Entry_From_Symbol_Table
   --
   --  <Description>
   --
   --    Delete_Entry_From_Symbol_Table removes the specified symbol table
   --    entry from the symbol table if it currently is in the symbol table.
   --    The entry is actually deleted from the list, as opposed to having
   --    its "visibility" field marked as removed.
   --
   --==========================================================================
   procedure Delete_Entry_From_Symbol_Table
       (Context      : in out Context_Definition;
        Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      Previous_Symtbl_Entry : Ref_Symbol_Table_Entry;
      Current_Symtbl_Entry  : Ref_Symbol_Table_Entry;

   begin
      Previous_Symtbl_Entry := Get_Top_Of_Symtbl (Context);
      Current_Symtbl_Entry := Get_Top_Of_Symtbl (Context);

      -- Locate the symbol table entry in the symbol table
      while not Is_Empty (Current_Symtbl_Entry) and then
         Current_Symtbl_Entry /= Symtbl_Entry loop

         Previous_Symtbl_Entry := Current_Symtbl_Entry;
         Current_Symtbl_Entry := Get_Next_Entry (Current_Symtbl_Entry);
      end loop;

      -- If the entry was found, remove it from the symbol table
      if not Is_Empty (Current_Symtbl_Entry) then
         if Previous_Symtbl_Entry /= Get_Top_Of_Symtbl (Context) then
            Set_Next_Entry (Previous_Symtbl_Entry,
                            Get_Next_Entry (Symtbl_Entry));
         else
            Set_Top_Of_Symtbl (Context, Get_Next_Entry (Symtbl_Entry));
         end if;

         Unhide_Entry (Context, Current_Symtbl_Entry);
         Make_Entry_Removed (Current_Symtbl_Entry); --raf
      end if;

   end Delete_Entry_From_Symbol_Table;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Add_Package_Decls_To_Symbol_Table
   --
   --  <Description>
   --
   --    Adds declarations onto the symbol table so that they may be
   --    visible during parsing of the package body, or that the body
   --    declarations may be visible for debugging purposes.  When parsing
   --    the package body for the package specification, the objects that were
   --    made visible by use statements within the package specification must
   --    be made visible also for the package body.  The Process_Use_Stmts
   --    should be set to True when the package spec declarations are being
   --    added to the symbol table for the parsing of the package body.
   --
   --==========================================================================
   procedure Add_Package_Decls_To_Symbol_Table
       (Context              : in out Context_Control.Context_Definition;
        Package_Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Decl_Kind            : in Pkg_Decl_Kind := Pkg_Spec;
        Process_Use_Stmts    : in Boolean       := False) is separate;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Delete_Package_Decls_From_Symbol_Table
   --
   --  <Description>
   --
   --    Removes declarations from the symbol table associated with a
   --    package.  This, in essence, is an "undo" of executing an Ada
   --    use statement.
   --    Symbol table entries are actually unlinked from the symbol table.
   --
   --==========================================================================
   procedure Delete_Package_Decls_From_Symbol_Table
       (Context              : in out Context_Control.Context_Definition;
        Package_Symtbl_Entry : in Ref_Symbol_Table_Entry) is separate;

   procedure Initialize_Symbol_Table is separate;

end Symbol_Table_Routines;
