--|$Log:	stset.a,v $
--|Revision 8.0  91/01/17  10:46:30  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Symtbl_Entry_Set_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--
--    This package contains the set of routines that stores values into fields
--    of symbol table entries.  All storage of values into components of
--    symbol table entries should be made through the routines provided within
--    this package.  This package (in conjunction with a similar package for
--    retrieving from symbol table entry components) provides the mechanism for
--    hiding the makeup and names of symbol table entry components.  Changes to
--    the structure and component names for symbol table entries would only
--    affect this package of routines, rather than all the other components
--    that deal with symbol table information.
--
--    The routines within the package are ordered with respect to the
--    definition of components for a symbol table entry.  The routines that
--    store values into components that are common to all symbol table entries
--    come first, followed by components that are applicable to different
--    variants of the symbol table entry.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Statement_Database;
use Statement_Database;

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Storage_Management;
use Storage_Management;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Misc_Support_Routines;
use Misc_Support_Routines;

with Error_Handler;
use Error_Handler;

with Help_Pkg;
use Help_Pkg;
package body Symtbl_Entry_Set_Routines is

   -- ::::::::::::::::::::::
   -- : Entry_Kind ==> ALL :
   -- ::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Entry_Kind
   --
   --  <Description>
   --
   --    Set the kind of entry associated with the symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Entry_Kind (Symtbl_Entry : in out Ref_Symbol_Table_Entry;
                             Item         : in Type_Entry_Enumeration) is

      -- Valid for: ALL symbol table entries
      --   component ==> Entry_Kind (discriminant)

      -- While entry_kind is technically valid for all symbol table entries,
      -- only particular symbol table entry kinds may change from one kind to
      -- another.  See below for the pairings.

      New_Symtbl_Entry : Ref_Symbol_Table_Entry;

   begin
      -- If the entry kind of the symbol table entry matches the intended
      -- new entry kind, nothing needs to be done.
      if Symtbl_Entry.Entry_Kind = Item then
         return ;
      end if;

      -- Only particular sets of [From, To] entry kinds are handled. These are:
      --    from: Subtype_Entry    --> to: Derived_Type_Entry
      --    from: Subtype_Entry    --> to: Object_Entry
      --    from: Subtype_Entry    --> to: Formal_Param_Entry
      -- Since the entry kind is the discriminant of the symbol table entry
      -- record, the entire record must be modified.

      if Symtbl_Entry.Entry_Kind = Subtype_Entry and Item = Derived_Type_Entry
          then
         New_Symtbl_Entry :=
            new Symbol_Table_Entry'
               (Entry_Kind                    => Derived_Type_Entry,
                Next_Entry                    => Symtbl_Entry.Next_Entry,
                Symbol_Name                   => Symtbl_Entry.Symbol_Name,
                Item_Level                    => Symtbl_Entry.Item_Level,
                Enclosing_Package => Symtbl_Entry.Enclosing_Package,
                Visibility                    => Symtbl_Entry.Visibility,
                Statement_Number              => Symtbl_Entry.Statement_Number,
                Base_Type                     => Symtbl_Entry.Base_Type,
                Constraint_Info               => Symtbl_Entry.Constraint_Info,
                Other_Type_Help               => Symtbl_Entry.Other_Type_Help,
                Dtype_Derived_Subprogram_List =>
                   Symtbl_Entry.Dtype_Derived_Subprogram_List);

      elsif Symtbl_Entry.Entry_Kind = Subtype_Entry and Item = Object_Entry
             then
         New_Symtbl_Entry :=
            new Symbol_Table_Entry'
               (Entry_Kind          => Object_Entry,
                Next_Entry          => Symtbl_Entry.Next_Entry,
                Symbol_Name         => Symtbl_Entry.Symbol_Name,
                Item_Level          => Symtbl_Entry.Item_Level,
                Enclosing_Package   => Symtbl_Entry.Enclosing_Package,
                Visibility          => Symtbl_Entry.Visibility,
                Statement_Number    => Symtbl_Entry.Statement_Number,
                Object_Type         => Symtbl_Entry.Base_Type,
                Object_Constraint   => Symtbl_Entry.Constraint_Info,
                Object_Mode         => In_Out_Mode,
                Object_Value_Offset => 0);

      elsif Symtbl_Entry.Entry_Kind = Type_Entry and Item = Object_Entry
             then
         -- Only case of this is an anonymous array object decl.
         New_Symtbl_Entry :=
            new Symbol_Table_Entry'
               (Entry_Kind          => Object_Entry,
                Next_Entry          => Symtbl_Entry.Next_Entry,
                Symbol_Name         => Symtbl_Entry.Symbol_Name,
                Item_Level          => Symtbl_Entry.Item_Level,
                Enclosing_Package   => Symtbl_Entry.Enclosing_Package,
                Visibility          => Symtbl_Entry.Visibility,
                Statement_Number    => Symtbl_Entry.Statement_Number,
                Object_Type         => Symtbl_Entry,
                Object_Constraint   => null,
                Object_Mode         => In_Out_Mode,
                Object_Value_Offset => 0);

      elsif Symtbl_Entry.Entry_Kind = Subtype_Entry and
            Item = Formal_Param_Entry then
         New_Symtbl_Entry :=
            new Symbol_Table_Entry'
               (Entry_Kind         => Formal_Param_Entry,
                Next_Entry         => Symtbl_Entry.Next_Entry,
                Symbol_Name        => Symtbl_Entry.Symbol_Name,
                Item_Level         => Symtbl_Entry.Item_Level,
                Enclosing_Package  => Symtbl_Entry.Enclosing_Package,
                Visibility         => Symtbl_Entry.Visibility,
                Statement_Number   => Symtbl_Entry.Statement_Number,
                Param_Type         => Symtbl_Entry.Base_Type,
                Param_Mode         => In_Mode,
                Param_Init         => null,
                Next_Param         => null,
                Param_Value_Offset => 0,
                Preceded_By_Comma  => False);

      else
         Show_Error
             (2001, Internal, "Set_Entry_Kind - invalid [From, To] pair");

      end if;

      -- NOTE: the "old" symbol table entry could now be discarded

      Symtbl_Entry := New_Symtbl_Entry;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Entry_Kind");

   end Set_Entry_Kind;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Next_Entry
   --
   --  <Description>
   --
   --    Set the next entry field of the Symtbl_Entry node to Item
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Next_Entry (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                             Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: ALL symbol table entries
      --   component ==> Next_Entry

   begin
      Symtbl_Entry.Next_Entry := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Next_Entry");

   end Set_Next_Entry;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Symbolic_Name
   --
   --  <Description>
   --
   --    Set the symbolic name associated with the symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Symbolic_Name (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                Item         : in Ref_String) is

      -- Valid for: ALL symbol table entries
      --   component ==> Symbol_Name

   begin
      Symtbl_Entry.Symbol_Name := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Symbolic_Name");

   end Set_Symbolic_Name;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Level
   --
   --  <Description>
   --
   --    Set the static level associated with the symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Level (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                        Item         : in Ace_Natural) is

      -- Valid for: ALL symbol table entries
      --   component ==> Item_Level

   begin
      Symtbl_Entry.Item_Level := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Level");

   end Set_Level;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Enclosing_Package
   --
   --  <Description>
   --
   --    Set the enclosing package field of the Symtbl_Entry node to Item
   --    (which should be a package symbol table entry)
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Enclosing_Package (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                    Item         : in Ref_Symbol_Table_Entry)
                                   is

      -- Valid for: ALL symbol table entries
      --   component ==> Enclosing_Package

   begin
      Symtbl_Entry.Enclosing_Package := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Enclosing_Package");

   end Set_Enclosing_Package;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Make_Entry_Visible
   --
   --  <Description>
   --
   --    Set the visibility field of the symbol table entry such that the
   --    item is visible
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Make_Entry_Visible (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Valid for: ALL symbol table entries
      --   component ==> Visibility

   begin
      Symtbl_Entry.Visibility := Visible;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Make_Entry_Visible");

   end Make_Entry_Visible;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Make_Entry_Hidden
   --
   --  <Description>
   --
   --    Set the visibility field of the symbol table entry such that the
   --    item is hidden
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Make_Entry_Hidden (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Valid for: ALL symbol table entries
      --   component ==> Visibility

   begin
      Symtbl_Entry.Visibility := Hidden;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Make_Entry_Hidden");

   end Make_Entry_Hidden;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Make_Entry_Removed
   --
   --  <Description>
   --
   --    Set the visibility field of the symbol table entry such that the
   --    item is "removed" from the symbol table
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Make_Entry_Removed (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Valid for: ALL symbol table entries
      --   component ==> Visibility

   begin
      Symtbl_Entry.Visibility := Removed;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Make_Entry_Removed");

   end Make_Entry_Removed;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Statement_Number
   --
   --  <Description>
   --
   --    Set the statement number associated with the symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Statement_Number (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                   Item         : in Ace_Natural) is

      -- Valid for: ALL symbol table entries
      --   component ==> Statement_Number

   begin
      Symtbl_Entry.Statement_Number := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Statement_Number");

   end Set_Statement_Number;

   -- -------------------------------------------------------------------------

   -- ::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Package_Entry :
   -- ::::::::::::::::::::::::::::::::
   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Context_Clause_Body
   --
   --  <Description>
   --
   --    Set the reference to the statement database node associated with the
   --    context clause within the implemenation portion of the package
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Context_Clause_Body (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                      Item         : in Ref_Statement_Database)
                                     is
      -- Valid for: Package_Entry
      --   component ==> Package_Spec_Decls

   begin
      case Symtbl_Entry.Entry_Kind is
         when Package_Entry    =>
            Symtbl_Entry.Pkg_Body_Context_Clause := Convert_To_Fake (Item);
         when others =>
            Show_Error (2000, Internal, "Set_Context_Clause_Body 1");
            raise Program_Error;
      end case;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Context_Clause_Body 2");
   end Set_Context_Clause_Body;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Context_Clause
   --
   --  <Description>
   --
   --    Set the reference to the statement database node associated with the
   --    context clause within the specification portion of the package
   --    or subprogram.
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Context_Clause (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                 Item         : in Ref_Statement_Database) is
      -- Valid for: Package_Entry
      --   component ==> Package_Spec_Decls

   begin
      case Symtbl_Entry.Entry_Kind is
         when Package_Entry    =>
            Symtbl_Entry.Pkg_Spec_Context_Clause := Convert_To_Fake (Item);
         when Procedure_Entry  | Function_Entry    =>
            Symtbl_Entry.Subprg_Context_Clause := Convert_To_Fake (Item);
         when others =>
            Show_Error (2000, Internal, "Set_Context_Clause 1");
            raise Program_Error;
      end case;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Context_Clause 2");
   end Set_Context_Clause;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Package_Spec_Decls
   --
   --  <Description>
   --
   --    Set the reference to the statement database node associated with the
   --    declarations within the specification portion of the package
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Package_Spec_Decls (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                     Item         : in Ref_Statement_Database)
                                    is
      -- Valid for: Package_Entry
      --   component ==> Package_Spec_Decls

   begin
      Symtbl_Entry.Package_Spec_Decls := Convert_To_Fake (Item);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Package_Spec_Decls");

   end Set_Package_Spec_Decls;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Package_Body_Decls
   --
   --  <Description>
   --
   --    Set the reference to the statement database node associated with the
   --    declarations within the body  portion of the package
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Package_Body_Decls (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                     Item         : in Ref_Statement_Database)
                                    is
      -- Valid for: Package_Entry
      --   component ==> Package_Body_Decls

   begin
      Symtbl_Entry.Package_Body_Decls := Convert_To_Fake (Item);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Package_Body_Decls");

   end Set_Package_Body_Decls;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Has_Body
   --
   --  <Description>
   --
   --    Set the fact that the package has a body
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Has_Body (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Valid for: Package_Entry
      --    component ==> Has_Body

   begin
      Symtbl_Entry.Has_Body := True;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Has_Body");

   end Set_Has_Body;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Head_Of_Statements_List
   --
   --  <Description>
   --
   --    Set the head of the list of statements for the executable
   --    statements associated with a package symbol table entry, a
   --    subprogram symbol table entry, or a label symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Head_Of_Statements_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Statement_Database) is

      -- Valid for: Package_Entry
      --   component ==> Package_Stmts
      -- Valid for: Procedure_Entry, Function_Entry
      --   component ==> Statements_List
      -- Valid for: Label_Entry
      --   component ==> Label_Stmt

      -- The declarations_list component must be converted from a
      -- Ref_Statement_Database_Fake to a Ref_Statement_Database

   begin
      case Symtbl_Entry.Entry_Kind is
         when Procedure_Entry  | Function_Entry    =>
            Symtbl_Entry.Statements_List := Convert_To_Fake (Item);

         when Label_Entry =>
            Symtbl_Entry.Label_Stmt := Convert_To_Fake (Item);

         when Package_Entry    =>
            Symtbl_Entry.Package_Stmts := Convert_To_Fake (Item);

         when others =>
            Show_Error (2000, Internal, "Set_Head_Of_Statements_List1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Head_Of_Statements_List");

   end Set_Head_Of_Statements_List;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Derived_Subprogram_List
   --
   --  <Description>
   --
   --    Set the head of the list of derived subprograms that are
   --    associated with a package.  Also, for types and derived types,
   --    set the head of the list of subprograms which are derivable
   --    from this type (or derived type).
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Derived_Subprogram_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_List_Of_Subprograms) is

      -- Valid for: Package_Entry
      --   component ==> Pkg_Derived_Subprogram_List
      -- Valid for: Type_Entry
      --   component ==> Type_Derived_Subprogram_List
      -- Valid for: Derived_Type_Entry
      --   component ==> DType_Derived_Subprogram_List

   begin
      case Symtbl_Entry.Entry_Kind is
         when Package_Entry    =>
            Symtbl_Entry.Pkg_Derived_Subprogram_List := Item;

         when Type_Entry  =>
            Symtbl_Entry.Type_Derived_Subprogram_List := Item;

         when Derived_Type_Entry    =>
            Symtbl_Entry.Dtype_Derived_Subprogram_List := Item;

         when others =>
            Show_Error (2000, Internal, "Set_Derived_Subprogram_List1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Derived_Subprogram_List");

   end Set_Derived_Subprogram_List;

   -- -------------------------------------------------------------------------

   -- ::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Procedure_Entry :
   -- :                Function_Entry  :
   -- ::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Subprg_Has_Decl
   --
   --  <Description>
   --
   --    Sets flag to indicate subprogram has an explicit declaration
   --
   --  <Called By>
   --
   --    Build_Subprogram_Spec_Decl
   --
   --==========================================================================
   procedure Set_Subprg_Has_Decl (Symtbl_Entry : in Ref_Symbol_Table_Entry) is
   begin
      Symtbl_Entry.Has_Explicit_Decl := True;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Subprg_Has_Decl");
   end Set_Subprg_Has_Decl;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Parameter_Count
   --
   --  <Description>
   --
   --    Set the number of parameters associated with a subprogram
   --    symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Parameter_Count (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                  Item         : in Ace_Natural) is
      -- Valid for: Procedure_Entry, Function_Entry
      --    component ==> Parameter_Count
   begin
      Symtbl_Entry.Parameter_Count := Item;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Parameter_Count");
   end Set_Parameter_Count;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Number_Of_Default_Parameters
   --
   --  <Description>
   --
   --    Set the number of parameters associated with a subprogram
   --    symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Number_Of_Default_Parameters
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ace_Natural) is
      -- Valid for: Procedure_Entry, Function_Entry
      -- record component name: Default_Parameter_Count
   begin
      Symtbl_Entry.Default_Parameter_Count := Item;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Default_Parameter_Count");
   end Set_Number_Of_Default_Parameters;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Head_Of_Parameters_List
   --
   --  <Description>
   --
   --    Set the reference to the first formal parameter associated with
   --    a subprogram symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Head_Of_Parameters_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: Procedure_Entry, Function_Entry
      --   component ==> Head_Param_List

   begin
      Symtbl_Entry.Head_Param_List := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Head_Of_Parameters_List");

   end Set_Head_Of_Parameters_List;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Tail_Of_Parameters_List
   --
   --  <Description>
   --
   --   Set the reference to the last formal parameter associated with
   --   a subprogram symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Tail_Of_Parameters_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: Procedure_Entry, Function_Entry
      --   component ==> Tail_Param_List

   begin
      Symtbl_Entry.Tail_Param_List := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Tail_Of_Parameters_List");

   end Set_Tail_Of_Parameters_List;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Return_Type
   --
   --  <Description>
   --
   --    For a function subprogram, set the reference to the return
   --    type associated with a function subprogram symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Return_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                              Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: Function_Entry
      --   component ==> Return_Type

      -- Ignored for: Procedure_Entry
      --   component ==> Return_Type -- don't care, set it any way

   begin
      Symtbl_Entry.Return_Type := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Return_Type");

   end Set_Return_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Head_Of_Declarations_List
   --
   --  <Description>
   --
   --    Set the head of the list of statements for the declaration
   --    statements associated with a subprogram symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Head_Of_Declarations_List
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Statement_Database) is

      -- Valid for: Procedure_Entry, Function_Entry
      --   component ==> Declarations_List

      -- The declarations_list component must be converted from a
      -- Ref_Statement_Database to a Ref_Statement_Database_Fake

   begin
      Symtbl_Entry.Declarations_List := Convert_To_Fake (Item);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Head_Of_Declarations_List");

   end Set_Head_Of_Declarations_List;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Value_Offset
   --
   --  <Description>
   --
   --    Set the reference to the data storage location in which the value
   --    associated with the item (object, parameter, or function) is stored
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Value_Offset (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                               Item         : in Ace_Natural) is

      -- Valid for: Function_Entry
      --   component ==> Function_Value_Offset
      -- Valid for: Object_Entry
      --   component ==> Object_Value_Offset
      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Value_Offset

      -- Ignored for: Procedure_Entry
      --   component ==> Function_Value_Offset

   begin
      case Symtbl_Entry.Entry_Kind is
         when Function_Entry   =>
            Symtbl_Entry.Function_Value_Offset := Item;

         when Object_Entry     =>
            Symtbl_Entry.Object_Value_Offset := Item;

         when Formal_Param_Entry    =>
            Symtbl_Entry.Param_Value_Offset := Item;

         when Procedure_Entry  =>
            null; -- no reason to store this for a procedure

         when others =>
            Show_Error (2000, Internal, "Set_Value_Offset1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Value_Offset");

   end Set_Value_Offset;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_As_Built_In_Subprogram
   --
   --  <Description>
   --
   --    Associate an internal numeric identifier with an ACE built-in
   --    subprogram
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_As_Built_In_Subprogram
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ace_Natural) is

      -- Valid for: Procedure_Entry, Function_Entry
      --   component ==> Is_Built_In_Subprogram

   begin
      Symtbl_Entry.Is_Built_In_Subprogram := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_As_Built_In_Subprogram");

   end Set_As_Built_In_Subprogram;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Derived_Subprogram_Parent
   --
   --  <Description>
   --
   --    Set the reference to the parent of a derived subprogram.
   --    (A non-derived subprogram has a null parent.)
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Derived_Subprogram_Parent
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Symbol_Table_Entry) is
      -- Valid for: Procedure_Entry, Function_Entry
      --    component ==> Derived_Subprogram_Parent
   begin
      Symtbl_Entry.Derived_Subprogram_Parent := Item;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Derived_Subprogram_Parent");
   end Set_Derived_Subprogram_Parent;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Function_Entry
   --
   --  <Description>
   --
   --    Set the reference to the parameterless function call symtbl entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Function_Entry (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                 Item         : in Ref_Symbol_Table_Entry) is
      -- Valid for: Enumeration_Literal_Entry
      --    component ==> Func_Entry
   begin
      Symtbl_Entry.Func_Entry := Item;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Function_Entry");
   end Set_Function_Entry;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Enumeration_Literal_Entry
   --
   --  <Description>
   --
   --    Set the reference to the enumeration literal entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Enumeration_Literal_Entry
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Symbol_Table_Entry) is
      -- Valid for: Procedure_Entry, Function_Entry
      --    component ==> Enum_Lit_Entry
   begin
      Symtbl_Entry.Enum_Lit_Entry := Item;
   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Enumeration_Literal_Entry");
   end Set_Enumeration_Literal_Entry;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Object_Entry :
   -- :::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type
   --
   --  <Description>
   --
   --    Set the reference to the type symbol table entry associated with
   --    either an object, a parameter, an enumeration literal, or a record
   --    component
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                       Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: Object_Entry
      --   component ==> Object_Type
      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Type
      -- Valid for: Enumeration_Literal_Entry
      --   component ==> Enumeration_Type
      -- Valid for: Record_Component_Entry
      --   component ==> Component_Type

   begin
      case Symtbl_Entry.Entry_Kind is
         when Object_Entry     =>
            Symtbl_Entry.Object_Type := Item;

         when Formal_Param_Entry    =>
            Symtbl_Entry.Param_Type := Item;

         when Enumeration_Literal_Entry  =>
            Symtbl_Entry.Enumeration_Type := Item;

         when Record_Component_Entry     =>
            Symtbl_Entry.Component_Type := Item;

         when others =>
            Show_Error (2000, Internal, "Set_Type1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Type");

   end Set_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Constraint
   --
   --  <Description>
   --
   --    Set the constraint for the item (object, subtype or derived type) to
   --    refer to a constraint entry already constructed
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Constraint (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                             Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: Object_Entry
      --   component ==> Object_Constraint
      -- Valid for: Subtype_Entry, Derived_Type_Entry
      --   component ==> Constraint_Info

   begin
      case Symtbl_Entry.Entry_Kind is
         when Object_Entry     =>
            Symtbl_Entry.Object_Constraint := Item;

         when Subtype_Entry    | Derived_Type_Entry     =>
            Symtbl_Entry.Constraint_Info := Item;

         when others =>
            Show_Error (2000, Internal, "Set_Constraint1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Constraint");

   end Set_Constraint;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Mode_To_Default_In
   --
   --  <Description>
   --
   --    Set the mode for the item (object or parameter) to DEFAULT_IN
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Mode_To_Default_In (Symtbl_Entry : in Ref_Symbol_Table_Entry)
                                    is

      -- Valid for: Object_Entry
      --   component ==> Object_Mode
      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Mode

      -- Note: although valid for object entries, should only be used for
      -- formal parameters.

   begin
      case Symtbl_Entry.Entry_Kind is
         when Object_Entry     =>
            Symtbl_Entry.Object_Mode := Default_In_Mode;

         when Formal_Param_Entry    =>
            Symtbl_Entry.Param_Mode := Default_In_Mode;

         when others =>
            Show_Error (2000, Internal, "Set_Mode_To_Default_In1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Mode_To_Default_In");

   end Set_Mode_To_Default_In;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Mode_To_In
   --
   --  <Description>
   --
   --    Set the mode for the item (object or parameter) to IN
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Mode_To_In (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Valid for: Object_Entry
      --   component ==> Object_Mode
      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Mode

   begin
      case Symtbl_Entry.Entry_Kind is
         when Object_Entry     =>
            Symtbl_Entry.Object_Mode := In_Mode;

         when Formal_Param_Entry    =>
            Symtbl_Entry.Param_Mode := In_Mode;

         when others =>
            Show_Error (2000, Internal, "Set_Mode_To_In1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Mode_To_In");

   end Set_Mode_To_In;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Mode_To_Out
   --
   --  <Description>
   --
   --    Set the mode for the item (object or parameter) to OUT
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Mode_To_Out (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Valid for: Object_Entry
      --   component ==> Object_Mode
      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Mode

   begin
      case Symtbl_Entry.Entry_Kind is
         when Object_Entry     =>
            Symtbl_Entry.Object_Mode := Out_Mode;

         when Formal_Param_Entry    =>
            Symtbl_Entry.Param_Mode := Out_Mode;

         when others =>
            Show_Error (2000, Internal, "Set_Mode_To_Out1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Mode_To_Out");

   end Set_Mode_To_Out;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Mode_To_In_Out
   --
   --  <Description>
   --
   --    Set the mode for the item (object or parameter) to IN OUT
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Mode_To_In_Out (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Valid for: Object_Entry
      --   component ==> Object_Mode
      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Mode

   begin
      case Symtbl_Entry.Entry_Kind is
         when Object_Entry     =>
            Symtbl_Entry.Object_Mode := In_Out_Mode;

         when Formal_Param_Entry    =>
            Symtbl_Entry.Param_Mode := In_Out_Mode;

         when others =>
            Show_Error (2000, Internal, "Set_Mode_To_In_Out1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Mode_To_In_Out");

   end Set_Mode_To_In_Out;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Formal_Param_Entry :
   -- :::::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Initialization_Code
   --
   --  <Description>
   --
   --    Set the initialization code location for the formal parameter or
   --    the record component
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Initialization_Code (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                      Item         : in Ref_Statement_Database)
                                     is

      -- Valid for: Formal_Param_Entry
      --   component ==> Param_Init
      -- Valid for: Record_Component_Entry
      --   component ==> Component_Init

   begin
      case Symtbl_Entry.Entry_Kind is
         when Formal_Param_Entry    =>
            Symtbl_Entry.Param_Init := Convert_To_Fake (Item);

         when Record_Component_Entry     =>
            Symtbl_Entry.Component_Init := Convert_To_Fake (Item);

         when others =>
            Show_Error (2000, Internal, "Set_Initialization_Code1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Initialization_Code");

   end Set_Initialization_Code;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Next_Parameter
   --
   --  <Description>
   --
   --    Set the symbol table entry for the next parameter following this
   --    parameter in the specification of the subprogram
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Next_Parameter (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                 Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: Formal_Param_Entry
      --   component ==> Next_Param

   begin
      Symtbl_Entry.Next_Param := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Next_Parameter");

   end Set_Next_Parameter;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Preceded_By_Comma
   --
   --  <Description>
   --
   --    Set the symbol table entry for the formal parameter to indicate
   --    that this parameter name was preceded by a comma in the textual
   --    definition of the subprogram declaration
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Preceded_By_Comma (Symtbl_Entry : in Ref_Symbol_Table_Entry)
                                   is

      -- Valid for: Formal_Param_Entry
      --   component ==> Preceded_By_Comma

   begin
      Symtbl_Entry.Preceded_By_Comma := True;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Preceded_By_Comma");

   end Set_Preceded_By_Comma;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Type_Entry :
   -- :::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_State_To_Incomplete
   --
   --  <Description>
   --
   --    Set the state of the this type identifier to be an incomplete
   --    type definition
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Type_State_To_Incomplete
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Valid for: Type_Entry
      --   component ==> Concreteness

   begin
      Symtbl_Entry.Concreteness := Incomplete;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Type_State_To_Incomplete");

   end Set_Type_State_To_Incomplete;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_State_To_Full
   --
   --  <Description>
   --
   --    Set the state of the this type identifier to be a full type definition
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Type_State_To_Full (Symtbl_Entry : in Ref_Symbol_Table_Entry)
                                    is

      -- Valid for: Type_Entry
      --   component ==> Concreteness

   begin
      Symtbl_Entry.Concreteness := Full;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Type_State_To_Full");

   end Set_Type_State_To_Full;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_State_To_Private
   --
   --  <Description>
   --
   --    Set the state of the this type identifier to be a private
   --    type definition
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Type_State_To_Private
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Valid for: Type_Entry
      --   component ==> Concreteness

   begin
      Symtbl_Entry.Concreteness := Only_Private;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Type_State_To_Private");

   end Set_Type_State_To_Private;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_State_To_Limited_Private
   --
   --  <Description>
   --
   --    Set the state of the this type identifier to be a limited private
   --    type definition
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Type_State_To_Limited_Private
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Valid for: Type_Entry
      --   component ==> Concreteness

   begin
      Symtbl_Entry.Concreteness := Limited_Private;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Type_State_To_Limited_Private");

   end Set_Type_State_To_Limited_Private;

   -- -------------------------------------------------------------------------

   -- ::::::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Subtype_Entry       :
   -- :                Derived_Type_Entry  :
   -- ::::::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Entry_Base_Type
   --
   --  <Description>
   --
   --    Set the base type definition associated with the subtype or the
   --    derived type symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Entry_Base_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                  Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: Subtype_Entry, Derived_Type_Entry
      --    component ==> Base_Type

   begin
      Symtbl_Entry.Base_Type := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Entry_Base_Type");

   end Set_Entry_Base_Type;

   -- -------------------------------------------------------------------------

   -- ::::::::::::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Enumeration_Literal_Entry :
   -- ::::::::::::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Literal_Value
   --
   --  <Description>
   --
   --    Set the internal numeric value assigned to the enumeration literal
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Literal_Value (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                Item         : in Ace_Natural) is

      -- Valid for: Enumeration_Literal_Entry
      --    component ==> Literal_Value

   begin
      Symtbl_Entry.Literal_Value := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Literal_Value");

   end Set_Literal_Value;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Successor
   --
   --  <Description>
   --
   --    Set the symbol table entry for the successor to this enumeration
   --    literal
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Successor (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                            Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: Enumeration_Literal_Entry
      --    component ==> Succ_Enum_Literal

   begin
      Symtbl_Entry.Succ_Enum_Literal := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Successor");

   end Set_Successor;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Predecessor
   --
   --  <Description>
   --
   --    Return the symbol table entry for the predecessor to this
   --    enumeration literal
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Predecessor (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                              Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: Enumeration_Literal_Entry
      --    component ==> Pred_Enum_Literal

   begin
      Symtbl_Entry.Pred_Enum_Literal := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Predecessor");

   end Set_Predecessor;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Record_Component_Entry :
   -- :::::::::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Next_Record_Component
   --
   --  <Description>
   --
   --    Set the symbol table entry for the next record component following
   --    this record component
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Next_Record_Component
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: Record_Component_Entry
      --    component ==> Next_Record_Component

   begin
      Symtbl_Entry.Next_Record_Component := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Next_Record_Component");

   end Set_Next_Record_Component;

   -- ::::::::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Simple_Name_Entry     :
   -- :   Component - Simple_Name_Construct  :
   -- ::::::::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Simple_Name_Construct
   --
   --  <Description>
   --
   --    Set the symbol table entry for Simple_Name_Entry to the appropriate
   --    statement database node.
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Simple_Name_Construct
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ref_Statement_Database) is
   begin
      Symtbl_Entry.Simple_Name_Construct := Convert_To_Fake (Item);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Simple_Name_Construct");
   end Set_Simple_Name_Construct;
   -- Simple_Name_Construct

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Type_Entry     :
   -- :   Component - Type_Definition :
   -- :::::::::::::::::::::::::::::::::

   -- :::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Type_Entry     :
   -- :   Component - Type_Definition :
   -- :                Type_Kind      :
   -- :::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_To_Enumeration_Type
   --
   --  <Description>
   --
   --    Create a type definition node that defines an enumeration type
   --    and associate it with the symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Type_To_Enumeration_Type
       (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition.Type_Kind

   begin
      Symtbl_Entry.Type_Definition :=
         new Type_Definition_Node (Enumeration_Type);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Type_To_Enumeration_Type");

   end Set_Type_To_Enumeration_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_To_Integer_Type
   --
   --  <Description>
   --
   --    Create a type definition node that defines an integer type
   --    and associate it with the symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Type_To_Integer_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry
                                       ) is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition.Type_Kind

   begin
      Symtbl_Entry.Type_Definition := new Type_Definition_Node (Integer_Type);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Type_To_Integer_Type");

   end Set_Type_To_Integer_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_To_Real_Type
   --
   --  <Description>
   --
   --    Create a type definition node that defines a real type
   --    and associate it with the symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Type_To_Real_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry)
                                   is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition.Type_Kind

   begin
      Symtbl_Entry.Type_Definition := new Type_Definition_Node (Real_Type);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Type_To_Real_Type");

   end Set_Type_To_Real_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_To_Record_Type
   --
   --  <Description>
   --
   --    Create a type definition node that defines a record type
   --    and associate it with the symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Type_To_Record_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry)
                                     is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition.Type_Kind

   begin
      Symtbl_Entry.Type_Definition := new Type_Definition_Node (Record_Type);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Type_To_Record_Type");

   end Set_Type_To_Record_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_To_Array_Type
   --
   --  <Description>
   --
   --    Create a type definition node that defines an array type
   --    and associate it with the symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Type_To_Array_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry)
                                    is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition.Type_Kind

   begin
      Symtbl_Entry.Type_Definition := new Type_Definition_Node (Array_Type);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Type_To_Array_Type");

   end Set_Type_To_Array_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Type_To_Access_Type
   --
   --  <Description>
   --
   --    Create a type definition node that defines an access type
   --    and associate it with the symbol table entry
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Type_To_Access_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry)
                                     is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition.Type_Kind

   begin
      Symtbl_Entry.Type_Definition := new Type_Definition_Node (Access_Type);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Type_To_Access_Type");

   end Set_Type_To_Access_Type;

   -- :::::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Type_Entry       :
   -- :   Component - Type_Definition   :
   -- :                Enumeration_Type :
   -- :::::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Literal_Count
   --
   --  <Description>
   --
   --    Set the number of enumeration literals associated with an
   --    enumeration type definition
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Literal_Count (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                Item         : in Ace_Natural) is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Enumeration_Type

   begin
      Symtbl_Entry.Type_Definition.Literal_Count := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Literal_Count");

   end Set_Literal_Count;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_First_Literal
   --
   --  <Description>
   --
   --    Set the symbol table entry for the first enumeration literal
   --    associated with an enumeration type definition
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_First_Literal (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Enumeration_Type

   begin
      Symtbl_Entry.Type_Definition.First_Literal := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_First_Literal");

   end Set_First_Literal;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Last_Literal
   --
   --  <Description>
   --
   --    Set the symbol table entry for the last enumeration literal
   --    associated with an enumeration type definition
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Last_Literal (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                               Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Enumeration_Type

   begin
      Symtbl_Entry.Type_Definition.Last_Literal := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Last_Literal");

   end Set_Last_Literal;

   -- :::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Type_Entry     :
   -- :   Component - Type_Definition :
   -- :                Integer_Type   :
   -- :::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Lower_Bound
   --
   --  <Description>
   --
   --    Set the lower bound associated with an integer type definition
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Lower_Bound (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                              Item         : in Ace_Integer) is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Integer_Type

   begin
      Symtbl_Entry.Type_Definition.Lower_Bound := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Lower_Bound");

   end Set_Lower_Bound;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Upper_Bound
   --
   --  <Description>
   --
   --    Set the upper bound associated with an integer type definition
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Upper_Bound (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                              Item         : in Ace_Integer) is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Integer_Type

   begin
      Symtbl_Entry.Type_Definition.Upper_Bound := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Upper_Bound");

   end Set_Upper_Bound;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Type_Entry     :
   -- :   Component - Type_Definition :
   -- :                Record_Type    :
   -- :::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Record_Component
   --
   --  <Description>
   --
   --    Set the first record component associated with a record type
   --    definition
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Record_Component (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                   Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Record_Type

   begin
      Symtbl_Entry.Type_Definition.Record_Component := Item;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Record_Component");

   end Set_Record_Component;

   -- -------------------------------------------------------------------------

   -- :::::::::::::::::::::::::::::::::
   -- : Entry_Kind ==> Type_Entry     :
   -- :   Component - Type_Definition :
   -- :                Array_Type     :
   -- :::::::::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_As_Constrained
   --
   --  <Description>
   --
   --    Set the array to be a constrained array for an array type
   --    definition
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_As_Constrained (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Array_Type

   begin
      Symtbl_Entry.Type_Definition.Is_Constrained := True;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_As_Constrained");

   end Set_As_Constrained;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_As_Unconstrained
   --
   --  <Description>
   --
   --    Set the array to be an unconstrained array for an array type
   --    definition
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_As_Unconstrained (Symtbl_Entry : in Ref_Symbol_Table_Entry) is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Array_Type

   begin
      Symtbl_Entry.Type_Definition.Is_Constrained := False;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_As_Unconstrained");

   end Set_As_Unconstrained;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Index_Type
   --
   --  <Description>
   --
   --    Set the symbol table entry for the index associated with the
   --    array type definition
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Index_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                             Item         : in Ref_Statement_Database) is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Array_Type

   begin
      Symtbl_Entry.Type_Definition.Array_Index_Type :=
          Convert_To_Fake(Item);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Index_Type");

   end Set_Index_Type;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Element_Type
   --
   --  <Description>
   --
   --    Set the symbol table entry for the element associated with the
   --    array type definition
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Element_Type (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                               Item         : in Ref_Statement_Database) is

      -- Valid for: Type_Entry
      --   component ==> Type_Definition, where Type_Kind = Array_Type

   begin
      Symtbl_Entry.Type_Definition.Array_Element_Type := Convert_To_Fake(Item);

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Element_Type");

   end Set_Element_Type;

   -- -------------------------------------------------------------------------

   -- ::::::::::::::::::::::::::
   -- : Constraint_Node fields :
   -- ::::::::::::::::::::::::::

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Lower_Bound
   --
   --  <Description>
   --
   --    Set the lower bound associated with an array type definition,
   --    or a constraint definition
   --
   --    Note: a constraint definition may exist for an object, a subtype, a
   --          derived type, or a constraint being built during tree building.
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Lower_Bound (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                              Item         : in Ref_Statement_Database) is

      -- Valid for: Constraint_Entry
      --   component ==> Constraint_Ref,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Object_Entry
      --   component ==> Object_Constraint,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Subtype_Entry, Derived_Type_Entry
      --   component ==> Constraint_Info,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint

      -- The lower bound must be converted from a Ref_Statement_Database
      -- to a Ref_Statement_Database_Fake

   begin
      case Symtbl_Entry.Entry_Kind is

         when Constraint_Entry =>
            Symtbl_Entry.Constraint_Ref.Lower_Bound :=
               Convert_To_Fake (Item);

         when Object_Entry     =>
            Symtbl_Entry.Object_Constraint.Constraint_Ref.Lower_Bound :=
               Convert_To_Fake (Item);

         when Subtype_Entry    | Derived_Type_Entry     =>
            Symtbl_Entry.Constraint_Info.Constraint_Ref.Lower_Bound :=
               Convert_To_Fake (Item);

         when others =>
            Show_Error (2000, Internal, "Set_Lower_Bound_C1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Lower_Bound_C");

   end Set_Lower_Bound;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Lower_Bound_Value_Offset
   --
   --  <Description>
   --
   --    Set the storage management offset location in which the lower
   --    bound associated with an array type definition, or a constraint
   --    definition is stored
   --
   --    Note: a constraint definition may exist for an object, a subtype, a
   --          derived type, or a constraint being built during tree building.
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Lower_Bound_Value_Offset
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ace_Natural) is

      -- Valid for: Constraint_Entry
      --   component ==> Constraint_Ref,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Object_Entry
      --   component ==> Object_Constraint,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Subtype_Entry, Derived_Type_Entry
      --   component ==> Constraint_Info,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint

   begin
      case Symtbl_Entry.Entry_Kind is

         when Constraint_Entry =>
            Symtbl_Entry.Constraint_Ref.Lower_Bound_Value_Offset := Item;

         when Object_Entry     =>
            Symtbl_Entry.Object_Constraint.Constraint_Ref.
               Lower_Bound_Value_Offset := Item;

         when Subtype_Entry    | Derived_Type_Entry     =>
            Symtbl_Entry.Constraint_Info.Constraint_Ref.
               Lower_Bound_Value_Offset := Item;

         when others =>
            Show_Error (2000, Internal, "Set_Lower_Bound_Value_Offset_1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Lower_Bound_Value_Offset");

   end Set_Lower_Bound_Value_Offset;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Upper_Bound
   --
   --  <Description>
   --
   --    Set the upper bound associated with an array type definition,
   --    or a constraint definition
   --
   --    Note: a constraint definition may exist for an object, a subtype, a
   --          derived type, or a constraint being built during tree building.
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Upper_Bound (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                              Item         : in Ref_Statement_Database) is

      -- Valid for: Constraint_Entry
      --   component ==> Constraint_Ref,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Object_Entry
      --   component ==> Object_Constraint,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Subtype_Entry, Derived_Type_Entry
      --   component ==> Constraint_Info,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint

      -- The upper bound must be converted from a Ref_Statement_Database
      -- to a Ref_Statement_Database_Fake

   begin
      case Symtbl_Entry.Entry_Kind is

         when Constraint_Entry =>
            Symtbl_Entry.Constraint_Ref.Upper_Bound :=
               Convert_To_Fake (Item);

         when Object_Entry     =>
            Symtbl_Entry.Object_Constraint.Constraint_Ref.Upper_Bound :=
               Convert_To_Fake (Item);

         when Subtype_Entry    | Derived_Type_Entry     =>
            Symtbl_Entry.Constraint_Info.Constraint_Ref.Upper_Bound :=
               Convert_To_Fake (Item);

         when others =>
            Show_Error (2000, Internal, "Set_Upper_Bound_C1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Upper_Bound_C");

   end Set_Upper_Bound;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Upper_Bound_Value_Offset
   --
   --  <Description>
   --
   --    Set the storage management offset location in which the upper
   --    bound associated with an array type definition, or a constraint
   --    definition is stored
   --
   --    Note: a constraint definition may exist for an object, a subtype, a
   --          derived type, or a constraint being built during tree building.
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Upper_Bound_Value_Offset
       (Symtbl_Entry : in Ref_Symbol_Table_Entry;
        Item         : in Ace_Natural) is

      -- Valid for: Constraint_Entry
      --   component ==> Constraint_Ref,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Object_Entry
      --   component ==> Object_Constraint,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint
      -- Valid for: Subtype_Entry, Derived_Type_Entry
      --   component ==> Constraint_Info,
      --                            where Constraint_Kind = Range_Constraint
      --                               or Constraint_Kind = Index_Constraint

   begin
      case Symtbl_Entry.Entry_Kind is

         when Constraint_Entry =>
            Symtbl_Entry.Constraint_Ref.Upper_Bound_Value_Offset := Item;

         when Object_Entry     =>
            Symtbl_Entry.Object_Constraint.Constraint_Ref.
               Upper_Bound_Value_Offset := Item;

         when Subtype_Entry    | Derived_Type_Entry     =>
            Symtbl_Entry.Constraint_Info.Constraint_Ref.
               Upper_Bound_Value_Offset := Item;

         when others =>
            Show_Error (2000, Internal, "Set_Upper_Bound_Value_Offset_1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Upper_Bound_Value_Offset");

   end Set_Upper_Bound_Value_Offset;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Next_Constraint
   --
   --  <Description>
   --
   --    Set the next constraint node associated with the index of an array
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error
   --==========================================================================
   procedure Set_Next_Constraint (Symtbl_Entry : in Ref_Symbol_Table_Entry;
                                  Item         : in Ref_Symbol_Table_Entry) is

      -- Valid for: Constraint_Entry
      --   component ==> Constraint_Ref,
      --                            where Constraint_Kind = Index_Constraint
      -- Valid for: Object_Entry
      --   component ==> Object_Constraint,
      --                            where Constraint_Kind = Index_Constraint

   begin
      case Symtbl_Entry.Entry_Kind is

         when Constraint_Entry =>
            Symtbl_Entry.Constraint_Ref.Next_Constraint := Item;

         when Object_Entry     =>
            Symtbl_Entry.Object_Constraint.Constraint_Ref.Next_Constraint :=
               Item;

         when others =>
            Show_Error (2000, Internal, "Set_Next_Constraint_1");

      end case;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Next_Constraint");

   end Set_Next_Constraint;

   -- -------------------------------------------------------------------------

   -- -------------------------------------------------------------------------

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Set_Help
   --
   --  <Description>
   --
   --    Set the help field of the symbol table entry to the associated
   --    information.
   --
   --==========================================================================
   procedure Set_Help (St_Entry : in Ref_Symbol_Table_Entry) is
      -- valid for Package_Entry, Procedure_Entry, Function_Entry, and types
      Help_Tmp : Ref_Help;
   begin
      if (St_Entry.Entry_Kind = Package_Entry) then
         Help_Tmp := Dq_Pkg_Help;
      else
         Help_Tmp := Get_Help;
      end if;

      if not Is_Empty (Help_Tmp) then
         case St_Entry.Entry_Kind is
            when Package_Entry =>
               St_Entry.Pkg_Help := Help_Tmp;
            when Procedure_Entry    | Function_Entry    =>
               St_Entry.Subprg_Help := Help_Tmp;
            when Type_Entry    =>
               St_Entry.Type_Help := Help_Tmp;
            when Subtype_Entry | Derived_Type_Entry     =>
               St_Entry.Other_Type_Help := Help_Tmp;
            when others   =>
               Show_Error (1001, Internal, "Set_Help");
         end case;
      end if;

   exception
      when Constraint_Error    =>
         Show_Error (2000, Internal, "Set_Help");
   end Set_Help;

end Symtbl_Entry_Set_Routines;
