--|$Log:	onlyx.a,v $
--|Revision 8.0  91/01/17  10:48:27  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - Window_System [Body]
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    This package contains the generic set of windowing system functions
--    that ACE will make use of.  This package retains the information
--    that determines which (if any) windowing system ACE is executing
--    within and invokes the proper routine based upon the windowing
--    system.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Debugger;

with X_Window_System;

package body Window_System is

   type Window_System_Kind is (None, Sun, X);

   Current_Ws : Window_System_Kind := None;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Running_Under_Windowing_System
   --
   --  <Description>
   --
   --    This function determines whether ACE is running under a windowing
   --    system.  If so, True is returned; otherwise, false.
   --
   --==========================================================================
   function Running_Under_Windowing_System return Boolean is
   begin
      return (Current_Ws /= None);
   end Running_Under_Windowing_System;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Observe_Subprograms_Window
   --
   --  <Description>
   --
   --    This procedrue creates a subprogram observation window (when the
   --    pragma observe (subprograms, on) is executed).
   --
   --==========================================================================
   procedure Create_Observe_Subprograms_Window is
   begin
      case Current_Ws is
         when Sun    =>
            null;
         when X =>
            X_Window_System.Create_Observe_Subprograms_Window
               ;
         when None   =>
            null;
      end case;
   end Create_Observe_Subprograms_Window;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Observe_Tasks_Window
   --
   --  <Description>
   --
   --    This procedure creates an observation window for tasks
   --
   --==========================================================================
   procedure Create_Observe_Tasks_Window is
   begin
      case Current_Ws is
         when Sun    =>
            null;
         when X =>
            X_Window_System.Create_Observe_Tasks_Window
               ;
         when None   =>
            null;
      end case;
   end Create_Observe_Tasks_Window;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    End_Of_Displaying_Lines
   --
   --  <Description>
   --
   --    This procedure notifies the window that the end of displaying lines
   --    within the window has been reached.
   --
   --==========================================================================
   procedure End_Of_Displaying_Lines is
   begin
      case Current_Ws is
         when Sun    =>
            null;
         when X =>
            X_Window_System.End_Of_Displaying_Lines;
         when None   =>
            null;
      end case;
   end End_Of_Displaying_Lines;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Synchonize_With_Textwindow
   --
   --  <Description>
   --
   --    This procedure allows synchronization to be accomplished with the
   --    text window when needed.
   --
   --==========================================================================
   procedure Synchonize_With_Textwindow is
   begin
      case Current_Ws is
         when Sun    =>
            null;
         when X =>
            X_Window_System.Synchronize;
         when None   =>
            null;
      end case;
   end Synchonize_With_Textwindow;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Clear_Textwindow
   --
   --  <Description>
   --
   --    This procedure clears the text from within a text window.
   --
   --==========================================================================
   procedure Clear_Textwindow is
   begin
      case Current_Ws is
         when Sun    =>
            null;
         when X =>
            X_Window_System.Clear_Textwindow;
         when None   =>
            null;
      end case;
   end Clear_Textwindow;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Show_Statement
   --
   --  <Description>
   --
   --    This procedure is the method by which an Ada statement within the
   --    window is highlighted (SHOWn) to the user.
   --
   --==========================================================================
   procedure Show_Statement (Str : in String) is
   begin
      case Current_Ws is
         when Sun    =>
            null;
         when X =>
            X_Window_System.Find_String (Str);
         when None   =>
            null;
      end case;
   end Show_Statement;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Kill_Window
   --
   --  <Description>
   --
   --    This routine kills the window.
   --
   --==========================================================================
   procedure Kill_Window is
   begin
      if Debugger.Any_Windows_Created then
         case Current_Ws is
            when Sun =>
               null;
            when X   =>
               X_Window_System.Kill_Window;
            when None     =>
               null;
         end case;
      end if;
   end Kill_Window;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Display_Line
   --
   --  <Description>
   --
   --    This procedure displays a line of text within the window.
   --
   --==========================================================================
   procedure Display_Line (Str : in String) is
   begin
      case Current_Ws is
         when Sun    =>
            null;
         when X =>
            X_Window_System.Display_Line (Str);
         when None   =>
            null;
      end case;
   end Display_Line;

   -- ======================================================================

   function Is_Under_X return Boolean;
   pragma INTERFACE (C, Is_Under_X);
   pragma LINKNAME (Is_Under_X, "is_under_X");

begin
   -- At initialization, this package determines which windowing system
   -- ACE is executing under.
   if Is_Under_X then
      Current_Ws := X;
   else
      Current_Ws := None;
   end if;
end Window_System;
