--|$Log:	system.a,v $
--|Revision 8.0  91/01/17  10:48:33  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - System_Dependent_Routines
--
---------------------------------------------------------------
--
--
-- Abstract
--
--  This package contains all the code that needs to be ported and
--  the system dependent routines.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------
with System;
use System;

with Literals;
use Literals;

with Text_Io;
use Text_Io;

with Error_Handler;
use Error_Handler;

with Ooe;
use Ooe;

with Ace_Text_Io;
use Ace_Text_Io;

with Ace_Universal_Types;
with Sequential_Io;
with Direct_Io;
with System_Environment;
with Unchecked_Conversion;
with Ace;

package body System_Dependent_Routines is


   Home                : Directory_Object             := Directory_Object'LAST;
   Current             : Directory_Object             := Directory_Object'LAST;

   type Status_Code is new Thirty_Two_Int;

   Interrupted         : Thirty_Two_Int               := 0;

   Command_Line_Number : System_Environment.Argc_Type := 0;

   Arg_Count : System_Environment.Argc_Type := System_Environment.Get_Argc;
   Arg_Value : System_Environment.Argv_Type := System_Environment.Get_Argv;



   --==========================================================================
   -- <Subprogram Name>
   --
   -- Flush
   --
   -- <Description>
   --
   -- This routine will "flush" all current input from the input
   -- queue.  This is a system dependent function, that needs to
   -- interface with the host O/S.
   --
   -- <Called By>
   --
   -- Error recovery routines
   --
   --==========================================================================

   procedure Flush is
      procedure C_Flush;
      pragma INTERFACE (C, C_Flush);
   begin
      C_Flush;
   end Flush;


   --==========================================================================
   -- <Subprogram Name>
   --
   -- System
   --
   -- <Description>
   --
   -- Spawns a process that will execute the contents of the string
   -- Command.
   --
   --==========================================================================

   procedure System (Command : String;
                     Args    : String         := "";
                     How     : Execution_Mode := Foreground) is

      procedure Csystem (Command : Cstring);
      pragma INTERFACE (C, Csystem);
      pragma LINKNAME (Csystem, "system");

      Foreground_String : constant String := Command & " " & Args;
      Background_String : constant String := Command & " " & Args & " &";

   begin
      if How = Foreground then
         Csystem (Convert_To_Cstring (Foreground_String));
      else
         Csystem (Convert_To_Cstring (Background_String));
      end if;
   end System;


   --==========================================================================
   -- <Subprogram Name>
   --
   -- Edit
   --
   -- <Description>
   --
   -- This routine calls a host operating system editor.
   --
   -- <Called By>
   --
   -- ACE_File_System
   --
   --==========================================================================

   procedure Edit (Obj : in out File_Object;
                   How : in Execution_Mode := Foreground) is
      File_Name : constant String := Text_Io.Name (File_Entries (Obj));
   begin
      Ace_File_System.Close (Obj);
      System ("ace_edit", File_Name, How);
      Ace_File_System.Open (Obj, Ace_File_System.In_Object, File_Name);
   end Edit;


   --==========================================================================
   -- <Subprogram Name>
   --
   -- Compile
   --
   -- <Description>
   --
   -- This routine compiles an Ada program in the host operating system.
   --
   -- <Called By>
   --
   -- Ace_File_System
   --
   --==========================================================================

   procedure Compile (Obj  : in File_Object;
                      Form : in String) is
   begin
      -- ace_ada is a host O/S shell script that will set up
      -- the proper environment for compilation, and compile
      -- the unit.  Form is used to send options to the Ada compiler
      System ("ace_ada", Form & ' ' & Text_Io.Name (File_Entries (Obj)),
              Foreground);
   end Compile;


   --==========================================================================
   -- <Subprogram Name>
   --
   -- Rename
   --
   -- <Description>
   --
   -- Renames a file from Obj1 to Obj2
   --
   -- <Called By>
   --
   -- Ace_File_System
   --
   --==========================================================================
   procedure Rename (Obj1, Obj2 : in out Ref_String) is

      Status : Status_Code;

      function Crename (From, To : Cstring) return Status_Code;
      pragma INTERFACE (C, Crename);
      pragma LINKNAME (Crename, "rename");

   begin
      Status := Crename (Convert_To_Cstring (Convert_To_String (Obj1)),
                         Convert_To_Cstring (Convert_To_String (Obj2)));
      if Status /= 0 then
         Show_Error (3007, Warning,
                     "Rename (" & Convert_To_String (Obj1) & " ," &
                     Convert_To_String (Obj2) & ");");
      end if;
   end Rename;


   --==========================================================================
   -- <Subprogram Name>
   --
   -- Chdir
   --
   -- <Description>
   --
   -- Changes the current working directory
   --
   -- <Called By>
   --
   -- Ace_File_System
   --
   --==========================================================================

   procedure Chdir (Obj1 : in Ref_String) is

      Status : Status_Code;

      function Cchdir (Dir : Cstring) return Status_Code;
      pragma INTERFACE (C, Cchdir);
      pragma LINKNAME (Cchdir, "chdir");

   begin
      Status := Cchdir (Convert_To_Cstring (Convert_To_String (Obj1)));
      if Status /= 0 then
         Show_Error (3007, Warning,
                     "Set_Directory (" & Convert_To_String (Obj1) & ");");
      end if;
   end Chdir;


   --==========================================================================
   -- <Subprogram Name>
   --
   -- Mkdir
   --
   -- <Description>
   --
   -- Creates the directory specified
   --
   -- <Called By>
   --
   -- Ace_File_System
   --
   --==========================================================================

   procedure Mkdir (Obj1 : in Ref_String) is

      Status : Status_Code;

      function Cmkdir (Dir  : Cstring;
                       Mode : Thirty_Two_Int) return Status_Code;
      pragma INTERFACE (C, Cmkdir);
      pragma LINKNAME (Cmkdir, "mkdir");

      Mode : constant Thirty_Two_Int := 8#777#;

   begin
      Status := Cmkdir (Convert_To_Cstring (Convert_To_String (Obj1)), Mode);
      if Status /= 0 then
         Show_Error (3007, Warning,
                     "Create (" & Convert_To_String (Obj1) & ");");
      end if;
   end Mkdir;


   package Unix_Directories is

      type Sixteen_Int is  range 0 .. (16 ** 2) - 1;
      for Sixteen_Int'SIZE use 16;

      type Dirdesc is
         record
            Fd    : Thirty_Two_Int;
            Loc   : Thirty_Two_Int;
            Size  : Thirty_Two_Int;
            Bsize : Thirty_Two_Int;
            Off   : Thirty_Two_Int;
            Buf   : Thirty_Two_Int;
         end record;

      for Dirdesc use
         record --at mod 16;
            Fd        at 0 range 0 .. 31;
            Loc       at 4 range 0 .. 31;
            Size      at 8 range 0 .. 31;
            Bsize     at 12 range 0 .. 31;
            Off       at 16 range 0 .. 31;
            Buf       at 20 range 0 .. 31;
         end record;

      for Dirdesc'SIZE use 24 * Standard.System.Storage_Unit;
      type Dirdesc_Access is access Dirdesc;

      type Direct is
         record
            Off   : Thirty_Two_Int;
            Fileno : Thirty_Two_Int;
            Reclen : Sixteen_Int;
            Namlen : Sixteen_Int;
            Name   : String (1 .. 256);
         end record;

      for Direct use
         record -- at mod 16;
            Off      at 0 range 0 .. 31;
            Fileno   at 4 range 0 .. 31;
            Reclen   at 8 range 0 .. 15;
            Namlen   at 10 range 0 .. 15;
            Name     at 12 range 0 .. 2047;
         end record;

      for Direct'SIZE use 268 * Standard.System.Storage_Unit;
      type Direct_Access is access Direct;

      function Opendir (Filename : Cstring) return Dirdesc_Access;
      pragma INTERFACE (C, Opendir);

      function Readdir (Dirfd : Dirdesc_Access) return Direct_Access;
      pragma INTERFACE (C, Readdir);

      function Closedir (Dirfd : Dirdesc_Access) return Status_Code;
      pragma INTERFACE (C, Closedir);

   end Unix_Directories;


   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Opendir
   --
   --    Last Mod:    Thu Nov 30 16:30:06 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Opendir (Filename : String) return Ace_Integer is
      use Unix_Directories;
      function Convert_To_Ace_Integer is new Unchecked_Conversion
          (Dirdesc_Access, Ace_Integer);
   begin
      return Convert_To_Ace_Integer (Opendir (Convert_To_Cstring (Filename)));
   end Opendir;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Readdir
   --
   --    Last Mod:    Thu Nov 30 16:30:02 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Readdir (Dirfd : Ace_Integer) return String is
      use Unix_Directories;
      function Convert_To_Dirdesc_Access is new Unchecked_Conversion
          (Ace_Integer, Dirdesc_Access);
      Dir : Direct_Access;
   begin
      Dir := Readdir (Convert_To_Dirdesc_Access (Dirfd));
      return Dir.Name (1 .. (Integer (Dir.Namlen)));
   end Readdir;

   --  --------------------------------------------------------------------
   --  <Subprogram Name>
   --    Closedir
   --
   --    Last Mod:    Thu Nov 30 16:29:55 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  --------------------------------------------------------------------
   function Closedir (Dirfd : Ace_Integer) return Ace_Integer is
      use Unix_Directories;
      function Convert_To_Dirdesc_Access is new Unchecked_Conversion
          (Ace_Integer, Dirdesc_Access);
   begin
      return Ace_Integer (Closedir (Convert_To_Dirdesc_Access (Dirfd)));
   end Closedir;


   --==========================================================================
   -- <Subprogram Name>
   --
   -- Directory_Lister
   --
   -- <Description>
   --
   -- This procedure will list all the files contained in the directory
   --
   -- <Called By>
   --
   -- Ace_File_System
   --
   --==========================================================================

   procedure Directory_Lister (Obj : Ref_String) is


      use Unix_Directories;

      -- Create an array to hold (hopefully) enough file names
      type Dir_Array is array (Integer range <>) of Cstring;
      Dir_List           : Dir_Array (1 .. 256);

      Directory_Notfound : exception;

      File_Count, J      : Integer;
      Num_Rows           : Integer;
      Status             : Status_Code;

      procedure Load_List (Which_Dir : in String;
                           File_List : out Dir_Array;
                           Num_Files : out Integer) is
         Dir_Ptr    : Dirdesc_Access;
         Dir_Name   : Cstring;
         Dir        : Direct_Access;
         File_Count : Integer;

         function Convert_To_Int is new Unchecked_Conversion
             (Direct_Access, Thirty_Two_Int);

      begin
         Dir_Name := Convert_To_Cstring (Which_Dir);
         Dir_Ptr := Opendir (Dir_Name);
         if (Dir_Ptr = null) then
            raise Directory_Notfound;
         end if;
         File_Count := 0;
         loop
            -- Read in all of the file names
            Dir := Readdir (Dir_Ptr);
            exit  when (Convert_To_Int (Dir) = Thirty_Two_Int (0));
            File_Count := File_Count + 1;
            File_List (File_Count) :=
               Convert_To_Cstring (Dir.Name (1 .. (Integer (Dir.Namlen))));
         end loop;
         Status := Closedir (Dir_Ptr);
         Num_Files := File_Count;
      end Load_List;

      procedure Sort_Filenames (File_List : in out Dir_Array) is
         Sort_Index : Integer;
         Fname_Tmp  : Cstring;
      begin
         for I in 2 .. File_List'LAST loop
            Fname_Tmp := File_List (I);
            Sort_Index := I;
            while (Sort_Index > 1) loop
               exit  when (Convert_To_Astring (File_List (Sort_Index - 1)) <=
                          Convert_To_Astring (Fname_Tmp));
               File_List (Sort_Index) := File_List (Sort_Index - 1);
               Sort_Index := Sort_Index - 1;
            end loop;
            File_List (Sort_Index) := Fname_Tmp;
         end loop;
      end Sort_Filenames;

   begin
      Load_List (Convert_To_String (Obj), Dir_List, File_Count);
      Sort_Filenames (Dir_List (1 .. File_Count));
      Num_Rows := File_Count / 2;

      -- File names will be displayed in two columns.. hopefully
      -- names will be less than 40 characters.

      -- account for odd number of files...
      if (File_Count mod 2 = 1) then
         Num_Rows := Num_Rows + 1;
      end if;

      Text_Io.Set_Col (1);

      for I in 1 .. (Num_Rows) loop
         J := I + Num_Rows;
         Text_Io.Put (Convert_To_Astring (Dir_List (I)));
         if (I + Num_Rows <= File_Count) then
            Text_Io.Set_Col (40);
            Text_Io.Put_Line (Convert_To_Astring (Dir_List (J)));
         end if;
      end loop;

      Text_Io.New_Line;

   exception
      when Directory_Notfound  =>
         raise Ace_Text_Io.Name_Error;

   end Directory_Lister;


   --==========================================================================
   -- <Subprogram Name>
   --
   -- Copy
   --
   -- <Description>
   --
   -- Copies Obj1 to Obj2
   --
   -- <Called By>
   --
   -- Ace_File_System
   --
   --==========================================================================

   procedure Copy (Obj1, Obj2 : in out Ref_String) is

      -- this routine will read chunks of 256 bytes until
      -- there are no more chunks, then it reads each byte until
      -- the end.

      -- Create a type that represents what composes files.
      -- For most systems that is an 8 bit object, but who knows.
      type Byte is  range 0 .. 255;
      for Byte'SIZE use 8;

      type Block is array (1 .. 256) of Byte;

      -- Create an IO package for our type
      package Block_Io is new Sequential_Io (Block);
      use Block_Io;

      package Byte_Io is new Direct_Io (Byte);

      -- File objects
      From_File  : Block_Io.File_Type;
      To_File    : Block_Io.File_Type;

      -- Object to be copied
      Object     : Block;

      Byte_Count : Ace_Integer := 0;

   begin

      -- Open From file.
      Block_Io.Open (From_File, In_File, Convert_To_String (Obj1));
      begin
         -- Open To file.  If it exists delete it.
         Block_Io.Open (To_File, Out_File, Convert_To_String (Obj2));
         Block_Io.Delete (To_File);
         Block_Io.Create (To_File, Out_File, Convert_To_String (Obj2));
      exception
         when Block_Io.Name_Error   =>
            -- File didn't exists, so we create it.
            Block_Io.Create (To_File, Out_File, Convert_To_String (Obj2));
      end;

      -- Copy From File to To File
      while not End_Of_File (From_File) loop
         Block_Io.Read (From_File, Object);
         Block_Io.Write (To_File, Object);
         Byte_Count := Byte_Count + 256;
      end loop;

      Block_Io.Close (From_File);
      Block_Io.Close (To_File);

   exception
      when Block_Io.End_Error  =>

         declare

            use Byte_Io;
            From_File_Byte : Byte_Io.File_Type;
            To_File_Byte   : Byte_Io.File_Type;
            Byte_Object    : Byte;

         begin

            Block_Io.Close (To_File);
            Block_Io.Close (From_File);

            Byte_Io.Open (From_File_Byte, In_File, Convert_To_String (Obj1));
            Byte_Io.Open (To_File_Byte, Out_File, Convert_To_String (Obj2));

            Byte_Count := Byte_Count + 1;

            Byte_Io.Read (From_File_Byte, Byte_Object,
                          Byte_Io.Positive_Count (Byte_Count));
            Byte_Io.Write (To_File_Byte, Byte_Object,
                           Byte_Io.Positive_Count (Byte_Count));

            loop
               Byte_Io.Read (From_File_Byte, Byte_Object);
               Byte_Io.Write (To_File_Byte, Byte_Object);
            end loop;

         exception

            when Byte_Io.End_Error  =>

               Byte_Io.Close (From_File_Byte);
               Byte_Io.Close (To_File_Byte);

         end;

   end Copy;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Get_Directory_Name
   --
   -- <Description>
   --
   -- This routine takes as a parameter a relative directory name, and
   -- returns the full directory name. For example, it will change "." to
   -- "/mn1/wpl/.".
   --
   -- <Called By>
   --
   -- Ace_File_System
   --
   --==========================================================================
   function Get_Directory_Name (Str : String) return String is
      File : Text_Io.File_Type;
   begin
      Text_Io.Open (File, In_File, Str, "");
      declare
         Return_String : constant String := Text_Io.Name (File);
      begin
         Text_Io.Close (File);
         return Return_String;
      end;
   end Get_Directory_Name;


   --==========================================================================
   -- <Subprogram Name>
   --
   -- Current_Directory
   --
   -- <Description>
   --
   -- Returns the File_OBject associated with the user current directory.
   --
   -- <Called By>
   --
   -- Ace_File_System
   --
   --==========================================================================

   function Current_Directory return Directory_Object is

      procedure Cgetwd (Buff : Cstring);
      pragma INTERFACE (C, Cgetwd);
      pragma LINKNAME (Cgetwd, "getwd");

      function Convert_To_Cstring is new Unchecked_Conversion
          (Address, Cstring);

      Holder : String (1 .. 1024); -- TBD internal limit on path.

      function Find_Null (Str : String) return Integer is
      begin
         for I in Str'FIRST .. Str'LAST loop
            if Str (I) = Ascii.Nul then
               return I - 1;
            end if;
         end loop;
      end Find_Null;

   begin
      if Is_Open (Current) then
         Close (Current);
      end if;
      Cgetwd (Convert_To_Cstring (Holder'ADDRESS));
      Open (Current, Holder (1 .. Find_Null (Holder)), "");
      return Current;
   end Current_Directory;


   --==========================================================================
   -- <Subprogram Name>
   --
   -- Home_Directory
   --
   -- <Description>
   --
   -- Returns the File_Object associated with the users home directory
   -- (Startup directory).
   --
   -- <Called By>
   --
   -- ACE_File_System
   --
   --==========================================================================

   function Home_Directory return Directory_Object is

      function Getenv (Var : String) return String is

         function Cgetenv (Var_Name : Cstring) return Cstring;
         pragma INTERFACE (C, Cgetenv);
         pragma LINKNAME (Cgetenv, "getenv");

      begin
         return Convert_To_Astring (Cgetenv (Convert_To_Cstring (Var)));
      end Getenv;

   begin
      if not Is_Open (Home) then
         Open (Home, Getenv ("HOME"), "");
      end if;
      return Home;
   end Home_Directory;


   --==========================================================================
   -- <Subprogram Name>
   --
   -- Command_Line_Argument
   --
   -- <Description>
   --
   -- This is an Ada compiler dependent subprogram, that will return
   -- a string that contains the first command line argument.  This
   -- routine will have to re-written for each Ada compilation system.
   -- Currently this routine will work for VADS, and Alsys (since we
   -- have a routine that mimics that Alsys "System_Environment" for
   -- VADS).
   --
   -- <Called By>
   --
   -- Pre_Parser
   --
   --==========================================================================
   function Command_Line_Argument return String is
      use System_Environment;
   begin
      Command_Line_Number := Command_Line_Number + 1;
      if Arg_Count > Command_Line_Number then
         return Arg_Value (Command_Line_Number).all;
      else
         return "";
      end if;
   end Command_Line_Argument;

   --==========================================================================
   -- <Subprogram Name>
   --
   -- Abort_Check
   --
   -- <Description>
   --
   -- This routine will check if the user has typed an abort (i.e., ^C).
   -- If so, then setup for another abort, and raise interrupt_error.
   --
   -- <Called By>
   --
   -- External
   --
   --==========================================================================
   procedure Set_Signal (Flag : Standard.System.Address);
   pragma INTERFACE (C, Set_Signal);

   function Abort_Check return Boolean is
      Interrupted_Save : Thirty_Two_Int := Interrupted;
   begin
      Interrupted := 0;
      return Interrupted_Save = 1;
   end Abort_Check;

   --  -----------------------------------------------------------------------
   --  <Package Name>
   --    CPU
   --
   --    Last Mod:    Tue Nov 28 16:07:01 1989
   --    Author:      William P Loftus <wpl@huey>
   --
   --  <Description>
   --
   --  Provide the CPU time to the user (and operations for printing it).
   --  -----------------------------------------------------------------------
   package body Cpu is

      type Ref_Duration is access Duration;

      function Convert_To_Time is new Unchecked_Conversion
          (Ref_Duration, Time);

      function Convert_To_Integer is new Unchecked_Conversion
          (Address, Thirty_Two_Int);

      function Convert_To_Duration (A_Time : Time) return Duration is
         function Convert_To_Ref_Duration is new Unchecked_Conversion
             (Time, Ref_Duration);
      begin
         return Convert_To_Ref_Duration (A_Time).all;
      end Convert_To_Duration;

      function Clock return Time is

         function Times (Param : Thirty_Two_Int) return Thirty_Two_Int;
         pragma INTERFACE (C, Times);

         type Time_Record is
            record
               User_Time, System_Time, Children_User_Time, Children_System_Time
                : Thirty_Two_Int;
            end record;

         Answer : Time_Record;
         Total  : Thirty_Two_Int;

      begin
         Total := Times (Convert_To_Integer (Answer'ADDRESS));
         return
            Convert_To_Time
                (new Duration'(
                    Duration (Standard."/" (Duration (Answer.User_Time),
                                            Duration (60.0)))));
      end Clock;


      function "-" (Stop_Time, Start_Time : Time) return Time is
      begin
         return
            Convert_To_Time (new Duration'(Convert_To_Duration (Stop_Time) -
                                   Convert_To_Duration (Start_Time)));
      end "-";

      procedure Put_Time (A_Time : in Time) is
         package Duration_Io is new Text_Io.Fixed_Io (Duration);
      begin
         Duration_Io.Put (Convert_To_Duration (A_Time), 2, 3, 0);
      end Put_Time;

   end Cpu;

begin
   -- set up ^C handler.
   Set_Signal (Interrupted'ADDRESS);
end System_Dependent_Routines;
