--|$Log:	user.a,v $
--|Revision 8.0  91/01/17  10:46:33  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - User_Pkg
--
---------------------------------------------------------------
--
--
-- Abstract
--
--   This package constains the input routines for ACE.
--
--   The input for ACE is complicated.  The following is a description of
--   what goes on:
--
--   Places input can come from:
--
--     1) Standard_Input  (The users terminal)
--     2) A named file    (A file named by the user that is loaded into ACE)
--     3) "Unput" buffer  (The buffer used to scan input ahead of its time)
--
--   If there is input in the Unput buffer (object Buf) then we take input
--   from there.
--   If there is input in the Standard_Input buffer (object Abuf) then
--   we get another input line from "Ace_Current_Input" which is sometimes
--   the Standard_Input of the users terminal and sometimes a named file.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with Text_Io;
use Text_Io;

with Ace_Universal_Types;

with Misc;
use Misc;

with Error_Handler;
use Error_Handler;

with Ace;
use Ace;

with System;

with Ace_Text_Io;
with Ooe;
with Unchecked_Conversion;
with Ace_Command_Line_Input;

package body User_Pkg is


   -- Buf is used to hold input that has been read in and
   -- "put" back on the input stream.
   Bufsize    : constant Integer := 1024;
   subtype Buf_Ind is Natural range 0 .. Bufsize;
   Buf        : String (1 .. Bufsize);
   Sptr       : Buf_Ind          := 0;

   -- Abuf is used to hold a line sometimes from
   -- Standard_Input and sometimes from a File object.
   Abuf       : String (1 .. Bufsize);
   Aptr       : Buf_Ind          := 0;
   Line_Lng   : Buf_Ind          := 0;

   -- Should we echo input?
   Echo_Input : Boolean          := False;

   -- EOF is OK or NOT.
   Eof_Ok     : Boolean          := True;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Echo
   --
   --  <Description>
   --
   --    Echo sets an object that will flag the input routine to put_line
   --    everything that is read into ACE.
   --
   --==========================================================================
   procedure Echo (Bool : Boolean) is
   begin
      Echo_Input := Bool;
   end Echo;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Unput
   --
   --  <Description>
   --
   --    Unput puts a character back into the input stream.
   --
   --  <Called By>
   --
   --    The lexical routines, and the pre_parser.
   --
   --==========================================================================
   procedure Unput (C : in Character) is
   begin
      Sptr       := Sptr + 1;
      Buf (Sptr) := C;
   end Unput;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Lower
   --
   --  <Description>
   --
   --     The lower procedure take a line of Ada input and turns all
   --   capital letters to small letters (except letters in Strings or
   --   character constants).  This lets us have a smaller scanner (we
   --   don't have to worry about case-insensitivity).
   --
   --  <Called By>
   --
   --    Get_Another_Line
   --
   --  <Exceptions Handled>
   --
   --    Constraint_Error  -- raised by indexing out of a String
   --
   --==========================================================================
   procedure Lower (Str : in out String) is
      I : Positive;
   begin
      I := Str'FIRST;
      while I <= Str'LAST loop
         case Str (I) is
            when 'A' .. 'Z'    =>
               Str (I) := Character'VAL (Character'POS (Str (I)) +
                                         Character'POS ('a') -
                                         Character'POS ('A'));
            when '-' =>
               if I /= Str'LAST and then Str (I + 1) = '-' then
                  -- return if we see the beginning of a Comment
                  return ;
               end if;
            when '"' =>
            Find_Double:
               loop
                  I := I + 1;
                  exit Find_Double when I > Str'LAST or else Str (I) = '"';
               end loop Find_Double;
            when ''' =>
               -- check for character literal
               if Str (I + 2) = ''' then
                  I := I + 2; -- Skip it
               end if;
            when others   =>
               null;
         end case;
         I := I + 1;
      end loop;
   exception
      when Constraint_Error    =>
         Show_Error (3001, Fatal, Str);
   end Lower;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Reset_Buffer
   --
   --  <Description>
   --
   --     Called from outside to reset the buffer (usually after an error
   --   has been detected).  This routine throws away the current line of input.
   --
   --==========================================================================
   procedure Reset_Buffer is
   begin
      Line_Lng := 0;
      Abuf (1) := Newline;
      Sptr     := 0;
      --Quick_Block_Io.Clear_Block;
   end Reset_Buffer;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Allow_EOF and Disallow_EOF
   --
   --  <Description>
   --
   -- When we are reading from a file sometimes it is OK to get an
   -- end_of_file, such as when a person uses the environment
   -- command "get_file"
   -- However, when we are using the Interpret command the end_of_file
   -- will be significant. These routines control whether or not
   -- the end_of_file will signal an error, or the continuation of
   -- input from Standard_Ace_Input.
   --
   --==========================================================================
   procedure Allow_Eof is
   begin
      Eof_Ok := True;
   end Allow_Eof;

   procedure Disallow_Eof is
   begin
      Eof_Ok := False;
   end Disallow_Eof;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Get_Another_Line
   --
   --  <Description>
   --
   --     This routine gets a line of input either from Standard_Input or
   --     a file.
   --
   --
   --  <Called By>
   --
   --     Input
   --
   --  <Exceptions Raised>
   --
   --     Implicitly Constraint_Error, and End_Error
   --
   --  <Exceptions_Handled>
   --
   --     End_Error, and Constraint_Error
   --
   --==========================================================================
   procedure Get_Another_Line is

   begin
      if Ace_Text_Io."/=" (Ace_Text_Io.Current_Ace_Input,
                          Ace_Text_Io.Standard_Ace_Input) then
        Ace_Text_Io.Get_Line (Ace_Text_Io.Current_Ace_Input, Abuf, Line_Lng);
      else
        Ace_Command_Line_Input.Command_Input.Get_Line
           (Ooe.File_Entries (Ace_Text_Io.Current_Ace_Input), Abuf, Line_Lng);
      end if;
      if Line_Lng = Abuf'LAST then
         if Echo_Input then
            Ace_Text_Io.Put (Ace_Text_Io.Current_Ace_Output, Abuf);
         end if;
      else
         if Echo_Input then
            Ace_Text_Io.Put_Line (Ace_Text_Io.Current_Ace_Output,
                                  Abuf (1 .. Line_Lng));
         end if;
         Line_Lng        := Line_Lng + 1;
         Abuf (Line_Lng) := Newline;
      end if;
      Lower (Abuf (1 .. Line_Lng));
   exception
      when Constraint_Error    =>
         Show_Error (3010, Warning, "Get_Another_Line");
         Abuf (1) := Newline;
         Line_Lng := 1;
      when End_Error =>
         if Eof_Ok then
            -- When we are reading from a file sometimes it is OK to get an
            -- end_of_file, such as when a person uses the environment
            -- command "get_file"
            Ace_Text_Io.Set_Ace_Input (Ace_Text_Io.Standard_Ace_Input);
            Disallow_Eof;
         else
            -- However, when we are using the Interpret command the end_of_file
            -- will be significant.
            raise End_Error;
         end if;
         Reset_Buffer;
         Abuf (1) := Newline;
         Set_Prompt_On;
   end Get_Another_Line;
   pragma INLINE (Get_Another_Line);

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Input
   --
   --  <Description>
   --
   --    This is the main input routine for ACE.  Input is gotten from
   --    one of three places  1) The "unput" buffer, 2) the input buffer,
   --    and 3) Get_Another_Line.
   --
   --  <Exceptions Raised>
   --
   --    Implicit End_Error (when user types EOF - ^D).
   --    Ace_Universal_Types.End_Ace
   --
   --  <Exceptions Handled>
   --
   --    End_Error
   --
   --==========================================================================
   function Input return Character is
      Tchar : Character;
   begin
      if Sptr > 0 then
         -- Check our Unput buffer
         Tchar := Buf (Sptr);
         Sptr := Sptr - 1;
      elsif Aptr < Line_Lng then
         -- Check our Stdin buffer
         Aptr := Aptr + 1;
         Tchar := Abuf (Aptr);
      else
         -- Need another line from Stdin
         Get_Another_Line;
         if Line_Lng = 0 then
            Aptr  := 0;
            Tchar := Newline;
         else
            Aptr := 1;
            Tchar := Abuf (1);
         end if;
      end if;

      if Tchar = Newline then
         Lineno := Lineno + 1;
      end if;

      return Tchar;

   exception
      when End_Error =>
         raise Ace_Universal_Types.End_Ace;
   end Input;
   pragma INLINE (Input);

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Output
   --
   --  <Description>
   --
   --    Print a character
   --
   --  <Called By>
   --
   --    Lex_Pkg
   --
   --==========================================================================
   procedure Output (C : in Character) is
   begin
      if C = Newline then
         New_Line;
      else
         Put (C);
      end if;
   end Output;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Init_Buf
   --
   --  <Description>
   --
   --    Initializes the "Unput" input buffer
   --
   --  <Called By>
   --
   --    Lex_Pkg
   --==========================================================================
   procedure Init_Buf is
   begin
      Sptr := 0;
   end Init_Buf;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Yy_Wrap
   --
   --  <Description>
   --
   --    Returns true meaning "Wrap Up" when an error is detected by
   --    the lexical analyzer.
   --
   --  <Called By>
   --
   --    Lex_Pkg
   --
   --==========================================================================
   function Yywrap return Boolean is
   begin
      return True;
   end Yywrap;

end User_Pkg;
