--|$Log:	xadt.a,v $
--|Revision 8.0  91/01/17  10:46:34  cm
--|
--|
--|Copyright (c) 1990, Unisys Corporation.
--|
--|Ada Command Environment (ACE).
--|
--|This software, developed under the Software Technology for Adaptable,
--|Reliable Systems (STARS) program, is approved for release under Distribution
--|"A" of the Scientific and Technical Information Program Classification 
--|Scheme (DoD Directive 5230.24) unless otherwise indicated.  Sponsored by the
--|U.S. Defense Advanced Research Projects Agency (DARPA) under contract 
--|F19628-88-D-0031, the STARS program is supported by the military services, 
--|SEI, and MITRE, with the U.S. Air Force as the executive contracting agent.
--|
--|Permission to use, copy, modify, and comment on this software and its
--|documentation for purposes stated under Distribution "A" and without fee
--|is hereby granted.  This software retains Contractor indemnification to
--|the Government regarding copyrights pursuant to the above referenced
--|STARS contract.  The Government disclaims all responsibility against
--|liability, including costs and expenses for violation of proprietary
--|rights, or copyrights arising out of the creation or use of this
--|software.
--|
--|In addition, the Government and Unisys disclaim all warranties with
--|regard to this software, including all implied warranties of
--|merchantability and fitness, and in no event shall the Government or
--|Unisys be liable for any special, indirect or consequential damages or
--|any damages whatsoever resulting from the loss of use, data, or profits,
--|whether in action of contract, negligence or other tortuous action,
--|arising in connection with the use or performance of this software.
--|
--
-- Package Name - X_Window_Adt [Body]
--
---------------------------------------------------------------
--
--
-- Abstract
--
--    This package provides the ACE Abstract Data Type implementation for
--    the X Window System.  The routines within this package interface
--    with the X-Ada bindings to implement a selected portion of window
--    creation, deletion, and text/drawing manipulation operations.
--
-- Update Log - :
--
-- End of Header
----------------------------------------------------------------------------

with X_Windows;
use X_Windows;

with Symbol_Table_Structures;
use Symbol_Table_Structures;

with Symtbl_Entry_Get_Routines;
use Symtbl_Entry_Get_Routines;

with Ace_Universal_Types;
use Ace_Universal_Types;

with Text_Io;
use Text_Io;

package body X_Window_Adt is

   use Events;

   subtype Range_Of_Window_Objects is Ace_Window_Type range 0 .. 10;

   Window_Objects_Free_List : array (Range_Of_Window_Objects) of Boolean :=
      (others => True);

   Ace_Display              : array (Range_Of_Window_Objects) of Display;
   Ace_Window               : array (Range_Of_Window_Objects) of Window;

   use Graphic_Output;

   Ace_Gc : array (Range_Of_Window_Objects) of Graphic_Context;
   Ace_Dash_Gc : array (Range_Of_Window_Objects) of Graphic_Context;
   Ace_Gcv : array (Range_Of_Window_Objects) of Gc_Value_Record;

   Window_Attributes        : Set_Window_Attributes_Record;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Allocate_New_Window_Object
   --
   --  <Description>
   --
   --    This function allocates a window object from the free list and
   --    returns its identity when one exists.
   --
   --  <Called By>
   --
   --    Create_Window
   --
   --  <Exceptions Raised>
   --
   --    Use_Error, when free list is empty
   --
   --=======================================================================
   function Allocate_New_Window_Object return Ace_Window_Type is
   begin
      -- find first free window object (if any)
      for I in Range_Of_Window_Objects'FIRST .. Range_Of_Window_Objects'LAST
       loop
         if Window_Objects_Free_List (I) then
            Window_Objects_Free_List (I) := False;
            return (I);
         end if;
      end loop;
      -- past range of available window objects
      -- show_error ??
      raise Use_Error;
   end Allocate_New_Window_Object;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Free_Destroyed_Window_Object
   --
   --  <Description>
   --
   --    This procedure returns a destroyed window identifier back onto the
   --    free list of window identifiers.
   --
   --  <Called By>
   --
   --    Destroy_Window
   --
   --==========================================================================
   procedure Free_Destroyed_Window_Object (Window : in out Ace_Window_Type) is
   begin
      Window_Objects_Free_List (Window) := True;
      Window := Range_Of_Window_Objects'FIRST;
   end Free_Destroyed_Window_Object;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Create_Window
   --
   --  <Description>
   --
   --    This procedure creates a new window object under the X Window System.
   --
   --==========================================================================
   procedure Create_Window (Window              : in out Ace_Window_Type;
                            Horizontal_Position : in Ace_Coordinate := 0;
                            Vertical_Position   : in Ace_Coordinate := 0;
                            Horizontal_Size     : in Ace_Pixels     := 100;
                            Vertical_Size       : in Ace_Pixels     := 100) is
   begin
      Window := Allocate_New_Window_Object;

      Ace_Display (Window) := Open_Display ("");

      Ace_Window (Window) :=
         Create_Simple_Window
             (Ace_Display (Window), Default_Root_Window (Ace_Display (Window)),
              (Coordinate (Horizontal_Position),
               Coordinate (Vertical_Position),
               Pixels (Horizontal_Size),
               Pixels (Vertical_Size)), 3,
              Black_Pixel (Ace_Display (Window),
                           Default_Screen (Ace_Display (Window))),
              White_Pixel (Ace_Display (Window),
                           Default_Screen (Ace_Display (Window))));

      Window_Attributes.Backing_Store    := Always;

      Change_Window_Attributes (Ace_Display (Window), Ace_Window (Window),
                                Wa_Backing_Store, Window_Attributes);

      Map_Window (Ace_Display (Window), Ace_Window (Window));

      Flush (Ace_Display (Window));

      -----
      -- do normal GC
      --
      Ace_Gcv (Window).Logical_Operation := Gx_Set;

      Ace_Gc (Window) := Create_Gc (Ace_Display (Window), Ace_Window (Window),
                                    Gc_Function, Ace_Gcv (Window));
      -----

      -----
      -- do dashed line GC
      --
      Ace_Gcv (Window).Line_Style        := Line_On_Off_Dash;

      Ace_Dash_Gc (Window) := Create_Gc (Ace_Display (Window),
                                         Ace_Window (Window),
                                         (Gc_Function or Gc_Line_Style),
                                         Ace_Gcv (Window));

      Set_Background (Ace_Display (Window), Ace_Dash_Gc (Window),
                      White_Pixel (Ace_Display (Window),
                                   Default_Screen (Ace_Display (Window))));
      Set_Foreground (Ace_Display (Window), Ace_Dash_Gc (Window),
                      Black_Pixel (Ace_Display (Window),
                                   Default_Screen (Ace_Display (Window))));
      -----

   end Create_Window;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Destroy_Window
   --
   --  <Description>
   --
   --    This procedure destroys an X Window that was previously created.
   --
   --==========================================================================
   procedure Destroy_Window (Window : in out Ace_Window_Type) is
   begin
      Unmap_Window (Ace_Display (Window), Ace_Window (Window));
      Destroy_Window (Ace_Display (Window), Ace_Window (Window));
      Flush (Ace_Display (Window));
      Free_Destroyed_Window_Object (Window);
   end Destroy_Window;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Clear_Window
   --
   --  <Description>
   --
   --    This procedure clears the all the elements from within the window.
   --
   --==========================================================================
   procedure Clear_Window (Window : in Ace_Window_Type) is
   begin
      Clear_Window (Ace_Display (Window), Ace_Window (Window));
      Flush (Ace_Display (Window));
   end Clear_Window;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Draw_Line
   --
   --  <Description>
   --
   --    This procedure draws a line within an X window as specified by the
   --    X and Y coordinate values.
   --
   --==========================================================================
   procedure Draw_Line (Window         : in Ace_Window_Type;
                        X1, Y1, X2, Y2 : in Ace_Coordinate) is
   begin
      Draw_Line (Ace_Display (Window), Ace_Window (Window), Ace_Gc (Window),
                 (Coordinate (X1),
                  Coordinate (Y1)), (Coordinate (X2),
                                     Coordinate (Y2)));
      Flush (Ace_Display (Window));
   end Draw_Line;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Draw_Dashed_Line
   --
   --  <Description>
   --
   --    This procedure draws a dashed line within an X window as specified
   --    by the X and Y coordinate values.
   --
   --==========================================================================
   procedure Draw_Dashed_Line (Window         : in Ace_Window_Type;
                               X1, Y1, X2, Y2 : in Ace_Coordinate) is
   begin
      Draw_Line (Ace_Display (Window), Ace_Window (Window),
                 Ace_Dash_Gc (Window), (Coordinate (X1),
                                        Coordinate (Y1)), (Coordinate (X2),
                                                           Coordinate (Y2)));
      Flush (Ace_Display (Window));
   end Draw_Dashed_Line;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Draw_Text
   --
   --  <Description>
   --
   --    This procedure draws text within an X window at the given coordinates.
   --
   --==========================================================================
   procedure Draw_Text (Window : in Ace_Window_Type;
                        X, Y   : in Ace_Coordinate;
                        Text   : in String) is
   begin
      Draw_String (Ace_Display (Window), Ace_Window (Window), Ace_Gc (Window),
                   (Coordinate (X),
                    Coordinate (Y)), Text);
      Flush (Ace_Display (Window));
   end Draw_Text;

   --==========================================================================
   --  <Subprogram Name>
   --
   --    Draw_Rectangle_Builtin
   --
   --  <Description>
   --
   --    This procedure draws a rectangle within an X window by making use of
   --    the draw rectangle function provided by X.  Rectangle is drawn using
   --    the X, Y coordinates and the width and height of the rectangle in
   --    pixels.
   --
   --==========================================================================
   procedure Draw_Rectangle_Builtin (Window        : in Ace_Window_Type;
                                     X, Y          : in Ace_Coordinate;
                                     Width, Height : in Ace_Pixels) is
   begin
      Draw_Rectangle (Ace_Display (Window), Ace_Window (Window),
                      Ace_Gc (Window), (Coordinate (X),
                                        Coordinate (Y),
                                        Coordinate (Width),
                                        Coordinate (Height)));
      Flush (Ace_Display (Window));
   end Draw_Rectangle_Builtin;

begin
   Window_Objects_Free_List (Range_Of_Window_Objects'FIRST) := False;
end X_Window_Adt;
